#include "xsw.h"
#include "net.h"


void AutoIntervalTune(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Handles automatic adjusting of the network interval.
 *
 *	Calls to this function must be timmed correctly.
 */
void AutoIntervalTune(void)
{
	long d;
	unsigned long total_bytes;
	xsw_net_parms_struct *net = &net_parms;
	xsw_net_stat_struct *nstat = &net_stat;
	xsw_option_struct *opt = &option;

	/* Must be connected or logging in */
	if(net->connection_state == XSW_CON_STATE_NOT_CONNECTED)
	{
	    /* Not connected or not logged in, so reset counters */
	    nstat->sx_ilast = nstat->sx_interval;
	    nstat->rx_ilast = nstat->rx_interval;
	    nstat->sx_interval = 0l;
	    nstat->rx_interval = 0l;
	    return;
	}

	/* Is automatic interval tunning on? */
	if(!auto_interval_tune.state)
	{
	    /* Automatic interval tunning is not on, so reset counters */
	    nstat->sx_ilast = nstat->sx_interval;
	    nstat->rx_ilast = nstat->rx_interval;
	    nstat->sx_interval = 0l;
	    nstat->rx_interval = 0l;
	    return;
	}


	/* Get total bytes recieved and sent */
	total_bytes = nstat->rx_interval + nstat->sx_interval;

	/* Convert total_bytes from per 1000 ms interval to per
	 * auto_interval_tune.interval ms interval
	 *
	 * This should not be needed since auto_interval_tune.interval
	 * should be 1000 ms but we check this just incase
	 */
	if(auto_interval_tune.interval > 0l)
	    total_bytes = total_bytes * 1000l / auto_interval_tune.interval;


	/* Calculate the delta of bandwidth and actual total bytes
	 * sent and received
	 */
	d = (long)total_bytes - (long)opt->net_bandwidth;
	if(d < 0l)
	{
	    /* Under bandwidth, so decrease interval */
	    net->net_int -= 50l;
	}
	else
	{
	    /* Over bandwidth, so increase interval */
	    net->net_int += 50l;
	}


	/* Sanitize interval */
	if(net->net_int > XSW_SWSERV_MAX_UPDATE_INT) 
	    net->net_int = XSW_SWSERV_MAX_UPDATE_INT;
	if(net->net_int < XSW_SWSERV_MIN_UPDATE_INT)  
	    net->net_int = XSW_SWSERV_MIN_UPDATE_INT;

	/* If in energy saver mode then the interval must be
	 * XSW_SWSERV_DEF_UPDATE_INT or greater
	 */
	if(option.energy_saver_mode)
	{
	    if(net->net_int < XSW_SWSERV_DEF_UPDATE_INT)
		net->net_int = XSW_SWSERV_DEF_UPDATE_INT;
	}


	/* Send interval update to server */
	NetSendSetInterval();
	if(!XSW_CON_IS_CONNECTED(net))
	    return;


	/* Reset sx and rx byte counters */
	nstat->sx_ilast = nstat->sx_interval;
	nstat->rx_ilast = nstat->rx_interval;

	nstat->sx_interval = 0l;
	nstat->rx_interval = 0l;
}
