#include "blitting.h"


static void BlitBufClearPixel8(void *p, const u_int32_t v);
static void BlitBufClearPixel15(void *p, const u_int32_t v);
static void BlitBufClearPixel16(void *p, const u_int32_t v);
static void BlitBufClearPixel24(void *p, const u_int32_t v);
static void BlitBufClearPixel32(void *p, const u_int32_t v);

static void BlitBufClearArea(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	const u_int32_t v,
	void (*pixel_op_func)(void *, const u_int32_t)
);

void BlitBufClear(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	const WColorStruct *c
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void BlitBufClearPixel8(void *p, const u_int32_t v)
{
	*(u_int8_t *)p = (u_int8_t)v;
}

static void BlitBufClearPixel15(void *p, const u_int32_t v)
{
	*(u_int16_t *)p = (u_int16_t)v;
}

static void BlitBufClearPixel16(void *p, const u_int32_t v)
{
	*(u_int16_t *)p = (u_int16_t)v;
}

static void BlitBufClearPixel24(void *p, const u_int32_t v)
{
	u_int8_t *p8 = (u_int8_t *)p;
	p8[0] = (u_int8_t)((v & 0x000000FF) >> 0);
	p8[1] = (u_int8_t)((v & 0x0000FF00) >> 8);
	p8[2] = (u_int8_t)((v & 0x00FF0000) >> 16);
}

static void BlitBufClearPixel32(void *p, const u_int32_t v)
{
	*(u_int32_t *)p = v;
}


static void BlitBufClearArea(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	const u_int32_t v,
	void (*pixel_op_func)(void *, const u_int32_t)
)
{
	int y;
	u_int8_t *ptr, *end;

	if((tar_buf == NULL) || (tar_width == 0) || (tar_height == 0))
	    return;

	for(y = 0; y < (int)tar_height; y++)
	{
	    ptr = tar_buf + (y * (int)tar_bpl);
	    end = ptr + (tar_width * bpp);
	    while(ptr < end)
	    {
		pixel_op_func(ptr, v);
		ptr += bpp;
	    }
	}
}


void BlitBufClear(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	const WColorStruct *c
)
{
	u_int32_t v;
	WColorStruct def_c;
	void (*pixel_op_func)(void *, u_int32_t v);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

	if(c == NULL)
	{
	    def_c.a = 0x00;
	    def_c.r = 0x00;
	    def_c.g = 0x00;
	    def_c.b = 0x00;
	    c = &def_c;
	}

	switch(depth)
	{
	  case 8:
	    v = PACK8TO8(c->r, c->g, c->b);
	    break;
	  case 15:
	    v = PACK8TO15(c->r, c->g, c->b);
	    break;
	  case 16:
	    v = PACK8TO16(c->r, c->g, c->b);
	    break;
	  case 24:
	  case 32:
	    v = PACK8TO32(c->a, c->r, c->g, c->b);
	    break;
	  default:
	    v = 0x00000000;
	    break;
	}

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufClearPixel8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufClearPixel15;
		break;
	      case 16:
		pixel_op_func = BlitBufClearPixel16;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    pixel_op_func = BlitBufClearPixel24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:
		pixel_op_func = BlitBufClearPixel24;
		break;
	      case 32:
		pixel_op_func = BlitBufClearPixel32;
		break;
	      default:
		return; 
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufClearArea(
	    bpp, depth,
	    tar_buf,
	    tar_width, tar_height, tar_bpl,
	    v, pixel_op_func
	);
}
