#include "blitting.h"


static void BlitBufFadePixel8(void *p, const float gamma);
static void BlitBufFadePixel15(void *p, const float gamma);
static void BlitBufFadePixel16(void *p, const float gamma);
static void BlitBufFadePixel24(void *p, const float gamma);
static void BlitBufFadePixel32(void *p, const float gamma);

static void BlitBufGlowPixelAdditive8(void *p, const WColorStruct *c);
static void BlitBufGlowPixelAdditive15(void *p, const WColorStruct *c);
static void BlitBufGlowPixelAdditive16(void *p, const WColorStruct *c);
static void BlitBufGlowPixelAdditive24(void *p, const WColorStruct *c);
static void BlitBufGlowPixelAdditive32(void *p, const WColorStruct *c);


static void BlitBufFadeArea(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int tar_apply_width, unsigned int tar_apply_height,
	float gamma,
	void (*pixel_op_func)(void *, const float)
);
void BlitBufFade(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y, 
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int tar_apply_width, unsigned int tar_apply_height,
	float gamma			/* 0.1 to 1.0 */
);

static void BlitBufGlowArea(
        unsigned int bpp,
        u_int8_t *tar_buf,
        int tar_x, int tar_y,
        unsigned int tar_width, unsigned int tar_height,
        unsigned int tar_bpl,
        unsigned int tar_apply_width, unsigned int tar_apply_height,
        const WColorStruct *c,
        void (*pixel_op_func)(void *, const WColorStruct *)
);
void BlitBufGlow(
        unsigned int bpp, unsigned int depth,
        u_int8_t *tar_buf,
        int tar_x, int tar_y,
        unsigned int tar_width, unsigned int tar_height,
        unsigned int tar_bpl,
        unsigned int tar_apply_width, unsigned int tar_apply_height,
        const WColorStruct *c
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void BlitBufFadePixel8(void *p, const float gamma)
{
	const u_int8_t tv = *(u_int8_t *)p;

	*(u_int8_t *)p = PACK8TO8(
	    (u_int8_t)(((tv & 0xE0)) * gamma), 
	    (u_int8_t)(((tv & 0x1C) << 2) * gamma),
	    (u_int8_t)(((tv & 0x03) << 5) * gamma)
	);
}

static void BlitBufFadePixel15(void *p, const float gamma)
{
	const u_int16_t tv = *(u_int16_t *)p;

	*(u_int16_t *)p = PACK8TO15(
	    (u_int8_t)(((tv & 0x7C00) >> 7) * gamma),
	    (u_int8_t)(((tv & 0x03E0) >> 2) * gamma),
	    (u_int8_t)(((tv & 0x001F) << 3) * gamma)
	);
}

static void BlitBufFadePixel16(void *p, const float gamma)
{
	const u_int16_t tv = *(u_int16_t *)p;

	*(u_int16_t *)p = PACK8TO16(
	    (u_int8_t)(((tv & 0xF800) >> 8) * gamma),
	    (u_int8_t)(((tv & 0x07E0) >> 3) * gamma),
	    (u_int8_t)(((tv & 0x001F) << 3) * gamma)
	);
}

static void BlitBufFadePixel24(void *p, const float gamma)
{
	u_int8_t *p8 = (u_int8_t *)p;
	p8[0] = (u_int8_t)(p8[0] * gamma);
	p8[1] = (u_int8_t)(p8[1] * gamma);
	p8[2] = (u_int8_t)(p8[2] * gamma);
}

static void BlitBufFadePixel32(void *p, const float gamma)
{
	u_int8_t *p8 = (u_int8_t *)p;
	p8[0] = (u_int8_t)(p8[0] * gamma);
	p8[1] = (u_int8_t)(p8[1] * gamma);
	p8[2] = (u_int8_t)(p8[2] * gamma);
}


static void BlitBufGlowPixelAdditive8(void *p, const WColorStruct *c)
{
        const u_int16_t tv = *(u_int8_t *)p;

        *(u_int8_t *)p = PACK8TO8(
            (u_int8_t)MIN(
(u_int16_t)(tv & 0xE0) + (u_int16_t)c->r,
                0x00FF
            ),
            (u_int8_t)MIN(
(u_int16_t)((tv & 0x1C) << 3) + (u_int16_t)c->g,
                0x00FF
            ),
            (u_int8_t)MIN(   
(u_int16_t)((tv & 0x03) << 6) + (u_int16_t)c->b,
                0x00FF
            )
        );
}

static void BlitBufGlowPixelAdditive15(void *p, const WColorStruct *c)
{
        const u_int16_t tv = *(u_int16_t *)p;

        *(u_int16_t *)p = PACK8TO15(
            (u_int8_t)MIN(
(u_int16_t)((tv & 0x7C00) >> 7) + (u_int16_t)c->r,
                0xFF
            ),      
            (u_int8_t)MIN(
(u_int16_t)((tv & 0x03E0) >> 2) + (u_int16_t)c->g,
                0xFF
            ),      
            (u_int8_t)MIN(   
(u_int16_t)((tv & 0x001F) << 3) + (u_int16_t)c->b,
                0xFF
            )
        );
}

static void BlitBufGlowPixelAdditive16(void *p, const WColorStruct *c)
{
        const u_int16_t tv = *(u_int16_t *)p;

        *(u_int16_t *)p = PACK8TO16(
            (u_int8_t)MIN(
(u_int16_t)((tv & 0xF800) >> 8) + (u_int16_t)c->r,
                0xFF
            ),
            (u_int8_t)MIN(
(u_int16_t)((tv & 0x07E0) >> 3) + (u_int16_t)c->g,
                0xFF
            ),
            (u_int8_t)MIN(
(u_int16_t)((tv & 0x001F) << 3) + (u_int16_t)c->b,
                0xFF
            )
        );   
}
 
static void BlitBufGlowPixelAdditive24(void *p, const WColorStruct *c)
{
        u_int8_t *p8 = (u_int8_t *)p;

        p8[2] = (u_int8_t)MIN(
            (u_int16_t)p8[2] + (u_int16_t)c->r,      
            0x00FF
        );
        p8[1] = (u_int8_t)MIN(
            (u_int16_t)p8[1] + (u_int16_t)c->g,      
            0x00FF
        );
        p8[0] = (u_int8_t)MIN(
            (u_int16_t)p8[0] + (u_int16_t)c->b,
            0x00FF
        );        
}

static void BlitBufGlowPixelAdditive32(void *p, const WColorStruct *c)
{
        u_int8_t *p8 = (u_int8_t *)p;

        p8[2] = (u_int8_t)MIN(
            (u_int16_t)p8[2] + (u_int16_t)c->r,
            0x00FF
        );
        p8[1] = (u_int8_t)MIN(
            (u_int16_t)p8[1] + (u_int16_t)c->g,
            0x00FF
        );
        p8[0] = (u_int8_t)MIN(
            (u_int16_t)p8[0] + (u_int16_t)c->b,
            0x00FF
        );
}


static void BlitBufFadeArea(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int tar_apply_width, unsigned int tar_apply_height,
	float gamma,
	void (*pixel_op_func)(void *, const float)
)
{
	int y;
	u_int8_t *tar_ptr, *tar_edge, *tar_end;

	if((tar_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (tar_apply_width == 0) || (tar_apply_height == 0)
	)
	    return;

	/* Is the area to apply entirely off the target buffer? */
	if((tar_x >= (int)tar_width) ||
	   (tar_y >= (int)tar_height) ||
	   ((tar_x + (int)tar_apply_width) <= 0) ||
	   ((tar_y + (int)tar_apply_height) <= 0)
	)
	    return;

	/* Sanitize starting coordinates */
	if(tar_x < 0)
	{
	    tar_apply_width = MAX((int)tar_apply_width + tar_x, 0);
	    tar_x = 0;
	}
	if(tar_y < 0)
	{
	    tar_apply_height = MAX((int)tar_apply_height + tar_y, 0);
	    tar_y = 0;
	}

	if((tar_x + (int)tar_apply_width) > (int)tar_width)
	    tar_apply_width = (int)tar_width - tar_x;

	if((tar_y + (int)tar_apply_height) > (int)tar_height)
	    tar_apply_height = (int)tar_height - tar_y;


	/* Get the starting buffer pointer */
	tar_ptr = tar_buf +
	    (tar_y * (int)tar_bpl) +
	    (tar_x * (int)bpp);

	/* Get the edge buffer pointer */
	tar_edge = tar_ptr + (tar_apply_width * (int)bpp);

	/* Get the end buffer pointer */
	tar_end = tar_buf +
	    ((tar_y + (int)tar_apply_height - 1) * (int)tar_bpl) +
	    ((tar_x + (int)tar_apply_width) * (int)bpp);


	/* Begin blitting */
	y = tar_y;
	while(tar_ptr < tar_end)
	{
	    if(tar_ptr >= tar_edge)
	    {
		y++;
		tar_ptr = tar_buf +
		    (y * (int)tar_bpl) +
		    (tar_x * (int)bpp);
		tar_edge = tar_ptr + (tar_apply_width * (int)bpp);
		continue;
	    }

	    pixel_op_func(tar_ptr, gamma);

	    tar_ptr += bpp;
	}
}

void BlitBufFade(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int tar_apply_width, unsigned int tar_apply_height,
	float gamma			/* 0.1 to 1.0 */
)
{
	void (*pixel_op_func)(void *, const float);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

	if(gamma < 0.0f)
	    gamma = 0.0f;
	else if(gamma > 1.0f)
	    gamma = 1.0f;

	/* Any gamma value of 1.0 or above will produce no affect */
	if(gamma >= 1.0f)
	    return;

	switch(bpp)
	{
	  case 1:     
	    pixel_op_func = BlitBufFadePixel8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufFadePixel15;
		break;
	      case 16:
		pixel_op_func = BlitBufFadePixel16;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    pixel_op_func = BlitBufFadePixel24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:  
		pixel_op_func = BlitBufFadePixel24;
		break;
	      case 32:
		pixel_op_func = BlitBufFadePixel32;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufFadeArea(
	    bpp,
	    tar_buf, tar_x, tar_y, tar_width, tar_height, tar_bpl,
	    tar_apply_width, tar_apply_height,
	    gamma, pixel_op_func
	);
}


static void BlitBufGlowArea(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int tar_apply_width, unsigned int tar_apply_height,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const WColorStruct *)
)
{
	int y;
	u_int8_t *tar_ptr, *tar_edge, *tar_end;

	if((tar_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (tar_apply_width == 0) || (tar_apply_height == 0)
	)
	    return;

	/* Is the area to apply entirely off the target buffer? */
	if((tar_x >= (int)tar_width) ||
	   (tar_y >= (int)tar_height) ||
	   ((tar_x + (int)tar_apply_width) <= 0) ||
	   ((tar_y + (int)tar_apply_height) <= 0)
	)
	    return;

	/* Sanitize starting coordinates */
	if(tar_x < 0)
	{
	    tar_apply_width = MAX((int)tar_apply_width + tar_x, 0);
	    tar_x = 0;
	}
	if(tar_y < 0)
	{
	    tar_apply_height = MAX((int)tar_apply_height + tar_y, 0);
	    tar_y = 0;
	}

	if((tar_x + (int)tar_apply_width) > (int)tar_width)
	    tar_apply_width = (int)tar_width - tar_x;

	if((tar_y + (int)tar_apply_height) > (int)tar_height)
	    tar_apply_height = (int)tar_height - tar_y;


	/* Get the starting buffer pointer */
	tar_ptr = tar_buf +
	    (tar_y * (int)tar_bpl) +
	    (tar_x * (int)bpp);

	/* Get the edge buffer pointer */
	tar_edge = tar_ptr + (tar_apply_width * (int)bpp);

	/* Get the end buffer pointer */
	tar_end = tar_buf +
	    ((tar_y + (int)tar_apply_height - 1) * (int)tar_bpl) +
	    ((tar_x + (int)tar_apply_width) * (int)bpp);


	/* Begin blitting */
	y = tar_y;
	while(tar_ptr < tar_end)
	{
	    if(tar_ptr >= tar_edge)
	    {
		y++;
		tar_ptr = tar_buf +
		    (y * (int)tar_bpl) +
		    (tar_x * (int)bpp);
		tar_edge = tar_ptr + (tar_apply_width * (int)bpp);
		continue;
	    }

	    pixel_op_func(tar_ptr, c);

	    tar_ptr += bpp;
	}
}

void BlitBufGlow(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int tar_apply_width, unsigned int tar_apply_height,
	const WColorStruct *c
)
{
	void (*pixel_op_func)(void *, const WColorStruct *);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

        if(c == NULL)
	    return;

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufGlowPixelAdditive8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufGlowPixelAdditive15;
		break;
	      case 16:
		pixel_op_func = BlitBufGlowPixelAdditive16;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    pixel_op_func = BlitBufGlowPixelAdditive24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:  
		pixel_op_func = BlitBufGlowPixelAdditive24;
		break;
	      case 32:
		pixel_op_func = BlitBufGlowPixelAdditive32;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufGlowArea(
	    bpp,
	    tar_buf, tar_x, tar_y, tar_width, tar_height, tar_bpl,
	    tar_apply_width, tar_apply_height,
	    c, pixel_op_func
	);
}
