#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "blitting.h"
#include "xswimgrefs.h"
#include "gctl.h"
#include "net.h"
#include "bridgewin.h"
#include "xsw.h"


static void BridgeDrawBlitImageAbsolute(
	shared_image_t *tar_img, image_t *src_img,
	const int x, const int y
);
static void BridgeDrawBlitImageAbsoluteCode(
	shared_image_t *tar_img, const int src_img_code,
	const int x, const int y
);
static void BridgeDrawBlitImageCursor(
	shared_image_t *tar_img, image_t *src_img,
	const int x, const int y,
	const WColorStruct *c
);
static void BridgeDrawBlitImageCursorCode(
	shared_image_t *tar_img, const int src_img_code,
	const int x, const int y,
	const WColorStruct *c
);
static void BridgeDrawBlitImgRef(
	shared_image_t *tar_img,
	imgref_struct *imgref,
	const int x, const int y,
	const int frame_num,
	const float zoom,
	const float vis,
	const bool_t draw_vector_lights,
	const bool_t draw_strobe_lights
);
static void BWDP_BACKGROUND(
	int obj_num, xsw_object_struct *obj,
	pixmap_t pixmap,
	const int panel_num		/* One of BW_PANEL_* */
);
static void BWDP_PNAME(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PSHIELDFREQ(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PWEAPONFREQ(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PCOMCHANNEL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PSCORE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PHULL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PPOWER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PSHIELDS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PVIS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PDMGCTL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window 
);

static void BWDP_SNAME(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SHULL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w, 
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SPOWER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SSHIELDS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SVIS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SBEARING(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SDISTANCE_BAR(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SDISTANCE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
#if 0
static void BWDP_SDMGCTL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
#endif

static void BWDP_PINAME(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PWLOCK(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PWEAPONS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);

static void BWDP_SEMPIRE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window  
);
static void BWDP_SWEAPONS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);

static void BWDP_PTHROTTLEL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PTHROTTLER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PTHROTTLE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PVELOCITY(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PANTIMATTER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PENGINESTATE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PHEADING(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_PTHRUSTVECTOR(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);

static void BWDP_SANTIMATTER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);
static void BWDP_SHEADING(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
);

void BridgeWinDrawPanel(
	const int obj_num,
	const int panel_num		/* One of BW_PANEL_* */
);
void BridgeWinDrawAll(xsw_bridge_win_struct *bw);

void BridgeDrawMessages(xsw_bridge_win_struct *bw);
void BridgeQueueDrawMessages(xsw_bridge_win_struct *bw);


#define RADIANS_TO_DEGREES(r) ((r) * 180 / PI)


/* Bridge console panels, sizes must match those in
 * bridgeevent.cpp
 */
#define BW_CONSOLE_PANEL_WIDTH		150
#define BW_CONSOLE_PANEL_HEIGHT		130


#define FTOI_RNDOFF(x)	((int)((x) + 0.5f))

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/* Maximum chars per console label */
#define BW_LABEL_MAX			80

/* Bar "curved end" margin (in pixels) */
#define BW_BAR_MARGIN			4

/* Bar color coefficient values */
#define BW_ANTIMATTER_WARNING_COEFF	0.40
#define BW_ANTIMATTER_DANGER_COEFF	0.20


/*
 *	Draws the source image to the target image using the absolute
 *	blitting method.
 */
static void BridgeDrawBlitImageAbsolute(
	shared_image_t *tar_img, image_t *src_img,
	const int x, const int y
)
{
	if((tar_img == NULL) || (src_img == NULL))
	    return;

	BlitBufAbsolute(
	    tar_img->bytes_per_pixel, tar_img->depth,
	    (u_int8_t *)tar_img->data,
	    (u_int8_t *)src_img->data,
	    x, y,
	    tar_img->width, tar_img->height,
	    tar_img->bytes_per_line,
	    0, 0,
	    src_img->width, src_img->height,
	    src_img->bytes_per_line,
	    src_img->width, src_img->height,
	    1,		/* Zoom */
	    1		/* Magnification */
	);
}

static void BridgeDrawBlitImageAbsoluteCode(
	shared_image_t *tar_img, const int src_img_code,
	const int x, const int y
)
{
	if(!ImgIsLoaded(src_img_code))
	    return;

	BridgeDrawBlitImageAbsolute(
	    tar_img, xsw_image[src_img_code]->image,
	    x, y
	);
}

/*
 *	Draws the source image to the target image using the cursor
 *	blitting method.
 */
static void BridgeDrawBlitImageCursor(
	shared_image_t *tar_img, image_t *src_img,
	const int x, const int y,
	const WColorStruct *c
)
{
	if((tar_img == NULL) || (src_img == NULL))
	    return;

	BlitBufCursor(
	    tar_img->bytes_per_pixel, tar_img->depth,
	    (u_int8_t *)tar_img->data,
	    (u_int8_t *)src_img->data,
	    x, y,
	    tar_img->width, tar_img->height, tar_img->bytes_per_line,
	    src_img->width, src_img->height, src_img->bytes_per_line,
	    c
	);
}

static void BridgeDrawBlitImageCursorCode(
	shared_image_t *tar_img, const int src_img_code,
	const int x, const int y,
	const WColorStruct *c
)
{
	if(!ImgIsLoaded(src_img_code))
	    return;

	BridgeDrawBlitImageCursor(
	    tar_img, xsw_image[src_img_code]->image,
	    x, y,
	    c
	);
}

/*
 *	Draws the imgref to the specified target image.
 */
static void BridgeDrawBlitImgRef(
	shared_image_t *tar_img,
	imgref_struct *imgref,
	const int x, const int y,
	const int frame_num,
	const float zoom,
	const float vis,
	const bool_t draw_vector_lights,
	const bool_t draw_strobe_lights
)
{
	if((tar_img == NULL) || (imgref == NULL))
	    return;

	if(imgref->merge_mode == IMGREF_MERGE_ADDITIVE)
	    BlitBufAdditive(
		tar_img->bytes_per_pixel, tar_img->depth,
		(u_int8_t *)tar_img->data,
		(u_int8_t *)imgref->image_data,
		x, y,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,
		0, frame_num * imgref->fheight,
		imgref->width, imgref->height, 0,
		imgref->fwidth, imgref->fheight,
		zoom, 1.0f
	    );
	else
	    BlitBufNormal(
		tar_img->bytes_per_pixel, tar_img->depth,
		(u_int8_t *)tar_img->data,
		(u_int8_t *)imgref->image_data,
		x, y,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,
		0, frame_num * imgref->fheight,
		imgref->width, imgref->height, 0,
		imgref->fwidth, imgref->fheight,
		zoom, vis, 1.0f
	    );

	/* Draw point lights */
	if(imgref->nlights > 0)
	{
	    const int	n = imgref->nlights,
			x2 = x + (int)(imgref->fwidth * zoom / 2),
			y2 = y + (int)(imgref->fheight * zoom / 2);
	    const float theta = frame_num * 0.39269908;
	    int i, x3, y3;
	    float theta2;
	    WColorStruct c;
	    imgref_light_struct *light;

	    for(i = 0; i < n; i++)
	    {
		light = imgref->light[i];
		if(light == NULL)
		    continue;

		/* Skip vector lights? */
		if(!draw_vector_lights)
		{
		    if((light->strobe_off_int <= 0l) &&
		       (light->strobe_on_int <= 0l)
		    )
			continue;
		}

		/* Skip strobes? */
		if(!draw_strobe_lights)
		{
		    if((light->strobe_off_int > 0l) ||
		       (light->strobe_on_int > 0l)
		    )
			continue;
		}

		/* Calculate the color of light or strobe */
		c.a = light->a;
		c.r = light->r;
		c.g = light->g;
		c.b = light->b;

		theta2 = SANITIZERADIANS(light->theta + theta);
		x3 = x2 + FTOI_RNDOFF(
		    (float)sin((double)theta2) * light->radius *
		    zoom
		);
		y3 = y2 - FTOI_RNDOFF(
		    (float)cos((double)theta2) * light->radius *
		    zoom
		);

		BlitBufPointLight(
		    tar_img->bytes_per_pixel, tar_img->depth,
		    tar_img->data,
		    x3, y3,
		    tar_img->width, tar_img->height,
		    tar_img->bytes_per_line,
		    1,			/* Radius */
		    &c			/* Color */
		);
	    }
	}
}

/*
 *	Draws the background for the panel specified by panel_num on
 *	to the pixmap specified by pixmap.
 *
 *	The obj_num and obj specifies the object who's stats are to
 *	used when drawing the panel.
 *
 *	The pixmap specifies the pixmap to draw on.
 *
 *	The panel_num specifies which panel to draw, it must be one
 *	of BW_PANEL_*.
 */
static void BWDP_BACKGROUND(
	const int obj_num, xsw_object_struct *obj,
	pixmap_t pixmap,
        const int panel_num		/* One of BW_PANEL_* */
)
{
	bool_t cannot_draw_bkg = False;
	int i, bg_img_code, imgref_num;
	image_t *src_img = NULL,
		*tar_img = NULL;
	shared_image_t *shared_tar_img = NULL;
	imgref_struct *imgref;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	if(bw == NULL)
	    return;

	/* Get this object's imgref */
	if(obj != NULL)
	{
	    imgref_num = obj->imgref;
	    imgref = ImgRefGetPtr(xsw_imgref, total_imgrefs, imgref_num);
	    if(!ImgRefIsLoaded(imgref))
		imgref = NULL;
	}
	else
	{
	    imgref = NULL;
	}


	switch(panel_num)
	{
	  /* Player stats panel section 1 */
	  case BW_PANEL_P1:
	    /* Draw the panel's background image */
	    bg_img_code = IMG_CODE_STATS_CON1;
	    tar_img = ImgIsLoaded(bg_img_code) ?
		xsw_image[bg_img_code]->image : NULL;
	    OSWPutImageToDrawable(tar_img, pixmap);
	    break;

	  /* Player stats panel section 2 (uses a medium image) */
	  case BW_PANEL_P2:
	    /* Get this panel's image */
	    shared_tar_img = bw->pan_p2_img;

	    /* Draw the panel's background image */
	    BridgeDrawBlitImageAbsoluteCode(
		shared_tar_img, IMG_CODE_STATS_CON2,
		0, 0
	    );

	    /* Draw the vessel image */
	    if(imgref != NULL)
	    {
		int x, y, frame_num = 12;
		float zoom = 1.0f;
		unsigned int zoomed_width, zoomed_height;

		if(frame_num >= imgref->total_frames)
		    frame_num = 0;
		if(frame_num < 0)
		    frame_num = 0;

		if(imgref->fwidth > BW_CONSOLE_PANEL_WIDTH)
		{
		    zoom = (float)BW_CONSOLE_PANEL_WIDTH /
			(float)imgref->fwidth;
		    if((unsigned int)(imgref->fheight * zoom) >
			BW_CONSOLE_PANEL_HEIGHT
		    )
			zoom = (float)BW_CONSOLE_PANEL_HEIGHT /
			    (float)imgref->fheight;
		}
		else if(imgref->fheight > BW_CONSOLE_PANEL_HEIGHT)
		{
		    zoom = (float)BW_CONSOLE_PANEL_HEIGHT /
			(float)imgref->fheight;
		    if((unsigned int)(imgref->fwidth * zoom) >
			BW_CONSOLE_PANEL_WIDTH
		    )
			zoom = (float)BW_CONSOLE_PANEL_WIDTH /
			    (float)imgref->fwidth;
		}
		zoomed_width = (unsigned int)(imgref->fwidth * zoom);
		zoomed_height = (unsigned int)(imgref->fheight * zoom); 

		x = (BW_CONSOLE_PANEL_WIDTH - (int)zoomed_width) / 2;
		y = (BW_CONSOLE_PANEL_HEIGHT - (int)zoomed_height) / 2;

		BridgeDrawBlitImgRef(
		    shared_tar_img, imgref,
		    x, y, frame_num, zoom, 1.0f,
		    obj->lighting & SW_OBJ_LIGHT_VECTOR,
		    False
		);
	    }

	    /* Draw this panel's readout outlines if the object is
	     * known
	     */
	    if((shared_tar_img != NULL) && (obj != NULL))
	    {
		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_HULL,
		    6, 6,
		    &xsw_color.bpol_hull_cv
		);

		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_POWER,
		    119, 6,
		    &xsw_color.bpol_power_cv
		);

		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_SHIELDS,
		    6, 62,
		    &xsw_color.bpol_shields_cv
		);

		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_VIS,
		    119, 62,
		    &xsw_color.bpol_vis_cv
		);

		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_DMGCTL,
		    30, 105,
		    &xsw_color.bpol_dmgctl_cv
		);
	    }
	    OSWPutSharedImageToDrawable(shared_tar_img, pixmap);
	    break;

	  /* ****************************************************** */
	  /* Player stats panel section 3 (uses a medium image) */
	  case BW_PANEL_P3:
	    /* Get this panel's image */
	    shared_tar_img = bw->pan_p3_img;

	    /* Draw this panel's background image */
	    BridgeDrawBlitImageAbsoluteCode(
		shared_tar_img, IMG_CODE_STATS_CON3,
		0, 0
	    );

	    /* Draw this panel's readout outlines if the object is
	     * known
	     */
	    if((shared_tar_img != NULL) && (obj != NULL))
	    {
		/* Antimatter */
		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_ANTIMATTER,
		    5, 20,
		    &xsw_color.bpol_throttle_cv
		);

		/* Left Throttle */
		i = IMG_CODE_BPANEL_OL_LTHROTTLE;
		if(ImgIsLoaded(i))
		{
		    src_img = xsw_image[i]->image;
		    if(src_img != NULL)
			BridgeDrawBlitImageCursor(
			    shared_tar_img, src_img,
			    6, 35,
			    &xsw_color.bpol_throttle_cv
			);
		}

		/* Right Throttle */
		i = IMG_CODE_BPANEL_OL_RTHROTTLE;
		if(ImgIsLoaded(i))
		{
		    src_img = xsw_image[i]->image;
		    if(src_img != NULL)
			BridgeDrawBlitImageCursor(
			    shared_tar_img, src_img,
			    BW_CONSOLE_PANEL_WIDTH -
				(int)src_img->width - 6,
			    35,
			    &xsw_color.bpol_throttle_cv
			);
		}

		/* Velocity ruler */
		BlitBufLine(
		    shared_tar_img->bytes_per_pixel, shared_tar_img->depth,
		    shared_tar_img->data,
		    6, 10,
		    shared_tar_img->width, shared_tar_img->height,
		    shared_tar_img->bytes_per_line,
		    0.5f * PI,		/* Bearing */
		    138,
		    1,			/* Broadness */
		    &xsw_color.bp_dark_outline_cv
		);
		BlitBufLine(
		    shared_tar_img->bytes_per_pixel, shared_tar_img->depth,
		    shared_tar_img->data,
		    6, 9,
		    shared_tar_img->width, shared_tar_img->height,
		    shared_tar_img->bytes_per_line,
		    0.0f * PI,		/* Bearing */ 
		    7,
		    1,			/* Broadness */
		    &xsw_color.bp_normal_outline_cv
		);
		BlitBufLine(
		    shared_tar_img->bytes_per_pixel, shared_tar_img->depth,
		    shared_tar_img->data,
		    144, 9,
		    shared_tar_img->width, shared_tar_img->height,
		    shared_tar_img->bytes_per_line,
		    0.0f * PI,		/* Bearing */ 
		    7,
		    1,			/* Broadness */
		    &xsw_color.bp_normal_outline_cv
		);

		/* Thrust vector outline */
		i = IMG_CODE_BPANEL_OL_THRUSTVECTOR;
		if(ImgIsLoaded(i))
		{
		    src_img = xsw_image[i]->image;
		    if(src_img != NULL)
			BridgeDrawBlitImageCursor(
			    shared_tar_img, src_img,
			    40 + (30 / 2) -
				((int)src_img->width / 2),
			    92 + (30 / 2) -
				((int)src_img->height / 2),
			    &xsw_color.bpol_throttle_cv
			);
		}

		/* Engine state outline */
/*
	int x = 75;
	int y = 92;
	unsigned int width = 40;
	unsigned int height = 30;
 */

	    }
	    OSWPutSharedImageToDrawable(shared_tar_img, pixmap);
	    break;

	  /* ****************************************************** */
	  /* Player stats panel section 4 */
	  case BW_PANEL_P4:
	    /* Draw this panel's background image */
	    bg_img_code = IMG_CODE_STATS_CON4;
	    tar_img = ImgIsLoaded(bg_img_code) ?
		xsw_image[bg_img_code]->image : NULL;
	    OSWPutImageToDrawable(tar_img, pixmap);
	    break;

	  /* ***************************************************** */
	  /* Subject stats panel section 1 */
	  case BW_PANEL_S1:
	    /* Get this panel's image */
	    shared_tar_img = bw->pan_s1_img;

	    /* Draw this panel's background image */
	    BridgeDrawBlitImageAbsoluteCode(
		shared_tar_img, IMG_CODE_SRO_CON1,
		0, 0
	    );

	    /* Draw vessel image */
	    if(imgref != NULL)
	    {
		int x, y, frame_num = 12;
		float zoom = 1.0f;
		unsigned int zoomed_width, zoomed_height;

		if(frame_num >= imgref->total_frames)
		    frame_num = 0;
		if(frame_num < 0)
		    frame_num = 0;

		if(imgref->fwidth > BW_CONSOLE_PANEL_WIDTH)
		{
		    zoom = (float)BW_CONSOLE_PANEL_WIDTH /
			(float)imgref->fwidth;
		    if((unsigned int)(imgref->fheight * zoom) >
			BW_CONSOLE_PANEL_HEIGHT
		    )
			zoom = (float)BW_CONSOLE_PANEL_HEIGHT /
			    (float)imgref->fheight;
		}
		else if(imgref->fheight > BW_CONSOLE_PANEL_HEIGHT)
		{
		    zoom = (float)BW_CONSOLE_PANEL_HEIGHT /
			(float)imgref->fheight;
		    if((unsigned int)(imgref->fwidth * zoom) >
			BW_CONSOLE_PANEL_WIDTH
		    )
			zoom = (float)BW_CONSOLE_PANEL_WIDTH /
			    (float)imgref->fwidth;
		}
		zoomed_width = (unsigned int)(imgref->fwidth * zoom);
		zoomed_height = (unsigned int)(imgref->fheight * zoom);

		x = (BW_CONSOLE_PANEL_WIDTH - (int)zoomed_width) / 2;
		y = (BW_CONSOLE_PANEL_HEIGHT - (int)zoomed_height) / 2;

		BridgeDrawBlitImgRef(
		    shared_tar_img, imgref,
		    x, y, frame_num, zoom, 1.0f,
		    obj->lighting & SW_OBJ_LIGHT_VECTOR,
		    False
		);
	    }

	    /* Draw this panel's readout outlines if the object is
	     * known
	     */
	    if((shared_tar_img != NULL) && (obj != NULL))
	    {
		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_HULL,
		    6, 16,
		    &xsw_color.bpol_hull_cv
		);

		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_POWER,
		    119, 16,
		    &xsw_color.bpol_power_cv
		);

		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_SHIELDS,
		    6, 72,
		    &xsw_color.bpol_shields_cv
		);

		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_VIS,
		    119, 72,
		    &xsw_color.bpol_vis_cv
		);

#if 0
		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_DMGCTL,
		    30, 106,
		    &xsw_color.bpol_dmgctl_cv
		);
#endif

		/* Draw distance line */
		BlitBufLine(
		    shared_tar_img->bytes_per_pixel, shared_tar_img->depth,
		    (u_int8_t *)shared_tar_img->data,
		    40, 129,
		    shared_tar_img->width, shared_tar_img->height,
		    shared_tar_img->bytes_per_line,
		    0.5f * PI,		/* Bearing */
		    70,
		    1,			/* Broadness */
		    &xsw_color.bp_dark_outline_cv
		);
		BlitBufLine(
		    shared_tar_img->bytes_per_pixel, shared_tar_img->depth,
		    shared_tar_img->data,
		    40, 128,
		    shared_tar_img->width, shared_tar_img->height,
		    shared_tar_img->bytes_per_line,
		    0.0 * PI,		/* Bearing */
		    5,
		    1,			/* Broadness */
		    &xsw_color.bp_normal_outline_cv
		);
		BlitBufLine(
		    shared_tar_img->bytes_per_pixel, shared_tar_img->depth,
		    shared_tar_img->data,
		    110, 128,
		    shared_tar_img->width, shared_tar_img->height,
		    shared_tar_img->bytes_per_line,
		    0.0 * PI,		/* Bearing */
		    5,
		    1,			/* Broadness */
		    &xsw_color.bp_normal_outline_cv
		);
	    }
	    OSWPutSharedImageToDrawable(shared_tar_img, pixmap);
	    break;

	  /* ***************************************************** */
	  /* Subject stats panel section 2 */
	  case BW_PANEL_S2:
	    /* Get this panel's image */
	    shared_tar_img = bw->pan_s2_img;

	    /* Draw this panel's background image */
	    BridgeDrawBlitImageAbsoluteCode(
		shared_tar_img, IMG_CODE_SRO_CON2,
		0, 0
	    );

	    /* Draw this panel's readout outlines if the object is
	     * known
	     */
	    if((shared_tar_img != NULL) && (obj != NULL))
	    {
		/* Antimatter */
		BridgeDrawBlitImageCursorCode(
		    shared_tar_img, IMG_CODE_BPANEL_OL_ANTIMATTER,
		    5, 20,
		    &xsw_color.bpol_throttle_cv
		);
	    }

	    OSWPutSharedImageToDrawable(shared_tar_img, pixmap);
	    break;

	  /* ***************************************************** */
	  /* Subject stats panel section 3 */
	  case BW_PANEL_S3:
	    /* Get this panel's image */
	    bg_img_code = IMG_CODE_SRO_CON3;
	    tar_img = ImgIsLoaded(bg_img_code) ?
		xsw_image[bg_img_code]->image : NULL;
	    OSWPutImageToDrawable(tar_img, pixmap);
	    break;

	  /* ***************************************************** */
	  default:
	    fprintf(
		stderr,
"BWDP_BACKGROUND(): Unsupported panel detail code `%i'.\n",
		panel_num
	    );
	    break;
	}

	if(cannot_draw_bkg)
	    OSWClearPixmap(
		pixmap,
		bw->pan_p1_width,
		bw->pan_p1_height,
		osw_gui[0].black_pix
	    );
}

/*
 *	Player name.
 */
static void BWDP_PNAME(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 5, y = 5;
	const unsigned int width = 145, height = 16;
	image_t *img;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return;

	/* Draw background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    int lbearing, ascent;
	    char *name;

	    switch(option.show_formal_label)
	    {
	      case XSW_SHOW_FORMAL_LABEL_ALWAYS:
		name = STRDUP(DBGetFormalNameStr(obj_num));
		break;
	      case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
		name = (obj->owner == net_parms.player_obj_num) ?
		    STRDUP(DBGetFormalNameStr(obj_num)) :
		    STRDUP(obj->name);
		break;
	      default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
		name = STRDUP(obj->name);
		break;
	    }
	    OSWSetForeground(pix);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, name, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
	        pixmap,
	        x - lbearing,
	        y + ascent,
	        name
	    );
	    free(name);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
	        w, pixmap,
	        x, y,
	        width, height,
	        x, y
	    );
}

/*
 *	Player shield frequency.
 */
static void BWDP_PSHIELDFREQ(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 5, y = 25;
	const unsigned int width = 145, height = 12;
	image_t *img;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return;

	/* Draw background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    char *s = (char *)malloc(40 + 80);
	    int lbearing, ascent;
	    sprintf(
		s,
		"Shield Freq: %.2f tHz",
	        obj->shield_frequency
	    );
	    OSWSetForeground(pix);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );
	    free(s);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player weapon frequency.
 */
static void BWDP_PWEAPONFREQ(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 5, y = 37;
	const unsigned int width = 145, height = 12;
	image_t *img;
	xsw_player_values_struct *pvals = &player_values;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return;

	/* Draw background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    char *s = (char *)malloc(40 + 80);
	    int lbearing, ascent;
	    sprintf(
		s,
"Weapon Freq: %.2f tHz",
		pvals->weapon_freq
	    );
	    OSWSetForeground(pix);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );
	    free(s);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player communications channel.
 */
static void BWDP_PCOMCHANNEL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 5, y = 49;
	const int width = 145, height = 12;
	image_t *img;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return;

	/* Draw background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    char *s = (char *)malloc(40 + 80);
	    int lbearing, ascent;
	    sprintf(
		s,
		"Com Channel: %.2f tHz",
		(float)obj->com_channel / 100.0
	    );
	    OSWSetForeground(pix);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );
	    free(s);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player score:
 */
static void BWDP_PSCORE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 5, y = 63;
	const int width = 145, height = (12 * 5) + (2 * 1);
	image_t *img;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return; 

	/* Draw background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    xsw_score_struct *score = obj->score;
	    char *s = (char *)malloc(40 + 40 + 80);
	    int y_pos = y, lbearing, ascent;
	    sprintf(
		s,
		"Credits: %.2f",
		(score != NULL) ? (float)score->credits : 0.0
	    );
	    OSWSetForeground(pix);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y_pos + ascent,
		s
	    );
	    y_pos += 12;

	    sprintf(
		s,
		"RMU: %.2f(%.2f)",
		(score != NULL) ? (float)score->rmu : 0.0,
		(score != NULL) ? (float)score->rmu_max : 0.0
	    );
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y_pos + ascent,
		s
	    );
	    y_pos += 12 + 2;

	    sprintf(
		s,
		"Dmg Given: %.2f",
		(score != NULL) ? (float)score->damage_given : 0.0
	    );
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y_pos + ascent,
		s
	    );
	    y_pos += 12;

	    sprintf(
		s,
		"Dmg Recv: %.2f",
		(score != NULL) ? (float)score->damage_received : 0.0
	    );
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y_pos + ascent,
		s
	    );
	    y_pos += 12;

	    sprintf(
		s,
		"Kills: %i",
		(score != NULL) ? (int)score->kills : 0
	    );
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y_pos + ascent,
		s
	    );

	    free(s);
	}
	 
	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}


/*
 *	Player hull.
 */
static void BWDP_PHULL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 6, y = 6;
	const unsigned int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_p2_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, ascent, ht, hv;
	    const float c = (obj->hp_max > 0.0f) ?
		MAX((obj->hp / obj->hp_max), 0.0f) : 0.0f;
	    char s[40];

	    if(c > XSW_HP_WARNING_COEFF)
		OSWSetForeground(pix);
	    else if(c > XSW_HP_DANGER_COEFF)
		OSWSetForeground(xsw_color.bp_warning);
	    else
		OSWSetForeground(xsw_color.bp_danger);

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    OSWDrawSolidRectangle(
		pixmap,
		x,
		y + BW_BAR_MARGIN + (ht - hv),
		6,
		(unsigned int)hv
	    );

	    if(obj->hp_max > 0.0)
	    {
		OSWSetFont(font);
	        sprintf(s, "%.0f", obj->hp);
	        OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, NULL
	        );
	        OSWDrawString(
		    pixmap,
		    x + 10 + 15 - lbearing,
		    y + ascent,
		    s
	        );
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player core power.
 */
static void BWDP_PPOWER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 75, y = 6;
	const unsigned int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_p2_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, swidth, ascent, hv, ht;
	    const float c = (obj->power_max > 0.0) ?
		MAX((obj->power / obj->power_max), 0.0f) : 0.0f;
	    char s[40];

	    if(c > XSW_POWER_WARNING_COEFF)
		OSWSetForeground(pix);
	    else if(c > XSW_POWER_DANGER_COEFF)
		OSWSetForeground(xsw_color.bp_warning);
	    else
		OSWSetForeground(xsw_color.bp_danger);

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    OSWDrawSolidRectangle(
		pixmap,
		137,
		y + BW_BAR_MARGIN + (ht - hv),
		6,
		(unsigned int)hv
	    );

	    if(obj->power_max > 0.0)
	    {
		OSWSetFont(font);
		sprintf(s, "%.0f", obj->power);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, &swidth, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    137 - 15 - swidth - lbearing,
		    y + ascent,
		    s
		);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y, 
		width, height,
		x, y  
	    );
}

/*
 *	Player shields.
 */
static void BWDP_PSHIELDS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 6, y = 62;
	const int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_p2_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, ascent, descent, font_height, hv, ht;
	    const float c = (obj->power_max > 0.0f) ?
		MAX((obj->power / obj->power_max), 0.0f) : 0.0f;
	    char s[40];

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    if(obj->shield_state == SW_SHIELD_STATE_UP)
	    {
		if(c > XSW_POWER_WARNING_COEFF)
		    OSWSetForeground(pix);
		else if(c > XSW_POWER_DANGER_COEFF)
		    OSWSetForeground(xsw_color.bp_warning);
		else
		    OSWSetForeground(xsw_color.bp_danger);
		OSWDrawSolidRectangle(
		    pixmap,
		    x,
	 	    y + BW_BAR_MARGIN + ((ht - hv) / 2),
		    6,
		    (unsigned int)hv
		);
		OSWSetFont(font);
		sprintf(s, "%.0f", obj->power);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    x + 10 + 15 - lbearing,
		    y + (int)height - font_height + ascent,
		    s
		);
	    }
	    else if(obj->shield_state == SW_SHIELD_STATE_DOWN)
	    {
#if 0
		OSWSetForeground(pix);
		OSWSetFont(font);
		strcpy(s, "Down");
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    x + 10 + 15 - lbearing,
		    y + (int)height - font_height + ascent,
		    s
		);
#endif
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y, 
		width, height,
		x, y  
	    );
}

/*
 *	Player visibility.
 */
static void BWDP_PVIS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 75, y = 62;
	const unsigned int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_p2_img;
	if(img == NULL)
	    return; 

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, swidth, ascent, descent, font_height, hv, ht;
	    float c = MAX(DBGetObjectVisibility(obj_num), 0.0);
	    char s[40];

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    OSWSetForeground(pix);
	    OSWDrawSolidRectangle(
		pixmap,
		137,
		y + BW_BAR_MARGIN + (ht - hv),
		6,
		(unsigned int)hv
	    );

	    OSWSetFont(font);
	    sprintf(s, "%.0f", c * 100.0);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWDrawString(
		pixmap,
		137 - 15 - swidth - lbearing,
		y + (int)height - font_height + ascent,
		s
	    );
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y, 
		width, height,
		x, y  
	    );
}

/*
 *	Player damage control.
 */
static void BWDP_PDMGCTL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 30, y = 122;
	const unsigned int width = 42, height = 8;
	shared_image_t *img = xsw_bridge_win->pan_p2_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    if(obj->damage_control == SW_DMGCTL_STATE_ON)
	    {
		int hv, ht;
		float c;

		if(obj->power_max != 0.0)
		    c = obj->power / obj->power_max;
		else
		    c = 1.0;

		ht = (int)width - (2 * BW_BAR_MARGIN);
		hv = (int)(ht * c);

		OSWSetForeground(pix);
		OSWDrawSolidRectangle(
		    pixmap,
		    x + BW_BAR_MARGIN,
		    y + 1,
		    (unsigned int)hv,
		    6
		);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y, 
		width, height,
		x, y  
	    );
}


/*
 *	Subject name.
 */
static void BWDP_SNAME(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 5, y = 0;
	const unsigned int width = 145, height = 16;
	shared_image_t *img = xsw_bridge_win->pan_s1_img;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    int lbearing, ascent;
	    char *name;

	    switch(option.show_formal_label)
	    {
	      case XSW_SHOW_FORMAL_LABEL_ALWAYS:
		name = STRDUP(DBGetFormalNameStr(obj_num));
		break;
	      case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
		name = (obj->owner == net_parms.player_obj_num) ?
		    STRDUP(DBGetFormalNameStr(obj_num)) :
		    STRDUP(obj->name);
		break;
	      default:  /* XSW_SHOW_FORMAL_LABEL_NEVER */
		name = STRDUP(obj->name);
		break;
	    }
	    OSWSetForeground(pix);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, name, -1,
		&lbearing, NULL, NULL, &ascent, NULL
	    );
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		name
	    );
	    free(name);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject hull.
 */
static void BWDP_SHULL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 6, y = 16;
	const unsigned int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_s1_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);
		
	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, ascent, hv, ht;
	    const float c = (obj->hp_max > 0.0f) ?
		MAX((obj->hp / obj->hp_max), 0.0f) : 0.0f;
	    char s[40];

	    if(c > XSW_HP_WARNING_COEFF)
		OSWSetForeground(pix);
	    else if(c > XSW_HP_DANGER_COEFF)
		OSWSetForeground(xsw_color.bp_warning);
	    else
		OSWSetForeground(xsw_color.bp_danger);

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    OSWDrawSolidRectangle(
		pixmap,
		x,
		y + BW_BAR_MARGIN + (ht - hv),
		6,
		(unsigned int)hv
	    );

	    if(obj->hp_max > 0.0)
	    {
		OSWSetFont(font);
		sprintf(s, "%.0f", obj->hp);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    x + 10 + 15 - lbearing,
		    y + ascent,
		    s
		);
	    }
	}
	
	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject core power.
 */
static void BWDP_SPOWER( 
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 75, y = 16;
	const unsigned int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_s1_img;
	if(img == NULL)
	    return;
		
	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, swidth, ascent, hv, ht;
	    const float c = (obj->power_max > 0.0f) ?
		MAX((obj->power / obj->power_max), 0.0f) : 0.0f;
	    char s[40];

	    if(c > XSW_POWER_WARNING_COEFF)
		OSWSetForeground(pix);
	    else if(c > XSW_POWER_DANGER_COEFF)
		OSWSetForeground(xsw_color.bp_warning);
	    else
		OSWSetForeground(xsw_color.bp_danger);

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    OSWDrawSolidRectangle(
		pixmap,
		137,
		y + BW_BAR_MARGIN + (ht - hv),
		6,
		(unsigned int)hv
	    );

	    if(obj->power_max > 0.0)
	    {
		OSWSetFont(font);
		sprintf(s, "%.0f", obj->power);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, &swidth, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    137 - 15 - swidth - lbearing,
		    y + ascent,
		    s
		);
	    }
	}
	
	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject shields.
 */
static void BWDP_SSHIELDS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 6, y = 72;
	const unsigned int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_s1_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, ascent, descent, font_height, hv, ht;
	    const float c = (obj->power_max > 0.0f) ?
		MAX((obj->power / obj->power_max), 0.0f) : 0.0f;
	    char s[80];

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    if(obj->shield_state == SW_SHIELD_STATE_UP)
	    {
		if(c > XSW_POWER_WARNING_COEFF)
		    OSWSetForeground(pix);
		else if(c > XSW_POWER_DANGER_COEFF)
		    OSWSetForeground(xsw_color.bp_warning);
		else
		    OSWSetForeground(xsw_color.bp_danger);
		OSWDrawSolidRectangle(
		    pixmap,
		    x,
		    y + BW_BAR_MARGIN + ((ht - hv) / 2),
		    6,
		    (unsigned int)hv
		);

		OSWSetFont(font);
		sprintf(s, "%.0f", obj->power);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    x + 10 + 15 - lbearing,
		    y + (int)height - font_height + ascent,
		    s
		);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject visibility.
 */
static void BWDP_SVIS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 75, y = 72;
	const unsigned int width = 70, height = 50;
	shared_image_t *img = xsw_bridge_win->pan_s1_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int lbearing, swidth, ascent, descent, font_height, hv, ht;
	    float c = MAX(DBGetObjectVisibility(obj_num), 0);
	    char s[40];

	    OSWSetForeground(pix);

	    ht = (int)height - (2 * BW_BAR_MARGIN);
	    hv = (int)(ht * c);

	    OSWDrawSolidRectangle(
		pixmap,
		137,
		y + BW_BAR_MARGIN + (ht - hv),
		6,
		(unsigned int)hv
	    );

	    OSWSetFont(font);
	    sprintf(s, "%.0f", c * 100.0);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWDrawString(
		pixmap,
		137 - 15 - swidth - lbearing,
		y + (int)height - font_height + ascent,
		s
	    );
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject bearing.
 */
static void BWDP_SBEARING(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int	x = 10,
			y = 0;
	const unsigned int width = 40, height = 18;
	shared_image_t *img = xsw_bridge_win->pan_s2_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int ref_obj_num = net_parms.player_obj_num;
	    if(!DBIsObjectGarbage(ref_obj_num))
	    {
		bool_t in_same_sector;
		int	tx[2], ty[2],
			lbearing, ascent, descent, font_height;
		unsigned int	twidth = height - 6,
				theight = height - 6;
		float bearing;
		char s[40];
		xsw_object_struct *ref_obj = xsw_object[ref_obj_num];

		/* Check if objects are in same the sector */
		if((obj->sect_x == ref_obj->sect_x) &&
		   (obj->sect_y == ref_obj->sect_y) &&
		   (obj->sect_z == ref_obj->sect_z)
		)
		    in_same_sector = True;
		else
		    in_same_sector = False;

		/* Get bearing */
		if(in_same_sector)
		    bearing = SWMathCoordinateDeltaVector(
			obj->x - ref_obj->x,
			obj->y - ref_obj->y
		    );
		else
		    bearing = SWMathCoordinateDeltaVector(
			obj->sect_x - ref_obj->sect_x,
			obj->sect_y - ref_obj->sect_y 
		    );

		/* Adjust bearing if scanner is oriented to object */
		if(xsw_bridge_win->scanner_orient == XSW_SCANNER_ORIENT_PLAYER)
		    bearing = SANITIZERADIANS(
			bearing - ref_obj->heading
		    );

		/* Get center of bearing tick in window coordinates */
		tx[0] = x + ((int)twidth / 2);
		ty[0] = y + ((int)height / 2);

		/* Draw bearing circle */
		OSWSetForeground(xsw_color.bp_dark_outline);
		OSWDrawArc(
		    pixmap,
		    tx[0] - ((int)twidth / 2),
		    ty[0] - ((int)theight / 2),
		    twidth, theight,
		    0.0 * PI, 2.0 * PI
		);

		/* Get end of bearing tick */
		tx[1] = tx[0] + (int)((float)sin((double)bearing) * (int)(twidth / 2));
		ty[1] = ty[0] - (int)((float)cos((double)bearing) * (int)(theight / 2));

		/* Draw bearing tick */
		OSWSetForeground(xsw_color.bp_light_outline);
		OSWDrawLine(
		    pixmap,
		    tx[0], ty[0],
		    tx[1], ty[1]
		);

		/* Draw bearing value */
		OSWSetForeground(pix);
		sprintf(s, "%.0f'", bearing * 180 / PI);
		OSWSetFont(font);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    tx[0] + ((int)twidth / 2) + 5 - lbearing,
		    y + (((int)height - font_height) / 2) + ascent,
		    s
		);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject distance bar.
 */
static void BWDP_SDISTANCE_BAR(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
) 
{
	const int x = 40, y = 123;
	const unsigned int width = 70, height = 7;
	shared_image_t *img = xsw_bridge_win->pan_s1_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height   
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int ref_obj_num = net_parms.player_obj_num;
	    if(!DBIsObjectGarbage(ref_obj_num))
	    {
		float c, d;
		xsw_object_struct *ref_obj = xsw_object[ref_obj_num];
	
		/* In same sector? */
		if((obj->sect_x == ref_obj->sect_x) &&
		   (obj->sect_y == ref_obj->sect_y) &&
		   (obj->sect_z == ref_obj->sect_z)
		)
		{
		    /* Calculate distance */
		    d = SWMath3DDistance(
			obj->x - ref_obj->x,
			obj->y - ref_obj->y,
			obj->z - ref_obj->z
		    );
		    c = MIN(d / ref_obj->scanner_range, 1.0);
		    if(c < 0.0)
			c = 0.0;

		    OSWSetForeground(pix);
		    OSWDrawLine(
			pixmap, 
			(int)(x + (c * (float)width)),
			y,
			(int)(x + (c * (float)width)),
			y + 5
		    );
		}
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y, 
		width, height,
		x, y
	    );
}

/*
 *	Subject distance.
 */
static void BWDP_SDISTANCE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int       x = 150 - 40 - 10,
			y = 0;
	const unsigned int width = 40, height = 18;
	shared_image_t *img = xsw_bridge_win->pan_s2_img;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    int ref_obj_num = net_parms.player_obj_num;
	    if(!DBIsObjectGarbage(ref_obj_num))
	    {
		int lbearing, swidth, ascent, descent, font_height;
		float c, d;
		const char *units_str;
		char s[80];
		xsw_object_struct *ref_obj = xsw_object[ref_obj_num];

		/* In same sector? */
		if((obj->sect_x == ref_obj->sect_x) &&
		   (obj->sect_y == ref_obj->sect_y) &&
		   (obj->sect_z == ref_obj->sect_z)
		)
		{
		    sw_legend_struct *legend = &sw_legend;
		    sw_units_struct units;

		    /* Calculate distance */
		    d = SWMath3DDistance(
			obj->x - ref_obj->x,
			obj->y - ref_obj->y,
			obj->z - ref_obj->z
		    );
		    c = MIN(d / ref_obj->scanner_range, 1.0f);
		    if(c < 0.0f)
			c = 0.0f;

		    units.ru_to_au = legend->ru_to_au;
		    switch(option.units)
		    {
		      case XSW_UNITS_METRIC:
			d = SWConvertRUToAU(&units, d);
			units_str = "au";
			break; 
		      case XSW_UNITS_ENGLISH:
			d = SWConvertRUToKiloMiles(&units, d);
			units_str = "kmiles";
			break;
		      case XSW_UNITS_PARSECS:
			d = SWConvertRUToMP(&units, d);
			units_str = "mp";
			break;
		      default:
			units_str = "ru";
			break;
		    }
		    if(d >= 100.0f)
			sprintf(s, "%.0f %s", d, units_str);
		    else if(d >= 10.0f)
			sprintf(s, "%.1f %s", d, units_str);
		    else if(d >= 1.0f)
			sprintf(s, "%.2f %s", d, units_str);
		    else if(d >= 0.1f)
			sprintf(s, "%.3f %s", d, units_str);
		    else
			sprintf(s, "%.4f %s", d, units_str);
		    OSWSetForeground(pix);
		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, s, -1,
			&lbearing, NULL, &swidth, &ascent, &descent
		    );
		    font_height = ascent + descent;
		    OSWDrawString(
			pixmap,
			x + MAX((((int)width - swidth) / 2), 0) - lbearing,
			y + (((int)height - font_height) / 2) + ascent,
			s
		    );
		}
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

#if 0
/*
 *	Subject damage control.
 */
static void BWDP_SDMGCTL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 30, y = 122;
	const unsigned int width = 42, height = 8;
	float c;
	shared_image_t *img_ptr = xsw_bridge_win->pan_s1_img;
	if(img_ptr == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img_ptr, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    if(obj->damage_control == SW_DMGCTL_STATE_ON)
	    {
		if(obj->power_max != 0.0)
		    c = (float)obj->power /
			(float)obj->power_max;
		else
		    c = 1.0;

		OSWSetForeground(pix);
		OSWDrawSolidRectangle(
		    pixmap,
		    x + BW_BAR_MARGIN,
		    y + 2,
		    static_cast<unsigned int>(((int)width - (2 * BW_BAR_MARGIN)) * c),
		    6
		);
	    }
	}
	
	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}
#endif


/*
 *	Player intercepting object name.
 */
static void BWDP_PINAME(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 7, y = 0;
	const unsigned int width = 144, height = 12;
	image_t *img;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return;

	/* Draw the background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw texts */
	if(obj != NULL)
	{
	    int lbearing, ascent;
	    char *s;
	    const char *name;
	    const int i_obj_num = obj->intercepting_object;
	    if(!DBIsObjectGarbage(i_obj_num))
	    {
		OSWSetFont(font);
		OSWSetForeground(pix);

		switch(option.show_formal_label)
		{
		  case XSW_SHOW_FORMAL_LABEL_ALWAYS:
		    name = DBGetFormalNameStr(i_obj_num);
		    s = (char *)malloc(STRLEN(name) + 80);
		    sprintf(
			s,
			"I: %s",
			(name != NULL) ? name : ""
		    );
		    break;
		  case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
		    if(obj_num == xsw_object[i_obj_num]->owner)
			name = DBGetFormalNameStr(i_obj_num);
		    else
			name = xsw_object[i_obj_num]->name;
		    s = (char *)malloc(STRLEN(name) + 80);
		    sprintf(
			s,
			"I: %s",
			(name != NULL) ? name : ""
		    );
		    break;
		  default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
		    name = xsw_object[i_obj_num]->name;
		    s = (char *)malloc(STRLEN(name) + 80);
		    sprintf(
			s,
			"I: %s",
			(name != NULL) ? name : ""
		    );
		    break;
		}

		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    x - lbearing,
		    y + ascent,
		    s
		);
		free(s);
	    }
	    /* Intercepting unknown object? */
	    else if(i_obj_num > -1)
	    {
		OSWSetFont(font);
		OSWSetForeground(pix);

		s = (char *)malloc(80 + 40);
		sprintf(s, "I: Object #%i", i_obj_num);

		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    x - lbearing,
		    y + ascent,
		    s
		);
		free(s);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y 
	    );
}

/*
 *	Player lock on name.
 */
static void BWDP_PWLOCK(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 7, y = 14;
	const unsigned int width = 144, height = 12;
	image_t *img;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return;

	/* Draw the background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw texts */
	if(obj != NULL) 
	{
	    int lbearing, ascent;
	    char *s;
	    const char *name;
	    const int locked_on = obj->locked_on;
	    if(!DBIsObjectGarbage(locked_on))
	    {
		OSWSetFont(font);
		OSWSetForeground(pix);

		switch(option.show_formal_label)
		{
		  case XSW_SHOW_FORMAL_LABEL_ALWAYS:
		    name = DBGetFormalNameStr(locked_on);
		    s = (char *)malloc(STRLEN(name) + 80);
		    sprintf(
			s,
			"WL: %s",
			(name != NULL) ? name : ""
		    );
		    break;
		  case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
		    if(obj_num == xsw_object[locked_on]->owner)
			name = DBGetFormalNameStr(locked_on);
		    else
			name = xsw_object[locked_on]->name;
		    s = (char *)malloc(STRLEN(name) + 80);
		    sprintf(
			s,
			"WL: %s",
			(name != NULL) ? name : ""
		    );
		    break;
		  default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
		    name = xsw_object[locked_on]->name;
		    s = (char *)malloc(STRLEN(name) + 80);
		    sprintf(
			s,
			"WL: %s",
			(name != NULL) ? name : ""
		    );
		    break;
		}

		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    x - lbearing,
		    y + ascent,
		    s
		);
		free(s);
	    }
	    /* Locked on unknown object? */
	    else if(locked_on > -1)
	    {
		OSWSetFont(font);
		OSWSetForeground(pix);

		s = (char *)malloc(40 + 80);
		sprintf(s, "WL: Object #%i", locked_on);

		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    x - lbearing,
		    y + ascent,
		    s
		);
		free(s);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player weapons.
 */
static void BWDP_PWEAPONS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 7, y = 28;
	const unsigned int width = 144, height = 150 - y;
	int i;
	image_t *img;
	xsw_weapon_struct *wep;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_player_values_struct *pvals = &player_values;

	if(!ImgIsLoaded(bkg_img_code))
	    return;

	img = xsw_image[bkg_img_code]->image;

	/* Draw the background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Note: foreground color is not set by the specified pix */


	/* Draw weapons listing */
	if(obj != NULL)
	{
	    const time_t cur_ms = cur_millitime;
	    int	input_area_num,
		lbearing, swidth, ascent, descent, font_height;
	    char *s, flags_str[40];
	    rectangle_t rect;
	    xsw_input_area_struct *a; 

	    OSWSetFont(font);

	    /* Draw the weapons yield bar? */
	    if(pvals->weapons_online)
	    {
		const unsigned int	yield_bar_width = 80,
					yield_bar_height = 6;
		const float yield = CLIP(pvals->weapon_yield, 0.0f, 1.0f);
		int bwidth;

		s = "Yld:";
		OSWSetForeground(xsw_color.bp_standard_text);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    x - lbearing,
		    y + ascent,
		    s
		);

		if(yield > 0.35f)
		    OSWSetForeground(xsw_color.bp_danger);
		else
		    OSWSetForeground(xsw_color.bp_warning);

		bwidth = (unsigned int)FTOI_RNDOFF(
		    yield * (float)yield_bar_width
		);
		OSWDrawSolidRectangle(
		    pixmap,
		    30,
		    y + FTOI_RNDOFF(
			(float)(font_height - (int)yield_bar_height) / 2.0f
		    ),
		    bwidth, yield_bar_height
		);

		s = (char *)malloc(40);
		OSWSetForeground(xsw_color.bp_standard_text);
		sprintf(
		    s,
		    "%i%%",
		    FTOI_RNDOFF(yield * 100.0f)
		);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, NULL
		);
		OSWDrawString(
		    pixmap,
		    30 + yield_bar_width + 2 - lbearing,
		    y + ascent,
		    s
		);
		free(s);
	    }
	    else
	    {
		OSWSetForeground(xsw_color.bp_withdrawn_text);
	    }

	    /* Set up the clip rectangle to draw weapons listing */
	    rect.x = x;
	    rect.y = y;
	    rect.width = width;
	    rect.height = height;
	    OSWSetClipRectangle(&rect);

	    /* Draw each weapon */
	    for(i = 0; i < obj->total_weapons; i++)
	    {
		wep = obj->weapons[i];
		if(wep == NULL)
		    continue;

		/* Get the input area that corresponds to this weapon */
		input_area_num = BW_INPUT_AREA_PSELWEP1 + i;
		if((input_area_num > BW_INPUT_AREA_PSELWEP9) ||
		   (input_area_num < 0) ||
		   (input_area_num >= bw->ninput_areas)
		)
		    continue;

		a = bw->input_area[input_area_num];

		/* Format the weapon string by its emission type
		 *
		 * The prefixes are as follows:
		 *
		 *	^ = Insufficient power to use weapon
		 *	~ = Unable to use weapon due to cloak being up
		 *	! = Weapon is in a refractory period
		 */
		switch(wep->emission_type)
		{
		  case SW_WEP_EMISSION_PULSE:
		    *flags_str = '\0';
		    if(wep->create_power > obj->power)
			strcat(flags_str, "^");
		    if((obj->cloak_state == SW_CLOAK_STATE_UP) ?
			(!(wep->flags & SW_WEP_FLAG_FIRE_CLOAKED)) : False
		    )
			strcat(flags_str, "~");
		    s = (char *)malloc(
			STRLEN(flags_str) + STRLEN(wep->name) +
			40 + 80
		    );
		    sprintf(
			s,
			"%s%s: %.0f%%",
			flags_str,
			wep->name,
			(obj->power_max != 0.0f) ?
			    (obj->power / obj->power_max * 100.0f) : 0.0f
		    );
		    break;
		  case SW_WEP_EMISSION_PROJECTILE:
		    *flags_str = '\0';
		    if(cur_ms < (wep->delay + wep->last_used))
			strcat(flags_str, "!");
		    if(wep->create_power > obj->power)
			strcat(flags_str, "^");
		    if((obj->cloak_state == SW_CLOAK_STATE_UP) ?
			(!(wep->flags & SW_WEP_FLAG_FIRE_CLOAKED)) : False
		    )
			strcat(flags_str, "~");
		    s = (char *)malloc(
			STRLEN(flags_str) + STRLEN(wep->name) +
			40 + 40 + 80
		    );
		    sprintf(
			s,
			"%s%s: %d(%d)",
			flags_str,
			wep->name,
			wep->amount,
			wep->max
		    );
		    break;
		  default:	/* SW_WEP_EMISSION_STREAM */
		    *flags_str = '\0';
		    if(wep->create_power > obj->power)
			strcat(flags_str, "^");
		    if((obj->cloak_state == SW_CLOAK_STATE_UP) ?
			(!(wep->flags & SW_WEP_FLAG_FIRE_CLOAKED)) : False
		    )
			strcat(flags_str, "~");
		    s = (char *)malloc(
			STRLEN(flags_str) + STRLEN(wep->name) +
			40 + 80
		    );
		    sprintf(
			s,
			"%s%s: %.0f%%",
			flags_str,
			wep->name,
			(obj->power_max != 0.0f) ?
			    (obj->power / obj->power_max * 100.0f) : 0.0f
		    );
		    break;
		}
		if(pvals->weapons_online)
		{
		    /* This weapon is currently selected? */
		    if(obj->selected_weapon == i)
		    {
			/* Use selected weapon color */
			OSWSetForeground(xsw_color.bp_bold_text);
		    }
		    else
		    {
			/* Weapon is not selected, use standard color
			 * (or withdrawn color if this projectile
			 * weapon is out of stock
			 */
			if((wep->emission_type == SW_WEP_EMISSION_PROJECTILE) &&
			   (wep->amount == 0) && (wep->max > 0)
			)
			    OSWSetForeground(xsw_color.bp_withdrawn_text);
			else
			    OSWSetForeground(xsw_color.bp_standard_text);
		    }
		}
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    a->x - lbearing,
		    a->y + (((int)a->height - font_height) / 2) + ascent,
		    s
		);
		free(s);
	    }

	    OSWSetClipRectangle(NULL);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}


/*
 *	Subject empire.
 */
static void BWDP_SEMPIRE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = (150 / 2) - 20, y = 0;
	const unsigned int width = 40, height = 20;
	image_t *img;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return; 
	      
	/* Draw background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw empire text */
	if((obj != NULL) && (net_parms.player_obj_ptr != NULL))
	{
	    /* Get color to represent empire */
	    switch(UNVMatchIFFPtr(obj, net_parms.player_obj_ptr))
	    {
	      case SW_IFF_FRIENDLY:
		pix = xsw_color.bp_friendly;
		break;
	      case SW_IFF_HOSTILE:
		pix = xsw_color.bp_hostile;
		break;
	      default:	/* SW_IFF_UNKNOWN */
		pix = xsw_color.bp_unknown;
		break;
	    }
	    OSWSetForeground(pix);

	    if(*obj->empire != '\0')
	    {
		int lbearing, swidth, ascent, descent, font_height;
		char *s = (char *)malloc(strlen(obj->empire) + 40);
		sprintf(s, "(%s)", obj->empire);
		OSWSetFont(font);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    x + (((int)width - swidth) / 2) - lbearing,
		    y + (((int)height - font_height) / 2) + ascent,
		    s
		);
		free(s);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject weapons.
 */
static void BWDP_SWEAPONS(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 7, y = 0;
	const unsigned int width = 144, height = 150 - y;
	int i, y_pos;
	char text[80 + SW_OBJ_NAME_MAX];
	image_t *img; 
	xsw_weapon_struct *wep;

	if(ImgIsLoaded(bkg_img_code))
	    img = xsw_image[bkg_img_code]->image;
	else
	    return;

	/* Draw the background */
	OSWPutImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Note: foreground color is not set by value of pix */

	/* Draw weapons listing */
	if(obj != NULL)
	{
	    int lbearing, ascent, descent, font_height;

	    OSWSetFont(font);

	    /* Set starting y position for drawing weapons listing */
	    y_pos = y;
	    for(i = 0; i < obj->total_weapons; i++)
	    {
		wep = obj->weapons[i];
		if(wep == NULL)
		    continue;

		*text = '\0';
		switch(wep->emission_type)
		{
		  case SW_WEP_EMISSION_PULSE:
		    sprintf(
			text,
			"%s: %.0f%%",
			wep->name,
			(obj->power_max != 0.0) ?
			    (obj->power / obj->power_max * 100.0) : 0.0
		    );
		    break;

		  case SW_WEP_EMISSION_PROJECTILE:
		    sprintf(
			text,
			"%s: %d(%d)",
			wep->name,
			wep->amount,
			wep->max
		    );
		    break;

		  default:	/* SW_WEP_EMISSION_STREAM */
		    sprintf(
			text,
			"%s: %.0f%%",
			wep->name,
			(obj->power_max != 0.0) ?
			    (obj->power / obj->power_max * 100.0) : 0.0
		    );
		    break;
		}
		if(True)
		{
		    /* This weapon is currently selected? */
		    if(obj->selected_weapon == i)
		    {
			/* Use selected weapon color */
			OSWSetForeground(xsw_color.bp_bold_text);
		    }
		    else
		    {
			/* Weapon is not selected, use standard color
			 * (or withdrawn color if this projectile
			 * weapon is out of stock
			 */
			if((wep->emission_type == SW_WEP_EMISSION_PROJECTILE) &&
			   (wep->amount == 0) && (wep->max > 0)
			)
			    OSWSetForeground(xsw_color.bp_withdrawn_text);
			else
			    OSWSetForeground(xsw_color.bp_standard_text);
		    }
		}
		OSWGetTextBounds(
		    font, text, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    x - lbearing,
		    y_pos + ((12 - font_height) / 2) + ascent,
		    text
		);
		y_pos += 12;
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}


/*
 *	Player throttle.
 */
static void BWDP_PTHROTTLEL(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 6, y = 35;
	const unsigned int width = 10, height = 90;
	shared_image_t *img = xsw_bridge_win->pan_p3_img;
	if(img == NULL)
	    return; 

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL) 
	{
	    const float throttle = CLIP(obj->throttle, 0.0, 1.0);
	    const int	h = (int)height - (2 * BW_BAR_MARGIN),
			th = (int)(h * throttle);

	    if((obj->thrust_dir >= (PI * 0.4)) &&
	       (obj->thrust_dir <= (PI * 1.6))
	    )
	        OSWSetForeground(pix);
	    else
		OSWSetForeground(xsw_color.bpol_throttle_rev);

	    if(th > 0)
		OSWDrawSolidRectangle(
		    pixmap,
		    x,
		    y + BW_BAR_MARGIN + (h - th),
		    8,
		    (unsigned int)th
		);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y  
	    );
}

static void BWDP_PTHROTTLER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w, 
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 134, y = 35;
	const unsigned int width = 10, height = 90;
	shared_image_t *img_ptr = xsw_bridge_win->pan_p3_img;
	if(img_ptr == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img_ptr, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */ 
	if(obj != NULL)
	{
	    const float throttle = CLIP(obj->throttle, 0.0, 1.0);
	    const int	h = (int)height - (2 * BW_BAR_MARGIN),
			th = (int)(h * throttle);

	    if((obj->thrust_dir >= (PI * 0.4)) &&
	       (obj->thrust_dir <= (PI * 1.6))
	    )
		OSWSetForeground(pix);
	    else
		OSWSetForeground(xsw_color.bpol_throttle_rev); 

	    if(th > 0)
		OSWDrawSolidRectangle(
		    pixmap,
		    x + 2,
		    y + BW_BAR_MARGIN + (h - th),
		    8,
		    (unsigned int)th
		);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

static void BWDP_PTHROTTLE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w, 
	int bkg_img_code,
	bool_t put_to_window
)
{
	BWDP_PTHROTTLEL(
	    obj_num, obj,
	    font, pix,
	    pixmap, w,
	    bkg_img_code,
	    put_to_window
	);

	BWDP_PTHROTTLER(
	    obj_num, obj,
	    font, pix,
	    pixmap, w,
	    bkg_img_code,
	    put_to_window
	);
}

/*
 *	Player velocity.
 */
static void BWDP_PVELOCITY(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 6, y = 0;
	const unsigned int width = 138, height = 10;
	float s;
	char text[80];
	shared_image_t *img_ptr = xsw_bridge_win->pan_p3_img;
	if(img_ptr == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img_ptr, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	if(obj != NULL)
	{
	    sw_legend_struct *legend = &sw_legend;
	    sw_units_struct units;
	    int lbearing, swidth, ascent;
	    float	vc = (float)((obj->velocity_max > 0.0f) ?
		(obj->velocity / obj->velocity_max) : 0.0f
	    ), warning_vc = 0.0f;

	    if(!DBIsObjectGarbage(obj->loc_obj_num))
	    {
		xsw_object_struct *area_obj = xsw_object[obj->loc_obj_num];
		if(area_obj->loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
		{
		    if((area_obj->velocity_max > 0.0) &&
		       (area_obj->velocity_max < obj->velocity_max)
		    )
			warning_vc = (float)((obj->velocity_max > 0.0) ?
			    (area_obj->velocity_max / obj->velocity_max) : 0.0
			);
		}
	    }

	    /* Warning tick */
	    if(warning_vc > 0.0)
	    {
		OSWSetForeground(xsw_color.bp_warning);
		OSWDrawLine(
		    pixmap,
		    (int)(x + (width * warning_vc)),
		    y + 2,
		    (int)(x + (width * warning_vc)),
		    y + 9
		);
	    }

	    /* Tick */
	    OSWSetForeground(pix);
	    OSWDrawLine(
		pixmap,
		(int)(x + (width * vc)),
		y + 2,
		(int)(x + (width * vc)),
		y + 9
	    );

	    /* Label */
	    units.ru_to_au = legend->ru_to_au;
	    switch(option.units)
	    {
	      case XSW_UNITS_METRIC:
		s = SWConvertRUPerCycleToAUPerSecond(&units, obj->velocity);
		if(s >= 100.0f)
		    sprintf(text, "%.0f au/s", s);
		else if(s >= 10.0f)
		    sprintf(text, "%.1f au/s", s);
		else if(s >= 1.0f)
		    sprintf(text, "%.2f au/s", s);
		else if(s >= 0.1f)
		    sprintf(text, "%.3f au/s", s);
		else
		    sprintf(text, "%.4f au/s", s);
		break;
	      case XSW_UNITS_ENGLISH:
		s = SWConvertRUPerCycleToKiloMilesPerSecond(&units, obj->velocity);
		if(s >= 100.0f)
		    sprintf(text, "%.0f kmiles/s", s);
		else if(s >= 10.0f)
		    sprintf(text, "%.1f kmiles/s", s);
		else if(s >= 1.0f)
		    sprintf(text, "%.2f kmiles/s", s);
		else if(s >= 0.1f)
		    sprintf(text, "%.3f kmiles/s", s);
		else
		    sprintf(text, "%.4f kmiles/s", s);
		break;
	      case XSW_UNITS_PARSECS:
		s = SWConvertRUPerCycleToMPPerSecond(&units, obj->velocity);
		if(s >= 100.0f)
		    sprintf(text, "%.0f mp/s", s);
		else if(s >= 10.0f)
		    sprintf(text, "%.1f mp/s", s);
		else if(s >= 1.0f)
		    sprintf(text, "%.2f mp/s", s);
		else if(s >= 0.1f)
		    sprintf(text, "%.3f mp/s", s);
		else
		    sprintf(text, "%.4f mp/s", s);
		break;
	      default:
		s = SWConvertRUPerCycleToRUPerSecond(&units, obj->velocity);
		if(s >= 100.0f)
		    sprintf(text, "%.0f ru/s", s);
		else if(s >= 10.0f)
		    sprintf(text, "%.1f ru/s", s);
		else if(s >= 1.0f)
		    sprintf(text, "%.2f ru/s", s);
		else if(s >= 0.1f)
		    sprintf(text, "%.3f ru/s", s);
		else
		    sprintf(text, "%.4f ru/s", s);
		break;
	    }
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, &swidth, &ascent, NULL
	    );
	    OSWSetFont(font);
	    OSWDrawString(
		pixmap,
		((vc > 0.5) ?
		    (x + 5) : (x + (int)width - swidth - 5)) - lbearing,
		y + ascent,
		text
	    );
	}

	/* Put to window */
	if(put_to_window)
	{
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
	}
}

/*
 *	Player antimatter (just the bar).
 */
static void BWDP_PANTIMATTER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 5, y = 20;
	const unsigned int width = 140, height = 6;
	shared_image_t *img = xsw_bridge_win->pan_p3_img;
	sw_legend_struct *legend = &sw_legend;
	if(img == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    const float c = (obj->antimatter_max > 0.0f) ?
		(obj->antimatter / obj->antimatter_max) : 0.0f;
	    const int	x2 = x + BW_BAR_MARGIN,
			y2 = y;
	    const unsigned int	w2 = (unsigned int)(
		((int)width - (2 * BW_BAR_MARGIN)) * c
	    ),
				h2 = height;
	    pixel_t fg_pix;

	    if((legend->antimatter_low_warning > 0.0f) &&
	       (obj->antimatter_max > legend->antimatter_low_warning)
	    )
	    {
		const float	v_warning = legend->antimatter_low_warning,
				v_danger = v_warning / 2.0f;
		if(obj->antimatter > v_warning)
		    fg_pix = pix;
		else if(obj->antimatter > v_danger)
		    fg_pix = xsw_color.bp_warning;
		else
		    fg_pix = xsw_color.bp_danger;
	    }
	    else
	    {
		if(c > BW_ANTIMATTER_WARNING_COEFF)
		    fg_pix = pix;
		else if(c > BW_ANTIMATTER_DANGER_COEFF)
		    fg_pix = xsw_color.bp_warning;
		else
		    fg_pix = xsw_color.bp_danger;
	    }
	    OSWSetForeground(fg_pix);
	    OSWDrawSolidRectangle(pixmap, x2, y2, w2, h2);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player engine state.
 */
static void BWDP_PENGINESTATE(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 75, y = 92;
	const unsigned int width = 55, height = 30;
	shared_image_t *img_ptr = xsw_bridge_win->pan_p3_img;
	if(img_ptr == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img_ptr, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	if(obj != NULL)
	{
	    int	xc = x, yc = y,
		lbearing, swidth, ascent, descent;
	    const char *s;

	    OSWSetFont(font);
	    OSWSetForeground(xsw_color.bp_normal_outline);

	    s = "Eng: ";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, &swidth, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		s
	    );
	    xc += swidth;

	    switch(obj->engine_state)
	    {
	      case SW_ENGINE_STATE_ON:
		s = "On";
		OSWSetForeground(xsw_color.bp_light_outline);
		break;
	      case SW_ENGINE_STATE_STARTING:
		s = "Init";
		OSWSetForeground(xsw_color.bp_normal_outline);
		break;
	      default:	/* SW_ENGINE_STATE_OFF and SW_ENGINE_STATE_NONE */
		s = "Off";
		OSWSetForeground(xsw_color.bp_dark_outline);
		break;
	    }
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, &swidth, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		s
	    );
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player heading and position.
 */
static void BWDP_PHEADING(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 18, y = 32;
	const unsigned int width = 114, height = 60;
	char text[256];
	shared_image_t *img_ptr = xsw_bridge_win->pan_p3_img;
	if(img_ptr == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img_ptr, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    int xc = x + 5, yc = y, lbearing, ascent, descent;

	    OSWSetFont(font);
	    OSWSetForeground(pix);

	    sprintf(
		text,
		"AM: %.2f",
		obj->antimatter
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text
	    );
	    yc += ascent + descent;

	    sprintf(
		text,
		"HD: %.2f'",
		RADTODEG(obj->heading)
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text
	    );
	    yc += ascent + descent;

	    sprintf(
		text,
		"Sect: %ld %ld %ld",
		obj->sect_x, obj->sect_y, obj->sect_z
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent, 
		text
	    );
	    yc += ascent + descent;

	    sprintf(
		text,
		"X: %.2f",
		obj->x
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent, 
		text
	    );
	    yc += ascent + descent;


	    sprintf(
		text,
		"Y: %.2f", 
		obj->y
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text
	    );
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Player thrust vector.
 */
static void BWDP_PTHRUSTVECTOR(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 40, y = 92;
	const unsigned int width = 30, height = 30;
	xsw_gctl_struct *gctl = xsw_gctl;
	shared_image_t *img_ptr = xsw_bridge_win->pan_p3_img;
	if(img_ptr == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img_ptr, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	if(obj != NULL)
	{
	    int	xc = x + ((int)width / 2) - 1,
		yc = y + ((int)height / 2);

	    float	theta = obj->thrust_dir,

			dx = (float)cos((double)((PI / 2.0f) - theta)) * 7.0f,
 			dy = (float)sin((double)((PI / 2.0f) - theta)) * -7.0f;

	    if((dx < 1.0f) && (dx > -1.0f))
		dx = 0.0f;

	    OSWSetForeground(pix);
	    OSWDrawLine(
		pixmap,
		xc, yc,
		(int)(xc + dx),
		(int)(yc + dy)
	    );

	    /* External dampers on? */
	    if((gctl != NULL) ? gctl->external_dampers : False)
	    {
		OSWDrawArc(
		    pixmap,
		    x,
		    (int)(y - ((int)height * 0.2)),
		    (int)width - 1,
		    (unsigned int)((int)height * 0.4),
		    PI,
		    (2 * PI)
		);
		OSWDrawArc(
		    pixmap,
		    x,
		    (int)(y + (int)height - ((int)height * 0.15)),
		    (int)width - 1,
		    (unsigned int)((int)height * 0.3),
		    PI, (2 * PI)
		);
	    }
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
} 


/*
 *	Subject antimatter (just the bar).
 */
static void BWDP_SANTIMATTER(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window 
)
{
	const int x = 5, y = 20;
	const unsigned int width = 140, height = 6;
	sw_legend_struct *legend = &sw_legend;
	shared_image_t *img = xsw_bridge_win->pan_s2_img;
	if(img == NULL)
	    return;

	/* Draw background */
	OSWPutSharedImageToDrawableSect(
	    img, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw bar */
	if(obj != NULL)
	{
	    const float c = (obj->antimatter_max > 0.0f) ?
		(obj->antimatter / obj->antimatter_max) : 0.0f;
	    const int   x2 = x + BW_BAR_MARGIN,
			y2 = y;
	    const unsigned int  w2 = (unsigned int)(
		((int)width - (2 * BW_BAR_MARGIN)) * c
	    ),
				h2 = height;
	    pixel_t fg_pix;

	    if((legend->antimatter_low_warning > 0.0f) &&
	       (obj->antimatter_max > legend->antimatter_low_warning)
	    )
	    {
		const float     v_warning = legend->antimatter_low_warning,
				v_danger = v_warning / 2.0f;
		if(obj->antimatter > v_warning)
		    fg_pix = pix;
		else if(obj->antimatter > v_danger)
		    fg_pix = xsw_color.bp_warning;
		else
		    fg_pix = xsw_color.bp_danger;
	    }
	    else
	    {
		if(c > BW_ANTIMATTER_WARNING_COEFF)
		    fg_pix = pix;
		else if(c > BW_ANTIMATTER_DANGER_COEFF)
		    fg_pix = xsw_color.bp_warning;
		else
		    fg_pix = xsw_color.bp_danger;
	    }
	    OSWSetForeground(fg_pix);
	    OSWDrawSolidRectangle(pixmap, x2, y2, w2, h2);
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}

/*
 *	Subject heading and location.
 */
static void BWDP_SHEADING(
	int obj_num, xsw_object_struct *obj,
	font_t *font, pixel_t pix,
	pixmap_t pixmap, win_t w,
	int bkg_img_code,
	bool_t put_to_window
)
{
	const int x = 18, y = 32;
	const unsigned int width = 114, height = 60;
	char text[256];
	shared_image_t *img_ptr = xsw_bridge_win->pan_s2_img;
	if(img_ptr == NULL)
	    return;

	/* Draw the background */
	OSWPutSharedImageToDrawableSect(
	    img_ptr, pixmap,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw text */
	if(obj != NULL)
	{
	    int xc = x + 5, yc = y, lbearing, ascent, descent;

	    OSWSetFont(font);
	    OSWSetForeground(pix);

	    sprintf(text, "AM: %.2f",
		obj->antimatter
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text
	    );
	    yc += ascent + descent;

	    sprintf(
		text,
		"HD: %.2f'",
		RADTODEG(obj->heading)
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text
	    );
	    yc += ascent + descent;

	    sprintf(
		text,
		"Sect: %ld %ld %ld",
		obj->sect_x, obj->sect_y, obj->sect_z
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text
	    );
	    yc += ascent + descent;

	    sprintf(
		text,
		"X: %.2f",
		obj->x   
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text   
	    );
	    yc += ascent + descent;

	    sprintf(
		text,
		"Y: %.2f",
		obj->y
	    );
	    OSWGetTextBounds(
		font, text, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    OSWDrawString(
		pixmap,
		xc - lbearing,
		yc + ascent,
		text   
	    );
	}

	/* Put to window */
	if(put_to_window)
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
}


/*
 *	Draws the bridge wnidow's console panel.
 *
 *	The obj_num specifies the object who's stats are to used when
 *	drawing the panel.
 *
 *	The panel_num specifies which panel to draw, it must be one
 *	of BW_PANEL_*.
 */
void BridgeWinDrawPanel(
	const int obj_num,
	const int panel_num		/* One of BW_PANEL_* */
)
{
	font_t *font;
	pixel_t pix;
	pixmap_t pixmap;
	win_t w;
	xsw_object_struct *obj;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	if(bw == NULL)
	    return;

	/* Get pointer to the specified object (which may be garbage) */
	obj = DBIsObjectGarbage(obj_num) ? NULL : xsw_object[obj_num];

	font = xsw_font.con_standard;
	pix = widget_global.editable_text_pix;
	pixmap = bw->pan_buf;

	switch(panel_num)
	{
	  /* ******************************************************* */
	  case BW_PANEL_P1:
	    w = bw->pan_p1;
	    BWDP_BACKGROUND(
		obj_num,
		obj,
		pixmap,
		panel_num
	    );
	    BWDP_PNAME(
		obj_num, obj,
		xsw_font.con_heading, xsw_color.bp_bold_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		False
	    );
	    BWDP_PSHIELDFREQ(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		False
	    );
	    BWDP_PWEAPONFREQ(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		False
	    );
	    BWDP_PCOMCHANNEL(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		False
	    );
	    BWDP_PSCORE(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		False
	    );
	    OSWPutDrawableToWindow(w, pixmap);
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_P2:
	    w = bw->pan_p2;
	    BWDP_BACKGROUND(
		obj_num,
		obj,
		pixmap,
		panel_num
	    );
	    BWDP_PHULL(
		obj_num, obj,
		font, xsw_color.bpol_hull,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		False
	    );
	    BWDP_PPOWER(
		obj_num, obj,
		font, xsw_color.bpol_power,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		False
	    );
	    BWDP_PSHIELDS(
		obj_num, obj,
		font, xsw_color.bpol_shields,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		False
	    );
	    BWDP_PVIS(
		obj_num, obj,
		font, xsw_color.bpol_vis,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		False
	    );
	    BWDP_PDMGCTL(
		obj_num, obj,
		font, xsw_color.bpol_dmgctl,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		False
	    );
	    OSWPutDrawableToWindow(w, pixmap);
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_P3:
	    w = bw->pan_p3;
	    BWDP_BACKGROUND(
		obj_num,
		obj,
		pixmap,
		panel_num
	    );
	    BWDP_PTHROTTLE(
		obj_num, obj,
		font, xsw_color.bpol_throttle,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		False
	    );
	    BWDP_PVELOCITY(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		False
	    );
	    BWDP_PANTIMATTER(
		obj_num, obj,
		font, xsw_color.bpol_power,	/* Use power color for now */
		pixmap, w,
		IMG_CODE_STATS_CON3,
		False
	    );
	    BWDP_PENGINESTATE(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		False
	    );
	    BWDP_PHEADING(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		False   
	    );
	    BWDP_PTHRUSTVECTOR(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		False
	    );
	    OSWPutDrawableToWindow(w, pixmap); 
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_P4:
	    w = bw->pan_p4;
	    BWDP_BACKGROUND(
		obj_num,
		obj,
		pixmap,
		panel_num
	    );
	    BWDP_PINAME(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON4,
		False
	    );
	    BWDP_PWLOCK(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON4,
		False
	    );
	    BWDP_PWEAPONS(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON4,
		False
	    );
	    OSWPutDrawableToWindow(w, pixmap);
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_S1:
	    w = bw->pan_s1;
	    BWDP_BACKGROUND(
		obj_num,
		obj,
		pixmap,
		panel_num
	    );
	    BWDP_SNAME(
		obj_num, obj,
		xsw_font.con_heading, xsw_color.bp_bold_text,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		False
	    );
	    BWDP_SHULL(
		obj_num, obj,
		font, xsw_color.bpol_hull,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		False
	    );
	    BWDP_SPOWER(
		obj_num, obj,
		font, xsw_color.bpol_power,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		False
	    );
	    BWDP_SVIS(
		obj_num, obj,
		font, xsw_color.bpol_vis,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		False
	    );
	    BWDP_SSHIELDS(
		obj_num, obj,
		font, xsw_color.bpol_shields,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		False
	    );
	    BWDP_SDISTANCE_BAR(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		False
	    );
#if 0
	    BWDP_SDMGCTL(
		obj_num, obj,
		font, xsw_color.bpol_dmgctl,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		False
	    );
#endif
	    OSWPutDrawableToWindow(w, pixmap);
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_S2:
	    w = bw->pan_s2;
	    BWDP_BACKGROUND(
		obj_num,
		obj,
		pixmap,
		panel_num
	    );
	    BWDP_SBEARING(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_SRO_CON2,
		False
	    );
	    BWDP_SEMPIRE(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_SRO_CON2,
		False
	    );
	    BWDP_SDISTANCE(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_SRO_CON2,
		False
	    );
	    BWDP_SANTIMATTER(
		obj_num, obj,
		font, xsw_color.bpol_power,	/* Use power color for now */
		pixmap, w,
		IMG_CODE_SRO_CON2,
		False
	    );
	    BWDP_SHEADING(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_SRO_CON2,
		False
	    );
	    OSWPutDrawableToWindow(w, pixmap);
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_S3:
	    w = bw->pan_s3;
	    BWDP_BACKGROUND(
		obj_num,
		obj,
		pixmap,
		panel_num
	    );
	    BWDP_SWEAPONS(
		obj_num, obj,
		font,
		0,		/* Pixel ignored */
		pixmap, w,
		IMG_CODE_SRO_CON3,
		False
	    );
	    OSWPutDrawableToWindow(w, pixmap);
	    break;


	  /* ******************************************************* */
	  case BW_PANEL_PNAME:
	    w = bw->pan_p1;
	    BWDP_PNAME(
		obj_num, obj,
		xsw_font.con_heading, xsw_color.bp_bold_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PSHIELDFREQ:
	    w = bw->pan_p1;
	    BWDP_PSHIELDFREQ(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		True
	    );
	    break;

	  case BW_PANEL_PWEAPONFREQ:
	    w = bw->pan_p1;
	    BWDP_PWEAPONFREQ(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		True
	    );
	    break;
	  
	  case BW_PANEL_PCOMCHANNEL:
	    w = bw->pan_p1;
	    BWDP_PCOMCHANNEL(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		True
	    );
	    break;

	  case BW_PANEL_PSCORE:
	    w = bw->pan_p1;
	    BWDP_PSCORE(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON1,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PHULL:
	    w = bw->pan_p2;
	    BWDP_PHULL(
		obj_num, obj,
		font, xsw_color.bpol_hull,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PPOWER:
	    w = bw->pan_p2;
	    BWDP_PPOWER(
		obj_num, obj,
		font, xsw_color.bpol_power,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PSHIELDS:
	    w = bw->pan_p2;
	    BWDP_PSHIELDS(
		obj_num, obj,
		font, xsw_color.bpol_shields,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PVIS:
	    w = bw->pan_p2;
	    BWDP_PVIS(
		obj_num, obj,
		font, xsw_color.bpol_vis,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PDMGCTL:
	    w = bw->pan_p2;
	    BWDP_PDMGCTL(
		obj_num, obj,
		font, xsw_color.bpol_dmgctl,
		pixmap, w,
		IMG_CODE_STATS_CON2,
		True
	    );
	    break;


	  /* ******************************************************* */
	  case BW_PANEL_SNAME:
	    w = bw->pan_s1;
	    BWDP_SNAME(
		obj_num, obj,
		xsw_font.con_heading, xsw_color.bp_bold_text,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SHULL:
	    w = bw->pan_s1;
	    BWDP_SHULL(
		obj_num, obj,
		font, xsw_color.bpol_hull,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SPOWER:
	    w = bw->pan_s1;
	    BWDP_SPOWER(
		obj_num, obj,
		font, xsw_color.bpol_power,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SSHIELDS:
	    w = bw->pan_s1;
	    BWDP_SSHIELDS(
		obj_num, obj,
		font, xsw_color.bpol_shields,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SVIS:
	    w = bw->pan_s1;
	    BWDP_SVIS(
		obj_num, obj,
		font, xsw_color.bpol_vis,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SBEARING:
	    w = bw->pan_s2;
	    BWDP_SBEARING(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_SRO_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SDISTANCE:
	    w = bw->pan_s1;
	    BWDP_SDISTANCE_BAR(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		True
	    );
	    w = bw->pan_s2;
	    BWDP_SDISTANCE(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_SRO_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SDMGCTL:
	    w = bw->pan_s1;
#if 0
	    BWDP_SDMGCTL(
		obj_num, obj,
		font, xsw_color.bpol_dmgctl,
		pixmap, w,
		IMG_CODE_SRO_CON1,
		True
	    );
#endif
	    break;


	  /* ******************************************************* */
	  case BW_PANEL_PINAME:
	    w = bw->pan_p4;
	    BWDP_PINAME(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON4,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PWLOCK:
	    w = bw->pan_p4;
	    BWDP_PWLOCK(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON4,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PWEAPONS:
	    w = bw->pan_p4;
	    BWDP_PWEAPONS(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON4,
		True
	    );
	    break;


	  /* ******************************************************* */
	  case BW_PANEL_SINAME:
	    w = bw->pan_s3;

	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SWLOCK:
	    w = bw->pan_s3;

	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SEMPIRE:
	    w = bw->pan_s2;
	    BWDP_SEMPIRE(
		obj_num, obj,
		font,
		0,		/* Pixel ignored */
		pixmap, w,
		IMG_CODE_SRO_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SWEAPONS:
	    w = bw->pan_s3;
	    BWDP_SWEAPONS(
		obj_num, obj,
		font,
		0,		/* Pixel ignored */
		pixmap, w,
		IMG_CODE_SRO_CON3,
		True
	    );
	    break;


	  /* ******************************************************* */ 
	  case BW_PANEL_PTHROTTLE:
	    w = bw->pan_p3;
	    BWDP_PTHROTTLE(
		obj_num, obj,
		font, xsw_color.bpol_throttle,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PVELOCITY:
	    w = bw->pan_p3;
	    BWDP_PVELOCITY(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PANTIMATTER:
	    w = bw->pan_p3;
	    BWDP_PANTIMATTER(
		obj_num, obj,
		font, xsw_color.bpol_power,	/* Use power color for now */
		pixmap, w,
		IMG_CODE_STATS_CON3,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PENGINESTATE:
	    w = bw->pan_p3;
	    BWDP_PENGINESTATE(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PHEADING:
	    w = bw->pan_p3;
	    BWDP_PHEADING(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_PTHRUSTVECTOR:
	    w = bw->pan_p3;
	    BWDP_PTHRUSTVECTOR(
		obj_num, obj,
		font, xsw_color.bp_light_outline,
		pixmap, w,
		IMG_CODE_STATS_CON3,
		True
	    );
	    break;


	  /* ******************************************************* */
	  case BW_PANEL_SANTIMATTER:
	    w = bw->pan_s2;
	    BWDP_SANTIMATTER(
		obj_num, obj,
		font, xsw_color.bpol_power,	/* Use power color for now */
		pixmap, w,
		IMG_CODE_SRO_CON2,
		True
	    );
	    break;

	  /* ******************************************************* */
	  case BW_PANEL_SHEADING:
	    w = bw->pan_s2;
	    BWDP_SHEADING(
		obj_num, obj,
		font, xsw_color.bp_standard_text,
		pixmap, w,
		IMG_CODE_SRO_CON2,
		True
	    );
	    break;
	}
}


/*
 *	Redraws the entire bridge window and the viewscreen.
 */
void BridgeWinDrawAll(xsw_bridge_win_struct *bw)
{
	xsw_net_parms_struct *net = &net_parms;
	const int player_obj_num = net->player_obj_num;
	const time_t	cur_ms = cur_millitime,
			cur_sec = cur_systime;
	const float tc = time_compensation;

	if(bw == NULL)
	    return;

	/* Selected weapon stats label on viewscreen */
	VSDrawUpdateWeaponLabel(
	    &bw->vs_weapon_image,
	    bw->vs_weapon_buf
	);

	/* Network stats labe on viewscreenl */
	VSDrawUpdateNetstatsLabel(
	    &bw->net_stats_image,
	    bw->net_stats_buf
	);

	/* Draw viewscreen */
	if(bw->viewscreen_vis_state != VisibilityFullyObscured)
	    VSDrawViewScreen(bw);

	/* Scanner labels */
	ScannerUpdateLabels(bw, player_obj_num);

	/* Scanner */
	if(bw->scanner_vis_state != VisibilityFullyObscured)
	    ScannerDraw(
		bw, player_obj_num,
		bw->scanner,
		bw->scanner_image,
		0,
		0,
		bw->scanner_zoom,
		cur_ms, cur_sec, tc
	    );

	/* Player stats console panels */
	if(bw->map_state)
	{
	    int tar_obj_num;
	    xsw_object_struct *player_obj = net->player_obj_ptr;

	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_P1);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_P2);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_P3);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_P4);

	    /* Get the subject object if valid */
	    tar_obj_num = (player_obj != NULL) ?
		player_obj->locked_on : -1;

	    /* Subject stats console panels */
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_S1);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_S2);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_S3);
	}
}

/*
 *	Redraws the small messages window.
 */
void BridgeDrawMessages(xsw_bridge_win_struct *bw)
{
	int	i, x, y, sb_x, sb_y,
		c_pos, line_spacing, lines_drawn, lines_visible,
		lbearing, cwidth, ascent, descent, font_height;
	unsigned int width, height;
	font_t *font;
	pixmap_t pixmap;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	scroll_bar_struct *sb;
	xsw_mesg_struct *msg;
	widget_global_struct *wglobal = &widget_global;

	if(bw == NULL)
	    return;

	line_spacing = MAX(bw->mesg_box_line_spacing, 1);
	cwidth = MAX(bw->mesg_box_char_width, 1);
	font = bw->mesg_box_font;
	w = bw->mesg_box;
	pixmap = bw->mesg_box_buf;
	width = bw->mesg_box_width;
	height = bw->mesg_box_height;
	lines_visible = FTOI_RNDOFF((float)height / (float)line_spacing) + 1;
	sb = &bw->mesg_box_sb;
	sb_x = (int)SBarGetPositionX(sb);
	sb_y = (int)SBarGetPositionY(sb);

	OSWSetFont(font);

	/* Redraw background */
	if(ImgIsAllocated(IMG_CODE_MESG_CON))
	    WidgetPutImageTile(
		pixmap, xsw_image[IMG_CODE_MESG_CON]->image,
		width, height
	    );
	else
	    OSWClearPixmap(
		pixmap, width, height,
		gui->black_pix
	    );

	/* Calculate the starting line and starting line coordinates */
	i = MAX((sb_y / line_spacing), 0);
	x = 10 - sb_x;
	y = 10 - (sb_y % line_spacing);
	if(i > 0)
	{
	    i--;
	    y -= line_spacing;
	}

	/* Draw the messages */
	lines_drawn = 0;
	while(lines_drawn < lines_visible)
	{
	    if(i >= total_messages)
		break;

	    msg = &mesg_list[i];
	    if(msg == NULL)
	    {
		i++;
		continue;
	    }

	    /* Is this message not selected? */
	    if((msg->sel_start < 0) || (msg->sel_end < 0))
	    {
		/* Draw unselected message */
		int x2, y2;
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->white_pix :
			msg->fg_pix
		);
		OSWGetTextBounds(
		    font, msg->message, -1,
		    &lbearing, NULL, NULL,
		    &ascent, &descent
		);
		font_height = ascent + descent;
		x2 = x;
		y2 = y + FTOI_RNDOFF(
		    (float)(line_spacing - font_height) / 2.0f
		);
		OSWDrawString(
		    pixmap,
		    x2 - lbearing,
		    y2 + ascent,
		    msg->message
		);
	    }
	    else
	    {
		/* Draw selected text background */
		int x2, y2;
		unsigned int w2, h2;

		OSWGetCharBounds(
		    font, msg->message[0],
		    &lbearing, NULL, NULL,
		    &ascent, &descent
		);
		font_height = ascent + descent;
		x2 = x + (msg->sel_start * cwidth);
		y2 = y;
		w2 = (unsigned int)MAX(
		    (msg->sel_end - msg->sel_start + 1) * cwidth,
		    1
		);
		h2 = line_spacing;

		OSWSetForeground(
		    wglobal->force_mono ?
			gui->white_pix :
			wglobal->surface_selected_pix
		);
		OSWDrawSolidRectangle(pixmap, x2, y2, w2, h2);

		/* Draw first unselected text segment */
		x2 = x;
		y2 = y + FTOI_RNDOFF(
		    (float)(line_spacing - font_height) / 2.0f
		);
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->white_pix :
			msg->fg_pix
		);
		OSWDrawStringLimited(
		    pixmap,
		    x2 - lbearing,
		    y2 + ascent,
		    msg->message,
		    msg->sel_start
		);
		/* Draw the selected text segment */
		c_pos = msg->sel_start;
		if((c_pos >= 0) && (c_pos < MSG_LIST_MSG_LEN))
		{
		    x2 = x + (c_pos * cwidth);
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->black_pix :
			    wglobal->selected_text_pix
		    );
		    OSWDrawStringLimited(
			pixmap,
			x2 - lbearing,
			y2 + ascent,
			msg->message + c_pos,
			msg->sel_end - c_pos + 1
		    );
		}
		/* Draw last unselected text segment */
		c_pos = msg->sel_end + 1;
		if((c_pos >= 0) && (c_pos < MSG_LIST_MSG_LEN))
		{
		    x2 = x + (c_pos * cwidth);
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    msg->fg_pix
		    );
		    OSWDrawStringLimited(
			pixmap,
			x2 - lbearing,
			y2 + ascent,
			msg->message + c_pos,
			MSG_LIST_MSG_LEN - msg->sel_end
		    );
		}
	    }

	    y += line_spacing;
	    i++;
	    lines_drawn++;
	}

	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the bridge window to draw messages.
 */
void BridgeQueueDrawMessages(xsw_bridge_win_struct *bw)
{
	if(bw == NULL)
	    return;

	WidgetQueueDrawAction(BridgeWinDrawMessagesCB, NULL, bw);
}
