#include <stdlib.h>
#include <string.h>

#include "../include/string.h"

#include "keymaps.h"
#include "xsw.h"


int CmdHandleInput(const char *cmd);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Handles a client name specified by cmd.
 */
int CmdHandleInput(const char *cmd)
{
	int status = 0;
	const char *s, *s_end;
	char *name, *arg;

	if(cmd == NULL)
	    return(-2);

	while(*cmd == ' ')
	    cmd++;

	if(*cmd == '\0')
	    return(-2);

	/* Parse name */
	s = cmd;
	s_end = strchr(s, ' ');
	if(s_end != NULL)
	{
	    const int len = s_end - s;
	    name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(name, s, len * sizeof(char));
	    name[len] = '\0';

	    s = s_end;
	    while(ISBLANK(*s))
		s++;
	}
	else
	{
	    name = STRDUP(s);
	    s = "";
	}
	if(name == NULL)
	    name = strdup("");
	strstrip(name);

	/* Parse argument */
	arg = STRDUP(s);
	if(arg == NULL)
	    arg = strdup("");
	strstrip(arg);

 
	/* Handle name by name */

	/* Debug */
	if(!strcasecmp(name, "debug"))
	{
	    status = CmdDebug(arg);
	}
	/* Test */
	else if(!strcasecmp(name, "test"))
	{
	    status = CmdTest(arg);
	}
	/* Help */
	else if(!strcasecmp(name, "help") ||
		!strcasecmp(name, "hel") ||
		!strcasecmp(name, "he") ||
		!strcasecmp(name, "h") ||
		!strcasecmp(name, "?")
	)
	{
	    status = CmdHelp(arg);
	}
	/* Version */
	else if(!strcasecmp(name, "version") ||
		!strcasecmp(name, "ver")
	)
	{
	    status = CmdVersion(arg);
	}
	/* Memory */
	else if(!strcasecmp(name, "memory") ||
		!strcasecmp(name, "mem")
	)
	{
	    status = CmdMemory(arg);
	}
	/* Automatic Interval */  
	else if(!strcasecmp(name, "autointerval") ||
		!strcasecmp(name, "autoint") ||
		!strcasecmp(name, "autoi") ||
		!strcasecmp(name, "aint")
	)
	{
	    status = CmdAutoInterval(arg);
	}
	/* Server Script */
	else if(!strcasecmp(name, "script") ||
		!strcasecmp(name, "serverscript") ||
		!strcasecmp(name, "servscript") ||
		!strcasecmp(name, "scr")
	)
	{
	    status = CmdServScript(arg);
	}
	/* Set */
	else if(!strcasecmp(name, "set"))
	{
	    status = CmdSet(arg);
	}
	/* Login Name */
	else if(!strcasecmp(name, "login_name") ||
		!strcasecmp(name, "loginname") ||
		!strcasecmp(name, "lname")
	)
	{
	    status = CmdLoginName(arg); 
	}
	/* Login Password */
	else if(!strcasecmp(name, "login_password") ||
		!strcasecmp(name, "login_pass") ||
		!strcasecmp(name, "loginpassword") ||
		!strcasecmp(name, "loginpass") ||
		!strcasecmp(name, "lpassword") ||
		!strcasecmp(name, "lpasswd") ||
		!strcasecmp(name, "lpass")
	)
	{
	    status = CmdLoginPassword(arg);
	}
	/* Log */
	else if(!strcasecmp(name, "log"))
	{
	    status = CmdLog(arg);
	}
	/* Connect */
	else if(!strcasecmp(name, "connect") ||
		!strcasecmp(name, "conn") ||
		!strcasecmp(name, "con")
	)
	{
	    status = CmdConnect(arg);
	}
	/* Disconnect */
	else if(!strcasecmp(name, "disconnect") ||
		!strcasecmp(name, "discon") ||
		!strcasecmp(name, "bye")
	)
	{
	    status = CmdDisconnect(arg);
	}
	/* Refresh */
	else if(!strcasecmp(name, "refresh") ||
		!strcasecmp(name, "ref")
	)
	{
	    status = CmdRefresh(arg);
	}
	/* Interval */
	else if(!strcasecmp(name, "interval") ||
		!strcasecmp(name, "int")
	)
	{
	    status = CmdNetInterval(arg);
	}
	/* Quit/Exit */
	else if(!strcasecmp(name, "quit") ||
		!strcasecmp(name, "exit")
	)
	{
	    /* No longer in use, see the key mapping which activates the
	     * mapping of the exit prompt
	     */
	    const keycode_t keycode = KeymapGetKeycode(XSW_KM_EXIT);
	    const char *arg = OSWGetKeyCodeName(keycode);
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"This command is depreciated, to exit press %s.",
		arg
	    );
	    XSWPrintWarning(s);
	    free(s);
	}
	/* Synctime */
	else if(!strcasecmp(name, "synctime") ||
		!strcasecmp(name, "synct") ||
		!strcasecmp(name, "sync") ||
		!strcasecmp(name, "syn")
	)
	{
	    status = CmdSynctime(arg);
	}
	else 
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such command or macro.",
		name
	    );
	    XSWPrintWarning(s);
	    free(s);
	}

	free(name);
	free(arg);

	return(status);
}
