#include <stdlib.h>
#include <string.h>

#include "../include/string.h"

#include "xsw.h"


int CmdHelp(const char *arg);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdHelp(const char *arg)
{
	/* Print commands list */
	if((arg != NULL) ? (*arg == '\0') : True)
	{
	    XSWPrint(
"aint       con  discon   help    interval  loginname  log"
	    );
	    XSWPrint(
"loginpass  mem  refresh  script  set       synctime   version"
	    );
	    XSWPrint(
"For additional help on a perticular command type `help [command]'"
	    );
	    return(0);
	}

	/* Check if argument is requesting for topic-specific help */

	/* Automatic Interval */
	if(!strcasecmp(arg, "autointerval") ||
	   !strcasecmp(arg, "autoint") ||
	   !strcasecmp(arg, "autoi") ||
	   !strcasecmp(arg, "aint")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [on|off]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Sets the automatic network interval tunning on or off."
	    );
	}
	/* Connect */
	else if(!strcasecmp(arg, "connect") ||
		!strcasecmp(arg, "conn") ||
		!strcasecmp(arg, "con")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [url]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Connects to universe specified in the [url]. If no argument is\
 specified then the universe list window will be mapped."
	    );
	}
	/* Disconnect */
	else if(!strcasecmp(arg, "disconnect") ||
		!strcasecmp(arg, "discon") ||
		!strcasecmp(arg, "bye")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Disconnects from current connected universe."
	    );
	}
	/* Help */
	else if(!strcasecmp(arg, "help") ||
		!strcasecmp(arg, "hel") ||
		!strcasecmp(arg, "he") ||
		!strcasecmp(arg, "h") ||
		!strcasecmp(arg, "?")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [command]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Prints detailed help about [command]. If [command] is not specified\
 then a list of commands is printed."
	    );
	}
	/* Interval */
	else if(!strcasecmp(arg, "interval") ||
		!strcasecmp(arg, "int")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [ms]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Sets the network update interval in milliseconds. This command is\
 only available when automatic interval is set to off."
	    );
	}
	/* Log */
	else if(!strcasecmp(arg, "log"))
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [on|off] [filename]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Turns logging on or off. If [filename] is specified when turning\
 logging on then the log will be written to the file specified\
 by [filename]."
	    );
	}
	/* Login Name */
	else if(!strcasecmp(arg, "login_name") ||
		!strcasecmp(arg, "loginname") ||
		!strcasecmp(arg, "lname")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [login_name]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Sets the default login name to be used whenever a URL does not\
 specify a login name."
	    );
	}
	/* Login Password */
	else if(!strcasecmp(arg, "login_password") ||
		!strcasecmp(arg, "login_pass") ||
		!strcasecmp(arg, "loginpassword") ||
		!strcasecmp(arg, "loginpass") ||
		!strcasecmp(arg, "lpassword") ||
		!strcasecmp(arg, "lpasswd") ||
		!strcasecmp(arg, "lpass")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [login_password]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Sets the default login password to be used whenever a URL does not\
 specify a login password."
	    );
	}
	/* Memory */
	else if(!strcasecmp(arg, "memory") ||
		!strcasecmp(arg, "mem")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [operation]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Prints the amount of memory used by this program. If [operation] is\
 specified then that memory operation will be performed. Available\
 values for [operation] are; refresh, reload."
	    );
	}
	/* Refresh */
	else if(!strcasecmp(arg, "refresh") ||
		!strcasecmp(arg, "ref")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Refreshes all information received from the server (must be\
 connected to use this command)."
	    );
	}
	/* Server Script */
	else if(!strcasecmp(arg, "script") ||
		!strcasecmp(arg, "serverscript") ||
		!strcasecmp(arg, "servscript") ||
		!strcasecmp(arg, "scr")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [path] [arg...]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Uploads a script file specified by [path] to the server. If no\
 arguments are specified then you will be prompted for a script\
 file."
	    );
	}
	/* Set */
	else if(!strcasecmp(arg, "set"))
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s [parameter][=<value>]",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Sets [parameter] to <value>. If no arguments are specified then a\
 list of all available parameters and their current values will be\
 printed."
	    );
	}
	/* Synctime */
	else if(!strcasecmp(arg, "synctime") ||
		!strcasecmp(arg, "synct") ||
		!strcasecmp(arg, "sync") ||
		!strcasecmp(arg, "syn")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Resets and syncronizes timmers. This action is usually performed\
 automatically by the program as needed."
	    );
	}
	/* Version */
	else if(!strcasecmp(arg, "version") ||
		!strcasecmp(arg, "ver")
	)
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"%s",
		arg
	    );
	    XSWPrint(s);
	    free(s);
	    XSWPrint(
"Prints the version of this program."
	    );
	}
	else
	{
	    char *s = (char *)malloc(STRLEN(arg) + 80);
	    sprintf(
		s,
"No help available for topic `%s'",
		arg
	    );
	    XSWPrintWarning(s);
	    free(s);
	}


	return(0);
}
