#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "xsw.h"


int CmdLog(const char *arg);


#define THIS_CMD_NAME			"log"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Turns logging on/off and sets log file name.
 */
int CmdLog(const char *arg)
{
	const char *s;
	bool_t state;
	char *path;
	xsw_option_struct *opt = &option;

	/* Print usage? */
	if((arg != NULL) ? (*arg == '\0') : True)
	{
	    char *s;

	    if(opt->log_client ||
	       opt->log_net ||
	       opt->log_errors
	    )
	    {
		struct stat stat_buf;

		if(!STRISEMPTY(fname.log) ?
		    !stat(fname.log, &stat_buf) : 0
		)
		{
		    s = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + STRLEN(fname.log) +
			40 + 80
		    );
		    sprintf(
			s,
"%s: on  log file: %s  log size:  %ld byte%s",
			THIS_CMD_NAME, fname.log,
			stat_buf.st_size,
			(stat_buf.st_size == 0l) ? "" : "s"
		    );
		}
		else
		{
		    s = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + STRLEN(fname.log) +
			80
		    );
		    sprintf(
			s,
"%s: on  log file: %s  log size:  0 bytes",
			THIS_CMD_NAME, fname.log
		    );
		}
	    }
	    else
	    {
		s = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 80
		);
		sprintf(
		    s,
"%s: off",
		    THIS_CMD_NAME
		);
	    }
	    XSWPrint(s);
	    free(s);

	    return(0);
	}

	/* Get the state */
	s = arg;
	while(ISBLANK(*s))
	    s++;
	state = strcasepfx(s, "on") ? True : False;

	/* Get the path */
	while(!ISBLANK(*s) && (*s != '\0'))
	    s++;
	while(ISBLANK(*s))
	    s++;
	if(*s != '\0')
	{
	    char *s2;

	    path = strdup(s);
	    s2 = STRDUP(PathSubHome(path));
	    free(path);
	    path = s2;
	    if(!ISPATHABSOLUTE(path))
	    {
		s2 = STRDUP(PrefixPaths(dname.home, path));
		free(path);
		path = s2;
	    }
	}
	else
	{
	    path = NULL;
	}

	/* Turn logging on? */
	if(state)
	{
	    bool_t log_file_exists;
	    char *s;
	    struct stat stat_buf;

	    /* Check if the log file exists and if it is then get
	     * its stats
	     */
	    if(STRISEMPTY(path) ?
		False : !stat(path, &stat_buf)
            )
	    {
		log_file_exists = True;
	    }
	    else
	    {
		log_file_exists = False;
	    }

	    /* Turn logging on */
	    opt->log_client = True;
	    opt->log_net = True;
	    opt->log_errors = True;

	    /* Set new log file? */
	    if(!STRISEMPTY(path))
	    {
		/* Set new log file */
		strncpy(fname.log, path, sizeof(fname.log));
		fname.log[sizeof(fname.log) - 1] = '\0';
	    }

	    /* Print current log state and path */
	    if(log_file_exists)
	    {
		s = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(fname.log) +
		    40 + 80
		);
		sprintf(
		    s,
"%s: on  log file: %s  log size:  %ld byte%s",
		    THIS_CMD_NAME, fname.log,
		    stat_buf.st_size,
		    (stat_buf.st_size == 0l) ? "" : "s"
		);
		XSWPrint(s);
		free(s);
#ifdef S_ISDIR
		if(S_ISDIR(stat_buf.st_mode))
		{
		    s = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + STRLEN(fname.log) +
			80
		    );
		    sprintf(
			s,
"%s: %s: Is a directory.",
			THIS_CMD_NAME, fname.log
		    );
		    XSWPrintWarning(s);
		    free(s);
		}
#endif
	    }
	    else
	    {
		s = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(fname.log) +
		    80
		);
		sprintf(
		    s,
"%s: on  log file: %s  log size:  0 bytes",
		    THIS_CMD_NAME, fname.log
		);
		XSWPrint(s);
		free(s);
	    }
	}
	else
	{
	    char *s;

	    /* Turn logging off */
	    opt->log_client = False;
	    opt->log_net = False;
	    opt->log_errors = False;

	    /* Print current log state */
	    s = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 80
	    );
	    sprintf(
		s,
"%s: off",
		THIS_CMD_NAME
	    );
	    XSWPrint(s);
	    free(s);
	}

	free(path);

	return(0);
}
