#include <stdlib.h>
#include <string.h>

#include "sndrefs.h"
#include "xsw.h"
#include "ocsnames.h"	/* Must be after xsw.h for types to be defined */


int CmdMemory(const char *arg);


#define THIS_CMD_NAME	"memory"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

 
/*
 *      Print memory stats, reload configuration, or refresh and reclaim
 *	memory.
 */
int CmdMemory(const char *arg)
{
	int events;
	char *s;
	xsw_mem_stat_struct buf;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	if(bw == NULL)
	    return(-1);

	/* Get memory stats */
	XSWGetMemUsage(&buf);

	/* Calculate memory consumed events */
	events = OSWEventsPending();
 
	/* Print memory statistics? */
	if(STRISEMPTY(arg))
	{
	    s = (char *)malloc(80);
	    sprintf(
		s,
"         total   intimgs    imgrefs     objs     ocsns"
	    );
	    XSWPrint(s);
	    free(s);

	    s = (char *)malloc(
		40 + 40 + 40 + 40 + 40 + 80
	    );
	    sprintf(
		s,
 "Mem:\
 %9ld\
 %9ld\
 %9ld\
 %9ld\
 %9ld",
		buf.total,
		buf.images,
		buf.imgrefs,
		buf.objects,
		buf.ocsns
	    );
	    XSWPrint(s);
	    free(s);

	    s = (char *)malloc(    
		40 + 40 + 40 + 40 + 40 + 80  
	    );    
	    sprintf(
		s,
 "Ent:\
 %9d\
 %9d\
 %9d\
 %9d\
 %9d",
		(int)(total_objects + total_imgrefs + total_ocsns),
		total_images,
		total_imgrefs,
		total_objects,
		total_ocsns
	    );
	    XSWPrint(s);
	    free(s);

	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
		"Usage: `%s [reload|refresh]'",
		THIS_CMD_NAME
	    );
	    XSWPrint(s);
	    free(s);
	}
	/* Reload */
	else if(!strcasecmp(arg, "reload"))
	{
	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Reloading...",
		THIS_CMD_NAME
	    );      
	    XSWPrint(s);
	    free(s);

	    s = (char *)malloc(
		STRLEN(THIS_CMD_NAME) +
		STRLEN(fname.main_page) + 80
	    );
	    sprintf(
		s,
"%s: Reloading `%s'",
		THIS_CMD_NAME, fname.main_page
	    );
	    XSWPrint(s);
	    free(s);
	    PageFileOpen(
		&bw->main_page,
		bw->viewscreen, bw->viewscreen_image,
		fname.main_page
	    );

	    s = (char *)malloc(     
		STRLEN(THIS_CMD_NAME) +
		STRLEN(fname.destroyed_page) + 80
	    );
	    sprintf(
		s,
"%s: Reloading `%s'",
		THIS_CMD_NAME, fname.destroyed_page
	    );
	    XSWPrint(s);
	    free(s);
	    PageFileOpen(
		&bw->destroyed_page,
		bw->viewscreen, bw->viewscreen_image,
		fname.destroyed_page
	    );

	    s = (char *)malloc(     
		STRLEN(THIS_CMD_NAME) +
		STRLEN(fname.ocsn) + 80
	    );
	    sprintf(
		s,
"%s: Reloading `%s'",
		THIS_CMD_NAME, fname.ocsn
	    );
	    XSWPrint(s);
	    free(s);
	    XSWOpenOCSNs(fname.ocsn);

	    s = (char *)malloc(
		STRLEN(THIS_CMD_NAME) +
		STRLEN(fname.imgrefs) + 80
	    );
	    sprintf(
		s,
"%s: Reloading `%s'",
		THIS_CMD_NAME, fname.imgrefs
	    );
	    XSWPrint(s);
	    free(s);
	    XSWOpenImgRefs(fname.imgrefs);

	    s = (char *)malloc(     
		STRLEN(THIS_CMD_NAME) +
		STRLEN(fname.sndrefs) + 80
	    );
	    sprintf(
		s,
"%s: Reloading `%s'",
		THIS_CMD_NAME, fname.sndrefs
	    );
	    XSWPrint(s);
	    free(s);
	    XSWOpenSndRefs(fname.sndrefs);

	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Reload done.",
		THIS_CMD_NAME
	    );
	    XSWPrint(s);
	    free(s);
	}
	/* Refresh */
	else if(!strcasecmp(arg, "refresh"))
	{
	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Refreshing...",
		THIS_CMD_NAME
	    );
	    XSWPrint(s);
	    free(s);

	    XSWReclaimMem(False, cur_millitime, cur_systime, time_compensation);

	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Refresh done.",
		THIS_CMD_NAME
	    );
	    XSWPrint(s);
	    free(s);
	}


	/* Recreate the selected weapon viewscreen label */
	VSDrawUpdateWeaponLabel(
	    &bw->vs_weapon_image,
	    bw->vs_weapon_buf
	);
 

	return(0);
}
