#include <stdlib.h>
#include <string.h>
#include "net.h"
#include "xsw.h"


int CmdNetInterval(const char *arg);


#define THIS_CMD_NAME	"interval"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdNetInterval(const char *arg)
{
	char *s;
	xsw_net_parms_struct *net = &net_parms;

	/* Not connected? */
	if(!XSW_CON_STATE_IS_CONNECTED(net))
	{
	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Not connected.",
		THIS_CMD_NAME
	    );
	    XSWPrintWarning(s);
	    free(s);
	    return(-1);
	}

	/* If no argument specified then print current interval */
	if((arg != NULL) ? (*arg == '\0') : True)
	{
	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
		"%s: %i ms",
		THIS_CMD_NAME, (int)net->net_int
	    );
	    XSWPrint(s);
	    free(s);
	    return(0);
	}

	/* Set new interval */
	net->net_int = atol(arg);
	if(net->net_int > XSW_SWSERV_MAX_UPDATE_INT)
	    net->net_int = XSW_SWSERV_MAX_UPDATE_INT;
	if(net->net_int < XSW_SWSERV_MIN_UPDATE_INT)
	    net->net_int = XSW_SWSERV_MIN_UPDATE_INT;


	/* Send interval to server */
	NetSendSetInterval();

	/* Redraw the bridge window */
	BridgeWinDrawAll(xsw_bridge_win);

	/* Print new interval */
	s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	sprintf(
	    s,
	    "%s: %i ms",
	    THIS_CMD_NAME, (int)net->net_int
	);
	XSWPrint(s);
	free(s);

	return(0);
}
