#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/prochandle.h"

#include "msgwin.h"
#include "xsw.h"


static void CmdServScriptSelectFileCB(void *widget, void *data);
int CmdServScript(const char *arg);


#define THIS_CMD_NAME	"serverscript"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Select file timeout callback.
 */
static void CmdServScriptSelectFileCB(void *widget, void *data)
{
	char *path, **paths_list;
	int npaths;
	win_t toplevel;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_msg_win_struct *mw = xsw_msg_win;

	path = STRDUP(PrefixPaths(
	    dname.scripts, "*"
	));

	if((mw != NULL) ? mw->map_state : False)
	    toplevel = mw->toplevel;
	else if((bw != NULL) ? bw->map_state : False)
	    toplevel = bw->toplevel;
	else
	    toplevel = None;

	if(XSWSelectFile(
	    "Select Server Script File",
	    "Select", "Cancel",
	    path,
	    &paths_list, &npaths,
	    toplevel
	))
	{
	    if(npaths > 0)
	    {
		const char *path = paths_list[0];
                char *parent = STRDUP(GetParentDir(path));
                if(ISPATHDIR(parent))
                {
                    strncpy(
                        dname.scripts, parent,
                        sizeof(dname.scripts)
                    );
                    dname.scripts[sizeof(dname.scripts) - 1] = '\0';
                }
                free(parent);
		ServScriptMapPrompt(path);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
}

int CmdServScript(const char *arg)
{
	int i, n, argc = 0;
	const char *s;
	char **argv = NULL, *path = NULL;

	/* If no arguments are specified then prompt for file */
	if((arg != NULL) ? (*arg == '\0') : True)
	{
	    WidgetIdleAdd(
		CmdServScriptSelectFileCB,
		NULL, NULL
	    );
	    return(0);
	}

	/* Get the path */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, ' ');
	    const int len = (s_end != NULL) ? (s_end - s) : strlen(s);
	    if(len > 0)
	    {
		path = (char *)malloc((len + 1) * sizeof(char));
		if(path != NULL)
		{
		    memcpy(path, s, len);
		    path[len] = '\0';
		}
	    }

	    s = s_end;
	    if(s != NULL)
	    {
		while(ISBLANK(*s))
		    s++;
	    }
	}
	if(path != NULL)
	{
	    char *s = STRDUP(PathSubHome(path));
	    free(path);
	    path = s;
	}

	/* Get the subsequent arguments */
	if(s != NULL)
	{
	    argv = ExecExplodeCommand(s, &argc);
	}

	/* Set the first argument to be that of the path */
	argc++;
	argv = (char **)realloc(argv, argc * sizeof(char *));
	if(argv == NULL)
	    argc = 0;
	for(i = argc - 1, n = 0; i > n; i--)
	    argv[i] = argv[i - 1];

	if((argv != NULL) && (argc >= 1))
	    argv[0] = STRDUP(path);


	/* Start server script */
	i = ServScriptStart(path, argc, argv);

	/* Do not delete argv, ServScript*() functions will do that */

	free(path);

	return(0);
}
