#include <stdlib.h>
#include <string.h>

#include "xsw.h"


int ImgLabelIsAllocated(xsw_imglabel_struct *label);
int ImgLabelIsLoaded(xsw_imglabel_struct *label);

xsw_imglabel_struct *ImgLabelNew(void);
void ImgLabelReset(xsw_imglabel_struct *label);
void ImgLabelDelete(xsw_imglabel_struct *label);


/*
 *	Checks if the label is allocated.
 */
int ImgLabelIsAllocated(xsw_imglabel_struct *label)
{
	return((label != NULL) ? 1 : 0);
}

/*
 *	Checks if the label is allocated and has an image loaded.
 */
int ImgLabelIsLoaded(xsw_imglabel_struct *label)
{
	if(ImgLabelIsAllocated(label))
	    return((label->image != NULL) ? 1 : 0);
	else
            return(0);
}


/*
 *	Creates a new label.
 */
xsw_imglabel_struct *ImgLabelNew(void)
{
	return((xsw_imglabel_struct *)calloc(
	    1, sizeof(xsw_imglabel_struct)
	));
}


/*
 *	Deletes the label's resources but not the label itself.
 */
void ImgLabelReset(xsw_imglabel_struct *label)
{
        if(!ImgLabelIsAllocated(label))
	    return;

	free(label->filename);
	label->filename = NULL;

	label->pos_by_percent = 0;
	label->size_by_percent = 0;

	label->x = 0;
	label->y = 0;

        OSWDestroyImage(&label->image);
}


/*
 *	Deletes the label and all its resources.
 */
void ImgLabelDelete(xsw_imglabel_struct *label)
{
	ImgLabelReset(label);
	free(label);
}
