#ifdef JS_SUPPORT

#include <stdlib.h>
#include <string.h>
#if defined(HAVE_LIBJSW)
# include <jsw.h>
#endif

#include "../include/string.h"

#include "jsmaps.h"
#include "gctl.h"
#include "jsmapwin.h"
#include "xsw.h"

#include "../include/oswkeyvals.h"


static void JSMWUnfocusAll(xsw_jsmap_win_struct *jsmw);

static void JSMWLoadAllJSMaps(xsw_jsmap_win_struct *jsmw);
static void JSMWApplySelected(xsw_jsmap_win_struct *jsmw);

static void JSMWFetchJSMap(xsw_jsmap_win_struct *jsmw, const int n);
static void JSMWApplyJSMap(xsw_jsmap_win_struct *jsmw, const int n);

static void JSMWAddPBCB(push_button_struct *btn, void *data);
static void JSMWRemovePBCB(push_button_struct *btn, void *data);

static void JSMWInitializePBCB(push_button_struct *btn, void *data);

static void JSMWResizeCB(void *widget, void *data);
static void JSMWDrawCB(void *widget, void *data);

static void JSMWJoystickCListActivateCB(
	clist_struct *list, void *data
);
static void JSMWJoystickCListSelectCB(
	clist_struct *list, int row_num, void *data
);
static void JSMWJoystickCListUnselectCB(
	clist_struct *list, int row_num, void *data
);
static void JSMWButtonsCListDataDestroyCB(void *data);
static void JSMWButtonsCListActivateCB(
	clist_struct *list, void *data
);
static void JSMWScanKeyPBCB(push_button_struct *btn, void *data);

static void JSMWOKPBCB(push_button_struct *btn, void *data);
static void JSMWApplyPBCB(push_button_struct *btn, void *data);
static void JSMWCancelPBCB(push_button_struct *btn, void *data);

xsw_jsmap_win_struct *JSMWNew(void);
void JSMWResize(xsw_jsmap_win_struct *jsmw);
void JSMWQueueResize(xsw_jsmap_win_struct *jsmw);
void JSMWDraw(xsw_jsmap_win_struct *jsmw, const int amount);
void JSMWQueueDraw(xsw_jsmap_win_struct *jsmw, const int amount);
int JSMWManage(xsw_jsmap_win_struct *jsmw, event_t *event);
void JSMWMap(xsw_jsmap_win_struct *jsmw, win_t toplevel);
void JSMWUnmap(xsw_jsmap_win_struct *jsmw);
void JSMWDelete(xsw_jsmap_win_struct *jsmw);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

/*
 *	Sizes (in pixels):
 */
#define JSMW_DEF_WIDTH			400
#define JSMW_DEF_HEIGHT			480

#define JSMW_BTN_WIDTH			70
#define JSMW_BTN_HEIGHT			28

#define JSMW_MARGIN			10
#define JSMW_PROMPT_HEIGHT		30

#define JSMW_SCANKEY_WIN_HEIGHT		70

#define JSMW_JS_LIST_HEIGHT		90
#define JSMW_BUTTON_LIST_HEIGHT		120

#define JSMW_DEF_JSDEVICE_NAME		"/dev/null"


/*
 *	Unfocuses all widgets.
 */
static void JSMWUnfocusAll(xsw_jsmap_win_struct *jsmw)
{
	if(jsmw == NULL)
	    return;

	WidgetSetFocus(&jsmw->js_list, False);
	WidgetSetFocus(&jsmw->device_prompt, False);
	WidgetSetFocus(&jsmw->turn_axis_prompt, False);
	WidgetSetFocus(&jsmw->throttle_axis_prompt, False);
	WidgetSetFocus(&jsmw->thrust_dir_axis_prompt, False);
	WidgetSetFocus(&jsmw->vs_zoom_axis_prompt, False);
	WidgetSetFocus(&jsmw->scanner_zoom_axis_prompt, False);
	WidgetSetFocus(&jsmw->aim_weapon_heading_prompt, False);
	WidgetSetFocus(&jsmw->buttons_list, False);
}


/*
 *	Loads all jsmaps.
 */
static void JSMWLoadAllJSMaps(xsw_jsmap_win_struct *jsmw)
{
	int i, row_num;
	const char *device_name;
	char *s, *v[2];
	clist_struct *clist;
	xsw_jsmap_struct *jsm;

	if(jsmw == NULL)
	    return;

	clist = &jsmw->js_list;

	/* Delete the current joysticks listing */
	CListClear(clist);

	/* Get new joysticks listing from the jsmaps list */
	for(i = 0; i < total_jsmaps; i++)
	{
	    jsm = xsw_jsmap[i];
	    if(jsm == NULL)
		continue;

	    row_num = CListRowAppend(clist);
	    if(row_num < 0)
		continue;

	    s = (char *)malloc(80 + 40);
	    sprintf(s, "Joystick %i", i + 1);
	    v[0] = s;

	    device_name = (jsm->device_name != NULL) ?
		jsm->device_name : JSMW_DEF_JSDEVICE_NAME;
	    s = (char *)malloc(STRLEN(device_name) + 80);
	    sprintf(s, "%s", device_name);
	    v[1] = s;

	    CListSetRowTexts(
		clist, row_num,
		(const char **)v, sizeof(v) / sizeof(char *)
	    );

	    free(v[0]);
	    free(v[1]);
	}
	CListAutosizeColumns(clist);

	/* Clear all prompts */
	PromptSetS(&jsmw->device_prompt, "");

	PromptSetS(&jsmw->turn_axis_prompt, "");
	PromptSetS(&jsmw->throttle_axis_prompt, "");
	PromptSetS(&jsmw->thrust_dir_axis_prompt, "");
	PromptSetS(&jsmw->vs_zoom_axis_prompt, "");
	PromptSetS(&jsmw->scanner_zoom_axis_prompt, "");
	PromptSetS(&jsmw->aim_weapon_heading_prompt, "");

	/* Clear the buttons list */
	CListClear(&jsmw->buttons_list);
}


/*
 *	Applies the values of the selected joystick (if any).
 */
static void JSMWApplySelected(xsw_jsmap_win_struct *jsmw)
{
	int i;

	if(jsmw == NULL)
	    return;

	i = CListGetLastSelectedRow(&jsmw->js_list);
	if(i > -1)
	    JSMWApplyJSMap(jsmw, i);
}


/*
 *	Fetches values from the jsmap specified by n.
 */
static void JSMWFetchJSMap(xsw_jsmap_win_struct *jsmw, const int n)
{
	int i, row_num;
	const char *keycode_name;
	char *s, *v[2];
	clist_struct *clist;
	xsw_jsmap_axis_struct *axis;
	xsw_jsmap_button_struct *btn, *btn2;
	xsw_jsmap_struct *jsm;

	if(jsmw == NULL)
	    return;

	/* Reset values first */

	/* Reset axis values */
	PromptSetS(&jsmw->turn_axis_prompt, "-1");
	PromptSetS(&jsmw->throttle_axis_prompt, "-1");
	PromptSetS(&jsmw->thrust_dir_axis_prompt, "-1");
	PromptSetS(&jsmw->vs_zoom_axis_prompt, "-1");
	PromptSetS(&jsmw->scanner_zoom_axis_prompt, "-1");
	PromptSetS(&jsmw->aim_weapon_heading_prompt, "-1");

	/* Clear the buttons list */
	CListClear(&jsmw->buttons_list);

	/* Get the jsmap */
	jsm = ((n >= 0) && (n < total_jsmaps)) ?
	    xsw_jsmap[n] : NULL;
	if(jsm == NULL)
	    return;

	/* Get axis values */
	for(i = 0; i < jsm->total_axises; i++)
	{
	    axis = jsm->axis[i];
	    if(axis == NULL)
		continue;

	    switch(axis->op)
	    {
	      case JSMAP_AXIS_OP_TURN:
	        PromptSetI(
		    &jsmw->turn_axis_prompt,
		    i
		);
		break;

	      case JSMAP_AXIS_OP_THROTTLE:
		PromptSetI(
		    &jsmw->throttle_axis_prompt,
		    i
		);
		break;

	      case JSMAP_AXIS_OP_THRUST_DIR:
		PromptSetI(
		    &jsmw->thrust_dir_axis_prompt,
		    i
		);
		break;

	      case JSMAP_AXIS_OP_SCANNER_ZOOM:
		PromptSetI(
		    &jsmw->scanner_zoom_axis_prompt,
		    i
		);
		break;

	      case JSMAP_AXIS_OP_VS_ZOOM:
		PromptSetI(
		    &jsmw->vs_zoom_axis_prompt,
		    i
		);
		break;

	      case JSMAP_AXIS_OP_AIM_WEAPON_HEADING:
		PromptSetI(
		    &jsmw->aim_weapon_heading_prompt,
		    i
		);
		break;

	      case JSMAP_AXIS_OP_AIM_WEAPON_PITCH:
		break;
	    }
	}


	/* Get buttons listing */
	clist = &jsmw->buttons_list;
	for(i = 0; i < jsm->total_buttons; i++)
	{
	    btn = jsm->button[i];
	    if(btn == NULL)
		continue;

	    row_num = CListRowAppend(clist);
	    if(row_num < 0)
		continue;

	    s = (char *)malloc(80 + 40);
	    sprintf(s, "Button %i", i + 1);
	    v[0] = s;

	    keycode_name = OSWGetKeyCodeName(btn->keycode);
	    s = (char *)malloc(STRLEN(keycode_name) + 80 + 80 + 80);
	    sprintf(   
		s,
		"%s%s%s%s",
		(btn->key_state & OSW_KEY_STATE_CTRL) ? "CTRL + " : "",
		(btn->key_state & OSW_KEY_STATE_ALT) ? "ALT + " : "",
		(btn->key_state & OSW_KEY_STATE_SHIFT) ? "SHIFT + " : "",
		keycode_name
	    );
	    strtoupper(s);
	    v[1] = s;

	    CListSetRowTexts(
		clist, row_num,
		(const char **)v, sizeof(v) / sizeof(char *)
	    );
	    free(v[0]);
	    free(v[1]);

	    /* Set data pointer on list item on colum 1 to
	     * be the value of (not point to) the button's
	     * generated keycode
	     */
	    btn2 = XSW_JSMAP_BUTTON(calloc(
		1, sizeof(xsw_jsmap_button_struct)
	    ));
	    if(btn2 != NULL)
	    {
		btn2->keycode = btn->keycode;
		btn2->key_state = btn->key_state;
		btn2->state = btn->state;
		CListRowSetData(
		    clist, row_num,
		    btn2, JSMWButtonsCListDataDestroyCB
		);
	    }
	}
	CListAutosizeColumns(clist);

	/* Update device prompt */
	PromptSetS(
	    &jsmw->device_prompt,
	    jsm->device_name
	);
}

/*
 *	Applies values from widgets into jsmap n.
 */
static void JSMWApplyJSMap(xsw_jsmap_win_struct *jsmw, const int n)
{
	int i;
	clist_struct *clist;
	xsw_jsmap_struct *jsm;
	xsw_jsmap_axis_struct *axis;
	xsw_jsmap_button_struct *btn, *btn2;

	if(jsmw == NULL)
	    return;

	/* Get the jsmap */
	jsm = ((n >= 0) && (n < total_jsmaps)) ?
	    xsw_jsmap[n] : NULL;
	if(jsm == NULL)
	    return;

	/* Reset axis values */
	for(i = 0; i < jsm->total_axises; i++)
	{
	    axis = jsm->axis[i];
	    if(axis == NULL)
		continue;

	    axis->op = JSMAP_AXIS_OP_NONE;
 	}


	/* Begin applying axis values */

	/* Turn axis number */
	i = PromptGetI(&jsmw->turn_axis_prompt);
	if((i >= 0) && (i < jsm->total_axises))
	{
	    axis = jsm->axis[i];
	    if(axis != NULL)
		axis->op = JSMAP_AXIS_OP_TURN;
	}

	/* Throttle axis number */
	i = PromptGetI(&jsmw->throttle_axis_prompt); 
	if((i >= 0) && (i < jsm->total_axises))
	{
	    axis = jsm->axis[i];
	    if(axis != NULL)
		axis->op = JSMAP_AXIS_OP_THROTTLE;
	}

	/* Thrust direction axis number */
	i = PromptGetI(&jsmw->thrust_dir_axis_prompt);
	if((i >= 0) && (i < jsm->total_axises))
	{
	    axis = jsm->axis[i];
	    if(axis != NULL)
		axis->op = JSMAP_AXIS_OP_THRUST_DIR;
	}

	/* Viewscreen zoom axis number */
	i = PromptGetI(&jsmw->vs_zoom_axis_prompt);
	if((i >= 0) && (i < jsm->total_axises))
	{
	    axis = jsm->axis[i];
	    if(axis != NULL)
		axis->op = JSMAP_AXIS_OP_VS_ZOOM;
	}

	/* Scanner zoom axis number */
	i = PromptGetI(&jsmw->scanner_zoom_axis_prompt);
	if((i >= 0) && (i < jsm->total_axises))
	{
	    axis = jsm->axis[i];
	    if(axis != NULL)
		axis->op = JSMAP_AXIS_OP_SCANNER_ZOOM;
	}

	/* Aim weapon heading axis number */
	i = PromptGetI(&jsmw->aim_weapon_heading_prompt);
	if((i >= 0) && (i < jsm->total_axises))
	{
	    axis = jsm->axis[i];
	    if(axis != NULL)
		axis->op = JSMAP_AXIS_OP_AIM_WEAPON_HEADING;
	}


	/* Apply button values */
	clist = &jsmw->buttons_list;
	for(i = 0; i < clist->total_rows; i++)
	{
	    if(i >= jsm->total_buttons)
		break;

	    btn = jsm->button[i];
	    if(btn == NULL)
		continue;

	    btn2 = XSW_JSMAP_BUTTON(CListRowGetData(clist, i));
	    if(btn2 == NULL)
		continue;

	    btn->keycode = btn2->keycode;
	    btn->key_state = btn2->key_state;
/*	    btn->state = btn2->state; */
	}

	/* Update device name */
	free(jsm->device_name);
	jsm->device_name = STRDUP(
	    PromptGetS(&jsmw->device_prompt)
	);

	/* Update the device column in the joysticks list */
	clist = &jsmw->js_list;
	CListSetItemText(
	    clist,
	    n,
	    1,
	    jsm->device_name
	);
	CListAutosizeColumns(clist);
}


/*
 *	Add callback.
 */
void JSMWAddPBCB(push_button_struct *btn, void *data)
{
	int n;
	clist_struct *clist;
	xsw_jsmap_struct *jsm;
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	clist = &jsmw->js_list;

	/* Unselect to apply selected jsmap */
	CListUnselectAllRows(clist);

	/* Allocate a new jsmap */
	if(total_jsmaps < 0)
	    total_jsmaps = 0;
	n = total_jsmaps;
	total_jsmaps = n + 1;
	xsw_jsmap = (xsw_jsmap_struct **)realloc(
	    xsw_jsmap,
	    total_jsmaps * sizeof(xsw_jsmap_struct *)
	);
	if(xsw_jsmap == NULL)
	{
	    total_jsmaps = 0;
	    return;
	}

	xsw_jsmap[n] = jsm = JSMapNew();
	if(jsm == NULL)
	    return;

#if defined(HAVE_LIBJSW)
	jsm->device_name = STRDUP(JSDefaultDevice);
#endif

	/* Refetch all jsmap values */
	JSMWLoadAllJSMaps(jsmw);

	/* Select joystick on joystick list widget */
	CListSelectRow(clist, n);
}

/*
 *	Remove callback.
 */
void JSMWRemovePBCB(push_button_struct *btn, void *data)
{
	int i;
	clist_struct *clist;
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data); 
	if(jsmw == NULL)
	    return;

	clist = &jsmw->js_list;
 
	/* Delete selected jsmap entry */
	i = CListGetLastSelectedRow(clist);
	if(i > -1)
	{
	    int n;

	    /* Unselect and apply, even though we are going to
	     * delete it, we do not want it's values being applied
	     * to another jsmap when it is deleted
	     */
	    CListUnselectRow(clist, i);

	    /* Delete the jsmap */
	    if((i >= 0) && (i < total_jsmaps))
		JSMapDelete(xsw_jsmap[i]);

	    /* Shift and reallocate pointers */
	    for(n = i; n < (total_jsmaps - 1); n++)
		xsw_jsmap[n] = xsw_jsmap[n + 1];

	    total_jsmaps--;
	    if(total_jsmaps > 0)
	    {
		xsw_jsmap = (xsw_jsmap_struct **)realloc(
		    xsw_jsmap,
		    total_jsmaps * sizeof(xsw_jsmap_struct *)
		);
		if(xsw_jsmap == NULL)
		{
		    total_jsmaps = 0;
		    return;
		}
	    }
	    else
	    {
		free(xsw_jsmap);
		xsw_jsmap = NULL;
		total_jsmaps = 0;
	    }
	}

	/* Refetch all jsmap values */
	JSMWLoadAllJSMaps(jsmw);
}


/*
 *	Initialize callback.
 */
static void JSMWInitializePBCB(push_button_struct *btn, void *data)
{
	int jsmap_num;
	clist_struct *clist;
	xsw_jsmap_struct *jsm;
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	clist = &jsmw->js_list;
	jsmap_num = CListGetLastSelectedRow(clist);

	/* Get the jsmap */
	jsm = ((jsmap_num >= 0) && (jsmap_num < total_jsmaps)) ?
	    xsw_jsmap[jsmap_num] : NULL;
	if(jsm == NULL)
	    return;

	/* Apply current values to the selected jsmap */
	JSMWApplyJSMap(jsmw, jsmap_num);

	/* Reinitialize the joystick */
	if(JSMapInitializeJoystick(
	    jsm,
	    jsm->device_name,
	    fname.js_calib
	))
	{
	    char *s = (char *)malloc(
		STRLEN(jsm->device_name) + (3 * 80)
	    );
	    sprintf(
		s,
"Unable to initialize joystick:\n\
\n\
    %s",
		jsm->device_name
	    );
	    XSWDialog(
		"Joystick Initialization Failed",
		XSW_IMG_DIALOG_ICON_ERROR,
		s,
		jsmw->toplevel
	    );
	    free(s);
	    return;
	}

	/* Refetch any changed values for the selected joystick */
	JSMWFetchJSMap(jsmw, jsmap_num);
}


/*
 *	Resize callback.
 */
static void JSMWResizeCB(void *widget, void *data)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(widget);
	if(jsmw == NULL)
	    return;

	JSMWResize(jsmw);
}

/*
 *	Draw callback.
 */
static void JSMWDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(widget);
	if(jsmw == NULL)
	    return;

	if(jsmw->map_state)
	    JSMWDraw(jsmw, amount);
}


/*
 *	Joystick clist activate callback.
 */
static void JSMWJoystickCListActivateCB(
	clist_struct *list, void *data
)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)                                 
	    return;
}

/*
 *	Joystick clist select callback.
 */
static void JSMWJoystickCListSelectCB(
	clist_struct *list, int row_num, void *data
)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	JSMWFetchJSMap(jsmw, row_num);
}

/*
 *	Joystick clist unselect callback.
 */
static void JSMWJoystickCListUnselectCB(
	clist_struct *list, int row_num, void *data
)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)                                 
	    return;

	JSMWApplyJSMap(jsmw, row_num);
}


/*
 *	Buttons clist row data destroy callback.  
 */
static void JSMWButtonsCListDataDestroyCB(void *data)
{
	xsw_jsmap_button_struct *btn = XSW_JSMAP_BUTTON(data);
	if(btn == NULL)
	    return;

	free(btn);
}

/*
 *	Buttons clist activate callback.
 */
static void JSMWButtonsCListActivateCB(
	clist_struct *list, void *data
)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	JSMWScanKeyPBCB(&jsmw->scan_key_btn, jsmw);
}

/*
 *	Callback handler for the scan key button.
 */
static void JSMWScanKeyPBCB(push_button_struct *btn, void *data)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	OSWGetWindowGeometry(
	    jsmw->toplevel, NULL, NULL, &width, &height, NULL
	);

	/* Move, resize, and map the scan key window */
	w = jsmw->scan_key_win;
	x = JSMW_MARGIN;
	y = ((int)height - JSMW_SCANKEY_WIN_HEIGHT) / 2;
	width = (unsigned int)MAX(
	    (int)width - (2 * JSMW_MARGIN),
	    1
	);
	height = JSMW_SCANKEY_WIN_HEIGHT;
	OSWMoveResizeWindow(w, x, y, width, height);
	OSWMapRaised(w);

	jsmw->scanning_key = True;
}

/*
 *	Callback handler for the OK button.
 */
static void JSMWOKPBCB(push_button_struct *btn, void *data)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	JSMWApplySelected(jsmw);
	JSMWUnmap(jsmw);
}

/*
 *	Callback handler for the apply button.
 */
static void JSMWApplyPBCB(push_button_struct *btn, void *data)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	JSMWApplySelected(jsmw);
}

/*
 *	Callback handler for the cancel button.
 */
static void JSMWCancelPBCB(push_button_struct *btn, void *data)
{
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(data);
	if(jsmw == NULL)
	    return;

	JSMWUnmap(jsmw);
}


/*
 *	Creates a new Joystick Map Window.
 */
xsw_jsmap_win_struct *JSMWNew(void)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	font_t *font;
	pixel_t pixel;
	pixmap_t pixmap;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	push_button_struct *btn;
	clist_struct *clist;
	widget_global_struct *wglobal = &widget_global;
	xsw_jsmap_win_struct *jsmw = XSW_JSMAP_WIN(calloc(
	    1, sizeof(xsw_jsmap_win_struct)
	));
	if(jsmw == NULL)
	    return(NULL);

	jsmw->map_state = False;
	jsmw->has_focus = False;
	jsmw->disabled = False;
	jsmw->freeze_count = 0;
	jsmw->busy_count = 0;
	jsmw->x = 0;
	jsmw->y = 0;
	jsmw->width = JSMW_DEF_WIDTH;
	jsmw->height = JSMW_DEF_HEIGHT;
	jsmw->visibility_state = VisibilityFullyObscured;
	jsmw->disabled = False;
	jsmw->scanning_key = False;

	/* Create toplevel */
	x = jsmw->x;
	y = jsmw->y;
	width = jsmw->width;
	height = jsmw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	jsmw->toplevel = w;
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);

	/* WM properties */
	if(ImgIsAllocated(XSW_IMG_ICON_OPTIONS))
	    pixmap = OSWCreatePixmapFromImage(
		xsw_image[XSW_IMG_ICON_OPTIONS]->image
	    );
	else
	    pixmap = wglobal->std_icon_pm;
	OSWSetWindowWMProperties(
	    w,
	    PROG_NAME_FULL ": Joystick Mappings",
	    "Joystick Mappings",	/* Icon title */
	    pixmap,			/* Icon */
	    False,			/* Let WM set coordinates? */
	    /* Coordinates */
	    jsmw->x, jsmw->y,
	    /* Min width and height */
	    jsmw->width, jsmw->height,
	    /* Max width and height */
	    jsmw->width, jsmw->height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);

	parent = w;

	/* Scan Key prompt window */
	x = JSMW_MARGIN;
	y = ((int)height - JSMW_SCANKEY_WIN_HEIGHT) / 2;
	width = (unsigned int)MAX(
	    (int)width - (2 * JSMW_MARGIN),
	    1
	);
	height = JSMW_SCANKEY_WIN_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	jsmw->scan_key_win = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | ButtonPressMask
	);

	/* Joysticks List */
	clist = &jsmw->js_list;
	x = JSMW_MARGIN;
	y = JSMW_MARGIN;
	CListInit(
	    clist, parent,
	    x, y,
	    (int)jsmw->width - (3 * JSMW_MARGIN) - JSMW_BTN_WIDTH,
	    JSMW_JS_LIST_HEIGHT
	);
	clist->activate_cb = JSMWJoystickCListActivateCB;
	clist->activate_data = jsmw;
	clist->select_cb = JSMWJoystickCListSelectCB;
	clist->select_data = jsmw;
	clist->unselect_cb = JSMWJoystickCListUnselectCB;
	clist->unselect_data = jsmw;
	font = wglobal->std_font;
	pixel = gui->black_pix;
	CListColumnAppend(
	    clist,
	    "Joystick Number",
	    font, pixel,
	    CLIST_COLUMN_FLAG_ALIGN_LEFT,
	    0			/* x position */
	);
	CListColumnAppend(
	    clist,
	    "Device",
	    font, pixel,
	    CLIST_COLUMN_FLAG_ALIGN_LEFT,
	    120			/* x position */
	);

	/* Add button */
	btn = &jsmw->add_btn;
	x = (int)jsmw->width - JSMW_BTN_WIDTH - JSMW_MARGIN;
	y = JSMW_MARGIN;
	PBtnInit(
	    btn, parent,
	    x, y,
	    JSMW_BTN_WIDTH, JSMW_BTN_HEIGHT,
	    "Add",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    JSMWAddPBCB, jsmw
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);
	

	/* Remove button */
	btn = &jsmw->remove_btn;
	y += JSMW_BTN_HEIGHT + 2;
	PBtnInit(
	    &jsmw->remove_btn, parent,
	    x, y,
	    JSMW_BTN_WIDTH, JSMW_BTN_HEIGHT,
	    "Remove",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    JSMWRemovePBCB, jsmw
	);
	accel_keys[0] = OSW_KEY_r;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Device prompt */
	x = JSMW_MARGIN;
	y = JSMW_MARGIN + JSMW_JS_LIST_HEIGHT + (JSMW_MARGIN / 2);
	PromptInit(
	    &jsmw->device_prompt, parent,
	    x, y,
	    (int)jsmw->width - (3 * JSMW_MARGIN) - JSMW_BTN_WIDTH,
	    JSMW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Device:",
	    PATH_MAX + NAME_MAX, 0
	);

	/* Initialize button */
	btn = &jsmw->initialize_btn;
	x = (int)jsmw->width - JSMW_BTN_WIDTH - JSMW_MARGIN;
	PBtnInit(
	    &jsmw->initialize_btn, parent,
	    x, y,
	    JSMW_BTN_WIDTH, JSMW_BTN_HEIGHT,
	    "Init",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    JSMWInitializePBCB, jsmw
	);
	accel_keys[0] = OSW_KEY_i;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Axis prompts */
	x = JSMW_MARGIN;
	y = JSMW_MARGIN + JSMW_JS_LIST_HEIGHT +
	    (JSMW_MARGIN / 2) + JSMW_PROMPT_HEIGHT + JSMW_MARGIN;
	PromptInit(
	    &jsmw->turn_axis_prompt, parent,
	    x, y,
	    120,
	    JSMW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Turn Axis:",
	    64, 0
	);

	x = JSMW_MARGIN + 120 + JSMW_MARGIN;
	PromptInit(
	    &jsmw->throttle_axis_prompt, parent,
	    x, y,
	    150,
	    JSMW_PROMPT_HEIGHT,  
	    PROMPT_STYLE_FLUSHED,
	    "Throttle Axis:",
	    64, 0
	);

	x = JSMW_MARGIN;
	y += JSMW_PROMPT_HEIGHT + (JSMW_MARGIN / 2);
	PromptInit(
	    &jsmw->thrust_dir_axis_prompt, parent,
	    x, y,
	    210,
	    JSMW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Thrust Direction Axis:",
	    64, 0
	);

	x = JSMW_MARGIN;
	y += (JSMW_PROMPT_HEIGHT + (JSMW_MARGIN / 2));
	PromptInit(
	    &jsmw->vs_zoom_axis_prompt, parent,
	    x, y,
	    170,
	    JSMW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Viewscreen Axis:",
	    64, 0
	);

	x = JSMW_MARGIN + 170 + JSMW_MARGIN; 
	PromptInit(
	    &jsmw->scanner_zoom_axis_prompt, parent,
	    x, y,
	    150,
	    JSMW_PROMPT_HEIGHT,  
	    PROMPT_STYLE_FLUSHED,
	    "Scanner Axis:",
	    64, 0
	);

	x = JSMW_MARGIN;
	y += (JSMW_PROMPT_HEIGHT + (JSMW_MARGIN / 2));
	PromptInit(
	    &jsmw->aim_weapon_heading_prompt, parent,
	    x, y,
	    210,
	    JSMW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Aim Weapon Heading Axis:",
	    64, 0
	);

	/* Link prompts togeather */
	jsmw->device_prompt.next = &jsmw->turn_axis_prompt;
	jsmw->device_prompt.prev = &jsmw->aim_weapon_heading_prompt;

	jsmw->turn_axis_prompt.next = &jsmw->throttle_axis_prompt;
	jsmw->turn_axis_prompt.prev = &jsmw->device_prompt;

	jsmw->throttle_axis_prompt.next = &jsmw->thrust_dir_axis_prompt;
	jsmw->throttle_axis_prompt.prev = &jsmw->turn_axis_prompt;

	jsmw->thrust_dir_axis_prompt.next = &jsmw->vs_zoom_axis_prompt;
	jsmw->thrust_dir_axis_prompt.prev = &jsmw->throttle_axis_prompt;

	jsmw->vs_zoom_axis_prompt.next = &jsmw->scanner_zoom_axis_prompt;
	jsmw->vs_zoom_axis_prompt.prev = &jsmw->thrust_dir_axis_prompt;

	jsmw->scanner_zoom_axis_prompt.next = &jsmw->aim_weapon_heading_prompt;
	jsmw->scanner_zoom_axis_prompt.prev = &jsmw->vs_zoom_axis_prompt;

	jsmw->aim_weapon_heading_prompt.next = &jsmw->device_prompt;
	jsmw->aim_weapon_heading_prompt.prev = &jsmw->scanner_zoom_axis_prompt;


	/* Buttons clist */
	clist = &jsmw->buttons_list;
	x = JSMW_MARGIN;
	y += JSMW_PROMPT_HEIGHT + JSMW_MARGIN;
	CListInit(
	    clist, parent,
	    x, y,
	    (int)jsmw->width - (3 * JSMW_MARGIN) - JSMW_BTN_WIDTH,
	    JSMW_BUTTON_LIST_HEIGHT
	);
	clist->activate_cb = JSMWButtonsCListActivateCB;
	clist->activate_data = jsmw;
	font = wglobal->std_font;
	pixel = gui->black_pix;
	CListColumnAppend(
	    clist,
	    "Button Number",
	    font, pixel,
	    CLIST_COLUMN_FLAG_ALIGN_LEFT,
	    0                   /* x position */
	);
	CListColumnAppend(
	    clist,
	    "Key",
	    font, pixel,
	    CLIST_COLUMN_FLAG_ALIGN_LEFT,
	    120			/* x position */
	);

	/* Scan key button */
	btn = &jsmw->scan_key_btn;
	x = (int)jsmw->width - JSMW_BTN_WIDTH - JSMW_MARGIN;
	y = JSMW_MARGIN + JSMW_JS_LIST_HEIGHT + (JSMW_MARGIN / 2) +
	    JSMW_PROMPT_HEIGHT + JSMW_MARGIN +
	    JSMW_PROMPT_HEIGHT + (JSMW_MARGIN / 2) +
	    JSMW_PROMPT_HEIGHT + (JSMW_MARGIN / 2) +
	    JSMW_PROMPT_HEIGHT + (JSMW_MARGIN / 2) +
	    JSMW_PROMPT_HEIGHT + JSMW_MARGIN;
	PBtnInit(
	    &jsmw->scan_key_btn, parent,
	    x, y,
	    JSMW_BTN_WIDTH, JSMW_BTN_HEIGHT,
	    "Scan Key",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    JSMWScanKeyPBCB, jsmw
	);
	accel_keys[0] = OSW_KEY_Return;
	accel_mods[0] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);


	/* OK button */
	btn = &jsmw->ok_btn;
	width = JSMW_BTN_WIDTH;
	height = JSMW_BTN_HEIGHT;
	x = JSMW_MARGIN;
	y = (int)jsmw->height - JSMW_MARGIN - height;
	PBtnInit(
	    &jsmw->ok_btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    JSMWOKPBCB, jsmw
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Apply button */
	btn = &jsmw->apply_btn;
	x = ((int)jsmw->width - (int)width) / 2;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Apply",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    JSMWApplyPBCB, jsmw
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Cancel button */
	btn = &jsmw->cancel_btn;
	x = (int)jsmw->width - (int)width - JSMW_MARGIN;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    JSMWCancelPBCB, jsmw
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);


	return(jsmw);
}

/*
 *	Resizes the jsmap window.
 */
void JSMWResize(xsw_jsmap_win_struct *jsmw)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(jsmw == NULL)
	    return;

	w = jsmw->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	jsmw->x = x;
	jsmw->y = y;
	jsmw->width = width;
	jsmw->height = height;


}

/*
 *	Queues the Joystick Mappings Window to resize.
 */
void JSMWQueueResize(xsw_jsmap_win_struct *jsmw)
{
	if(jsmw == NULL)
	    return;

	WidgetQueueResizeAction(
	    JSMWResizeCB, jsmw, NULL
	);

}

/*
 *	Draws the Joystick Mappings Window.
 */
void JSMWDraw(xsw_jsmap_win_struct *jsmw, const int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(jsmw == NULL)
	    return;

	/* Toplevel */
	if(amount == JSMW_DRAW_AMOUNT_COMPLETE)
	{
	    unsigned int width, height;
	    win_t w = jsmw->toplevel;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		int y;

		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);

		OSWSetForeground(wglobal->surface_shadow_pix);
		y = (int)height - JSMW_BTN_HEIGHT - 25;
		OSWDrawLine(
		    w,
		    0, y, (int)width, y
		);
		OSWSetForeground(wglobal->surface_highlight_pix);
		y = (int)height - JSMW_BTN_HEIGHT - 24;
		OSWDrawLine(
		    w,
		    0, y, (int)width, y
		);
	    }
	}

	/* Scan key window */
	if((amount == JSMW_DRAW_AMOUNT_COMPLETE) ||
	   (amount == JSMW_DRAW_AMOUNT_SCANKEY)
	)
	{
	    if(jsmw->scanning_key)
	    {
		int	x, y,
			lbearing, rbearing, ascent, descent,
			font_height;
		unsigned int width, height;
		const char *s;
		font_t *font = wglobal->std_font;
		win_t w = jsmw->scan_key_win;

		OSWGetWindowGeometry(
		    w, NULL, NULL, &width, &height, NULL
		);
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->white_pix :
			xsw_color.keymap_query_bg
		);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->black_pix :
			xsw_color.keymap_query_fg
		);

		OSWSetFont(font);
		x = JSMW_MARGIN;
		y = JSMW_MARGIN;
		s =
"Press a key on the keyboard to set the keycode.";
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, &rbearing, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    w,
		    x - lbearing,
		    y + ascent,
		    s
		);
		y += font_height + (JSMW_MARGIN / 2);

		s = "Press Button1 to cancel.";
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, &rbearing, NULL, &ascent, &descent
		);
		OSWDrawString(
		    w,
		    x - lbearing,
		    y + ascent,  
		    s
		);
		y += font_height;

		s = "Press Button3 to clear.";
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, &rbearing, NULL, &ascent, &descent
		);
		OSWDrawString(
		    w,
		    x - lbearing,
		    y + ascent,  
		    s
		);
	    }
	}
}

/*
 *	Queues the Joystick Mappings Window to resize.
 */
void JSMWQueueDraw(xsw_jsmap_win_struct *jsmw, const int amount)
{
	if(jsmw == NULL)
	    return;

	WidgetQueueDrawAction(
	    JSMWDrawCB, jsmw, (void *)amount
	);
} 

/*
 *	Manages the Joystick Mappings Window event.
 */
int JSMWManage(xsw_jsmap_win_struct *jsmw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	keycode_t keycode;
	key_state_t key_state;
	event_button_t *button;
	event_key_t *key;
	event_visibility_t *visibility;

	if((jsmw == NULL) || (event == NULL))
	    return(events_handled);

	if(!jsmw->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == jsmw->toplevel)
	    {
		JSMWQueueResize(jsmw);
		if(jsmw->map_state)
		    JSMWQueueDraw(jsmw, JSMW_DRAW_AMOUNT_COMPLETE);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case Expose:
	    if(w == jsmw->toplevel)
	    {
		JSMWQueueDraw(jsmw, JSMW_DRAW_AMOUNT_COMPLETE);
		events_handled++;
		return(events_handled);
	    }
	    else if(w == jsmw->scan_key_win)
	    {
		JSMWQueueDraw(jsmw, JSMW_DRAW_AMOUNT_SCANKEY);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!jsmw->has_focus)
		return(events_handled);
	    if(jsmw->scanning_key)
	    {
		OSWKBAutoRepeatOff();
		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!jsmw->has_focus)
		return(events_handled);

	    keycode = key->keycode;
	    key_state = key->state;

	    /* Scanning for keyboard input? */
	    if(jsmw->scanning_key)
	    {
		clist_struct *clist = &jsmw->buttons_list;
		int i = CListGetLastSelectedRow(clist);
		xsw_jsmap_button_struct *btn = XSW_JSMAP_BUTTON(
		    CListRowGetData(clist, i)
		);
		if(btn != NULL)
		{
		    const char *keycode_name = OSWGetKeyCodeName(keycode);
		    char *s = (char *)malloc(STRLEN(keycode_name) + 80 + 80 + 80);
		    sprintf(
			s,
			"%s%s%s%s",
			(key_state & OSW_KEY_STATE_CTRL) ? "CTRL + " : "",
			(key_state & OSW_KEY_STATE_ALT) ? "ALT + " : "",
			(key_state & OSW_KEY_STATE_SHIFT) ? "SHIFT + " : "",
			keycode_name
		    );
		    strtoupper(s);
		    CListSetItemText(
			clist,
			i, 1,
			s
		    );
		    free(s);
		    CListAutosizeColumns(clist);
		    btn->keycode = keycode;
		    btn->key_state = key_state;
		}

		OSWKBAutoRepeatOn();

		/* Unmap the scan key prompt */
		jsmw->scanning_key = False;
		OSWUnmapWindow(jsmw->scan_key_win);

		events_handled++;
		return(events_handled);
	    }
	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    /* Scanning for keyboard input? */
	    if(jsmw->scanning_key)
	    {
		clist_struct *clist = &jsmw->buttons_list;
		int i = CListGetLastSelectedRow(clist);
		xsw_jsmap_button_struct *btn = XSW_JSMAP_BUTTON(
		    CListRowGetData(clist, i)
		);

		/* Unmap the scan key prompt window */
		jsmw->scanning_key = False;
		OSWUnmapWindow(jsmw->scan_key_win);

		if(btn != NULL)
		{
		    /* Check which button was pressed */
		    switch(button->button)
		    {
		      case Button3:	/* Clear */
			btn->keycode = 0;
			btn->key_state = 0;
			btn->state = False;
			CListSetItemText(
			    clist,
			    i, 1,
			    OSWGetKeyCodeName(btn->keycode)
			);
			CListAutosizeColumns(clist);
			break;
		    }
		}
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;

	    break;

	  case MapNotify:
	    if(w == jsmw->toplevel)
	    {
		if(!jsmw->map_state)
		    JSMWMap(jsmw, None);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case UnmapNotify:
	    if(w == jsmw->toplevel)
	    {
		if(jsmw->map_state)
		    JSMWUnmap(jsmw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == jsmw->toplevel)
	    {
		jsmw->visibility_state = visibility->state;
		events_handled++;
		return(events_handled);
	    }
	    break;
 
	  case FocusIn:
	    if(w == jsmw->toplevel)
	    {
		jsmw->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == jsmw->toplevel)
	    {
		JSMWUnfocusAll(jsmw);
		jsmw->has_focus = False;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(jsmw->toplevel, event, &t))
	    {
		OSWSetInputFocus(jsmw->toplevel, RevertToParent, t);
		events_handled++;
		return(events_handled); 
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(jsmw->toplevel, event))
	    {
		JSMWUnmap(jsmw);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}


	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->initialize_btn, event);

	if(events_handled == 0)            
	    events_handled += WidgetManage(&jsmw->add_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->remove_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->scan_key_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->ok_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->apply_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->cancel_btn, event);


	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->js_list, event);


	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->device_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->turn_axis_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->throttle_axis_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->thrust_dir_axis_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->vs_zoom_axis_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->scanner_zoom_axis_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->aim_weapon_heading_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&jsmw->buttons_list, event);


	return(events_handled);
}

/*
 *	Maps the jsmap window.
 */
void JSMWMap(xsw_jsmap_win_struct *jsmw, win_t toplevel)
{
	win_t w;
	clist_struct *clist;

	if(jsmw == NULL)
	    return;

	/* Make sure global controller is set to joystick, if not
	 * then print warning
	 */
	if(option.controller != XSW_CONTROLLER_JOYSTICK)
	{
	    XSWDialog(
		"Joystick Controller Not Set",
		XSW_IMG_DIALOG_ICON_WARNING,
"You must set the Controller Type to `Joystick' before you\n\
can edit the joystick mappings.\n\
\n\
Go to Options->General and set the Controller Type to\n\
`Joystick' then click on `Apply'.",
		jsmw->toplevel
	    );
	    return;
	}

	if(jsmw->toplevel == None)
	    return;

	if(!jsmw->map_state)
	{
	    /* Fetch all jsmap values */
	    JSMWLoadAllJSMaps(jsmw);

	    /* Select first jsmap */
	    clist = &jsmw->js_list;
	    CListUnselectAllRows(clist);
	    CListSelectRow(clist, 0);
	}

	JSMWUnfocusAll(jsmw);

	w = jsmw->toplevel;
	OSWSetTransientFor(w, toplevel);
	if(toplevel != None)
	    WidgetCenterWindowToWindow(w, toplevel);
	else
	    WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);

	/* Map */
	if(!jsmw->map_state)
	{
	    OSWMapRaised(jsmw->toplevel);
	    jsmw->map_state = True;
	    WidgetMap(&jsmw->js_list);
	    WidgetMap(&jsmw->add_btn);   
	    WidgetMap(&jsmw->remove_btn);
	    WidgetMap(&jsmw->device_prompt);
	    WidgetMap(&jsmw->initialize_btn);
	    WidgetMap(&jsmw->turn_axis_prompt); 
	    WidgetMap(&jsmw->throttle_axis_prompt);
	    WidgetMap(&jsmw->thrust_dir_axis_prompt);
	    WidgetMap(&jsmw->vs_zoom_axis_prompt);
	    WidgetMap(&jsmw->scanner_zoom_axis_prompt);
	    WidgetMap(&jsmw->aim_weapon_heading_prompt);
	    WidgetMap(&jsmw->buttons_list);
	    WidgetMap(&jsmw->scan_key_btn);
	    jsmw->scanning_key = False;
	    OSWUnmapWindow(jsmw->scan_key_win);
	    WidgetMap(&jsmw->ok_btn);
	    WidgetMap(&jsmw->apply_btn);
	    WidgetMap(&jsmw->cancel_btn);
	}

	/* Focus */
	OSWGrabFocus(jsmw->toplevel);
}

/*
 *	Unmaps the jsmap window.
 */
void JSMWUnmap(xsw_jsmap_win_struct *jsmw)
{
	int i;

	if(jsmw == NULL)
	    return;

	OSWUnmapWindow(jsmw->toplevel);
	jsmw->map_state = False;
	jsmw->has_focus = False;

	WidgetUnmap(&jsmw->js_list);

	WidgetUnmap(&jsmw->add_btn);   
	WidgetUnmap(&jsmw->remove_btn);

	WidgetUnmap(&jsmw->device_prompt);
	WidgetUnmap(&jsmw->initialize_btn);

	WidgetUnmap(&jsmw->turn_axis_prompt);
	WidgetUnmap(&jsmw->throttle_axis_prompt);
	WidgetUnmap(&jsmw->thrust_dir_axis_prompt);
	WidgetUnmap(&jsmw->vs_zoom_axis_prompt);
	WidgetUnmap(&jsmw->scanner_zoom_axis_prompt);
	WidgetUnmap(&jsmw->aim_weapon_heading_prompt);

	WidgetUnmap(&jsmw->buttons_list);

	WidgetUnmap(&jsmw->scan_key_btn);
	jsmw->scanning_key = False;
	OSWUnmapWindow(jsmw->scan_key_win);

	WidgetUnmap(&jsmw->ok_btn);
	WidgetUnmap(&jsmw->apply_btn);
	WidgetUnmap(&jsmw->cancel_btn);

	/* Refetch the selected jsmap values so that current the changes
	 * do not get applied when the joysticks clist is cleared
	 */
	i = CListGetLastSelectedRow(&jsmw->js_list);
	if(i > -1)
	    JSMWFetchJSMap(jsmw, i);

	CListClear(&jsmw->js_list);
	CListClear(&jsmw->buttons_list);
}

/*
 *	Deletes the Joystick Map Window.
 */
void JSMWDelete(xsw_jsmap_win_struct *jsmw)
{
	if(jsmw == NULL)
	    return;

	CListClear(&jsmw->js_list);
	CListClear(&jsmw->buttons_list);

	WidgetDestroy(&jsmw->js_list);
	WidgetDestroy(&jsmw->add_btn);
	WidgetDestroy(&jsmw->remove_btn);
	WidgetDestroy(&jsmw->device_prompt);
	WidgetDestroy(&jsmw->initialize_btn);
	WidgetDestroy(&jsmw->turn_axis_prompt);
	WidgetDestroy(&jsmw->throttle_axis_prompt);
	WidgetDestroy(&jsmw->thrust_dir_axis_prompt);
	WidgetDestroy(&jsmw->vs_zoom_axis_prompt);
	WidgetDestroy(&jsmw->scanner_zoom_axis_prompt);
	WidgetDestroy(&jsmw->aim_weapon_heading_prompt);
	WidgetDestroy(&jsmw->buttons_list);
	WidgetDestroy(&jsmw->scan_key_btn);
	jsmw->scanning_key = False;
	OSWDestroyWindow(&jsmw->scan_key_win);
	WidgetDestroy(&jsmw->ok_btn);
	WidgetDestroy(&jsmw->apply_btn);
	WidgetDestroy(&jsmw->cancel_btn); 
	OSWDestroyWindow(&jsmw->toplevel);

	free(jsmw);
}

#endif	/* JS_SUPPORT */
