#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include "xsw.h"


void XSWLogFile(const char *path, const char *s);
void XSWLog(const char *s);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Writes the specified string to the log file specified by path.
 *
 *	The path can be NULL to indicate the default log file.
 */
void XSWLogFile(const char *path, const char *s)
{
	int len, bytes_written;
	FILE *fp;

	if(STRISEMPTY(path))
	    path = fname.log;
	if(STRISEMPTY(path) || STRISEMPTY(s))
	    return;

	/* Open or create the file for write-append */
	fp = fopen(path, "a");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(STRLEN(path) + STRLEN(error_msg) + 80);
		sprintf(
		    s,
"%s: %s.",
		    path, error_msg
		);
		XSWPrintWarning(s);
		free(s);
		free(error_msg);
	    }
	    return;
	}

	/* Append the line */
	len = strlen(s);
	bytes_written = fwrite(s, sizeof(char), len, fp);
	if(bytes_written < 0)
	{
	    const int error_code = errno;
            char *s, *error_msg = STRDUP(strerror(error_code));
            if(error_msg != NULL)
            {
                *error_msg = toupper(*error_msg);
                s = (char *)malloc(STRLEN(path) + STRLEN(error_msg) + 80);
                sprintf(
                    s,
"%s: %s.",
                    path, error_msg
                );
                XSWPrintWarning(s);
                free(s);
                free(error_msg);
            }
	}

	/* Close the file */
	fclose(fp);
}

/*
 *	Writes the specified string to the default log file.
 */
void XSWLog(const char *s)
{
	XSWLogFile(NULL, s);
}

