#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>  
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/prochandle.h"

#include "../include/swsoundcodes.h"

#include "../include/reality.h"
#include "../include/unvmain.h"
#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "keymaps.h"
#include "particle.h"
#include "xswimgrefs.h"
#include "ocsinfofio.h"
#include "imgrefsfio.h"

#include "gctl.h"
#include "sound.h"
#include "sndrefs.h"

#include "splash.h"
#include "bridgewin.h"
#include "vsmenu.h"
#include "msgwin.h"
#include "unvlist.h"
#include "ignoredobj.h"
#include "unvlistwin.h"
#include "unvpropdlg.h"
#include "unvlistfile.h"
#include "ecowin.h"
#include "starchartwin.h"
#include "optwin.h"
#include "keymapwin.h"
#ifdef JS_SUPPORT
# include "jsmaps.h"
# include "jsmapwin.h"
#endif
#include "vsmenuwin.h"
#include "xswinstall.h"
#include "cfgfile.h"
#include "net.h"
#include "xsw.h"

#include "ocsnames.h"	/* Must be after xsw.h for types to be defined */


/* Cursor Pixmaps */
#include "../include/cursors/xsw_scanner_lock.xpm"
#include "../include/cursors/xsw_text.xpm"
#include "../include/cursors/translate.xpm"
#include "../include/cursors/zoom.xpm"


win_t XSWGetWindowToplevel(void *win);

static void XSWSignalCB(int s);

mode_t XSWCurrentFileMode(void);
const char *XSWCurrentDirectory(void);

void XSWStartServer(const char *cmd);
void XSWOpenOCSNs(const char *path);
void XSWOpenImgRefs(const char *path);
void XSWOpenSndRefs(const char *path);

void XSWResetTimmers(
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static int XSW_LOAD_PIXEL_RGB(pixel_t *pix, WColorStruct c);
static int XSWLoadColors(void);
static void XSWFreeColors(void);

void XSWGetMemUsage(xsw_mem_stat_struct *buf);
void XSWReclaimMem(
	bool_t verbose,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static void XSWImgRefLoadCompletedCB(const int imgref_num, void *data);

int XSWManageGUI(
	void *data,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int XSWManageSound(void *data);
int XSWManageNet(void *data);

int XSWInit(int argc, char *argv[]);
void XSWMainIteration(void *data);
void XSWMain(void *data);
void XSWMainPop(void *data);
void XSWMainPopAll(void *data);
void XSWShutdown(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int		runlevel;

time_t		cur_millitime;
time_t		cur_systime;
time_t		lapsed_millitime;
float		time_compensation;
int		prompt_mode;

xsw_debug_struct	debug;
xsw_option_struct	option;
xsw_dname_struct	dname;
xsw_fname_struct	fname;
xsw_next_struct		next;

xsw_fps_counter_struct	fps_counter;
xsw_anim_timmer_struct	xsw_anim_timmer[XSW_MAX_ANIM_TIMMERS];

xsw_font_struct		xsw_font;
xsw_color_struct	xsw_color;
xsw_cursor_struct	xsw_cursor;

xsw_image_struct	**xsw_image;
int			total_images;

xsw_mesg_struct		*mesg_list;
int			total_messages;

unvitem_struct	**univ_item;
int		total_univ_items;

ignored_obj_struct	**ignored_obj_list;
int			total_ignored_objs;

imgref_struct	**xsw_imgref;
int		total_imgrefs;

xsw_object_struct	**xsw_object;
int			total_objects;

xsw_particle_struct	**xsw_particle;
int			total_particles;

xsw_object_struct	**inrange_xsw_object;
int			total_inrange_objects;

vs_object_label_struct	**vs_object_label;
int			total_vs_object_labels;

#ifdef JS_SUPPORT
xsw_jsmap_struct	**xsw_jsmap;
int			total_jsmaps;
#endif	/* JS_SUPPORT */

xsw_gctl_struct		*xsw_gctl;
xsw_player_values_struct	player_values;

sw_legend_struct	sw_legend;

scanner_contact_struct	**scanner_contact_list;
int			total_scanner_contacts;

xsw_ocsn_struct	**xsw_ocsn;
int		total_ocsns;

xsw_net_parms_struct	net_parms;
xsw_net_stat_struct	net_stat;

xsw_auto_interval_tune_struct	auto_interval_tune;

serv_script_struct	**serv_script;
int			total_serv_scripts;

xsw_sound_struct	sound;
sndref_struct	**xsw_sndref;
int		total_sndrefs;

xsw_bridge_win_struct	*xsw_bridge_win;
xsw_vsmenu_struct	*xsw_vsmenu;
xsw_msg_win_struct	*xsw_msg_win;
xsw_unvlist_win_struct	*xsw_unvlist_win;
xsw_unvprop_dlg_struct	*xsw_unvprop_dlg;
xsw_eco_win_struct	*xsw_ecow;
starchart_win_struct	*xsw_scht;

xsw_options_win_struct	*xsw_optwin;
xsw_keymap_win_struct	*xsw_keymap_win;
#ifdef JS_SUPPORT
xsw_jsmap_win_struct	*xsw_jsmap_win;
#endif
xsw_vsmenu_win_struct	*xsw_vsmenu_win;

dialog_struct	*xsw_dlg;
cdlg_struct	*xsw_cdlg;

fb_struct	*xsw_file_selector;


/*
 *	Returns the toplevel window for the specified window
 *	structure.
 */
win_t XSWGetWindowToplevel(void *win)
{
	if(win == NULL)
	    return(None);
	else if(win == xsw_bridge_win)
	    return(xsw_bridge_win->toplevel);
	else if(win == xsw_msg_win)
	    return(xsw_msg_win->toplevel);
	else if(win == (void *)xsw_unvlist_win)
	    return(xsw_unvlist_win->toplevel);
	else if(win == (void *)xsw_unvprop_dlg)
	    return(xsw_unvprop_dlg->toplevel);
	else if(win == (void *)xsw_ecow)
	    return(xsw_ecow->toplevel);
	else if(win == (void *)xsw_scht)
	    return(xsw_scht->toplevel);
	else if(win == (void *)xsw_optwin)
	    return(xsw_optwin->toplevel);
	else if(win == (void *)xsw_vsmenu_win)
	    return(xsw_vsmenu_win->toplevel);
	else
	    return(None);
}


/*
 *	Signal callback
 */
static void XSWSignalCB(int s)
{
	switch(s)
	{
	  case SIGHUP:
/* TODO reload configuration */

	    XSWRefresh(cur_millitime, cur_systime, time_compensation);
	    break;

	  case SIGINT:
	    XSWMainPopAll(NULL);
	    break;

	  case SIGQUIT:
	    XSWMainPopAll(NULL);
	    break;

	  case SIGABRT:
	    XSWMainPopAll(NULL);
	    break;

	  case SIGKILL:
	    XSWMainPopAll(NULL);
	    break;

	  case SIGPIPE:
	    break;

	  case SIGSEGV:
	    XSWPrintDebugStderr(1);
	    exit(1);
	    break;

	  case SIGTERM:
	    XSWMainPopAll(NULL);
	    break;

	  case SIGSTOP:
	    break;

	  case SIGCONT:
	    break;

	  default:
	    if(s > 0)
	    {
		char *msg = (char *)malloc(40 + 80);
		sprintf(
		    msg,
"Received unknown signal `%i'.\n",
		    s
		);
		XSWPrintWarning(msg);
		free(msg);
	    }
	    break;
	}
}


/*
 *	Returns the current permissions for files.
 */
mode_t XSWCurrentFileMode(void)
{
        const mode_t m = umask(0);
        umask(m);
        return(
            (~m) &
            (S_IRUSR | S_IWUSR |
             S_IRGRP | S_IWGRP |
             S_IROTH | S_IWOTH)
        );
}

/*
 *	Returns the current working directory.
 *
 *	The returned string must not be modified or deleted.
 */
const char *XSWCurrentDirectory(void)
{
        static char path[PATH_MAX];
        return(getcwd(path, sizeof(path)));
}


/*
 *	Runs the universe server.
 */
void XSWStartServer(const char *cmd)
{
	pid_t	p;
#ifdef __WIN32__
	p = Exec(".\\swserv.exe");
#else
	char	*s, *cwd,
		new_cwd[PATH_MAX];

	if(cmd == NULL)
	    return;

	/* Path must be executeable */
	if(!ISPATHEXECUTABLE(cmd))
	{
	    char *s = (char *)malloc(STRLEN(cmd) + 80 + 80);
	    sprintf(
		s,
"Unable to execute:\n\
\n\
    %s\n\
\n\
Program or script is not set executeable.",
		cmd
	    );
	    XSWDialog(
		"Execute Failed",
		XSW_IMG_DIALOG_ICON_WARNING,
		s,
		xsw_bridge_win->toplevel
	    );
	    free(s);
	    return;
	}

	/* Path must be absolute */
	if(!ISPATHABSOLUTE(cmd))
	{
	    char *s = (char *)malloc(STRLEN(cmd) + 80 + 80);
	    sprintf(
		s,
"Unable to execute:\n\
\n\
    %s\n\
\n\
Path not absolute.",
		cmd
	    );
	    XSWDialog(
		"Execute Failed", 
		XSW_IMG_DIALOG_ICON_WARNING,
		s,
		xsw_bridge_win->toplevel
	    );
	    free(s);
	    return;
	}

	/* ************************************************** */

	/* Record current working dir */
	cwd = STRDUP(XSWCurrentDirectory());

	/* Set new working dir */
	s = GetParentDir(cmd);
	if(s == NULL)
	{
	    free(cwd);
	    return;
	}

	strncpy(new_cwd, s, sizeof(new_cwd));
	new_cwd[sizeof(new_cwd) - 1] = '\0';
	chdir(new_cwd);


	/* Run server in background */
	p = Exec(cmd);


	/* Change back to previous working dir */
	chdir(cwd);
	free(cwd);

#endif
}

/*
 *	Opens the OCS names from the specified file.
 */
void XSWOpenOCSNs(const char *path)
{
	int i, total;
	char *full_path;
	ocs_info_struct **list;
	xsw_net_parms_struct *net = &net_parms;

	if(STRISEMPTY(path))
	    return;

	full_path = ISPATHABSOLUTE(path) ?
	    STRDUP(path) :
	    STRDUP(PrefixPaths(dname.etc, path));

	/* Delete all the existing OCS Names */
	OCSNClear(&xsw_ocsn, &total_ocsns);

	/* Open the OCS Names from file */
	list = NULL;
	total = 0;
	if(!OCSInfoFileOpen(
	    &list, &total,
	    full_path,
	    NULL, NULL,
	    XSWPrintWarning
	))
	{
	    int ocsn_num;
	    ocs_info_struct *ocs_info;
	    xsw_ocsn_struct *ocsn;

	    /* Parse the loaded OCS Info list and convert each OCS
	     * Info to an OCS Name
	     */
	    for(i = 0; i < total; i++)
	    {
		ocs_info = list[i];
		if(ocs_info == NULL)
		    continue;

		/* Create a new OCS Name for this OCS Info */
		ocsn_num = MAX(total_ocsns, 0);
		total_ocsns = ocsn_num + 1;
		xsw_ocsn = (xsw_ocsn_struct **)realloc(
		    xsw_ocsn,
		    total_ocsns * sizeof(xsw_ocsn_struct *)
		);
		xsw_ocsn[ocsn_num] = ocsn = XSW_OCSN(calloc(
		    1, sizeof(xsw_ocsn_struct)
		));
		if(ocsn == NULL)
		    break;

		ocsn->code = ocs_info->code;
		ocsn->name = STRDUP(ocs_info->name);
		if(!STRISEMPTY(ocs_info->icon_file))
		{
		    char *icon_file = ISPATHABSOLUTE(ocs_info->icon_file) ?
			STRDUP(ocs_info->icon_file) :
			STRDUP(PrefixPaths(
			    dname.images, ocs_info->icon_file
			));
		    ocsn->icon = WidgetLoadImageFromTgaFile(icon_file);
		    if(ocsn->icon == NULL)
		    {
			char *s = (char *)malloc(STRLEN(icon_file) + 80);
			sprintf(
			    s,
"%s: Unable to open image.",
			    icon_file
			);
			XSWPrintWarning(s);
			free(s);
		    }
		    free(icon_file);
		}


	    }

	    strncpy(fname.ocsn, full_path, sizeof(fname.ocsn));
	    fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';

	    /* Updated resources that would be affected by this change */

	    /* Remove weapons from player object so they get recreated */ 	if(!DBIsObjectGarbage(net_parms.player_obj_num))
	    UNVAllocObjectWeapons(
		net->player_obj_ptr,
		0 
	    );

	    /* Recreate the selected weapon viewscreen label */
	    VSDrawUpdateWeaponLabel(
		&xsw_bridge_win->vs_weapon_image,
		xsw_bridge_win->vs_weapon_buf
	    );
	}

	/* Delete all the loaded OCS Infos */
	for(i = 0; i < total; i++)
	    OCSInfoDelete(list[i]);
	free(list);

	free(full_path);
}

/*
 *	Opens the image references from file.
 */
void XSWOpenImgRefs(const char *path)
{
	char *full_path;

	if(STRISEMPTY(path))
	    return;

	full_path = ISPATHABSOLUTE(path) ?
	    STRDUP(path) :
	    STRDUP(PrefixPaths(dname.images, path));

	/* Delete all the existing OCS Names */
	ImgRefClear(&xsw_imgref, &total_imgrefs);

	/* Open the ImgRefs file */
	if(!ImgRefsFileOpen(
	    &xsw_imgref, &total_imgrefs,
	    full_path,
	    NULL, NULL,
	    XSWPrintWarning
	))
	{
	    int i;
	    imgref_struct *imgref;

	    /* If the default imgref was not defined in the imgrefs
	     * file then create one ourselves
	     */
	    if(ImgRefGetPtr(xsw_imgref, total_imgrefs, IMGREF_DEFAULT_IMGREF) == NULL)
		ImgRefLoadAsDefault(
		    &xsw_imgref, &total_imgrefs,
		    IMGREF_DEFAULT_IMGREF
		);

	    /* Iterate through the loaded imgrefs */
	    for(i = 0; i < total_imgrefs; i++)
	    {
		imgref = xsw_imgref[i];
		if(imgref == NULL)
		    continue;

		/* Complete the filename as needed */
		if(!STRISEMPTY(imgref->path))
		{
		    struct stat stat_buf;

		    if(!ISPATHABSOLUTE(imgref->path))
		    {
			char *full_path = STRDUP(PrefixPaths(
			    dname.images, imgref->path
			));
			free(imgref->path);
			imgref->path = full_path;
		    }

		    /* Check if the file does not exist? */
		    if((imgref->path != NULL) ?
		       stat(imgref->path, &stat_buf) : True
		    )
		    {
			const int error_code = errno;
			char *s, *error_msg = STRDUP(strerror(error_code));
			if(error_msg != NULL)
			{
			    *error_msg = toupper(*error_msg);
			    s = (char *)malloc(
				STRLEN(imgref->path) +
				STRLEN(error_msg) + 80
			    );
			    sprintf(
				s, "%s: %s",
				imgref->path, error_msg
			    );
			    XSWPrintWarning(s);
			    free(s);
			}
			free(error_msg);

			continue;	/* Skip the rest of this imgref */
		    }
		}

		/* Need to load this imgref's image when the imgrefs
		 * file was opened?
		 */
		if(imgref->options & IMGREF_STAY_LOADED)
		{
		    if(ImgRefLoad(&xsw_imgref, &total_imgrefs, i))
		    { 
			char *s = (char *)malloc(STRLEN(imgref->path) + 80);
			sprintf(
			    s,
"%s: Unable to open image.",
			    imgref->path
			);
			XSWPrintWarning(s);
			free(s);
		    }
		}

	    }

	    strncpy(fname.imgrefs, full_path, sizeof(fname.imgrefs));
	    fname.imgrefs[sizeof(fname.imgrefs) - 1] = '\0';
	}

	free(full_path);
}

/*
 *	Opens the sound references from file.
 */
void XSWOpenSndRefs(const char *path)
{
	char *full_path, *audio_mode_name = NULL;
	xsw_option_struct *opt = &option;

	if(path == NULL)
	    return;

	full_path = ISPATHABSOLUTE(path) ?
	    STRDUP(path) :
	    STRDUP(PrefixPaths(dname.sounds, path));

	/* Open the SndRefs from file */
	if(!SndRefsFileOpen(
	    &xsw_sndref, &total_sndrefs,
	    full_path,
	    &audio_mode_name,
	    NULL, NULL,
	    XSWPrintWarning
	))
	{
	    int i;
	    sndref_struct *sndref;

	    /* Audio mode specified? */
	    if(!STRISEMPTY(audio_mode_name))
	    {
		char *s;

		/* Switch to the specified Audio mode */
		switch(SoundChangeMode(&sound, audio_mode_name))
		{
		  case -1:
		  case -2:
		    s = (char *)malloc(
			STRLEN(audio_mode_name) + 80
		    );
		    sprintf(
			s,
"Unable to switch to Audio mode `%s'.",
			audio_mode_name
		    );
		    XSWPrintWarning(s);
		    free(s);
		    break;
		}
	    }

	    /* Iterate through the sndrefs list and complete any
	     * relative paths
	     */
	    for(i = 0; i < total_sndrefs; i++)
	    {
		sndref = xsw_sndref[i];
		if(sndref == NULL)
		    continue;

		if(STRISEMPTY(sndref->path))
		    continue;

		if(!ISPATHABSOLUTE(sndref->path))
		{
		    struct stat stat_buf;

		    char *full_path = STRDUP(PrefixPaths(
			dname.sounds, sndref->path
		    ));
		    free(sndref->path);
		    sndref->path = full_path;

		    /* Check if the file does not exist? */
		    if((full_path != NULL) ?
			stat(full_path, &stat_buf) : True
		    )
		    {
			const int error_code = errno;
			char *s, *error_msg = STRDUP(strerror(error_code));
			if(error_msg != NULL)
			{
			    *error_msg = toupper(*error_msg);
			    s = (char *)malloc(
				STRLEN(full_path) +  
				STRLEN(error_msg) + 80
			    );
			    sprintf(
				s, "%s: %s",
				full_path, error_msg
			    );
			    XSWPrintWarning(s);
			    free(s);
			}
			free(error_msg);

			continue;	/* Skip the rest of this sndref */
		    }
		}
	    }

	    /* Update the looping sounds as needed */
	    if(opt->sounds >= XSW_SOUNDS_ENGINE)
		XSWUpdateLoopSounds();

	    /* Start the background music as needed */
	    if(opt->music)
		XSWUpdateBackgroundMusic();

	    strncpy(fname.sndrefs, full_path, sizeof(fname.sndrefs));
	    fname.sndrefs[sizeof(fname.sndrefs) - 1] = '\0';
	}
	free(audio_mode_name);

	free(full_path);
}


/*
 *	Resets all timmers to the specified time values.
 */
void XSWResetTimmers(
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, n, m;
	xsw_weapon_struct *wep;
	xsw_object_struct *obj;
	imgref_struct *imgref_ptr;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	/* Next instances */
	next.update_check = cur_ms;
	next.viewscreen = cur_ms;
	next.consoles = cur_ms;
	next.memory_clean = cur_ms;
	next.player_pos_send = cur_ms;
	next.spawn_particles = cur_ms;
	next.wormhole_enter = cur_ms;

	/* Bridge window viewscreen redraw interval
	 *
	 * This may need to be reset because otherwise with a current
	 * time calculation may set this value too high
	 */
	if(bw != NULL)
	{
	    bw->viewscreen_int = 0l;
	}

	/* Animation Timmers */
	m = (int)(sizeof(xsw_anim_timmer) / sizeof(xsw_anim_timmer_struct));
	for(i = 0; i < m; i++)
	    xsw_anim_timmer[i].next = cur_ms;

	/* Auto Interval Tunning */
	auto_interval_tune.next = cur_ms;

	/* FPS Timmer */
	fps_counter.fcount = 0;
	fps_counter.lfcount = 0;
	fps_counter.next = cur_ms;


	/* Objects */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    obj->last_updated = cur_ms;
	    if(obj->birth_time > 0l)
		obj->birth_time = cur_ms;
	    if(obj->last_cloak > 0l)
		obj->last_cloak = cur_ms;
	    obj->animation.last_interval = cur_ms;

	    /* Auto destruct needs to be unset since there is no way
	     * to calculate it after the timmers being reset
	     */
	    obj->auto_destruct_next = 0l;

	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;

		wep->last_used = 0l;
	    }
	}

	/* ImgRefs */
	for(i = 0; i < total_imgrefs; i++)
	{
	    imgref_ptr = xsw_imgref[i];
	    if(imgref_ptr == NULL)
		continue;

	    /* Strobe timmer on point lights */
	    if(imgref_ptr->nlights > 0)
	    {
		const int m = imgref_ptr->nlights;
		int n;
		imgref_light_struct *light;
		for(n = 0; n < m; n++)
		{
		    light = imgref_ptr->light[n];
		    if(light == NULL)
			continue;

		    light->strobe_next = cur_ms;
		}
	    }
	}
}


/*
 *	Macro used by XSWLoadColors() to load colors.
 */
static int XSW_LOAD_PIXEL_RGB(pixel_t *pix, WColorStruct c)
{
	return(OSWLoadPixelRGB(pix, c.r, c.g, c.b));
}

/*
 *      Procedure to load all GUI pixel colors used by this program.
 */
int XSWLoadColors(void)
{
#define CS_TREF xsw_color

	/* Standard text colors */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.standard_text,
	    CS_TREF.standard_text_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bold_text,
	    CS_TREF.bold_text_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.withdrawn_text,
	    CS_TREF.withdrawn_text_cv
	);

	/* Bridge console panel readout colors */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_standard_text,
	    CS_TREF.bp_standard_text_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_bold_text,
	    CS_TREF.bp_bold_text_cv 
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_withdrawn_text,
	    CS_TREF.bp_withdrawn_text_cv
	);

	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_light_outline,
	    CS_TREF.bp_light_outline_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_normal_outline,
	    CS_TREF.bp_normal_outline_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_dark_outline,
	    CS_TREF.bp_dark_outline_cv
	);

	/* Warning, danger, and critical colors */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_warning,
	    CS_TREF.bp_warning_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_danger,
	    CS_TREF.bp_danger_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_critical,
	    CS_TREF.bp_critical_cv
	);

	/* Friendly, unknown, and hostile IFF colors */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_friendly,
	    CS_TREF.bp_friendly_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_unknown,
	    CS_TREF.bp_unknown_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bp_hostile,
	    CS_TREF.bp_hostile_cv
	);


	/* Bridge console panel readout outline colors */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bpol_hull,
	    CS_TREF.bpol_hull_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bpol_power,
	    CS_TREF.bpol_power_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bpol_vis,
	    CS_TREF.bpol_vis_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bpol_shields,
	    CS_TREF.bpol_shields_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bpol_dmgctl,
	    CS_TREF.bpol_dmgctl_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bpol_throttle,
	    CS_TREF.bpol_throttle_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.bpol_throttle_rev,
	    CS_TREF.bpol_throttle_rev_cv
	);

	/* Scanner marks */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.scmark_unknown,
	    CS_TREF.scmark_unknown_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.scmark_locked,
	    CS_TREF.scmark_locked_cv 
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.scmark_weapon,
	    CS_TREF.scmark_weapon_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.scmark_home, 
	    CS_TREF.scmark_home_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.scmark_area, 
	    CS_TREF.scmark_area_cv
	);


	/* Keymap edit window */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.keymap_query_bg,
	    CS_TREF.keymap_query_bg_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.keymap_query_fg,
	    CS_TREF.keymap_query_fg_cv
	);

	/* Starchart */
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.chart_bg,
	    CS_TREF.chart_bg_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.chart_grid,
	    CS_TREF.chart_grid_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.chart_sector_grid,
	    CS_TREF.chart_sector_grid_cv
	);
	XSW_LOAD_PIXEL_RGB(
	    &CS_TREF.chart_cross_hairs,
	    CS_TREF.chart_cross_hairs_cv
	);


	return(0);
#undef CS_TREF
}

/*
 *	Procedure to free all GUI pixel colors used by this program.
 */
static void XSWFreeColors(void)
{
	if(!IDC())
	    return;

	OSWDestroyPixel(&xsw_color.standard_text);
	OSWDestroyPixel(&xsw_color.bold_text);
	OSWDestroyPixel(&xsw_color.withdrawn_text);

	OSWDestroyPixel(&xsw_color.bp_standard_text);
	OSWDestroyPixel(&xsw_color.bp_bold_text);
	OSWDestroyPixel(&xsw_color.bp_withdrawn_text);

	OSWDestroyPixel(&xsw_color.bp_light_outline);
	OSWDestroyPixel(&xsw_color.bp_normal_outline);
	OSWDestroyPixel(&xsw_color.bp_dark_outline);

	OSWDestroyPixel(&xsw_color.bp_warning);
	OSWDestroyPixel(&xsw_color.bp_danger);
	OSWDestroyPixel(&xsw_color.bp_critical);

	OSWDestroyPixel(&xsw_color.bp_friendly);
	OSWDestroyPixel(&xsw_color.bp_unknown);
	OSWDestroyPixel(&xsw_color.bp_hostile);

	OSWDestroyPixel(&xsw_color.bpol_hull);
	OSWDestroyPixel(&xsw_color.bpol_power);
	OSWDestroyPixel(&xsw_color.bpol_vis);
	OSWDestroyPixel(&xsw_color.bpol_shields);
	OSWDestroyPixel(&xsw_color.bpol_dmgctl);
	OSWDestroyPixel(&xsw_color.bpol_throttle);
	OSWDestroyPixel(&xsw_color.bpol_throttle_rev);

	OSWDestroyPixel(&xsw_color.scmark_unknown);
	OSWDestroyPixel(&xsw_color.scmark_locked);
	OSWDestroyPixel(&xsw_color.scmark_weapon);
	OSWDestroyPixel(&xsw_color.scmark_home);
	OSWDestroyPixel(&xsw_color.scmark_area);

	OSWDestroyPixel(&xsw_color.keymap_query_bg);
	OSWDestroyPixel(&xsw_color.keymap_query_fg);

	OSWDestroyPixel(&xsw_color.chart_bg);
	OSWDestroyPixel(&xsw_color.chart_grid);
	OSWDestroyPixel(&xsw_color.chart_sector_grid);
	OSWDestroyPixel(&xsw_color.chart_cross_hairs);

	return;
}


/*
 *	Get dynamically allocated memory statistics used by
 *	this program.
 */
void XSWGetMemUsage(xsw_mem_stat_struct *buf)
{
	int i;
	long subtotal = 0;

	int events;
	xsw_image_struct **img_ptr;
	unvitem_struct **uli_ptr;
	xsw_object_struct *obj;
	imgref_struct *imgref;
	xsw_ocsn_struct *ocsn;
	vs_object_label_struct **vsol_ptr;
	scanner_contact_struct *sc;
	osw_gui_struct *gui = &osw_gui[0];

	if(buf == NULL)
	    return;

	/* GUI Events */
	if(IDC())
	{
	    events = OSWEventsPending();
	    buf->gui = (events * sizeof(event_t *)) +
		(events * sizeof(event_t));
	}
	else
	{
	    buf->gui = 0;
	}
	subtotal += buf->gui;


	/* Sound events */
	buf->sound = 0;
	subtotal += buf->sound;


	/* Program images */
	for(buf->images = 0, i = 0, img_ptr = xsw_image;
	    i < total_images;
	    i++, img_ptr++
	)
	{
	    if(*img_ptr == NULL)
		continue;

	    if((*img_ptr)->filename != NULL)
		buf->images += strlen((*img_ptr)->filename) + 1;

	    if((*img_ptr)->image != NULL)
	    {
		image_t *img = (*img_ptr)->image;
		buf->images += (int)(img->bytes_per_line * img->height);
	    }

	    buf->images += sizeof(xsw_image_struct);
	}
	buf->images += total_images * sizeof(xsw_image_struct *);

	subtotal += buf->images;


	/* Universe entries */
	for(buf->univ_entries = 0, i = 0, uli_ptr = univ_item;
	    i < total_univ_items;
	    i++, uli_ptr++
	)
	{
	    if(*uli_ptr == NULL)
		continue;

	    if((*uli_ptr)->alias != NULL)
	    {
		buf->univ_entries += strlen((*uli_ptr)->alias) + 1;
	    }
	    if((*uli_ptr)->url != NULL)
	    {
		buf->univ_entries += strlen((*uli_ptr)->url) + 1;
	    }
	    if((*uli_ptr)->comments != NULL)
	    {
		buf->univ_entries += strlen((*uli_ptr)->comments) + 1;
	    }
	    buf->univ_entries += sizeof(unvitem_struct);
	}

	buf->univ_entries += total_univ_items * sizeof(unvitem_struct *);
	subtotal += buf->univ_entries;


	/* XSW Objects */
	buf->objects = 0;
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->description != NULL)
		buf->objects += (strlen(obj->description) + 1) *
		    sizeof(char);

	    if(obj->elink != NULL)
		buf->objects += (strlen(obj->elink) + 1) *
		    sizeof(char);

	    if(obj->weapons != NULL)
	    {
		buf->objects += (obj->total_weapons *
		    sizeof(xsw_weapon_struct *)) +
		    (obj->total_weapons *
		    sizeof(xsw_weapon_struct));
	    }
	    if(obj->tractored_object != NULL)
	    {
		buf->objects += obj->total_tractored_objects *
		    sizeof(int);
	    }
	    if(obj->score != NULL)
	    {
		buf->objects += sizeof(xsw_score_struct);
	    }
	    if(obj->eco != NULL)
	    {
		buf->objects += sizeof(xsw_ecodata_struct);
	    }
	    buf->objects += sizeof(xsw_object_struct);
	}

	buf->objects += total_objects * sizeof(xsw_object_struct *);
	subtotal += buf->objects;


	/* Image set referances */
	for(buf->imgrefs = 0, i = 0; i < total_imgrefs; i++)
	{
	    imgref = xsw_imgref[i];
	    if(imgref == NULL)
		continue;

	    if(imgref->path != NULL)
	    {
		buf->imgrefs += strlen(imgref->path) + 1;
	    }
	    if(imgref->lib_data != NULL)
	    {
		/* Library data not NULL implies image data is unique
		 * to this imgref (not shared).
		 */
/* Need to calculate image data? */
		if(imgref->image_data != NULL)
		    buf->imgrefs += (
			imgref->width * imgref->height * 
		        MAX(gui->depth / 8, 1)
		    );
	    }
	    if(imgref->light != NULL)
	    {
		buf->imgrefs += (imgref->nlights *
		    sizeof(imgref_light_struct *)) +
		    (imgref->nlights *
		    sizeof(imgref_light_struct));
	    }

	    buf->imgrefs += sizeof(imgref_struct);
	}
	buf->imgrefs += total_imgrefs * sizeof(imgref_struct *);
	subtotal += buf->imgrefs;


	/* OCS Names */
	buf->ocsns = 0;
	for(i = 0; i < total_ocsns; i++)
	{
	    ocsn = xsw_ocsn[i];
	    if(ocsn == NULL)
		continue;

	    buf->ocsns += sizeof(xsw_ocsn_struct);
	    buf->ocsns += STRLEN(ocsn->name);
	    if(ocsn->icon != NULL)
	    {
		image_t *img = ocsn->icon;
		buf->ocsns += sizeof(image_t);
		buf->ocsns += img->bytes_per_line * img->height;
	    }
	}
	buf->ocsns += total_ocsns * sizeof(xsw_ocsn_struct *);
	subtotal += buf->ocsns;


	/* Viewscreen object labels */
	for(buf->vs_labels = 0, i = 0, vsol_ptr = vs_object_label;
	    i < total_vs_object_labels;
	    i++, vsol_ptr++
	)
	{
	    if(*vsol_ptr == NULL)
		continue;

	    if((*vsol_ptr)->image != NULL)
	    {
		image_t *img = (*vsol_ptr)->image;
		buf->vs_labels += (int)(img->bytes_per_line * img->height);
	    }
	    buf->vs_labels += sizeof(vs_object_label_struct);
	}
	buf->vs_labels += total_vs_object_labels *
	    sizeof(vs_object_label_struct *);
	subtotal += buf->vs_labels;


	/* Scanner contacts */
	buf->scanner_contacts = 0;
	for(i = 0; i < total_scanner_contacts; i++)
	{
	    sc = scanner_contact_list[i];
	    if(sc == NULL)
		continue;

	    buf->scanner_contacts += sizeof(scanner_contact_struct);
	}
	buf->scanner_contacts += total_scanner_contacts *
	    sizeof(scanner_contact_struct *);
	subtotal += buf->scanner_contacts;


	/* Total */
	buf->total = subtotal;
}


/*
 *	Refreshes used memory, updating listings and deleting out
 *	dated or unused resources (ie objects).
 */
void XSWReclaimMem(
	bool_t verbose,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	if(verbose) 
	    XSWPrint(
"Client: Reclaiming memory..."
	    );

	DBReclaim(cur_ms, cur_sec, tc);
	ImgRefReclaimMemory(
	    &xsw_imgref, &total_imgrefs,
	    xsw_object, total_objects,
	    xsw_particle, total_particles
	);
	VSLabelReclaimMemory();
	DBInRangeUpdate(
	    net_parms.player_obj_num,
	    cur_ms, cur_sec, tc
	);

	if(verbose)
	    XSWPrint(
"Client: Memory reclaim complete."
	    );

	/* Mark next time to reclaim memory */
	next.memory_clean = cur_ms + XSW_MEMORY_CLEAN_INTERVAL;
}


/*
 *	ImgRef load completed callback.
 */
static void XSWImgRefLoadCompletedCB(const int imgref_num, void *data)
{
	xsw_net_parms_struct *net = &net_parms;
	xsw_object_struct *obj = net->player_obj_ptr;

	/* Redraw bridge window console panels as needed */
	if(obj != NULL)
	{
	    const int locked_obj_num = obj->locked_on;

	    if(obj->imgref == imgref_num)
		BridgeWinDrawPanel(net->player_obj_num, BW_PANEL_P2);

	    if(!DBIsObjectGarbage(locked_obj_num))
	    {
		xsw_object_struct *locked_obj = xsw_object[locked_obj_num];
		if(locked_obj->imgref == imgref_num)
		    BridgeWinDrawPanel(locked_obj_num, BW_PANEL_S1);
	    }
	}
}


/*
 *	Manages GUI events.
 */
int XSWManageGUI(
	void *data,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	bool_t need_continue;
	int events_handled, total_events_handled = 0;
	event_t event;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	/* Manage GUI events */
	while(OSWEventsPending() > 0)
	{
	    need_continue = False;
	    events_handled = 0;

	    /* Fetch new event */
	    OSWWaitNextEvent(&event);

	    /* Send KeyPress and KeyRelease events to the game
	     * controller if the bridge window is in focus and its
	     * prompt is not set
	     */
	    if(((bw != NULL) ? bw->has_focus : False) &&
	       (prompt_mode == XSW_PROMPT_NONE) &&
	       ((event.type == KeyPress) || (event.type == KeyRelease))
	    )
	    {
		events_handled += GCtlManageGUIKeyEvent(
		    xsw_gctl, &event,
		    xsw_bridge_win
		);
		if(events_handled > 0)
		{
		    total_events_handled += events_handled;
		    break;
		}
	    }

	    /* Handle event by type */
	    switch(event.type)
	    {
	      case FocusIn:
		break;

	      case FocusOut:
		if((bw != NULL) ? bw->has_focus : False)
		{
		    GCtlResetKeys(xsw_gctl);
		}
		break;
	    }

	    if(need_continue)
		continue;

	    /* Manage widget event */
	    if(events_handled == 0)
		events_handled += WidgetManageGUI(&event);

	    /* Bridge */
	    if(events_handled == 0)
	        events_handled += BridgeManage(
		    bw, &event,
		    cur_ms, cur_sec, tc
		);

	    /* Message Window */
	    if(events_handled == 0)
		events_handled += MsgWinManage(
		    xsw_msg_win, &event,
		    cur_ms, cur_sec, tc
		);

	    /* Universe list window */
	    if(events_handled == 0)
		events_handled += UnvListWinManage(xsw_unvlist_win, &event);

	    /* Universe Edit Dialog */
	    if(events_handled == 0)
		events_handled += UnvPropDlgManage(xsw_unvprop_dlg, &event);

	    /* Economy Window */
	    if(events_handled == 0)
		events_handled += EcoWinManage(xsw_ecow, &event);

	    /* Starchart window */
	    if(events_handled == 0)
		events_handled += SChtManage(
		    xsw_scht, &event,
		    cur_ms, cur_sec, tc
		);

	    /* Options Window */
	    if(events_handled == 0)
		events_handled += OptWinManage(xsw_optwin, &event);

	    /* Keymap Window */
	    if(events_handled == 0)
		events_handled += KeymapWinManage(xsw_keymap_win, &event);

#ifdef JS_SUPPORT
	    /* Joystick Map Window */
	    if(events_handled == 0)
		events_handled += JSMWManage(xsw_jsmap_win, &event);
#endif	/* JS_SUPPORT */

	    /* VSMenu Configuration Window */
	    if(events_handled == 0)  
		events_handled += VSMWManage(xsw_vsmenu_win, &event);

	    /* Dialog */
	    if(events_handled == 0)
		events_handled += WidgetManage(xsw_dlg, &event);

	    /* Confirmation Dialog */
	    if(events_handled == 0)
		events_handled += CDlgManage(xsw_cdlg, &event);

	    /* File Selector */
	    if(events_handled == 0)
		events_handled += WidgetManage(xsw_file_selector, &event);


	    /* Add up total events handled so far */
	    total_events_handled += events_handled;
	}

	/* Process queued widget events */
	WidgetMainIteration(NULL);

	return(total_events_handled);
}

/*
 *	Manage sound server events.
 */
int XSWManageSound(void *data)
{
	int total_events_handled = 0;
	xsw_sound_struct *snd = &sound;
	xsw_option_struct *opt = &option;

	/* Manage sound events */
	total_events_handled += SoundManage(snd);

	/* Sound shut down during management? */
	if((opt->sounds > XSW_SOUNDS_NONE) &&
	   (snd->server_type == SOUND_SERVER_TYPE_NONE)
	)
	{
	    /* Turn off sound level and music */
	    opt->sounds = XSW_SOUNDS_NONE;
	    opt->music = False;

	    XSWPrintWarning(
"Sound has been shut down unexpectedly."
	    );
	}

	return(total_events_handled);
}


/*
 *	Manage network.
 */
int XSWManageNet(void *data)
{
	const time_t cur_sec = cur_systime;
	int total_events_handled = 0;
	xsw_net_parms_struct *net = &net_parms;

	/* Manage by connection state */
	switch(net->connection_state)
	{
	  case XSW_CON_STATE_NOT_CONNECTED:
	    break;

	  case XSW_CON_STATE_NEGOTIATING:
	    /* Warn if logging in is taking longer than it should */
	    if(((net->con_start + 15l) <= cur_sec) &&
	       (net->login_delay_warning_last < 15l)
	    )
	    {
		XSWPrint(
"Waited 15 seconds for server response, still waiting..."
		);
		net->login_delay_warning_last = 15l;
	    }
	    else if(((net->con_start + 30l) <= cur_sec) &&
	            (net->login_delay_warning_last < 30l)
	    )
	    {
		XSWPrint(
"Waited 30 seconds for server response, still waiting..."
		);
		net->login_delay_warning_last = 30l;
	    }
	    else if(((net->con_start + 45l) <= cur_sec) &&
		    (net->login_delay_warning_last < 45l)
	    )
	    {
		const keycode_t keycode = KeymapGetKeycode(XSW_KM_DISCONNECT);
		char *arg = STRDUP(OSWGetKeyCodeName(keycode));
		char *s = (char *)malloc(STRLEN(arg) + 80);
		strtoupper(arg);
		sprintf(
		    s,
"Waited 45 seconds (press %s a few times to give up)...",
		    arg
		);
		free(arg);
		XSWPrint(s);
		free(s);
		net->login_delay_warning_last = 45l;
	    }

	    /* Handle incoming data for login, version, or
	     * disconnect
	     */
	    NetHandleRecv();
	    if(!XSW_CON_IS_CONNECTED(net))
		break;

	    break;

	  case XSW_CON_STATE_CONNECTED:

	    /* Handle incoming data */
	    NetHandleRecv();
	    if(!XSW_CON_IS_CONNECTED(net) ||
	       !XSW_CON_STATE_IS_CONNECTED(net)
	    )
		break;

	    /* Time to adjust network interval? */
	    if(auto_interval_tune.next <= cur_millitime)
	    {
		AutoIntervalTune();
		if(!XSW_CON_IS_CONNECTED(net))
		    break;

		/* Schedual next auto interval tune */
		auto_interval_tune.next = cur_millitime +
		    auto_interval_tune.interval;
	    }

	    /* Time to send out player position? */
	    if(next.player_pos_send <= cur_millitime)
	    {
		const int player_obj_num = net->player_obj_num;

		/* Position */
		NetSendObjPose(player_obj_num);
		if(!XSW_CON_IS_CONNECTED(net))
		    break;

		/* Throttle */
		NetSendObjThrottle(player_obj_num);
		if(!XSW_CON_IS_CONNECTED(net))   
		    break;

		/* Schedual next player position send */
		next.player_pos_send = cur_millitime + net->net_int;
	    }

	    /* Manage server script uploads
	     *
	     * The timming is handled by ServScriptManage()
	     */
	    ServScriptManage();
	    if(!XSW_CON_IS_CONNECTED(net))   
		break;

	    break;

	  case XSW_CON_STATE_CLOSE_WAIT:

	    /* Handle incoming data for any subsequent data sent by
	     * the server and for recv() == 0
	     */
	    NetHandleRecv();
	    if(!XSW_CON_IS_CONNECTED(net))
		break;

	    break;
	}

	return(total_events_handled);
}


/*
 *	Initializes XSW.
 */
int XSWInit(int argc, char *argv[])
{
	int i, n, x, y, z;
	bool_t b;
	const char *arg;
	char	*s,
		cwd[PATH_MAX],
		home[PATH_MAX],
		local_sw_dir[PATH_MAX];
	struct stat stat_buf;
	osw_gui_struct *gui = &osw_gui[0];
	WColorStruct color;
	splash_win_struct *splash;
	xsw_anim_timmer_struct *anim_timmer;
	xsw_vsmenu_item_struct **vsmenu_item = NULL;
	int total_vsmenu_items = 0;
	xsw_sound_struct *snd = &sound;
	xsw_net_parms_struct *net = &net_parms;
	xsw_net_stat_struct *nstat = &net_stat;
	xsw_player_values_struct *pvals = &player_values;
	xsw_bridge_win_struct *bw;
	xsw_option_struct *opt = &option;

	/* Time zone */
	tzset();

#if defined(__WIN32__)
	/* Change working dir to be the user's home dir */
	chdir(CWD_STR);

	/* Get current, home and local shipwars directories */
	strncpy(cwd, CWD_STR, sizeof(cwd));
	strncpy(home, CWD_STR, sizeof(home));
	strncpy(local_sw_dir, CWD_STR, sizeof(local_sw_dir));
#else

	/* Get home directory */
	s = getenv("HOME");
	if(s == NULL)
	    s = "/";
	strncpy(home, s, sizeof(home));
	home[sizeof(home) - 1] = '\0';

	/* Get the current working directory */
	s = STRDUP(XSWCurrentDirectory());
	strncpy(cwd, (s != NULL) ? s : "/", sizeof(cwd));
	cwd[sizeof(cwd) - 1] = '\0';
	free(s);

	/* Get the local shipwars directory */
	if(ISPATHABSOLUTE(XSW_DEF_LOCAL_DATA_DIR))
	{
	    strncpy(
		local_sw_dir, XSW_DEF_LOCAL_DATA_DIR,
		sizeof(local_sw_dir)
	    );
	}
	else
	{
	    s = PrefixPaths(home, XSW_DEF_LOCAL_DATA_DIR);
	    strncpy(
		local_sw_dir, (s != NULL) ? s : home,
		sizeof(local_sw_dir)
	    );
	}
	local_sw_dir[sizeof(local_sw_dir) - 1] = '\0';
#endif


	/* Reset/fetch global variables to their default values */
	debug.level = DEBUG_LEVEL_NONE;
	debug.val = 0.0f;

	cur_millitime = OSWTimeMS();
	cur_systime = time(NULL);
	lapsed_millitime = 0l;

	time_compensation = 1.0f;

	next.update_check = cur_millitime;
	next.viewscreen = cur_millitime;
	next.consoles = cur_millitime;
	next.memory_clean = cur_millitime + XSW_MEMORY_CLEAN_INTERVAL;
	next.player_pos_send = cur_millitime;
	next.spawn_particles = cur_millitime;

	prompt_mode = XSW_PROMPT_NONE;

	strncpy(dname.startup, cwd, sizeof(dname.startup));
	dname.startup[sizeof(dname.startup) - 1] = '\0';
	strncpy(dname.home, home, sizeof(dname.home));
	dname.home[sizeof(dname.home) - 1] = '\0';

	s = PrefixPaths(home, XSW_DEF_LOCAL_DATA_DIR);
	strncpy(
	    dname.ltoplevel, (s != NULL) ? s : cwd,
	    sizeof(dname.ltoplevel)
	);
	dname.ltoplevel[sizeof(dname.ltoplevel) - 1] = '\0';

	strncpy(
	    dname.toplevel, XSW_DEF_GLOBAL_DATA_DIR,
	    sizeof(dname.toplevel)
	);
	dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';
	strncpy(
	    dname.etc, XSW_DEF_GLOBAL_ETC_DIR,
	    sizeof(dname.etc)
	);
	dname.etc[sizeof(dname.etc) - 1] = '\0';
	strncpy(
	    dname.images, XSW_DEF_GLOBAL_IMAGES_DIR,
	    sizeof(dname.images)
	);
	dname.images[sizeof(dname.images) - 1] = '\0';
	strncpy(
	    dname.sounds, XSW_DEF_GLOBAL_SOUNDS_DIR,
	    sizeof(dname.sounds)
	);
	dname.sounds[sizeof(dname.sounds) - 1] = '\0';
	strncpy(
	    dname.scripts, home,
	    sizeof(dname.scripts)
	);
	dname.scripts[sizeof(dname.scripts) - 1] = '\0';
	strncpy(
	    dname.downloads, home,
	    sizeof(dname.downloads)
	);
	dname.downloads[sizeof(dname.downloads) - 1] = '\0';
	strncpy(
	    dname.starchart, local_sw_dir,
	    sizeof(dname.starchart)
	);
	dname.starchart[sizeof(dname.starchart) - 1] = '\0';

	fps_counter.fcount = 0;
	fps_counter.lfcount = 0;
	fps_counter.interval = 1000l;		/* In ms */
	fps_counter.next = cur_millitime + fps_counter.interval;


	xsw_font.std = NULL;
	xsw_font.std_bold = NULL;
	xsw_font.con_heading = NULL;
	xsw_font.con_standard = NULL;
	xsw_font.con_message = NULL;
	xsw_font.vs_heading = NULL;
	xsw_font.vs_standard = NULL;


	opt->cfg_version_major = PROG_VERSION_MAJOR;
	opt->cfg_version_minor = PROG_VERSION_MINOR;
	opt->cfg_version_release = PROG_VERSION_RELEASE;

	opt->cpu_priority = XSW_CPU_PRIORITY_HIGHEST;
	opt->units = XSW_UNITS_METRIC;
	opt->async_image_loading = 1;
	opt->async_image_pixels = 500000;

	opt->def_scanner_orient = XSW_SCANNER_ORIENT_GC;

	opt->log_client = False;
	opt->log_net = False;
	opt->log_errors = False;

	opt->async_redraws = True;
	opt->show_formal_label = XSW_SHOW_FORMAL_LABEL_AS_NEEDED;
	opt->show_viewscreen_flags = XSW_SHOW_VIEWSCREEN_SELECTED_WEAPON |
	    XSW_SHOW_VIEWSCREEN_MARKINGS |
	    XSW_SHOW_VIEWSCREEN_OBJECT_LABELS;
	opt->show_lens_flares = True;
	opt->show_strobe_glow = True;
	opt->show_nebula_glow = True;
	opt->show_fires = True;
	opt->show_smoke_trails = True;
	opt->max_particles = 500;
	opt->net_bandwidth = XSW_DEF_NET_BANDWIDTH;
	opt->net_timeout = XSW_DEF_NET_TIMEOUT;
	opt->net_object_outdated_time = XSW_DEF_OBJECT_OUTDATED_TIME;
	opt->show_net_errors = False;
	opt->show_server_errors = True;
	opt->local_updates = True;
	opt->auto_zoom = True;
	opt->sounds = XSW_SOUNDS_NONE;
	opt->music = False;
	opt->controller = XSW_CONTROLLER_KEYBOARD;
	opt->cmd_line_set_controller = False;
	opt->throttle_mode = XSW_THROTTLE_MODE_NORMAL;
#ifdef JS_SUPPORT
	opt->focus_out_js_close = False;
#endif
	opt->energy_saver_mode = False;
	opt->scanner_limiting = True;
	opt->notify_scanner_contacts = True;

	opt->auto_map_eco_win = True;
	opt->auto_map_univ_list_win = True;
	opt->clear_chart_on_connect = True;
	opt->save_on_exit = True;

	opt->bridge_x = 0;
	opt->bridge_y = 0;
	opt->bridge_width = 0;
	opt->bridge_height = 0;

	opt->msgw_x = 0;
	opt->msgw_y = 0;
	opt->msgw_width = 0;
	opt->msgw_height = 0;

	opt->ulw_x = 0;
	opt->ulw_y = 0;
	opt->ulw_width = 0;
	opt->ulw_height = 0;

	opt->ecow_x = 0;
	opt->ecow_y = 0;
	opt->ecow_width = 0;
	opt->ecow_height = 0;

	opt->scht_x = 0;
	opt->scht_y = 0;
	opt->scht_width = 0;
	opt->scht_height = 0;


#ifdef HAVE_Y2
	snd->server_type = SOUND_SERVER_TYPE_YIFF;
#else
	snd->server_type = SOUND_SERVER_TYPE_NONE;
#endif
	snd->start_cmd = NULL;
	snd->con_arg = NULL;
	snd->con_data = NULL;
	snd->audio_mode_name = NULL;
	snd->playid_list = NULL;
	snd->nplayids = 0;
	snd->bg_music_playid = NULL;
	snd->bg_music_sndref_num = -1;
	snd->loop_tractor_beam_playid = NULL;

	pvals->weapons_online = 1;
	pvals->weapon_freq = 160.00;
	pvals->weapon_yield = 1.0;
	pvals->weapon_fire_heading = 0.0 * PI;
	pvals->weapon_fire_pitch = 0.0 * PI;
	pvals->auto_destruct_start_int = 20;
	pvals->auto_destruct_last_sec_left = 0l;
	pvals->warning_weapons_lock = 0;
	pvals->warning_incoming_fire = 0;

	UNVLegendReset(&sw_legend);

	scanner_contact_list = NULL;
	total_scanner_contacts = 0;

	vs_object_label = NULL;
	total_vs_object_labels = 0;

	xsw_ocsn = NULL;
	total_ocsns = 0;

	xsw_object = NULL;
	total_objects = 0;

	xsw_particle = NULL;
	total_particles = 0;

	inrange_xsw_object = NULL;
	total_inrange_objects = 0;

	xsw_imgref = NULL;
	total_imgrefs = 0;

	xsw_image = NULL;
	total_images = 0;

	mesg_list = NULL;
	total_messages = 0;

	serv_script = NULL;
	total_serv_scripts = 0;

	xsw_sndref = NULL;
	total_sndrefs = 0;

#ifdef JS_SUPPORT
	xsw_jsmap = NULL;
	total_jsmaps = 0;
#endif	/* JS_SUPPORT */

	/* Reset player object to none (-1) */
	DBSetPlayerObject(-1);


	/* Set default file names */
	strncpy(
	    fname.cfg,
	    PrefixPaths(local_sw_dir, XSW_DEF_CFG_FILE),
	    sizeof(fname.cfg)
	);
	fname.cfg[sizeof(fname.cfg) - 1] = '\0';
	strncpy(
	    fname.universes,
	    PrefixPaths(local_sw_dir, XSW_DEF_UNIVERSES_FILE),
	    sizeof(fname.universes)
	);
	fname.universes[sizeof(fname.universes) - 1] = '\0';

#ifdef JS_SUPPORT
	strncpy(
	    fname.js_calib,
#if defined(HAVE_LIBJSW) && defined(JSDefaultCalibration)
	    PrefixPaths(home, JSDefaultCalibration),
#else
	    PrefixPaths(home, ".joystick"),
#endif
	    sizeof(fname.js_calib)
	);
	fname.js_calib[sizeof(fname.js_calib) - 1] = '\0';
#endif	/* JS_SUPPORT */

	strncpy(
	    fname.main_page,
	    PrefixPaths(cwd, XSW_DEF_MAIN_PAGE_FILE),
	    sizeof(fname.main_page)
	);
	fname.main_page[sizeof(fname.main_page) - 1] = '\0';

	strncpy(
	    fname.destroyed_page,
	    PrefixPaths(cwd, XSW_DEF_DESTROYED_PAGE_FILE),
	    sizeof(fname.destroyed_page)
	);
	fname.destroyed_page[sizeof(fname.destroyed_page) - 1] = '\0';

	strncpy(
	    fname.ocsn,
	    PrefixPaths(cwd, XSW_DEF_OCSN_FILE),
	    sizeof(fname.ocsn)
	);
	fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';

	strncpy(
	    fname.imgrefs,
	    PrefixPaths(cwd, XSW_DEF_IMGREFS_FILE),
	    sizeof(fname.imgrefs)
	);
	fname.imgrefs[sizeof(fname.imgrefs) - 1] = '\0';

	strncpy(
	    fname.sndrefs,
	    PrefixPaths(cwd, XSW_DEF_SNDREFS_FILE),
	    sizeof(fname.sndrefs)
	);
	fname.sndrefs[sizeof(fname.sndrefs) - 1] = '\0';

	strncpy(
	    fname.log,
	    PrefixPaths(home, XSW_DEF_LOG_FILE),
	    sizeof(fname.log)
	);
	fname.log[sizeof(fname.log) - 1] = '\0';

	/* Reset net parameters */
	net->connection_state = XSW_CON_STATE_NOT_CONNECTED;
	net->socket = -1;
	net->flags = 0;
	net->player_obj_num = -1;
	net->player_obj_ptr = NULL;
	s = getenv("LOGNAME");
	net->login_name = (s != NULL) ?
	    strdup(s) : strdup(XSW_DEF_GUEST_LOGIN_NAME);
	net->login_password = strdup(XSW_DEF_GUEST_LOGIN_PASSWORD);
	net->con_start = 0l;
	net->login_delay_warning_last = 0l;
	net->title = NULL;
	net->address = strdup(XSW_DEF_SWSERV_ADDRESS);
	net->port = XSW_DEF_SWSERV_PORT;
	net->net_int = XSW_SWSERV_DEF_UPDATE_INT;
	net->disconnect_send_count = 0;
	net->bad_send_count = 0;

	*net->recvbuf = '\0';
	net->recvbuf_len = 0;


	nstat->rx_interval = 0l;
	nstat->sx_interval = 0l;
	nstat->rx_ilast = 0l;
	nstat->sx_ilast = 0l;


	auto_interval_tune.state = True;
	auto_interval_tune.interval = AINT_DEF_TUNE_INT;
	auto_interval_tune.next = cur_millitime +
	    auto_interval_tune.interval;


	/* Set up global color values */
	xsw_color.heading_arrow.a = 0xff;
	xsw_color.heading_arrow.r = 0x00;
	xsw_color.heading_arrow.g = 0xf0;
	xsw_color.heading_arrow.b = 0xff;

	xsw_color.lock_arrow.a = 0xff;
	xsw_color.lock_arrow.r = 0xff;
	xsw_color.lock_arrow.g = 0xff;
	xsw_color.lock_arrow.b = 0x10;

	xsw_color.lock_cursor.a = 0xff;
	xsw_color.lock_cursor.r = 0xff;
	xsw_color.lock_cursor.g = 0xff;
	xsw_color.lock_cursor.b = 0x10;

	xsw_color.stream_green.a = 0xff;
	xsw_color.stream_green.r = 0x00;
	xsw_color.stream_green.g = 0xff;
	xsw_color.stream_green.b = 0x78;

	xsw_color.stream_yellow.a = 0xff;
	xsw_color.stream_yellow.r = 0xff;
	xsw_color.stream_yellow.g = 0xf4;
	xsw_color.stream_yellow.b = 0x80;

	xsw_color.stream_orange.a = 0xff;
	xsw_color.stream_orange.r = 0xff;
	xsw_color.stream_orange.g = 0x90;
	xsw_color.stream_orange.b = 0x30;

	xsw_color.stream_purple.a = 0xff;
	xsw_color.stream_purple.r = 0xff;
	xsw_color.stream_purple.g = 0x70;
	xsw_color.stream_purple.b = 0xff;

	xsw_color.shield_blue.a = 0xff;
	xsw_color.shield_blue.r = 0xa0;
	xsw_color.shield_blue.g = 0xe0;
	xsw_color.shield_blue.b = 0xff;

	xsw_color.star_glow.a = 0xff;
	xsw_color.star_glow.r = 0xff;
	xsw_color.star_glow.g = 0xff;
	xsw_color.star_glow.b = 0xff;


	/* Game controller */
	xsw_gctl = NULL;

	/* Initialize the universe */
	UNVInit();


	/* Setup global animation timmer values */
	anim_timmer = &xsw_anim_timmer[XSW_ANIM_TIMMER_DEFAULT];
	anim_timmer->count = 0;
	anim_timmer->count_max = 8;
	anim_timmer->next = cur_millitime;
	anim_timmer->interval = 120l;

	anim_timmer = &xsw_anim_timmer[XSW_ANIM_TIMMER_GLOW_SHORT];
	anim_timmer->count = 0;
	anim_timmer->count_max = 8;
	anim_timmer->next = cur_millitime;
	anim_timmer->interval = 60l;

	anim_timmer = &xsw_anim_timmer[XSW_ANIM_TIMMER_GLOW_MEDIUM];
	anim_timmer->count = 0;
	anim_timmer->count_max = 8;
	anim_timmer->next = cur_millitime;
	anim_timmer->interval = 100l;

	anim_timmer = &xsw_anim_timmer[XSW_ANIM_TIMMER_GLOW_LONG];
	anim_timmer->count = 0;
	anim_timmer->count_max = 8;
	anim_timmer->next = cur_millitime;
	anim_timmer->interval = 140l;

	anim_timmer = &xsw_anim_timmer[XSW_ANIM_TIMMER_BLINK_SHORT];
	anim_timmer->count = 0;
	anim_timmer->count_max = 1;
	anim_timmer->next = cur_millitime;
	anim_timmer->interval = 300l;

	anim_timmer = &xsw_anim_timmer[XSW_ANIM_TIMMER_BLINK_LONG];
	anim_timmer->count = 0;
	anim_timmer->count_max = 1;
	anim_timmer->next = cur_millitime;
	anim_timmer->interval = 750l;


	univ_item = NULL;
	total_univ_items = 0;

	ignored_obj_list = NULL;
	total_ignored_objs = 0;


	/* Reset windows */
	xsw_bridge_win = NULL;
	xsw_vsmenu = NULL;
	xsw_msg_win = NULL;
	xsw_unvlist_win = NULL;
	xsw_unvprop_dlg = NULL;
	xsw_ecow = NULL;
	xsw_scht = NULL;

	xsw_optwin = NULL;
	xsw_keymap_win = NULL;
#ifdef JS_SUPPORT  
	xsw_jsmap_win = NULL;
#endif
	xsw_vsmenu_win = NULL;

	xsw_dlg = NULL;
	xsw_cdlg = NULL;
	xsw_file_selector = NULL;


	/* Handle command line parameters */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Configuration file */
	    if(strcasepfx(arg, "--cfg") ||
	       strcasepfx(arg, "-cfg") ||
	       strcasepfx(arg, "--rc") ||
	       strcasepfx(arg, "-rc") ||
	       !strcasecmp(arg, "-f")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.cfg, argv[i], sizeof(fname.cfg));
		    fname.cfg[sizeof(fname.cfg) - 1] = '\0';
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Controller type */
	    else if(strcasepfx(arg, "--con") ||
		    strcasepfx(arg, "-c")
	    )
	    {
		i++;
		if(i < argc)
		{
		    const char *con_val = argv[i];
		    if(strcasepfx(con_val, "k"))
		    {
			opt->controller = XSW_CONTROLLER_KEYBOARD;
			opt->cmd_line_set_controller = 1;
		    }
#ifdef JS_SUPPORT
		    else if(strcasepfx(con_val, "j"))
		    {
		        opt->controller = XSW_CONTROLLER_JOYSTICK;
			opt->cmd_line_set_controller = 1;
		    }
#endif	/* JS_SUPPORT */
		    else if(strcasepfx(con_val, "p") ||
		            strcasepfx(con_val, "m")
		    )
		    {
			opt->controller = XSW_CONTROLLER_POINTER;
			opt->cmd_line_set_controller = 1;
		    }
		    else
		    {
		        fprintf(
			    stderr,
			    "%s: Unsupported controller type.\n",
			    con_val
			);
		    }
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Site address and port */
	    else if(strcasepfx(arg, "shipwars://") ||
		    strcasepfx(arg, "swserv://") ||
		    strcasepfx(arg, "sw://")
	    )
	    {
		char *s, *url = strdup(arg);

		/* Get login name */
		s = StringParseName(url);
		if(s != NULL)
		{
		    free(net->login_name);
		    net->login_name = strdup(s);
		}
		/* Get login password */
		s = StringParsePassword(url);
		if(s != NULL)
		{
		    free(net->login_password);
		    net->login_password = strdup(s);
		}
		/* Get address */
		s = StringParseAddress(url);
		if(s != NULL)
		{
		    free(net->address);
		    net->address = strdup(s);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Unable to parse URL.\n",
			url
		    );
		}
		/* Get port */
		n = StringParsePort(url);
		if(n > -1)
		    net->port = n;

		/* Mark address being set at startup */
		net->flags |= XSW_CON_FLAG_ADDRESS_SET;

		free(url);
	    }
	}


	/* Open the configuration file
	 *
	 * Is it not in the current or the command line specified
	 * location? 
	 */
	if(stat(fname.cfg, &stat_buf))
	{
	    /* Not in current or command line specified location,
	     * check if it is in the local shipwars directory
	     */
	    s = PrefixPaths(local_sw_dir, XSW_DEF_CFG_FILE);
	    strncpy(
		fname.cfg, (s != NULL) ? s : XSW_DEF_CFG_FILE,
		sizeof(fname.cfg)
	    );
	    fname.cfg[sizeof(fname.cfg) - 1] = '\0';
	    if(stat(fname.cfg, &stat_buf))
	    {
		/* Not in the local shipwars directory, so assume it
		 * does not exist and install the global configuration
		 * file to the local shipwars directory
		 */
		char *src_path = STRDUP(PrefixPaths(
		    dname.etc, XSW_DEF_CFG_FILE
		));
		XSWDoInstallObject(
		    local_sw_dir,		/* Target */
		    src_path,			/* Source */
		    S_IRUSR | S_IWUSR | S_IXUSR,
		    XSW_INST_OBJ_FLAG_DIR |
		    XSW_INST_OBJ_FLAG_RECURSIVE, /* Options */
		    NULL			/* Data */
		);
		XSWDoInstallObject(
		    fname.cfg,			/* Target */
		    src_path,			/* Source */
		    S_IRUSR | S_IWUSR,
		    0,				/* Options */
		    NULL			/* Data */
		);
		free(src_path);
	    }
	}
	/* Open the configuration file */
	CfgFileOpen(fname.cfg, opt);

	/* Check if the universes file does not exist, its location
	 * was defined in the configuration file
	 */
	if(stat(fname.universes, &stat_buf))
	{
	    /* Does not exist, so install the global one to the
	     * location defined in the configuration file
	     */
	    char *src_path = STRDUP(PrefixPaths(
		dname.etc, XSW_DEF_UNIVERSES_FILE
	    ));
	    XSWDoInstallObject(
		fname.universes,	/* Target */
		src_path,		/* Source */
		S_IRUSR | S_IWUSR,
		0,			/* Options */
		NULL			/* Data */
	    );
	    free(src_path);
	}
	/* Open the universes list file */
	UnvListFileOpen(
	    fname.universes,
	    &univ_item, &total_univ_items,
	    NULL, NULL,
	    XSWPrintWarning
	);


	/* Now that configuration values have been loaded, check if the
	 * required directories exist
	 */
	if(stat(dname.toplevel, &stat_buf))
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
		*error_msg = toupper(*error_msg);
	    fprintf(
		stderr,
"%s: %s.\n",
		error_msg,
		dname.toplevel
	    );
	    fprintf(
		stderr,
"This is the global data files directory and must exist in order for\n\
this program to run, please check if it has been installed and that its\n\
permissions are set properly to allow it to be accessed by this program.\n"
	    );
	    free(error_msg);
	    return(1);
	}
	else if(stat(dname.images, &stat_buf))
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
		*error_msg = toupper(*error_msg);
	    fprintf(
		stderr,
"%s: %s.\n",
		error_msg,
		dname.images
	    );
	    fprintf(
		stderr,
"This is the global images directory and must exist in order for\n\
this program to run, please check if it has been installed and that its\n\
permissions are set properly to allow it to be accessed by this program.\n"
	    );
	    free(error_msg);
	    return(1);
	}
	else if(stat(dname.sounds, &stat_buf))
	{
	    /* Do not warn */
	}
	else if(stat(dname.etc, &stat_buf))
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
		*error_msg = toupper(*error_msg);
	    fprintf(
		stderr,
"%s: %s.\n",
		error_msg,
		dname.etc
	    );
	    fprintf(
		stderr,
"This is the global configuration directory and must exist in order for\n\
this program to run, please check if it has been installed and that its\n\
permissions are set properly to allow it to be accessed by this program.\n"
	    );
	    free(error_msg);
	    return(1);
	}


	/* Apply command line overrides to the configuration */

	/* Controller */
	if(opt->cmd_line_set_controller)
	{
	    GCtlDelete(xsw_gctl);
	    xsw_gctl = GCtlNew(opt->controller);
	    if(xsw_gctl == NULL)
	    {
		fprintf(
		    stderr,
"An error occured while initializing the game controller.\n\
\n\
Make sure that your game controllers are set up properly.\n\
\n\
You may set the controller by editing the configuration file:\n\
\n\
    %s\n\
\n\
By changing the value for parameter `Controller' to something such as\n\
`keyboard'.\n",
		    fname.cfg
		);
	    }
	}


	/* ******************************************************** */
	/* Initialize sound */
	if((opt->sounds > XSW_SOUNDS_NONE) &&
	   (snd->server_type > SOUND_SERVER_TYPE_NONE)
	)
	{
	    if(SoundInit(&sound, snd->server_type))
	    {
		/* An error occured while starting sound server */
		fprintf(
		    stderr,
"Your configuration specified sound support, however there was an\n\
error initializing the sound server. Please verify that the sound\n\
settings are correct and that your sound server is set up properly.\n"
		);

		/* Turn sounds off */
		opt->sounds = XSW_SOUNDS_NONE;
	    }
	    else
	    {
		/* Successfully initialized sound */

		/* Open the sndrefs from file */
		XSWOpenSndRefs(fname.sndrefs);
	    }
	}


	/* *********************************************************** */
	/* Connect to GUI */
	if(OSWGUIConnect(argc, argv))
	    return(-1);


	/* Make sure that the depth is supported */
#ifdef HAVE_X
	if((gui->depth != 8) &&
	   (gui->depth != 15) &&
	   (gui->depth != 16) &&
	   (gui->depth != 24) &&
	   (gui->depth != 32)
	)
	{
	    fprintf(
		stderr,
"%s does not support %i bit depth video mode.\n",
		PROG_NAME_FULL,
		gui->depth
	    );
	    fprintf(
		stderr,
"\nPlease set your video mode to one of the following bit depths:\n\
\n\
    8 bits\n\
    15 bits\n\
    16 bits\n\
    32 bits\n\
\n"
	    );

/*	    return(-1); */
	}
#endif	/* HAVE_X */


#ifdef HAVE_X
	/* Check if visual class is TrueColor or DirectColor */
	if(gui->visual != NULL)
	{
	    int vclass;
	    visual_t *vis = gui->visual;
#if defined(__cplusplus) || defined(c_plusplus)
	    vclass = vis->c_class;
#else
	    vclass = vis->class;
#endif

	    /* Unsupported visual class? */
	    if((vclass != TrueColor) && (vclass != PseudoColor))
	    {
		const char *vis_name;
		switch(vclass)
		{
		  case StaticGray:
		    vis_name = "StaticGray";
		    break;
		  case GrayScale:
		    vis_name = "GrayScale";
		    break;
		  case StaticColor:
		    vis_name = "StaticColor";
		    break;
		  case PseudoColor:
		    vis_name = "PseudoColor";
		    break;
		  case TrueColor:
		    vis_name = "TrueColor";
		    break;
		  case DirectColor:
		    vis_name = "DirectColor";
		    break;
		  default:
		    vis_name = "Unknown";
		    break;
 		}

		fprintf(
		    stderr,
"Warning: Visual class %s is not supported by this program.\n",
		    vis_name
		);

		fprintf(
		    stderr,
"\nThis program runs best with a Visual class of TrueColor. To correct\n\
this problem, specify your Visual value to \"TrueColor\" in your X server's\n\
configuration file.\n"
		);
	    }
	}
#endif	/* HAVE_X */

	/* Load fonts */

	/* Standard fonts */
	OSWLoadFont(&xsw_font.std, "7x14");
	OSWLoadFont(&xsw_font.std_bold, "7x14");

	/* Bridge and message window console fonts */
	OSWLoadFont(
	    &xsw_font.con_heading,
	    "-adobe-helvetica-bold-r-normal-*-14-*-*-*-p-*-iso8859-1"
	);
	if(xsw_font.con_heading == NULL)
	    OSWLoadFont(&xsw_font.con_heading, "7x14");

	OSWLoadFont(
	    &xsw_font.con_standard,
	    "-adobe-helvetica-medium-r-normal-*-10-*-*-*-p-*-iso8859-1"
	);
	if(xsw_font.con_standard == NULL)
	    OSWLoadFont(&xsw_font.con_standard, "6x10");

	/* Msgbox font, must be fixed width 7 by 14 */
	OSWLoadFont(&xsw_font.con_message, "7x14");

	/* Viewscreen fonts */
	OSWLoadFont(
	    &xsw_font.vs_heading,
	    "-adobe-helvetica-bold-r-normal-*-14-*-*-*-p-*-iso8859-1"
	);
	if(xsw_font.vs_heading == NULL)
	    OSWLoadFont(&xsw_font.vs_heading, "7x14");

	OSWLoadFont( 
	    &xsw_font.vs_standard,
	    "-adobe-helvetica-bold-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	if(xsw_font.vs_standard == NULL)
	    OSWLoadFont(&xsw_font.vs_heading, "7x14");


	/* Load colors */
	if(XSWLoadColors())
	    return(-1);


	/* Load cursors */
	color.a = 0xff;
	color.r = 0xff;
	color.g = 0xc0;
	color.b = 0xfa;

	xsw_cursor.scanner_lock = WidgetCreateCursorFromData(
	    xsw_scanner_lock_xpm,
	    8, 8,
	    &color, NULL
	);
	xsw_cursor.text = WidgetCreateCursorFromData(
	    xsw_text_xpm,
	    8, 8,
	    &color, NULL
	);
	xsw_cursor.wait = NULL;		/* No wait cursor yet */
	xsw_cursor.translate = WidgetCreateCursorFromData(
	    translate_xpm,
	    8, 8,
	    &color, NULL
	);
	xsw_cursor.zoom = WidgetCreateCursorFromData(
	    zoom_xpm,
	    8, 8,
	    &color, NULL
	);


	/* Keyboard control settings */
/* None */


#if defined(HAVE_X) && defined(USE_XSHM)
	/* Check for MIT SHM extension availablility */
	if(XQueryExtension(gui->display, "MIT-SHM", &x, &x, &x))
	{
	    if(XShmQueryVersion(gui->display, &y, &z, &b) == True)
	    {
#if 0
	        printf(
"XShm extention version %d.%d %s shared pixmaps.\n",
	            y, z, (b == True) ? "with" : "without"
	        );
#endif
	    }
	    else
	    {
		fprintf(
		    stderr,
"MIT-SHM: Warning: Unable to obtain version.\n"
		);
	    }
	}
#endif	/* defined(HAVE_X) && defined(USE_XSHM) */


	/* ********************************************************* */
	/* Initialize window widgets, groups and base images */

	/* Must initialized the widget globals before initializing
	 * any widgets
	 */
	if(WidgetInitGlobals(argc, argv))
	{
	    fprintf(
		stderr,
"An error occured while initializing this programs widget set.\n"
	    );
	    return(-1);
	}


	/* Create the splash window, this should be done before
	 * loading the images so the splash window's progress bar
	 * displays the image loading progress
	 */
	splash = SplashNew();

	/* Load program images:
	 *
	 * This must be performed first before any of the other
	 * window sets are initialized (except for the load status
	 * window which is initialized before the images).
	 */
	if(ImgLoadAll(splash))
	{
	    fprintf(
		stderr,
"An error occured while opening the program's images.\n\
Please check and verify that you have the images installed in their\n\
proper locations.\n"
	    );
	}

	/* Messages list */
	total_messages = 200;
	mesg_list = (xsw_mesg_struct *)realloc(
	    mesg_list,
	    total_messages * sizeof(xsw_mesg_struct)
	);
	if(mesg_list == NULL)
	    total_messages = 0;
	else
	    memset(mesg_list, 0x00, total_messages * sizeof(xsw_mesg_struct));


	/* Bridge window */
	xsw_bridge_win = bw = BridgeWinNew();
	if(bw == NULL)
	    return(-1);

	/* Bridge window main menu */
	PageFileOpen(
	    &bw->main_page,
	    bw->viewscreen, bw->viewscreen_image,
	    fname.main_page
	);

	/* Destroyed page configuration (load from file) */
	PageFileOpen(
	    &bw->destroyed_page,
	    bw->viewscreen, bw->viewscreen_image,
	    fname.destroyed_page
	);

	/* Viewscreen Menu */
	CfgFileOpenVSMenu(fname.cfg, &vsmenu_item, &total_vsmenu_items);
	xsw_vsmenu = VSMenuNew(vsmenu_item, total_vsmenu_items);
	VSMenuItemsListDelete(vsmenu_item, total_vsmenu_items);
	vsmenu_item = NULL;
	total_vsmenu_items = 0;
	if(xsw_vsmenu == NULL)
	    return(-1);

	/* Messages Window */
	xsw_msg_win = MsgWinNew();

	/* Universes List Window */
	xsw_unvlist_win = UnvListWinNew();

	/* Economy */
	xsw_ecow = EcoWinNew();

	/* StarChart */
	xsw_scht = SChtNew();

	/* Initialize the reality engine */
	REngInit();


	/* Open the OCS Names */
	XSWOpenOCSNs(fname.ocsn);

	/* Open the ImgRefs file */
	XSWOpenImgRefs(fname.imgrefs);


	/* Delete the splash window */
	SplashDelete(splash);


	/* Map the bridge window */
	BridgeWinMap(bw);


	/* ******************************************************* */
	/* Print greeting message on the message box */
	s = (char *)malloc(
	    STRLEN(PROG_NAME_FULL) + STRLEN(PROG_VERSION) + 80
	);
	sprintf(
	    s,
"%s Version %s.",
	    PROG_NAME_FULL,
	    PROG_VERSION
	);
	XSWPrint(s);
	free(s);

	if(True)
	{
	    keycode_t keycode = KeymapGetKeycode(XSW_KM_HELP);
	    char *arg = STRDUP(OSWGetKeyCodeName(keycode));
	    s = (char *)malloc(STRLEN(arg) + 80);
	    strtoupper(arg);
	    sprintf(
		s,
"Press %s for help.",
		arg
	    );
	    free(arg);
	    XSWPrint(s);
	    free(s);
	}


	/* ******************************************************** */
	/* Connect at startup? */
	if(net->flags & XSW_CON_FLAG_ADDRESS_SET)
	{
	    s = (char *)malloc(
		STRLEN(net->login_name) +
		STRLEN(net->login_password) +
		STRLEN(net->address) +
		40 + 80
	    );
	    sprintf(
		s,
"connect swserv://%s:%s@%s:%i",
		(net->login_name != NULL) ?
		    net->login_name : "",
		(net->login_password != NULL) ?
		    net->login_password : "",
		(net->address != NULL) ?
		    net->address : "",
		net->port
	    );
	    CmdHandleInput(s);
	    free(s);
	}
	else
	{
	    /* Map the universes list window? */
	    if(opt->auto_map_univ_list_win)
	        UnvListWinMap(xsw_unvlist_win);
	}

	return(0);
}

/*
 *	Manages one iteration of XSW.
 *
 *	Updates timmers and manages all resources.
 */
void XSWMainIteration(void *data)
{
	int i, m;
	time_t	cur_ms, cur_sec,
		new_systime = time(NULL),
		new_millitime = OSWTimeMS();
	float tc;
	xsw_anim_timmer_struct *anim_timmer;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	/* Check if timmers have cycled */
	if((new_systime < cur_systime) ||
	   (new_millitime < cur_millitime)
	)
	{
	    /* Reset timmers and time lapses */
	    lapsed_millitime = 0l;
	    time_compensation = 1.0f;
	    XSWResetTimmers(new_millitime, new_systime, time_compensation);
	}
	else
	{
	    /* Calculate time lapse ms and compensation coeff */
	    lapsed_millitime = new_millitime - cur_millitime;
	    time_compensation = (float)lapsed_millitime /
		(float)SW_CYCLE_LAPSE_MS;
#if 0
/* Allow cases where lapsed_millitime < SW_CYCLE_LAPSE_MS since
 * higher performance CPUs may required usleep(0)
 */
	    if(time_compensation < 1.0f)
		time_compensation = 1.0f;
#endif
	}

	/* Set current time */
	cur_systime = cur_sec = new_systime;
	cur_millitime = cur_ms = new_millitime;
	tc = time_compensation;

	/* Increment general animation timmers */
	m = (int)(sizeof(xsw_anim_timmer) / sizeof(xsw_anim_timmer_struct));
	for(i = 0; i < m; i++)
	{
	    anim_timmer = &xsw_anim_timmer[i];

	    /* Skip if timmer is not due for increment */
	    if(anim_timmer->next > cur_ms)
		continue;

	    anim_timmer->count++;
	    if(anim_timmer->count >= anim_timmer->count_max)
		anim_timmer->count = 0;

	    anim_timmer->next = cur_ms + anim_timmer->interval;
	}


	/* Manage events */

	/* Manage game controller */
	if(opt->controller == XSW_CONTROLLER_KEYBOARD)
	{
	    xsw_bridge_win_struct *bw = xsw_bridge_win;

	    /* Game controller is set to keyboard, so do update only
	     * if the bridge window is in focus and the prompt mode is
	     * set to none
	     */
	    if((prompt_mode == XSW_PROMPT_NONE) && bw->has_focus)
	        GCtlManageJoysticks(xsw_gctl);
	}
	else
	{
	    GCtlManageJoysticks(xsw_gctl);
	}

	/* Manage GUI events */
	XSWManageGUI(data, cur_ms, cur_sec, tc);

	/* Manage game controller
	 *
	 * This should be done after managing joystick and key events
	 */
	GCtlManage(
	    xsw_gctl,
	    cur_ms, cur_sec, tc
	);

	/* Manage sound */
	XSWManageSound(NULL);

	/* Manage network */
	XSWManageNet(NULL);

	/* Manage reality engine */
	REngManage(
	    net->player_obj_ptr,
	    cur_ms, cur_sec, tc
	);

	/* Manage ImgRefs */
	ImgRefManage(
	    &xsw_imgref, &total_imgrefs,
	    cur_ms, cur_sec, tc,
	    XSWImgRefLoadCompletedCB, NULL
	);

        /* Update the looping sounds as needed */
        if(opt->sounds >= XSW_SOUNDS_ENGINE)
            XSWUpdateLoopSounds();


	/* Redraw GUI windows that need to be updated regularly */

	/* Viewscreen */
	if(opt->energy_saver_mode ?
	    (next.viewscreen <= cur_ms) : True
	)
	{
	    const int player_obj_num = net->player_obj_num;
	    xsw_object_struct *player_obj = net->player_obj_ptr;
	    xsw_fps_counter_struct *fps = &fps_counter;
	    xsw_bridge_win_struct *bw = xsw_bridge_win;

	    /* Update frames per second */
	    fps->fcount++;
	    if(fps->next <= cur_ms)
	    {
		fps->lfcount = fps_counter.fcount;
		fps->fcount = 0;
		fps->next = cur_ms + fps->interval;
	    }

	    /* Redraw viewscreen */
	    if(bw->viewscreen_vis_state != VisibilityFullyObscured)
		VSDrawViewScreen(bw);

	    /* Redraw scanner */
	    if(bw->scanner_vis_state != VisibilityFullyObscured)
		ScannerDraw(
		    bw, player_obj_num,
		    bw->scanner,
		    bw->scanner_image,
		    0,
		    0,
		    bw->scanner_zoom,
		    cur_ms, cur_sec, tc
		);

	    /* Redraw bridge console panels that need to be redrawn
	     * regularly
	     */
	    if(XSW_CON_STATE_IS_CONNECTED(net) &&
	       (player_obj != NULL)
	    )
	    {
		const int tar_obj_num = player_obj->locked_on;

		BridgeWinDrawPanel(
		    player_obj_num,
		    BW_PANEL_PTHROTTLE
		);
		BridgeWinDrawPanel(
		    player_obj_num,
		    BW_PANEL_PVELOCITY
		);
		BridgeWinDrawPanel(
		    player_obj_num,
		    BW_PANEL_PHEADING
		);
		BridgeWinDrawPanel(
		    player_obj_num,
		    BW_PANEL_PTHRUSTVECTOR
		);

		/* Subject object */
		if(tar_obj_num > -1)
		{
		    BridgeWinDrawPanel(
			tar_obj_num,
			BW_PANEL_SBEARING
		    );
		    BridgeWinDrawPanel(
			tar_obj_num,
			BW_PANEL_SDISTANCE
		    );
		}
	    }


	    /* Calculate and schedual next viewscreen draw time */
	    if(opt->energy_saver_mode)
	    {
		/* In energy saver mode, draw slower */
		bw->viewscreen_int = MAX(
		    OSWTimeMS() - cur_ms + 1000l,
		    0l
		);
		next.viewscreen = OSWTimeMS() + bw->viewscreen_int;
	    }
	    else
	    {
		/* Calculate how long the above redraws took */
		bw->viewscreen_int = MAX(
		    OSWTimeMS() - cur_ms - SW_CYCLE_LAPSE_MS,
		    0l
		);
		next.viewscreen = cur_ms + bw->viewscreen_int;
	    }
	}


	/* Redraw consoles and do scanner contact check? */
	if(next.consoles <= cur_ms)
	{
	    int tar_obj_num;
	    const int player_obj_num = net->player_obj_num;
	    xsw_bridge_win_struct *bw = xsw_bridge_win;

	    /* Update the scanner contacts list */
	    ScHandleContacts(
		&scanner_contact_list, &total_scanner_contacts,
		player_obj_num
	    );

	    /* Redraw player object stats */
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PNAME);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PSHIELDFREQ);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PWEAPONFREQ);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PCOMCHANNEL);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PSCORE);

	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PHULL);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PPOWER);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PSHIELDS);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PVIS);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PDMGCTL);

	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PINAME);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PWLOCK);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PWEAPONS); 

	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PANTIMATTER);
#if 0
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PENGINESTATE);
	    BridgeWinDrawPanel(player_obj_num, BW_PANEL_PTHRUSTVECTOR);
#endif
	    if(net->player_obj_ptr != NULL)
		tar_obj_num = net->player_obj_ptr->locked_on;
	    else
		tar_obj_num = -1;

	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SNAME);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SEMPIRE);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SHULL);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SPOWER);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SSHIELDS);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SVIS);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SBEARING);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SDISTANCE);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SDMGCTL);

	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SINAME);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SWLOCK);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SWEAPONS);

	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SANTIMATTER);
	    BridgeWinDrawPanel(tar_obj_num, BW_PANEL_SHEADING);

	    /* Update the viewscreen's netstats label as needed */
	    if(opt->show_viewscreen_flags & XSW_SHOW_VIEWSCREEN_NETSTATS)
		VSDrawUpdateNetstatsLabel(
		    &bw->net_stats_image,
		    bw->net_stats_buf
		);

	    /* Schedual next time to redraw consoles and do scanner
	     * contacts check
	     */
	    next.consoles = cur_ms + XSW_CONSOLES_UPDATE_INTERVAL;
	}


	/* Manage the StarChart's timeouts */
	SChtTimeoutCB(xsw_scht);


	/* Delete any non-essantial unmapped windows */
	if((xsw_unvprop_dlg != NULL) ? 
	    !xsw_unvprop_dlg->map_state : False 
	)
	{
	    UnvPropDlgDelete(xsw_unvprop_dlg);
	    xsw_unvprop_dlg = NULL;
	}
	if((xsw_optwin != NULL) ? 
	    !xsw_optwin->map_state : False 
	)
	{
	    OptWinDelete(xsw_optwin);
	    xsw_optwin = NULL;
	}
	if((xsw_keymap_win != NULL) ?
	    !xsw_keymap_win->map_state : False
	)
	{
	    KeymapWinDelete(xsw_keymap_win);
	    xsw_keymap_win = NULL;
	}
#ifdef JS_SUPPORT
	if((xsw_jsmap_win != NULL) ?
	    !xsw_jsmap_win->map_state : False
	)
	{
	    JSMWDelete(xsw_jsmap_win);
	    xsw_jsmap_win = NULL;
	}
#endif	/* JS_SUPPORT */
	if((xsw_vsmenu_win != NULL) ?
	    !xsw_vsmenu_win->map_state : False
	)
	{
	    VSMWDelete(xsw_vsmenu_win);
	    xsw_vsmenu_win = NULL;
	}


	/* Clean/refresh memory? */
	if(next.memory_clean <= cur_ms)
	{
	    /* Clean, reclaim, memory and schedual next */
	    XSWReclaimMem(False, cur_ms, cur_sec, tc);

	    /* XSWReclaimMem() already scheduals next */
	}
}

/*
 *	Manages XSW.
 *
 *	Pushes an XSW main loop and manages and blocks until
 *	XSWMainPop*() is called.
 */
void XSWMain(void *data)
{
	const int entering_runlevel = runlevel;
	runlevel++;
	while(runlevel > entering_runlevel)
	{
	    switch(option.cpu_priority)
	    {
	      case XSW_CPU_PRIORITY_HIGHEST:
		/* No sleeping */
		break;
	      case XSW_CPU_PRIORITY_HIGH:
		usleep(0);
		break;
	      case XSW_CPU_PRIORITY_MEDIUM:
		usleep(SW_CYCLE_LAPSE_US / 2);
		break;
	      default:		/* XSW_CPU_PRIORITY_LOWEST */
		usleep(SW_CYCLE_LAPSE_US);
		break;
	    }

	    XSWMainIteration(data);
	}
}

/*
 *	Pops off the highest XSW main loop.
 */
void XSWMainPop(void *data)
{
	if(runlevel > 0)
	    runlevel--;
}

/*
 *	Pops off all XSW main loops.
 */
void XSWMainPopAll(void *data)
{
	runlevel = 0;
}  

/*
 *	Shuts down XSW.
 *
 *	The configuration is not saved by this call, you should call
 *	CfgFileSave() prior to this call if you want to save the
 *	configuration.
 */
void XSWShutdown(void)
{
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	/* Delete all the server script upload events */
	ServScriptDeleteAll();

	/* Close the connection properly */
	if(XSW_CON_IS_CONNECTED(net))
	{
	    if(XSW_CON_STATE_IS_CLOSE_WAIT(net))
	    {
		NetWaitClose(opt->net_timeout);
	    }
	    else
	    {
		NetSendLogout();
		NetShutdownConnection(XSW_CON_SHUTDOWN_RDWR, True);
		NetWaitClose(opt->net_timeout);
	    }
	}

	/* Delete any remaining network resources */
	XSWDisconnect();
	free(net->login_name);
	net->login_name = NULL;
	free(net->login_password);
	net->login_password = NULL;
	free(net->title);
	net->title = NULL;
	free(net->address);
	net->address = NULL;


	if(IDC())
	{
	    /* Windows */

	    VSMWDelete(xsw_vsmenu_win);
	    xsw_vsmenu_win = NULL;

#ifdef JS_SUPPORT
	    JSMWDelete(xsw_jsmap_win);
	    xsw_jsmap_win = NULL;
#endif
	    KeymapWinDelete(xsw_keymap_win);
	    xsw_keymap_win = NULL;

	    OptWinDelete(xsw_optwin);
	    xsw_optwin = NULL;

	    SChtDelete(xsw_scht);
	    xsw_scht = NULL;

	    EcoWinDelete(xsw_ecow);
	    xsw_ecow = NULL;

	    UnvPropDlgDelete(xsw_unvprop_dlg);
	    xsw_unvprop_dlg = NULL;

	    UnvListWinDelete(xsw_unvlist_win);
	    xsw_unvlist_win = NULL;

	    MsgWinDelete(xsw_msg_win);
	    xsw_msg_win = NULL;

	    VSMenuDelete(xsw_vsmenu);
	    xsw_vsmenu = NULL;

	    BridgeWinDelete(xsw_bridge_win);
	    xsw_bridge_win = NULL;

	    XSWDialogDelete(xsw_dlg);
	    xsw_dlg = NULL;
	    XSWConfirmationDialogDelete(xsw_cdlg);
	    xsw_cdlg = NULL;
	    XSWFileSelectorDelete(xsw_file_selector);
	    xsw_file_selector = NULL;

	    /* Messages list */
	    free(mesg_list);
	    mesg_list = NULL;
	    total_messages = 0;

	    /* ImgRefs */
	    ImgRefClear(&xsw_imgref, &total_imgrefs);

	    /* Images */
	    ImgUnloadAll();

	    /* Viewscreen object labels */
	    VSLabelDeleteAll();

	    /* Cursors */
	    WidgetDestroyCursor(&xsw_cursor.scanner_lock);
	    WidgetDestroyCursor(&xsw_cursor.text);
	    WidgetDestroyCursor(&xsw_cursor.wait);
	    WidgetDestroyCursor(&xsw_cursor.translate);
	    WidgetDestroyCursor(&xsw_cursor.zoom);

	    /* Fonts */
	    OSWUnloadFont(&xsw_font.std);
	    OSWUnloadFont(&xsw_font.std_bold);
	    OSWUnloadFont(&xsw_font.con_heading);
	    OSWUnloadFont(&xsw_font.con_standard);
	    OSWUnloadFont(&xsw_font.con_message);
	    OSWUnloadFont(&xsw_font.vs_heading);
	    OSWUnloadFont(&xsw_font.vs_standard);

	    /* Colors */
	    XSWFreeColors();

	    /* Widget globals */
	    WidgetDestroyGlobals();

	    /* Reset keyboard repeat */
	    OSWKBAutoRepeatOn();

	    /* Disconnect from the GUI */
	    OSWGUIDisconnect();
	}

	/* Shutdown the reality engine */
	REngShutdown();

	/* Shutdown and delete the game controller */
	GCtlDelete(xsw_gctl);
	xsw_gctl = NULL;

#ifdef JS_SUPPORT
	/* Delete all the joystick mappings */
	JSMapsListClear(xsw_jsmap, total_jsmaps);
	xsw_jsmap = NULL;
	total_jsmaps = 0;
#endif

	/* Delete all the sndrefs */
	SndRefDeleteAll(&xsw_sndref, &total_sndrefs);

	/* Shutdown the sound */
	SoundShutdown(&sound);

	/* Delete all the ignored objects list */
	IgnoredObjListClear(ignored_obj_list, total_ignored_objs);
	ignored_obj_list = NULL;
	total_ignored_objs = 0;

	/* Delete all the OCS Names */
	OCSNClear(&xsw_ocsn, &total_ocsns);

	/* Delete the inrange objects list */
	DBInRangeClear();

        /* Delete the scanner contacts list */
        ScListClear(scanner_contact_list, total_scanner_contacts);   
        scanner_contact_list = NULL;
        total_scanner_contacts = 0;

	/* Delete all the objects & particles */
	DBDeleteAllObjects();
	ParticleClear(&xsw_particle, &total_particles);

	/* Shutdown the universe */
	UNVShutdown();

	/* Delete the universes list */
	UnvListClear(univ_item, total_univ_items);
	univ_item = NULL;
	total_univ_items = 0;
}



int main(int argc, char *argv[])
{
	int i, status;
	const char *arg;

	runlevel = 0;

	/* Set the signals to watch for */
	signal(SIGHUP, XSWSignalCB);
	signal(SIGINT, XSWSignalCB);
	signal(SIGQUIT, XSWSignalCB);
	signal(SIGABRT, XSWSignalCB);
	signal(SIGSEGV, XSWSignalCB);
	signal(SIGPIPE, XSWSignalCB);
	signal(SIGTERM, XSWSignalCB);
	signal(SIGCONT, XSWSignalCB);
	signal(SIGSTOP, XSWSignalCB);

	/* Handle help and version arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!strcasecmp(arg, "--help") ||
	       !strcasecmp(arg, "-help") || 
	       !strcasecmp(arg, "--h") ||   
	       !strcasecmp(arg, "-h") ||
	       !strcasecmp(arg, "-?") ||
	       !strcasecmp(arg, "/?")
	    )
	    {
		XSWPrintHelpStdout();
		return(0);
	    }
	    /* Version */
	    else if(!strcasecmp(arg, "--version") ||
		    !strcasecmp(arg, "-version")
	    )
	    {
		XSWPrintVersionStdout();
		return(0);
	    }
	}

	/* Initialize */
	status = XSWInit(argc, argv);
	if(status)
	{
	    XSWShutdown();
	    return(-status);
	}

	/* Manage */
	XSWMain(NULL);

	/* Save the configuration to file? */
	if(option.save_on_exit)
	{
	    /* Unmap the bridge to save coordinates */
	    BridgeWinUnmap(xsw_bridge_win);

	    CfgFileSave(
		fname.cfg,
		&option,
		xsw_vsmenu
	    );
	    UnvListFileSave(
		fname.universes,
		univ_item, total_univ_items,
		xsw_unvlist_win,
		NULL, NULL,
		XSWPrintWarning
	    );
	}

	/* Shutdown */
	XSWShutdown();

	return(0);
}
