#include <stdlib.h>
#include <string.h>

//#include "../include/string.h"

#include "../include/reality.h"

#include "imagepaths.h"  
#include "keymaps.h"
#include "msglist.h"
#include "msgwin.h"
#include "net.h"
#include "xsw.h"
#include "xswaction.h"

#include "../include/oswkeyvals.h"


static void MsgWinResizeCB(void *widget, void *data);
static void MsgWinDrawCB(void *widget, void *data);
static void MsgWinSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
);

static void MsgWinMarkCoordinates(
	xsw_msg_win_struct *mw,
	int start_x, int start_y,
	int end_x, int end_y
);
static void MsgWinScrollTimeoutCB(void *widget, void *data);
static void MsgWinAddScrollTimeout(
	xsw_msg_win_struct *mw, const time_t interval_ms
);
static void MsgWinRemoveScrollTimeout(
	xsw_msg_win_struct *mw
);

static void MsgWinPromptActivateCB(
	prompt_struct *prompt, const char *s, void *data
);
static void MsgWinPromptCancelCB(
	prompt_struct *prompt, void *data
);

void MsgWinMapPromptMode(
	xsw_msg_win_struct *mw,
	const char *label, const char *val, const int mode
);

xsw_msg_win_struct *MsgWinNew(void);
void MsgWinResize(xsw_msg_win_struct *mw);
void MsgWinQueueResize(xsw_msg_win_struct *mw);
void MsgWinDraw(xsw_msg_win_struct *mw);
void MsgWinQueueDraw(xsw_msg_win_struct *mw);
int MsgWinManage(
	xsw_msg_win_struct *mw, event_t *event,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void MsgWinMap(xsw_msg_win_struct *mw);
void MsgWinUnmap(xsw_msg_win_struct *mw);
void MsgWinDelete(xsw_msg_win_struct *mw);


#define MSG_WIN_DEF_WIDTH		640
#define MSG_WIN_DEF_HEIGHT		480

#define MSG_WIN_MIN_WIDTH		100
#define MSG_WIN_MIN_HEIGHT		100

#define FTOI_RNDOFF(x)	((int)((x) + 0.5f))

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Message window resize callback.
 */
static void MsgWinResizeCB(void *widget, void *data)
{
	xsw_msg_win_struct *mw = XSW_MESG_WIN(data);
	if(mw == NULL)
	    return;

	MsgWinResize(mw);
}

/*
 *	Message window draw callback.
 */
static void MsgWinDrawCB(void *widget, void *data)
{
	xsw_msg_win_struct *mw = XSW_MESG_WIN(data);
	if(mw == NULL)
	    return;

	if(mw->map_state)
	    MsgWinDraw(mw);
}

/*
 *	Message window scroll bar value changed callback.
 */
static void MsgWinSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
)
{
	xsw_msg_win_struct *mw = XSW_MESG_WIN(data);
	if(mw == NULL)
	    return;

	if(&mw->sb == sb)
	{
	    if(mw->map_state)
		MsgWinQueueDraw(mw);
	}
}


/*
 *	Marks messages by the specified rectangular coordinates.
 *
 *	Scroll bar positions will be applied.
 */
static void MsgWinMarkCoordinates(
	xsw_msg_win_struct *mw,
	int start_x, int start_y,
	int end_x, int end_y
)
{
	int	sb_x, sb_y,
		start_line, end_line,
		start_line_x, end_line_x,
		margin_x, margin_y,
		line_spacing, cwidth;
	scroll_bar_struct *sb;

	if(mw == NULL)
	    return;

	margin_x = MSG_WIN_MARGIN;
	margin_y = MSG_WIN_MARGIN;
	line_spacing = MAX(mw->line_spacing, 1);
	cwidth = MAX(mw->char_width, 1);
	sb = &mw->sb;
	sb_x = (int)SBarGetPositionX(sb);
	sb_y = (int)SBarGetPositionY(sb);

	/* Calculate start and end lines */
	start_line = CLIP(
	    (start_y - margin_y + sb_y) / line_spacing,
	    0, (total_messages - 1)
	);
	end_line = CLIP(
	    (end_y - margin_y + sb_y) / line_spacing,
	    0, (total_messages - 1)
	);

	/* Calculate the start and end line x positions (do not clip) */
	start_line_x = (start_x - margin_x + sb_x) / cwidth;
	end_line_x = (end_x - margin_x + sb_x) / cwidth;

#define MSG_SEL_SWAP_AS_NEEDED(_msg_)	{	\
 if((_msg_)->sel_start > (_msg_)->sel_end) {	\
  const int i = (_msg_)->sel_end;		\
  (_msg_)->sel_end = (_msg_)->sel_start;	\
  (_msg_)->sel_start = i;			\
} }

	/* Single line? */
	if(start_line == end_line)
	{
	    xsw_mesg_struct *msg = &mesg_list[start_line];
	    const int line_len = strlen(msg->message);
	    if(start_line_x < end_line_x)
	    {
		msg->sel_start = CLIP(start_line_x, 0, (line_len - 1));
		msg->sel_end = CLIP(end_line_x - 1, 0, (line_len - 1));
	    }
	    else if(start_line_x > end_line_x)
	    {
		msg->sel_start = CLIP(end_line_x, 0, (line_len - 1));
		msg->sel_end = CLIP(start_line_x - 1, 0, (line_len - 1));
	    }
	}
	/* Select multiple lines top to bottom? */
	else if(start_line < end_line)
	{
	    xsw_mesg_struct *msg = &mesg_list[start_line];
	    int i, line_len = strlen(msg->message);
	    if(start_line_x < line_len)
	    {
		msg->sel_start = CLIP(start_line_x, 0, (line_len - 1));
		msg->sel_end = line_len - 1;
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }

	    for(i = start_line + 1; i < end_line; i++)
	    {
		msg = &mesg_list[i];
		msg->sel_start = 0;
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
	    }

	    msg = &mesg_list[end_line];
	    line_len = strlen(msg->message);
	    if(end_line_x >= 0)
	    {
		msg->sel_start = 0;
		msg->sel_end = CLIP(end_line_x, 0, (line_len - 1));
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }
	}
	/* Select multiple lines bottom to top? */
	else if(end_line < start_line)
	{
	    xsw_mesg_struct *msg = &mesg_list[start_line];
	    int i, line_len = strlen(msg->message);
	    if(start_line_x >= 0)
	    {
		msg->sel_start = 0;
		msg->sel_end = CLIP(start_line_x - 1, 0, (line_len - 1));
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }

	    for(i = start_line - 1; i > end_line; i--)
	    {
		msg = &mesg_list[i];
		msg->sel_start = 0;
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
	    }

	    msg = &mesg_list[end_line];
	    line_len = strlen(msg->message);
	    if(end_line_x < line_len)
	    {
		msg->sel_start = CLIP(end_line_x, 0, (line_len - 1));
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }
	}
#undef MSG_SEL_SWAP_AS_NEEDED
}

/*
 *	Scroll messages timeout.
 */
static void MsgWinScrollTimeoutCB(void *widget, void *data)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	widget_global_struct *wglobal = &widget_global;
	xsw_msg_win_struct *mw = XSW_MESG_WIN(widget);
	if(mw == NULL)
	    return;

	w = mw->toplevel;
	if(w == None)
	    return;

	OSWGetPointerCoords(w, NULL, NULL, &x, &y);
	width = mw->width;
	height = mw->height;

	if(y < 0)
	{
	    scroll_bar_struct *sb = &mw->sb;
	    const int   dy = 0 - y,
			prev_y_pos = (int)SBarGetPositionY(sb);
	    SBarSetPositionY(sb, (float)(prev_y_pos - dy));
	    mw->sel_start_y += prev_y_pos -
		(int)SBarGetPositionY(sb);
	    MsgWinMarkCoordinates(
		mw,
		mw->sel_start_x,
		mw->sel_start_y,
		x, y
	    );
	}
	else if(y >= (int)height)
	{
	    scroll_bar_struct *sb = &mw->sb;
	    const int   dy = (y + 1 - (int)height),
			prev_y_pos = (int)SBarGetPositionY(sb);
	    SBarSetPositionY(sb, (float)(prev_y_pos + dy));
	    mw->sel_start_y += prev_y_pos -
		(int)SBarGetPositionY(sb);
	    MsgWinMarkCoordinates(
		mw,
		mw->sel_start_x,
		mw->sel_start_y,
		x, y
	    );
	}

	/* Schedual next timeout */
	MsgWinAddScrollTimeout(
	    mw, wglobal->prompt_scroll_repeat_int
	);
}

/*
 *	Adds a scroll messages timeout.
 */
static void MsgWinAddScrollTimeout(
	xsw_msg_win_struct *mw, const time_t interval_ms
)
{
	if(mw == NULL)
	    return;

	WidgetTimeoutAdd(
	    MsgWinScrollTimeoutCB,
	    mw, NULL,
	    interval_ms
	);
}

/*
 *	Removes any scroll messages timeouts.
 */
static void MsgWinRemoveScrollTimeout(
	xsw_msg_win_struct *mw
)
{
	if(mw == NULL)
	    return;

	WidgetTimeoutRemoveByFuncWidget(
	    MsgWinScrollTimeoutCB, mw
	);
}


/*
 *	Prompt activate callback.
 */
static void MsgWinPromptActivateCB(
	prompt_struct *prompt, const char *value, void *data
)
{
	int mode;
	int obj_num;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;
	xsw_player_values_struct *pvals = &player_values;
	xsw_msg_win_struct *mw = XSW_MESG_WIN(data);
	if((value == NULL) || (mw == NULL))
	    return;

	mode = prompt_mode;

	/* Close the prompt and reset the prompt mode */
	WidgetUnmap(&mw->prompt);
	prompt_mode = XSW_PROMPT_NONE;

	switch(mode)
	{
	  case XSW_PROMPT_CONNECT:
	    XSWConnect(value);
	    break;

	  case XSW_PROMPT_CLIENT_CMD:
	    CmdHandleInput(value);
	    break;

	  case XSW_PROMPT_SERVER_CMD:
	    if(XSW_CON_STATE_IS_CONNECTED(net))
	    {
		NetSendExec(value);
	    }
	    else
	    {
		XSWPrintWarning("Not connected.");
	    }
	    break;

	  case XSW_PROMPT_MESSAGE:
	    obj = net->player_obj_ptr;
	    obj_num = net->player_obj_num;
	    if(obj != NULL)
	    {
		NetSendComMessage(
		    obj_num,
		    -1,
		    obj->com_channel,
		    value
		);
	    }
	    else
	    {
		XSWPrintWarning("Not connected or not logged in.");
	    }
	    break;

	  case XSW_PROMPT_SHIELD_FREQ:
	    obj = net->player_obj_ptr;
	    obj_num = net->player_obj_num;
	    if(obj != NULL)
	    {
		const float vf = CLIP(
		    ATOF(value), SW_FREQ_MIN, SW_FREQ_MAX
		);
		NetSendSetShields(
		    obj_num,
		    obj->shield_state,
		    vf
		);
		BridgeWinDrawPanel(
		    net->player_obj_num,
		    BW_PANEL_PSHIELDFREQ   
		);
	    }
	    else
	    {
		XSWPrintWarning("Not connected or not logged in.");
	    }
	    break;

	  case XSW_PROMPT_WEAPONS_FREQ:
	    obj = net->player_obj_ptr;
	    obj_num = net->player_obj_num;
	    if(obj != NULL)
	    {
		const float vf = CLIP(
		    ATOF(value), SW_FREQ_MIN, SW_FREQ_MAX
		);
		pvals->weapon_freq = vf;
		BridgeWinDrawPanel(
		    net->player_obj_num,
		    BW_PANEL_PWEAPONFREQ
		);
	    }
	    else
	    {
		XSWPrintWarning("Not connected or not logged in.");
	    }
	    break;

	  case XSW_PROMPT_INTERCEPT:
	    obj = net->player_obj_ptr;
	    obj_num = net->player_obj_num;
	    if(obj != NULL)
	    {
		NetSendIntercept(
		    obj_num,
		    !STRISEMPTY(value) ? value : "#off"
		);
	    }
	    else
	    {
		XSWPrintWarning("Not connected or not logged in.");
	    }
	    break;

	  /* Com channel */
	  case XSW_PROMPT_COM_CHANNEL:
	    obj = net->player_obj_ptr;
	    obj_num = net->player_obj_num;
	    if(obj != NULL)
	    {
		const int vi = (int)(CLIP(
		    ATOF(value), SW_FREQ_MIN, SW_FREQ_MAX
		) * 100.0f);
		NetSendSetChannel(
		    obj_num,
		    vi
		);
	    }
	    else
	    {
		XSWPrintWarning("Not connected or not logged in.");
	    }
	    break;
	}
}

/*
 *	Prompt cancel callback.
 */
static void MsgWinPromptCancelCB(
	prompt_struct *prompt, void *data
)
{
	xsw_msg_win_struct *mw = XSW_MESG_WIN(data);
	if(mw == NULL)
	    return;

	/* Close the prompt and reset the prompt mode */
	WidgetUnmap(&mw->prompt);
	prompt_mode = XSW_PROMPT_NONE;
}


/*
 *	Maps the message window's prompt.
 */
void MsgWinMapPromptMode(
	xsw_msg_win_struct *mw,
	const char *label, const char *val, const int mode
)
{
	prompt_struct *prompt = (mw != NULL) ? &mw->prompt : NULL;
	if(prompt == NULL)
	    return;

	if(label != NULL)
	    PromptSetLabel(prompt, label);

	PromptSetS(prompt, val);

	prompt_mode = mode;

	PromptMarkBuffer(prompt, PROMPT_POS_END);

	PromptMap(prompt);
	prompt->has_focus = True;
}


/*
 *	Creates a new message window.
 */
xsw_msg_win_struct *MsgWinNew(void)
{
	int x, y, cwidth, ascent, descent, font_height;
	unsigned int width, height;
	font_t *font;
	pixmap_t pixmap;
	prompt_struct *prompt;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	scroll_bar_struct *sb;
	widget_global_struct *wglobal = &widget_global;
	xsw_option_struct *opt = &option;
	xsw_msg_win_struct *mw = XSW_MESG_WIN(calloc(
	    1, sizeof(xsw_msg_win_struct)
	));
	if(mw == NULL)
	    return(NULL);

	font = xsw_font.con_message;
	OSWGetCharBounds(
	    font, 'X',
	    NULL, NULL, &cwidth, &ascent, &descent
	);
	font_height = ascent + descent;

	mw->map_state = False;
	mw->has_focus = False;
	mw->visibility_state = VisibilityFullyObscured;
	mw->x = opt->msgw_x;
	mw->y = opt->msgw_y;
	mw->width = opt->msgw_width;
	mw->height = opt->msgw_height;
	if(mw->width < MSG_WIN_MIN_WIDTH)
	    mw->width = MSG_WIN_DEF_WIDTH;
	if(mw->height < MSG_WIN_MIN_HEIGHT)
	    mw->height = MSG_WIN_DEF_HEIGHT;
	if((mw->x + (int)mw->width) < 0)
	    mw->x = 0;
	if((mw->y + (int)mw->height) < 0)
	    mw->y = 0;
	if(mw->x > (int)gui->display_width)
	    mw->x = 0;
	if(mw->y > (int)gui->display_height)
	    mw->y = 0;
	mw->font = font;

	mw->line_spacing = MAX(font_height, 1);
	mw->char_width = MAX(cwidth, 1);
	mw->longest_line = 0;

	/* Create toplevel */
	x = mw->x;
	y = mw->y;
	width = mw->width;
	height = mw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	mw->toplevel = w;
	mw->toplevel_buf = None;

	/* WM Properties */
	if(ImgIsAllocated(XSW_IMG_ICON_XSW))
	    pixmap = OSWCreatePixmapFromImage( 
		xsw_image[XSW_IMG_ICON_XSW]->image
	    );
	else
	    pixmap = wglobal->std_icon_pm;
	OSWSetWindowWMProperties(
	    w,
	    PROG_NAME " - Messages",	/* Title */
	    "Messages",		/* Icon title */
	    pixmap,		/* Icon */
	    False,		/* Let WM set coordinates? */
	    x, y,
	    MSG_WIN_MIN_WIDTH, MSG_WIN_MIN_HEIGHT,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_STANDARD,
	    0, NULL
	);

	WidgetSetWindowCursor(w, xsw_cursor.text);
	OSWSetWindowInput(
	    w,
	    OSW_EVENTMASK_TOPLEVEL |
	    ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask
	);

	parent = w;


	/* Create prompt */
	prompt = &mw->prompt;
	PromptInit(
	    prompt, parent,
	    16, MAX((int)mw->height - 46, 0),
	    MAX((int)mw->width - 32, 100), 30,
	    PROMPT_STYLE_RAISED,
	    "Command:",
	    XSW_DEF_PROMPT_BUF_LEN, XSW_DEF_PROMPT_HIST_BUFS
	);
	prompt->activate_cb = MsgWinPromptActivateCB;
	prompt->activate_data = mw;
	prompt->cancel_cb = MsgWinPromptCancelCB;
	prompt->cancel_data = mw;

	/* Create scroll bars */
	sb = &mw->sb;
	SBarInit(
	    sb, parent,
	    mw->width, mw->height,
	    mw->width,
	    mw->line_spacing * (total_messages + 5)
	);
	sb->x_step_inc = FTOI_RNDOFF(mw->width * 0.10f);
	sb->y_step_inc = mw->line_spacing;
	SBarSetPositionX(sb, 0.0f);
	SBarSetPositionY(sb, (float)sb->total_height);
	sb->value_changed_cb = MsgWinSBValueChangedCB;
	sb->value_changed_data = mw;

	/* Unmark all message buffers, this will reset them too */
	MsgListUnmarkAll();

	return(mw);
}

/*
 *	Resizes the message window.
 */
void MsgWinResize(xsw_msg_win_struct *mw)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	scroll_bar_struct *sb;

	if(mw == NULL)
	    return;

	w = mw->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	mw->x = x;
	mw->y = y;
	mw->width = width;
	mw->height = height;

	OSWDestroyPixmap(&mw->toplevel_buf);

	/* Move and resize the prompt */
	width = (unsigned int)MAX(
	    (int)mw->width - (2 * 10) - SCROLLBAR_YBAR_WIDTH,
	    100  
	);
	height = 30;
	x = 10;
	y = MAX(
	    (int)mw->height - (int)height - 10 - SCROLLBAR_XBAR_HEIGHT,
	    0
	);
	OSWMoveResizeWindow(
	    mw->prompt.toplevel, x, y, width, height
	);
	WidgetQueueResize(&mw->prompt);

	/* Resize the scroll bar */
	sb = &mw->sb;
	SBarQueueResize(
	    sb,
	    mw->width, mw->height,
	    (mw->longest_line * mw->char_width) +
		SCROLLBAR_YBAR_WIDTH + (2 * MSG_WIN_MARGIN),
	    mw->line_spacing * (total_messages + 5)
	);
	/* Scroll to end */
	SBarSetPositionX(sb, 0.0f);
	SBarSetPositionY(sb, (float)sb->total_height);

	/* Resize the messages window background image */
	ImgResize(
	    XSW_IMG_MESGWIN_BG,
	    mw->width,
	    mw->height
	);
}

/*
 *	Queues the message window to resize.
 */
void MsgWinQueueResize(xsw_msg_win_struct *mw)
{
	if(mw == NULL)
	    return;

	WidgetQueueResizeAction(MsgWinResizeCB, NULL, mw);
}

/*
 *	Draws the message window;
 */
void MsgWinDraw(xsw_msg_win_struct *mw)
{
	int sb_x, sb_y, line_spacing, cwidth;
	unsigned int width, height;
	pixmap_t pixmap;
	win_t w;
	image_t *img;
	scroll_bar_struct *sb;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(mw == NULL)
	    return;

	line_spacing = MAX(mw->line_spacing, 1);
	cwidth = MAX(mw->char_width, 1);
	w = mw->toplevel;
	width = MAX(mw->width, 1);
	height = MAX(mw->height, 1);
	sb = &mw->sb;
	sb_x = (int)SBarGetPositionX(sb);
	sb_y = (int)SBarGetPositionY(sb);

	if(mw->toplevel_buf == None)
	    OSWCreatePixmap(
		&mw->toplevel_buf, width, height
	    );
	pixmap = mw->toplevel_buf;

	/* Draw background */
	if(ImgIsAllocated(XSW_IMG_MESGWIN_BG))
	{
	    img = xsw_image[XSW_IMG_MESGWIN_BG]->image;
	    if((img != NULL) && !wglobal->force_mono)
		OSWPutImageToDrawable(img, pixmap);
	    else
		OSWClearPixmap(
		    pixmap, width, height,
		    gui->black_pix
		);
	}

	/* Draw text */
	if((w != None) && (pixmap != None))
	{
	    int	i, x, y, c_pos, lines_drawn, lines_visible,
		lbearing, ascent, descent, font_height;
	    font_t *font = mw->font;
	    xsw_mesg_struct *msg;

	    /* Calculate the starting line and its coordinates */
	    i = MAX((sb_y / line_spacing), 0);
	    x = MSG_WIN_MARGIN - sb_x;
	    y = MSG_WIN_MARGIN - (sb_y % line_spacing);
	    if(i > 0)
	    {
		i--;
		y -= line_spacing;
	    }

	    lines_drawn = 0;
	    lines_visible = MAX(
		FTOI_RNDOFF((float)height / (float)line_spacing),
		0
	    );

	    OSWSetFont(font); 

	    /* Draw the messages */
	    while(lines_drawn < lines_visible)
	    {
		if(i >= total_messages)
		    break;

		msg = &mesg_list[i];

		/* Is this message not selected? */
	        if((msg->sel_start < 0) || (msg->sel_end < 0))
	        {
		    /* Draw unselected text */
		    int x2, y2;
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    msg->fg_pix
		    );
		    OSWGetTextBounds(
			font, msg->message, -1,
			&lbearing, NULL, NULL,
			&ascent, &descent
		    );
		    font_height = ascent + descent;
		    x2 = x;
		    y2 = y + FTOI_RNDOFF(
			(float)(line_spacing - font_height) / 2.0f
		    );
		    OSWDrawString(
			pixmap,
			x2 - lbearing,
			y2 + ascent,
			msg->message
		    );
	        }
	        else
	        {
		    /* Draw selected text background */
		    int	x2, y2;
		    unsigned int w2, h2;

		    OSWGetCharBounds(
			font, msg->message[0],
			&lbearing, NULL, NULL,
			&ascent, &descent
		    );
		    font_height = ascent + descent;
		    x2 = x + (msg->sel_start * cwidth);
		    y2 = y;
		    w2 = (unsigned int)MAX(
			(msg->sel_end - msg->sel_start + 1) * cwidth,
			1
		    );
		    h2 = line_spacing;

		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    wglobal->surface_selected_pix
		    );
		    OSWDrawSolidRectangle(pixmap, x2, y2, w2, h2);

		    /* Draw first unselected text segment */
		    x2 = x;
		    y2 = y + FTOI_RNDOFF(
			(float)(line_spacing - font_height) / 2.0f
		    );
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    msg->fg_pix
		    );
		    OSWDrawStringLimited(
		        pixmap,
		        x2 - lbearing,
		        y2 + ascent,
		        msg->message,
		        msg->sel_start
		    );
		    /* Draw the selected text segment */
		    c_pos = msg->sel_start;
		    if((c_pos >= 0) && (c_pos < MSG_LIST_MSG_LEN))
	            {
			x2 = x + (c_pos * cwidth);
			OSWSetForeground(
			    wglobal->force_mono ?
				gui->black_pix :
				wglobal->selected_text_pix
			);
			OSWDrawStringLimited(
			    pixmap,
			    x2 - lbearing,
			    y2 + ascent,
			    msg->message + c_pos,
			    msg->sel_end - c_pos + 1
			);
		    }
		    /* Draw last unselected text segment */
		    c_pos = msg->sel_end + 1;
		    if((c_pos >= 0) && (c_pos < MSG_LIST_MSG_LEN))
		    {
			x2 = x + (c_pos * cwidth);
			OSWSetForeground(
			    wglobal->force_mono ?
				gui->white_pix :
				msg->fg_pix
			);
			OSWDrawStringLimited(
			    pixmap,
			    x2 - lbearing,
			    y2 + ascent,
			    msg->message + c_pos,
			    MSG_LIST_MSG_LEN - msg->sel_end
			);
		    }
	        }

		y += line_spacing;
		i++;
		lines_drawn++;
	    }
	}
	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the message window to draw.
 */
void MsgWinQueueDraw(xsw_msg_win_struct *mw)
{
	if(mw == NULL)
	    return;

	WidgetQueueDrawAction(MsgWinDrawCB, NULL, mw);
}

/*
 *	Manages the message window event.
 */
int MsgWinManage(
	xsw_msg_win_struct *mw, event_t *event,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int events_handled = 0;
	time_t t;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	event_visibility_t *visibility;
	xsw_net_parms_struct *net = &net_parms;

	if((mw == NULL) || (event == NULL))
	    return(events_handled);

	if(!mw->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == mw->toplevel)
	    {
		MsgWinQueueResize(mw);
		if(mw->map_state)
		{
		    MsgWinQueueDraw(mw);
		    SBarQueueDraw(&mw->sb);
		}
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == mw->toplevel)
	    {
		MsgWinQueueDraw(mw);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!mw->has_focus)
		return(events_handled);

	    key = &event->xkey;
	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Manage only the prompt if it is mapped */
	    if(mw->prompt.map_state)
	    {
		events_handled += WidgetManage(&mw->prompt, event);
		return(events_handled);
	    }

	    /* Skip modifier keys */
	    if(OSWIsModifierKey(keycode))
		return(events_handled);

	    /* Help */
	    if(keycode == KeymapGetKeycode(XSW_KM_HELP))
	    {
		XSWPrintHelp();
		events_handled++;
	    }
	    /* Exit */
	    else if(keycode == KeymapGetKeycode(XSW_KM_EXIT))
	    {
		XSWAction(
		    XSW_ACTION_EXIT_CONFIRM, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Vector Lights */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LIGHTS_VECTOR))
	    {
		XSWAction(
		    XSW_ACTION_LIGHTS_VECTOR, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Strobes */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LIGHTS_STROBE))
	    {
		XSWAction(
		    XSW_ACTION_LIGHTS_STROBE, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle External Lumination */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LIGHTS_LUMINATION))
	    {
		XSWAction(
		    XSW_ACTION_LIGHTS_LUMINATION, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Set Weapon Frequency */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_FREQ))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_FREQ, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Shields */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SHIELD))
	    {
		XSWAction(
		    XSW_ACTION_SHIELD, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Set Shield Frequency */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SHIELD_FREQ))
	    {
		XSWAction(
		    XSW_ACTION_SHIELD_FREQ, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Damage Control */
	    else if(keycode == KeymapGetKeycode(XSW_KM_DMGCTL))
	    {
		XSWAction(
		    XSW_ACTION_DMGCTL, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Cloak */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CLOAK))
	    {
		XSWAction(
		    XSW_ACTION_CLOAK, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Set Intercept */
	    else if(keycode == KeymapGetKeycode(XSW_KM_INTERCEPT_SET))
	    {
		XSWAction(
		    XSW_ACTION_INTERCEPT_SET, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Hail */
	    else if(keycode == KeymapGetKeycode(XSW_KM_HAIL))
	    {
		XSWAction(
		    XSW_ACTION_HAIL, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Set Com Channel */
	    else if(keycode == KeymapGetKeycode(XSW_KM_COM_CHANNEL_SET))
	    {
		XSWAction(
		    XSW_ACTION_COM_CHANNEL_SET, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Engine State */
	    else if(keycode == KeymapGetKeycode(XSW_KM_ENGINE_STATE))
	    {
		XSWAction(
		    XSW_ACTION_ENGINE_STATE, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Throttle Mode */
	    else if(keycode == KeymapGetKeycode(XSW_KM_THROTTLE_MODE))
	    {
		XSWAction(
		    XSW_ACTION_THROTTLE_MODE, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Lock Next */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LOCK_NEXT))
	    {
		XSWAction(
		    XSW_ACTION_LOCK_NEXT, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Lock Nearest */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LOCK_NEAREST))
	    {
		XSWAction(
		    XSW_ACTION_LOCK_NEAREST, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Unlock */  
	    else if(keycode == KeymapGetKeycode(XSW_KM_UNLOCK))
	    {
		XSWAction(
		    XSW_ACTION_UNLOCK, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Weapons Online/Offline */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPONS_ONLINE))
	    {
		XSWAction(
		    XSW_ACTION_WEAPONS_ONLINE, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Weapon Select Next */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_NEXT))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_NEXT, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Weapon Select Prev */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_PREV))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_PREV, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 1 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_1))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_1, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 2 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_2))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_2, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 3 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_3))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_3, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 4 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_4))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_4, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 5 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_5))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_5, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 6 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_6))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_6, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 7 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_7))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_7, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 8 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_8))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_8, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 9 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_9))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_9, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 10 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_10))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_10, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Tractor Beam Lock */
	    else if(keycode == KeymapGetKeycode(XSW_KM_TRACTOR_BEAM_LOCK))
	    {
		if(key->state & OSW_KEY_STATE_SHIFT)
		    XSWAction(   
			XSW_ACTION_TRACTOR_BEAM_UNLOCK, NULL,
			mw,
			cur_ms, cur_sec, tc
		    );
		else
		    XSWAction(
			XSW_ACTION_TRACTOR_BEAM_LOCK, NULL,
			mw,
			cur_ms, cur_sec, tc
		    );
		events_handled++;
	    }
	    /* Send Message */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SEND_MESSAGE))
	    {
		XSWAction(
		    XSW_ACTION_SEND_MESSAGE, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Auto Destruct */
	    else if(keycode == KeymapGetKeycode(XSW_KM_AUTO_DESTRUCT))
	    {
		XSWAction(
		    XSW_ACTION_AUTO_DESTRUCT, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Energy Saver Mode */
	    else if(keycode == KeymapGetKeycode(XSW_KM_ENERGY_SAVER_MODE))
	    {
		XSWAction(
		    XSW_ACTION_ENERGY_SAVER_MODE, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Net Interval Inc */
	    else if(keycode == KeymapGetKeycode(XSW_KM_NET_INTERVAL_INC))
	    {
		XSWAction(
		    XSW_ACTION_NET_INTERVAL_INC, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Net Interval Dec */
	    else if(keycode == KeymapGetKeycode(XSW_KM_NET_INTERVAL_DEC))
	    {
		XSWAction(
		    XSW_ACTION_NET_INTERVAL_DEC, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Economy */
	    else if(keycode == KeymapGetKeycode(XSW_KM_ECONOMY))
	    {
		XSWAction(
		    XSW_ACTION_ECONOMY, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Starchart */
	    else if(keycode == KeymapGetKeycode(XSW_KM_STARCHART))
	    {
		XSWAction(
		    XSW_ACTION_STARCHART, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Connect */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CONNECT))
	    {
		char *buf = (char *)malloc(
		    STRLEN(net->login_name) +
		    STRLEN(net->login_password) +
		    STRLEN(net->address) +
		    40 +
		    80
		);
		sprintf(
		    buf,
		    "swserv://%s:%s@%s:%i",
		    (net->login_name != NULL) ?
			net->login_name : "",
		    (net->login_password != NULL) ?
			net->login_password : "",
		    (net->address != NULL) ?
			net->address : "",
		    net->port
		);
		MsgWinMapPromptMode(
		    mw,
		    "Connect:",
		    buf,
		    XSW_PROMPT_CONNECT
		);
		free(buf);
		events_handled++;
	    }
	    /* Connect Last */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CONNECT_LAST))
	    {
		XSWAction(
		    XSW_ACTION_CONNECT_LAST, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Disconnect */
	    else if(keycode == KeymapGetKeycode(XSW_KM_DISCONNECT))
	    {
		XSWAction(
		    XSW_ACTION_DISCONNECT, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Refresh */
	    else if(keycode == KeymapGetKeycode(XSW_KM_REFRESH))
	    {
		XSWAction(
		    XSW_ACTION_REFRESH, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Client Command */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CLIENT_CMD))
	    {
		XSWAction(
		    XSW_ACTION_CLIENT_CMD, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Server Command */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SERVER_CMD))
	    {
		XSWAction(
		    XSW_ACTION_SERVER_CMD, NULL,
		    mw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }

	    /* Close */
	    else if(keyval == OSW_KEY_Escape)
	    {
		MsgWinUnmap(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case KeyRelease:
	    if(!mw->has_focus)
		return(events_handled);

	    key = &event->xkey;
	    keycode = key->keycode;

	    /* Manage only the prompt if it is mapped */
	    if(mw->prompt.map_state)
	    {
		events_handled += WidgetManage(&mw->prompt, event);
		return(events_handled);
	    }
	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == mw->toplevel)
	    {
		scroll_bar_struct *sb = &mw->sb;

		/* Set the scroll bar in focus */
		sb->has_focus = True;

		switch(button->button)
		{
		  case Button1:
		    MsgListUnmarkAll();
		    mw->sel_in_progress = True;
		    mw->sel_start_x = button->x;
		    mw->sel_start_y = button->y;
		    MsgWinQueueDraw(mw);
		    break;
		  case Button2:
		    break;
		  case Button3:
		    break;
		  case Button4:
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) - sb->y_step_inc
		    );
		    break;
		  case Button5:
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) + sb->y_step_inc
		    );
		    break;
		}

		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == mw->toplevel)
	    {
		if(mw->sel_in_progress)
		{
		    MsgWinRemoveScrollTimeout(mw);
		    MsgListCopyToClipboard(mw->toplevel);
		    mw->sel_in_progress = False;
		    MsgListQueueDrawAllWindows();
		}
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(w == mw->toplevel)
	    {
		/* Mark text */
		if(mw->sel_in_progress)
		{
		    MsgListUnmarkAll();
		    MsgWinMarkCoordinates(
			mw,
		        mw->sel_start_x,
		        mw->sel_start_y,
		        motion->x,
		        motion->y
		    );
		    MsgWinQueueDraw(mw);
		    events_handled++;
		}
	    }
	    break;

	  case LeaveNotify:
	    if(w == mw->toplevel)
	    {
		if(mw->sel_in_progress)
		{
		    MsgWinRemoveScrollTimeout(mw);
		    MsgWinAddScrollTimeout(mw, 0l);
		}
		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == mw->toplevel)
	    {
		mw->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == mw->toplevel)
	    {
		mw->has_focus = False;
		events_handled++;
	    }
	    break;

	  case MapNotify:
	    if(w == mw->toplevel)
	    {
		if(!mw->map_state)
		    MsgWinMap(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case UnmapNotify:
	    if(w == mw->toplevel)
	    {
		if(mw->map_state)
		    MsgWinUnmap(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == mw->toplevel)
	    {
		mw->visibility_state = visibility->state;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case SelectionClear:
	    if(event->xselectionclear.window == mw->toplevel)
	    {
		MsgListUnmarkAll();
		MsgListQueueDrawAllWindows();
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(mw->toplevel, event, &t))
	    {
		OSWSetInputFocus(mw->toplevel, RevertToParent, t);
		events_handled++;
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(mw->toplevel, event))
	    {
		MsgWinUnmap(mw);
		events_handled++;
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(&mw->sb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&mw->prompt, event);

	return(events_handled);
}

/*
 *	Maps the message window.
 */
void MsgWinMap(xsw_msg_win_struct *mw)
{
	win_t w;
	scroll_bar_struct *sb;

	if(mw == NULL)
	    return;

	w = mw->toplevel;
	if(w == None)
	    return;

	/* Reload the background image as needed */
	ImgOpenFile(XSW_IMG_MESGWIN_BG, IMGPATH_MESGWIN_BG, False);

	/* Resize the background image */
	ImgResize(
	    XSW_IMG_MESGWIN_BG,
	    mw->width,
	    mw->height
	);

	/* Scroll to end */
	sb = &mw->sb;
	SBarSetPositionY(sb, (float)sb->total_height);

	/* Map */
	if(!mw->map_state)
	{
	    OSWMapRaised(w);
	    mw->map_state = True;
	}

	/* Focus */
	OSWGrabFocus(w);
}

/*      
 *	Unmaps the message window.
 */
void MsgWinUnmap(xsw_msg_win_struct *mw)
{
	xsw_option_struct *opt = &option;

	if(mw == NULL)
	    return;

	/* Record the window position */
	OSWGetWindowRootOrigin(
	    mw->toplevel, &opt->msgw_x, &opt->msgw_y
	);
	opt->msgw_width = mw->width;
	opt->msgw_height = mw->height;

	WidgetUnmap(&mw->prompt);
	if(prompt_mode != XSW_PROMPT_NONE)
	    prompt_mode = XSW_PROMPT_NONE;

	/* Unmap */
	OSWUnmapWindow(mw->toplevel);
	mw->map_state = False;
	mw->has_focus = False;
	mw->visibility_state = VisibilityFullyObscured;

	/* Destroy big buffers */
	OSWDestroyPixmap(&mw->toplevel_buf);

	/* Unload message screen background */
	ImgUnload(XSW_IMG_MESGWIN_BG);
}

/*
 *	Deletes the message window.
 */
void MsgWinDelete(xsw_msg_win_struct *mw)
{
	if(mw == NULL)
	    return;

	MsgWinRemoveScrollTimeout(mw);

	WidgetDestroy(&mw->sb);
	WidgetDestroy(&mw->prompt);

	OSWDestroyPixmap(&mw->toplevel_buf);
	OSWDestroyWindow(&mw->toplevel);

	free(mw);
}
