#include <stdlib.h>
#include <string.h>

#include "xsw.h"		/* Has to be first to define image_t */

#include "ocsnames.h"


xsw_ocsn_struct *OCSNGetPtr(
	xsw_ocsn_struct **list, const int total, const int ocsn_num
);
int OCSNGetByCode(
	xsw_ocsn_struct **list, const int total, const int code
);

static void OCSNReset(xsw_ocsn_struct *ocsn);

void OCSNClear(
	xsw_ocsn_struct ***list, int *total
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the OCSN pointer.
 */
xsw_ocsn_struct *OCSNGetPtr(
	xsw_ocsn_struct **list, const int total, const int ocsn_num
)
{
	if((ocsn_num < 0) ||
	   (ocsn_num >= total) ||
	   (list == NULL)
	)
	    return(NULL);
	else
	    return(list[ocsn_num]);
}

/*
 *	Returns an OCS Name who's code matches the specified code.
 *
 *	Can return -1 on error.
 */
int OCSNGetByCode(
	xsw_ocsn_struct **list, const int total, const int code
)
{
	int i;
	xsw_ocsn_struct *ocsn;

	/* The OCS Name  code may not be be garbage */
	if(code <= OCS_TYPE_GARBAGE)
	    return(-1);

	for(i = 0; i < total; i++)
	{
	    ocsn = list[i];
	    if(ocsn == NULL)
		continue;

	    if(ocsn->code == code)
		return(i);
	}

	return(-1);
}


/*
 *	Resets the OCSN.
 */
static void OCSNReset(xsw_ocsn_struct *ocsn)
{
	if(ocsn == NULL)
	    return;

	ocsn->code = OCS_TYPE_GARBAGE;
	free(ocsn->name);
	ocsn->name = NULL;
	OSWDestroyImage(&ocsn->icon);
	OSWDestroyImage(&ocsn->eco_icon);
}

/*
 *	Deletes all the OCSNs.
 */
void OCSNClear(xsw_ocsn_struct ***list, int *total)
{
	int i;
	xsw_ocsn_struct *ocsn;

	for(i = 0; i < *total; i++)
	{
	    ocsn = (*list)[i];
	    OCSNReset(ocsn);
	    free(ocsn);
	}
	free(*list);
	*list = NULL;
	*total = 0;
}
