#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/swsoundcodes.h"

#include "gctl.h"
#include "sndrefs.h"
#include "sound.h"
#include "keymapwin.h"
#ifdef JS_SUPPORT
# include "jsmapwin.h"
#endif
#include "vsmenu.h"
#include "vsmenuwin.h"
#include "net.h"
#include "optwin.h"
#include "xsw.h"
#include "xswaction.h"
#include "cfgfile.h"


void OptWinFetchValues(xsw_options_win_struct *optwin);
void OptWinApplyChanges(xsw_options_win_struct *optwin);
void OptWinSetDefaults(xsw_options_win_struct *optwin);
int OptWinSaveChanges(xsw_options_win_struct *optwin);

void OptWinTestSoundPBCB(push_button_struct *btn, void *data);
void OptWinBrowseOCSNsPBCB(push_button_struct *btn, void *data);
void OptWinBrowseImgRefsPBCB(push_button_struct *btn, void *data);
void OptWinBrowseSndRefsPBCB(push_button_struct *btn, void *data);
void OptWinBrowseJSCalPBCB(push_button_struct *btn, void *data);
void OptWinKeymapPBCB(push_button_struct *btn, void *data);
#ifdef JS_SUPPORT
void OptWinJSMapPBCB(push_button_struct *btn, void *data);
#endif
void OptWinConfigureVSMenuPBCB(push_button_struct *btn, void *data);
void OptWinRefreshMemoryPBCB(push_button_struct *btn, void *data);

void OptWinOKPBCB(push_button_struct *btn, void *data);
void OptWinApplyPBCB(push_button_struct *btn, void *data);
void OptWinSavePBCB(push_button_struct *btn, void *data);
void OptWinDefaultsPBCB(push_button_struct *btn, void *data);
void OptWinCancelPBCB(push_button_struct *btn, void *data);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)

/* Is a within inclusive range of min and max? */
#define INRANGEINC(a,min,max)	((a >= min) && (a <= max))


/*
 *	Sets the options window values to that of the global options.
 */
void OptWinFetchValues(xsw_options_win_struct *optwin)
{
	int i;
	char *s;
	xsw_option_struct *opt = &option;

	if(optwin == NULL)
	    return;

	/* General */

	/* Show Universes List On Startup */
	CBtnSetActive(
	    &optwin->show_univ_list_win_cb,
	    opt->auto_map_univ_list_win
	);

	/* Show Economy Window On Orbit */
	CBtnSetActive(
	    &optwin->show_eco_win_cb,
	    opt->auto_map_eco_win
	);

	/* CPU Priority */
	PUListSetSelItem(
	    &optwin->cpu_priority_pulist,
	    opt->cpu_priority
	);

	/* Controller Type */
	switch(opt->controller)
	{
	  case XSW_CONTROLLER_POINTER:
	    i = 2;
	    break;
	  case XSW_CONTROLLER_JOYSTICK:
	    i = 1;
	    break;
	  default:	/* XSW_CONTROLLER_KEYBOARD */
	    i = 0;
	    break;
	}
	PUListSetSelItem(&optwin->controller_pulist, i);

/* Disable joystick option as needed */
/* Disable pointer */

	/* Throttle mode */
	switch(opt->throttle_mode)
	{
	  case XSW_THROTTLE_MODE_INCREMENTAL:
	    i = 2;
	    break;
	  case XSW_THROTTLE_MODE_BIDIRECTIONAL:
	    i = 1;
	    break;
	  default:	/* XSW_THROTTLE_MODE_NORMAL */
	    i = 0;
	    break;
	}
	RadioSetActive(&optwin->throttle_mode_radio, i);

	/* Units */
	switch(opt->units)
	{
	  case XSW_UNITS_XSW:
	    i = 3;
	    break;
	  case XSW_UNITS_PARSECS:
	    i = 2;
	    break;
	  case XSW_UNITS_ENGLISH:
	    i = 1;
	    break;
	  default:	/* XSW_UNITS_METRIC */
	    i = 0;
	    break;
	}
	PUListSetSelItem(&optwin->units_pulist, i);

	/* Auto zoom */
	CBtnSetActive(
	    &optwin->auto_zoom_cb,
	    opt->auto_zoom
	);

	/* Local updates */
	CBtnSetActive(
	    &optwin->local_updates_cb,
	    opt->local_updates
	);

	/* Scanner contacts */
	CBtnSetActive(
	    &optwin->notify_scanner_contacts_cb,
	    opt->notify_scanner_contacts
	);

#if 0
	/* OCS Names file */
	s = STRDUP(fname.ocsn);
	StripParentPath(s, dname.etc);
	PromptSetS(&optwin->ocsn_prompt, s);
	free(s);

	/* ImgRefs file */
	s = STRDUP(fname.imgrefs);
	StripParentPath(s, dname.images);
	PromptSetS(&optwin->imgrefs_prompt, s);
	free(s);

	/* SndRefs file */
	s = STRDUP(fname.sndrefs);
	StripParentPath(s, dname.sounds);
	PromptSetS(&optwin->sndrefs_prompt, s);
	free(s);
#endif

	/* Graphics */

	/* Show Viewscreen Selected Weapon */
	CBtnSetActive(
	    &optwin->show_viewscreen_selected_weapon_cb,
	    opt->show_viewscreen_flags & XSW_SHOW_VIEWSCREEN_SELECTED_WEAPON
	);

	/* Show Viewscreen Markings */
	CBtnSetActive(
	    &optwin->show_viewscreen_markings_cb,
	    opt->show_viewscreen_flags & XSW_SHOW_VIEWSCREEN_MARKINGS
	);

	/* Show Viewscreen Object Labels */
	CBtnSetActive(
	    &optwin->show_viewscreen_object_labels_cb,
	    opt->show_viewscreen_flags & XSW_SHOW_VIEWSCREEN_OBJECT_LABELS
	);

	/* Show Viewscreen Netstats */
	CBtnSetActive(
	    &optwin->show_viewscreen_netstats_cb,
	    opt->show_viewscreen_flags & XSW_SHOW_VIEWSCREEN_NETSTATS
	);

	/* Formal object name labels */
	switch(opt->show_formal_label)
	{
	  case XSW_SHOW_FORMAL_LABEL_ALWAYS:
	    i = 2;
	    break;
	  case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
	    i = 1;
	    break;
	  default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
	    i = 0;
	    break;
	}
	RadioSetActive(&optwin->show_formal_label_radio, i);

	/* Show Lens Flares */
	CBtnSetActive(
	    &optwin->show_lens_flares_cb,
	    opt->show_lens_flares
	);

	/* Show Nebula Glow */
	CBtnSetActive(
	    &optwin->show_nebula_glow_cb,
	    opt->show_nebula_glow
	);

	/* Show Strobe Glow */
	CBtnSetActive(
	    &optwin->show_strobe_glow_cb,
	    opt->show_strobe_glow
	);

	/* Show Fires */
	CBtnSetActive(
	    &optwin->show_fires_cb,
	    opt->show_fires
	);

	/* Show Smoke Trails */
	CBtnSetActive(
	    &optwin->show_smoke_trails_cb,
	    opt->show_smoke_trails
	);

	/* Max Particles */
	PromptSetI(
	    &optwin->max_particles_prompt,
	    opt->max_particles
	);

	/* Async image loading */
	CBtnSetActive(
	    &optwin->async_image_loading_cb,
	    opt->async_image_loading
	);

	/* Async image loading pixels per cycle prompt */
	PromptSetI(
	    &optwin->async_image_pixels_prompt,
	    opt->async_image_pixels
	);


	/* Sounds */

	/* Sounds level */
	switch(opt->sounds)
	{
	  case XSW_SOUNDS_ALL:
	    i = 3;
	    break;
	  case XSW_SOUNDS_ENGINE:
	    i = 2;
	    break;
	  case XSW_SOUNDS_EVENTS:
	    i = 1;
	    break;
	  default:	/* XSW_SOUNDS_NONE */
	    i = 0;
	    break;
	}
	RadioSetActive(&optwin->sounds_radio, i);

	/* Music */
	CBtnSetActive(
	    &optwin->music_cb,
	    opt->music
	);

	/* Sound server type */
	switch(sound.server_type)
	{
	  case SOUND_SERVER_TYPE_MIKMOD:
	    i = 3;
	    break;
	  case SOUND_SERVER_TYPE_ESOUND:
	    i = 2;
	    break;
	  case SOUND_SERVER_TYPE_YIFF:
	    i = 1;
	    break;
	  default:	/* SOUND_SERVER_TYPE_NONE */
	    i = 0;
	    break;
	}
	RadioSetActive(&optwin->server_type_radio, i);

	/* Sound server program name */
	PromptSetS(
	    &optwin->sound_server_prompt,
	    sound.start_cmd
	);

	/* Sound server connection argument */
	PromptSetS(
	    &optwin->sound_con_arg_prompt,
	    sound.con_arg
	);


	/* Network */

	/* Default address */
	PromptSetS(
	    &optwin->def_address_prompt,
	    net_parms.address
	);

	/* Default port */
	PromptSetI(
	    &optwin->def_port_prompt,
	    net_parms.port
	);

	/* Network Bandwidth */
	PromptSetL(
	    &optwin->net_bandwidth_prompt,
	    opt->net_bandwidth
	);

	/* Show network errors */
	CBtnSetActive(
	    &optwin->show_net_errors_cb,
	    opt->show_net_errors
	);

	/* Show server errors */
	CBtnSetActive(
	    &optwin->show_server_errors_cb,
	    opt->show_server_errors
	);

	/* Network Interval */
	PromptSetI(
	    &optwin->net_int_prompt,
	    (int)net_parms.net_int
	);

	/* Automatic interval tunning */
	CBtnSetActive(
	    &optwin->auto_interval_cb,
	    auto_interval_tune.state
	);

	/* Network Timeout */
	PromptSetL(
	    &optwin->net_timeout_prompt,
	    opt->net_timeout
	);

	/* Object Outdated Time */
	PromptSetL(
	    &optwin->net_object_outdated_prompt,
	    opt->net_object_outdated_time
	);


	/* Misc */

	/* Local toplevel directory */
	s = STRDUP(dname.ltoplevel);
	StripParentPath(s, dname.home);
	PromptSetS(&optwin->xsw_local_toplevel_path_prompt, s);
	free(s);

	/* Toplevel directory */
	PromptSetS(
	    &optwin->xsw_toplevel_path_prompt,
	    dname.toplevel
	);

	/* Etc directory */
	s = STRDUP(dname.etc);
	StripParentPath(s, dname.toplevel);
	PromptSetS(&optwin->xsw_etc_path_prompt, s);
	free(s);

	/* Images directory */
	s = STRDUP(dname.images);
	StripParentPath(s, dname.toplevel);
	PromptSetS(&optwin->xsw_images_path_prompt, s);
	free(s);

	/* Sounds directory */
	s = STRDUP(dname.sounds);
	StripParentPath(s, dname.toplevel);
	PromptSetS(&optwin->xsw_sounds_path_prompt, s);
	free(s);

	/* Downloads directory */
	s = STRDUP(dname.downloads);
	StripParentPath(s, dname.home);
	PromptSetS(&optwin->xsw_downloads_path_prompt, s);

#ifdef JS_SUPPORT
	/* Joystick calibration file location */
	s = STRDUP(fname.js_calib);
	StripParentPath(s, dname.home);
	PromptSetS(&optwin->js_calib_path_prompt, s);
	free(s);
#endif /* JS_SUPPORT */


	/* Update widgets to reflect current data values */
	OptWinUpdateWidgets(optwin);
}

/*
 *	Applies changes made in options win to all applicateable
 *	global variables.  Sanitizing will be performed on all values.
 */
void OptWinApplyChanges(xsw_options_win_struct *optwin)
{
	char init_sound;
/*	bool_t need_load; */
	int i;
	char *s;
	xsw_option_struct *opt = &option;

	if((optwin == NULL) || (opt == NULL))
	    return;

	/* General */

	/* Show Universes List On Startup */
	opt->auto_map_univ_list_win = CBtnGetActive(
	    &optwin->show_univ_list_win_cb
	);

	/* Show Economy Window On Orbit */  
	opt->auto_map_eco_win = CBtnGetActive(
	    &optwin->show_eco_win_cb
	);

	/* CPU Priority */
	opt->cpu_priority = PUListGetSelItem(
	    &optwin->cpu_priority_pulist
	);

	/* Controller Type */
	switch(PUListGetSelItem(&optwin->controller_pulist))
	{
	  case 2:
	    opt->controller = XSW_CONTROLLER_POINTER;
	    break;
	  case 1:
	    opt->controller = XSW_CONTROLLER_JOYSTICK;
	    break;
	  default:
	    opt->controller = XSW_CONTROLLER_KEYBOARD;
	    break;
	}
if(opt->controller == XSW_CONTROLLER_POINTER)
    opt->controller = XSW_CONTROLLER_KEYBOARD;
	GCtlDelete(xsw_gctl);
	xsw_gctl = GCtlNew(opt->controller);
	if(xsw_gctl == NULL)
	{
	    XSWPrintWarning(
"An error occured while initializing the game controller."
	    );
	}

	/* Throttle mode */
	switch(RadioGetActive(&optwin->throttle_mode_radio))
	{
	  case 2:
	    opt->throttle_mode = XSW_THROTTLE_MODE_INCREMENTAL;
	    break;
	  case 1:
	    opt->throttle_mode = XSW_THROTTLE_MODE_BIDIRECTIONAL;
	    break;
	  default:
	    opt->throttle_mode = XSW_THROTTLE_MODE_NORMAL;
	    break;
	}

	/* Units */
	switch(PUListGetSelItem(&optwin->units_pulist))
	{
	  case 3:
	    opt->units = XSW_UNITS_XSW;
	    break;
	  case 2:
	    opt->units = XSW_UNITS_PARSECS;
	    break;
	  case 1:
	    opt->units = XSW_UNITS_ENGLISH;
	    break;
	  default:
	    opt->units = XSW_UNITS_METRIC;
	    break;
	}

	/* Auto zoom */
	opt->auto_zoom = CBtnGetActive(
	    &optwin->auto_zoom_cb
	);

	/* Local updates */
	opt->local_updates = CBtnGetActive(
	    &optwin->local_updates_cb
	);

	/* Scanner contacts */
	opt->notify_scanner_contacts = CBtnGetActive(
	    &optwin->notify_scanner_contacts_cb
	);

	/* NOTE: ocsn, imgref, and sndref paths are updated in the
	 * Misc section
	 */


	/* Graphics */

	/* Show Viewscreen Selected Weapon */
	if(CBtnGetActive(&optwin->show_viewscreen_selected_weapon_cb))
	    opt->show_viewscreen_flags |= XSW_SHOW_VIEWSCREEN_SELECTED_WEAPON;
	else
	    opt->show_viewscreen_flags &= ~XSW_SHOW_VIEWSCREEN_SELECTED_WEAPON;

	/* Show Viewscreen Markings */
	if(CBtnGetActive(&optwin->show_viewscreen_markings_cb))
	    opt->show_viewscreen_flags |= XSW_SHOW_VIEWSCREEN_MARKINGS;
	else
	    opt->show_viewscreen_flags &= ~XSW_SHOW_VIEWSCREEN_MARKINGS;

	/* Show Viewscreen Object Labels */
	if(CBtnGetActive(&optwin->show_viewscreen_object_labels_cb))
	    opt->show_viewscreen_flags |= XSW_SHOW_VIEWSCREEN_OBJECT_LABELS;
	else
	    opt->show_viewscreen_flags &= ~XSW_SHOW_VIEWSCREEN_OBJECT_LABELS;

	/* Show Viewscreen Netstats */
	if(CBtnGetActive(&optwin->show_viewscreen_netstats_cb))
	    opt->show_viewscreen_flags |= XSW_SHOW_VIEWSCREEN_NETSTATS;
	else
	    opt->show_viewscreen_flags &= ~XSW_SHOW_VIEWSCREEN_NETSTATS;

	/* Formal object name labels */
	switch(RadioGetActive(&optwin->show_formal_label_radio))
	{
	  case 2:
	    opt->show_formal_label = XSW_SHOW_FORMAL_LABEL_ALWAYS;
	    break;
	  case 1:
	    opt->show_formal_label = XSW_SHOW_FORMAL_LABEL_AS_NEEDED;
	    break;
	  default:
	    opt->show_formal_label = XSW_SHOW_FORMAL_LABEL_NEVER;
	    break;
	}

	/* Show Lens Flares */
	opt->show_lens_flares = CBtnGetActive(
	    &optwin->show_lens_flares_cb
	);

	/* Show Nebula Glow */
	opt->show_nebula_glow = CBtnGetActive(
	    &optwin->show_nebula_glow_cb
	);

	/* Show Strobe Glow */
	opt->show_strobe_glow = CBtnGetActive(
	    &optwin->show_strobe_glow_cb
	);

	/* Show Fires */
	opt->show_fires = CBtnGetActive(
	    &optwin->show_fires_cb
	);   

	/* Show Smoke Trails */
	opt->show_smoke_trails = CBtnGetActive(
	    &optwin->show_smoke_trails_cb
	);

	/* Max Particles */
	opt->max_particles = MAX(
	    PromptGetI(&optwin->max_particles_prompt),
	    0
	);

	/* Async image loading */
	opt->async_image_loading = CBtnGetActive(
	    &optwin->async_image_loading_cb
	);

	/* Async image loading pixels per cycle prompt */
	opt->async_image_pixels = MAX(
	    PromptGetI(&optwin->async_image_pixels_prompt),
	    1
	);


	/* Sounds */

	/* Sound server type */
	i = RadioGetActive(&optwin->server_type_radio);
	if(INRANGEINC(i, 0, 3))
	{
	    /* Must shutdown sound server if there is a change
	     * in the type of sound server
	     */
	    if(sound.server_type != i)
	    {
		SoundShutdown(&sound);

		/* Now set new sound server type */
		sound.server_type = i;
	    }
	}

	/* Sound Server start command */
	s = PromptGetS(&optwin->sound_server_prompt);
	if(s != NULL)
	{
	    free(sound.start_cmd);
	    sound.start_cmd = STRDUP(s);
	}

	/* Sound server connect argument */
	s = PromptGetS(&optwin->sound_con_arg_prompt);
	if(s != NULL)
	{
	    free(sound.con_arg);
	    sound.con_arg = STRDUP(s);
	}

	/* Sound amount (must be checked after sound server type) */
	init_sound = 0;
	i = RadioGetActive(&optwin->sounds_radio);
	if(INRANGEINC(i, XSW_SOUNDS_NONE, XSW_SOUNDS_ALL))
	{
	    /* Sound server type set to none? */
	    if(sound.server_type <= SOUND_SERVER_TYPE_NONE)
	    {
		/* Sound server type is set to none, so sounds level
		 * must also be set to none
		 */
		RadioSetActive(&optwin->sounds_radio, 0);

		if(opt->sounds > XSW_SOUNDS_NONE)
		{
		    init_sound = -1;	/* Need to shutdown sound */
		    opt->sounds = XSW_SOUNDS_NONE;
		}
		else
		{
		    init_sound = 0;
		}
	    }
	    else
	    {
		/* Sound server type is set to something */

		/* Was off, need to turn on? */
		if((opt->sounds <= XSW_SOUNDS_NONE) &&
		   (i > XSW_SOUNDS_NONE)
		)
		    init_sound = 1;	/* Need to initialize sound */
		/* Was on, need to turn off? */
		else if((opt->sounds > XSW_SOUNDS_NONE) &&
		        (i <= XSW_SOUNDS_NONE)
		)
		    init_sound = -1;	/* Need to shutdown sound */
		else
		    init_sound = 0;	/* No change */

		opt->sounds = i;
	    }
	}


	/* Music */
	opt->music = CBtnGetActive(&optwin->music_cb);
	/* Change background music */
	if(opt->music)
	    XSWUpdateBackgroundMusic();
	else
	    SoundStopBackgroundMusic(&sound);


	/* Network */

	/* Default address */
	s = PromptGetS(&optwin->def_address_prompt);
	if(s != NULL)
	{
	    free(net_parms.address);
	    net_parms.address = STRDUP(s);
	}

	/* Default port */
	net_parms.port = PromptGetI(&optwin->def_port_prompt);

	/* Show network errors */
	opt->show_net_errors = CBtnGetActive(
	    &optwin->show_net_errors_cb
	);

	/* Show server errors */
	opt->show_server_errors = CBtnGetActive(
	    &optwin->show_server_errors_cb
	);

	/* Maximum net load */
	opt->net_bandwidth = PromptGetL(
	    &optwin->net_bandwidth_prompt
	);
	if(opt->net_bandwidth == 0l)
	{
	    opt->net_bandwidth = XSW_DEF_NET_BANDWIDTH;
	    PromptSetL(
		&optwin->net_bandwidth_prompt,
		opt->net_bandwidth
	    );
	}

	/* Network Interval */
	net_parms.net_int = PromptGetL(&optwin->net_int_prompt);
	if(net_parms.net_int <= 0l)
	{
	    net_parms.net_int = XSW_SWSERV_DEF_UPDATE_INT;
	    PromptSetL(
		&optwin->net_int_prompt,
		net_parms.net_int
	    );
	}

	/* Auto interval */
	auto_interval_tune.state = CBtnGetActive(
	    &optwin->auto_interval_cb
	);

	/* Network Timeout */
	opt->net_timeout = PromptGetL(&optwin->net_timeout_prompt);
	if(opt->net_timeout < 0l)
	{
	    opt->net_timeout = 0l;
	    PromptSetL(
		&optwin->net_timeout_prompt,
		opt->net_timeout
	    );
	}

	/* Network Object Outdated Time */
	opt->net_object_outdated_time = PromptGetL(
	    &optwin->net_object_outdated_prompt
	);


	/* Misc */

	/* Local Toplevel */
	s = PromptGetS(&optwin->xsw_local_toplevel_path_prompt);
	if(s != NULL)
	{
	    if(ISPATHABSOLUTE(s))
		strncpy(dname.ltoplevel, s, sizeof(dname.ltoplevel));
	    else
		strncpy(
		    dname.ltoplevel, PrefixPaths(dname.home, s),
		    sizeof(dname.ltoplevel)
		);
	}
	dname.ltoplevel[sizeof(dname.ltoplevel) - 1] = '\0';

	/* Toplevel */
	s = PromptGetS(&optwin->xsw_toplevel_path_prompt);
	if(s != NULL)
	{
	    /* Apply only if full path */
	    if(ISPATHABSOLUTE(s))
		strncpy(dname.toplevel, s, sizeof(dname.toplevel));
	    else
		PromptSetS(
		    &optwin->xsw_toplevel_path_prompt,
		    dname.toplevel
		);
	}
	dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';

	/* Etc */
	s = PromptGetS(&optwin->xsw_etc_path_prompt);
	if(s != NULL)
	{
	    if(ISPATHABSOLUTE(s))
		strncpy(dname.etc, s, sizeof(dname.etc));
	    else
		strncpy(
		    dname.etc, PrefixPaths(dname.toplevel, s),
		    sizeof(dname.etc)
		);
	}
	dname.etc[sizeof(dname.etc) - 1] = '\0';

	/* Images */
	s = PromptGetS(&optwin->xsw_images_path_prompt);
	if(s != NULL)
	{
	    if(ISPATHABSOLUTE(s))
		strncpy(dname.images, s, sizeof(dname.images));
	    else
		strncpy(
		    dname.images, PrefixPaths(dname.toplevel, s),
		    sizeof(dname.images)
		);
	}
	dname.images[sizeof(dname.images) - 1] = '\0';

	/* Sounds */
	s = PromptGetS(&optwin->xsw_sounds_path_prompt);
	if(s != NULL)
	{
	    if(ISPATHABSOLUTE(s))
		strncpy(dname.sounds, s, sizeof(dname.sounds));
	    else
		strncpy(
		    dname.sounds, PrefixPaths(dname.toplevel, s),
		    sizeof(dname.sounds)
		);
	}
	dname.sounds[sizeof(dname.sounds) - 1] = '\0';

	/* Downloads */
	s = PromptGetS(&optwin->xsw_downloads_path_prompt);
	if(s != NULL)
	{
	    if(ISPATHABSOLUTE(s))
		strncpy(dname.downloads, s, sizeof(dname.downloads));
	    else
		strncpy(
		    dname.downloads, PrefixPaths(dname.home, s),
		    sizeof(dname.downloads)
		);
	}
	dname.downloads[sizeof(dname.downloads) - 1] = '\0';

#ifdef JS_SUPPORT
	/* JS Calib */
	s = PromptGetS(&optwin->js_calib_path_prompt);
	if(s != NULL)
	{
	    if(ISPATHABSOLUTE(s))
	    {
		strncpy(
		    fname.js_calib, s,
		    sizeof(fname.js_calib)
		);
	    }
	    else
	    {
		strncpy(
		    fname.js_calib, PrefixPaths(dname.home, s),
		    sizeof(fname.js_calib)
		);
	    }
	}
	fname.js_calib[sizeof(fname.js_calib) - 1] = '\0';
#endif /* JS_SUPPORT */

#if 0
	/* Imgrefs file */
	need_load = False;
	s = PromptGetS(&optwin->imgrefs_prompt);   
	if(s != NULL)
	{
	    strstrip(s);

	    if(ISPATHABSOLUTE(s))
	    {
		if(strcmp(fname.imgrefs, s))
		    need_load = True;

		strncpy(fname.imgrefs, s, sizeof(fname.imgrefs));
	    }
	    else
	    {
		const char *s2 = PrefixPaths(dname.images, s);
		if(s2 != NULL)
		{
		    if(strcmp(fname.imgrefs, s2))
			need_load = True;

		    strncpy(fname.imgrefs, s2, sizeof(fname.imgrefs));
		}
	    }
	    fname.imgrefs[sizeof(fname.imgrefs) - 1] = '\0';
	}
	/* Load as needed */
	if(need_load)
	    XSWOpenImgRefs(fname.imgrefs);

	/* OCSNs file */
	need_load = False;
	s = PromptGetS(&optwin->ocsn_prompt);
	if(s != NULL)
	{
	    strstrip(s);

	    if(ISPATHABSOLUTE(s))
	    {
		if(strcmp(fname.ocsn, s))
		    need_load = True;

		strncpy(fname.ocsn, s, sizeof(fname.ocsn));
	    }
	    else
	    {
		const char *s2 = PrefixPaths(dname.etc, s);
		if(s2 != NULL)
		{
		    if(strcmp(fname.ocsn, s2))
			need_load = True;

		    strncpy(fname.ocsn, s2, sizeof(fname.ocsn));
		}
	    }
	    fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';
	}
	/* Load as needed */
	if(need_load)
	    XSWOpenOCSNs(fname.ocsn);


	/* Sndrefs file */
	need_load = False;
	s = PromptGetS(&optwin->sndrefs_prompt);
	if(s != NULL)
	{
	    strstrip(s);

	    if(ISPATHABSOLUTE(s))
	    {
		if(strcmp(fname.sndrefs, s))
		    need_load = True;

		strncpy(
		    fname.sndrefs, s,
		    sizeof(fname.sndrefs)
		);
	    }
	    else        
	    {
		const char *s2 = PrefixPaths(dname.sounds, s);
		if(s2 != NULL)
		{
		    if(strcmp(fname.sndrefs, s2))
			need_load = True;

		    strncpy(
			fname.sndrefs, s2,
			sizeof(fname.sndrefs)
		    );
		}
	    }
	    fname.sndrefs[sizeof(fname.sndrefs) - 1] = '\0';
	}
	/* Load as needed */
	if(need_load)
	    XSWOpenSndRefs(fname.sndrefs);
#endif


	/* Initialize or shut down sound */
	switch(init_sound)
	{
	  case 1:
	    if(sound.con_data == NULL)
	    {
		xsw_sound_struct *snd = &sound;

		/* Initialize sound */
	        if(SoundInit(snd, snd->server_type))
		{
		    /* Unable to initialize sound */
		    XSWDialog(
			"Sound Initialization Failed",
			XSW_IMG_DIALOG_ICON_ERROR,
"Unable to establish a connection to the sound server\n\
or unable to start the sound server. Please make sure\n\
that the sound server is running and that the specified\n\
address is correct.",
			optwin->toplevel
		    );
	        }
		else
		{
		    /* Sound successfully initialized */

		    /* Open the sndrefs from file */
		    XSWOpenSndRefs(fname.sndrefs);
 		}
	    }
	    break;

	  case 0:
	    if(opt->sounds >= XSW_SOUNDS_ENGINE)
		XSWUpdateLoopSounds();
	    break;

	  case -1:
	    /* Delete all sndrefs */
	    SndRefDeleteAll(&xsw_sndref, &total_sndrefs);
	    /* Shutdown sound */
	    SoundShutdown(&sound);
	    break;
	}

	/* Reset has changes marker */
	optwin->has_changes = False;

	/* Update widgets to reflect current data values */
	OptWinUpdateWidgets(optwin);

	/* Refresh */
	XSWRefresh(cur_millitime, cur_systime, time_compensation);
}

/*
 *      Set default values.
 */
void OptWinSetDefaults(xsw_options_win_struct *optwin)
{
	if(optwin == NULL)
	    return;

	/* General */
	CBtnSetActive(&optwin->show_univ_list_win_cb, True);
	CBtnSetActive(&optwin->show_eco_win_cb, True);
	PUListSetSelItem(
	    &optwin->cpu_priority_pulist,
	    XSW_CPU_PRIORITY_MEDIUM
	);
	PUListSetSelItem(
	    &optwin->controller_pulist,
	    0			/* XSW_CONTROLLER_KEYBOARD */
	);
	RadioSetActive(
	    &optwin->throttle_mode_radio,
	    0			/* XSW_THROTTLE_MODE_NORMAL */
	);
	PUListSetSelItem(
	    &optwin->units_pulist,
	    0			/* XSW_UNITS_METRIC */
	);
	CBtnSetActive(&optwin->auto_zoom_cb, True);
	CBtnSetActive(&optwin->local_updates_cb, True);
	CBtnSetActive(&optwin->notify_scanner_contacts_cb, True);
#if 0
	PromptSetS(
	    &optwin->ocsn_prompt,
	    XSW_DEF_OCSN_FILE
	);
	PromptSetS(
	    &optwin->imgrefs_prompt,
	    XSW_DEF_IMGREFS_FILE
	);
	PromptSetS(
	    &optwin->sndrefs_prompt,
	    XSW_DEF_SNDREFS_FILE
	);
#endif

	/* Graphics */
	CBtnSetActive(&optwin->show_viewscreen_selected_weapon_cb, True);
	CBtnSetActive(&optwin->show_viewscreen_markings_cb, True);
	CBtnSetActive(&optwin->show_viewscreen_object_labels_cb, True);
	CBtnSetActive(&optwin->show_viewscreen_netstats_cb, False);
	RadioSetActive(
	    &optwin->show_formal_label_radio,
	    1			/* XSW_SHOW_FORMAL_LABEL_AS_NEEDED */
	);
	CBtnSetActive(&optwin->show_lens_flares_cb, True);
	CBtnSetActive(&optwin->show_nebula_glow_cb, True);
	CBtnSetActive(&optwin->show_strobe_glow_cb, True);
	CBtnSetActive(&optwin->show_fires_cb, True);
	CBtnSetActive(&optwin->show_smoke_trails_cb, True);
	PromptSetI(
	    &optwin->max_particles_prompt,
	    500
	);
	CBtnSetActive(&optwin->async_image_loading_cb, True);
	PromptSetI(
	    &optwin->async_image_pixels_prompt,
	    500000
	);

	/* Sounds */
	RadioSetActive(
	    &optwin->sounds_radio,
	    0			/* XSW_SOUNDS_NONE */
	);
	CBtnSetActive(&optwin->music_cb, False);
	RadioSetActive(
	    &optwin->server_type_radio,
	    0			/* SOUND_SERVER_TYPE_NONE */
	);
	PromptSetS(
	    &optwin->sound_server_prompt,
	    ""
	);
	PromptSetS(
	    &optwin->sound_con_arg_prompt,
	    ""
	);
	CBtnSetActive(&optwin->flip_stereo_cb, False);

	/* Network */
	PromptSetS(
	    &optwin->def_address_prompt,
	    XSW_DEF_SWSERV_ADDRESS
	);
	PromptSetI(
	    &optwin->def_port_prompt,
	    XSW_DEF_SWSERV_PORT
	);
	CBtnSetActive(&optwin->show_net_errors_cb, False);
	CBtnSetActive(&optwin->show_server_errors_cb, True);
	PromptSetI(
	    &optwin->net_bandwidth_prompt,
	    XSW_DEF_NET_BANDWIDTH
	);
	PromptSetI(
	    &optwin->net_int_prompt,
	    1000
	);
	PromptSetI(
	    &optwin->net_timeout_prompt,
	    XSW_DEF_NET_TIMEOUT
	);
	PromptSetI(
	    &optwin->net_object_outdated_prompt,
	    XSW_DEF_OBJECT_OUTDATED_TIME
	);
	CBtnSetActive(&optwin->auto_interval_cb, True);

	/* Misc */
	PromptSetS(
	    &optwin->xsw_local_toplevel_path_prompt,
	    PrefixPaths(dname.home, XSW_DEF_LOCAL_DATA_DIR)
	);
	PromptSetS(
	    &optwin->xsw_toplevel_path_prompt,
	    XSW_DEF_GLOBAL_DATA_DIR
	);
	PromptSetS(
	    &optwin->xsw_etc_path_prompt,
	    XSW_DEF_GLOBAL_ETC_DIR
	);
	PromptSetS(
	    &optwin->xsw_images_path_prompt,
	    XSW_DEF_GLOBAL_IMAGES_DIR
	);
	PromptSetS( 
	    &optwin->xsw_sounds_path_prompt,
	    XSW_DEF_GLOBAL_SOUNDS_DIR
	);
	PromptSetS( 
	    &optwin->xsw_downloads_path_prompt,
	    dname.home
	);
#ifdef JS_SUPPORT
	PromptSetS( 
	    &optwin->js_calib_path_prompt,
#if defined(HAVE_LIBJSW) && defined(JSDefaultCalibration)
	    PrefixPaths(dname.home, JSDefaultCalibration)
#else
	    PrefixPaths(dname.home, ".joystick")
#endif
	);
#endif	/* JS_SUPPORT */


	/* Mark has modifications */
	optwin->has_changes = True;

	/* Update widgets to reflect current data values */
	OptWinUpdateWidgets(optwin);
}

/*
 *      Applies all widget values to global and then saves changes.
 */
int OptWinSaveChanges(xsw_options_win_struct *optwin)
{
	if(optwin == NULL)
	    return(-1);

	/* Apply changes first */
	OptWinApplyChanges(optwin);

	/* Save changes */
	if(CfgFileSave(
	    fname.cfg,
	    &option,
	    xsw_vsmenu
	))
	{
	    char *s = (char *)malloc(STRLEN(fname.cfg) + 256);
	    sprintf(
		s,
"An error occured while saving the options to the file:\n\
\n\
    %s",
		fname.cfg
	    );
	    XSWDialog(
		"Save Error",
		XSW_IMG_DIALOG_ICON_ERROR,
		s,
		optwin->toplevel
	    );
	    free(s);
	}
	else
	{
	    XSWDialog(
		"Options Saved",
		XSW_IMG_DIALOG_ICON_INFO, 
		"Options successfully saved.",
		optwin->toplevel
	    );
	}

	/* Reset has changes marker */
	optwin->has_changes = False;

	/* Update widgets to reflect current data values */
	OptWinUpdateWidgets(optwin);

	return(0);
}


/*
 *	Test sound button callback.
 */
void OptWinTestSoundPBCB(push_button_struct *btn, void *data)
{
	xsw_sound_struct *snd = &sound;
	xsw_option_struct *opt = &option;
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if((optwin == NULL) || (opt == NULL))
	    return;

	/* Check if the sound server is initialized */
	if((snd->con_data == NULL) ||
	   (opt->sounds <= XSW_SOUNDS_NONE) ||
	   (snd->server_type <= SOUND_SERVER_TYPE_NONE)
	)
	{
	    XSWDialog(
		"Sound Not Initialized",
		XSW_IMG_DIALOG_ICON_WARNING,
"Sound has not been initialized.\n\
\n\
To initialize sound:\n\
\n\
1. Select the Sound Server Type.\n\
2. Set the Sounds to All.\n\
3. Click on Apply.",
		optwin->toplevel
	    );
	    return;
	}

	/* Begin sound test */
	if(snd != NULL)
	{
	    const char *server_type_str, *sounds_level_str;
	    char *s;
	    void *playid;

	    /* Play the default XSW SndRef */
	    playid = SoundPlay(
		&sound,
		SW_SNDREF_XSW_LOGO,		/* SndRef */
		1.0,				/* Left volume */
		1.0,				/* Right volume */
		0,				/* Effects */
		0				/* Priority */
	    );
	    if(playid == NULL)
	    {
		playid = SoundPlay(
		    &sound,
		    SW_SNDREF_DEFAULT,		/* SndRef */
		    1.0,			/* Left volume */
		    1.0,			/* Right volume */
		    0,				/* Effects */
		    0				/* Priority */
		);
	    }

	    /* Check the sound play result */
	    if(playid != NULL)
	    {
		switch(snd->server_type)
		{
		  case SOUND_SERVER_TYPE_YIFF:
		    server_type_str = "YIFF";
		    break;
		  case SOUND_SERVER_TYPE_ESOUND:
		    server_type_str = "EsounD";
		    break;
		  case SOUND_SERVER_TYPE_MIKMOD:
		    server_type_str = "MikMod";
		    break;
		  default:
		    server_type_str = "None";
		    break;
		}
		switch(opt->sounds)
		{
		  case XSW_SOUNDS_EVENTS:
		    sounds_level_str = "Events";
		    break;
		  case XSW_SOUNDS_ENGINE:
		    sounds_level_str = "Engine";
		    break;
		  case XSW_SOUNDS_ALL:
		    sounds_level_str = "All";
		    break;
		  default:
		    sounds_level_str = "None";
		    break;
		}
		s = (char *)malloc(
		    STRLEN(server_type_str) +
		    STRLEN(snd->con_arg) +
		    STRLEN(snd->audio_mode_name) +
		    STRLEN(sounds_level_str) +
		    (8 * 80)
		);
		sprintf(
		    s,
"Sound test successful.\n\
\n\
Sound Server Type: %s\n\
Connected To: %s\n\
Audio Mode: %s\n\
\n\
Sounds Level: %s\n\
Music: %s",
		    server_type_str,
		    (snd->con_arg != NULL) ?
			snd->con_arg : "n/a",
		    (snd->audio_mode_name != NULL) ?
			snd->audio_mode_name : "n/a",
		    sounds_level_str,
		    opt->music ? "On" : "Off"
		);
		XSWDialog(
		    "Sound Test Successful",
		    XSW_IMG_DIALOG_ICON_INFO,
		    s,
		    optwin->toplevel
		);
		free(s);
	    }
	    else
	    {
		XSWDialog(
		    "Sound Test Failed",
		    XSW_IMG_DIALOG_ICON_WARNING,
		    "Error playing sound.",
		    optwin->toplevel
		);
	    }
	}
}


/*
 *	Browser OCSN file callback.
 */
void OptWinBrowseOCSNsPBCB(push_button_struct *btn, void *data)
{
#if 0
	char *path, **paths_list;
	int npaths;
	prompt_struct *prompt;
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	prompt = &optwin->ocsn_prompt;
	path = STRDUP(PrefixPaths(dname.etc, XSW_OCSN_EXT_MASK));

	if(XSWSelectFile(
	    "Select OCS Names File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    optwin->toplevel
	))
	{ 
	    if(npaths > 0)
	    {
		PromptSetS(prompt, paths_list[0]);
		PromptMarkBuffer(prompt, PROMPT_POS_END);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
#endif
}

/*
 *	Browse ImgRefs file callback.
 */
void OptWinBrowseImgRefsPBCB(push_button_struct *btn, void *data)
{
#if 0
	char *path, **paths_list;
	int npaths;
	prompt_struct *prompt;
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	prompt = &optwin->imgrefs_prompt;
	path = STRDUP(PrefixPaths(dname.images, XSW_IMGREFS_EXT_MASK));

	if(XSWSelectFile(
	    "Select ImgRefs File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    optwin->toplevel
	))
	{
	    if(npaths > 0)
	    {
		PromptSetS(prompt, paths_list[0]);
		PromptMarkBuffer(prompt, PROMPT_POS_END);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
#endif
}

/*
 *	Browser SndRefs file callback.
 */
void OptWinBrowseSndRefsPBCB(push_button_struct *btn, void *data)
{
#if 0
	char *path, **paths_list;
	int npaths;
	prompt_struct *prompt;
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	prompt = &optwin->sndrefs_prompt;
	path = STRDUP(PrefixPaths(dname.sounds, XSW_SNDREFS_EXT_MASK));

	if(XSWSelectFile(
	    "Select SndRefs File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    optwin->toplevel
	))
	{
	    if(npaths > 0)
	    {
		PromptSetS(prompt, paths_list[0]);
		PromptMarkBuffer(prompt, PROMPT_POS_END);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
#endif
}

#ifdef JS_SUPPORT
/*
 *	Browser joystick calibration file callback.
 */
void OptWinBrowseJSCalPBCB(push_button_struct *btn, void *data)
{
	char *path, **paths_list;
	int npaths;
	prompt_struct *prompt;
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	prompt = &optwin->js_calib_path_prompt;
	path = STRDUP(GetParentDir(PromptGetS(prompt)));
	if(!ISPATHABSOLUTE(path))
	{
	    free(path);
	    path = STRDUP(dname.home);
	}

	if(XSWSelectFile(
	    "Select Joystick Calibration File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    optwin->toplevel
	))
	{
	    if(npaths > 0)
	    {
		PromptSetS(prompt, paths_list[0]);
		PromptMarkBuffer(prompt, PROMPT_POS_END);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
}
#endif	/* JS_SUPPORT */


/*
 *	Map Keyboard callback.
 */
void OptWinKeymapPBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	XSWAction(
	    XSW_ACTION_KEYMAP, NULL,
	    optwin,
	    cur_millitime, cur_systime, time_compensation
	);
}

#ifdef JS_SUPPORT
/*
 *	Map Joystick callback.
 */
void OptWinJSMapPBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

        XSWAction(
	    XSW_ACTION_JSMAP, NULL,
	    optwin,
	    cur_millitime, cur_systime, time_compensation
	);
}
#endif	/* JS_SUPPORT */

/*
 *	Configure VSMenu callback.
 */
void OptWinConfigureVSMenuPBCB(push_button_struct *btn, void *data)
{
        xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
        if(optwin == NULL)
            return;

        XSWAction(
	    XSW_ACTION_VSMENU_WIN, NULL,
	    optwin,
	    cur_millitime, cur_systime, time_compensation
	);
}

/*
 *	Reclaim memory button callback.
 */
void OptWinRefreshMemoryPBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)                              
	    return;       

	/* Clean, reclaim, memory and schedual next */
	XSWReclaimMem(False, cur_millitime, cur_systime, time_compensation);

	/* Update widgets to reflect current data values */
	OptWinUpdateWidgets(optwin);
}


/*
 *	OK callback.
 */
void OptWinOKPBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	OptWinApplyChanges(optwin);
	OptWinUnmap(optwin);
}

/*
 *	Apply callback.
 */
void OptWinApplyPBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	OptWinApplyChanges(optwin);
}

/*
 *	Save callback.
 */
void OptWinSavePBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	OptWinSaveChanges(optwin);
}

/*
 *	Defaults callback.
 */
void OptWinDefaultsPBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	OptWinSetDefaults(optwin);
}

/*
 *	Cancel callback.
 */
void OptWinCancelPBCB(push_button_struct *btn, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(data);
	if(optwin == NULL)
	    return;

	OptWinUnmap(optwin);
}
