#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "../include/unvmath.h"

#include "net.h"
#include "xsw.h"

int ScannerButtonMatch(
	xsw_bridge_win_struct *bw,
	int x, int y,			/* In window coordinates */
	int start_obj_num
);

void VSChangePage(
	xsw_bridge_win_struct *bw, page_struct *new_page
);
int VSButtonMatch(xsw_bridge_win_struct *bw, event_t *event);
int VSManage(xsw_bridge_win_struct *bw, event_t *event);
void VSMap(xsw_bridge_win_struct *bw);
void VSUnmap(xsw_bridge_win_struct *bw);
void VSDestroy(xsw_bridge_win_struct *bw);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Matches an object by a button press coordinate on the scanner.
 *
 *      Returns matched object number or -1 for no match.
 */
int ScannerButtonMatch(
	xsw_bridge_win_struct *bw,
	int x, int y,			/* In window coordinates */
	int start_obj_num
)
{
	int i, obj_num;
	xsw_object_struct *obj, *tar_obj;
	float	width, height, shortest_len,
		scanner_scale,		/* Scanner zoom */
		obj_size_ru,
		rel_x, rel_y,
		dx, dy,
		rx, ry,
		sin_theta, cos_theta,
		scanner_range,		/* Effective scaner range in ru */
		scanner_factor;		/* RU to scanner units factor */
float win_x = 0.0f;
float win_y = 0.0f;

	if(bw == NULL)
	    return(-1);

	/* Get the object that our scanner matching will be relative
	 * to
	 */
	obj_num = net_parms.player_obj_num;
	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	width = bw->scanner_width;
	height = bw->scanner_height;
	shortest_len = MIN(width, height);
	if(shortest_len <= 0.0)
	    return(-1);

	scanner_scale = CLIP(bw->scanner_zoom, 0.001, 1.0);

	/* Get the current effective scanner range in real units */
	scanner_range = DBGetObjectEffectiveScannerRangePtr(obj); 
	if(scanner_range <= 0.0f)
	    return(-1);

	/* Calculate the scanner factor */
	scanner_factor = MAX(
	    (((shortest_len / 2) / scanner_range) / scanner_scale),
	    0.0
	);

	/* Get the object's relative position in real units */
	rel_x = obj->x;
	rel_y = obj->y;

	/* Get sin_theta and cos_theta */
	sin_theta = (float)sin((double)obj->heading);
	cos_theta = (float)cos((double)obj->heading);


	/* Iterate through the objects list (skip stream weapon and
	 * area objects)
	 */
	for(i = MAX(start_obj_num, 0); i < total_objects; i++)
	{
	    tar_obj = xsw_object[i];
	    if((tar_obj == NULL) || (tar_obj == obj))
		continue;

	    /* Skip the following object types */
	    if((tar_obj->type <= SW_OBJ_TYPE_GARBAGE) ||
	       (tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	       (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON) ||
	       (tar_obj->type == SW_OBJ_TYPE_AREA)
	    )
		continue;

	    /* Skip outdated objects */
	    if((tar_obj->last_updated + option.net_object_outdated_time) <
	       cur_millitime
	    )
		continue;

	    /* Object not within effective scanner range? */
	    if(!DBIsObjectInEffectiveScannerRangePtr(
		obj, tar_obj, scanner_range
	    ))
		continue;

	    /* This object is within scanner range, now calculate the
	     * position of the object in window coordinates
	     */
	    dx = (tar_obj->x - rel_x) * scanner_factor;
	    dy = (tar_obj->y - rel_y) * scanner_factor;

	    switch(bw->scanner_orient)
	    {
	      case XSW_SCANNER_ORIENT_PLAYER:
		rx = (cos_theta * dx) + (sin_theta * -1 * dy);  
		ry = ((sin_theta * dx) + (cos_theta * dy)) * -1;
		break;
	      default:		/* XSW_SCANNER_ORIENT_GC */
		rx = dx;
		ry = dy * -1;
		break;
	    }

	    dx = rx + (width / 2);
	    dx += win_x;

	    dy = ry + (height / 2);
	    dy += win_y;

	    /* dx and dy are now center of target object relative to
	     * player object which is center of plane
	     */
	    if(True)
	    {
		int d = (int)SWMath3DDistance(dx - x, dy - y, 0.0);
		if(d <= 4)
		    return(i);
	    }
	}

	/* Iterate through only the area objects */
	for(i = MAX(start_obj_num, 0); i < total_objects; i++)
	{
	    tar_obj = xsw_object[i];
	    if((tar_obj == NULL) || (tar_obj == obj))
		continue;

	    /* Skip non-area objects */
	    if(tar_obj->type != SW_OBJ_TYPE_AREA)
		continue;

	    /* Skip outdated objects */
	    if((tar_obj->last_updated + option.net_object_outdated_time) <
	       cur_millitime
	    )
		continue;

	    /* Object not within effective scanner range? */
	    if(!DBIsObjectInEffectiveScannerRangePtr(
		obj, tar_obj, scanner_range
	    ))
		continue;

	    /* This object is within scanner range, now calculate the
	     * position of the object in window coordinates
	     */
	    dx = (tar_obj->x - rel_x) * scanner_factor;
	    dy = (tar_obj->y - rel_y) * scanner_factor;
	    obj_size_ru = MAX(((float)tar_obj->size / 1000.0f), 0.0f);

	    switch(bw->scanner_orient)
	    {
	      case XSW_SCANNER_ORIENT_PLAYER:
		rx = (cos_theta * dx) + (sin_theta * -1 * dy);
		ry = ((sin_theta * dx) + (cos_theta * dy)) * -1;
		break;
	      default:          /* XSW_SCANNER_ORIENT_GC */
		rx = dx;
		ry = dy * -1;
		break;
	    }

	    dx = rx + (width / 2);
	    dx += win_x;

	    dy = ry + (height / 2);
	    dy += win_y;

	    /* dx and dy are now center of target object relative to
	     * player object which is center of plane
	     */
	    if(True)
	    {
		int d = (int)SWMath3DDistance(dx - x, dy - y, 0.0);
		if(d <= (int)(obj_size_ru * scanner_factor))
		    return(i);
	    }
	}


	return(-1);
}


/*
 *	Changes the page being displayed on the viewscreen to the page
 *	specified by new_page.
 */
void VSChangePage(
	xsw_bridge_win_struct *bw, page_struct *new_page
)
{
	bool_t mapped = False;
	win_t w;
	shared_image_t *image;

	if((bw == NULL) || (new_page == NULL))
	    return;

	w = bw->viewscreen;
	image = bw->viewscreen_image;

	/* Record if page was currently mapped */
	if(bw->cur_page != NULL)
	{
	    if(bw->cur_page->map_state)
		mapped = True;
	}

	/* Unmap as needed */
	if(mapped)
	    PageUnmap(bw->cur_page, w, image);

	/* Set pointer to new current page */
	bw->cur_page = new_page;

	/* Map new page if previous was mapped */
	if(mapped)
	    PageMap(bw->cur_page, w, image);
}

/*
 *      Handles the viewscreen button event.
 *
 *	The button event is assumed to have occured on the viewscreen.
 *
 *      Returns the object that is matched or -1 on error.
 */
int VSButtonMatch(xsw_bridge_win_struct *bw, event_t *event)
{
	float x_rel, y_rel;
	int x_vtrans, y_vtrans;
	unsigned int width, height;
	float distance;
	int obj_num;
	xsw_object_struct *obj, *camera_obj;
	event_button_t *button;

	if((bw == NULL) || (event == NULL))
	    return(-1);

	button = &event->xbutton;
	width = bw->viewscreen_width;
	height = bw->viewscreen_height;

	/* Skip if the player object is garbage */
	camera_obj = net_parms.player_obj_ptr;
	if(camera_obj == NULL)
	    return(-1);

	/* Calculate x_vtrans and y_vtrans (in units of pixels) */
	if(camera_obj->velocity_max > 0.0)
	{
	    x_vtrans = static_cast<int>(SWMathRectangularToPolarX(
		camera_obj->velocity_heading,
		camera_obj->velocity /
		camera_obj->velocity_max *
		(float)((int)width / 2.5)
	    ));
	    y_vtrans = static_cast<int>(SWMathRectangularToPolarY(
		camera_obj->velocity_heading,
		camera_obj->velocity /
		camera_obj->velocity_max *
		(float)((int)height / 2.5)
	    ));
	}
	else
	{
	    x_vtrans = 0;
	    y_vtrans = 0;
	}

	/* Calculate x_rel and y_rel */
	x_rel = ((float)(button->x - ((int)width / 2) +
	    x_vtrans) / 1000 / bw->viewscreen_zoom) +
	    camera_obj->x;

	y_rel = ((float)((int)height - button->y - 
	    ((int)height / 2) + y_vtrans) / 1000 /
	    bw->viewscreen_zoom) + camera_obj->y;

	/* Iterate through all objects */
	for(obj_num = 0; obj_num < total_objects; obj_num++)
	{
	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		continue;

	    /* Skip following object types */
	    if((obj->type <= SW_OBJ_TYPE_GARBAGE) ||
	       (obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	       (obj->type == SW_OBJ_TYPE_SPHEREWEAPON) ||
	       (obj->type == SW_OBJ_TYPE_AREA)
	    )
		continue;

	    /* Skip itself */
	    if(obj == camera_obj)
		continue;
   
	    /* Skip if object is not in the same sector */
	    if(!SWMath3DInSameSectorPtr(camera_obj, obj))
		continue;

	    /* Skip outdated objects */
	    if((obj->last_updated + option.net_object_outdated_time) <
	       cur_millitime
	    )
		continue;
	
	    distance = SWMath3DDistance(
		obj->x - x_rel,
		obj->y - y_rel,
		0
	    ) * 1000.0;

	    if(distance < (float)obj->size)
		break;
	}
	if(obj_num < total_objects)
	    return(obj_num);
	else
	    return(-1);
}

/*
 *	Manages the viewscreen event.
 */
int VSManage(xsw_bridge_win_struct *bw, event_t *event)
{
	int events_handled = 0;

	if((bw == NULL) || (event == NULL))
	    return(events_handled);



	return(events_handled);
}

/*
 *	Maps the viewscreen.
 */
void VSMap(xsw_bridge_win_struct *bw)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(bw == NULL)
	    return;

	/* Already mapped? */
	if(bw->viewscreen_map_state)
	    return;

	/* Change cursor */
	if(xsw_cursor.scanner_lock != NULL)
	    OSWSetWindowCursor(
	        bw->viewscreen,
		xsw_cursor.scanner_lock->cursor
	    );

	/* Change selected watched events */
	OSWSetWindowInput(
	    bw->viewscreen,
	    ButtonPressMask | VisibilityChangeMask
	);


	/* Mark as mapped */
	bw->viewscreen_map_state = True;

#if defined(HAVE_X) && defined(USE_XSHM)
	/* Reset viewscreen image "put" completion mark to False */
	if(bw->viewscreen_image != NULL)
	{
	    shared_image_t *simg = bw->viewscreen_image;
	    if(gui->def_use_shm)
	    {
		simg->in_progress = False;
	    }
	}
#endif	/* defined(HAVE_X) && defined(USE_XSHM) */
}

/*
 *	Unmaps the viewscreen.
 */
void VSUnmap(xsw_bridge_win_struct *bw)
{
	if(bw == NULL)
	    return;

	/* Already unmapped? */
	if(!bw->viewscreen_map_state)
	    return;

	/* Mark as unmapped */
	bw->viewscreen_map_state = False;
}

/*
 *	Destroys the viewscreen.
 */
void VSDestroy(xsw_bridge_win_struct *bw)
{

}
