#include <stdlib.h>
#include <string.h>

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "vsmenu.h"
#include "xsw.h"
#include "xswaction.h"


xsw_vsmenu_item_struct *VSMenuItemNew(void);
xsw_vsmenu_item_struct *VSMenuItemDup(xsw_vsmenu_item_struct *item);
void VSMenuItemDelete(xsw_vsmenu_item_struct *item);

void VSMenuItemsListDelete(
	xsw_vsmenu_item_struct **items_list, const int nitems
);

static xsw_vsmenu_item_struct **VSMenuSetDefaultItems(
	int *nitems_rtn
);

static void VSMenuActivateCB(
	menu_struct *menu, int id, void *data
);

xsw_vsmenu_struct *VSMenuNew(
	xsw_vsmenu_item_struct **items_list, const int nitems
);
void VSMenuDraw(xsw_vsmenu_struct *m);
int VSMenuManage(xsw_vsmenu_struct *m, event_t *event);
void VSMenuMap(xsw_vsmenu_struct *m);
void VSMenuUnmap(xsw_vsmenu_struct *m);
void VSMenuDelete(xsw_vsmenu_struct *m);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


xsw_vsmenu_item_struct *VSMenuItemNew(void)
{
	return(XSW_VSMENU_ITEM(calloc(
	    1, sizeof(xsw_vsmenu_item_struct)
	)));
}

xsw_vsmenu_item_struct *VSMenuItemDup(xsw_vsmenu_item_struct *item)
{
	xsw_vsmenu_item_struct *tar_item;

	if(item == NULL)
	    return(NULL);

	tar_item = VSMenuItemNew();
	if(tar_item == NULL)
	    return(NULL);

	tar_item->type = item->type;
	tar_item->action = item->action;
	tar_item->text = STRDUP(item->text);
	tar_item->icon_path = STRDUP(item->icon_path);

	return(tar_item);
}

void VSMenuItemDelete(xsw_vsmenu_item_struct *item)
{
        if(item == NULL)
            return;

        free(item->text);
	free(item->icon_path);
	free(item);
}

/*
 *	Deletes the list of Viewscreen Menu Items.
 */
void VSMenuItemsListDelete(
	xsw_vsmenu_item_struct **items_list, const int nitems
)
{
	int i;

	for(i = 0; i < nitems; i++)
	    VSMenuItemDelete(items_list[i]);

	free(items_list);
}


/*
 *	Returns a list of detault menu items.
 */
static xsw_vsmenu_item_struct **VSMenuSetDefaultItems(
	int *nitems_rtn
)
{
	xsw_vsmenu_item_struct **items_list = NULL;
	int nitems = 0;

#define APPEND_ITEM(_type_,_text_,_action_)	{	\
 xsw_vsmenu_item_struct *item;				\
 const int i = nitems;					\
 nitems = i + 1;					\
 items_list = (xsw_vsmenu_item_struct **)realloc(	\
  items_list,						\
  nitems * sizeof(xsw_vsmenu_item_struct *)		\
 );							\
 items_list[i] = item = VSMenuItemNew();		\
 item->type = (_type_);					\
 item->text = STRDUP(_text_);				\
 item->action = (_action_);				\
}

	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    PROG_NAME_FULL " Commands",
	    XSW_ACTION_NONE
	);
	APPEND_ITEM(
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    XSW_ACTION_NONE
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Connect...",
	    XSW_ACTION_CONNECT
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Connect Last",
	    XSW_ACTION_CONNECT_LAST
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Refresh",
	    XSW_ACTION_REFRESH
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Toggle Auto Interval",
	    XSW_ACTION_AUTO_INTERVAL
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Disconnect",
	    XSW_ACTION_DISCONNECT
	);
	APPEND_ITEM(
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    XSW_ACTION_NONE
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Economy...",
	    XSW_ACTION_ECONOMY
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Star Chart...",
	    XSW_ACTION_STARCHART
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Script...",
	    XSW_ACTION_SERVER_SCRIPT
	);
	APPEND_ITEM(
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    XSW_ACTION_NONE
	);
	APPEND_ITEM(
	    MENU_ITEM_LABEL,
	    "Options...",
	    XSW_ACTION_OPTIONS
	);

#undef APPEND_ITEM

	if(nitems_rtn != NULL)
	    *nitems_rtn = nitems;

	return(items_list);
}


/*
 *	Menu activate callback.
 */
static void VSMenuActivateCB(
	menu_struct *menu, int id, void *data
)
{
	const int action = id;
	xsw_vsmenu_struct *m = XSW_VSMENU(data);
	if(m == NULL)
	    return;

	XSWAction(
	    action, NULL, xsw_bridge_win,
	    cur_millitime, cur_systime, time_compensation
	);
}


/*
 *	Creates a new Viewscreen Menu.
 *
 *	The items_list and nitems specify the list of menu items, if
 *	they are NULL and 0 then a default set of menu items will be
 *	used instead.
 */
xsw_vsmenu_struct *VSMenuNew(
	xsw_vsmenu_item_struct **items_list, const int nitems
)
{
	win_t parent = osw_gui[0].root_win;
	menu_struct *menu;
	xsw_vsmenu_struct *m = XSW_VSMENU(calloc(1, sizeof(xsw_vsmenu_struct)));
	if(m == NULL)
	    return(NULL);

	m->item = NULL;
	m->total_items = 0;

	/* Initialize the menu */
	menu = &m->menu;
	MenuInit(
	    menu, parent,
	    VSMenuActivateCB, m
	);

	/* Menu items specified? */
	if(nitems > 0)
	{
	    /* Copy the menu items list */
	    int i;
	    m->total_items = nitems;
	    m->item = (xsw_vsmenu_item_struct **)malloc(
		m->total_items * sizeof(xsw_vsmenu_item_struct *)
	    );
	    for(i = 0; i < m->total_items; i++)
		m->item[i] = VSMenuItemDup(items_list[i]);
	}
	else
	{
	    /* No menu items specified, create default menu items */
	    m->item = VSMenuSetDefaultItems(&m->total_items);
	}

	/* Create the menu items on the menu */
	if(m->total_items > 0)
	{
	    const int n = m->total_items;
	    int i;
	    xsw_vsmenu_item_struct *item;

	    for(i = 0; i < n; i++)
	    {
		item = m->item[i];
		if(item == NULL)
		    continue;

/* Load the icon file */

		MenuItemAppend(
		    menu,
		    item->type,
		    item->text,
		    None, None,
		    item->action
		);
	    }
	}

	return(m);
}

/*
 *	Draws the Viewscreen Menu.
 */
void VSMenuDraw(xsw_vsmenu_struct *m)
{
	if(m == NULL)
	    return;

	WidgetQueueDraw(&m->menu);
}

/*
 *	Manages the Viewscreen Menu.
 */
int VSMenuManage(xsw_vsmenu_struct *m, event_t *event)
{
	int events_handled = 0;

	if((m == NULL) || (event == NULL))
	    return(0);

	if(events_handled == 0)
	    events_handled += MenuManage(&m->menu, event);

	return(events_handled);
}

/*
 *	Maps the Viewscreen Menu.
 */
void VSMenuMap(xsw_vsmenu_struct *m)
{
	int x, y;
	osw_gui_struct *gui = &osw_gui[0];

	if(m == NULL)
	    return;

	/* Get the current pointer position */
	OSWGetPointerCoords(
	    gui->root_win,
	    NULL, NULL,
	    &x, &y
	);

	/* Map the menu */
	MenuMapCoordinates(&m->menu, x, y);
}

/*
 *	Unmaps the Viewscreen Menu.
 */
void VSMenuUnmap(xsw_vsmenu_struct *m)
{
	if(m == NULL)
	    return;

	MenuUnmap(&m->menu);
}

/*
 *	Deletes the Viewscreen Menu.
 */
void VSMenuDelete(xsw_vsmenu_struct *m)
{
	if(m == NULL)
	    return;

	VSMenuItemsListDelete(m->item, m->total_items);
	WidgetDestroy(&m->menu);
	free(m);
}
