#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/cfgfmt.h"
#include "../include/osw-x.h"
#include "../include/widget.h"

#include "vsmenu.h"
#include "xsw.h"


int VSMenuFileOpen(
	const char *path,
	xsw_vsmenu_item_struct ***list, int *total_rtn
);
int VSMenuFileSave(
	const char *path,
	xsw_vsmenu_item_struct **list, const int total
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Opens the Viewscreen Menu configuration from the specified
 *	file.
 */
int VSMenuFileOpen(
	const char *path,
	xsw_vsmenu_item_struct ***list, int *total_rtn
)
{
	int lines_read = 0;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	unsigned long file_size;
	struct stat stat_buf;

	if(STRISEMPTY(path) || (list == NULL) ||
	   (total_rtn == NULL)
	)
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for reading");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    XSWPrintWarning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;
	}

#define WARN_NO_SUCH_PARM(_parm_)		{	\
 char *s = (char *)malloc(				\
  STRLEN(path) + 40 + STRLEN(_parm_) + 80		\
 );							\
 sprintf(						\
  s,							\
"%s: Line %i: Unsupported parameter `%s'.",		\
  path, lines_read, (_parm_)				\
 );							\
 XSWPrintWarning(s);					\
 free(s);						\
}

#define WARN_NO_SUCH_FILE(_v_,_error_code_)	{	\
 const int error_code = (_error_code_);			\
 char *s, *error_msg = STRDUP(strerror(error_code));	\
 if(error_msg == NULL)					\
  error_msg = strdup("no such file");			\
 *error_msg = toupper(*error_msg);			\
							\
 s = (char *)malloc(					\
  STRLEN(path) + 40 + STRLEN(parm) + STRLEN(_v_) +	\
  STRLEN(error_msg) + 80				\
 );							\
 sprintf(						\
  s,							\
"%s: Line %i: %s: Warning: %s: %s.",			\
  path, lines_read, parm, (_v_), error_msg		\
 );							\
 XSWPrintWarning(s);					\
 free(s);						\
							\
 free(error_msg);					\
}

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';

	    /* BeginVSMenuItem */
	    if(!strcasecmp(parm, "BeginVSMenuItem"))
	    {
		int i;
		xsw_vsmenu_item_struct *item;

		i = MAX(*total_rtn, 0);
		*total_rtn = i + 1;

		*list = (xsw_vsmenu_item_struct **)realloc(
		    *list,
		    (*total_rtn) * sizeof(xsw_vsmenu_item_struct *)
		);
		if(*list == NULL)
		{
		    *total_rtn = 0;
		    continue;
		}

		(*list)[i] = item = VSMenuItemNew();
		if(item == NULL)
		    break;

		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* Type */
		    if(!strcasecmp(parm, "Type"))
		    {
			item->type = ATOI(val);
		    }
		    /* Text */
		    else if(!strcasecmp(parm, "Text") ||
			    !strcasecmp(parm, "Name")
		    )
		    {
			free(item->text);
			item->text = STRDUP(val);
		    }
		    /* Action */
		    else if(!strcasecmp(parm, "Action") ||
			    !strcasecmp(parm, "ActionCode")
		    )
		    {
			item->action = ATOI(val);
		    }
		    /* EndVSMenuItem */
		    else if(!strcasecmp(parm, "EndVSMenuItem"))
		    {
			break;
		    }
		    /* Unknown parameter */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
	    /* Unknown parameter */
	    else
	    {
		WARN_NO_SUCH_PARM(parm);
	    }
	}
	free(s);

#undef WARN_NO_SUCH_PARM
#undef WARN_NO_SUCH_FILE

	/* Close the file */
	fclose(fp);

	return(0);
}


/*
 *	Saves the Viewscreen Menu configuration to the specified file.
 */
int VSMenuFileSave(
	const char *path,
	xsw_vsmenu_item_struct **list, const int total
)
{
	FILE *fp;

	/* Open the file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for writing");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    XSWPrintError(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	/* Header */
	fprintf(
	    fp,
"# %s Version %s Viewscreen Menu Configuration File\n\
#\n\
\n",
	    PROG_NAME_FULL,
	    PROG_VERSION
	);

	if(total > 0)
	{
	    int i;
	    xsw_vsmenu_item_struct *item;

	    for(i = 0; i < total; i++)
	    {
		item = list[i];
		if(item == NULL)
		    continue;

		fprintf(
		    fp,
		    "BeginVSMenuItem\n"
		);
		fprintf(
		    fp,
		    "\tType = %i\n",
		    item->type
		);
		if(!STRISEMPTY(item->text))
		    fprintf(
			fp,
			"\tText = %s\n",
			item->text
		    );
		fprintf(
		    fp,
		    "\tAction = %i\n",
		    item->action
		);
		fprintf(
		    fp,
		    "EndVSMenuItem\n\n"
		);
	    }
	}

	/* Close the file */
	fclose(fp);

	return(0);
}
