#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "../include/disk.h"

#include "xswinstall.h"
#include "xsw.h"


int XSWDoInstallObject(
	const char *target, const char *source,
	mode_t m,
	unsigned int flags,
	void *data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Installs local files and directories.
 */
int XSWDoInstallObject(
	const char *target, const char *source,
	mode_t m,
	unsigned int flags,
	void *data
)
{
	int status = 0;

	if(flags & XSW_INST_OBJ_FLAG_DIR)
	{
	    /* Create new directory */
	    if(target == NULL)
		return(-2);

	    /* Recursivly? */
	    if(flags & XSW_INST_OBJ_FLAG_RECURSIVE)
	    {
		/* Create new directory recursivly */
		char	*s, *full_path,
			cwd[PATH_MAX];
		struct stat stat_buf;

		s = STRDUP(XSWCurrentDirectory());
		strncpy(cwd, (s != NULL) ? s : "/", sizeof(cwd));
		cwd[sizeof(cwd) - 1] = '\0';
		free(s);

		/* Complete path */
		full_path = ISPATHABSOLUTE(target) ?
		    strdup(target) :
		    STRDUP(PrefixPaths(cwd, target));

		if(full_path == NULL)
		    return(-3);

		/* Get pointer to beginning of target path but skip
		 * first deliminator character
		 */
		s = full_path + 1;
		while(True)
		{
		    if(s == NULL)
			break;

		    s = strchr(s, DIR_DELIMINATOR);
		    if(s != NULL)
			*s = '\0';

		    if(stat(full_path, &stat_buf))
		    {
			if(mkdir(full_path, m))
			    break;
		    }

		    if(s != NULL)
		    {
			*s = DIR_DELIMINATOR;
			s++;
		    }
		}

		free(full_path);
	    }
	    else
	    {
		status = mkdir(target, m);
	    }
	}
	else
	{
	    /* Copy file */
	    if((target == NULL) || (source == NULL))
		return(-2);

	    status = CopyObject(
		target,
		source,
		NULL
	    );
	    if(status == 0)
	    {
		chmod(target, m);
	    }
	}

	return(status);
}
