#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/tga.h"
#include "../include/osw-x.h"

#include "xswscrshot.h"


int XSWScrShotWindow(const char *path, drawable_t d);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
#define IS_NUM_ODD(n)	((long)(n) & 1)


/*
 *	Saves the contents of the window to the specified tga image
 *	file.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error or memory allocation error.
 *	-7	Unable to get image data from window.
 */
int XSWScrShotWindow(const char *path, drawable_t d)
{
	int status;
	unsigned int width, height;
	image_t *img;
	tga_data_struct *td;

	if(STRISEMPTY(path) || (d == None))
	    return(-2);

	if(!OSWGetPixmapGeometry(d, &width, &height, NULL))
	    return(-7);

	if((width <= 1) || (height <= 1))
	    return(-2);

	if(IS_NUM_ODD(width))
	    width -= 1;
	if(IS_NUM_ODD(height))
	    height -= 1;

	img = OSWGetImage(d, 0, 0, width, height);
	if(img == NULL)
	    return(-7);

	/* Save to tga image */
	td = (tga_data_struct *)calloc(
	    1, sizeof(tga_data_struct)
	);
	td->fp = NULL;

	td->id_field_len = 0;
	td->cmap_type = TgaColormapNone;
	td->img_type = TgaImageDataURGB;

	td->cmap_first_color_index = 0;
	td->cmap_total_colors = 0;
	td->cmap_entry_size = 24;

	td->x = 0;
	td->y = 0;
	td->width = img->width;
	td->height = img->height;
	td->depth = 24;
	td->flags = (tga_image_flags)0;

	td->file_size = 0l;
	td->data_size = 0l;

	td->cur_load_pixel = -1;
	td->header_data = NULL;
	td->data = (u_int8_t *)img->data;
	td->data_depth = img->depth;
	td->data_bits_per_pixel = img->bits_per_pixel;

	status = TgaWriteToFile(
	   path,		/* Path to the image file */
	   td,			/* Tga data */
	   24			/* 24 bits on file */
	);

	free(td);

	OSWDestroyImage(&img);

	return(status ? -1 : 0);
}
