#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/strexp.h"
#include "../include/fio.h"   
#include "../include/disk.h"
#include "../include/prochandle.h"
#include "../include/mf.h"


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Gets the free and total memory statistics and puts them
 *	at the location pointed to in buf.
 *
 *	Returns 0 on success and -1 on error.
 */
int MemoryFree(mf_stat_struct *buf)
{
	int i, strc;
	char *s, *s2, *tmp_path, **strv;
	pid_t p;
	FILE *fp;

	if(buf == NULL)
	    return(-1);

	memset(buf, 0, sizeof(mf_stat_struct));

	/* Get tempory file name */
	s = getenv("TMPDIR");
	if(STRISEMPTY(s))
#if defined(P_tmpdir)
	    s = P_tmpdir;
#elif defined(WIN32)
	    s = "C:\\TEMP";
#else
	    s = "/tmp";
#endif
	/* Create template path */
	tmp_path = STRDUP(PrefixPaths(s, "XXXXXX"));
	if(tmp_path == NULL)
	    return(-1);

	mkstemp(tmp_path);

	/* Run df program, blocking execution */
	p = ExecBAO(FREE_CMD, tmp_path);
	if(p < 0)
	{
	    free(tmp_path);
	    return(-1);
	}

	/* Open the tempory file for reading */
	fp = fopen(tmp_path, "rb");
	if(fp == NULL)
	{
	    unlink(tmp_path);
	    free(tmp_path);
	    return(-1);
	}

	/* Skip past first line */
	s = FReadNextLineAlloc(fp, '#');
	if(s == NULL)
	{
	    fclose(fp);
	    unlink(tmp_path);
	    free(tmp_path);
	    return(-1);
	}

	free(s);	/* Discard first line */

	/* Begin reading each line */
	s = FReadNextLineAlloc(fp, '#');
	while(s != NULL)
	{
	    /* Begin parsing line */

	    /* Remove carrage returns at the end of the line */
	    s2 = strchr(s, '\n');
	    if(s2 != NULL)
		*s2 = '\0';
	    s2 = strchr(s, '\r');
	    if(s2 != NULL)
		*s2 = '\0';


	    /* Explode line */
	    strv = strexp(s, &strc);
	    if(strc <= 0)
	    {
		/* Free current line and read next line */
		free(s);
		s = FReadNextLineAlloc(fp, '#');

		continue;
	    }


	    /* Check what stats this line gives us */
	    if(!strcasecmp(strv[0], "Mem:"))
	    {
	        /* Total */
	        buf->total = ((strc >= 2) ? atol(strv[1]) : 0) * 1024;

	        /* Used */
	        buf->used = ((strc >= 3) ? atol(strv[2]) : 0) * 1024;

		/* Free */
		buf->free = ((strc >= 4) ? atol(strv[3]) : 0) * 1024;

		/* Shared */
		buf->shared = ((strc >= 5) ? atol(strv[4]) : 0) * 1024;

		/* Buffers */
		buf->buffers = ((strc >= 6) ? atol(strv[5]) : 0) * 1024;

		/* Cached */
		buf->cached = ((strc >= 7) ? atol(strv[6]) : 0) * 1024;
	    }
	    else if(!strcasecmp(strv[0], "Swap:"))
	    {
		/* Swap total */
		buf->swap_total = ((strc >= 2) ? atol(strv[1]) : 0) * 1024;

		/* Swap used */
		buf->swap_used = ((strc >= 3) ? atol(strv[2]) : 0) * 1024;

		/* Swap free */
		buf->swap_free = ((strc >= 4) ? atol(strv[3]) : 0) * 1024;
	    }

	    /* Free exploded strings */
	    for(i = 0; i < strc; i++)
		free(strv[i]);
	    free(strv);
	    strv = NULL;


	    /* Free current line and read next line */
	    free(s);
	    s = FReadNextLineAlloc(fp, '#');
	}

	/* Close and remove the tempory file */
	fclose(fp);
	unlink(tmp_path);
	free(tmp_path);

	return(0);
}


/*
 *	Returns the amount of free physical memory in bytes.
 *
 *	If buf is not NULL, then the memory stats will be copped to
 *	it's address.
 *
 *	Returns 0 on error.
 */
off_t mf(mf_stat_struct *buf)
{
	mf_stat_struct lbuf;
	if(buf == NULL)
	{
	    MemoryFree(&lbuf);
	    return(lbuf.free);
	}
	else
	{
	    MemoryFree(buf);
	    return(buf->free);
	}
}




