/*
		     Shipwars Extended Network Codes


	These are network codes used in extended CS commands.

	Each one of these codes is prefixed (after a CS_CODE_EXT)
	to each segment of data transmitted to a CS_CODE_EXT.

	Example:

	"80 30 [argument(s)...]"

	80 is CS_CODE_EXT

	30 is SWEXTCMD_STDOBJVALS

 */

#ifndef SWNETCODES_H
#define SWNETCODES_H


/*
 *	path
 */
#define SWEXTCMD_SETOCSN                10

/*
 *	ru_to_au
 *	antimatter_low_warning
 *	asteroid_field_damage_rate
 *	auto_destruct_damage_range
 *	cloak_antimatter_rate
 *	com_range
 *	damage_control_rate
 *	disarmed_weapon_lifespan
 *	engine_antimatter_rate
 *	on_fire_hp_coeff
 *	tractor_beam_antimatter_rate
 *	tractor_beam_range
 *	transporter_range
 */
#define SWEXTCMD_SET_LEGEND		11

/*
 *	title
 */
#define SWEXTCMD_SETTITLE		12


/* 13 to 19 are reserved */


/* 20 to 29 are reserved */


/*
 *	obj_num loc_type loc_obj_num locked_on intercepting_object
 *	thrust_rev_state thrust_dir thrust throttle
 *	lighting hp power antimatter shield_state
 *	selected_weapon cloak_state cloak_strength visibility
 *	damage_control
 */
#define SWEXTCMD_STDOBJVALS		30

/*
 *	obj_num type imageset owner size scanner_range
 *	velocity_max thrust_power turnrate hp_max power_max
 *	power_purity core_efficency antimatter_max
 *	total_weapons visibility
 */
#define SWEXTCMD_STDOBJMAXS		31

/*
 *	obj_num weapon_num
 *	ocs_code emission_type amount max
 *	damage_hull damage_shields range create_power
 *	delay last_used fire_sndref flags
 */
#define SWEXTCMD_STDWEPVALS		32


/* 33 to 39 are reserved */


/*
 *	obj_num;name;empire
 */
#define SWEXTCMD_SETOBJNAME		40

/*
 *	obj_num sect_x sect_y sect_z
 */
#define SWEXTCMD_SETOBJSECT		41

/* 
 *      obj_num sect_x sect_y sect_z
 */
#define SWEXTCMD_SETFOBJSECT		42

/*
 *	obj_num throttle
 */
#define SWEXTCMD_SETTHROTTLE            43

/*
 *	obj_num selected_weapon
 */
#define SWEXTCMD_SETWEAPON		44

/*
 *	obj_num arg
 *
 *	arg can be the name or number of the object to be intercepted
 *	or "#off" to turn intercept off.
 */
#define SWEXTCMD_SETINTERCEPT		45

/*
 *      obj_num tar_obj_num
 *
 *      tar_obj_num can be -1 for unlock or -2 for lock next.
 */
#define SWEXTCMD_SETWEPLOCK		46

/*
 *	obj_num shield_state shield_frequency
 */
#define SWEXTCMD_SETSHIELDS		47

/*
 *	obj_num damage_control
 */
#define SWEXTCMD_SETDMGCTL		48

/*
 *	obj_num cloak_state
 */
#define SWEXTCMD_SETCLOAK		49

/*
 *	obj_num shield_visibility
 */
#define SWEXTCMD_SETSHIELDVIS		50

/*
 *	obj_num lighting
 *
 *	Lighting is NOT a single mask value but the lighting's
 *	OR'ed mask value.
 */
#define SWEXTCMD_SETLIGHTING		51

/*
 *	obj_num channel
 */
#define SWEXTCMD_SETCHANNEL		52

/*
 *	Client sends:
 *
 *	obj_num
 *
 *	Server sends:
 *
 *      obj_num
 *      credits rmu rmu_max damage_given
 *	damage_recieved kills
 */
#define SWEXTCMD_SETSCORE		54

/* 
 *	obj_num engine_state
 */
#define SWEXTCMD_SETENGINE		55

/*
 *	obj_num op offset heading pitch bank
 */
#define SWEXTCMD_SET_OBJ_PARTICLE_TRAIL	56

#define SW_PARTICLE_TRAIL_OP_FIRE_OFF	0
#define SW_PARTICLE_TRAIL_OP_FIRE_ON	1
#define SW_PARTICLE_TRAIL_OP_SMOKE_OFF	2
#define SW_PARTICLE_TRAIL_OP_SMOKE_ON	3

/*
 *	flags imgref_num sect_x sect_y sect_z x y z
 *	size_min size_max life_span ref_obj_num
 */
#define SWEXTCMD_CREATE_EFFECT		57

#define SW_EFFECT_FLAG_DRIFT		(1 << 3)	/* Smoke puffs or fire particles */


/*
 *	type sect_x sect_y sect_z x y z size_min size_max
 *	life_span damage_hull damage_shield ref_obj_num
 */
#define SWEXTCMD_CREATE_SHOCKWAVE	58


/* 59 is reserved */


/*
 *	obj_num
 */
#define SWEXTCMD_REQNAME                60

/*
 *	obj_num
 */
#define SWEXTCMD_REQSECT		61


/*
 *	obj_num
 *	sect_x sect_y sect_z
 *	x y z
 *	heading pitch bank
 *	velocity
 *	velocity_heading velocity_pitch velocity_bank
 *	freq yield
 */
#define SWEXTCMD_FIREWEAPON		100

/*
 *	src_obj tar_obj
 */
#define SWEXTCMD_TRACTORBEAMLOCK	101

/*
 *	src_obj tar_obj bearing channel
 */
#define SWEXTCMD_HAIL			102

/*
 *      src_obj tar_obj bearing channel;message
 */
#define SWEXTCMD_COMMESSAGE		103

/*
 *	src_obj tar_obj
 */
#define SWEXTCMD_WORMHOLEENTER		104

/*
 *	src_obj tar_obj
 *
 *	Note: client sends this to server, server sends back a forward
 *	which the client will handle and do the actual bamf.
 */
#define SWEXTCMD_ELINKENTER		105

/*
 *	src_obj tar_obj
 *
 *	When server sends to client it is a notify of a weapon being
 *	disarmed.
 */
#define SWEXTCMD_WEPDISARM		106

/*
 *	obj_num sec
 *
 *	If sec is zero then it means cancel.
 *
 *	When server sends to client, it is a notify of auto destruct.
 */
#define SWEXTCMD_AUTODESTRUCT		107


/*
 *	obj_num;description
 */
#define SWEXTCMD_SETDESCRIPTION		115

/*
 *	obj_num
 */
#define SWEXTCMD_REQDESCRIPTION		116


/*
 *	wep_obj_num
 *	tar_obj_num
 *	total_damage
 *	bearing
 *	structure_damage
 *	shield_damage
 *
 *	Bearing is in radians relative from tar_obj_num to wep_obj_num.
 */
#define SWEXTCMD_NOTIFYHIT		120

/*
 *	reason
 *	destroyed_obj_num
 *	destroyer_obj_num
 *	destroyer_obj_owner_num
 */
#define SWEXTCMD_NOTIFYDESTROY		121

#define SW_DESTROY_REASON_UNSPECIFIC	0	/* Just plain destroyed */
#define SW_DESTROY_REASON_SELF_DESTRUCT	1
#define SW_DESTROY_REASON_COLLISION	2


/*
 *	obj_num
 */
#define SWEXTCMD_ECO_REQVALUES		140

/*
 *	obj_num flags tax_general tax_friend tax_hostile
 */
#define SWEXTCMD_ECO_SETVALUES		141

/*
 *	obj_num ocs_code sell_price buy_price
 *	product_amount product_max;product_name
 */
#define SWEXTCMD_ECO_SETPRODUCTVALUES	142

/*
 *      customer_obj proprietor_obj amount magic_number;product_name
 */
#define SWEXTCMD_ECO_BUY		143

/*
 *      customer_obj proprietor_obj ocs_code amount
 */
#define SWEXTCMD_ECO_SELL		144

/*
 *      customer_obj proprietor_obj ocs_code amount
 */
#define SWEXTCMD_ECO_TRADE		145



/*
 *	obj_num
 *	type isref_num size
 *	sect_x sect_y sect_z
 *	x, y, z
 *	heading, pitch, bank
 */
#define SWEXTCMD_STARCHART_ADD_OBJECT	160

/*
 *	obj_num;name
 */
#define SWEXTCMD_STARCHART_SET_OBJECT_NAME	161

/*
 *      obj_num;empire
 */
#define SWEXTCMD_STARCHART_SET_OBJECT_EMPIRE	162

/*
 *	obj_num
 */
#define SWEXTCMD_STARCHART_SET_OBJECT_RECYCLE	163


#endif /* SWNETCODES_H */
