#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/strexp.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "mon.h"
#include "conwin.h"

#include "../include/oswkeyvals.h"


static void ConWinUnfocusPrompt(prompt_struct *prompt);
static void ConWinRefocusPrompts(
	con_win_struct *cw, prompt_struct *focus_prompt
);

static void ConWinConnectPBCB(push_button_struct *btn, void *data);
static void ConWinCancelPBCB(push_button_struct *btn, void *data);

int ConWinInit(
	con_win_struct *cw,
        void (*connect_cb)(            
                const char *, 
                int,
                const char *, 
                const char *, 
                void *
        ), void *connect_data
);
void ConWinDraw(con_win_struct *cw);
int ConWinManage(con_win_struct *cw, event_t *event);
void ConWinMap(con_win_struct *cw);
void ConWinUnmap(con_win_struct *cw);
void ConWinDestroy(con_win_struct *cw);


#define CW_DEF_WIDTH	380
#define CW_DEF_HEIGHT	165

#define CW_BTN_WIDTH	70
#define CW_BTN_HEIGHT	28


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void ConWinUnfocusPrompt(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return;

	PromptUnmarkBuffer(prompt, PROMPT_POS_END);
	prompt->has_focus = False;

	if(prompt->map_state)
	    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
}

static void ConWinRefocusPrompts(
	con_win_struct *cw,
	prompt_struct *focus_prompt
)
{
	if(cw == NULL)
	    return;

	ConWinUnfocusPrompt(&cw->address_prompt);
	ConWinUnfocusPrompt(&cw->port_prompt);
	ConWinUnfocusPrompt(&cw->name_prompt);
	ConWinUnfocusPrompt(&cw->password_prompt);

	PromptMarkBuffer(focus_prompt, PROMPT_POS_END);
	focus_prompt->has_focus = True;

	if(focus_prompt->map_state)
	    PromptQueueDraw(focus_prompt, PROMPT_DRAW_AMOUNT_TEXT);
}


static void ConWinConnectPBCB(push_button_struct *btn, void *data)
{
	con_win_struct *cw = (con_win_struct *)data;
	if(cw == NULL)
	    return;

	if(cw->connect_cb != NULL)
	{
	    char *address = STRDUP(PromptGetS(&cw->address_prompt));
	    int port = PromptGetI(&cw->port_prompt);
	    char *login_name = STRDUP(PromptGetS(&cw->name_prompt));
	    char *login_password = STRDUP(PromptGetS(&cw->password_prompt));
	    cw->connect_cb(
		address,
		port,
		login_name,
		login_password,
		cw->connect_data
	    );
	    free(address);
	    free(login_name);
	    free(login_password);
	}

	ConWinUnmap(cw);
}

static void ConWinCancelPBCB(push_button_struct *btn, void *data)
{
	con_win_struct *cw = (con_win_struct *)data;
	if(cw == NULL)
	    return;
 
	ConWinUnmap(cw);
}


int ConWinInit(
	con_win_struct *cw,
        void (*connect_cb)(
		const char *,
		int,
		const char *,
		const char *,
		void *
        ), void *connect_data
)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	win_t w, parent = osw_gui[0].root_win;
	push_button_struct *btn;

	if(cw == NULL)
	    return(-1);

	cw->map_state = False;
	cw->has_focus = False;
	cw->x = 0;
	cw->y = 0;
	cw->width = CW_DEF_WIDTH;
	cw->height = CW_DEF_HEIGHT;
	cw->visibility_state = VisibilityFullyObscured;

	cw->connect_cb = connect_cb;
	cw->connect_data = connect_data;

	/* Toplevel */
	x = cw->x;
	y = cw->y;
	width = cw->width;
	height = cw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	cw->toplevel = w;
	WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_POINTER);

	OSWSetWindowWMProperties(
	    w,
	    "Connect To Universe",	/* Title */
	    "Connect",			/* Icon title */
	    (mon_image.monitor_icon_pm != None) ?
		mon_image.monitor_icon_pm :
		widget_global.std_icon_pm,
	    False,		/* Let WM set coordinates? */
	    cw->x, cw->y,
	    cw->width, cw->height,
	    cw->width, cw->height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(
	    w,
	    OSW_EVENTMASK_TOPLEVEL | ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask | ExposureMask |
	    EnterWindowMask | LeaveWindowMask
	);

	parent = w;

	/* Address prompt */
	PromptInit(
	    &cw->address_prompt, parent,
	    10, 10 + (30 * 0),
	    (int)cw->width - 130,
	    30,
	    PROMPT_STYLE_FLUSHED,
	    "Address:",
	    HOST_NAME_MAX, 0
	);

	/* Port prompt */
	PromptInit(
	    &cw->port_prompt, parent,
	    (int)cw->width - 130 + 15,
	    10 + (30 * 0),
	    130 - 25,
	    30,
	    PROMPT_STYLE_FLUSHED,
	    "Port:",
	    40, 0
	);

	/* Login name prompt */
	PromptInit(
	    &cw->name_prompt, parent,
	    10,
	    10 + (30 * 1) + 5,
	    (int)cw->width - 20,
	    30,
	    PROMPT_STYLE_FLUSHED,
	    "    Name:",
	    80, 0
	);

	/* Login password prompt */
	PromptInit(
	    &cw->password_prompt, parent,
	    10,
	    10 + (30 * 2) + 5, 
	    (int)cw->width - 20, 
	    30,
	    PROMPT_STYLE_FLUSHED,
	    "Password:",
	    80, 0
	);

	/* Link prompts togeather */
	cw->address_prompt.next = &cw->port_prompt;
	cw->address_prompt.prev = &cw->password_prompt;

	cw->port_prompt.next = &cw->name_prompt;
	cw->port_prompt.prev = &cw->address_prompt;

	cw->name_prompt.next = &cw->password_prompt;
	cw->name_prompt.prev = &cw->port_prompt;

	cw->password_prompt.next = &cw->address_prompt;
	cw->password_prompt.prev = &cw->name_prompt;


	/* Connect button */
	btn = &cw->connect_btn;
	width = CW_BTN_WIDTH;
	height = CW_BTN_HEIGHT;
	x = 10;
	y = (int)cw->height - (int)height - 10;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Connect",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    ConWinConnectPBCB, cw
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_Return;
	accel_mods[1] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 2);

	/* Cancel button */
	btn = &cw->cancel_btn;
	x = (int)cw->width - (int)width - 10;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    ConWinCancelPBCB, cw
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);


	return(0);
}


void ConWinDraw(con_win_struct *cw)
{
	int y;
	unsigned int width, height;
	font_t *prev_font;
	pixmap_t pixmap;
	win_t w;

	if(cw == NULL)
	    return;

	/* Map as needed */
	if(!cw->map_state)
	{
	    OSWMapRaised(cw->toplevel);
	    cw->map_state = True;
	    WidgetMap(&cw->address_prompt);
	    WidgetMap(&cw->port_prompt);
	    WidgetMap(&cw->name_prompt);
	    WidgetMap(&cw->password_prompt);
	    WidgetMap(&cw->connect_btn);
	    WidgetMap(&cw->cancel_btn);
	}

	/* Recreate buffers as needed */
	if(cw->toplevel_buf == 0)
	{
	    OSWGetWindowGeometry(
		cw->toplevel, NULL, NULL, &width, &height, NULL
	    );
	    if(OSWCreatePixmap(
		&cw->toplevel_buf, width, height
	    ))
		return;
	}

	prev_font = OSWQueryCurrentFont();


	/* Redraw toplevel */
	if(True)
	{
	    w = cw->toplevel;
	    pixmap = cw->toplevel_buf;
	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    if(widget_global.force_mono)
	    {
		OSWClearPixmap(
		    pixmap,
		    width, height,
		    osw_gui[0].black_pix
		);
	    }
	    else
	    {
		WidgetPutImageTile(
		    pixmap,
		    widget_global.std_bkg_img,
		    width, height
		);

		OSWSetForeground(widget_global.surface_shadow_pix);
		y = (int)height - CW_BTN_HEIGHT - 25;
		OSWDrawLine(
		    pixmap,
		    0, y, width, y
		);

		OSWSetForeground(widget_global.surface_highlight_pix);
		y = (int)height - CW_BTN_HEIGHT - 25 + 1;
		OSWDrawLine(
		    pixmap,
		    0, y, width, y
		);
	    }

	    OSWPutDrawableToWindow(w, pixmap);
	}

	OSWSetFont(prev_font);
}


int ConWinManage(con_win_struct *cw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	keycode_t keycode;
	win_t w;

	if((cw == NULL) || (event == NULL))
	    return(events_handled);

	if(!cw->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	/* Manage events for monitor normally */
	switch(event->type)
	{
          case Expose:
            if(w == cw->toplevel)
            {
                events_handled++;
            }
            break;

	  case KeyPress:
	    if(!cw->has_focus)
		break;

	    keycode = event->xkey.keycode;

	    break;

	  case KeyRelease:
	    if(!cw->has_focus)
		break;

	    keycode = event->xkey.keycode;

	    break;

	  case FocusIn:
	    if(w == cw->toplevel)
	    {
		cw->has_focus = True;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case FocusOut:
	    if(w == cw->toplevel)
	    {
		cw->has_focus = False;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case MapNotify:
	    if(w == cw->toplevel)
	    {
		if(!cw->map_state)
		    ConWinMap(cw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case UnmapNotify:
	    if(w == cw->toplevel)
	    {
		if(cw->map_state)
		    ConWinUnmap(cw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus */
	    if(OSWEventIsTakeFocus(cw->toplevel, event, &t))
	    {
		OSWGrabFocusTime(cw->toplevel, t);
		events_handled++;
                return(events_handled);
	    }
	    /* WM Delete Window */
	    else if(OSWEventIsDeleteWindow(cw->toplevel, event))
	    {
		ConWinUnmap(cw);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	if(events_handled > 0)
	    ConWinDraw(cw);

	if(events_handled == 0)
	    events_handled += WidgetManage(&cw->connect_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&cw->cancel_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&cw->address_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&cw->port_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&cw->name_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&cw->password_prompt, event);

	return(events_handled);
}


void ConWinMap(con_win_struct *cw)
{
	if(cw == NULL)
	    return;

	cw->map_state = False;
	ConWinDraw(cw);

	ConWinRefocusPrompts(
	    cw,
	    &cw->address_prompt
	);

	OSWSetModal(cw->toplevel);
	OSWGrabFocus(cw->toplevel);
}

void ConWinUnmap(con_win_struct *cw)
{
	if(cw == NULL)
	    return;

	OSWUnmapWindow(cw->toplevel);
	cw->map_state = False;
	cw->has_focus = False;
	cw->visibility_state = VisibilityFullyObscured;

        WidgetUnmap(&cw->connect_btn);
        WidgetUnmap(&cw->cancel_btn);
        WidgetUnmap(&cw->address_prompt);
        WidgetUnmap(&cw->port_prompt);
        WidgetUnmap(&cw->name_prompt);
        WidgetUnmap(&cw->password_prompt);

	OSWDestroyPixmap(&cw->toplevel_buf);
}


void ConWinDestroy(con_win_struct *cw)
{
	if(cw == NULL)
	    return;

	WidgetDestroy(&cw->connect_btn);
	WidgetDestroy(&cw->cancel_btn);
	WidgetDestroy(&cw->address_prompt);
	WidgetDestroy(&cw->port_prompt);
	WidgetDestroy(&cw->name_prompt);
	WidgetDestroy(&cw->password_prompt);
	OSWDestroyWindow(&cw->toplevel);
	OSWDestroyPixmap(&cw->toplevel_buf);

	cw->connect_cb = NULL;
	cw->connect_data = NULL;
}
