#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "mon.h"
#include "mesgwin.h"

#include "../include/oswkeyvals.h"


void MesgWinMarkCoordinates(
	mesgwin_struct *mw,
	int start_x, int start_y,
	int end_x, int end_y
);
void MesgWinUnmarkAll(mesgwin_struct *mw);
void MesgWinPutDDE(mesgwin_struct *mw);

void MesgWinAppend(
	mesgwin_struct *mw, const char *msg, pixel_t pix
);

static void MesgWinResizeCB(void *widget, void *data);
static void MesgWinDrawCB(void *widget, void *data);
static void MesgWinSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
);

int MesgWinInit(mesgwin_struct *mw, int argc, char **argv);
void MesgWinResize(mesgwin_struct *mw);
void MesgWinQueueResize(mesgwin_struct *mw);
void MesgWinDraw(mesgwin_struct *mw);
void MesgWinQueueDraw(mesgwin_struct *mw);
int MesgWinManage(mesgwin_struct *mw, event_t *event);
void MesgWinMap(mesgwin_struct *mw);
void MesgWinUnmap(mesgwin_struct *mw);
void MesgWinDestroy(mesgwin_struct *mw);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


#define MESGWIN_MARGIN		10

#define MESGWIN_CHAR_WIDTH	7
#define MESGWIN_CHAR_HEIGHT	14

#define MESGWIN_LINE_SPACING	(MESGWIN_CHAR_HEIGHT + 2)

#define MESGWIN_DEF_WIDTH	640
#define MESGWIN_DEF_HEIGHT	480

#define MESGWIN_DEF_LINES	200	/* Lines */
#define MESGWIN_DEF_LINE_LEN	256	/* Characters */


/*
 *	Updates the message selection with the specified values in
 *	window coordinates.
 */
void MesgWinMarkCoordinates(
	mesgwin_struct *mw,
	int start_x, int start_y,
	int end_x, int end_y
)
{
	int	start_line, end_line,
		start_line_x, end_line_x,
		margin_x, margin_y, line_spacing;
	scroll_bar_struct *sb;

	if(mw == NULL)
	    return;

	sb = &mw->sb;
	margin_x = MESGWIN_MARGIN;
	margin_y = MESGWIN_MARGIN;
	line_spacing = MESGWIN_LINE_SPACING;
	if(line_spacing <= 0)
	    return;

	/* Calculate start and end lines */
	start_line = CLIP(
	    ((start_y - margin_y + (int)SBarGetPositionY(sb)) / line_spacing),
	    0, (mw->total_messages - 1)
	);
	end_line = CLIP(
	    ((end_y - margin_y + (int)SBarGetPositionY(sb)) / line_spacing),
	    0, (mw->total_messages - 1)
	);

	/* Calculate the start and end line x positions (do not clip) */
	start_line_x = (start_x - margin_x + (int)SBarGetPositionX(sb)) /
	    MESGWIN_CHAR_WIDTH;
	end_line_x = (end_x - margin_x + (int)SBarGetPositionX(sb)) /
	    MESGWIN_CHAR_WIDTH;

#define MSG_SEL_SWAP_AS_NEEDED(_msg_)	{	\
 if((_msg_)->sel_start > (_msg_)->sel_end) {	\
  const int i = (_msg_)->sel_end;		\
  (_msg_)->sel_end = (_msg_)->sel_start;	\
  (_msg_)->sel_start = i;			\
} }

	/* Single line? */
	if(start_line == end_line)
	{
	    mesgwin_mesg_struct *msg = mw->message[start_line];
	    const int line_len = strlen(msg->message);
	    if(start_line_x < end_line_x)
	    {
		msg->sel_start = CLIP(start_line_x, 0, (line_len - 1));
		msg->sel_end = CLIP(end_line_x - 1, 0, (line_len - 1));
	    }
	    else if(start_line_x > end_line_x)
	    {
		msg->sel_start = CLIP(end_line_x, 0, (line_len - 1));
		msg->sel_end = CLIP(start_line_x - 1, 0, (line_len - 1));
	    }
	}
	/* Select multiple lines top to bottom? */
	else if(start_line < end_line)
	{
	    mesgwin_mesg_struct *msg = mw->message[start_line];
	    int i, line_len = strlen(msg->message);
	    if(start_line_x < line_len)
	    {
		msg->sel_start = CLIP(start_line_x, 0, (line_len - 1));
		msg->sel_end = line_len - 1;
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }

	    for(i = start_line + 1; i < end_line; i++)
	    {
		msg = mw->message[i];
		msg->sel_start = 0;
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
	    }

	    msg = mw->message[end_line];
	    line_len = strlen(msg->message);
	    if(end_line_x >= 0)
	    {
		msg->sel_start = 0;
		msg->sel_end = CLIP(end_line_x, 0, (line_len - 1));
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }
	}
	/* Select multiple lines bottom to top? */
	else if(end_line < start_line)
	{
	    mesgwin_mesg_struct *msg = mw->message[start_line];
	    int i, line_len = strlen(msg->message);
	    if(start_line_x >= 0)
	    {
		msg->sel_start = 0;
		msg->sel_end = CLIP(start_line_x - 1, 0, (line_len - 1));
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }
	     
	    for(i = start_line - 1; i > end_line; i--)
	    {
		msg = mw->message[i];
		msg->sel_start = 0;
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
	    }

	    msg = mw->message[end_line];
	    line_len = strlen(msg->message);
	    if(end_line_x < line_len)
	    {
		msg->sel_start = CLIP(end_line_x, 0, (line_len - 1));
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }
	}
#undef MSG_SEL_SWAP_AS_NEEDED
}

/*
 *	Unmarks all lines in message window.
 */
void MesgWinUnmarkAll(mesgwin_struct *mw)
{
	int i;
	mesgwin_mesg_struct *msg;

	if(mw == NULL)
	    return;

	for(i = 0; i < mw->total_messages; i++)
	{
	    msg = mw->message[i];
	    if(msg == NULL)
		continue;

	    msg->sel_start = -1;
	    msg->sel_end = -1;
	}
}


/*
 *	Puts the selected message to the clipboard.
 */
void MesgWinPutDDE(mesgwin_struct *mw)
{
	int i, len;
	char *buf = NULL;
	mesgwin_mesg_struct *msg;
	int buf_pos = 0;
	int buf_len = 0;

	if(mw == NULL)
	    return;

	for(i = 0; i < mw->total_messages; i++)
	{
	    msg = mw->message[i];
	    if(msg == NULL)
		continue;

	    if(msg->message == NULL)
		continue;

	    if((msg->sel_start < 0) || (msg->sel_end < 0))
		continue;

	    /* Sanitize selected start and end positions */
	    if(msg->sel_start >= msg->length)
		msg->sel_start = msg->length - 1;
	    if(msg->sel_end >= msg->length)
		msg->sel_end = msg->length - 1;

	    /* Get the length of the selection on this line */
	    len = msg->sel_end - msg->sel_start + 1;
	    if(len < 1)
		continue;

	    if((msg->sel_start + len) >= msg->length)
		len = msg->length - msg->sel_start - 1;
	    if(len <= 0)
		continue;

	    /* If this is second or greater line, add \n to buffer */
	    if(buf_len > 0)
	    {
		buf_len += 1;
		buf = (char *)realloc(buf, (buf_len + 1) * sizeof(char));
		if(buf == NULL)
		    return;

		buf_pos = buf_len - 1;
		buf[buf_pos] = '\n';
	    }

	    /* Increase buffer size */
	    buf_pos = buf_len;
	    buf_len += len;
	    buf = (char *)realloc(buf, (buf_len + 1) * sizeof(char));
	    if(buf == NULL)
		return;

	    memcpy(
		buf + buf_pos,
		msg->message + msg->sel_start,
		len
	    );
	}

	/* Null terminate */
	if(buf_len > 0)
	{
	    buf[buf_len] = '\0';
	    OSWDDESetString(mw->toplevel, buf);
	}

	free(buf);
}


/*
 *	Appends a line to the message window's messages.
 */
void MesgWinAppend(
	mesgwin_struct *mw, const char *msg, pixel_t pix
) 
{
	int i, m, len, longest_line = 0;
	mesgwin_mesg_struct *src_msg, *tar_msg;

	if((mw == NULL) || (msg == NULL))
	    return;

	/* Shift messages */
	m = mw->total_messages - 1;
	for(i = 0; i < m; i++)
	{
	    tar_msg = mw->message[i];
	    src_msg = mw->message[i + 1];

	    if((src_msg == NULL) ||
	       (tar_msg == NULL)
	    )
		continue;

	    if((src_msg->message == NULL) ||
	       (tar_msg->message == NULL)
	    )
		continue;

	    memcpy(
		tar_msg->message,
		src_msg->message,
		MIN(tar_msg->length,
		    src_msg->length
		) * sizeof(char)
	    );

	    /* Update the longest line */
	    len = strlen(tar_msg->message);
	    if(len > longest_line)
		longest_line = len;

	    tar_msg->pixel = src_msg->pixel;
	    tar_msg->sel_start = src_msg->sel_start;
	    tar_msg->sel_end = src_msg->sel_end;

	    /* Do not copy msg->length */
	}

	i = mw->total_messages - 1;
	tar_msg = (i >= 0) ? mw->message[i] : NULL;
	if(tar_msg == NULL)
	    return;

	if(tar_msg->message == NULL)
	    return;

	strncpy(
	    tar_msg->message,
	    msg,
	    tar_msg->length
	);
	tar_msg->message[tar_msg->length - 1] = '\0';


	len = strlen(tar_msg->message);
	if(len > longest_line)
	    longest_line = len;

	tar_msg->sel_start = -1;
	tar_msg->sel_end = -1;

	tar_msg->pixel = pix;


	/* Update the longest line */
	mw->longest_line = longest_line;

	/* Resize the scroll bar */
	SBarQueueResize(
	    &mw->sb,
	    mw->width, mw->height,
	    (mw->longest_line * MESGWIN_CHAR_WIDTH) +
		SCROLLBAR_YBAR_WIDTH + 20,
	    (mw->total_messages * MESGWIN_LINE_SPACING) +
		SCROLLBAR_XBAR_HEIGHT + 20
	);
	if(mw->map_state)
	{
	    WidgetQueueDraw(&mw->sb);
	    MesgWinQueueDraw(mw);
	}
}


/*
 *	Resize callback.
 */
static void MesgWinResizeCB(void *widget, void *data)
{
	mesgwin_struct *mw = MESGWIN(data);
	if(mw == NULL)
	    return;

	MesgWinResize(mw);
}

/*
 *	Draw callback.
 */
static void MesgWinDrawCB(void *widget, void *data)
{
	mesgwin_struct *mw = MESGWIN(data);
	if(mw == NULL)
	    return;

	if(mw->map_state)
	    MesgWinDraw(mw);
}

/*
 *	Scroll bar value changed callback
 */
static void MesgWinSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
)
{
	mesgwin_struct *mw = MESGWIN(data);
	if(mw == NULL)
	    return;

	if(mw->map_state)
	    MesgWinQueueDraw(mw);
}


/*
 *	Initializes the message window.
 */
int MesgWinInit(mesgwin_struct *mw, int argc, char **argv)
{
	int i, x, y;
	unsigned int width, height;
	scroll_bar_struct *sb;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	mesgwin_mesg_struct *msg;
	widget_global_struct *wglobal = &widget_global;

	if(mw == NULL)
	    return(-1);

	mw->map_state = False;
	mw->has_focus = False;
	mw->visibility_state = VisibilityFullyObscured;
	mw->x = gui->def_toplevel_x;
	mw->y = gui->def_toplevel_y;
	mw->width = MESGWIN_DEF_WIDTH;
	mw->height = MESGWIN_DEF_HEIGHT;
	mw->bkg = NULL;
	mw->message = NULL;
	mw->total_messages = 0;
	mw->longest_line = 0;
	mw->sel_in_progress = False;
	mw->sel_start_x = 0;
	mw->sel_start_y = 0;

	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    if(argv[i] == NULL)
		continue;


	}

	/* Allocate message list */
	mw->total_messages = MESGWIN_DEF_LINES;
	mw->message = (mesgwin_mesg_struct **)calloc(
	    mw->total_messages, sizeof(mesgwin_mesg_struct *)
	);
	if(mw->message != NULL)
	{
	    for(i = 0; i < mw->total_messages; i++)
	    {
		mw->message[i] = msg = (mesgwin_mesg_struct *)calloc(
		    1, sizeof(mesgwin_mesg_struct)
		);
		if(msg == NULL)
		    continue;

		msg->length = MESGWIN_DEF_LINE_LEN;
		msg->message = (char *)calloc(
		    msg->length, sizeof(char)
		);
		if(msg->message == NULL)
		    msg->length = 0;

		msg->sel_start = -1;
		msg->sel_end = -1;

		msg->pixel = gui->white_pix;
	    }
	}

	/* Create toplevel */
	x = mw->x;
	y = mw->y;
	width = mw->width;
	height = mw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	mw->toplevel = w;
	WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);
	OSWSetWindowWMProperties(
	    w,
	    "Messages",
	    "Messages",
	    (mon_image.monitor_icon_pm != None) ?
		mon_image.monitor_icon_pm :
		wglobal->std_icon_pm,
	    False,	/* Let WM set coordinates? */
	    mw->x, mw->y,
	    100, 100,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_STANDARD,
	    0, NULL
	);
	OSWSetWindowInput(
	    w,
	    OSW_EVENTMASK_TOPLEVEL | ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask | ExposureMask
	);
	WidgetSetWindowCursor(w, mon_cursor.text);

	parent = w;

	/* Scroll bars */
	width = mw->width;
	height = mw->height;
	sb = &mw->sb;
	SBarInit(
	    sb, parent,
	    width, height,
	    (mw->longest_line * MESGWIN_CHAR_WIDTH) + 30,
	    (mw->total_messages * MESGWIN_LINE_SPACING) + 30
	);
	sb->value_changed_cb = MesgWinSBValueChangedCB;
	sb->value_changed_data = mw;


	return(0);
}

/*
 *	Resizes the Message Window.
 */
void MesgWinResize(mesgwin_struct *mw)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	scroll_bar_struct *sb;

	if(mw == NULL)
	    return;

	w = mw->toplevel;
	OSWGetWindowGeometry(
	    w, &x, &y, &width, &height, NULL
	);

	mw->x = x;
	mw->y = y;

	if((width == mw->width) && (height == mw->height))
	    return;

	mw->width = width;
	mw->height = height;

	sb = &mw->sb;
	SBarQueueResize(
	    sb,
	    width, height,
	    (mw->longest_line * MESGWIN_CHAR_WIDTH) +
		SCROLLBAR_YBAR_WIDTH + 20,
	    (mw->total_messages * MESGWIN_LINE_SPACING) +
		SCROLLBAR_XBAR_HEIGHT + 20
	);
	SBarSetPositionX(sb, 0.0f);
	SBarSetPositionY(sb, (float)((int)sb->total_height - (int)sb->height));

	OSWSyncSharedImage(mw->bkg, mw->toplevel_buf);
	OSWDestroySharedImage(&mw->bkg);
	
	OSWDestroyPixmap(&mw->toplevel_buf);
}

/*
 *	Queues the Message Window to resize.
 */
void MesgWinQueueResize(mesgwin_struct *mw)
{
	if(mw == NULL)
	    return;

	WidgetQueueResizeAction(
	    MesgWinResizeCB, NULL, mw
	);
}

/*
 *	Draws the Message Window.
 */
void MesgWinDraw(mesgwin_struct *mw)
{
	unsigned int width, height;
	win_t w;
	pixmap_t pixmap;
	image_t *img;
	shared_image_t *simg;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(mw == NULL)
	    return;

	/* Map as needed */
	if(!mw->map_state)
	{
	    OSWMapRaised(mw->toplevel);
	    mw->map_state = True;
	}

	/* Recreate buffers as needed */
	if(mw->toplevel_buf == None)
	{
	    OSWGetWindowGeometry(
		mw->toplevel, NULL, NULL, &width, &height, NULL
	    );
	    if(OSWCreatePixmap(
		&mw->toplevel_buf, width, height
	    ))
		return;
	}

	/* Resize background? */
	if(mw->bkg == NULL)
	{
	    OSWGetWindowGeometry(
		mw->toplevel, NULL, NULL, &width, &height, NULL
	    );
	    if(OSWCreateSharedImage(&simg, width, height))
		return;

	    if(mon_image.mesg_bkg != NULL)
	    {
		img = mon_image.mesg_bkg;

		WidgetResizeImageBuffer(
		    gui->depth,
		    simg->data,				/* Target */
		    (const u_int8_t *)img->data,	/* Source */
		    simg->width, simg->height,
		    simg->bytes_per_line,
		    img->width, img->height,
		    img->bytes_per_line
		);
	    }

	    mw->bkg = simg;
	}

	if(True)
	{
	    int	i, x, y,
		lines_drawn, lines_visible,
		lbearing, swidth, ascent, descent,
		font_height;
	    const char *s;
	    font_t *font = mon_font.messages;
	    scroll_bar_struct *sb = &mw->sb;
	    mesgwin_mesg_struct *msg;

	    w = mw->toplevel;
	    pixmap = mw->toplevel_buf;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Draw background */
	    if(wglobal->force_mono)
		OSWClearPixmap(
		    pixmap, width, height,
		    gui->black_pix
		);
	    else
		OSWPutSharedImageToDrawable(mw->bkg, pixmap);

	    /* Calculate the starting line */
	    i = MAX(
		(int)SBarGetPositionY(sb) / MESGWIN_LINE_SPACING,
		0
	    );

	    x = MESGWIN_MARGIN - (int)SBarGetPositionX(sb);
	    y = MESGWIN_MARGIN -
		((int)SBarGetPositionY(sb) % MESGWIN_LINE_SPACING);

	    lines_drawn = 0;
	    lines_visible = (int)height / MESGWIN_LINE_SPACING;

	    while(lines_drawn < lines_visible)
	    {
		if(i >= mw->total_messages)
		    break;

		msg = mw->message[i];
		if(msg == NULL)
		{
		    y += MESGWIN_LINE_SPACING;
 		    i++;
		    lines_drawn++;
		    continue;
		}

		OSWSetFont(font);

		/* This line not selected? */
		if((msg->sel_start < 0) || (msg->sel_end < 0))
		{
		    /* Draw unselected text */
		    if(wglobal->force_mono)
			OSWSetForeground(gui->white_pix);
		    else
			OSWSetForeground(msg->pixel);

		    s = msg->message;
		    OSWGetTextBounds(
			font, s, -1,
			&lbearing, NULL, NULL, &ascent, NULL
		    );
		    OSWDrawString(
			pixmap,
			x - lbearing,
			y + ascent,
			s
		    );
		}
		else
		{
		    int n, x_s2 = x, x_s3 = x, len;

		    /* Draw first unmarked text */
		    s = msg->message;
		    len = msg->sel_start;
		    if(len > 0)
		    {
			if(wglobal->force_mono)
			    OSWSetForeground(gui->white_pix);
			else
			    OSWSetForeground(msg->pixel);
			OSWGetTextBounds(
			    font, s, len,
			    &lbearing, NULL, &swidth, &ascent, NULL
			);
			OSWDrawStringLimited(
			    pixmap,
			    x - lbearing,
			    y + ascent,
			    s, len
			);
			x_s2 += swidth - lbearing;
			x_s3 = x_s2;
		    }

		    /* Draw marked text */
		    n = msg->sel_start;
		    if((n >= 0) && (n < msg->length))
		    {
			s = msg->message + n;
			len = msg->sel_end - n + 1;
			if(len > 0)
			{
			    OSWGetTextBounds(
				font, s, len,
				&lbearing, NULL, &swidth, &ascent, &descent
			    );
			    font_height = ascent + descent;

			    /* Need to adjust x_s2 to the start
			     * coordinates if there was no unmarked
			     * text before the marked text
			     */
			    if(n == 0)
			    {
				x_s2 -= lbearing;
				x_s3 = x_s2;
			    }

			    /* Draw the selection highlight */
			    if(wglobal->force_mono)
				OSWSetForeground(gui->white_pix);
			    else
				OSWSetForeground(wglobal->surface_selected_pix);
			    OSWDrawSolidRectangle(
				pixmap,
				x_s2, y,
				swidth, MESGWIN_LINE_SPACING
			    );

			    /* Draw the selected text */
			    if(wglobal->force_mono)
				OSWSetForeground(gui->black_pix);
			    else
				OSWSetForeground(wglobal->selected_text_pix);
			    OSWDrawStringLimited(
				pixmap,
				x_s2,
				y + ascent,
				s, len
			    );
			    x_s3 += swidth;
			}
		    }

		    /* Draw last unmarked text */
		    n = msg->sel_end + 1;
		    if((n >= 0) && (n < msg->length))
		    {
			s = msg->message + n;
			len = msg->length - n;
			if(len > 0)
			{
			    if(wglobal->force_mono)
				OSWSetForeground(gui->white_pix);
			    else
				OSWSetForeground(msg->pixel);

			    OSWGetTextBounds(
				font, s, len,
				&lbearing, NULL, NULL, &ascent, NULL
			    );
			    OSWDrawStringLimited(
				pixmap,
				x_s3,
				y + ascent,
				s, len
			    );
			}
		    }
		}

		/* Increment positions */
		y += MESGWIN_LINE_SPACING;
		i++;
		lines_drawn++;
	    }

	    /* Put buffer to window */
	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the Message Window to draw.
 */
void MesgWinQueueDraw(mesgwin_struct *mw)
{
	if(mw == NULL)
	    return;

	WidgetQueueDrawAction(
	    MesgWinDrawCB, NULL, mw
	);
}

/*
 *	Manages the Message Window event.
 */
int MesgWinManage(mesgwin_struct *mw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;

	if((mw == NULL) || (event == NULL))
	    return(events_handled);

	if(!mw->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	/* Manage events for monitor normally */
	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == mw->toplevel)
	    {
		MesgWinQueueResize(mw);
		WidgetQueueDraw(&mw->sb);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == mw->toplevel)
	    {
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!mw->has_focus)
		return(events_handled);

	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    if(keyval == OSW_KEY_Escape)
	    {
		MesgWinUnmap(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == mw->toplevel)
	    {
		/* Put scroll bar into focus */
		mw->sb.has_focus = True;

		/* Start mark text */
		if(button->button == Button1)
		{
		    MesgWinUnmarkAll(mw);

		    mw->sel_in_progress = True;
		    mw->sel_start_x = (int)button->x;
		    mw->sel_start_y = (int)button->y;

		    MesgWinQueueDraw(mw);

		    events_handled++;
		}
		return(events_handled);
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == mw->toplevel)
	    {
		MesgWinPutDDE(mw);
		mw->sel_in_progress = False;

		MesgWinQueueDraw(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(w == mw->toplevel)
	    {
		/* Mark text */
		if(mw->sel_in_progress)
		{
		    MesgWinUnmarkAll(mw);

		    MesgWinMarkCoordinates(
			mw,
			mw->sel_start_x,
			mw->sel_start_y,
			(int)motion->x,
			(int)motion->y
		    );

		    MesgWinQueueDraw(mw);

		    events_handled++;
		    return(events_handled);
		}
	    }
	    break;

	  case FocusIn:
	    if(w == mw->toplevel)
	    {
		mw->has_focus = True;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case FocusOut:
	    if(w == mw->toplevel)
	    {
		mw->has_focus = False;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case VisibilityNotify:
	    if(w == mw->toplevel)
	    {
		mw->visibility_state = event->xvisibility.state;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case MapNotify:
	    if(w == mw->toplevel)
	    {
		if(!mw->map_state)
		    MesgWinMap(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case UnmapNotify:
	    if(w == mw->toplevel)
	    {
		if(mw->map_state)
		    MesgWinUnmap(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus */
	    if(OSWEventIsTakeFocus(mw->toplevel, event, &t))
	    {
		OSWGrabFocusTime(mw->toplevel, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window */
	    else if(OSWEventIsDeleteWindow(mw->toplevel, event))
	    {
		MesgWinUnmap(mw);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	if(events_handled > 0)
	{
	    MesgWinQueueDraw(mw);
	    return(events_handled);
	}

	/* Manage scroll bar */
	if(events_handled == 0)
	    events_handled += WidgetManage(&mw->sb, event);

	return(events_handled);
}

/*
 *	Maps the Message window.
 */
void MesgWinMap(mesgwin_struct *mw)
{
	scroll_bar_struct *sb;

	if(mw == NULL)
	    return;

	sb = &mw->sb;
	SBarSetPositionX(sb, 0.0f);
	SBarSetPositionY(sb, (float)((int)sb->total_height - (int)sb->height));

	mw->map_state = False;
	MesgWinDraw(mw);

	OSWGrabFocus(mw->toplevel);
	mw->sb.has_focus = True;
}

/*
 *	Unmaps the Message Window.
 */
void MesgWinUnmap(mesgwin_struct *mw)
{
	if(mw == NULL)
	    return;

	OSWUnmapWindow(mw->toplevel);
	mw->map_state = False;
	mw->has_focus = False;
	mw->visibility_state = VisibilityFullyObscured;

	OSWSyncSharedImage(mw->bkg, mw->toplevel_buf);
	OSWDestroySharedImage(&mw->bkg);

	OSWDestroyPixmap(&mw->toplevel_buf);
}

/*
 *	Destroys the Message Window.
 */
void MesgWinDestroy(mesgwin_struct *mw)
{
	int i;

	if(mw == NULL)
	    return;

	for(i = 0; i < mw->total_messages; i++)
	{
	    if(mw->message[i] == NULL)
		continue;

	    free(mw->message[i]->message);
	    free(mw->message[i]);
	}
	free(mw->message);
	mw->message = NULL;
	mw->total_messages = 0;

	WidgetDestroy(&mw->sb);

	OSWSyncSharedImage(mw->bkg, mw->toplevel_buf);
	OSWDestroySharedImage(&mw->bkg);

	OSWDestroyWindow(&mw->toplevel);
	OSWDestroyPixmap(&mw->toplevel_buf);

	mw->map_state = False;
	mw->has_focus = False;
	mw->x = 0;
	mw->y = 0;
	mw->width = 0;
	mw->height = 0;
	mw->visibility_state = VisibilityFullyObscured;
}
