#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/widget.h"

/* Cursor pixmaps */
#include "../include/cursors/text_cur.xpm"
#include "../include/cursors/no_way.xpm"

/* Push button background Images */
#include "../include/images/button_armed.h"
#include "../include/images/button_unarmed.h"
#include "../include/images/button_highlighted.h"

/* Scale bar handles */
#include "../include/images/scalebar_h.h"
#include "../include/images/scalebar_v.h"

/* Background images */
#include "../include/images/bg_standard.h"

/* Menu Bar images */
#include "../include/images/menu_bar_bg.h"

/* Menu images */
#include "../include/images/menu_item.h"
#include "../include/images/menu_item_highlighted.h"

/* Tab images */
#include "../include/images/tab_unselected.h"
#include "../include/images/tab_selected.h"

/* Toggle button images */
#include "../include/images/check_button_unarmed.h"
#include "../include/images/check_button_armed.h"
#include "../include/images/check_button_highlighted.h"


/* Icon images */
#include "../include/icons/browsefiles.h"
#include "../include/icons/default.h"
#include "../include/icons/icon_file_20x20.h"
#include "../include/icons/icon_folder_closed_20x20.h"
#include "../include/icons/icon_folder_opened_20x20.h"
#include "../include/icons/icon_link_20x20.h"
#include "../include/icons/icon_pipe_20x20.h"
#include "../include/icons/icon_device_block_20x20.h"
#include "../include/icons/icon_device_character_20x20.h"
#include "../include/icons/icon_executable_20x20.h"
#include "../include/icons/parent_dir.h"
#include "../include/icons/home_dir.h"
#include "../include/icons/refresh.h"
#include "../include/icons/pulist_map.h"


/* Widget Global Initize & Shutdown */
int WidgetInitGlobals(const int argc, char **argv);
void WidgetDestroyGlobals(void);

/* Widget Manage GUI */
int WidgetGUIEventsPending(void);
int WidgetManageGUI(event_t *event);

/* Process Queued Actions and Timeouts */
int WidgetActionsPending(void);
void WidgetMainIteration(void *data);

/* Registery */
int WidgetRegister(void *widget, const int type);
int WidgetGetType(void *widget);
const char *WidgetGetTypeName(void *widget);
bool_t WidgetIsRegistered(void *widget);
void WidgetUnregister(void *widget);

/* Widget Nexus */
win_t WidgetGetToplevel(void *widget);
win_t WidgetGetParent(void *widget);
bool_t WidgetHasFocus(void *widget);
void WidgetSetFocus(void *widget, const bool_t focus);
bool_t WidgetDisabled(void *widget);
void WidgetSetDisabled(void *widget, const bool_t disabled);
void WidgetQueueResize(void *widget);
void WidgetQueueDraw(void *widget);
int WidgetManage(void *widget, event_t *event);
bool_t WidgetMapped(void *widget);
void WidgetMap(void *widget);
void WidgetUnmap(void *widget);
visibility_t WidgetVisibility(void *widget);
void WidgetDestroy(void *widget);

/* Queue Action */
void WidgetQueueResizeAction(
	void (*func)(void *, void *),
	void *widget, void *data
);
void WidgetQueueDrawAction(
	void (*func)(void *, void *),
	void *widget, void *data
);

/* Timeout */
void WidgetTimeoutAdd(
	void (*func)(void *, void *),
	void *widget, void *data,
	const time_t interval_ms
);
void WidgetIdleAdd(                                    
        void (*func)(void *, void *),                  
        void *widget, void *data
);
void WidgetTimeoutRemoveByFunc(void (*func)(void *, void *));
void WidgetTimeoutRemoveByWidget(void *widget);
void WidgetTimeoutRemoveByFuncWidget(
	void (*func)(void *, void *),
	void *widget
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


widget_global_struct	widget_global;


/*
 *	Initializes global resources for the widgets.
 */
int WidgetInitGlobals(const int argc, char **argv)
{
	int i;
	WColorStruct c_fg, c_bg;
	osw_gui_struct *gui = &osw_gui[0];
	depth_t depth = gui->depth;
	widget_global_struct *wglobal = &widget_global;


	/* Check if the depth is one that we support */
	if((depth != 8) &&
	   (depth != 15) &&
	   (depth != 16) &&
	   (depth != 24) &&
	   (depth != 32)
	)
	{
	    fprintf(
		stderr,
"WidgetInitGlobals(): Warning, depth %i is not supported.\n",
		depth
	    );
	}

	/* Begin initializing values */

	wglobal->force_mono = False;

	wglobal->double_click_int = WIDGET_DEF_DOUBLE_CLICK_INT;
	wglobal->tooltips_map_int = WIDGET_DEF_TOOLTIPS_MAP_INT;
	wglobal->rename_click_int = WIDGET_DEF_RENAME_CLICK_INT;
	wglobal->pulist_repeat_int = WIDGET_DEF_PULIST_SCROLL_REPEAT_INT;
	wglobal->prompt_scroll_repeat_int = WIDGET_DEF_PROMPT_SCROLL_REPEAT_INT;
	wglobal->sbar_scroll_repeat_delay = WIDGET_DEF_SBAR_SCROLL_REPEAT_DELAY;
	wglobal->sbar_scroll_repeat_int = WIDGET_DEF_SBAR_SCROLL_REPEAT_INT;


	/* Widgets regeristry list */
	wglobal->widget = NULL;
	wglobal->nwidgets = 0;


	/* Queued resizes */
	wglobal->queued_resize = NULL;
	wglobal->nqueued_resizes = 0;

	/* Queued draws */
	wglobal->queued_draw = NULL;
	wglobal->nqueued_draws = 0;

	/* Timeouts */
	wglobal->timeout = NULL;
	wglobal->ntimeouts = 0; 


	/* Parse arguments */
	for(i = 0; i < argc; i++)
	{
	    if(argv[i] == NULL)
		continue;

	    /* Force mono */
	    if(!strcmp(argv[i], "--force_mono") ||
	       !strcmp(argv[i], "-force_mono") ||
	       !strcmp(argv[i], "--mono") ||
	       !strcmp(argv[i], "-mono")
	    )
	    {
		wglobal->force_mono = True;
	    }
	}


	/* Begin loading colors */

	/* Surface colors */
	if(OSWLoadPixelCLSP(&wglobal->surface_normal_pix, WIDGET_CLSP_SURFACE_NORMAL))
	    wglobal->surface_normal_pix = gui->black_pix;
	if(OSWLoadPixelCLSP(&wglobal->surface_editable_pix, WIDGET_CLSP_SURFACE_EDITABLE))
	    wglobal->surface_editable_pix = gui->black_pix;
	if(OSWLoadPixelCLSP(&wglobal->surface_selected_pix, WIDGET_CLSP_SURFACE_SELECTED))
	    wglobal->surface_selected_pix = gui->white_pix;
	if(OSWLoadPixelCLSP(&wglobal->surface_shadow_pix, WIDGET_CLSP_SURFACE_SHADOW))
	    wglobal->surface_shadow_pix = gui->white_pix;
	if(OSWLoadPixelCLSP(&wglobal->surface_highlight_pix, WIDGET_CLSP_SURFACE_HIGHLIGHT))
	    wglobal->surface_highlight_pix = gui->white_pix;

	/* Scroll Bar and Scale Bar colors */
	if(OSWLoadPixelCLSP(&wglobal->scroll_bar_bg_pix, WIDGET_CLSP_SBAR_BG))
	    wglobal->scroll_bar_bg_pix = gui->black_pix;
	if(OSWLoadPixelCLSP(&wglobal->scroll_bar_frame_pix, WIDGET_CLSP_SBAR_FRAME))
	    wglobal->scroll_bar_frame_pix = gui->white_pix;
	if(OSWLoadPixelCLSP(&wglobal->scroll_bar_btn_pix, WIDGET_CLSP_SBAR_BUTTON))
	    wglobal->scroll_bar_btn_pix = gui->white_pix;
	if(OSWLoadPixelCLSP(&wglobal->scroll_bar_handle_pix, WIDGET_CLSP_SBAR_SLIDER))
	    wglobal->scroll_bar_handle_pix = gui->white_pix;

	/* Text colors */
	if(OSWLoadPixelCLSP(&wglobal->normal_text_pix, WIDGET_CLSP_TEXT_NORMAL))
	    wglobal->normal_text_pix = gui->white_pix;
	if(OSWLoadPixelCLSP(&wglobal->editable_text_pix, WIDGET_CLSP_TEXT_EDITABLE))
	    wglobal->editable_text_pix = gui->white_pix;
	if(OSWLoadPixelCLSP(&wglobal->selected_text_pix, WIDGET_CLSP_TEXT_SELECTED)) 
	    wglobal->selected_text_pix = gui->black_pix;
	if(OSWLoadPixelCLSP(&wglobal->disabled_text_pix, WIDGET_CLSP_TEXT_DISABLED))
	    wglobal->disabled_text_pix = gui->white_pix;

	/* Tooltips colors */
	if(OSWLoadPixelCLSP(&wglobal->hint_bkg_pix, WIDGET_CLSP_TOOLTIPS_BG))
	    wglobal->hint_bkg_pix = gui->white_pix;
	if(OSWLoadPixelCLSP(&wglobal->hint_text_pix, WIDGET_CLSP_TOOLTIPS_TEXT))
	    wglobal->hint_text_pix = gui->black_pix;

	/* Progress Bar colors */
	if(OSWLoadPixelCLSP(&wglobal->pbar_text_pix, WIDGET_CLSP_PBAR_TEXT))
	    wglobal->pbar_text_pix = gui->white_pix;


	/* Begin loading cursors */
	c_fg.a = 0xff;
	c_fg.r = 0xff;
	c_fg.g = 0xff;
	c_fg.b = 0xff;
	c_bg.a = 0xff;
	c_bg.r = 0x00;
	c_bg.g = 0x00;
	c_bg.b = 0x00;

	/* Horizontal Split */
#ifdef XC_sb_h_double_arrow
	wglobal->h_split_wcur = WidgetCreateCursorFromID(
	    XC_sb_h_double_arrow
	);
#endif

	/* Vertical Split */
#ifdef XC_sb_v_double_arrow
	wglobal->v_split_wcur = WidgetCreateCursorFromID(
	    XC_sb_v_double_arrow
	);
#endif

	/* Text */
	wglobal->text_wcur = WidgetCreateCursorFromData(
	    text_cur_xpm,
	    8, 8,
	    &c_fg, &c_bg
	);

	/* No way */
	wglobal->no_way_wcur = WidgetCreateCursorFromData(
	    no_way_xpm,
	    0, 0,
	    &c_fg, &c_bg
	);


	/* Begin loading fonts */

	OSWLoadFont(
	    &wglobal->pbtn_font,
"-adobe-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	if(wglobal->pbtn_font == NULL)
	    OSWLoadFont(&wglobal->pbtn_font, "7x14");

	OSWLoadFont(
	    &wglobal->menu_font,
"-adobe-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	if(wglobal->menu_font == NULL)
	    OSWLoadFont(&wglobal->menu_font, "7x14");

	OSWLoadFont(
	    &wglobal->prompt_label_font,
"-adobe-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	if(wglobal->prompt_label_font == NULL)
	    OSWLoadFont(&wglobal->prompt_label_font, "7x14");

	OSWLoadFont(&wglobal->prompt_text_font, "7x14");

	OSWLoadFont(&wglobal->scale_bar_font, "6x12");

	OSWLoadFont(
	    &wglobal->std_font,
"-adobe-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	if(wglobal->std_font == NULL)
	    OSWLoadFont(&wglobal->std_font, "7x14");


	/* Begin loading images */

	/* Standard background */
	wglobal->std_bkg_pm = WidgetLoadPixmapFromTgaData(
	    bg_standard_tga, NULL
	);

	/* Standard icon */
	wglobal->std_icon_pm = WidgetLoadPixmapFromTgaData(
	    default_icon_tga, NULL
	);
 
	/* Standard background */
	wglobal->std_bkg_img = WidgetLoadImageFromTgaData(
	    bg_standard_tga
	);

	/* Button background images */
	if(!wglobal->force_mono)
	    wglobal->button_unarmed_img = WidgetLoadImageFromTgaData(
		button_unarmed_tga
	    );
	if(!wglobal->force_mono)
	    wglobal->button_armed_img = WidgetLoadImageFromTgaData(
		button_armed_tga
	    );
	if(!wglobal->force_mono)
	    wglobal->button_highlighted_img = WidgetLoadImageFromTgaData(
		button_highlighted_tga
	    );

	/* Scale Bar handles */
	wglobal->scalebar_h_img = WidgetLoadImageFromTgaData(
	    scalebar_h_tga
	);
	wglobal->scalebar_v_img = WidgetLoadImageFromTgaData(
	    scalebar_v_tga
	);

	/* Menu Bar */
	wglobal->menu_bar_bg_img = WidgetLoadImageFromTgaData(
	    menu_bar_bg_tga
	);

	/* Menu Item */
	wglobal->menu_item_img = WidgetLoadImageFromTgaData(
	    menu_item_tga
	);
	wglobal->menu_item_highlighted_img = WidgetLoadImageFromTgaData(
	    menu_item_highlighted_tga
	);

	/* Popup List Map icon */
	wglobal->pulist_map_pixmap = WidgetLoadPixmapFromTgaData(
	    pulist_map_tga, &wglobal->pulist_map_mask 
	);

	/* Parent Directory icon */
	wglobal->parent_dir_pixmap = WidgetLoadPixmapFromTgaData(
	    parent_dir_tga, &wglobal->parent_dir_mask
	);

	/* Home Dir icon */
	wglobal->home_dir_pixmap = WidgetLoadPixmapFromTgaData(
	    home_dir_tga, &wglobal->home_dir_mask
	);

	/* Refresh icon */
	wglobal->refresh_pixmap = WidgetLoadPixmapFromTgaData(
	    refresh_tga,
	    &wglobal->refresh_mask
	);

	/* Browse Files icon */
	wglobal->browse_files_pixmap = WidgetLoadPixmapFromTgaData(
	    browsefiles_tga,
	    &wglobal->browse_files_mask
	);

	/* 20x20 list icons */
	wglobal->icon_file_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_file_20x20_tga,
	    &wglobal->icon_file_20x20_mask
	);
	wglobal->icon_folder_closed_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_folder_closed_20x20_tga,
	    &wglobal->icon_folder_closed_20x20_mask
	);
	wglobal->icon_folder_opened_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_folder_opened_20x20_tga,
	    &wglobal->icon_folder_opened_20x20_mask
	);
	wglobal->icon_link_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_link_20x20_tga,
	    &wglobal->icon_link_20x20_mask
	);
	wglobal->icon_pipe_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_pipe_20x20_tga,
	    &wglobal->icon_pipe_20x20_mask
	);
	wglobal->icon_device_block_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_device_block_20x20_tga,
	    &wglobal->icon_device_block_20x20_mask
	);
	wglobal->icon_device_character_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_device_character_20x20_tga,
	    &wglobal->icon_device_character_20x20_mask
	);
	wglobal->icon_executable_20x20_pixmap = WidgetLoadPixmapFromTgaData(
	    icon_executable_20x20_tga, 
	    &wglobal->icon_executable_20x20_mask
	);

	/* Toggle button unarmed */
	if(!wglobal->force_mono)
	{
	    wglobal->check_button_unarmed_img = WidgetLoadImageFromTgaData(
		check_button_unarmed_tga
	    );
	    wglobal->check_button_armed_img = WidgetLoadImageFromTgaData(
		check_button_armed_tga
	    );
	    wglobal->check_button_highlighted_img = WidgetLoadImageFromTgaData(
		check_button_highlighted_tga
	    );
	}

	/* Tab normal */
	wglobal->tab_unselected_img = WidgetLoadImageFromTgaData(
	    tab_unselected_tga
	);
	/* Tab selected */
	wglobal->tab_selected_img = WidgetLoadImageFromTgaData(
	    tab_selected_tga
	);


	/* Tooltips */
	wglobal->tooltips = TOOLTIPS(calloc(
	    1, sizeof(tooltips_struct)
	));
	TooltipsInit(wglobal->tooltips);


	/* Mark as initialized */
	wglobal->is_init = True;


	return(0);
}

/*
 *	Destroys all global widget resources.
 */
void WidgetDestroyGlobals(void)
{
	int i, widgets_not_destroyed = 0;
	widget_register_struct *reg;
	widget_global_struct *wglobal = &widget_global;

	/* Queued resizes */
	for(i = 0; i < wglobal->nqueued_resizes; i++)
	    free(wglobal->queued_resize[i]);
	free(wglobal->queued_resize);
	wglobal->queued_resize = NULL;
	wglobal->nqueued_resizes = 0;

	/* Queued draws */
	for(i = 0; i < wglobal->nqueued_draws; i++)
	    free(wglobal->queued_draw[i]);
	free(wglobal->queued_draw);
	wglobal->queued_draw = NULL;
	wglobal->nqueued_draws = 0;

	/* Timeouts */
	for(i = 0; i < wglobal->ntimeouts; i++)
	    free(wglobal->timeout[i]);
	free(wglobal->timeout);
	wglobal->timeout = NULL;
	wglobal->ntimeouts = 0; 


	/* Tooltips */
	WidgetDestroy(wglobal->tooltips);
	free(wglobal->tooltips);
	wglobal->tooltips = NULL;


	/* Cursors */
	WidgetDestroyCursor(&wglobal->h_split_wcur);
	WidgetDestroyCursor(&wglobal->v_split_wcur);
	WidgetDestroyCursor(&wglobal->text_wcur);
	WidgetDestroyCursor(&wglobal->no_way_wcur);


	/* Background tile pixmaps */
	OSWDestroyPixmap(&wglobal->std_bkg_pm);
	OSWDestroyPixmap(&wglobal->std_icon_pm);


	/* Dialog, list, etc images */
	OSWDestroyImage(&wglobal->std_bkg_img);
	OSWDestroyImage(&wglobal->menu_bar_bg_img);
	OSWDestroyImage(&wglobal->menu_item_img);
	OSWDestroyImage(&wglobal->menu_item_highlighted_img);

	OSWDestroyImage(&wglobal->button_unarmed_img);
	OSWDestroyImage(&wglobal->button_armed_img);
	OSWDestroyImage(&wglobal->button_highlighted_img);

	OSWDestroyImage(&wglobal->scalebar_h_img);
	OSWDestroyImage(&wglobal->scalebar_v_img);

	OSWDestroyPixmap(&wglobal->pulist_map_pixmap);
	OSWDestroyBitmap(&wglobal->pulist_map_mask);

	OSWDestroyPixmap(&wglobal->parent_dir_pixmap);
	OSWDestroyBitmap(&wglobal->parent_dir_mask);

	OSWDestroyPixmap(&wglobal->home_dir_pixmap);
	OSWDestroyBitmap(&wglobal->home_dir_mask);

	OSWDestroyPixmap(&wglobal->refresh_pixmap);
	OSWDestroyBitmap(&wglobal->refresh_mask);

	OSWDestroyPixmap(&wglobal->browse_files_pixmap);
	OSWDestroyBitmap(&wglobal->browse_files_mask);

	OSWDestroyPixmap(&wglobal->icon_file_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_file_20x20_mask);
	OSWDestroyPixmap(&wglobal->icon_folder_closed_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_folder_closed_20x20_mask);
	OSWDestroyPixmap(&wglobal->icon_folder_opened_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_folder_opened_20x20_mask);
	OSWDestroyPixmap(&wglobal->icon_link_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_link_20x20_mask);
	OSWDestroyPixmap(&wglobal->icon_pipe_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_pipe_20x20_mask);
	OSWDestroyPixmap(&wglobal->icon_device_block_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_device_block_20x20_mask);
	OSWDestroyPixmap(&wglobal->icon_device_character_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_device_character_20x20_mask);
	OSWDestroyPixmap(&wglobal->icon_executable_20x20_pixmap);
	OSWDestroyBitmap(&wglobal->icon_executable_20x20_mask);

	OSWDestroyImage(&wglobal->check_button_unarmed_img);
	OSWDestroyImage(&wglobal->check_button_armed_img);
	OSWDestroyImage(&wglobal->check_button_highlighted_img);

	OSWDestroyImage(&wglobal->tab_unselected_img);
	OSWDestroyImage(&wglobal->tab_selected_img);


	/* Color pixels */
	OSWDestroyPixel(&wglobal->surface_normal_pix);
	OSWDestroyPixel(&wglobal->surface_editable_pix);
	OSWDestroyPixel(&wglobal->surface_selected_pix);
	OSWDestroyPixel(&wglobal->surface_shadow_pix);
	OSWDestroyPixel(&wglobal->surface_highlight_pix);

	OSWDestroyPixel(&wglobal->normal_text_pix);
	OSWDestroyPixel(&wglobal->editable_text_pix);
	OSWDestroyPixel(&wglobal->selected_text_pix);
	OSWDestroyPixel(&wglobal->disabled_text_pix);

	OSWDestroyPixel(&wglobal->scroll_bar_bg_pix);
	OSWDestroyPixel(&wglobal->scroll_bar_frame_pix);
	OSWDestroyPixel(&wglobal->scroll_bar_btn_pix);
	OSWDestroyPixel(&wglobal->scroll_bar_handle_pix);

	OSWDestroyPixel(&wglobal->pbar_text_pix);

	OSWDestroyPixel(&wglobal->hint_bkg_pix);
	OSWDestroyPixel(&wglobal->hint_text_pix);


	/* Fonts */
	OSWUnloadFont(&wglobal->pbtn_font);
	OSWUnloadFont(&wglobal->menu_font);
	OSWUnloadFont(&wglobal->prompt_label_font);
	OSWUnloadFont(&wglobal->prompt_text_font);
	OSWUnloadFont(&wglobal->scale_bar_font);
	OSWUnloadFont(&wglobal->std_font);


	/* Delete the widget registery list, warn if any widgets are
	 * still allocated and not destroyed
	 */
	for(i = 0; i < wglobal->nwidgets; i++)
	{
	    reg = wglobal->widget[i];
	    if(reg == NULL)
		continue;

	    if(reg->widget != NULL)
	    {
		void *widget = reg->widget;
		char *type_name = STRDUP(WidgetGetTypeName(widget));
		char *s = (char *)malloc(STRLEN(type_name) + 40 + 40 + 80);

		sprintf(
		    s,
"WidgetDestroyGlobals(): Warning: %s widget #%i %p was not destroyed.",
		    type_name, i, widget
		);
		fprintf(stderr, "%s\n", s);
		free(s);
		free(type_name);

		widgets_not_destroyed++;
	    }

	    free(reg);
	    wglobal->widget[i] = reg = NULL;
	}
	free(wglobal->widget);
	wglobal->widget = NULL;
	wglobal->nwidgets = 0;
	if(widgets_not_destroyed > 0)
	{
	    fprintf(
		stderr,
"WidgetDestroyGlobals(): %i registered widgets were not destroyed.\n\
Use WidgetDestroy() to destroy them next time.\n",
		widgets_not_destroyed
	    );
	}

	/* Mark as uninitizlied */
	wglobal->is_init = False;
}


/*
 *	Returns the number of pending GUI events.
 */
int WidgetGUIEventsPending(void)
{
	return(OSWEventsPending());
}

/*
 *	Manages all widget events.
 *
 *	Returns the number of events managed.
 */
int WidgetManageGUI(event_t *event)
{
	int events_handled = 0;
	widget_global_struct *wglobal = &widget_global;

	if(event == NULL)
	    return(events_handled);

	if(events_handled == 0)
	    events_handled += WidgetManage(wglobal->tooltips, event);

	return(events_handled);
}


/*
 *	Returns the number of pending queued actions and timeouts.
 */
int WidgetActionsPending(void)
{
	widget_global_struct *wglobal = &widget_global;
	return(
	    wglobal->nqueued_resizes +
	    wglobal->nqueued_draws +
	    wglobal->ntimeouts
	);
}

/*
 *	Processes all queued resizes, all queued draws, and all
 *	timeouts that are due.
 */
void WidgetMainIteration(void *data)
{
	int total_events_handled = 0;
	widget_global_struct *wglobal = &widget_global;


#define REALLOC_LIST(_list_,_total_,_type_) {	\
 if((_total_) > 0) {				\
  (_list_) = (_type_ **)realloc(		\
   (_list_),					\
   (_total_) * sizeof(_type_ *)			\
  );						\
  if((_list_) == NULL)				\
   (_total_) = 0;				\
 } else {					\
  (_total_) = 0;				\
  free(_list_);					\
  (_list_) = NULL;				\
 }						\
}

#define POP_ACTION(_list_,_total_,_n_)	{	\
 int _i;					\
 (_total_)--;					\
 for(_i = (_n_); _i < (_total_); _i++)		\
  (_list_)[_i] = (_list_)[_i + 1];		\
}
	/* Process all queued resize actions */
	if(wglobal->nqueued_resizes > 0)
	{
	    int events_handled = 0;
	    wqaction_struct *act;

	    while(wglobal->nqueued_resizes > 0)
	    {
		/* Get the action and remove its reference from the
		 * actions list
		 */
		act = wglobal->queued_resize[0];
		POP_ACTION(
		    wglobal->queued_resize,
		    wglobal->nqueued_resizes,
		    0
		);

		/* Call the action's callback and then delete the
		 * action
		 */
		if((act != NULL) ? (act->func != NULL) : False)
		    act->func(act->widget, act->data);
		free(act);

		events_handled++;
/* Possible bug, if a queued event queues another event for the same
 * function it may cause infinate recursion here, should we break out
 * of the loop after each processing of a queued event?
 */
	    }
	    if(events_handled > 0)
	    {
		REALLOC_LIST(
		    wglobal->queued_resize,
		    wglobal->nqueued_resizes,
		    wqaction_struct
		);
	    }

	    total_events_handled += events_handled;
	}

	/* Process all queued draw actions */
	if(wglobal->nqueued_draws > 0)
	{
	    int events_handled = 0;
	    wqaction_struct *act;

	    while(wglobal->nqueued_draws > 0)
	    {
                /* Get the action and remove its reference from the
                 * actions list 
                 */
                act = wglobal->queued_draw[0];
                POP_ACTION(
                    wglobal->queued_draw,
                    wglobal->nqueued_draws,
                    0
                );

                /* Call the action's callback and then delete the
                 * action
                 */
		if((act != NULL) ? (act->func != NULL) : False)
		    act->func(act->widget, act->data);
		free(act);

		events_handled++;
	    }
            if(events_handled > 0)
            {
                REALLOC_LIST(
                    wglobal->queued_draw,
                    wglobal->nqueued_draws,
                    wqaction_struct
                );
            }

            total_events_handled += events_handled; 
	}

	/* Process all due timeouts */
	if(wglobal->ntimeouts > 0)
	{
	    int i, events_handled = 0;
	    const time_t cur_ms = OSWTimeMS();
	    wqaction_struct *act;

	    for(i = 0; i < wglobal->ntimeouts; i++)
	    {
		act = wglobal->timeout[i];
		if(act == NULL)
		    continue;

		/* Is this timeout due to be called? */
		if(act->timeout_next <= cur_ms)
		{
		    /* Remove this from the list before calling the
		     * callback so that the callback does mess with
		     * this
		     */
		    POP_ACTION(
			wglobal->timeout,
			wglobal->ntimeouts,
			i
		    );

                    /* Call the action's callback and then delete the
                     * action
                     */
		    if(act->func != NULL)
			act->func(act->widget, act->data);
		    free(act);

		    events_handled++;
		    i--;
		    continue;
		}
	    }
	    if(events_handled > 0)
	    {
		REALLOC_LIST(
		    wglobal->timeout,
		    wglobal->ntimeouts,
		    wqaction_struct
		);
	    }

	    total_events_handled += events_handled;
	}
#undef POP_ACTION
#undef REALLOC_LIST
}


/*
 *	Adds the specified widget to the widget regeristry list.
 */
int WidgetRegister(void *widget, const int type)
{
	int i;
	widget_register_struct *reg;
	widget_global_struct *wglobal = &widget_global;

	i = MAX(wglobal->nwidgets, 0);
	wglobal->nwidgets = i + 1;
	wglobal->widget = (widget_register_struct **)realloc(
	    wglobal->widget,
	    wglobal->nwidgets * sizeof(widget_register_struct *)
	);
	if(wglobal->widget == NULL)
	{
	    wglobal->nwidgets = 0;
	    return(-3);
	}

	wglobal->widget[i] = reg = (widget_register_struct *)calloc(
	    1, sizeof(widget_register_struct)
	);
	if(reg == NULL)
	    return(-3);

	reg->widget = widget;
	reg->type = type;

	return(i);
}

/*
 *	Returns the widget's type.
 *
 *	If the widget is NULL then WIDGET_TYPE_UNKNOWN will be
 *	returned;
 */
int WidgetGetType(void *widget)
{
	return(
	    (widget != NULL) ?
		(*((int *)widget)) : WIDGET_TYPE_UNKNOWN
	);
}

/*
 *	Returns the type name of ths specified widget.
 *
 *	This function never returns NULL.
 *
 *	If the widget's type is unknown then "Unknown" is returned.
 */
const char *WidgetGetTypeName(void *widget)
{
	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CHECK_BUTTON:
	    return("CheckButton");
	    break;
	  case WIDGET_TYPE_CLIST:
	    return("CList");
	    break;
	  case WIDGET_TYPE_DIALOG:
	    return("Dialog");
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    return("FileBrowser");
	    break;
	  case WIDGET_TYPE_LABEL:
	    return("Label");
	    break;
	  case WIDGET_TYPE_LIST:
	    return("List");
	    break;
	  case WIDGET_TYPE_MENU:
	    return("Menu");
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    return("MenuBar");
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    return("PopupList");
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    return("ProgressBar");
	    break;
	  case WIDGET_TYPE_PROMPT:
	    return("Prompt");
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    return("PushButton");
	    break;
	  case WIDGET_TYPE_RADIO:
	    return("Radio");
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    return("ScaleBar");
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    return("Tooltips");
	    break;
	  case WIDGET_TYPE_VIEWER:
	    return("Viewer");
	    break;
	  default:
	    return("Unknown");
	    break;
	}
}

/*
 *	Checks if the specified widget is registered.
 */
bool_t WidgetIsRegistered(void *widget)
{
	int i;
	widget_register_struct *reg;
	widget_global_struct *wglobal = &widget_global;   

	if(widget == NULL)
	    return(False);

	for(i = 0; i < wglobal->nwidgets; i++)
	{
	    reg = wglobal->widget[i];
	    if(reg == NULL)
		continue;

	    if(reg->widget == widget)
		return(True);
	}

	return(False);
}

/*
 *	Removes the specified widget from the widget regeristry list.
 */
void WidgetUnregister(void *widget)
{
	int i;
	widget_register_struct *reg;
	widget_global_struct *wglobal = &widget_global;

	for(i = 0; i < wglobal->nwidgets; i++)
	{
	    reg = wglobal->widget[i];
	    if(reg == NULL)
		continue;

	    /* Is this the specified widget? */
	    if(reg->widget == widget)
	    {
		/* Unregister */
		int n;

		/* Remove this widget from the list */
		wglobal->nwidgets--;
		for(n = i; n < wglobal->nwidgets; n++)
		    wglobal->widget[n] = wglobal->widget[n + 1];

		/* Delete the regeristry for this widget */
		free(reg);

		i--;
		continue;
	    }
	}

	/* Reallocate list */
	if(wglobal->nwidgets > 0)
	{
	    wglobal->widget = (widget_register_struct **)realloc(
		wglobal->widget,
		wglobal->nwidgets * sizeof(widget_register_struct *)
	    );
	    if(wglobal->widget == NULL)
		wglobal->nwidgets = 0;
	}
	else
	{
	    wglobal->nwidgets = 0;
	    free(wglobal->widget);
	    wglobal->widget = NULL;
	}
}


/*
 *	Returns the toplevel window for the specified widget.
 */
win_t WidgetGetToplevel(void *widget)
{
	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CLIST:
	    return(((clist_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_DIALOG:
	    return(((dialog_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    return(((fb_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_LABEL:
	    return(((label_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_LIST:
	    return(((list_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_MENU:
	    return(((menu_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    return(((menu_bar_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    return(((popup_list_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    return(((progress_bar_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_PROMPT:
	    return(((prompt_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    return(((push_button_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    return(((scale_bar_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    /* Return the y scroll bar's toplevel */
	    return(((scroll_bar_struct *)widget)->y_toplevel);
	    break;
	  case WIDGET_TYPE_RADIO:
	    return(((radio_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_CHECK_BUTTON:
	    return(((check_button_struct *)widget)->toplevel);
	    break;
	  case WIDGET_TYPE_VIEWER:
	    return(((file_viewer_struct *)widget)->toplevel);
	    break;
	  default:
	    return(None);
	    break;
	}
}

/*
 *	Returns the widget's toplevel window parent.
 */
win_t WidgetGetParent(void *widget) 
{
	return(OSWGetWindowParent(
	    WidgetGetToplevel(widget)
	));
}


/*
 *	Returns the widget's focus state.
 */
bool_t WidgetHasFocus(void *widget)
{
	widget_struct *common = WIDGET(widget);
	return((common != NULL) ? common->has_focus : False);
}

/*
 *	Sets the widget's focus state.
 */
void WidgetSetFocus(void *widget, const bool_t focus)
{
	widget_struct *common = WIDGET(widget);
	if(common == NULL)
	    return;

	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CHECK_BUTTON:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_CLIST:
	    if(common->has_focus != focus)
	    {
		clist_struct *clist = CLIST(widget);

		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);

		WidgetSetFocus(clist->sb, False);
	    }
	    break;
	  case WIDGET_TYPE_DIALOG:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    if(common->has_focus != focus)
	    {
		fb_struct *fb = FB(widget);

		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);

		WidgetSetFocus(fb->location_pulist, False);
		WidgetSetFocus(fb->list_win_sb, False);
		WidgetSetFocus(fb->dir_win_sb, False);
		WidgetSetFocus(fb->file_win_sb, False);
		WidgetSetFocus(fb->prompt, False);
	    }
	    break;
	  case WIDGET_TYPE_LABEL:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_LIST:
	    if(common->has_focus != focus)
	    {
		list_struct *list = LIST(widget);

		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);

		WidgetSetFocus(list->sb, False);
	    }
	    break;
	  case WIDGET_TYPE_MENU:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    if(common->has_focus != focus)
	    {
		popup_list_struct *pulist = POPUP_LIST(widget);

		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);

		WidgetSetFocus(pulist->map_btn, False);
	    }
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_PROMPT:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_RADIO:
	    if(common->has_focus != focus)
	    {
		int i;
		radio_struct *radio = RADIO(widget);

		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);

		for(i = 0; i < radio->ncheck_buttons; i++)
		    WidgetSetFocus(radio->check_button[i], False);
	    }
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	  case WIDGET_TYPE_VIEWER:
	    if(common->has_focus != focus)
	    {
		common->has_focus = focus;
		if(common->map_state)
		    WidgetQueueDraw(widget);
	    }
	    break;
	}
}

/*
 *	Returns the widget's disabled state.
 */
bool_t WidgetDisabled(void *widget)
{
	widget_struct *common = WIDGET(widget);
	return((common != NULL) ? common->disabled : False);
}

/*
 *	Sets the widget's disabled state.
 */
void WidgetSetDisabled(void *widget, const bool_t disabled)
{
	widget_struct *common = WIDGET(widget);
	if(common == NULL)
	    return;

	if(common->disabled != disabled)
	{
	    common->disabled = disabled;
	    if(common->map_state)
		WidgetQueueDraw(widget);
	}
}


/*
 *	Queues the widget to resize.
 */
void WidgetQueueResize(void *widget)
{
	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CHECK_BUTTON:
	    CBtnQueueResize((check_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_CLIST:
	    CListQueueResize((clist_struct *)widget);
	    break;
	  case WIDGET_TYPE_DIALOG:
	    DialogQueueResize((dialog_struct *)widget);
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    FBrowserQueueResize((fb_struct *)widget);
	    break;
	  case WIDGET_TYPE_LABEL:
	    LabelQueueResize((label_struct *)widget);
	    break;
	  case WIDGET_TYPE_LIST:
	    ListWinQueueResize((list_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU:
	    MenuQueueResize((menu_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    MenuBarQueueResize((menu_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    PUListQueueResize((popup_list_struct *)widget, -1);
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    PBarQueueResize((progress_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROMPT:
	    PromptQueueResize((prompt_struct *)widget);
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    PBtnQueueResize((push_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_RADIO:
	    RadioQueueResize((radio_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
/*
	    ScaleBarQueueResize((scale_bar_struct *)widget);
 */
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    if(True)
	    {
		scroll_bar_struct *sb = (scroll_bar_struct *)widget;
		SBarQueueResize(
		    sb,
		    sb->width, sb->height,
		    sb->total_width, sb->total_height
		);
	    }
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    TooltipsQueueResize((tooltips_struct *)widget);
	    break;
	  case WIDGET_TYPE_VIEWER:
/*
	    ViewerQueueResize((file_viewer_struct *)widget);
 */
	    break;

	}
}

/*
 *	Queues the widget to draw.
 */
void WidgetQueueDraw(void *widget)
{
	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CHECK_BUTTON:
	    CBtnQueueDraw((check_button_struct *)widget, CBTN_DRAW_AMOUNT_COMPLETE);
	    break;
	  case WIDGET_TYPE_CLIST:
	    CListQueueDraw((clist_struct *)widget, CLIST_DRAW_AMOUNT_COMPLETE);
	    break;
	  case WIDGET_TYPE_DIALOG:
	    DialogQueueDraw((dialog_struct *)widget);
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    FBrowserQueueDraw((fb_struct *)widget, FB_DRAW_COMPLETE);
	    break;
	  case WIDGET_TYPE_LABEL:
	    LabelQueueDraw((label_struct *)widget);
	    break;
	  case WIDGET_TYPE_LIST:
	    ListWinQueueDraw((list_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU:
	    MenuQueueDraw((menu_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    MenuBarQueueDraw((menu_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    PUListQueueDraw((popup_list_struct *)widget, PULIST_DRAW_AMOUNT_COMPLETE);
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    PBarQueueDraw((progress_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROMPT:
	    PromptQueueDraw((prompt_struct *)widget, PROMPT_DRAW_AMOUNT_COMPLETE);
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    PBtnQueueDraw((push_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_RADIO:
	    RadioQueueDraw((radio_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    ScaleBarQueueDraw((scale_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    SBarQueueDraw((scroll_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    TooltipsQueueDraw((tooltips_struct *)widget);
	    break;
	  case WIDGET_TYPE_VIEWER:
/*
	    ViewerQueueDraw((file_viewer_struct *)widget);
 */
	    break;
	}
}

/*
 *	Manages the widget event.
 */
int WidgetManage(void *widget, event_t *event)
{
	int events_handled = 0;

	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CHECK_BUTTON:
	    events_handled = CBtnManage((check_button_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_CLIST:
	    events_handled = CListManage((clist_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_DIALOG:
	    events_handled = DialogManage((dialog_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    events_handled = FBrowserManage((fb_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_LABEL:
	    events_handled = LabelManage((label_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_LIST:
	    events_handled = ListWinManage((list_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_MENU:
	    events_handled = MenuManage((menu_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    events_handled = MenuBarManage((menu_bar_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    events_handled = PUListManage((popup_list_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    events_handled = PBarManage((progress_bar_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_PROMPT:
	    events_handled = PromptManage((prompt_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    events_handled = PBtnManage((push_button_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_RADIO:
	    events_handled = RadioManage((radio_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    events_handled = ScaleBarManage((scale_bar_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    events_handled = SBarManage((scroll_bar_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    events_handled = TooltipsManage((tooltips_struct *)widget, event);
	    break;
	  case WIDGET_TYPE_VIEWER:
	    events_handled = ViewerManage((file_viewer_struct *)widget, event);
	    break;
	}

	return(events_handled);
}

/*
 *	Returns the widget's map state.
 */
bool_t WidgetMapped(void *widget)
{
	widget_struct *common = WIDGET(widget);
	return((common != NULL) ? common->map_state : False);
}

/*
 *	Maps the widget.
 */
void WidgetMap(void *widget)
{
	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CHECK_BUTTON:
	    CBtnMap((check_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_CLIST:
	    CListMap((clist_struct *)widget);  
	    break;
	  case WIDGET_TYPE_DIALOG:
	    DialogMap((dialog_struct *)widget);
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    FBrowserMap((fb_struct *)widget);
	    break;
	  case WIDGET_TYPE_LABEL:
	    LabelMap((label_struct *)widget);
	    break;
	  case WIDGET_TYPE_LIST:
	    ListWinMap((list_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU:
	    MenuMap((menu_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    MenuBarMap((menu_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    PUListMap((popup_list_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    PBarMap((progress_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROMPT:
	    PromptMap((prompt_struct *)widget);
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    PBtnMap((push_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_RADIO:
	    RadioMap((radio_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    ScaleBarMap((scale_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    /* ScrollBar widget has no map function */
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    TooltipsMap((tooltips_struct *)widget);
	    break;
	  case WIDGET_TYPE_VIEWER:
	    ViewerMap((file_viewer_struct *)widget);
	    break;
	}
}

/*
 *	Unmaps the widget.
 */
void WidgetUnmap(void *widget)
{
	switch(WidgetGetType(widget))
	{
	  case WIDGET_TYPE_CHECK_BUTTON:
	    CBtnUnmap((check_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_CLIST:
	    CListUnmap((clist_struct *)widget);
	    break;
	  case WIDGET_TYPE_DIALOG:
	    DialogUnmap((dialog_struct *)widget);
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    FBrowserUnmap((fb_struct *)widget);
	    break;
	  case WIDGET_TYPE_LABEL:
	    LabelUnmap((label_struct *)widget);
	    break;            
	  case WIDGET_TYPE_LIST:
	    ListWinUnmap((list_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU:
	    MenuUnmap((menu_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    MenuBarUnmap((menu_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    PUListUnmap((popup_list_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    PBarUnmap((progress_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROMPT:
	    PromptUnmap((prompt_struct *)widget);
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    PBtnUnmap((push_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_RADIO:
	    RadioUnmap((radio_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    ScaleBarUnmap((scale_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    /* ScrollBar widget has no unmap function */
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    TooltipsUnmap((tooltips_struct *)widget);
	    break;     
	  case WIDGET_TYPE_VIEWER:
	    ViewerUnmap((file_viewer_struct *)widget);
	    break;
	}
}

/*
 *	Returns the widget's visibility state.
 */
visibility_t WidgetVisibility(void *widget)
{
	widget_struct *common = WIDGET(widget);
	return((common != NULL) ? common->visibility_state : VisibilityFullyObscured);
}

/*
 *	Destroys the widget.
 */
void WidgetDestroy(void *widget) 
{
	int type;
	widget_global_struct *wglobal = &widget_global;

	if(widget == NULL)
	    return;

	type = WidgetGetType(widget);
	if(type == WIDGET_TYPE_UNKNOWN)
	{
	    fprintf(
		stderr,
"WidgetDestroy(): Widget %p: Warning: Unable to get the widget's type.\n",
		widget
	    );
	}

	/* Remove all queued resizes and draws for this widget */
	if(wglobal->nqueued_resizes > 0)
	{
	    int i;
	    wqaction_struct *act;
	    for(i = 0; i < wglobal->nqueued_resizes; i++)
	    {
		act = wglobal->queued_resize[i];
		if((act != NULL) ? (act->widget == widget) : False)
		{
		    int n;
		    wglobal->nqueued_resizes--;
		    for(n = i; n < wglobal->nqueued_resizes; n++)
			wglobal->queued_resize[n] = wglobal->queued_resize[n + 1];
		    free(act);
		    i--;      
		    continue;
		}
	    }
	}
	if(wglobal->nqueued_draws > 0)
	{
	    int i;
	    wqaction_struct *act;
	    for(i = 0; i < wglobal->nqueued_draws; i++)
	    {
		act = wglobal->queued_draw[i];
		if((act != NULL) ? (act->widget == widget) : False)
		{
		    int n;
		    wglobal->nqueued_draws--;
		    for(n = i; n < wglobal->nqueued_draws; n++)
			wglobal->queued_draw[n] = wglobal->queued_draw[n + 1];
		    free(act);
		    i--;
		    continue;
		}
	    }
	}

	/* Remove all timeouts for this widget */
	WidgetTimeoutRemoveByWidget(widget);

	/* Unregister this widget */
	WidgetUnregister(widget);

	/* Destroy this widget */
	switch(type)
	{
	  case WIDGET_TYPE_UNKNOWN:
	    /* Already destroyed */
	    break;
	  case WIDGET_TYPE_CHECK_BUTTON:
	    CBtnDestroy((check_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_CLIST:
	    CListDestroy((clist_struct *)widget);  
	    break;
	  case WIDGET_TYPE_DIALOG:
	    DialogDestroy((dialog_struct *)widget);
	    break;
	  case WIDGET_TYPE_FILE_BROWSER:
	    FBrowserDestroy((fb_struct *)widget);
	    break;
	  case WIDGET_TYPE_LABEL:
	    LabelDestroy((label_struct *)widget);
	    break;
	  case WIDGET_TYPE_LIST:
	    ListWinDestroy((list_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU:
	    MenuDestroy((menu_struct *)widget);
	    break;
	  case WIDGET_TYPE_MENU_BAR:
	    MenuBarDestroy((menu_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_POPUP_LIST:
	    PUListDestroy((popup_list_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROGRESS_BAR:
	    PBarDestroy((progress_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_PROMPT:
	    PromptDestroy((prompt_struct *)widget);
	    break;
	  case WIDGET_TYPE_PUSH_BUTTON:
	    PBtnDestroy((push_button_struct *)widget);
	    break;
	  case WIDGET_TYPE_RADIO:
	    RadioDestroy((radio_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCALE_BAR:
	    ScaleBarDestroy((scale_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_SCROLL_BAR:
	    SBarDestroy((scroll_bar_struct *)widget);
	    break;
	  case WIDGET_TYPE_TOOLTIPS:
	    TooltipsDestroy((tooltips_struct *)widget);
	    break;     
	  case WIDGET_TYPE_VIEWER:
	    ViewerDestroy((file_viewer_struct *)widget);
	    break;
	  default:
	    fprintf(
		stderr,
"WidgetDestroy(): Widget %p: Warning: Unsupported widget type `%i'.\n",
		widget,
		type
	    );
	    break;
	}
}


/*
 *	Queues a resize.
 *
 *	If the specified values have already been queued then it will
 *	not be queued again.
 */
void WidgetQueueResizeAction(
	void (*func)(void *, void *),
	void *widget, void *data
)
{
	int i;
	wqaction_struct *act;
	widget_global_struct *wglobal = &widget_global;

	if(func == NULL)
	    return;

	/* Check if this resize has already been queued */
	for(i = 0; i < wglobal->nqueued_resizes; i++)
	{
	    act = wglobal->queued_resize[i];
	    if(act == NULL)
		continue;

	    if((act->func == func) &&
	       (act->widget == widget) &&
	       (act->data == data)
	    )
	    {
		/* Swap it to the end */
		const int m = wglobal->nqueued_resizes - 1;
		int n;
		for(n = i; n < m; n++)
		    wglobal->queued_resize[n] = wglobal->queued_resize[n + 1];
		wglobal->queued_resize[m] = act;
		return;
	    }
	}

	/* Queue this resize */
	i = wglobal->nqueued_resizes;
	wglobal->nqueued_resizes = i + 1;
	wglobal->queued_resize = (wqaction_struct **)realloc(
	    wglobal->queued_resize,
	    wglobal->nqueued_resizes * sizeof(wqaction_struct *)
	);
	if(wglobal->queued_resize == NULL)
	{
	    wglobal->nqueued_resizes = 0;
	    return;
	}
	wglobal->queued_resize[i] = act = (wqaction_struct *)calloc(
	    1, sizeof(wqaction_struct)
	);
	if(act != NULL)
	{
	    act->func = func;
	    act->widget = widget;
	    act->data = data;
	}
}

/*
 *	Queues a draw.
 *
 *	If the specified values have already been queued then it will
 *	not be queued again.
 */
void WidgetQueueDrawAction(
	void (*func)(void *, void *),
	void *widget, void *data
)
{
	int i;
	wqaction_struct *act;  
	widget_global_struct *wglobal = &widget_global;

	if(func == NULL)
	    return;

	/* Check if this draw has already been queued */
	for(i = 0; i < wglobal->nqueued_draws; i++)
	{                                 
	    act = wglobal->queued_draw[i];
	    if(act == NULL)
		continue;

	    if((act->func == func) && 
	       (act->widget == widget) &&
	       (act->data == data)
	    )
	    {
		/* Swap it to the end */
		const int m = wglobal->nqueued_draws - 1;
		int n;
		for(n = i; n < m; n++)
		    wglobal->queued_draw[n] = wglobal->queued_draw[n + 1];
		wglobal->queued_draw[m] = act;
		return;
	    }
	}

	/* Queue this draw */
	i = wglobal->nqueued_draws;
	wglobal->nqueued_draws = i + 1;
	wglobal->queued_draw = (wqaction_struct **)realloc(
	    wglobal->queued_draw,
	    wglobal->nqueued_draws * sizeof(wqaction_struct *)
	);
	if(wglobal->queued_draw == NULL)
	{
	    wglobal->nqueued_draws = 0;
	    return;
	}
	wglobal->queued_draw[i] = act = (wqaction_struct *)calloc(
	    1, sizeof(wqaction_struct)
	);
	if(act != NULL)
	{
	    act->func = func;
	    act->widget = widget;
	    act->data = data;
	}
}

/*
 *	Adds a timeout.
 *
 *	The interval_ms specifies the minimum amount of time to wait
 *	in milliseconds before processing this timeout. If interval_ms
 *	is 0 then this timeout will be processed on the very next call
 *	to WidgetMainIteration().
 */
void WidgetTimeoutAdd(
	void (*func)(void *, void *),
	void *widget, void *data,
	const time_t interval_ms
)
{
	int i;
	wqaction_struct *act;
	widget_global_struct *wglobal = &widget_global;

	if(func == NULL)
	    return;

	/* Queue this draw */
	i = wglobal->ntimeouts;
	wglobal->ntimeouts = i + 1;
	wglobal->timeout = (wqaction_struct **)realloc(
	    wglobal->timeout,
	    wglobal->ntimeouts * sizeof(wqaction_struct *)
	);
	if(wglobal->timeout == NULL)
	{
	    wglobal->ntimeouts = 0;
	    return;
	}
	wglobal->timeout[i] = act = (wqaction_struct *)calloc(
	    1, sizeof(wqaction_struct)
	);
	if(act != NULL)  
	{
	    act->func = func;
	    act->widget = widget;
	    act->data = data;
	    act->timeout_interval = interval_ms;
	    act->timeout_next = OSWTimeMS() + interval_ms;
	}
}

/*
 *	Same as WidgetTimeoutAdd() except that the inverval is set to
 *	0.
 */
void WidgetIdleAdd(
        void (*func)(void *, void *),
        void *widget, void *data 
)
{
	WidgetTimeoutAdd(func, widget, data, 0l);
}

/*
 *	Removes all timeouts that refer to the specified function.
 */
void WidgetTimeoutRemoveByFunc(void (*func)(void *, void *))
{
	int i;
	wqaction_struct *act;
	widget_global_struct *wglobal = &widget_global;

	for(i = 0; i < wglobal->ntimeouts; i++)
	{
	    act = wglobal->timeout[i];
	    if(act == NULL) 
		continue;

	    /* This timeout's function matches the specified function? */
	    if(act->func == func)
	    {
		/* Remove this timeout */
		int n;
		free(act);
		wglobal->ntimeouts--;
		for(n = i; n < wglobal->ntimeouts; n++)
		    wglobal->timeout[n] = wglobal->timeout[n + 1];
		i--;
		continue;
	    }
	}
}

/*
 *	Removes all timeouts that have the specified widget.
 */
void WidgetTimeoutRemoveByWidget(void *widget)
{
	int i;
	wqaction_struct *act;
	widget_global_struct *wglobal = &widget_global;

	for(i = 0; i < wglobal->ntimeouts; i++)
	{
	    act = wglobal->timeout[i];
	    if(act == NULL)
		continue;

	    /* This timeout's widget matches the specified widget? */
	    if(act->widget == widget)
	    {
		/* Remove this timeout */
		int n;
		free(act);
		wglobal->ntimeouts--;
		for(n = i; n < wglobal->ntimeouts; n++)
		    wglobal->timeout[n] = wglobal->timeout[n + 1];
		i--;
		continue;
	    }
	}
}

/*
 *	Removes all timeouts that refer to the specified function
 *	and widget.
 */
void WidgetTimeoutRemoveByFuncWidget(
	void (*func)(void *, void *),
	void *widget
)
{
	int i;
	wqaction_struct *act;
	widget_global_struct *wglobal = &widget_global;

	for(i = 0; i < wglobal->ntimeouts; i++)
	{
	    act = wglobal->timeout[i];
	    if(act == NULL)
		continue;

	    /* This timeout's function and widget matches the
	     * specified function and widget?
	     */
	    if((act->func == func) && (act->widget == widget))
	    {
		/* Remove this timeout */
		int n;
		free(act);
		wglobal->ntimeouts--;
		for(n = i; n < wglobal->ntimeouts; n++)
		    wglobal->timeout[n] = wglobal->timeout[n + 1];
		i--;
		continue;
	    }
	}
}
