#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static int CmdBootConnectionNumber(
	const int con_num, const int tar_con_num, const char force,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int CmdBootPlayer(
	const int con_num, const char *player_name, const char force,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

int CmdBoot(const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc);


#define THIS_CMD_NAME		"boot"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static int CmdBootConnectionNumber(
	const int con_num, const int tar_con_num, const char force,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	sws_connection_struct *tar_con =
	    ((tar_con_num >= 0) && (tar_con_num < total_connections)) ?
		connection[tar_con_num] : NULL;
	int tar_obj_num;
	char *sndbuf, *name_booter;

	if(tar_con == NULL)
	{
	    sndbuf = (char *)malloc(40 + 80);
	    sprintf(
		sndbuf,
"Connection %i: No such connection.",
		tar_con_num + 1
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	tar_obj_num = tar_con->object_num;

	/* Get formal names of objects */
	name_booter = STRDUP(DBGetFormalNameStr(con_obj_num));

	/* Notify connection */
	if(DBIsObjectGarbage(tar_obj_num))
	{
	    sndbuf = (char *)malloc(40 + 80);
	    sprintf(
		sndbuf,
"Booted connection %i.",
		tar_con_num + 1
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}
	else
	{
	    char *name_booted = STRDUP(DBGetFormalNameStr(tar_obj_num));
	    sndbuf = (char *)malloc(STRLEN(name_booted) + 40 + 80);
	    sprintf(
		sndbuf,
"Booted %s on connection %i.",
		name_booted, tar_con_num + 1
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name_booted);
	}

	/* Log */
	if(sysparm.log_general)   
	{
	    if(DBIsObjectGarbage(tar_obj_num))
	    {
		char *s = (char *)malloc(                                   
		    STRLEN(name_booter) + 40 + 80
		);
		sprintf(
		    s,
"%s: Booted: Connection %i.",
		    name_booter, tar_con_num
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	    else
	    {
		char *name_booted = STRDUP(DBGetFormalNameStr(tar_obj_num));
		char *s = (char *)malloc(
		    STRLEN(name_booter) + STRLEN(name_booted) + 40 + 80
		);
		sprintf(
		    s,
"%s: Booted: `%s' on connection %i.",
		    name_booter, name_booted, tar_con_num
		);
		LogAppend(fname.log, s);
		free(s);
		free(name_booted);
	    }
	}

	/* Boot connection
	 *
	 * Note that if the connection that was booted was this
	 * connection then this connection should be considered
	 * invalid and not referenced again
	 */
	if(force)
	    NetCloseConnection(
		tar_con_num,
		cur_ms, cur_sec, tc
	    );
	else
	    NetShutdownConnection(
		tar_con_num, SWS_CON_SHUTDOWN_WR, 1,
		cur_ms, cur_sec, tc
	    );

	free(name_booter);

	return(0);
}


static int CmdBootPlayer(
	const int con_num, const char *player_name, const char force,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	int obj_num, tar_con_num;
	char *sndbuf, *name_booter, *name_booted;
	xsw_object_struct *obj, *con_obj = xsw_object[con_obj_num];

	/* Match the player object */
	if(!strcasecmp(player_name, "me"))
	{
	    obj_num = con_obj_num;
	}
	else if(!strcasecmp(player_name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else
	{   
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		player_name, SW_OBJ_TYPE_PLAYER
	    );
	}     

	/* No such player? */
	if(DBIsObjectGarbage(obj_num))
	{
	    sndbuf = (char *)malloc(STRLEN(player_name) + 80);
	    sprintf(
		sndbuf,
"%s: No such player.",
		player_name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	obj = xsw_object[obj_num];

	/* Is that player not connected? */
	tar_con_num = ConGetByObject(obj_num);
	if(tar_con_num < 0)
	{
	    sndbuf = (char *)malloc(STRLEN(obj->name) + 80);
	    sprintf(
		sndbuf,
"%s: Not connected.",
		obj->name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Get formal names of objects */
	name_booter = STRDUP(DBGetFormalNameStr(con_obj_num));
	name_booted = STRDUP(DBGetFormalNameStr(obj_num));

	/* Notify connection */
	sndbuf = (char *)malloc(STRLEN(name_booted) + 40 + 80);
	sprintf(
	    sndbuf,
"Booted %s on connection %i.",
	    name_booted, tar_con_num + 1
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);

	/* Log */
	if(sysparm.log_general)
	{
	    char *s = (char *)malloc(
		STRLEN(name_booter) + STRLEN(name_booted) + 40 + 80
	    );
	    sprintf(
		s,
"%s: Booted: `%s' on connection %i.",
		name_booter, name_booted, tar_con_num + 1
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	/* Boot connection
	 *
	 * Note that if the connection that was booted was this
	 * connection then this connection should be considered
	 * invalid and not referenced again
	 */
	if(force)
	    NetCloseConnection(
		tar_con_num,
		cur_ms, cur_sec, tc
	    );
	else
	    NetShutdownConnection(
		tar_con_num, SWS_CON_SHUTDOWN_WR, 1,
		cur_ms, cur_sec, tc
	    );

	/* Are we still connected and logged in? */
	if(ConIsLoggedIn(con_num))
	{
	    /* Warn if there are subsequent connections */
	    const int tar_con_num = ConGetByObject(obj_num);
	    if((tar_con_num > -1) && !DBIsObjectGarbage(obj_num))
	    {
		obj = xsw_object[obj_num];
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(obj->name) + 80
		);
		sprintf(
		    sndbuf,
"%s: Warning: %s still has additional connections.",
		    THIS_CMD_NAME, obj->name
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	}   

	free(name_booter);
	free(name_booted);

	return(0);
}

/*
 *      Boot off a player or connection.
 */
int CmdBoot(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int status;
	char force = 0;
	char *sndbuf;
	const char *s;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Print usage? */ 
	if(STRISEMPTY(arg))
	{
	    sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		sndbuf,
"Usage: `%s [--force] [player|me|this] [con_num]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(0);
	}

	/* Get the arguments */

	s = arg;
	while(ISBLANK(*s))
	    s++;

	while(*s == '-')
	{
	    if(strcasepfx(s, "--force") ||
	       strcasepfx(s, "-force") ||
	       strcasepfx(s, "--f") ||
	       strcasepfx(s, "-f")
	    )
		force = 1;

	    while((*s != '\0') && !ISBLANK(*s))
		s++;
	    while(ISBLANK(*s))
		s++;
	}

	/* Permission check */
	if(con_obj->permission.uid > SWS_ACCESS_UID_BOOT)
	{
	    sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_BOOT
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Get the name or connection number */
	if(isdigit(*s))
	{
	    const int tar_con_num = atoi(s) - 1;
	    status = CmdBootConnectionNumber(
		con_num, tar_con_num, force,
		cur_ms, cur_sec, tc
	    );
	}
	else
	{
	    char *player_name = STRDUP(s);
	    status = CmdBootPlayer(
		con_num, player_name, force,
		cur_ms, cur_sec, tc
	    );
	    free(player_name);
	}

	return(status);
}
