#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdChown(const int con_num, const char *arg);


#define THIS_CMD_NAME	"chown"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdChown(const int con_num, const char *arg)
{
	int tar_obj_num, new_owner_obj_num;
	const char *s;
	char *tar_obj_name, *new_owner_obj_name, *con_obj_name;
	xsw_object_struct *tar_obj, *new_owner_obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <object|me|this>[=<owner|me|this>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Parse the target object name */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    tar_obj_name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(tar_obj_name, s, len);
	    tar_obj_name[len] = '\0';
	    strstrip(tar_obj_name);

	    s = (s_end != NULL) ? s_end + 1 : NULL;
	}
	else
	{
	    tar_obj_name = NULL;
	}

	/* Parse the new owner object name */
	if(s != NULL)
	{
	    const int len = STRLEN(s);
	    new_owner_obj_name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(new_owner_obj_name, s, len);
	    new_owner_obj_name[len] = '\0';
	    strstrip(new_owner_obj_name);

	    s = NULL;
	}
	else
	{
	    new_owner_obj_name = NULL;
	}

	if(STRISEMPTY(tar_obj_name) || STRISEMPTY(new_owner_obj_name))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <object|me|this>[=<owner|me|this>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(-2);
	}

	/* Match the target object */
	if(!strcasecmp(tar_obj_name, "me"))
	{
	    tar_obj_num = con_obj_num;
	}
	else if(!strcasecmp(tar_obj_name, "this"))
	{
	    tar_obj_num = con_obj->locked_on;
	}
	else
	{
	    tar_obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		tar_obj_name, -1
	    );
	}
	if(DBIsObjectGarbage(tar_obj_num))
	{
	    char *s = (char *)malloc(STRLEN(tar_obj_name) + 80);
	    sprintf(
		s,
"%s: No such object.",
		tar_obj_name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(-1);
	}

	tar_obj = xsw_object[tar_obj_num];

	/* Players may not be chowned */
	if(tar_obj->type == SW_OBJ_TYPE_PLAYER)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Permission denied: Players always own themselves.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(-1);
	}


	/* Match the new owner object */
	if(!strcasecmp(new_owner_obj_name, "me"))
	{
	    new_owner_obj_num = con_obj_num;
	}
	else if(!strcasecmp(new_owner_obj_name, "this"))
	{
	    new_owner_obj_num = con_obj->locked_on;
	}
	else
	{
	    new_owner_obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		new_owner_obj_name, SW_OBJ_TYPE_PLAYER
	    );
	}
	if(DBIsObjectGarbage(new_owner_obj_num))
	{
	    char *s = (char *)malloc(STRLEN(new_owner_obj_name) + 80);
	    sprintf(
		s,
"%s: No such player.",
		new_owner_obj_name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(-1);
	}

	new_owner_obj = xsw_object[new_owner_obj_num];

	/* Not a player object? */
	if(new_owner_obj->type != SW_OBJ_TYPE_PLAYER)
	{
	    char *s = (char *)malloc(STRLEN(new_owner_obj_name) + 80);
	    sprintf(
		s,
"%s: Not a player.",
		new_owner_obj_name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(-1);
	}

	/* Check if the target object's current owner is the same as
	 * the new owner
	 */
	if(tar_obj->owner == new_owner_obj_num)
	{
	    char	*name1 = STRDUP(DBGetFormalNameStr(tar_obj_num)),
			*name2 = STRDUP(DBGetFormalNameStr(new_owner_obj_num)),
			*s = (char *)malloc(
		STRLEN(name1) + STRLEN(name2) + 80
	    );
	    sprintf(
		s,
"%s: Already owned by `%s'.",
		name1, name2
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name1);
	    free(name2);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(0);
	}


	/* Check permissions, allowed to chown? */
	if(con_obj->permission.uid > SWS_ACCESS_UID_CHOWN)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_CHOWN
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(-1); 
	}

	/* If the target object is owned by someone else, then check
	 * if the connection object is allowed to chown objects
	 * belonging to someone else
	 */
	if((con_obj->permission.uid > SWS_ACCESS_UID_CHOWNO) &&
	   (tar_obj->owner > -1) &&
	   (tar_obj->owner != con_obj_num)
	)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_CHOWNO
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(tar_obj_name);
	    free(new_owner_obj_name);
	    return(-1);
	}


	/* Chown the object */
	tar_obj->owner = new_owner_obj_num;


	/* Get formal names of the objects */
	con_obj_name = STRDUP(DBGetFormalNameStr(con_obj_num));
	free(tar_obj_name);
	tar_obj_name = STRDUP(DBGetFormalNameStr(tar_obj_num));
	free(new_owner_obj_name);
	new_owner_obj_name = STRDUP(DBGetFormalNameStr(new_owner_obj_num));

	/* Log and print response */
	if(1)
	{
	    char *s = (char *)malloc(
		STRLEN(tar_obj_name) +
		STRLEN(new_owner_obj_name) +
		80
	    );
	    sprintf(
		s,
"Chowned `%s' to `%s'.",
		tar_obj_name,
		new_owner_obj_name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	}

	if(sysparm.log_general)
	{
	    char *s = (char *)malloc(
		STRLEN(con_obj_name) +
		STRLEN(tar_obj_name) +
		STRLEN(new_owner_obj_name) +
		80
	    );
	    sprintf(
		s,
"%s: Chowned `%s' to `%s'.",
		con_obj_name, tar_obj_name,
		new_owner_obj_name
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	free(con_obj_name);
	free(tar_obj_name);
	free(new_owner_obj_name);

	return(0);
}
