#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/imgrefs.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdCreate(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME           "create"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdCreate(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int type, new_obj_num;
	const char *s;
	char	*name, *type_str,
		*name1, *name2;
	xsw_object_struct *new_obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <name>[=<type>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Check if object's permission allows create object */
	if(con_obj->permission.uid > SWS_ACCESS_UID_CREATE)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_CREATE
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Parse the name */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(name, s, len);
	    name[len] = '\0';
	    strstrip(name);

	    s = (s_end != NULL) ? s_end + 1 : NULL;
	}
	else
	{
	    name = NULL;
	}

	/* Parse the type */
	if(s != NULL)
	{
	    const int len = STRLEN(s);
	    type_str = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(type_str, s, len);
	    type_str[len] = '\0';
	    strstrip(type_str);

	    s = NULL;
	}
	else
	{
	    type_str = NULL;
	}
	if(STRISEMPTY(type_str))
	{
	    free(type_str);
	    type_str = STRDUP(SW_OBJ_TYPE_NAME_STATIC);
	}

	if(STRISEMPTY(name))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <name>[=<type>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(type_str);
	    return(-1);
	}

	/* Is the specified name invalid? */
	if(DBValidateObjectName(name)) 
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: Invalid name.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(type_str);
	    return(-2);
	}

	/* Get the type */
	if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_STATIC))
	    type = SW_OBJ_TYPE_STATIC;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_DYNAMIC))
	    type = SW_OBJ_TYPE_DYNAMIC;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_CONTROLLED))
	    type = SW_OBJ_TYPE_CONTROLLED;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_PLAYER))
	    type = SW_OBJ_TYPE_PLAYER;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_WEAPON))
	    type = SW_OBJ_TYPE_WEAPON;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_STREAMWEAPON))
	    type = SW_OBJ_TYPE_STREAMWEAPON;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_SPHEREWEAPON))
	    type = SW_OBJ_TYPE_SPHEREWEAPON;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_HOME))
	    type = SW_OBJ_TYPE_HOME;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_AREA))
	    type = SW_OBJ_TYPE_AREA;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_ANIMATED))
	    type = SW_OBJ_TYPE_ANIMATED;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_WORMHOLE))
	    type = SW_OBJ_TYPE_WORMHOLE;
	else if(!strcasecmp(type_str, SW_OBJ_TYPE_NAME_ELINK))
	    type = SW_OBJ_TYPE_ELINK;
	else
	    type = SW_OBJ_TYPE_GARBAGE;

	/* Invalid type? */
	if(type <= SW_OBJ_TYPE_GARBAGE)
	{
	    char *s = (char *)malloc(STRLEN(type_str) + 80);
	    sprintf(
		s,
"%s: Invalid type.",
		type_str
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(type_str);
	    return(-2);
	}

	/* Trying to create a player? */
	if(type == SW_OBJ_TYPE_PLAYER)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Unable to create players, use `createplayer' instead.",
 		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(type_str);
	    return(-2);
	}

	/* Create the new object */
	new_obj_num = DBCreateObject(
	    IMGREF_DEFAULT_IMGREF,	/* ImgRef */
	    type,			/* Type */
	    con_obj_num,		/* Owner */
	    con_obj->sect_x, con_obj->sect_y, con_obj->sect_z,
	    con_obj->x, con_obj->y, con_obj->z,
	    con_obj->heading, con_obj->pitch, con_obj->bank,
	    cur_ms, cur_sec, tc
	);
	if(DBIsObjectGarbage(new_obj_num))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: DBCreateObject(): Unable to create object.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(type_str);
	    return(-1);
	}

	new_obj = xsw_object[new_obj_num];

	/* Set the name of the new object */
	strncpy(new_obj->name, name, sizeof(new_obj->name));
	new_obj->name[sizeof(new_obj->name) - 1] = '\0';

	/* Notify all connections about the new object */
	NetSendCreateObject(-1, new_obj_num);
	NetSendObjectName(-1, new_obj_num);

	/* Print and log */
	name1 = STRDUP(DBGetFormalNameStr(new_obj_num));
	name2 = STRDUP(DBGetFormalNameStr(con_obj_num));

	if(1)
	{
	    char *s = (char *)malloc(STRLEN(name1) + 80);
	    sprintf(
		s,
"Created `%s'.",
		name1
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	}

	if(sysparm.log_events)
	{
	    char *s = (char *)malloc(STRLEN(name2) + STRLEN(name1) + 80);
	    sprintf(
		s,
"%s: Created `%s'.",
		name2,
		name1
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	free(name1);
	free(name2);

	free(name);
	free(type_str);

	return(0);
}
