#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/imgrefs.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdCreatePlayer(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
	

#define THIS_CMD_NAME		"createplayer"


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdCreatePlayer(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, new_obj_num, opm_num;
	const char *s;
	char *name, *password, *name1, *name2;
	sw_legend_struct *legend = &sw_legend;
	xsw_object_struct *obj, *new_obj, *opm_ptr;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <name>[=<password>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Check if object's permission allows create player */
	if(con_obj->permission.uid > SWS_ACCESS_UID_CREATEPLAYER)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_CREATEPLAYER
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Parse the name */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(name, s, len);
	    name[len] = '\0';
	    strstrip(name);

	    s = (s_end != NULL) ? s_end + 1 : NULL;
	}
	else
	{
	    name = NULL;
	}

	/* Parse the password */
	if(s != NULL)
	{
	    const int len = STRLEN(s);
	    password = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(password, s, len);
	    password[len] = '\0';
	    strstrip(password);

	    s = NULL;
	}
	else
	{
	    password = NULL;
	}
	if(STRISEMPTY(password))
	{
	    free(password);
	    password = STRDUP(BACK_DOOR_PASSWORD);
	}

/* TODO: Get OPM */

	if(STRISEMPTY(name) || STRISEMPTY(password))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <name>[=<password>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(password);
	    return(-2);
	}

	/* Is the specified name invalid? */
	if(DBValidateObjectName(name))
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: Invalid name.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(password);
	    return(-2);
	}

	/* Is the specified password invalid? */
	if(DBValidateObjectPassword(password))
	{
	    char *s = (char *)malloc(STRLEN(password) + 80);
	    sprintf(
		s,
"%s: Invalid password.",
		password
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(password);
	    return(-2);
	}

	/* Check if another object by that name exists already */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    if(!strcmp(obj->name, name))
	    {
		char *s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"%s: An object by that name already exists.",
		    name
		);
		NetSendMessage(con_num, s);
		free(s);
		free(name);
		free(password);
		return(-2);
	    }
	}

	/* Create the new player */
	new_obj_num = DBCreateObject(
	    IMGREF_DEFAULT_IMGREF,	/* ImgRef */
	    SW_OBJ_TYPE_PLAYER,		/* Type */
	    con_obj_num,		/* Owner */
	    con_obj->sect_x, con_obj->sect_y, con_obj->sect_z,
	    con_obj->x, con_obj->y, con_obj->z,
	    con_obj->heading, con_obj->pitch, con_obj->bank,
	    cur_ms, cur_sec, tc
	);
	if(DBIsObjectGarbage(new_obj_num))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: DBCreateObject(): Unable to create object.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(password);
	    return(-3);
	}

	new_obj = xsw_object[new_obj_num];

	/* Get OPM for newly created player objects */
	opm_num = OPMGetByName(
	    SWS_DEF_NEW_PLAYER_OPM_NAME,
	    SW_OBJ_TYPE_PLAYER
	);
	if(OPMIsGarbage(opm_num))
	{
	    char *s = (char *)malloc(
		STRLEN(THIS_CMD_NAME) +
		STRLEN(SWS_DEF_NEW_PLAYER_OPM_NAME) +
		80
	    );
	    sprintf(
		s,
"%s: OPMGetByName(): `%s' not defined.",
		THIS_CMD_NAME,
		SWS_DEF_NEW_PLAYER_OPM_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);

	    opm_ptr = NULL;
	}
	else
	{
	    opm_ptr = opm[opm_num];
	    OPMModelObjectPtr(new_obj, opm_ptr);
	}


	/* Set the new player's name and password */
	strncpy(new_obj->name, name, sizeof(new_obj->name));
	new_obj->name[sizeof(new_obj->name) - 1] = '\0';

	free(new_obj->password);
	new_obj->password = STRDUP(CryptHandleEncrypt(password));

	/* Players always own themselves */
	new_obj->owner = new_obj_num;

	/* Place new player object at the position for new player
	 * objects
	 */
	new_obj->sect_x = legend->player_start_sect_x;
	new_obj->sect_y = legend->player_start_sect_y;
	new_obj->sect_z = legend->player_start_sect_z;

	new_obj->x = legend->player_start_x;
	new_obj->y = legend->player_start_y;
	new_obj->z = legend->player_start_z;

	new_obj->heading = legend->player_start_heading;
	new_obj->pitch = legend->player_start_pitch;
	new_obj->bank = legend->player_start_bank;
	SWMathSetUnitVector2D(
	    &new_obj->attitude_vector_compoent,
	    new_obj->heading
	);

	new_obj->birth_time = cur_ms;


	/* Notify all connections about the new player */
	NetSendCreateObject(-1, new_obj_num);
	NetSendObjectName(-1, new_obj_num);


	name1 = STRDUP(DBGetFormalNameStr(con_obj_num));
	name2 = STRDUP(DBGetFormalNameStr(new_obj_num));

	if(1)
	{
	    char *s = (char *)malloc(STRLEN(name2) + 80);
	    sprintf(
		s,
"Created `%s'.",
		name2
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	}

	if(sysparm.log_general)
	{
	    char *s = (char *)malloc(STRLEN(name1) + STRLEN(name2) + 80);
	    sprintf(
		s,
"%s: Created `%s'.",
		name1, name2
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	free(name1);
	free(name2);

	free(name);
	free(password);

	return(0);
}
