#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdETA(const int con_num, const char *arg);


#define THIS_CMD_NAME	"eta"  

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdETA(const int con_num, const char *arg)
{
	int dest_obj_num;
	float distance, velocity, eta;
	xsw_object_struct *dest_obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Get the destination object */
	dest_obj_num = con_obj->intercepting_object;
	if(DBIsObjectGarbage(dest_obj_num))
	    return(0);

	dest_obj = xsw_object[dest_obj_num];

	/* Get the velocity */
	velocity = con_obj->velocity;

	/* The velocity must be non-zero */
	if(velocity <= 0.0f)
	    return(0);

	/* Check if in same sector */   
	if(!SWMath3DInSameSectorPtr(
	    con_obj, dest_obj
	))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Destination not in current sector.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(0);
	}

	/* Calculate the distance between the objects in real units */
	distance = SWMath3DDistance(
	    dest_obj->x - con_obj->x,
	    dest_obj->y - con_obj->y,
	    dest_obj->z - con_obj->z
	);

	/* Calculate the eta in units of cycles */
	eta = distance / velocity;

	/* Convert the eta from cycles to milliseconds */
	eta = eta * 1000.0f / (float)SW_CYCLE_LAPSE_MS;

	/* Convert the eta from milliseconds to seconds */
	eta = eta / 1000.0f;

	if(eta > 0.0f)
	{
	    const char *time_str = StringFormatTimePeriod((time_t)eta);
	    char *s = (char *)malloc(
		STRLEN(dest_obj->name) + STRLEN(time_str) + 80
	    );
	    sprintf(
		s,
"ETA to `%s' in %s",
		dest_obj->name, time_str
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	}

	return(0);
}
