#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdExamine(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"examine"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdExamine(
        const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, obj_num, wep_num;
	const char *s, *s2;
	char *name, *name1, *name2, *sndbuf, lifespan_str[80];
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

#if 0
	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
		"Usage: `%s <object|me|this>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(s);
	    return(-1);
	}
#endif

	/* Check if this object is allowed to examine? */
	if(con_obj->permission.uid > SWS_ACCESS_UID_EXAMINE)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_EXAMINE
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Parse object name from argument */
	s = arg;
	name = STRDUP(s);
	strstrip(name);

	/* Match the object */
	if(STRISEMPTY(name))
	{
	    if(DBIsObjectGarbage(con_obj->locked_on))
		obj_num = con_obj_num;
	    else
		obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con->object_num;
	}
	else if(!strcasecmp(name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}

	obj = xsw_object[obj_num];


	/* Check if allowed to examine other? */
	if((obj->owner != con_obj_num) &&
	   (con_obj->permission.uid > SWS_ACCESS_UID_EXAMINEO)
	)
	{
	    if(DBIsObjectGarbage(obj->owner))
	    {
		NetSendMessage(
		    con_num,
THIS_CMD_NAME ": *No owner*"
		);
		free(name);
		return(-1);
	    }
	    else
	    {
		xsw_object_struct *owner_obj = xsw_object[obj->owner];
		char *s = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(owner_obj->name) + 80
		);
		sprintf(
		    s,
"%s: Owner: %s",
		    THIS_CMD_NAME, owner_obj->name
		);
		NetSendMessage(con_num, s);
		free(s);
		free(name);
		return(-1);
	    }
	}

	/* Print information */

	/* Line 1: Name, type, uid, and credits */
	name1 = STRDUP(DBGetFormalNameStr(obj_num));
	name2 = STRDUP(DBGetFormalNameStr(obj->owner));
	if(obj->lifespan > 0l)
	    sprintf(
		lifespan_str,
		"Lifespan: %ld ms (%ld ms)",
		(obj->birth_time + obj->lifespan) - cur_ms,
		obj->lifespan
	    );
	else
	    *lifespan_str = '\0';
	s = DBGetTypeName(obj->type);
	sndbuf = (char *)malloc(
	    STRLEN(name1) + STRLEN(name2) + STRLEN(s) + 40 +
	    STRLEN(obj->empire) + STRLEN(lifespan_str) +
	    80
	);
	sprintf(
	    sndbuf,
"%s  Owner: %s  Type: %s  UID: %i  Empire: %s  %s",
	    name1, name2, s, obj->permission.uid,
	    obj->empire, lifespan_str
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);
	free(name1);
	free(name2);

	/* Line 1.25: (optional for Elink objects) */
	if(obj->type == SW_OBJ_TYPE_ELINK)
	{
	    sndbuf = (char *)malloc(STRLEN(obj->elink) + 80);
	    sprintf(
		sndbuf,
"External Link: %s",
		((obj->elink == NULL) ? "*unlinked*" : obj->elink)
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	/* Line 1.5: (optional for AI/Controlled) */
	if(obj->type == SW_OBJ_TYPE_CONTROLLED)
	{
	    sndbuf = (char *)malloc(
		STRLEN(SW_AI_STR_FOLLOW_FRIEND) +
		STRLEN(SW_AI_STR_FOLLOW_UNKNOWN) +
		STRLEN(SW_AI_STR_FOLLOW_HOSTILE) +
		STRLEN(SW_AI_STR_FIRE_FRIEND) +
		STRLEN(SW_AI_STR_FIRE_UNKNOWN) +
		STRLEN(SW_AI_STR_FIRE_HOSTILE) +
		80
	    );
	    sprintf(
		sndbuf,
 "AI Flags: %s%s%s%s%s%s",
		((obj->ai_flags & SW_AI_FOLLOW_FRIEND) ?
		    SW_AI_STR_FOLLOW_FRIEND " " : ""
		),
		((obj->ai_flags & SW_AI_FOLLOW_UNKNOWN) ?
		    SW_AI_STR_FOLLOW_UNKNOWN " " : ""
		),
		((obj->ai_flags & SW_AI_FOLLOW_HOSTILE) ?
		    SW_AI_STR_FOLLOW_HOSTILE " " : ""
		),
		((obj->ai_flags & SW_AI_FIRE_FRIEND) ?
		    SW_AI_STR_FIRE_FRIEND " " : ""
		),
		((obj->ai_flags & SW_AI_FIRE_UNKNOWN) ?
		    SW_AI_STR_FIRE_UNKNOWN " " : ""
		),
		((obj->ai_flags & SW_AI_FIRE_HOSTILE) ?
		    SW_AI_STR_FIRE_HOSTILE " " : ""
		)
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	/* Line 2: Sector, position, location type, and size */
	switch(obj->loc_type)
	{
	  case SW_LOC_TYPE_SPACE_TERRITORY:
	    s = SW_LOC_TYPE_STR_SPACE_TERRITORY;
	    break;
	  case SW_LOC_TYPE_NEBULA:
	    s = SW_LOC_TYPE_STR_NEBULA;
	    break;
	  case SW_LOC_TYPE_ASTEROID_FIELD:
	    s = SW_LOC_TYPE_STR_ASTEROID_FIELD;
	    break;
	  case SW_LOC_TYPE_PULSAR:
	    s = SW_LOC_TYPE_STR_PULSAR;
	    break;
	  default:	/* SW_LOC_TYPE_SPACE */
	    s = SW_LOC_TYPE_STR_SPACE;
	    break;
	}
	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 + 40 + 40 + 40 + 40 + STRLEN(s) +
	    (2 * 80)
	);
	sprintf(
	    sndbuf,
"ImgRef: %i  Size: %ld su\
  Sector: %ld %ld %ld  Position: %.2f %.2f %.2f  Location Type: %s",
	    obj->imgref, obj->size,
	    obj->sect_x, obj->sect_y, obj->sect_z,
	    obj->x, obj->y, obj->z,
	    s
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);

	/* Line 3: Attitude, velocity vector */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 + (2 * 80)
	);
	sprintf(
	    sndbuf,
"Attitude: %.2f' %.2f' %.2f'  Velocity Vector: %.2f' %.2f' %.2f'\
  Turn Rate: %.5f",
	    RADTODEG(obj->heading),
	    RADTODEG(obj->pitch),
	    RADTODEG(obj->bank),
	    RADTODEG(obj->velocity_heading),
	    0.0 * PI,
	    0.0 * PI,
	    obj->turnrate
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);

	/* Line 4: Velocity, thrust power, throttle, engine state */
	switch(obj->engine_state)
	{
	  case SW_ENGINE_STATE_OFF:
	    s = SW_ENGINE_STATE_STR_OFF;
	    break;
	  case SW_ENGINE_STATE_STARTING:
	    s = SW_ENGINE_STATE_STR_STARTING;
	    break;
	  case SW_ENGINE_STATE_ON:
	    s = SW_ENGINE_STATE_STR_ON;
	    break;
	  default:
	    s = SW_ENGINE_STATE_STR_NONE;
	    break;
	}
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + STRLEN(s) + (2 * 80)
	);
	sprintf(
	    sndbuf,
"Velocity: %.5f(%.5f)  Thrust Power: %.5f  Throttle: %.0f%%\
  Engine State: %s",
	    obj->velocity, obj->velocity_max,
	    obj->thrust_power,
	    obj->throttle * 100.0f,
	    s
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);


	/* Line 5: Scanner Range, Locked On and Intercept */
	name1 = STRDUP(DBGetFormalNameStr(obj->locked_on));
	name2 = STRDUP(DBGetFormalNameStr(obj->intercepting_object));
	sndbuf = (char *)malloc(
	    40 + STRLEN(name1) + STRLEN(name2) + 80
	);
	sprintf(
	    sndbuf,
"Scanner Range: %.2f ru  Locked On: %s  Intercept: %s",
	    obj->scanner_range, name1, name2
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);
	free(name1);
	free(name2);
		
	/* Line 6: Hull, power, purity, core efficiency */
	sndbuf = (char *)malloc(                                   
	    40 + 40 + 40 + 40 + 40 + 40 + (2 * 80)
	);
	sprintf(
	    sndbuf,
"Hull: %.2f(%.2f)  Power: %.2f(%.2f)  Purity: %i%%\
  Core Efficency: %.4f",
	    obj->hp, obj->hp_max,
	    obj->power, obj->power_max,
	    (int)(obj->power_purity * 100.0f),
	    obj->core_efficency
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);

	/* Line 7: Antimatter, shields, cloak and visibility */
	switch(obj->shield_state)
	{
	  case SW_SHIELD_STATE_UP:
	    s = SW_SHIELD_STATE_STR_UP;
	    break;
	  case SW_SHIELD_STATE_DOWN:
	    s = SW_SHIELD_STATE_STR_DOWN;
	    break;
	  default:
	    s = SW_SHIELD_STATE_STR_NONE;
	    break;
	}
	switch(obj->cloak_state)
	{
	  case SW_CLOAK_STATE_UP:
	    s2 = SW_CLOAK_STATE_STR_UP;
	    break;
	  case SW_CLOAK_STATE_DOWN:
	    s2 = SW_CLOAK_STATE_STR_DOWN;
	    break;
	  default:
	    s2 = SW_CLOAK_STATE_STR_NONE;
	    break;
	}
	sndbuf = (char *)malloc(
	    40 + 40 + STRLEN(s) + STRLEN(s2) + 40 + 40 + (2 * 80)
	);
	sprintf(
	    sndbuf,
"AM: %.2f(%.2f)  Shields: %s  Cloak: %s  Cloak Strength: %i%%\
  Vis: %i%%",
	    obj->antimatter, obj->antimatter_max,
	    s, s2,
	    (int)(obj->cloak_strength * 100.0f),
	    (int)(obj->visibility * 100.0f)
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);
	    
	/* Print weapons */
	if(obj->weapons != NULL)
	{
	    xsw_weapon_struct *wep;

	    for(wep_num = 0; wep_num < obj->total_weapons; wep_num++)
	    {
		wep = obj->weapons[wep_num];
		if(wep == NULL)
		    continue;

		sndbuf = (char *)malloc(
		    40 + 40 + 40 + 40 + 40 + 40 + 40 + 40 + 40 + 40 +
		    40 + (3 * 80)
		);
		sprintf(
		    sndbuf,
"Wep %i [%s%s%s%s%s%s%s%s%s]:\
  OCS: %i  EmtType: %i  Amt: %i(%i)  Dmg(H/S): %.2f/%.2f\
  Rng: %.2f  CPwr: %.2f  Delay: %ld ms  FireSndRef: %i",
		    wep_num,
		(wep->flags & SW_WEP_FLAG_NO_LOCK) ? "L" : "",
		(wep->flags & SW_WEP_FLAG_FIRE_CLOAKED) ? "C" : "",
		(wep->flags & SW_WEP_FLAG_NO_FIRE_SOUND) ? "Q" : "",
		(wep->flags & SW_WEP_FLAG_FIXED) ? "F" : "",
		(wep->flags & SW_WEP_FLAG_PENETRATE_SHIELD) ? "P" : "",
		(wep->flags & SW_WEP_FLAG_DESTROYABLE) ? "D" : "",
		(wep->flags & SW_WEP_FLAG_SMOKE_TRAIL) ? "T" : "",
		(wep->flags & SW_WEP_FLAG_SMOKE_PUFF_FIRED) ? "S" : "",
		(wep->flags & SW_WEP_FLAG_SMOKE_PUFF_FIRED) ? "W" : "",
		    wep->ocs_code,
		    wep->emission_type,
		    wep->amount,
		    wep->max,
		    wep->damage_hull, wep->damage_shields,
		    (float)wep->range / 1000.0f,
		    wep->create_power,
		    wep->delay,
		    wep->fire_sndref
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	}
 
	/* Scores */
	if(obj->score != NULL)  
	{
	    xsw_score_struct *score = obj->score;
	    sndbuf = (char *)malloc(
		40 + 40 + 40 + 40 + 40 + 40 + (2 * 80)
	    );
	    sprintf(
		sndbuf,
"Credits: %.2f  RMU: %.2f(%.2f)  DmgGiv: %.2f  DmgRec: %.2f  Kills: %i",
		score->credits,
		score->rmu,
		score->rmu_max,
		score->damage_given,
		score->damage_received,
		score->kills
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	/* Economy */
	if(obj->eco != NULL)
	{
	    xsw_ecodata_struct *eco = obj->eco;
	    xsw_ecoproduct_struct *product;

	    sndbuf = (char *)malloc(
		40 + 40 + 40 + 40 + 80
	    );
	    sprintf(
		sndbuf,
"Eco [%s%s%s%s%s]: TaxG: %i%%  TaxF: %i%%  TaxH: %i%%",
		((eco->flags & ECO_FLAG_OPEN) ? "O" : ""),
		((eco->flags & ECO_FLAG_BUY_OK) ? "B" : ""),
		((eco->flags & ECO_FLAG_SELL_OK) ? "S" : ""),
		((eco->flags & ECO_FLAG_TRADE_OK) ? "T" : ""),
		((eco->flags & ECO_FLAG_INTRODUCE_OK) ? "I" : ""),
		(int)((eco->tax_general - 1.0f) * 100.0f),
		(int)((eco->tax_friend - 1.0f) * 100.0f),
		(int)((eco->tax_hostile - 1.0f) * 100.0f)
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    for(i = 0; i < eco->total_products; i++)
	    {
		product = eco->product[i];
		if(product == NULL)
		    continue;

		sndbuf = (char *)malloc(
		    STRLEN(product->name) + 40 + 40 + 40 + 40 + 40 + 80
		);
		if(product->amount_max > 0.0f)
		    sprintf(
			sndbuf,
"EcoProd `%s':  OCS: %i  Buy: %.2f  Sell: %.2f  Amt: %.2f(%.2f)",
			product->name,
			product->ocs_code,
			product->buy_price,
			product->sell_price,
			product->amount,
			product->amount_max
		    );
		else
		    sprintf(
			sndbuf,
"EcoProd `%s':  OCS: %i  Buy: %.2f  Sell: %.2f  Amt: *unlimited*",
			product->name,
			product->ocs_code,
			product->buy_price,
			product->sell_price
		    );
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	}

	/* Description */
	if(!STRISEMPTY(obj->description))
	{
	    sndbuf = (char *)malloc(
		STRLEN(obj->description) + 80
	    );
	    sprintf(
		sndbuf,
"Description: %s",
		obj->description
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	free(name);

	return(0);
}
