#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmath.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdLink(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"link"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdLink(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num, dest_obj_num;
	const char *s;
	char *sndbuf, *name = NULL, *dest_name = NULL;
	xsw_object_struct *obj, *dest_obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		sndbuf,
"Usage: `%s <object|me|this>[=<destination|me|this>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}


	/* Get the object's name */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    if(len > 0)
	    {
		name = (char *)malloc((len + 1) * sizeof(char));
		if(name != NULL)
		{
		    memcpy(name, s, len);
		    name[len] = '\0';
		}
	    }

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	/* Get the destination */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    if(len > 0)
	    {
		dest_name = (char *)malloc((len + 1) * sizeof(char));
		if(dest_name != NULL)
		{      
		    memcpy(dest_name, s, len);
		    dest_name[len] = '\0';
		}
	    }

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	if(name == NULL)
	    name = strdup("");
	if(dest_name == NULL)
	    dest_name = strdup("");

	strstrip(name);
	strstrip(dest_name);


	/* Match the object */
	if(STRISEMPTY(name))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con_obj_num;
	}
        else if(!strcasecmp(name, "this"))
        {
            obj_num = con_obj->locked_on;
        }
	else
	{
	    obj_num = UNVMatchObjectByName(
	        xsw_object, total_objects,
	        name, -1
	    );
	}
	if(DBIsObjectGarbage(obj_num))
	{
	    sndbuf = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		sndbuf,
"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name);
	    free(dest_name);
	    return(-1);
	}
	obj = xsw_object[obj_num];


	/* Match the destination object (if it does not exist then set
	 * dest_obj to NULL)
	 */
	if(STRISEMPTY(dest_name))
	{
	    dest_obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(dest_name, "me"))
	{
	    dest_obj_num = con_obj_num;
	}
        else if(!strcasecmp(dest_name, "this"))
        {
            dest_obj_num = con_obj->locked_on;
        }
	else
	{
	    dest_obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		dest_name, -1
	    );
	}
	dest_obj = DBIsObjectGarbage(dest_obj_num) ?
	    NULL : xsw_object[dest_obj_num];


	/* Does the connection own or can set the object? */
	if(SWS_ACCESS_UID_SET < con_obj->permission.uid)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_SET
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name);
	    free(dest_name);
	    return(-1);
	}
	if((con_obj->owner != con_obj_num) &&
	   (SWS_ACCESS_UID_SETO < con_obj->permission.uid)
	)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80  
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_SETO
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name);
	    free(dest_name);
	    return(-1);
	}
	if(obj->permission.uid < con_obj->permission.uid)
	{
	    char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + STRLEN(name1) +
		40 + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: %s access level %i exceeds your access level %i.",
		THIS_CMD_NAME,
		name1,
		obj->permission.uid,
		con_obj->permission.uid
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name1);
	    free(name);
	    free(dest_name);
	    return(-1);
	}


	/* Link by the object's type */
	switch(obj->type)
	{
	  case SW_OBJ_TYPE_WORMHOLE:
	    if(dest_obj != NULL)
	    {
		/* Link to */
		obj->intercepting_object = dest_obj_num;

		/* Calculate source object heading to target object */
		if(SWMath3DInSameSectorPtr(obj, dest_obj))
		{ 
		    obj->heading = SWMathCoordinateDeltaVector(
			dest_obj->x - obj->x,
			dest_obj->y - obj->y
		    );
		    obj->pitch = 0.0f * PI;
		    obj->bank = 0.0f * PI;
		    SWMathSetUnitVector2D(
			&obj->attitude_vector_compoent,
			obj->heading
		    );
		}
		else
		{
		    obj->heading = SWMathCoordinateDeltaVector(
			dest_obj->sect_x - obj->sect_x,
			dest_obj->sect_y - obj->sect_y
		    );
		    obj->pitch = 0.0f * PI;
		    obj->bank = 0.0f * PI;
		    SWMathSetUnitVector2D(
			&obj->attitude_vector_compoent,
			obj->heading
		    );
		}
	    }
	    else
	    {
		/* Unlink */
		obj->intercepting_object = -1;
		obj->heading = 0.0f * PI;
		obj->pitch = 0.0f * PI;
		obj->bank = 0.0f * PI;
		SWMathSetUnitVector2D(
		    &obj->attitude_vector_compoent,
		    obj->heading
		);
	    }
	    NetSendObjectValues(-1, obj_num);
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	    if(dest_obj != NULL)
	    {
		char	*name1 = STRDUP(DBGetFormalNameStr(obj_num)),
			*name2 = STRDUP(DBGetFormalNameStr(dest_obj_num));
		sndbuf = (char *)malloc(
		    STRLEN(name1) + STRLEN(name2) + 80
		);
		sprintf(
		    sndbuf,
"Linked `%s' to `%s'.",
		    name1, name2
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name1);
		free(name2);
	    }
	    else
	    {
		char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
		sndbuf = (char *)malloc(STRLEN(name1) + 80);
		sprintf(
		    sndbuf,
"Unlinked `%s'.",
		    name1
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name1);
	    }
	    if(sysparm.log_general)
	    {
		if(dest_obj != NULL)
		{
		    char        *name1 = STRDUP(DBGetFormalNameStr(con_obj_num)),
				*name2 = STRDUP(DBGetFormalNameStr(obj_num)),
				*name3 = STRDUP(DBGetFormalNameStr(dest_obj_num)),
				*s = (char *)malloc(
				    STRLEN(name1) + STRLEN(name2) +
				    STRLEN(name3) + 80
				);
		    sprintf(
			s,
"%s: Linked `%s' to `%s'.",
			name1, name2, name3
		    );
		    LogAppend(fname.log, s);
		    free(s);
		    free(name1);
		    free(name2);
		    free(name3);
		}
		else
		{
		    char        *name1 = STRDUP(DBGetFormalNameStr(con_obj_num)),
				*name2 = STRDUP(DBGetFormalNameStr(obj_num)),
				*s = (char *)malloc(
				    STRLEN(name1) + STRLEN(name2) + 80
				);
		    sprintf(
			s,
"%s: Unlinked `%s'.",
			name1, name2
		    );
		    LogAppend(fname.log, s);
		    free(s);
		    free(name1);
		    free(name2);
		}
	    }
	    break;

	  case SW_OBJ_TYPE_ELINK:
	    free(obj->elink);
	    obj->elink = STRISEMPTY(dest_name) ?
		NULL : STRDUP(dest_name);
            NetSendObjectValues(-1, obj_num);
	    if(obj->elink != NULL)
	    {
		char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
		sndbuf = (char *)malloc(
		    STRLEN(name1) + STRLEN(obj->elink) + 80
		);
		sprintf(
		    sndbuf,
"Linked `%s' to `%s'.",
		    name1, obj->elink
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name1);
	    }
	    else
	    {
		char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
		sndbuf = (char *)malloc(STRLEN(name1) + 80);
		sprintf(
		    sndbuf,
"Unlinked `%s'.",
		    name1
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name1); 
	    }
	    if(sysparm.log_general)
	    {
		if(obj->elink != NULL)
		{
		    char        *name1 = STRDUP(DBGetFormalNameStr(con_obj_num)),
				*name2 = STRDUP(DBGetFormalNameStr(obj_num)),
				*s = (char *)malloc(
				    STRLEN(name1) + STRLEN(name2) +
				    STRLEN(obj->elink) + 80
				);
		    sprintf(
			s,
"%s: Linked `%s' to `%s'.",
			name1, name2, obj->elink
		    );
		    LogAppend(fname.log, s);
		    free(s);
		    free(name1);
		    free(name2);
		}
		else
		{
		    char	*name1 = STRDUP(DBGetFormalNameStr(con_obj_num)),
				*name2 = STRDUP(DBGetFormalNameStr(obj_num)),
				*s = (char *)malloc(
				    STRLEN(name1) + STRLEN(name2) + 80
				);
		    sprintf(
			s,
"%s: Unlinked `%s'.",
			name1, name2
		    );
		    LogAppend(fname.log, s);
		    free(s);
		    free(name1);
		    free(name2);
		}
	    }
	    break;

	  default:
	    if(1)
	    {
		char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
		sndbuf = (char *)malloc(STRLEN(name1) + 80);
		sprintf(
		    sndbuf,
"%s: Not a linkable object type.",
		    name1
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name1);
	    }
	    break;
	}

	free(name);
	free(dest_name);

	return(0);
}
