#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h" 


static void CmdNetstatPrintStandard(const int con_num);
static void CmdNetstatPrintListening(const int con_num);
static void CmdNetstatPrintPlayer(const int con_num, const int obj_num);
static xsw_object_struct **CmdNetstatGetConList(int *total_rtn);

int CmdNetstat(const int con_num, const char *arg);


#define THIS_CMD_NAME	"netstat"

#define SPACES		"                                 "

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Prints the netstats list of player objects and aux connections.
 */
static void CmdNetstatPrintStandard(const int con_num)
{
	int i, obj_num, total;
	const char *name, *state_str;
	char *s, *sndbuf;
	xsw_object_struct **list;
	sws_connection_struct *con;
	sws_aux_connection_struct *aux_con;

	/* Heading */
	NetSendMessage(
	    con_num,
"\
Con     \
Player                   \
Address                  \
RX        \
SX        \
ErrRX   \
ErrSX   \
State"
	);

	/* Get a list of connected player objects */
	list = CmdNetstatGetConList(&total);
free(list);	/* Not used yet */

	/* Print the list of connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
		name = "*Unknown*";
	    else
		name = DBGetFormalNameStr(obj_num);
	    if(STRISEMPTY(name))
		name = "*Unknown*";

	    sndbuf = (char *)malloc(
		40 +
		STRLEN(name) +
		STRLEN(con->conhost) +
		40 + 40 + 40 + 40 +
		40 + 40 + STRLEN(SPACES) +
		(2 * 80)
	    );

	    /* Connection */
	    s = sndbuf;
	    sprintf(s, "%i", i + 1);
	    strcat(s, SPACES);
	    s += 8;

	    /* Player */
	    strcpy(s, name);
	    strcat(s, SPACES);
	    s += 25;

	    /* Address */
	    strcpy(s, (con->conhost != NULL) ? con->conhost : "*Unknown*");
	    strcat(s, SPACES);
	    s += 25;

	    /* RX (relative to client) */
	    sprintf(s, "%ld", con->bytes_sent);
	    strcat(s, SPACES);
	    s += 10;

	    /* SX (relative to client) */
	    sprintf(s, "%ld", con->bytes_received);
	    strcat(s, SPACES);
	    s += 10;

	    /* ErrRX (relative to client) */
	    sprintf(s, "%ld", con->errors_sent);
	    strcat(s, SPACES);
	    s += 8;

	    /* ErrSX (relative to client) */
	    sprintf(s, "%ld", con->errors_received);
	    strcat(s, SPACES);
	    s += 8;

	    /* State */
	    if(con->flags & SWS_CON_FLAG_CLOSE_WAIT)
		state_str = "CLOSE_WAIT";
	    else if(con->object_num < 0)
		state_str = "NEGOTIATING";
	    else
		state_str = "ESTABLISHED";
	    strcpy(s, state_str);

	    /* Print */
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	/* Heading */
	NetSendMessage(
	    con_num,
"\
AUX Con \
Address                  \
RX        \
SX        \
ErrRX   \
ErrSX   \
State"
	);

	/* Iterate through aux connections list */
	for(i = 0; i < total_aux_connections; i++)
	{
	    aux_con = aux_connection[i];
	    if(aux_con == NULL)
		continue;

	    if(aux_con->socket < 0)
		continue;

	    sndbuf = (char *)malloc(
		40 +
		STRLEN(aux_con->conhost) +
		40 + 40 + 40 + 40 +
		40 + STRLEN(SPACES) +
		(2 * 80)
	    );

	    /* Connection */
	    s = sndbuf;
	    sprintf(s, "%i", i + 1);
	    strcat(s, SPACES);
	    s += 8;

	    /* Foreign Address */
	    strcpy(s, (aux_con->conhost != NULL) ? aux_con->conhost : "*Unknown*");
	    strcat(s, SPACES);
	    s += 25;

	    /* RX (relative to client) */
	    sprintf(s, "%ld", aux_con->bytes_sent);
	    strcat(s, SPACES);
	    s += 10;

	    /* SX (relative to client) */
	    sprintf(s, "%ld", aux_con->bytes_received);
	    strcat(s, SPACES);
	    s += 10;

	    /* ErrRX (relative to client) */
	    sprintf(s, "%ld", aux_con->errors_sent);
	    strcat(s, SPACES);
	    s += 8;

	    /* ErrSX (relative to client) */
	    sprintf(s, "%ld", aux_con->errors_received);
	    strcat(s, SPACES);
	    s += 8;

	    /* State */
	    if(aux_con->flags & SWS_AUX_CON_FLAG_CLOSE_WAIT)
		state_str = "CLOSE_WAIT";
	    else
		state_str = "ESTABLISHED";
	    strcpy(s, state_str);

	    /* Print */
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}
}

/*
 *	Prints the netstats list of listening sockets.
 */
static void CmdNetstatPrintListening(const int con_num)
{
	int i;
	const char *sv;
	char *s, *sndbuf;
	incoming_socket_struct *insock;

	/* Heading */
	NetSendMessage(con_num,
"\
Type      \
Port      \
Int"
	);

	/* Print each incoming socket */
	for(i = 0; i < total_incoming_sockets; i++)
	{
	    insock = incoming_socket[i];
	    if(insock == NULL)
		continue;

	    if(insock->socket < 0)
		continue;

	    sndbuf = (char *)malloc(
		40 + 40 + 40 +
		STRLEN(SPACES) + (2 * 80)
	    );

	    /* Type */
	    s = sndbuf;
	    switch(insock->type)
	    {
	      case SWS_INCOMING_SOCKET_TYPE_AUXSTATS:
		sv = "aux";
		break;
	      case SWS_INCOMING_SOCKET_TYPE_STANDARD:
		sv = "std";
		break;
	      default:
		sv = "*Unknown*";
		break;
	    }
	    strcpy(s, sv);
	    strcat(s, SPACES);
	    s += 10;

	    /* Port */
	    sprintf(s, "%i", insock->port);
	    strcat(s, SPACES);
	    s += 10;

	    /* Interval */
	    sprintf(s, "%ld", sysparm.int_new_connection_poll);

	    /* Print */
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}
}

/*
 *	Prints the netstats of the specified player object.
 */
static void CmdNetstatPrintPlayer(const int con_num, const int obj_num)
{
	int i, printed;
	const char *name, *state_str;
	char *s, *sndbuf;
	sws_connection_struct *con;

	/* Heading */
	NetSendMessage(
	    con_num,
"\
Con     \
Player                   \
Address                  \
RX        \
SX        \
ErrRX   \
ErrSX   \
State"
	);

	/* Get object's connection */
	printed = 0;
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    if(con->object_num != obj_num)
		continue;

	    name = DBGetFormalNameStr(con->object_num);
	    if(STRISEMPTY(name))
		name = "*Unknown*";

	    sndbuf = (char *)malloc(
		40 +
		STRLEN(name) +
		STRLEN(con->conhost) +
		40 + 40 + 40 + 40 +
		40 + 40 + STRLEN(SPACES) +
		(2 * 80)
	    );

	    /* Connection */
	    s = sndbuf;
	    sprintf(s, "%i", i + 1);
	    strcat(s, SPACES);
	    s += 8;

	    /* Player */
	    strcpy(s, name);
	    strcat(s, SPACES);
	    s += 25;

	    /* Address */ 
	    strcpy(s, (con->conhost != NULL) ? con->conhost : "*Unknown*");
	    strcat(s, SPACES);
	    s += 25;

	    /* RX (relative to client) */
	    sprintf(s, "%ld", con->bytes_sent);
	    strcat(s, SPACES);
	    s += 10;

	    /* SX (relative to client) */
	    sprintf(s, "%ld", con->bytes_received);
	    strcat(s, SPACES);
	    s += 10;

	    /* ErrRX (relative to client) */
	    sprintf(s, "%ld", con->errors_sent);
	    strcat(s, SPACES);
	    s += 8;

	    /* ErrSX (relative to client) */
	    sprintf(s, "%ld", con->errors_received);
	    strcat(s, SPACES);
	    s += 8;

	    /* State */
	    if(con->flags & SWS_CON_FLAG_CLOSE_WAIT)
		state_str = "CLOSE_WAIT";
	    else if(con->object_num < 0)
		state_str = "NEGOTIATING";
	    else
		state_str = "ESTABLISHED";
	    strcpy(s, state_str);

	    /* Print */
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    printed++;
	}

	/* Nothing printed? */
	if(!printed)
	{
	    name = DBGetFormalNameStr(obj_num);
	    if(STRISEMPTY(name))
		name = "*Unknown*";

	    sndbuf = (char *)malloc(
		40 +
		STRLEN(name) +
		40 +
		40 + 40 + 40 + 40 +
		40 + 40 + STRLEN(SPACES) +
		(2 * 80)
	    );

	    /* Connection */
	    s = sndbuf;
	    strcpy(s, "-");
	    strcat(s, SPACES);
	    s += 8;

	    /* Player */
	    strcpy(s, name);
	    strcat(s, SPACES);
	    s += 25;

	    /* Address */
	    strcpy(s, "-");
	    strcat(s, SPACES);
	    s += 25;

	    /* RX */
	    strcpy(s, "0");
	    strcat(s, SPACES);
	    s += 10;

	    /* SX */
	    strcpy(s, "0");
	    strcat(s, SPACES);
	    s += 10;

	    /* ErrRX */
	    strcpy(s, "0");
	    strcat(s, SPACES);
	    s += 8;

	    /* ErrSX */
	    strcpy(s, "0");
	    strcat(s, SPACES);
	    s += 8;

	    /* State */
	    strcpy(s, "NOT_CONNECTED");

	    /* Print */
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}
}

/*
 *	Returns a list of player objects that are connected.
 *
 *	The returned list must be deleted by the calling function, but
 *	not each object.
 */
static xsw_object_struct **CmdNetstatGetConList(int *total_rtn)
{
	int i, n, obj_num, total = 0;
	xsw_object_struct *obj, **list = NULL;
	sws_connection_struct *con;

	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if((con->socket < 0) ||
	       (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    )
		continue;

	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
		continue;

	    obj = xsw_object[obj_num];

	    n = total;
	    total++;
	    list = DBReallocList(list, total);
	    if(list == NULL)
	    {
		total = 0;
		break;
	    }

	    list[n] = obj;
	}

	if(total_rtn != NULL)
	    *total_rtn = total;

	return(list);
}


int CmdNetstat(const int con_num, const char *arg)
{
	int obj_num;
	char *sndbuf;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Insufficient permission? */
	if(con_obj->permission.uid > SWS_ACCESS_UID_NETSTAT)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_NETSTAT
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* No arguments specified? */
	if(STRISEMPTY(arg))
	{
	    /* Print standard listing */
	    CmdNetstatPrintStandard(con_num);

	    /* Print usage at end */
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 80
	    );
	    sprintf(
		sndbuf,
"Usage: `%s [player|me|this] [-l] [-c]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}
	else
	{
	    const char *s = arg;

	    if(*s == '-')
	    {
		if(strcasepfx(s, "-c"))
		{
		    /* Print connections */
		    CmdNetstatPrintStandard(con_num);
		}
		else if(strcasepfx(s, "-l"))
		{
		    /* Print listening sockets */
		    CmdNetstatPrintListening(con_num);
		}
		else
		{
		    /* Print usage */
		    sndbuf = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + 80
		    );
		    sprintf(
			sndbuf,
"Usage: `%s [player|me|this] [-l] [-c]'",
			THIS_CMD_NAME
		    );
		    NetSendMessage(con_num, sndbuf);
		    free(sndbuf);
		}
	    }
	    else
	    {
		/* Print netstats of player */
		char *name = STRDUP(s);

		if(!strcasecmp(name, "me"))
		{
		    obj_num = con_obj_num;
		}
		else if(!strcasecmp(name, "this"))
		{
		    obj_num = con_obj->locked_on;
		}
		else
		{
		    obj_num = UNVMatchObjectByName(
		        xsw_object, total_objects,
			name, SW_OBJ_TYPE_PLAYER
		    );
		}
		if(DBIsObjectGarbage(obj_num))
		{
		    sndbuf = (char *)malloc(
			STRLEN(name) + 80
		    );
		    sprintf(
			sndbuf,
"%s: No such player.",
			name
		    );
		    NetSendMessage(con_num, sndbuf);
		    free(sndbuf);
		}
		else
		{
		    CmdNetstatPrintPlayer(con_num, obj_num);
		}

		free(name);
	    }
	}

	return(0);
}
