#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdRecover(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"recover"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

#define SPACES_60	"                                                           "


int CmdRecover(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num, owner_obj_num;
	char *s, *name, *name1, *name2;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;

	/* No arguments specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    /* Print the list of recycled objects */
	    int i, nobjs = 0;
	    char *owner_name;

	    for(i = 0; i < total_recycled_objects; i++)
	    {
		obj = recycled_xsw_object[i];
		if(obj == NULL)
		    continue;

		if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		    continue;

		if(nobjs == 0)
		{
		    /* There is at least one object, so print heading */
		    NetSendMessage(
			con_num,
"Name                     Type                Owner"
		    );
		}

		owner_obj_num = obj->owner;
		owner_name = STRDUP(DBGetFormalNameStr(owner_obj_num));

		if(obj->type == SW_OBJ_TYPE_PLAYER)
		{
		    char *type_name = STRDUP(DBGetTypeName(obj->type));
		    s = (char *)malloc(
			STRLEN(obj->name) + STRLEN(type_name) +
			STRLEN("-") + STRLEN(SPACES_60) + 80
		    );
		    strcpy(s, obj->name);
		    strcat(s, SPACES_60);
		    strcpy(s + 25, type_name);
		    strcat(s, SPACES_60);
		    strcpy(s + 45, "-");
		    NetSendMessage(con_num, s);
		    free(s);
		    free(type_name);
		}
		else
		{
		    char *type_name = STRDUP(DBGetTypeName(obj->type));
		    s = (char *)malloc(
			STRLEN(obj->name) + STRLEN(type_name) +
			STRLEN(owner_name) + STRLEN(SPACES_60) + 80
		    );
		    strcpy(s, obj->name);
		    strcat(s, SPACES_60);
		    strcpy(s + 25, type_name);
		    strcat(s, SPACES_60);
		    strcpy(s + 45, owner_name);
		    NetSendMessage(con_num, s);
		    free(s);
		    free(type_name);
		}

		free(owner_name);

		nobjs++;
	    }
	    s = (char *)malloc(40 + 40 + 80);
	    if(nobjs > 0)
		sprintf(
		    s,
"%i recycled %s (max is %i).",
		    nobjs,
		    (nobjs == 1) ? "object" : "objects",
		    sysparm.recycled_objects_max
		);
	    else
		sprintf(
		    s,
"No objects recycled (max is %i).",
		    sysparm.recycled_objects_max
		);
	    NetSendMessage(con_num, s);
	    free(s);

	    s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <search_string>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);

	    return(0);
	}

	/* Get the name of the object to recover */
	name = STRDUP(arg);
	strstrip(name);

	/* Attempt to recover the object */
	obj_num = DBRecoverRecycledObjectByName(
	    name,
	    con_obj_num		/* Owner */
	);

	/* Unable to recover? */
	if(DBIsObjectGarbage(obj_num))
	{
	    s = (char *)malloc(STRLEN(name) + 80);
	    switch(obj_num)
	    {
	      case -2:
		sprintf(
		    s,
"%s: No such object recycled.",
		    name
		);
		break;
	      case -3:
                sprintf(
                    s,
"%s: Memory allocation error.",
                    name
                );
                break;
	      case -7:
                sprintf(
                    s,
"%s: Insufficient permission to recover this object.",
                    name
                );
                break;
	      default:
		sprintf(
		    s,
"%s: Not found or unable to recover.",
		    name
		);
		break;
	    }
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}


	/* Get object names */
	name1 = STRDUP(DBGetFormalNameStr(con_obj_num));
	name2 = STRDUP(DBGetFormalNameStr(obj_num));

	/* Print response */
	s = (char *)malloc(STRLEN(name2) + 80);
	sprintf(
	    s,
"Recovered `%s'.",
	    name2
	);
	NetSendMessage(con_num, s);
	free(s);

	/* Log recovery */
	if(sysparm.log_general)
	{
	    s = (char *)malloc(STRLEN(name1) + STRLEN(name2) + 80);
	    sprintf(
		s,
"%s: Recovered `%s'.",
		name1, name2
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	/* Send updates to all connections */
	NetSendCreateObject(-1, obj_num);
	NetSendObjectName(-1, obj_num);
	NetSendObjectMaximums(-1, obj_num);
	NetSendObjectValues(-1, obj_num);

	free(name1);
	free(name2);
	free(name);

	return(0);
}
