#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static int CmdRecycleWeapons(
	const int con_num,
	const int con_obj_num, xsw_object_struct *con_obj,
        const time_t cur_ms, const time_t cur_sec, const float tc  
);
static int CmdRecycleDeadWeapons(
	const int con_num,
	const int con_obj_num, xsw_object_struct *con_obj,
        const time_t cur_ms, const time_t cur_sec, const float tc  
);

int CmdRecycle(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc  
);


#define THIS_CMD_NAME	"recycle"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Recycle all weapons.
 *
 *	All permission checks will be tested and handled accordingly.
 */
static int CmdRecycleWeapons(
	const int con_num,
	const int con_obj_num, xsw_object_struct *con_obj,
        const time_t cur_ms, const time_t cur_sec, const float tc  
)
{
	int i;
	char *s, *name1, *name2;
	xsw_object_struct *obj;

	/* Record the connection object's name */
	name1 = STRDUP(DBGetFormalNameStr(con_obj_num));

	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;
  
	    if((obj->type != SW_OBJ_TYPE_WEAPON) &&
	       (obj->type != SW_OBJ_TYPE_STREAMWEAPON) &&
	       (obj->type != SW_OBJ_TYPE_SPHEREWEAPON)
	    )
		continue; 
 
	    /* Owner by other and can recycle other? */
	    if((con_obj->permission.uid > SWS_ACCESS_UID_RECYCLEO) &&
	       (obj->owner > -1) &&   
	       (obj->owner != con_obj_num)
	    )
		continue;
	 
	    /* Cannot recycle something with a higher uid */
	    if(con_obj->permission.uid > obj->permission.uid)
		continue;
	
	    /* Get name of object we're about to recycle */
	    name2 = STRDUP(DBGetFormalNameStr(i));

	    /* Notify connection of recycling this object */
	    s = (char *)malloc(STRLEN(name2) + 80);
	    sprintf(
		s,
"Recycled `%s'.",
		name2
	    );
	    NetSendMessage(con_num, s);
	    free(s);

	    if(sysparm.log_general)
	    {
		s = (char *)malloc(STRLEN(name1) + STRLEN(name2) + 80);
		sprintf(
		    s,
"%s: Recycled `%s'.",
		    name1, name2
		);
		LogAppend(fname.log, s);
		free(s);
	    }

	    free(name2);

	    /* Recycle the object */
	    DBRecycleObject(i, cur_ms, cur_sec, tc);
	
	    /* Notify all connections about object being recycled */
	    NetSendRecycleObject(-1, i);
	}

	free(name1);

	return(0);
}

/*
 *	Recycle disarmed or dead weapons.
 *
 *	All permission checks will be tested and handled accordingly.
 */
static int CmdRecycleDeadWeapons(
	const int con_num,
	const int con_obj_num, xsw_object_struct *con_obj,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i;
	char *s, *name1, *name2;
	xsw_object_struct *obj;

	/* Record the connection object's name */
	name1 = STRDUP(DBGetFormalNameStr(con_obj_num));

	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if((obj->type != SW_OBJ_TYPE_WEAPON) &&
	       (obj->type != SW_OBJ_TYPE_STREAMWEAPON) &&
	       (obj->type != SW_OBJ_TYPE_SPHEREWEAPON)
	    )
		continue;

	    /* Is this weapon disarmed or dead? */
	    if(obj->antimatter <= 0.0f)
		continue;

	    /* Owner by other and can recycle other? */
	    if((con_obj->permission.uid > SWS_ACCESS_UID_RECYCLEO) &&
	       (obj->owner > -1) &&
	       (obj->owner != con_obj_num)
	    )
		continue;

	    /* Cannot recycle something with a higher uid */
	    if(con_obj->permission.uid > obj->permission.uid)
		continue;


	    /* Get name of object we're about to recycle */
	    name2 = STRDUP(DBGetFormalNameStr(i));

	    /* Notify connection of recycling this object */
	    s = (char *)malloc(STRLEN(name2) + 80);
	    sprintf(
		s,
"Recycled `%s'.",
		name2
	    );
	    NetSendMessage(con_num, s);
	    free(s);

	    if(sysparm.log_general)
	    {
		s = (char *)malloc(STRLEN(name1) + STRLEN(name2) + 80);
		sprintf(
		    s,
"%s: Recycled `%s'.",
		    name1, name2
		);
		LogAppend(fname.log, s);
		free(s);
	    }

	    free(name2);

	    /* Recycle the object */
	    DBRecycleObject(i, cur_ms, cur_sec, tc);

	    /* Notify all connections about object being recycled */
	    NetSendRecycleObject(-1, i);
	}

	free(name1);

	return(0);
}


int CmdRecycle(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc  
)
{
	int obj_num;
	const char *s;
	char *name, *name1, *name2;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* No argument specified? */
	if((arg == NULL) ? 1 : (*arg == '\0'))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s [<object|me|this>|#weapons|#dead_weapons]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Connection allowed to recycle? */
	if(con_obj->permission.uid > SWS_ACCESS_UID_RECYCLE)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_RECYCLE
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Parse the name or first argument */
	s = arg;
	name = STRDUP(s);
	strstrip(name);

	if(STRISEMPTY(name))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s [<object|me|this>|#weapons|#dead_weapons]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}

	/* Recycle all weapons? */
	if(!strcasecmp(name, "#weapons"))
	{
	    const int status = CmdRecycleWeapons(
		con_num,
		con_obj_num, con_obj,
		cur_ms, cur_sec, tc
	    );
	    free(name);
	    return(status);
	}
	/* Recycle all dead weapons? */
	else if(!strcasecmp(name, "#dead_weapons"))
	{
	    const int status = CmdRecycleDeadWeapons(
		con_num,
		con_obj_num, con_obj,
		cur_ms, cur_sec, tc
	    );
	    free(name);
	    return(status);
	}

	/* Match the object */
	if(!strcasecmp(name, "me"))
	{
	    obj_num = con_obj_num;
	}
	else if(!strcasecmp(name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}
	if(DBIsObjectGarbage(obj_num))
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-2);
	}

	obj = xsw_object[obj_num];

	/* Is the object a player? */
	if(obj->type == SW_OBJ_TYPE_PLAYER)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Unable to recycle players, use `recycleplayer' instead.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}

	/* If object owned by other, allowed to recycle others? */
	if((con_obj->permission.uid > SWS_ACCESS_UID_RECYCLEO) &&
	   (obj->owner > -1) &&
	   (obj->owner != con_obj_num)
	)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_RECYCLEO
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}
	/* Is the object's UID greater than the connection's object? */
	else if(con_obj->permission.uid > obj->permission.uid)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 40 + 80);
	    sprintf(
		s,
"%s: Target object access level %i exceeds your access level %i.",
		THIS_CMD_NAME,
		obj->permission.uid,
		con_obj->permission.uid
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}


	name1 = STRDUP(DBGetFormalNameStr(con_obj_num));
	name2 = STRDUP(DBGetFormalNameStr(obj_num));

	if(1)
	{
	    char *s = (char *)malloc(STRLEN(name2) + 80);
	    sprintf(
		s,
"Recycled `%s'.",
		name2
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	}

	if(sysparm.log_general)
	{
	    char *s = (char *)malloc(STRLEN(name1) + STRLEN(name2) + 80);
	    sprintf(
		s,
"%s: Recycled `%s'.",
		name1, name2
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	/* Save object into recycled objects buffer */
	DBSaveRecycledObject(obj_num);

	/* Actually recycle the object */
	DBRecycleObject(obj_num, cur_ms, cur_sec, tc);

	/* Notify all connections about object being recycled */
	NetSendRecycleObject(-1, obj_num);

	free(name);
	free(name1);
	free(name2);

	return(0);
}
