#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdRecyclePlayer(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"recycleplayer"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

 
int CmdRecyclePlayer(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num;
	const char *s;
	char *name, *name1, *name2;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <player|me|this>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Parse the name */
	s = arg;
	name = STRDUP(s);
	strstrip(name);

	if(STRISEMPTY(name))
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,  
"Usage: `%s <player|me|this>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}

	/* Match the object */
	if(!strcasecmp(arg, "me"))
	{
	    obj_num = con_obj_num;
	}
	else if(!strcasecmp(arg, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		arg, SW_OBJ_TYPE_PLAYER
	    );
	}
	if(DBIsObjectGarbage(obj_num))
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such player.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}

	obj = xsw_object[obj_num];

	/* Attempting to recycle player #0? */
	if(obj_num == 0)
	{
	    char	*name1 = STRDUP(DBGetFormalNameStr(obj_num)),
			*s = (char *)malloc(
			    STRLEN(THIS_CMD_NAME) + STRLEN(name1) + 80
			);
	    sprintf(
		s,
"%s: Player #0 may never be recycled.",
		name1
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name1);
	    free(name);
	    return(-1);
	}

	/* Recycling the connection's object? */
	if(obj_num == con_obj_num)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: You may not recycle yourself.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}

	/* Permission check */
	if(con_obj->permission.uid > SWS_ACCESS_UID_RECYCLEPLAYER)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_RECYCLEPLAYER
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    return(-1);
	}

	name1 = STRDUP(DBGetFormalNameStr(con_obj_num));
	name2 = STRDUP(DBGetFormalNameStr(obj_num));

	/* Print response */
	if(1)
	{
	    char *s = (char *)malloc(STRLEN(name2) + 80);
	    sprintf(
		s,
"Recycled `%s'.",
		name2
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	}

	/* Log */
	if(sysparm.log_general)
	{
	    char *s = (char *)malloc(STRLEN(name1) + STRLEN(name2) + 80);
	    sprintf(
		s,
"%s: Recycled `%s'.",
		name1, name2
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	free(name1);
	free(name2);
	free(name);

	/* Recycle player */
	DBSaveRecycledObject(obj_num);
	DBRecycleObject(obj_num, cur_ms, cur_sec, tc);
	NetSendRecycleObject(-1, obj_num);

	return(0);
}
