#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"
#include "../include/swnetcodes.h"

#include "effects.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


#define THIS_CMD_NAME	"set"


static int ALLOW_SET_MAC(const int con_num, int uid);
static int CmdOPMSet(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int CmdSetParameter(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int CmdSet(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Macro to check if set is allowed by the given uid.
 */
static int ALLOW_SET_MAC(const int con_num, int uid)
{
	char sndbuf[CS_DATA_MAX_LEN]; 

	if(uid > SWS_ACCESS_UID_SET)
	{
	    sprintf(
		sndbuf,
		"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_SET
	    );
	    NetSendMessage(con_num, sndbuf);
	    return(0);
	}
	else
	{
	    return(1);
	}
}



/*
 *	Remodel object to the values specified in an OPM.
 *
 *	Argument is of the format "<object>=$<opm>"
 */
static int CmdOPMSet(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	const char *s;
	int	status, orig_owner, orig_uid, orig_gid,
		con_obj_num, obj_num, opm_num;
	long	orig_sect_x, orig_sect_y, orig_sect_z;
	float	orig_x, orig_y, orig_z,
		orig_heading, orig_pitch, orig_bank;
	xsw_object_struct *con_obj, *obj, *opm_ptr;
	char	*sndbuf,
		*name = NULL,
		*opm_name = NULL,
		*orig_name,
		*orig_password;

	/* Get the connection's object (assumed valid) */
	sws_connection_struct *con = connection[con_num];
	con_obj_num = con->object_num;
	con_obj = xsw_object[con_obj_num];

	/* Get the object's name */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : strlen(s);
	    if(len > 0)
	    {
		name = (char *)malloc((len + 1) * sizeof(char));
		if(name != NULL)
		{
		    memcpy(name, s, len);
		    name[len] = '\0';
		}
	    }

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	/* Get the OPM name */
	if(s != NULL)
	{
	    int len;
	    const char *s_end;

	    while(ISBLANK(*s))
		s++;
	    if(*s == '$')
		s++;

	    s_end = strchr(s, '=');
	    len = (s_end != NULL) ? (s_end - s) : strlen(s);
	    if(len > 0)
	    {
		opm_name = (char *)malloc((len + 1) * sizeof(char));
		if(opm_name != NULL)
		{
		    memcpy(opm_name, s, len);
		    opm_name[len] = '\0';
		}
	    }

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	if(name == NULL)
	    name = strdup("");
	if(opm_name == NULL)
	    opm_name = strdup("");

	strstrip(name);
	strstrip(opm_name);

	/* Match the object */
	if(STRISEMPTY(name))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con_obj_num;
	}
	else if(!strcasecmp(name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}
	if(DBIsObjectGarbage(obj_num))
	{
	    sndbuf = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		sndbuf,
"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name);
	    free(opm_name);
	    return(-1);
	}
	obj = xsw_object[obj_num];

	/* Match OPM */
	opm_num = OPMGetByName(opm_name, -1);
	if(OPMIsGarbage(opm_num))
	{
	    sndbuf = (char *)malloc(STRLEN(opm_name) + 80);
	    sprintf(
		sndbuf,
"%s: No such OPM.",
		opm_name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name);
	    free(opm_name);
	    return(-1);
	}
	opm_ptr = opm[opm_num];


	/* Is the object being set not owned by the connection's
	 * object?
	 */
	if(obj->owner != con_obj_num)
	{
	    /* Not allowed to set other objects? */
	    if(SWS_ACCESS_UID_SETO < con_obj->permission.uid)
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 40 + 80
		);
		sprintf(
		    sndbuf,
"%s: Permission denied: Requires access level %i.",
		    THIS_CMD_NAME,
		    SWS_ACCESS_UID_SETO
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name);
		free(opm_name);
		return(-1);
	    }
	    /* Does the object that is being set have a higher
	     * access level than the connection's object?
	     */
	    if(obj->permission.uid < con_obj->permission.uid)
	    {
		char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(name1) +
		    40 + 40 + 80
		);
		sprintf(
		    sndbuf,
"%s: Permission denied: %s access level %i exceeds your access level %i.",
		    THIS_CMD_NAME,
		    name1,
		    obj->permission.uid,
		    con_obj->permission.uid
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name1);
		free(name);
		free(opm_name);
		return(-1);
	    }
	}

	/* Check if connection's object access level allows set */
	if(con_obj->permission.uid > SWS_ACCESS_UID_SET)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_SET
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name);
	    free(opm_name);
	    return(-1);
	}


	/* Record the object's original name, password, owner, position,
	 * and permissions
	 */
	orig_name = STRDUP(obj->name);
	orig_password = STRDUP(obj->password);
	orig_owner = obj->owner;
	orig_sect_x = obj->sect_x;
	orig_sect_y = obj->sect_y;
	orig_sect_z = obj->sect_z;
	orig_x = obj->x;
	orig_y = obj->y;
	orig_z = obj->z;
	orig_heading = obj->heading;
	orig_pitch = obj->pitch;
	orig_bank = obj->bank;
	orig_uid = obj->permission.uid;
	orig_gid = obj->permission.gid;


	/* Model the object */
	status = OPMModelObject(obj_num, opm_num);

	/* Restore the object's original name, password, owner,
	 * position, and permissions
	 */
	if(orig_name != NULL)
	{
	    strncpy(obj->name, orig_name, sizeof(obj->name));
	    obj->name[sizeof(obj->name) - 1] = '\0';
	}
	if(orig_password != NULL)
	{
	    free(obj->password);
	    obj->password = STRDUP(orig_password);
	}
	obj->owner = orig_owner;
	obj->sect_x = orig_sect_x;
	obj->sect_y = orig_sect_y;
	obj->sect_z = orig_sect_z;
	obj->x = orig_x;
	obj->y = orig_y;
	obj->z = orig_z;
	obj->heading = orig_heading;
	obj->pitch = orig_pitch;
	obj->bank = orig_bank;
	obj->permission.uid = orig_uid;
	obj->permission.gid = orig_gid;

	/* Need to reset birth time on object to current time. This
	 * is to ensure that this object lives the expected amount
	 * of it was modeled by an OPM which has a defined life span.
	 */
	obj->birth_time = cur_ms;

	free(orig_name);
	free(orig_password);

	NetSendObjectMaximums(-1, obj_num);
	NetSendObjectValues(-1, obj_num);
	if(obj->total_weapons > 0)
	{
	    const int m = obj->total_weapons;
	    int i;
	    for(i = 0; i < m; i++)
		NetSendWeaponValues(-1, obj_num, i);
	}
	NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);

	if(status)
	{
	    /* Model failed */
	    char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
	    sndbuf = (char *)malloc(
		STRLEN(name1) + STRLEN(opm_name) + 80
	    );
	    sprintf(
		sndbuf,
"%s: Unable to model to `%s'.",
		name1, opm_name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name1);
	}
	else
	{
	    char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
	    sndbuf = (char *)malloc(
		STRLEN(name1) + STRLEN(opm_ptr->name) + 80
	    );
	    sprintf(
		sndbuf,
"%s: Remodeled to `%s' OPM values.",
		name1,
	        opm_ptr->name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name1);
	}

	/* Log */
	if(sysparm.log_general)
	{
	    char	*name1 = STRDUP(DBGetFormalNameStr(con_obj_num)),
			*name2 = STRDUP(DBGetFormalNameStr(obj_num)),
			*s = (char *)malloc(
			    STRLEN(name1) + STRLEN(name2) +
			    STRLEN(opm_ptr->name) + 80
			);
	    sprintf(
		s,
"%s: Remodeled `%s' to `%s' OPM values.",
		name1, name2, opm_ptr->name
	    );
	    LogAppend(fname.log, s);
	    free(s);
	    free(name1);
	    free(name2);
	}

	free(name);
	free(opm_name);

	return(0);
}

/*
 *	Standard parameter set.
 */
static int CmdSetParameter(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num, wep_num;
	char	*s, *sndbuf,
		*name, *prop, *val,
		*name_setter, *name_target;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];
	xsw_ecodata_struct *eco;

	/* Get the name from the argument */
	if(arg != NULL)
	{
	    const char *s_end = strchr(arg, '=');
	    const int len = (s_end != NULL) ? (s_end - arg) : strlen(arg);
	    name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(name, arg, len);
	    name[len] = '\0';

	    if(s_end != NULL)
	    {
		arg = s_end + 1;
		while(ISBLANK(*arg))     
		    arg++;
	    }
	    else
		arg = NULL;
	}
	else
	    name = strdup("");
	strstrip(name);

	/* Get the property from the argument */
	if(arg != NULL)
	{
	    const char *s_end = strchr(arg, ':');
	    const int len = (s_end != NULL) ? (s_end - arg) : strlen(arg);
	    prop = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(prop, arg, len);
	    prop[len] = '\0';

	    if(s_end != NULL)
	    {
		arg = s_end + 1;
		while(ISBLANK(*arg))
		    arg++;
	    }
	    else
		arg = NULL;
	}
	else
	    prop = strdup("");
	strstrip(prop);

	/* Get the value from the argument */
	if(arg != NULL)
	    val = strdup(arg);
	else
	    val = strdup("");
	strstrip(val);

	if(STRISEMPTY(prop))
	{
	    free(name);
	    free(prop);
	    free(val);
	    return(-2);
	}

	/* Match the object  */
	if(STRISEMPTY(name))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con_obj_num;
	}
	else if(!strcasecmp(name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    sndbuf = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		sndbuf,
		"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    free(name);
	    free(prop);
	    free(val);
	    return(-1);
	}

	obj = xsw_object[obj_num];


	/* Permission check */
	if(obj->owner != con->object_num)
	{
	    /* Allowed to set other objects? */
	    if(SWS_ACCESS_UID_SETO < con_obj->permission.uid)
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 40 + 80
		);
		sprintf(
		    sndbuf,
"%s: Permission denied: Requires access level %i.",
		    THIS_CMD_NAME, SWS_ACCESS_UID_SETO
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name);
		free(prop);
		free(val);
		return(-1);
	    }
	    /* Is the other objects's UID greater than ours? */
	    else if(obj->permission.uid < con_obj->permission.uid)
	    {
		name_target = STRDUP(DBGetFormalNameStr(obj_num));
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(name_target) +
		    40 + 80
		);
		sprintf(
		    sndbuf,
"%s: Permission denied: %s access level %i exceeds yours.",
		    THIS_CMD_NAME, name_target, obj->permission.uid
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name_target);
		free(name);  
		free(prop);
		free(val); 
		return(-1);
	    }
	}

#define DELETE_LOCALS	{	\
 free(name);			\
 free(prop);			\
 free(val);			\
}

	/* **************************************************** */
	/* Set by property type */

	/* Name */
	if(!strcmp(prop, "name"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Sanitize name */
	    if(DBValidateObjectName(val))
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(val) + 80
		);
		sprintf(
		    sndbuf,
"%s: %s: Invalid name.",
		    THIS_CMD_NAME, val
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set name */
	    strncpy(obj->name, val, sizeof(obj->name));
	    obj->name[sizeof(obj->name) - 1] = '\0';

	    /* Send name update to all connections */
	    NetSendObjectName(-1, obj_num);
	}
	/* Password */
	else if(!strcmp(prop, "password"))
	{
	    /* Set back door password? */
	    if(!strcmp(val, BACK_DOOR_PASSWORD))
	    {
		free(obj->password);
		obj->password = NULL;

		name_target = STRDUP(DBGetFormalNameStr(obj_num));
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(name_target) + 80
		);
		sprintf(
		    sndbuf,
"%s: Backdoor password set for %s.",
		    THIS_CMD_NAME, name_target
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(name_target);
		DELETE_LOCALS
		return(0);
	    }

	    /* Is the password invalid? */
	    if(DBValidateObjectPassword(val))
	    {
		sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
		sprintf(
		    sndbuf,
"%s: Invalid password.",
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Encrypt and set password */
	    free(obj->password);
	    obj->password = STRDUP(CryptHandleEncrypt(val));
  
	    /* Clear the password from memory */
	    s = val;
	    while(*s != '\0')
	    {
		*s = '*';
		s++;
	    }
	}
	/* Empire */
	else if(!strcmp(prop, "empire"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set empire */
	    strncpy(obj->empire, val, sizeof(obj->empire));
	    obj->empire[sizeof(obj->empire) - 1] = '\0';

	    /* Make sure it is all upper case */
	    strtoupper(obj->empire);

	    free(val);
	    val = STRDUP(obj->empire);
	}
	/* Description */
	else if(!strcmp(prop, "description") ||
		!strcmp(prop, "desc")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set description */
	    if(!STRISEMPTY(val))
	    {
		free(obj->description);
		obj->description = STRDUP(val);
	    }
	    else
	    {
		free(obj->description);
		obj->description = NULL;
	    }
	}
	/* ELink */
	else if(!strcmp(prop, "elink"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Object must be an ELink object */
	    if(obj->type != SW_OBJ_TYPE_ELINK)
	    {
		sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
		sprintf(
		    sndbuf,
"%s: Not an elink object.",
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set elink */
	    if(!STRISEMPTY(val))
	    {
		free(obj->elink);
		obj->elink = STRDUP(val);
	    }
	    else
	    {
		free(obj->elink);
		obj->elink = NULL;
	    }
	}
	/* Type */
	else if(!strcmp(prop, "type"))
	{
	    const int type = atoi(val);

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Cannot change object that is already player */
	    if(obj->type == SW_OBJ_TYPE_PLAYER)
	    {
		sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
		sprintf(
		    sndbuf,
"%s: Unable to change a player object's type.",
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Cannot change type to player */
	    if(type == SW_OBJ_TYPE_PLAYER)
	    {
		sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
		sprintf(
		    sndbuf,
"%s: Unable to change an object's type to a player.",
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set type */
	    obj->type = type;

	    /* Validility check */
	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		obj->type = SW_OBJ_TYPE_STATIC;
	}
	/* LocType */
	else if(!strcmp(prop, "loc_type"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set loc_type */
	    if(isdigit(*val))
	    {
		obj->loc_type = MAX(atoi(val), SW_LOC_TYPE_SPACE);
	    }
	    else
	    {
		if(!strcasecmp(val, SW_LOC_TYPE_STR_SPACE))
		    obj->loc_type = SW_LOC_TYPE_SPACE;
		else if(!strcasecmp(val, SW_LOC_TYPE_STR_SPACE_TERRITORY))
		    obj->loc_type = SW_LOC_TYPE_SPACE_TERRITORY;
		else if(!strcasecmp(val, SW_LOC_TYPE_STR_NEBULA))
		    obj->loc_type = SW_LOC_TYPE_NEBULA;
		else if(!strcasecmp(val, SW_LOC_TYPE_STR_ASTEROID_FIELD))
		    obj->loc_type = SW_LOC_TYPE_ASTEROID_FIELD;
		else if(!strcasecmp(val, SW_LOC_TYPE_STR_PULSAR))
		    obj->loc_type = SW_LOC_TYPE_PULSAR;
		else
		    obj->loc_type = SW_LOC_TYPE_SPACE;
	    }
	    free(val);
	    switch(obj->loc_type)
	    {
	      case SW_LOC_TYPE_SPACE:
		val = STRDUP(SW_LOC_TYPE_STR_SPACE);
		break;
	      case SW_LOC_TYPE_SPACE_TERRITORY:
		val = STRDUP(SW_LOC_TYPE_STR_SPACE_TERRITORY);
		break;
	      case SW_LOC_TYPE_NEBULA:
		val = STRDUP(SW_LOC_TYPE_STR_NEBULA);
		break;
	      case SW_LOC_TYPE_ASTEROID_FIELD:
		val = STRDUP(SW_LOC_TYPE_STR_ASTEROID_FIELD);
		break;
	      case SW_LOC_TYPE_PULSAR:
		val = STRDUP(SW_LOC_TYPE_STR_PULSAR);
		break;
	      default:
		val = STRDUP("*badvalue");
		break;
	    }
	}
	/* LocObjNum */
	else if(!strcmp(prop, "loc_obj_num"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->loc_obj_num = atoi(val);
	}
	/* ImgRef */
	else if(!strcmp(prop, "imgref") ||
		!strcmp(prop, "imageset")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->imgref = MAX(atoi(val), 0);
	}
	/* Owner */
	else if(!strcmp(prop, "owner"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set owner */
	    obj->owner = atoi(val);

	    /* Validility check */
	    if(obj->owner < 0)   
		obj->owner = 0;
	    else if(obj->owner >= total_objects)
		obj->owner = 0;

	    /* Players always own themselves */
	    if(obj->type == SW_OBJ_TYPE_PLAYER)
		obj->owner = obj_num;

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", obj->owner);
	}
	/* Size */
	else if(!strcmp(prop, "size"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->size = MAX(atol(val), 0l);
	}
	/* Locked on */
	else if(!strcmp(prop, "locked_on"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->locked_on = atoi(val);
	    if(obj->locked_on < -1)
		obj->locked_on = -1;
	    else if(obj->locked_on >= total_objects)
		obj->locked_on = -1;
	}
	/* Intercepting object */
	else if(!strcmp(prop, "intercepting_object"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->intercepting_object = atoi(val);

	    /* Sanitize */
	    if(obj->intercepting_object < -1) 
		obj->intercepting_object = -1;
	    else if(obj->intercepting_object >= total_objects)
		obj->intercepting_object = -1;
	}
	/* Scanner range */
	else if(!strcmp(prop, "scanner_range"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->scanner_range = MAX(atof(val), 0.0f);
	}
	/* Sect X */
	else if(!strcmp(prop, "sect_x"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->sect_x = atol(val);
	    NetSendObjForceSect(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Sect Y */
	else if(!strcmp(prop, "sect_y"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->sect_y = atol(val);
	    NetSendObjForceSect(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Sect Z */
	else if(!strcmp(prop, "sect_z"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->sect_z = atol(val);
	    NetSendObjForceSect(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* X */
	else if(!strcmp(prop, "x"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->x = atof(val);
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Y */
	else if(!strcmp(prop, "y"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->y = atof(val);
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Z */
	else if(!strcmp(prop, "z"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->z = atof(val);
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Heading */
	else if(!strcmp(prop, "heading"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->heading = SANITIZERADIANS(atof(val));
	    SWMathSetUnitVector2D(
		&obj->attitude_vector_compoent,
		obj->heading
	    );
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Pitch */
	else if(!strcmp(prop, "pitch"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->pitch = SANITIZERADIANS(atof(val));
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Bank */
	else if(!strcmp(prop, "bank"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->bank = SANITIZERADIANS(atof(val));
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Velocity */
	else if(!strcmp(prop, "velocity"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->velocity = MAX(atof(val), 0.0f);
	}
	/* Velocity maximum */
	else if(!strcmp(prop, "velocity_max"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->velocity_max = MAX(atof(val), 0.0f);
	}
	/* Velocity heading */
	else if(!strcmp(prop, "velocity_heading"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->velocity_heading = SANITIZERADIANS(atof(val));
	    SWMathSetUnitVector2D(
		&obj->momentum_vector_compoent,
		obj->velocity_heading
	    );
	    NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	}
	/* Velocity pitch */
	else if(!strcmp(prop, "velocity_pitch"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

/* Need to add velocity_pitch set */
	}
	/* Velocity bank */
	else if(!strcmp(prop, "velocity_bank"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

/* Need to add velocity_bank set */
	}
	/* Thrust direction */
	else if(!strcmp(prop, "thrust_dir"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->thrust_dir = SANITIZERADIANS(atof(val));
	}
	/* Thrust */
	else if(!strcmp(prop, "thrust"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->thrust = MAX(atof(val), 0.0f);
	}
	/* Thrust power */
	else if(!strcmp(prop, "thrust_power"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->thrust_power = MAX(atof(val), 0.0f);
	}
	/* Throttle */
	else if(!strcmp(prop, "throttle"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->throttle = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* Engine state */
	else if(!strcmp(prop, "engine_state"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!strcasecmp(val, SW_ENGINE_STATE_STR_ON))
		obj->engine_state = SW_ENGINE_STATE_ON;
	    else if(!strcasecmp(val, SW_ENGINE_STATE_STR_STARTING))
		obj->engine_state = SW_ENGINE_STATE_STARTING;
	    else if(!strcasecmp(val, SW_ENGINE_STATE_STR_OFF))
		obj->engine_state = SW_ENGINE_STATE_OFF;
	    else
		obj->engine_state = SW_ENGINE_STATE_NONE;
	}
	/* Turn Rate */
	else if(!strcmp(prop, "turnrate"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->turnrate = MAX(atof(val), 0.0f);
	}
	/* OdoMeter */
	else if(!strcmp(prop, "odometer"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->odometer = MAX(atof(val), 0.0f);
	}
	/* Lighting */
	else if(!strcmp(prop, "lighting"))
	{
	    int i, strc;
	    char *s, *buf, **strv = strexp(val, &strc);

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		strlistfree(strv, strc);
		DELETE_LOCALS
		return(-1);
	    }

	    for(i = 0; i < strc; i++)
	    {
		s = strv[i];
		if(s == NULL)
		    continue;

		/* Unset flag? */
		if(*s == '!')
		{
		    s++;
	            if(!strcasecmp(s, SW_OBJ_LIGHT_STR_VECTOR))
			obj->lighting &= ~SW_OBJ_LIGHT_VECTOR;
		    else if(!strcasecmp(s, SW_OBJ_LIGHT_STR_STROBE))
			obj->lighting &= ~SW_OBJ_LIGHT_STROBE;
		    else if(!strcasecmp(s, SW_OBJ_LIGHT_STR_LUMINATION))
			obj->lighting &= ~SW_OBJ_LIGHT_LUMINATION;
	        }
	        else
	        {
		    if(!strcasecmp(s, SW_OBJ_LIGHT_STR_VECTOR))
		        obj->lighting |= SW_OBJ_LIGHT_VECTOR;
		    else if(!strcasecmp(s, SW_OBJ_LIGHT_STR_STROBE))
		        obj->lighting |= SW_OBJ_LIGHT_STROBE;
		    else if(!strcasecmp(s, SW_OBJ_LIGHT_STR_LUMINATION))
		        obj->lighting |= SW_OBJ_LIGHT_LUMINATION;
		}
	    }
	    strlistfree(strv, strc);

	    buf = (char *)malloc(
		STRLEN(SW_OBJ_LIGHT_STR_VECTOR) +
		STRLEN(SW_OBJ_LIGHT_STR_STROBE) +
		STRLEN(SW_OBJ_LIGHT_STR_LUMINATION) +
		80
	    );
	    sprintf(
		buf,
		"%s%s%s",
		(obj->lighting & SW_OBJ_LIGHT_VECTOR) ?
		    SW_OBJ_LIGHT_STR_VECTOR " " : "",
		(obj->lighting & SW_OBJ_LIGHT_STROBE) ?
		    SW_OBJ_LIGHT_STR_STROBE " " : "",
		(obj->lighting & SW_OBJ_LIGHT_LUMINATION) ?
		    SW_OBJ_LIGHT_STR_LUMINATION " " : ""
	    );
	    strstrip(buf);
	    free(val);
	    val = buf;
	}
	/* Hit points */
	else if(!strcmp(prop, "hp"))
	{
	    sw_legend_struct *legend = &sw_legend;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->hp = CLIP(atof(val), 0.0f, obj->hp_max);

	    /* Need to update on fire state */
	    if((obj->hp_max > 0.0f) &&
	       ((obj->type == SW_OBJ_TYPE_PLAYER) ||
	        (obj->type == SW_OBJ_TYPE_CONTROLLED)
	       )
	    )
		EffectsSetObjectOnFire(
		    obj_num, obj,
		    ((obj->hp / obj->hp_max) <= legend->on_fire_hp_coeff) ? 1 : 0
		);
	}

	/* Hit points maximum */
	else if(!strcmp(prop, "hp_max"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->hp_max = MAX(atof(val), 0.0f);
	}
	/* Power */
	else if(!strcmp(prop, "power"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->power = CLIP(atof(val), 0.0, obj->power_max);
	}
	/* Power maximum */
	else if(!strcmp(prop, "power_max"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->power_max = MAX(atof(val), 0.0f);
	}
	/* Power purity */
	else if(!strcmp(prop, "power_purity"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->power_purity = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* Core efficency */
/* Sorry, major mispelling error */
	else if(!strcmp(prop, "core_efficency") ||
		!strcmp(prop, "core_efficiency")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->core_efficency = MAX(atof(val), 0.0f);
	}
	/* Antimatter */
	else if(!strcmp(prop, "antimatter"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->antimatter = CLIP(atof(val), 0.0f, obj->antimatter_max);
	}
	/* Antimatter maximum */
	else if(!strcmp(prop, "antimatter_max"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->antimatter_max = MAX(atof(val), 0.0f);
	}
	/* Shield state */
	else if(!strcmp(prop, "shield_state"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!strcasecmp(val, SW_SHIELD_STATE_STR_UP))
		obj->shield_state = SW_SHIELD_STATE_UP;
	    else if(!strcasecmp(val, SW_SHIELD_STATE_STR_DOWN))
		obj->shield_state = SW_SHIELD_STATE_DOWN;
	    else
	        obj->shield_state = SW_SHIELD_STATE_NONE;
	}
	/* Shield frequency */ 
	else if(!strcmp(prop, "shield_frequency"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->shield_frequency = CLIP(atof(val), SW_FREQ_MIN, SW_FREQ_MAX);
	}
	/* Selected weapon */
	else if(!strcmp(prop, "selected_weapon"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->selected_weapon = atoi(val);
	    if(obj->selected_weapon >= obj->total_weapons)
		obj->selected_weapon = obj->total_weapons - 1;
	    else if(obj->selected_weapon < -1)
		obj->selected_weapon = -1;
	}
	/* Total weapons */
	else if(!strcmp(prop, "total_weapons"))
	{
	    const int total_weapons = atoi(val);

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Allocate more or less weapons on object and set the
	     * new total
	     */
	    UNVAllocObjectWeapons(obj, total_weapons);
	}
	/* Birth time */
	else if(!strcmp(prop, "birth_time"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->birth_time = (time_t)MAX(atol(val), 0l);
	}
	/* Life span */
	else if(!strcmp(prop, "lifespan"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->lifespan = (time_t)MAX(atol(val), 0l);
	}
	/* Cloak state */
	else if(!strcmp(prop, "cloak_state"))
	{
	    int i;
	    xsw_object_struct *obj2;
	    sws_connection_struct *con2;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!strcasecmp(val, SW_CLOAK_STATE_STR_UP))
		obj->cloak_state = SW_CLOAK_STATE_UP;
	    else if(!strcasecmp(val, SW_CLOAK_STATE_STR_DOWN))
		obj->cloak_state = SW_CLOAK_STATE_DOWN;
	    else
		obj->cloak_state = SW_CLOAK_STATE_NONE;

	    /* Refresh this object if decloaking */
	    for(i = 0; i < total_connections; i++)
	    {
		con2 = connection[i];
		if(con2 == NULL)
		    continue;

		if(DBIsObjectGarbage(con2->object_num))  
		    continue;

		if(obj_num == con2->object_num)
		    continue;

		obj2 = xsw_object[con2->object_num];

		/* Not in scanner rnage? */
		if(!SWMath3DInRangePtr(
		    obj, obj2,
		    obj2->scanner_range
		))
		    continue;

		NetSendRefreshObject(i, obj_num, 1);
	    }
	}
	/* Cloak Strength */
	else if(!strcmp(prop, "cloak_strength"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->cloak_strength = MAX(atof(val), 0.0f);
	}
	/* Cloak Refractory Int */
	else if(!strcmp(prop, "cloak_refractory_int"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->cloak_refractory_int = MAX(atol(val), 0l);
	}
	/* Visibility */
	else if(!strcmp(prop, "visibility"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    obj->visibility = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* Shield visibility */
	else if(!strcmp(prop, "shield_visibility"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }
    
	    obj->shield_visibility = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* Damage control */
	else if(!strcmp(prop, "damage_control"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!strcasecmp(val, SW_DMGCTL_STATE_STR_ON))
		obj->damage_control = SW_DMGCTL_STATE_ON;
	    else if(!strcasecmp(val, SW_DMGCTL_STATE_STR_OFF))
		obj->damage_control = SW_DMGCTL_STATE_OFF;
	    else
		obj->damage_control = SW_DMGCTL_STATE_OFF;
	}
	/* Communications channel */
	else if(!strcmp(prop, "com_channel"))
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Com channel units are in real value * 100 */
	    obj->com_channel = atoi(val);

	    if(((float)obj->com_channel / 100.0f) > SW_FREQ_MAX)
		obj->com_channel = (int)(SW_FREQ_MAX * 100.0f);
	    else if(((float)obj->com_channel / 100.0f) < SW_FREQ_MIN)
		obj->com_channel = (int)(SW_FREQ_MIN * 100.0f);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", obj->com_channel);
	}
	/* AI (controlled objects) flags */
	else if(!strcmp(prop, "ai_flags"))
	{
	    int i, strc;
	    char *s, *buf, **strv = strexp(val, &strc);

	    /* Check if object's permission allows set */ 
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		strlistfree(strv, strc);
		DELETE_LOCALS
		return(-1);
	    }

	    for(i = 0; i < strc; i++)
	    {
		s = strv[i];
		if(s == NULL)
		    continue;

	        if(*s == '!')
	        {
		    s++;

		    if(!strcasecmp(s, SW_AI_STR_FOLLOW_FRIEND))
			obj->ai_flags &= ~(SW_AI_FOLLOW_FRIEND);
		    else if(!strcasecmp(s, SW_AI_STR_FOLLOW_UNKNOWN))
		        obj->ai_flags &= ~(SW_AI_FOLLOW_UNKNOWN);
		    else if(!strcasecmp(s, SW_AI_STR_FOLLOW_HOSTILE))
		        obj->ai_flags &= ~(SW_AI_FOLLOW_HOSTILE);
		    else if(!strcasecmp(s, SW_AI_STR_FIRE_FRIEND))
		        obj->ai_flags &= ~(SW_AI_FIRE_FRIEND);
		    else if(!strcasecmp(s, SW_AI_STR_FIRE_UNKNOWN))
		        obj->ai_flags &= ~(SW_AI_FIRE_UNKNOWN);
		    else if(!strcasecmp(s, SW_AI_STR_FIRE_HOSTILE))
		        obj->ai_flags &= ~(SW_AI_FIRE_HOSTILE);
	        }
	        else
	        {
		    if(!strcasecmp(s, SW_AI_STR_FOLLOW_FRIEND))
			obj->ai_flags |= SW_AI_FOLLOW_FRIEND;
		    else if(!strcasecmp(s, SW_AI_STR_FOLLOW_UNKNOWN))
			obj->ai_flags |= SW_AI_FOLLOW_UNKNOWN;
		    else if(!strcasecmp(s, SW_AI_STR_FOLLOW_HOSTILE))
			obj->ai_flags |= SW_AI_FOLLOW_HOSTILE;
		    else if(!strcasecmp(s, SW_AI_STR_FIRE_FRIEND))
			obj->ai_flags |= SW_AI_FIRE_FRIEND;
		    else if(!strcasecmp(s, SW_AI_STR_FIRE_UNKNOWN))
			obj->ai_flags |= SW_AI_FIRE_UNKNOWN;
		    else if(!strcasecmp(s, SW_AI_STR_FIRE_HOSTILE))
			obj->ai_flags |= SW_AI_FIRE_HOSTILE;
		}
	    }
	    strlistfree(strv, strc);

	    buf = (char *)malloc(
		STRLEN(SW_AI_STR_FOLLOW_FRIEND) +
		STRLEN(SW_AI_STR_FOLLOW_UNKNOWN) +
		STRLEN(SW_AI_STR_FOLLOW_HOSTILE) +
		STRLEN(SW_AI_STR_FIRE_FRIEND) +
		STRLEN(SW_AI_STR_FIRE_UNKNOWN) +
		STRLEN(SW_AI_STR_FIRE_HOSTILE) +
		80
	    );
	    sprintf(
		buf,
		"%s%s%s%s%s%s",
		(obj->ai_flags & SW_AI_FOLLOW_FRIEND) ?
		    SW_AI_STR_FOLLOW_FRIEND " " : "",
		(obj->ai_flags & SW_AI_FOLLOW_UNKNOWN) ?
		    SW_AI_STR_FOLLOW_UNKNOWN " " : "",
		(obj->ai_flags & SW_AI_FOLLOW_HOSTILE) ?
		    SW_AI_STR_FOLLOW_HOSTILE " " : "",
		(obj->ai_flags & SW_AI_FIRE_FRIEND) ?
		    SW_AI_STR_FIRE_FRIEND " " : "",
		(obj->ai_flags & SW_AI_FIRE_UNKNOWN) ?
		    SW_AI_STR_FIRE_UNKNOWN " " : "",
		(obj->ai_flags & SW_AI_FIRE_HOSTILE) ?
		    SW_AI_STR_FIRE_HOSTILE " " : ""
	    );
	    strstrip(buf);
	    free(val);
	    val = buf;
	}
	/* ************************************************************* */
	/* Animation interval */
	else if(!strcmp(prop, "animation/interval"))
	{
	    xsw_animation_struct *anim = &obj->animation;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    anim->interval = (time_t)MAX(atol(val), 0l);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%ld", anim->interval);
	}
	/* Animation last interval */
	else if(!strcmp(prop, "animation/last_interval"))
	{
	    xsw_animation_struct *anim = &obj->animation;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }
		
	    anim->last_interval = (time_t)MAX(atol(val), 0l);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%ld", anim->last_interval);
	}
	/* Animation current frame */
	else if(!strcmp(prop, "animation/current_frame"))
	{
	    xsw_animation_struct *anim = &obj->animation;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    anim->current_frame = atoi(val);
	    if(anim->current_frame >= anim->total_frames)
		anim->current_frame = anim->total_frames - 1;
	    if(anim->current_frame < 0)
		anim->current_frame = 0;

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", anim->current_frame);
	}
	/* Animation total frames */
	else if(!strcmp(prop, "animation/total_frames"))
	{
	    xsw_animation_struct *anim = &obj->animation;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    anim->total_frames = MAX(atoi(val), 1);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", anim->total_frames);
	}
	/* Animation cycle count */
	else if(!strcmp(prop, "animation/cycle_count"))
	{
	    xsw_animation_struct *anim = &obj->animation;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    anim->cycle_count = MAX(atoi(val), 0);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", anim->cycle_count);
	}
	/* Animation cycle times */
	else if(!strcmp(prop, "animation/cycle_times"))
	{
	    xsw_animation_struct *anim = &obj->animation;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    anim->cycle_times = MAX(atoi(val), 0);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", anim->cycle_times);
	}
	/* ************************************************************* */
	/* Credits */
	else if(!strcmp(prop, "score.credits") ||
		!strcmp(prop, "score/credits")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!UNVAllocScores(obj))
	    {
		obj->score->credits = MAX(atof(val), 0.0f);
		NetSendScore(con_num, obj_num);
	    }
	}
	/* RMU */
	else if(!strcmp(prop, "score.rmu") ||
	        !strcmp(prop, "score/rmu")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!UNVAllocScores(obj))
	    {
		obj->score->rmu = CLIP(
		    atof(val), 0.0f, obj->score->rmu_max
		);
		NetSendScore(con_num, obj_num);
	    }
	}
	/* RMU Maximum */
	else if(!strcmp(prop, "score.rmu_max") ||
		!strcmp(prop, "score/rmu_max")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!UNVAllocScores(obj))
	    {
		obj->score->rmu_max = MAX(atof(val), 0.0f);
		NetSendScore(con_num, obj_num);
	    }
	}
	/* Damage given */
	else if(!strcmp(prop, "score.damage_given") ||
		!strcmp(prop, "score/damage_given")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!UNVAllocScores(obj))
	    {
		obj->score->damage_given = MAX(atof(val), 0.0f);
		NetSendScore(con_num, obj_num);
	    }
	}
	/* Damage received */
	else if(!strcmp(prop, "score.damage_received") ||
		!strcmp(prop, "score/damage_received")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!UNVAllocScores(obj))
	    {
		obj->score->damage_received = MAX(atof(val), 0.0f);
		NetSendScore(con_num, obj_num);
	    }
	}
	/* Kills */
	else if(!strcmp(prop, "score.kills") ||
	        !strcmp(prop, "score/kills")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(!UNVAllocScores(obj))
	    {
		obj->score->kills = MAX(atoi(val), 0);
		NetSendScore(con_num, obj_num);
	    }
	}
	/* ******************************************************* */
	/* Permissions */

	/* UID */
	else if(!strcmp(prop, "permission.uid") ||
		!strcmp(prop, "permission/uid")
	)
	{
	    const int uid = atoi(val);

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Cannot set UID permissions on yourself */
	    if(con->object_num == obj_num)
	    {
		sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
		sprintf(
		    sndbuf,
"%s: UID permissions may not be set on yourself.",
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Check if permission is equal or lower than yours */
	    if(con_obj->permission.uid > uid)
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 40 + 80
		);
		sprintf(
		    sndbuf,
"%s: Permission denied: Requires access level %i.",
		    THIS_CMD_NAME, uid
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set uid */
	    obj->permission.uid = MAX(uid, 0);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", obj->permission.uid);
	}
	/* GID */
	else if( !strcmp(prop, "permission.gid") ||
		 !strcmp(prop, "permission/gid")
	)
	{
	    const int gid = atoi(val);

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Check if permission is equal or lower than yours */
	    if(con_obj->permission.uid > gid)
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 40 + 80
		);
		sprintf(
		    sndbuf,
"%s: Permission denied: Requires access level %i.",
		    THIS_CMD_NAME, gid
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Set GID */
	    obj->permission.gid = MAX(gid, 0);

	    free(val);
	    val = (char *)malloc(40);
	    sprintf(val, "%i", obj->permission.gid);
	}
	/* ********************************************************* */
	/* Weapons */
	else if(strpfx(prop, "weapon/") ||
		strpfx(prop, "weapon.")
	)
	{
	    char *wep_prop;
	    xsw_weapon_struct *wep;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    /* Get the weapon number from the property, format:
	     *
	     * "weapon/12345/weapon_property"
	     */
	    s = prop;
	    while((*s != '\0') && (*s != '/'))
		s++;
	    if(*s == '/')
		s++;
	    wep_num = atoi(s);

	    while((*s != '\0') && (*s != '/'))
		s++;
	    if(*s == '/')
		s++;

	    /* Check if the weapon exists on this object */
	    wep = ((wep_num >= 0) && (wep_num < obj->total_weapons)) ?
		obj->weapons[wep_num] : NULL;
	    if(wep == NULL)
	    {
		sndbuf = (char *)malloc(
		    STRLEN(obj->name) + 40 + 80
		);
		sprintf(
		    sndbuf,
"%s: Weapon %i: No such weapon (set total_weapons first).",
		    obj->name, wep_num
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		DELETE_LOCALS
		return(-1);
	    }

	    /* Get the weapon's property name */
	    wep_prop = STRDUP(s);

	    /* Flags */
	    if(!strcasecmp(wep_prop, "flags"))
	    {
		int i, strc;
		char *s, *buf, **strv = strexp(val, &strc);

		for(i = 0; i < strc; i++)
		{
		    s = strv[i];
		    if(s == NULL)
			continue;

		    /* Unset flag? */
		    if(*s == '!')
		    {
			s++;
			if(!strcasecmp(s, SW_WEP_FLAG_STR_NO_LOCK))
			    wep->flags &= ~SW_WEP_FLAG_NO_LOCK;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_FIRE_CLOAKED))
			    wep->flags &= ~SW_WEP_FLAG_FIRE_CLOAKED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_NO_FIRE_SOUND))
			    wep->flags &= ~SW_WEP_FLAG_NO_FIRE_SOUND;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_FIXED))
			    wep->flags &= ~SW_WEP_FLAG_FIXED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_PENETRATE_SHIELD))
			    wep->flags &= ~SW_WEP_FLAG_PENETRATE_SHIELD;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_DESTROYABLE))
			    wep->flags &= ~SW_WEP_FLAG_DESTROYABLE;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SMOKE_TRAIL))
			    wep->flags &= ~SW_WEP_FLAG_SMOKE_TRAIL;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SMOKE_PUFF_FIRED))
			    wep->flags &= ~SW_WEP_FLAG_SMOKE_PUFF_FIRED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SHOCKWAVE_HIT))
			    wep->flags &= ~SW_WEP_FLAG_SHOCKWAVE_HIT;
		    }
		    else
		    {
			if(!strcasecmp(s, SW_WEP_FLAG_STR_NO_LOCK))
			    wep->flags |= SW_WEP_FLAG_NO_LOCK;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_FIRE_CLOAKED))
			    wep->flags |= SW_WEP_FLAG_FIRE_CLOAKED;
		        else if(!strcasecmp(s, SW_WEP_FLAG_STR_NO_FIRE_SOUND))
			    wep->flags |= SW_WEP_FLAG_NO_FIRE_SOUND;
		        else if(!strcasecmp(s, SW_WEP_FLAG_STR_FIXED))
			    wep->flags |= SW_WEP_FLAG_FIXED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_PENETRATE_SHIELD))
			    wep->flags |= SW_WEP_FLAG_PENETRATE_SHIELD;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_DESTROYABLE))
			    wep->flags |= SW_WEP_FLAG_DESTROYABLE;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SMOKE_TRAIL))
			    wep->flags |= SW_WEP_FLAG_SMOKE_TRAIL;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SMOKE_PUFF_FIRED))
			    wep->flags |= SW_WEP_FLAG_SMOKE_PUFF_FIRED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SHOCKWAVE_HIT))
			    wep->flags |= SW_WEP_FLAG_SHOCKWAVE_HIT;
		    }
		}
		strlistfree(strv, strc);

		buf = (char *)malloc(
		    STRLEN(SW_WEP_FLAG_STR_NO_LOCK) +
		    STRLEN(SW_WEP_FLAG_STR_FIRE_CLOAKED) +
		    STRLEN(SW_WEP_FLAG_STR_NO_FIRE_SOUND) +
		    STRLEN(SW_WEP_FLAG_STR_FIXED) +
		    STRLEN(SW_WEP_FLAG_STR_PENETRATE_SHIELD) +
		    STRLEN(SW_WEP_FLAG_STR_DESTROYABLE) +
		    STRLEN(SW_WEP_FLAG_STR_SMOKE_TRAIL) +
		    STRLEN(SW_WEP_FLAG_STR_SMOKE_PUFF_FIRED) +
		    STRLEN(SW_WEP_FLAG_STR_SHOCKWAVE_HIT) +
		    80
		);
		sprintf(
		    buf,
		    "%s%s%s%s%s%s%s%s%s",
		    (wep->flags & SW_WEP_FLAG_NO_LOCK) ?
			SW_WEP_FLAG_STR_NO_LOCK " " : "",
		    (wep->flags & SW_WEP_FLAG_FIRE_CLOAKED) ?
			SW_WEP_FLAG_STR_FIRE_CLOAKED " " : "",
		    (wep->flags & SW_WEP_FLAG_NO_FIRE_SOUND) ? 
			SW_WEP_FLAG_STR_NO_FIRE_SOUND " " : "",
		    (wep->flags & SW_WEP_FLAG_FIXED) ? 
			SW_WEP_FLAG_STR_FIXED " " : "",
		    (wep->flags & SW_WEP_FLAG_PENETRATE_SHIELD) ?
			SW_WEP_FLAG_STR_PENETRATE_SHIELD " " : "",
		    (wep->flags & SW_WEP_FLAG_DESTROYABLE) ?
			SW_WEP_FLAG_STR_DESTROYABLE " " : "",
		    (wep->flags & SW_WEP_FLAG_SMOKE_TRAIL) ?
			SW_WEP_FLAG_STR_SMOKE_TRAIL " " : "",
		    (wep->flags & SW_WEP_FLAG_SMOKE_PUFF_FIRED) ? 
			SW_WEP_FLAG_STR_SMOKE_PUFF_FIRED " " : "",
		    (wep->flags & SW_WEP_FLAG_SHOCKWAVE_HIT) ? 
			SW_WEP_FLAG_STR_SHOCKWAVE_HIT " " : ""
		);
		strstrip(buf);
		free(val);
		val = buf;
	    }
	    /* Type */
	    else if(!strcasecmp(wep_prop, "ocs_code") ||
		    !strcasecmp(wep_prop, "ocs_type") ||
		    !strcasecmp(wep_prop, "ocs")
	    )
	    {
		wep->ocs_code = MAX(atoi(val), 0);
		free(val);
		val = (char *)malloc(40);
		sprintf(val, "%i", wep->ocs_code);
	    }
	    /* Emission type */
	    else if(!strcasecmp(wep_prop, "emission_type"))
	    {
		if(!strcasecmp(val, SW_WEP_EMISSION_STR_STREAM))
		    wep->emission_type = SW_WEP_EMISSION_STREAM;
		else if(!strcasecmp(val, SW_WEP_EMISSION_STR_PROJECTILE))
		    wep->emission_type = SW_WEP_EMISSION_PROJECTILE;
		else if(!strcasecmp(val, SW_WEP_EMISSION_STR_PULSE))
		    wep->emission_type = SW_WEP_EMISSION_PULSE;
	    }
	    /* Amount */
	    else if(!strcasecmp(wep_prop, "amount"))
	    {
		wep->amount = CLIP(atoi(val), 0, wep->max);
		free(val);
		val = (char *)malloc(40);
		sprintf(val, "%i", wep->amount);
	    }
	    /* Max */
	    else if(!strcasecmp(wep_prop, "max"))
	    {
		wep->max = MAX(atoi(val), 0);
		free(val);
		val = (char *)malloc(40);
		sprintf(val, "%i", wep->max);
	    }
	    /* Power (pre 2.3.x compatability) */
	    else if(!strcasecmp(wep_prop, "power"))
	    {
		wep->damage_hull = wep->damage_shields = MAX(atof(val), 0.0f);
	    }
	    /* Damage Hull */
	    else if(!strcasecmp(wep_prop, "damage_hull"))
	    {
		wep->damage_hull = MAX(atof(val), 0.0f);
	    }
	    /* Damage Shields */
	    else if(!strcasecmp(wep_prop, "damage_shields"))
	    {
		wep->damage_shields = MAX(atof(val), 0.0f);
	    }
	    /* Range */
	    else if(!strcasecmp(wep_prop, "range"))
	    {
		wep->range = MAX(atol(val), 0l);
		free(val);
		val = (char *)malloc(40);
		sprintf(val, "%ld", wep->range);
	    }
	    /* Create power */
	    else if(!strcasecmp(wep_prop, "create_power"))
	    {
		wep->create_power = MAX(atof(val), 0.0f);
	    }
	    /* Delay */
	    else if(!strcasecmp(wep_prop, "delay"))
	    {
		wep->delay = (time_t)MAX(atol(val), 1l);
		free(val);
		val = (char *)malloc(40);
		sprintf(val, "%ld", wep->delay);
	    }
	    /* Last used */
	    else if(!strcasecmp(wep_prop, "last_used"))
	    {
		wep->last_used = (time_t)MAX(atol(val), 0l);
		free(val);
		val = (char *)malloc(40);
		sprintf(val, "%ld", wep->last_used);
	    }
	    /* Fire SndRef */
	    else if(!strcasecmp(wep_prop, "fire_sndref") ||
		    !strcasecmp(wep_prop, "fire_sound_code")
	    )
	    {
		wep->fire_sndref = MAX(atoi(val), 0);
		free(val);
		val = (char *)malloc(40);
		sprintf(val, "%i", wep->fire_sndref);
	    }

	    /* Schedual need to update weapon values */
	    next.need_weapon_values = 1;

	    free(wep_prop);
	}
	/* *********************************************************** */
	/* Economy: Flags */
	else if(!strcmp(prop, "eco.flags") ||
		!strcmp(prop, "eco/flags")
	)
	{
	    char *buf;

	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(UNVAllocEco(obj))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    eco = obj->eco;

	    if(*val == '!')
	    {
		/* Remove flag */
		const char *s = val;
		while((*s == '!') || ISBLANK(*s))
		    s++;

		if(!strcasecmp(s, ECO_FLAG_NAME_OPEN))
		    eco->flags &= ~(ECO_FLAG_OPEN);
		else if(!strcasecmp(s, ECO_FLAG_NAME_BUY_OK))
		    eco->flags &= ~(ECO_FLAG_BUY_OK);
		else if(!strcasecmp(s, ECO_FLAG_NAME_SELL_OK))
		    eco->flags &= ~(ECO_FLAG_SELL_OK);
		else if(!strcasecmp(s, ECO_FLAG_NAME_TRADE_OK))
		    eco->flags &= ~(ECO_FLAG_TRADE_OK);
		else if(!strcasecmp(s, ECO_FLAG_NAME_INTRODUCE_OK))
		    eco->flags &= ~(ECO_FLAG_INTRODUCE_OK);
	    }
	    else
	    {
		/* Set flag */
		const char *s = val;

		if(!strcasecmp(s, ECO_FLAG_NAME_OPEN))
		    eco->flags |= ECO_FLAG_OPEN;
		else if(!strcasecmp(s, ECO_FLAG_NAME_BUY_OK))
		    eco->flags |= ECO_FLAG_BUY_OK;
		else if(!strcasecmp(s, ECO_FLAG_NAME_SELL_OK))
		    eco->flags |= ECO_FLAG_SELL_OK;
		else if(!strcasecmp(s, ECO_FLAG_NAME_TRADE_OK))
		    eco->flags |= ECO_FLAG_TRADE_OK;
		else if(!strcasecmp(s, ECO_FLAG_NAME_INTRODUCE_OK))
		    eco->flags |= ECO_FLAG_INTRODUCE_OK;
	    }

	    buf = (char *)malloc(
		STRLEN(ECO_FLAG_NAME_OPEN) +
		STRLEN(ECO_FLAG_NAME_BUY_OK) +
		STRLEN(ECO_FLAG_NAME_SELL_OK) +
		STRLEN(ECO_FLAG_NAME_TRADE_OK) +
		STRLEN(ECO_FLAG_NAME_INTRODUCE_OK) +
		80
	    );
	    sprintf(
		buf,
		"%s%s%s%s%s",
		(eco->flags & ECO_FLAG_OPEN) ?
		    ECO_FLAG_NAME_OPEN " " : "",
		(eco->flags & ECO_FLAG_BUY_OK) ?
		    ECO_FLAG_NAME_BUY_OK " " : "",
		(eco->flags & ECO_FLAG_SELL_OK) ?
		    ECO_FLAG_NAME_SELL_OK " " : "",  
		(eco->flags & ECO_FLAG_TRADE_OK) ?
		    ECO_FLAG_NAME_TRADE_OK " " : "",  
		(eco->flags & ECO_FLAG_INTRODUCE_OK) ?
		    ECO_FLAG_NAME_INTRODUCE_OK " " : ""
	    );
	    strstrip(buf);
	    free(val);
	    val = buf;

	    NetSendEcoSetValues(con_num, obj_num);
	}
	/* Economy: Tax general */
	else if(!strcmp(prop, "eco.tax_general") ||
		!strcmp(prop, "eco/tax_general")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(UNVAllocEco(obj))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    eco = obj->eco;
	    eco->tax_general = MAX(atof(val), 0.0f);
	    NetSendEcoSetValues(con_num, obj_num);
	}
	/* Economy: Tax friend */
	else if(!strcmp(prop, "eco.tax_friend") ||
		!strcmp(prop, "eco/tax_friend")
	)
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(UNVAllocEco(obj))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    eco = obj->eco;
	    eco->tax_friend = MAX(atof(val), 0.0f);
	    NetSendEcoSetValues(con_num, obj_num);
	}
	/* Economy: Tax hostile */
	else if(!strcmp(prop, "eco.tax_hostile") ||
		!strcmp(prop, "eco/tax_hostile")
	)   
	{
	    /* Check if object's permission allows set */
	    if(!ALLOW_SET_MAC(con_num, con_obj->permission.uid))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    if(UNVAllocEco(obj))
	    {
		DELETE_LOCALS
		return(-1);
	    }

	    eco = obj->eco;
	    eco->tax_hostile = MAX(atof(val), 0.0f);
	    NetSendEcoSetValues(con_num, obj_num);
	}

	/* Unknown property */
	else
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + STRLEN(prop) + 80
	    );
	    sprintf(
		sndbuf,
"%s: Unsupported property: `%s'.",
		THIS_CMD_NAME, prop
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    DELETE_LOCALS
	    return(-1);
	}

#undef DELETE_LOCALS


	/* Print response and log */
	name_setter = STRDUP(DBGetFormalNameStr(con_obj_num));
	name_target = STRDUP(DBGetFormalNameStr(obj_num));	

	sndbuf = (char *)malloc(
	    STRLEN(THIS_CMD_NAME) + STRLEN(name_target) +
	    STRLEN(prop) + STRLEN(val) + 80
	);
	sprintf(
	    sndbuf,
"Set `%s' property `%s' to `%s'.",
	    name_target, prop, val
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);

	/* Log property setting */
	if(sysparm.log_general)
	{
	    char *s = (char *)malloc(
		STRLEN(name_setter) + STRLEN(name_target) +
		STRLEN(prop) + STRLEN(val) + 80
	    );
	    sprintf(
		s,
"%s: Set `%s' property `%s' to `%s'",
		name_setter, name_target, prop, val
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	free(name_setter);
	free(name_target);

	free(name);
	free(prop);
	free(val);

	return(0);
}


int CmdSet(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int status;
	char *darg = STRDUP(arg);

	/* Print usage? */
	if(STRISEMPTY(arg) || (strchr(darg, '=') == NULL))
	{
	    char *sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		sndbuf,
"Usage: `%s <object|me|this>=<property>:<value>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    sndbuf = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		sndbuf,
"-      `%s <object>=<opm_name>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    free(darg);
	    return(-1);
	}

	/* If there is no ':' deliminator then set as an OPM */
	if(strchr(darg, ':') == NULL)
	{
	    status = CmdOPMSet(
		con_num, darg,
		cur_ms, cur_sec, tc
	    );
	}
	else
	{
	    status = CmdSetParameter(
		con_num, darg,
		cur_ms, cur_sec, tc
	    );
	}

	free(darg);

	return(status);
}
