#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmath.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdSignup(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"signup"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdSignup(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, opm_num;
	const char *name;
	char    *s, *sndbuf,
		player_name[SW_OBJ_NAME_MAX],
		vessel_name[SW_OBJ_NAME_MAX];
	xsw_object_struct *opm_ptr, *tar_obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Signup allowed? */
	if(!sysparm.allow_guest_signup)
	{
	    sndbuf = (char *)malloc(
		STRLEN(sysparm.guest_login_name) + 80
	    );
	    sprintf(
		sndbuf,
"%s signups are not allowed.",
		sysparm.guest_login_name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Print usage? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    int nvessels = 0;

	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 80
	    );
	    sprintf(
		sndbuf,
		"Usage: `%s <name>=<vessel_type>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    /* Print list of vessels */
	    for(i = 0; i < total_opms; i++)
	    {
		opm_ptr = opm[i];
		if(opm_ptr == NULL)
		    continue;

		if(opm_ptr->type != SW_OBJ_TYPE_PLAYER)
		    continue;

		name = opm_ptr->name;

		/* Skip new player and guest opms */
		if(!strcmp(name, SWS_DEF_NEW_PLAYER_OPM_NAME) ||
		   !strcmp(name, sysparm.guest_login_name)
		)
		    continue;

		if(nvessels == 0)
		{
		    sndbuf = (char *)malloc(80);
		    sprintf(
			sndbuf,
			"Available vessel types:"
		    );
		    NetSendMessage(con_num, sndbuf);
		    free(sndbuf);
		}
		sndbuf = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    sndbuf,
		    "    %s",
		    name
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		nvessels++;
	    }
	    if(nvessels > 0)
	    {
		sndbuf = (char *)malloc(40 + 80);
		sprintf(
		    sndbuf,
		    "%i vessel%s listed",
		    nvessels,
		    (nvessels == 1) ? "" : "s"
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    return(-1);
	}

	/* Already signed up? */
	if(!(con->flags & SWS_CON_FLAG_IS_GUEST))
	{
	    sndbuf = (char *)malloc(80);
	    sprintf(
		sndbuf,
"You are already signed up."
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Parse player and vessel name */
	strncpy(player_name, arg, sizeof(player_name));
	player_name[sizeof(player_name) - 1] = '\0';
	s = strchr(player_name, '=');
	if(s != NULL)
	    *s = '\0';
	strstrip(player_name);

	while((*arg != '=') && (*arg != '\0'))
	    arg++;
	if(*arg == '=')
	    arg++;
	strncpy(vessel_name, arg, sizeof(vessel_name));
	vessel_name[sizeof(vessel_name) - 1] = '\0';
	s = strchr(vessel_name, '=');
	if(s != NULL)
	    *s = '\0';
	strstrip(vessel_name);


	/* Check if the name is valid for a player to have */
	if(DBValidateObjectName(player_name))
	{
	    sndbuf = (char *)malloc(80);
	    sprintf(
		sndbuf,
"The specified name is not valid for a player's name."
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}
	/* The player name may also not have the guest name prefix */
	if(strpfx(player_name, sysparm.guest_login_name))
	{
	    sndbuf = (char *)malloc(80);
	    sprintf(
		sndbuf,
"The specified name is not valid for a player's name."
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Check if the player name already exists */
	for(i = 0; i < total_objects; i++)
	{
	    tar_obj = xsw_object[i];
	    if(tar_obj == NULL)
		continue;
	    if(tar_obj->type != SW_OBJ_TYPE_PLAYER)
		continue;

	    if(!strcmp(tar_obj->name, player_name))
	    {
		sndbuf = (char *)malloc(STRLEN(player_name) + 80);
		sprintf(
		    sndbuf,
"%s: An object by that name already exists.",
		    player_name
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		return(-1);
	    }
	}

	/* Get the OPM */
	for(i = 0; i < total_opms; i++)
	{
	    opm_ptr = opm[i];
	    if(opm_ptr == NULL)
		continue;

	    if(opm_ptr->type != SW_OBJ_TYPE_PLAYER)
		continue;

	    name = opm_ptr->name;

	    /* Skip new player and guest opms */
	    if(!strcmp(name, SWS_DEF_NEW_PLAYER_OPM_NAME) ||
	       !strcmp(name, sysparm.guest_login_name)
	    )
		continue;

	    if(!strcmp(vessel_name, name))
		break;
	}
	if(i < total_opms)
	{
	    opm_num = i;
	}
	else
	{
	    sndbuf = (char *)malloc(
		STRLEN(vessel_name) + STRLEN(THIS_CMD_NAME) + 80
	    );
	    sprintf(
		sndbuf,
"%s: No such vessel (type `%s' with no arguments for a list).",
		vessel_name,
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Got valid player name and vessel OPM
	 *
	 * Model the guest connection's object to the vessel OPM
	 */
	if(con_obj != NULL)
	{
	    int wep_num;
	    const float shield_frequency = con_obj->shield_frequency;
	    const int com_channel = con_obj->com_channel;

	    if(OPMModelObject(con_obj_num, opm_num))
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + STRLEN(vessel_name) + 80
		);
		sprintf(
		    sndbuf,
"%s: Unable to model vessel `%s'.",
		    THIS_CMD_NAME, vessel_name
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		return(-1);
	    }

	    con_obj->last_connected = cur_sec;
	    con_obj->shield_frequency = shield_frequency;
	    con_obj->com_channel = com_channel;
	    con_obj->owner = con_obj_num;

	    /* Update object values to connections */
	    NetSendObjForceSect(-1, con_obj_num, cur_ms, cur_sec, tc);
	    NetSendObjectForcePose(-1, con_obj_num, cur_ms, cur_sec, tc);
	    NetSendObjectMaximums(-1, con_obj_num);
	    for(wep_num = 0; wep_num <= con_obj->total_weapons; wep_num++)
		NetSendWeaponValues(-1, con_obj_num, wep_num);
	    NetSendScore(-1, con_obj_num);
	}

	/* Successfully modeled the guest connection's object to
	 * the selected vessel OPM
	 *
	 * Set the connection as no longer being a guest connection
	 * and the player's anme
	 */
	con->flags &= ~SWS_CON_FLAG_IS_GUEST;

	strncpy(con_obj->name, player_name, sizeof(con_obj->name));
	con_obj->name[sizeof(con_obj->name) - 1] = '\0';
	NetSendObjectName(-1, con_obj_num);

	free(con_obj->password);
	con_obj->password = NULL;

	NetSendObjectDescription(con_num, con_obj_num);


	/* Send refresh to other connection's objects that are within
	 * scanner range
	 */
	if(con_obj != NULL)
	{
	    xsw_object_struct *obj2;
	    sws_connection_struct *con2;

	    for(i = 0; i < total_connections; i++)
	    {
		con2 = connection[i];
		if(con2 == NULL)
		    continue;

		if((con2->object_num < 0) ||
		   (con2->flags & SWS_CON_FLAG_CLOSE_WAIT)
		)
		    continue;

		if(i == con_num)
		    continue;

		if(DBIsObjectGarbage(con2->object_num))
		    continue;

		obj2 = xsw_object[con2->object_num];

                if(con_obj->owner != con2->object_num)
                {
                    if(DBGetObjectVisibilityPtr(con_obj) <= 0.0f)
                        continue;
                }

		/* Not in scanner range? */
		if(!SWMath3DInRangePtr(
		    con_obj, obj2,
		    obj2->scanner_range
		))
		    continue;

		NetSendRefreshObject(i, con_obj_num, 1);
	    }
	}

	/* Notify the connection about the successful signup */
	sndbuf = (char *)malloc(
	    STRLEN(player_name) + STRLEN(vessel_name) + 80
	);    
	sprintf(
	    sndbuf,
"Successfully signed up `%s' as vessel `%s'.",
	    player_name, vessel_name
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);

	/* Notify the other connections about the signup? */
	if(sysparm.con_notify && (con_obj != NULL))
	{
	    sws_connection_struct *con2;
	    sndbuf = (char *)malloc(
		STRLEN(player_name) + STRLEN(con_obj->empire) +
		STRLEN(vessel_name) + 80
	    );
	    sprintf(
		sndbuf,
"%s has signed up with the %s empire as vessel `%s'.",
		player_name, con_obj->empire, vessel_name
	    );
	    for(i = 0; i < total_connections; i++)
	    {
		con2 = connection[i];
		if(con2 == NULL)
		    continue;

		if(i == con_num)
		    continue;

		if(con2->object_num < 0)
		    continue;

		NetSendMessage(i, sndbuf);
	    }
	    free(sndbuf);
	}

	/* Log property setting */
	if(sysparm.log_general)
	{
	    char	*name = STRDUP(DBGetFormalNameStr(con_obj_num)),
			*s = (char *)malloc(
			    STRLEN(name) + STRLEN(vessel_name) + 80
			);
	    sprintf(
		s,
"%s: Signed up as vessel `%s'.",
		name, vessel_name
	    );
	    LogAppend(fname.log, s);
	    free(s);
	    free(name);
	}

	return(0);
}
