#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "siteban.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdSiteBan(const int con_num, const char *arg);


#define THIS_CMD_NAME	"siteban"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdSiteBan(const int con_num, const char *arg)
{
	int op = 0;
	const char *s;
	char *sndbuf, *address;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	if(con_obj->permission.uid > SWS_ACCESS_UID_SITEBAN)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_SITEBAN
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Get the operation */
	s = arg;
	if(s != NULL)
	{
	    char c;

	    while(ISBLANK(*s))
		s++;

	    c = tolower(*s);
	    if((c == '+') ||
	       (c == 'a')
	    )
		op = 1;
	    else if((c == '-') ||
		    (c == 'd') ||
		    (c == 'r')
	    )
		op = 2;

	    while((*s != '\0') && !ISBLANK(*s))
		s++;
	    while(ISBLANK(*s))
		s++;
	}

	/* Get the address */
	if(s != NULL)
	{
	    int len;
	    const char *s_end;

	    s_end = s;
	    while((*s_end != '\0') && !ISBLANK(*s_end))
		s_end++;

	    len = s_end - s;
	    address = (char *)malloc((len + 1) * sizeof(char));
	    if(address != NULL)
	    {
		memcpy(address, s, len);
		address[len] = '\0';
	    }

	    s = s_end;
	    while(ISBLANK(*s))
		s++;
	}
	else
	{
	    address = strdup("");
	}

	/* Check which operation to perform */
	switch(op)
	{
	  case 1:	/* Add entry */
	    if(STRISEMPTY(address))
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 80
		);
		sprintf(
		    sndbuf,
"Usage: `%s [+ <ip>] [- <ip>]'",
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    else
	    {
		siteban_ip_union ip;

	        StringParseIP(
		    address,
		    &(ip.part_u8[0]),
		    &(ip.part_u8[1]),
		    &(ip.part_u8[2]),
		    &(ip.part_u8[3])
	        );
	        if(SiteBanAdd(&ip, 0) < 0)
		{
		    sndbuf = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + 40 + 40 + 40 + 40 + 80
		    );
	            sprintf(
			sndbuf,
"%s: Error adding IP `%i.%i.%i.%i'.",
		        THIS_CMD_NAME,
		        ip.part_u8[0],
		        ip.part_u8[1],
		        ip.part_u8[2],
		        ip.part_u8[3]
		    );
		}
	        else
		{
		    sndbuf = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + 40 + 40 + 40 + 40 + 80
		    );
		    sprintf(
			sndbuf,
"%s: Added `%i.%i.%i.%i' to the siteban list.", 
			THIS_CMD_NAME,
			ip.part_u8[0],
			ip.part_u8[1],
			ip.part_u8[2],
			ip.part_u8[3]
		    );
		}
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    break;

	  case 2:       /* Remove entry */
	    if(STRISEMPTY(address))
	    {
		sndbuf = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 80
		);
		sprintf(
		    sndbuf,
"Usage: `%s [+ <ip>] [- <ip>]'",
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    else
 	    {
		siteban_ip_union ip;

		StringParseIP(
		    address,
		    &(ip.part_u8[0]),  
		    &(ip.part_u8[1]), 
		    &(ip.part_u8[2]), 
		    &(ip.part_u8[3])
		);
		if(SiteBanRemoveIP(&ip) < 0)
		{
		    sndbuf = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + 40 + 40 + 40 + 40 + 80
		    );
		    sprintf(
			sndbuf,
"%s: No such IP `%i.%i.%i.%i' in the site ban list.",
			THIS_CMD_NAME,
			ip.part_u8[0],
			ip.part_u8[1],
			ip.part_u8[2],
			ip.part_u8[3]
		    );
		}
		else
		{
		    sndbuf = (char *)malloc(
			STRLEN(THIS_CMD_NAME) + 40 + 40 + 40 + 40 + 80
		    );
		    sprintf(
			sndbuf,
"%s: Removed IP `%i.%i.%i.%i' from siteban list.",
			THIS_CMD_NAME,
			ip.part_u8[0],
			ip.part_u8[1],
			ip.part_u8[2],
			ip.part_u8[3]
		    );
		}
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    break;

	  default:	/* List siteban IPs */
	    if(total_sitebans > 0)
	    {
		int i, nsitebans = 0;
		siteban_struct *sb;
		siteban_ip_union *ip;

		for(i = 0; i < total_sitebans; i++)
		{
		    sb = siteban[i];
		    if(sb == NULL)
			continue;

		    ip = &sb->ip;
		    sndbuf = (char *)malloc(
			40 + 40 + 40 + 40 + 80
		    );
		    sprintf(
			sndbuf,
"%i.%i.%i.%i",
			ip->part_u8[0],
			ip->part_u8[1],
			ip->part_u8[2],
			ip->part_u8[3]
		    );
		    NetSendMessage(con_num, sndbuf);
		    free(sndbuf);
		    nsitebans++;
		}
		sndbuf = (char *)malloc(40 + 80);
		sprintf(   
		    sndbuf,
"*** %i Sitebanned %s ***",
		    nsitebans,
		    ((nsitebans == 1) ? "Address" : "Addresses")
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    else
	    {
		const int nsitebans = 0;
		sndbuf = (char *)malloc(40 + 80);
		sprintf(
		    sndbuf,
"*** %i Sitebanned %s ***",
		    nsitebans,
		    ((nsitebans == 1) ? "Address" : "Addresses")
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 80
	    );
	    sprintf(
		sndbuf,
"Usage: `%s [+ <ip>] [- <ip>]'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    break;
	}

	free(address);

	return(0);
}
