#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdSyncTime(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"synctime"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdSyncTime(
        const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Check if object's permission allows sync time */
	if(con_obj->permission.uid > SWS_ACCESS_UID_SYNCTIME)
	{
	    char *s = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_SYNCTIME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Log */
	if(sysparm.log_general)
	{
	    char *name = STRDUP(DBGetFormalNameStr(con_obj_num));
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: Syncronized timmers.",
		name
	    );
	    LogAppend(fname.log, s);
	    free(s);
	    free(name);
	}

	/* Warn all connections about timming reset */
	NetSendMessage(
	    -1,
"Server: Syncronizing timmers..."
	);

	/* Reset all timmers */
	SWSResetTimmers(cur_ms, cur_sec, tc);

	/* Warn all connections about timming reset */
	NetSendMessage(
	    -1,
"Server: Timmers syncronized."
	);

	return(0);
}
