#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int CmdUptime(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"uptime"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int CmdUptime(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	const time_t start_time = sws_start_time;
	time_t uptime;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Permission check */
	if(con_obj->permission.uid > SWS_ACCESS_UID_UPTIME)
	{
	    char *s = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_UPTIME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Calculate the uptime */
	uptime = ((start_time > 0l) && (start_time <= cur_sec)) ?
	    (cur_sec - start_time) : 0l;

	if(uptime > 0l)
	{
	    char	*s, start_time_str[80],
			*uptime_str = STRDUP(StringFormatTimePeriod(uptime));
	    struct tm *tm_ptr = localtime(&start_time);

	    if(uptime > (365l * 86400l))
		strftime(
		    start_time_str, sizeof(start_time_str),
		    "%b %d %Y", tm_ptr
		);
	    else
		strftime(
		    start_time_str, sizeof(start_time_str),  
		    "%b %d %H:%M", tm_ptr
		);
	    s = (char *)malloc(
		STRLEN(start_time_str) + STRLEN(uptime_str) + 80
	    );
	    sprintf(
		s,
"Start: %s  Up: %s",
		start_time_str, uptime_str
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(uptime_str);
	}
	else
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"%s: Unable to obtain information.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	}

	return(0);
}

