#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "swserv.h"


static int CmdWhoSortLastConnectedCB(const void *a, const void *b);
static xsw_object_struct **CmdWhoGetConList(int *total_rtn);

int CmdWho(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"who"

#define SPACES		"                                           "

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sort by last connected time callback.
 *
 *	The newer connections will be listed first.
 */
static int CmdWhoSortLastConnectedCB(const void *a, const void *b)
{
	xsw_object_struct	*o1 = *((xsw_object_struct **)a),
				*o2 = *((xsw_object_struct **)b);
	return((int)(o2->last_connected - o1->last_connected));
}

/*
 *	Returns a list of player objects that are connected.
 *
 *	The returned list must be deleted by the calling function, but
 *	not each object.
 */
static xsw_object_struct **CmdWhoGetConList(int *total_rtn)
{
	int i, n, obj_num, total = 0;
	xsw_object_struct *obj, **list = NULL;
	sws_connection_struct *con;

	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if((con->socket < 0) ||
	       (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    )
		continue;

	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
		continue;

	    obj = xsw_object[obj_num];

	    n = total;
	    total++;
	    list = DBReallocList(list, total);
	    if(list == NULL)
	    {
		total = 0;
		break;
	    }

	    list[n] = obj;
	}

	if(total_rtn != NULL)
	    *total_rtn = total;

	return(list);
}


int CmdWho(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char *sndbuf, num_str[40];

	/* Print all connections? */
	if(STRISEMPTY(arg))
	{
	    int i, n, total, obj_num, players_connected = 0;
	    const char *name;
	    char *s;
	    xsw_object_struct *obj, **list;
	    sws_connection_struct *con;

	    /* Get the list of connected player objects */
	    list = CmdWhoGetConList(&total);

	    /* Sort the list */
	    if(list != NULL)
		qsort(
		    list, total,
		    sizeof(xsw_object_struct *),
		    CmdWhoSortLastConnectedCB
		);

	    /* Print the list of players that are connected */
	    for(i = 0; i < total; i++)
	    {
		obj = list[i];
		if(obj == NULL)
		    continue;

		obj_num = DBGetObjectNumByPtr(obj);
		if(obj_num < 0)
		    continue;

		n = ConGetByObject(obj_num);
		if(n < 0)
		    continue;

		con = connection[n];

		name = obj->name;
		sndbuf = (char *)malloc(
		    STRLEN(name) + STRLEN(obj->empire) +
		    40 + 40 + 40 +
		    STRLEN(SPACES) + 80
		);

		/* Player Name */
		s = sndbuf;
		*s = '\0';
		strcat(s, name);
		strcat(s, SPACES);

		/* Empire */
		s = sndbuf + 24;
		*s = '\0';
		strcat(s, obj->empire);
		strcat(s, SPACES);

		/* Client Type Code */
		s = sndbuf + 34;
		*s = '\0'; 
		sprintf(num_str, "%i", con->client_type);
		strcat(s, num_str);
		strcat(s, SPACES);

		/* Connected Time */
		s = sndbuf + 44;
		*s = '\0';
		strcat(s, StringFormatTimePeriod(cur_sec - con->contime));
		strcat(s, SPACES);

		/* Is Guest */
		if(con->flags & SWS_CON_FLAG_IS_GUEST)
		{
		    s = sndbuf + 68;
		    *s = '\0';
		    strcat(s, "Guest");
		}

		/* Print */
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);

		players_connected++;
	    }

	    /* Print footer */
	    sndbuf = (char *)malloc(40 + 40 + 80);
	    sprintf(
		sndbuf,
"*** %i %s Connected ***",
		players_connected,
		(players_connected == 1) ? "Player" : "Players"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    free(list);
	}
	/* Print all connections with formal names and host names */
	else if(!strcasecmp(arg, "*"))
	{
	    int i, n, obj_num, total, players_connected = 0;
	    char *s;
	    const char *name;
	    xsw_object_struct *obj, **list;
	    sws_connection_struct *con;

	    /* Get the list of connected player objects */
	    list = CmdWhoGetConList(&total);

	    /* Sort the list */
	    if(list != NULL)
		qsort(
		    list, total,
		    sizeof(xsw_object_struct *),
		    CmdWhoSortLastConnectedCB
		);

	    /* Print the list of players that are connected */
	    for(i = 0; i < total; i++)
	    {
		obj = list[i];
		if(obj == NULL)
		    continue;

		obj_num = DBGetObjectNumByPtr(obj);
		if(obj_num < 0)
		    continue;

		n = ConGetByObject(obj_num);
		if(n < 0)
		    continue;

		con = connection[n];

		name = DBGetFormalNameStr(obj_num);
		if(name == NULL)
		    name = obj->name;

		sndbuf = (char *)malloc(
		    STRLEN(name) + STRLEN(obj->empire) +
		    40 + STRLEN(con->conhost) + 40 +
		    STRLEN(SPACES) + 80
		);

		/* Player Name */
		s = sndbuf;
		*s = '\0';
		strcat(s, name);
		strcat(s, SPACES);

		/* Empire */
		s = sndbuf + 24;
		*s = '\0';
		strcat(s, obj->empire);
		strcat(s, SPACES);

		/* Client Type Code */
		s = sndbuf + 34;
		*s = '\0';
		sprintf(num_str, "%i", con->client_type);
		strcat(s, num_str);
		strcat(s, SPACES);

		/* Host */
		s = sndbuf + 44;
		*s = '\0';
		strcat(s, (con->conhost != NULL) ? con->conhost : "");
		strcat(s, SPACES);

		/* Is Guest */
		if(con->flags & SWS_CON_FLAG_IS_GUEST)
		{
		    s = sndbuf + 68;
		    *s = '\0';
		    strcat(s, "Guest");
		}

		/* Print */
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);

		players_connected++;
	    }

	    /* Print footer */
	    sndbuf = (char *)malloc(40 + 40 + 80);
	    sprintf(
		sndbuf,
"*** %i %s Connected ***",
		players_connected,
		(players_connected == 1) ? "Player" : "Players"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    free(list);
	}
	/* Print a specific player */
	else
	{
	    int	tar_con_num,
		obj_num = UNVMatchObjectByName(
		    xsw_object, total_objects,
		    arg, SW_OBJ_TYPE_PLAYER
		);
	    xsw_object_struct *obj;

	    if(DBIsObjectGarbage(obj_num))
	    {
		sndbuf = (char *)malloc(STRLEN(arg) + 80);
		sprintf(
		    sndbuf,
"%s: No such player.",
		    arg
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		return(-1);
	    }

	    obj = xsw_object[obj_num];

	    /* Get first connection that references the object */
	    tar_con_num = ConGetByObject(obj_num);
	    if(tar_con_num > -1)
	    {
		char *s;
		sws_connection_struct *con = connection[tar_con_num];

		sndbuf = (char *)malloc(
		    STRLEN(obj->name) + STRLEN(obj->empire) +
		    40 + 40 + 40 +
		    STRLEN(SPACES) + 80
		);

		/* Player Name */
		s = sndbuf;
		*s = '\0';
		strcat(s, obj->name);
		strcat(s, SPACES);

		/* Empire */
		s = sndbuf + 24;
		*s = '\0';
		strcat(s, obj->empire);
		strcat(s, SPACES);

		/* Client Type Code */
		s = sndbuf + 34;
		*s = '\0';
		sprintf(num_str, "%i", con->client_type);
		strcat(s, num_str);
		strcat(s, SPACES);

		/* Connected Time */
		s = sndbuf + 44;
		*s = '\0';
		strcat(s, StringFormatTimePeriod(cur_sec - con->contime));
		strcat(s, SPACES);

		/* Is Guest */
		if(con->flags & SWS_CON_FLAG_IS_GUEST)
		{
		    s = sndbuf + 68;
		    *s = '\0';
		    strcat(s, "Guest");
		}

		/* Print */
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
	    }
	    else
	    {
		sndbuf = (char *)malloc(
		    STRLEN(obj->name) + 80
		);
		sprintf(
		    sndbuf,
"%s: Not connected.",
		    obj->name
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		return(-1);
	    }
	}

	return(0);
}
