#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/imgrefs.h"
#include "../include/swnetcodes.h"

#include "effects.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int EffectsCreateDebrisObject(
	const int obj_num, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void EffectsCreateExplosionSmall(
	const int obj_num, xsw_object_struct *obj,
        long sect_x, long sect_y, long sect_z,
        float x, float y, float z,
        float heading, float pitch, float bank
);
void EffectsCreateExplosionMedium(
	const int obj_num, xsw_object_struct *obj,
        long sect_x, long sect_y, long sect_z,
        float x, float y, float z,
        float heading, float pitch, float bank
);
void EffectsCreateExplosionLarge(
	const int obj_num, xsw_object_struct *obj,
        long sect_x, long sect_y, long sect_z,
        float x, float y, float z,
        float heading, float pitch, float bank
);
void EffectsCreateShockwave(
	const int obj_num, xsw_object_struct *obj,
	long size_min, long size_max,
	float damage_hull, float damage_shield
);

void EffectsSetObjectOnFire(
	const int obj_num, xsw_object_struct *obj,
	const char is_on_fire
);
void EffectsSetObjectSmokeTrail(
	const int obj_num, xsw_object_struct *obj,
	const char has_smoke_trail
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a debris object relative to the specified object.
 *
 *	The debris OPM to be used will be determined by the size and
 *	other attributes of the specified object.
 */
int EffectsCreateDebrisObject(
	const int obj_num, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int opm_num, debris_num;
	char *debris_name;
	xsw_object_struct *opm_ptr;
	sw_legend_struct *legend = &sw_legend;

	if(obj == NULL)
	    return(-1);

	/* Get the appropriate debris OPM based on the specified
	 * object's size
	 */
	if(obj->size < 40l)
	    opm_num = sws_std_opm.debris_50_num;
	else if(obj->size < 90l)
	    opm_num = sws_std_opm.debris_100_num;
	else if(obj->size < 140l)
	    opm_num = sws_std_opm.debris_150_num;
	else if(obj->size < 190l)
	    opm_num = sws_std_opm.debris_200_num;
	else if(obj->size < 240l)
	    opm_num = sws_std_opm.debris_250_num;
	else
	    opm_num = sws_std_opm.debris_300_num;
	if(opm_num < 0)
	    opm_num = sws_std_opm.debris_num;
	opm_ptr = OPMIsGarbage(opm_num) ? NULL : opm[opm_num];
	if(opm_ptr == NULL)
	    return(-1);

	debris_name = (char *)malloc(STRLEN(obj->name) + 80);
	if(debris_name == NULL)
	    return(-3);
	if(!STRISEMPTY(obj->name))
	    sprintf(
		debris_name,
"Debris of %s",
		obj->name
	    );
	else
	    strcpy(debris_name, "Debris");

	debris_num = DBCreateObjectByOPMPtr(
	    opm_ptr,
	    debris_name,
	    SW_OBJ_TYPE_STATIC,
	    legend->lost_found_owner,
	    obj->sect_x, obj->sect_y, obj->sect_z,
	    obj->x, obj->y, obj->z,
	    obj->heading, obj->pitch, 0.0f * PI,
	    cur_ms, cur_sec, tc
	);
	if(!DBIsObjectGarbage(debris_num))
	{
	    xsw_object_struct *debris = xsw_object[debris_num];

	    strncpy(  
		debris->empire, obj->empire,
		sizeof(debris->empire)
	    );
	    debris->empire[sizeof(debris->empire) - 1] = '\0';

	    NetSendCreateObject(-1, debris_num);  
	    NetSendObjectName(-1, debris_num);
	}

	free(debris_name);

	return(debris_num);
}


/*
 *	Creates a small explosion.
 */
void EffectsCreateExplosionSmall(
	const int obj_num, xsw_object_struct *obj,
        long sect_x, long sect_y, long sect_z,
        float x, float y, float z,
        float heading, float pitch, float bank
)
{
	if(obj == NULL)
	    return;

	NetSendCreateEffect(
	    -1,
	    0,
	    IMGREF_EXPLOSION_SMALL,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    64l, 64l, 1200l,
	    obj_num
	);
}

/*
 *	Creates a medium explosion.
 */
void EffectsCreateExplosionMedium(
	const int obj_num, xsw_object_struct *obj,
        long sect_x, long sect_y, long sect_z,
        float x, float y, float z,
        float heading, float pitch, float bank
)
{
	if(obj == NULL)
	    return;

	NetSendCreateEffect(
	    -1,
	    0,
	    IMGREF_EXPLOSION_MEDIUM,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    200l, 200l, 1200l,
	    obj_num
	);
} 

/*
 *	Creates a large explosion,
 */
void EffectsCreateExplosionLarge(
	const int obj_num, xsw_object_struct *obj,
        long sect_x, long sect_y, long sect_z,
        float x, float y, float z,
        float heading, float pitch, float bank
)
{
	if(obj == NULL)
	    return;

	NetSendCreateEffect(
	    -1,
	    0,
	    IMGREF_EXPLOSION_LARGE,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    350l, 350l, 3000l,
	    obj_num
	);
}

/*
 *	Creates a shockwave.
 */
void EffectsCreateShockwave(
	const int obj_num, xsw_object_struct *obj,
	long size_min, long size_max,
	float damage_hull, float damage_shield
)
{
	if(obj == NULL)
	    return;

	NetSendCreateShockwave(
	    -1,
	    -1,
	    obj->sect_x, obj->sect_y, obj->sect_z,
	    obj->x, obj->y, obj->z,
	    size_min, size_max,
	    3000l, damage_hull, damage_shield,
	    obj_num
	);
}


/*
 *      Sets the object's on fire state.
 */
void EffectsSetObjectOnFire(
	const int obj_num, xsw_object_struct *obj,
	const char is_on_fire
)
{
	if(obj == NULL)
	    return;

	if(is_on_fire)
	{
	    obj->ext_flags |= SWS_OBJ_FLAG_ON_FIRE;
	    NetSendSetObjParticleTrail(
		-1, obj_num, SW_PARTICLE_TRAIL_OP_FIRE_ON,
		0.0f, 0.0f, 0.0f, 0.0f 
	    );
	}
	else
	{
	    obj->ext_flags &= ~SWS_OBJ_FLAG_ON_FIRE;
	    NetSendSetObjParticleTrail(
		-1, obj_num, SW_PARTICLE_TRAIL_OP_FIRE_OFF,
		0.0f, 0.0f, 0.0f, 0.0f
	    );
	}
}

/*
 *      Sets the object's smoke trail state.
 */
void EffectsSetObjectSmokeTrail(
	const int obj_num, xsw_object_struct *obj,
	const char has_smoke_trail
)
{
	if(obj == NULL)
	    return;

	if(has_smoke_trail)
	{
	    obj->ext_flags |= SWS_OBJ_FLAG_SMOKE_TRAIL;
	    NetSendSetObjParticleTrail(
		-1, obj_num, SW_PARTICLE_TRAIL_OP_SMOKE_ON,
		0.0f, 0.0f, 0.0f, 0.0f
	    );
	}
	else
	{
	    obj->ext_flags &= ~SWS_OBJ_FLAG_SMOKE_TRAIL;
	    NetSendSetObjParticleTrail(
		-1, obj_num, SW_PARTICLE_TRAIL_OP_SMOKE_OFF,
		0.0f, 0.0f, 0.0f, 0.0f
	    );
	}
}

