#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/netio.h"
#include "../include/swnetcodes.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "siteban.h"
#include "cmd.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int NetTotalGuestsConnected(void);

int NetManageNewConnections(
	const int socket,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void NetCloseSocket(const int con_num, const int socket);
static void NetLogoutConnection(
	const int con_num,
        const time_t cur_ms, const time_t cur_sec, const float tc 
);
void NetShutdownConnection(
	const int con_num, const int how, const char is_initiator,
        const time_t cur_ms, const time_t cur_sec, const float tc 
);
void NetCloseConnection(
	const int con_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
);

static int NetSendDataIteration(
	const int con_num,
	const char *data,
	const int priority,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendDataToConnection(const int con_num, const char *sndbuf);
void NetSendDataWithInScannerRange(
	const char *sndbuf,
	const long sect_x, const long sect_y, const long sect_z,
	const float x, const float y, const float z,
	const long size
);

void NetManageSend(
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static void NetHandleExtCmd(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void NetManageRecvConnection(
	const int con_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetManageRecv(
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the current number of guest connections.
 */
int NetTotalGuestsConnected(void)
{
	int i, nguests = 0;
	sws_connection_struct *con;

	for(i = 0; i < total_connections; i++)
	{
	    if(!ConIsLoggedIn(i))
		continue;

	    con = connection[i];
	    if(con->flags & SWS_CON_FLAG_IS_GUEST)
		nguests++;
	}

	return(nguests);
}


/*
 *	Checks the listening socket for incoming connections and
 *	processes any new connections.
 *
 *	Returns non-zero on error.
 */
int NetManageNewConnections(
	const int socket,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, n, con_num;
	char *sndbuf;
	sws_connection_struct *con;

	int sin_size;
	int new_socket;
	struct sockaddr_in foreign_addr;
	siteban_ip_union ip;

	/* No incoming connections? */
	if(!NetIsSocketReadable(socket, NULL))
	    return(0);

	/* Answer new connection */
	sin_size = sizeof(struct sockaddr_in);
	new_socket = accept(
	    socket,
	    (struct sockaddr *)&foreign_addr,
	    (socklen_t *)&sin_size
	);
	if(new_socket == -1)
	    return(-1);

	/* Set socket nonblocking */
	if(fcntl(new_socket, F_SETFL, O_NONBLOCK) == -1)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(sysparm.log_errors && (error_msg != NULL))
	    {
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(
		    40 + STRLEN(error_msg) + 80
		);
		sprintf(
		    s,
"New connection: Set socket O_NONBLOCK error: Socket %i: %s.",
		    new_socket,
		    error_msg
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	    free(error_msg);
	}


	/* Get IP (in nbo) */
	ip.whole = foreign_addr.sin_addr.s_addr;

	/* Is this IP banned? */
	if(SiteBanIsBanned(&ip, 0))
	{
	    sndbuf = (char *)malloc(
		40 + 40 + 40 + 40 + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%i Your address %i.%i.%i.%i has been banned.\n",
		CS_CODE_LIVEMESSAGE,
		ip.part_u8[0],
		ip.part_u8[1],
		ip.part_u8[2],
		ip.part_u8[3]
	    );
	    NetSend(new_socket, sndbuf, strlen(sndbuf));
	    free(sndbuf);

	    sndbuf = (char *)malloc(40 + 80);
	    sprintf(
		sndbuf,
"%i\n",
		CS_CODE_LOGOUT  
	    );
	    NetSend(new_socket, sndbuf, strlen(sndbuf));
 	    free(sndbuf);

	    /* Close socket */
	    NetCloseSocket(-1, new_socket);

	    /* Pretend that we did not do anything */
	    return(0);
	}


	/* Do we have room for new connection? */
	n = 0;
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    n++;
	}
	if(n >= sysparm.max_connections)
	{
	    sndbuf = (char *)malloc(40 + 40 + 80);
	    sprintf(
		sndbuf,
"%i Maximum of %i connections reached.\n",
		CS_CODE_LIVEMESSAGE,
		sysparm.max_connections
	    );
	    NetSend(new_socket, sndbuf, strlen(sndbuf));
	    free(sndbuf);

	    sndbuf = (char *)malloc(40 + 80);
	    sprintf(
		sndbuf,
"%i\n",
		CS_CODE_LOGOUT
	    );
	    NetSend(new_socket, sndbuf, strlen(sndbuf));
	    free(sndbuf);

	    /* Close socket */
	    NetCloseSocket(-1, new_socket);

	    return(-3);
	}


	/* Get a new connection descriptor for new connection */
	con_num = ConNew();
	if(ConIsAllocated(con_num))
	{
	    con = connection[con_num];
	}
	else
	{
	    if(sysparm.log_errors)
	    {
		char *s = (char *)malloc(80 + 40);
		sprintf(
		    s,
"New connection: Unable to create a new connection (total = %i)",
		    total_connections
		);
	        LogAppend(fname.log, s);
		free(s);
	    }

	    /* Close socket */
	    NetCloseSocket(-1, new_socket);

	    return(-3);
	}


	/* Reset some statistics for the new connection */
	con->contime = time(NULL);

	con->obj_update_int = SWS_DEF_OBJECT_UPDATE_INT;
	con->obj_update_next = cur_ms;

	con->socket = new_socket;

	free(con->conhost);
	con->conhost = STRDUP(inet_ntoa(foreign_addr.sin_addr));


	/* Set the connection as not having an object, this indicates
	 * that it is connected but not logged in
	 *
	 * Once the connection is logged in then we will give it an
	 * object
	 */
	con->object_num = -1;


	/* Log newly allocated descriptor for connection */
	if(!sysparm.console_quiet)
	    fprintf(
		stdout,
		"Connection %i: Got connection from `%s'.\n",
		con_num,
		(con->conhost != NULL) ? con->conhost : ""
	    );
	if(sysparm.log_net)
	{
	    char *s = (char *)malloc(
		40 + STRLEN(con->conhost) + 80
	    );
	    sprintf(
		s,
"Connection %i: Got connection from `%s'.",
		con_num,
		(con->conhost != NULL) ? con->conhost : ""
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	/* Send version and request for login */
	NetSendVersion(con_num);
	NetSendLogin(con_num);

	return(0);
}


/*
 *	Closes the socket.
 *
 *	This call will block until the socket is closed and report
 *	any errors involving the closing of the socket.
 */
void NetCloseSocket(const int con_num, const int socket)
{
	int error_code;
	const time_t	timeout_dsec = 30l,
			start_time = time(NULL);

	if(socket < 0)
	    return;

	while(close(socket) != 0)
	{
	    error_code = errno;

	    /* Give up if timmed out */
	    if(time(NULL) > (start_time + timeout_dsec))
	    {
		if(sysparm.log_errors)
		{
		    char *msg = (char *)malloc(40 + 40 + 40 + (3 * 80));
		    if(con_num > -1)
			sprintf(
			    msg,
"Connection %i:\
 Error closing socket: Socket %i:\
 Timmed out (waited more than %ld seconds).",
			    con_num,
			    socket,
			    timeout_dsec
			);
		    else
			sprintf(
			    msg,
"New connection:\
 Error closing socket %i:\
 Timmed out (waited more than %ld seconds).",
			    socket,
			    timeout_dsec
			);
		    LogAppend(fname.log, msg);
 		    free(msg);
		}
		break;
	    }

	    if(error_code == EAGAIN)
	    {
		sleep(1);
		continue;
	    }
	    else
	    {
		char *error_msg = STRDUP(strerror(error_code));
		if(error_msg != NULL)
		{
		    *error_msg = toupper(*error_msg);
		    if(con_num > -1)
			fprintf(
			    stderr,
"Connection %i: Close socket %i error: %s.\n",
			    con_num,
			    socket,
			    error_msg
		        );
		    else
			fprintf(
			    stderr,
"New connection: Close socket %i error: %s.\n",
			    socket,
			    error_msg
			);

		    if(sysparm.log_errors)
		    {
			char *msg = (char *)malloc(
			    40 + 40 + STRLEN(error_msg) + 80
			);
			if(con_num > -1)
			    sprintf(
			        msg,
"Connection %i: Close socket %i error: %s.",
			        con_num,
			        socket,
			        error_msg
			    );
			else
			    sprintf(
				msg,
"New connection: Close socket %i error: %s.",
				socket,
				error_msg
			    );
			LogAppend(fname.log, msg);
			free(msg);
		    }

		    free(error_msg);
		}
		break;
	    }
	}
}


/*
 *	Notifies other connections about this connection being logged
 *	out and its socket being shut down.
 *
 *	Logs this connection as disconnecting.
 *
 *	Sets post-logout values to the connection's object.
 */
static void NetLogoutConnection(
	const int con_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num;
	sws_connection_struct *con;

	/* Notify console of disconnect as needed */
	if(!sysparm.console_quiet && ConIsConnected(con_num))
	{
	    con = connection[con_num];
	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
		fprintf(
		    stdout,
"Connection %i: Shutting down socket...\n",
		    con_num
		);
	    else
		fprintf(
		    stdout,
"Connection %i: %s logged out and shutting down socket...\n",
		    con_num,
		    DBGetFormalNameStr(obj_num)
		);
	}

	/* Log logout and shutting down of connection? */
	if(sysparm.log_net && ConIsConnected(con_num))
	{
	    char *s;

	    con = connection[con_num];
	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
	    {
		s = (char *)malloc(40 + 80);
		sprintf(
		    s,
"Connection %i: Shutting down socket...",
		    con_num
		);
	    }
	    else
	    {
		const char *name = DBGetFormalNameStr(obj_num);
		s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"Connection %i: %s logged out and shutting down socket...",
		    con_num,
		    name
		);
	    }
	    LogAppend(fname.log, s);
	    free(s);
	}

	if(!ConIsConnected(con_num))
	    return;

	con = connection[con_num];

	/* Set the post-logout values to the connection's object and
	 * other connections if the connection is logged in (still
	 * has an object)
	 *
	 * Note that this function can be called from DBRecycleObject()
	 * with con->object_num set to -1 so this check is to prevent
	 * a recursive call to DBRecycleObject()
	 */
	obj_num = con->object_num;
	if(!DBIsObjectGarbage(obj_num))
	{
	    int i;
	    xsw_object_struct   *obj = xsw_object[obj_num],
				*tar_obj;

	    /* Notify all other connections of disconnect? */
	    if(sysparm.con_notify)
	    {
		char    *s = strdup("%name has disconnected."),
			*mesg = strsub(s, "%name", obj->name);
		free(s);
		sws_connection_struct *con2;

		for(i = 0; i < total_connections; i++)
		{
		    con2 = connection[i];
		    if(con2 == NULL)
			continue;

		    if(i == con_num)
			continue;

		    if((con2->socket < 0) ||
		       (con2->flags & SWS_CON_FLAG_CLOSE_WAIT)
		    )
			continue;

		    if(DBIsObjectGarbage(con2->object_num))
			continue;

		    NetSendMessage(i, mesg);
		}

		free(mesg);
	    }

	    /* Mark this object as not connected */
	    obj->ext_flags &= ~SWS_OBJ_FLAG_CONNECTED;

	    /* Set the last connected time on this object */
	    obj->last_connected = cur_sec;

	    /* Mark disconnceted players hidden at home objects? */
	    if(sysparm.hide_players_home)
	    {
		/* Iterate through the objects list and check if this
		 * object is at a home object
		 */
		for(i = 0; i < total_objects; i++)
		{
		    tar_obj = xsw_object[i];
		    if(tar_obj == NULL)
			continue;

		    /* Skip non-HOME objects */
		    if(tar_obj->type != SW_OBJ_TYPE_HOME)
			continue;

		    /* In contact with this home object? */
		    if(SWMath3DInContactPtr(tar_obj, obj))
		    {
			/* Mark it hidden from connections */
			obj->ext_flags |= SWS_OBJ_FLAG_HIDEFROMCON;
			break;
		    }
		}
	    }

	    /* If this object has been hidden then notify all
	     * connections that it has been recycled so that they will
	     * no longer reference it locally
	     */
	    if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
	    {
		for(i = 0; i < total_connections; i++)
		{
		    if(i == con_num)
			continue;

		    NetSendRecycleObject(i, obj_num);
		}
	    }

	    /* Set the object's throttle to 0 so that it does not
	     * consume antimatter while disconnected
	     */
	    obj->throttle = 0.0f;
	}

	/* Unset the connection's object now, marking the connection
	 * as no longer being logged in (logged out) so that subsequent
	 * calls below do not logout this connection's object again
	 */
	con->object_num = -1;

	con->flags &= ~SWS_CON_FLAG_FSECT_WAIT;
	con->flags &= ~SWS_CON_FLAG_FPOSE_WAIT;

	/* Was this a guest? */
	if((con->flags & SWS_CON_FLAG_IS_GUEST) &&
	   !DBIsObjectGarbage(obj_num)
	)
	{
	    /* Recycle this object since guests are not to be
	     * saved after they disconnect
	     */
	    NetSendRecycleObject(-1, obj_num);
	    DBRecycleObject(obj_num, cur_ms, cur_sec, tc);
	}
}

/*
 *	Logs out the connection and shuts down the connection's socket.
 *
 *	Unsets the connection's object (logging it out).
 *
 *	Shuts down the connection's socket and marks the connection as
 *	CLOSE_WAIT.
 */
void NetShutdownConnection(
	const int con_num, const int how, const char is_initiator,
        const time_t cur_ms, const time_t cur_sec, const float tc 
)
{
	int lhow;
	sws_connection_struct *con;

	if(!ConIsAllocated(con_num))
	    return;

	con = connection[con_num];

	/* Mark this connection as the initiator of the disconnect? */
	if(is_initiator)
	    con->flags |= SWS_CON_FLAG_DISCON_INITIATOR;

	/* Mark the connection's socket as CLOSE_WAIT, it is important
	 * to do this now before any subsequent calls or else they
	 * will attempt to call this function to shut down this
	 * connection and cause infinate recursion
	 */
	con->flags |= SWS_CON_FLAG_CLOSE_WAIT;

	/* Logout this connection */
	NetLogoutConnection(con_num, cur_ms, cur_sec, tc);

	/* No longer connected from the logout? */
	if(!ConIsConnected(con_num))
	    return;

	con = connection[con_num];

	/* Shutdown the socket */
	switch(how)
	{
	  case SWS_CON_SHUTDOWN_WR:
	    lhow = SHUT_WR;
	    break;
	  case SWS_CON_SHUTDOWN_RDWR:
	    lhow = SHUT_RDWR;
	    break;
	  default:
	    lhow = SHUT_RD;
	    break;
	}
	if(shutdown(con->socket, lhow))
	{
	    int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		*error_msg = toupper(*error_msg);

		fprintf(
		    stderr,
"Connection %i: Shutdown socket error: Socket %i: %s.\n",
		    con_num,
		    con->socket,
		    error_msg
		);

		if(sysparm.log_errors)
		{
		    char *s = (char *)malloc(
			40 + 40 + STRLEN(error_msg) + 80
		    );
		    sprintf(
			s,
"Connection %i: Shutdown socket %i error: %s.",
			con_num,
			con->socket,
			error_msg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
	    }
	    free(error_msg);
	}

	/* Schedual stats export to now in order to export the
	 * connections list now
	 */
	next.stats_export = cur_sec;
}

/*
 *	Shuts down the connection's socket, logs out the connection,
 *	closes the connection's socket, and finally recycles the
 *	connection.
 */
void NetCloseConnection(
	const int con_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	sws_connection_struct *con;

	if(!ConIsAllocated(con_num))
	    return;

	con = connection[con_num];

	/* Is the connection's socket connected? */
	if(con->socket > -1)
	{
	    /* Shut down the socket only if it is not already on
	     * CLOSE_WAIT
	     */
	    if(!(con->flags & SWS_CON_FLAG_CLOSE_WAIT))
		NetShutdownConnection(
		    con_num, SWS_CON_SHUTDOWN_RDWR, 1,
		    cur_ms, cur_sec, tc
		);

	    /* Close the socket (may be forcably if it was already on
	     * CLOSE_WAIT)
	     */
	    NetCloseSocket(con_num, con->socket);
	    con->socket = -1;

	    /* Notify console of disconnect as needed */
	    if(!sysparm.console_quiet)
		fprintf(
		    stdout,
"Connection %i: Disconnected.\n",
		    con_num
		);

	    /* Log disconnect */
	    if(sysparm.log_net)
	    {
		char *s = (char *)malloc(40 + 80);
		sprintf(
		    s,
"Connection %i: Disconnected.",
		    con_num
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	}

	/* Recycle this connection */
	ConRecycle(con_num);
}


/*
 *	Sends the data to the connection.
 *
 *	If there are any queued buffers on this connection then they
 *	will be sent first before the specified data.
 *
 *	The priority can be 1 or 2 (where 1 is the highest).
 *
 *	A priority of 1 means to attempt to send the data, if it is
 *	not possible then queue it, in which case it will be attempted
 *	to be sent in a future call to this function.
 *
 *	A priority of 2 means to attempt to send the data, if it is
 *	not possible then just discard it.
 *
 *	Returns:
 *
 *	0	Data sent successfully.
 *	-1	General error.
 *	-2	Bad value.
 *	-3	Systems error.
 *	-4	Interrupted by signal.
 *	-5	Send failed and the connection has been closed by this
 *		function.
 */
static int NetSendDataIteration(
	const int con_num,
	const char *data,
	const int priority,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int	s, status, buf_len,
		bytes_sent = 0,
		cur_qbufs;
	sws_connection_struct *con;

	if(data == NULL)
	    return(-2);

	if(*data == '\0')
	    return(0);

	/* Not connected? */
	if(!ConIsConnected(con_num))
	    return(-1);

	con = connection[con_num];

	/* Do not send if the connection is waiting to close */
	if(con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    return(-1);

	s = con->socket;

	/* Get length of data to be sent */
	buf_len = strlen(data);
	if(buf_len >= CS_DATA_MAX_LEN)
	    return(-2);

	/* Send any queued buffers first */
	cur_qbufs = con->conbuf_items;
	if(cur_qbufs > 0)
	{
	    int	i,
		total_qbufs = SWS_MAX_CON_BUF_ITEMS;
	    sws_conbuf_struct *qbuf = con->conbuf;

	    if(cur_qbufs > total_qbufs) 
		cur_qbufs = total_qbufs;

	    /* Send each queued line */
	    for(i = 0; i < cur_qbufs; i++)
	    {
		status = NetSend(
		    s,
		    qbuf[i].buffer, 
		    strlen(qbuf[i].buffer)
		);
		if(status > 0)
		{
		    bytes_sent += status;
		}
		else
		{
		    const int error_code = errno;
		    switch(error_code)
		    {
		      case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
		      case EWOULDBLOCK:
#endif
		      case ENOBUFS:	/* Out of memory, try again */
			break;

		      case EPIPE:	/* Connection has been shut down */
			con->flags |= SWS_CON_FLAG_CLOSE_WAIT;
			NetCloseConnection(con_num, cur_ms, cur_sec, tc);
			return(-5);
			break;

		      case EBADF:	/* Invalid descriptor */
		      case ENOTSOCK:	/* Descriptor is not a socket */
			con->errors_sent++;
			NetCloseConnection(con_num, cur_ms, cur_sec, tc);
			return(-5);
			break;

		      case EINTR:	/* Signal occured */
			con->errors_sent++;
			break;

		      case ENOMEM:	/* Out of memory */
			con->errors_sent++;
			break;

		      default:
			con->errors_sent++;
			break;
		    }
		    break;
		}
	    }
	    if(i < cur_qbufs)
	    {
		/* Not able to send out all queued buffers, shift the
		 * remaining queued buffers
		 */
		if(i > 0)
		{
		    int n, k;

		    for(n = 0, k = i;
			k < total_qbufs;
			n++, k++
		    )
		    {
		        if(k >= cur_qbufs)
			    break;

		        memcpy(
			    qbuf[n].buffer,		/* Target */
			    qbuf[k].buffer,		/* Source */
			    CS_DATA_MAX_LEN
		        );
		    }

		    /* Set new number of queued buffers */
		    con->conbuf_items = n;
		}
	    }
	    else
	    {
		/* All queued buffers sent, reset number of queued
		 * buffers on this connection
		 */
		con->conbuf_items = 0;
	    }
	}


	/* Send out the specified data by priority
	 *
	 *      0	Highest, force send for up to 5 seconds (not
	 *		supported yet)
	 *      1	Moderate, try to send. If cannot then put into
	 *		connection's buffer
	 *	2	Lowest, try to send, If cannot then forget it
	 */
	switch(priority)
	{
	  case 0:
	    fprintf(
		stderr,
"NetSendDataIteration(): Priority 0 not supported.\n"
	    );
	    break;

	  case 1:	/* High priority */
	    status = NetSend(s, data, buf_len);
	    if(status > 0)
	    {
		bytes_sent += status;
	    }
	    else
	    {
		const int error_code = errno;
		switch(error_code)
		{
		  case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
		  case EWOULDBLOCK:
#endif
		  case ENOBUFS:		/* Out of memory, try again */
		    /* Queue this data for later send */
		    cur_qbufs = con->conbuf_items;
		    if(cur_qbufs < SWS_MAX_CON_BUF_ITEMS)
		    {
			strcpy(
			    con->conbuf[cur_qbufs].buffer,
			    data
			);
			cur_qbufs++;
			con->conbuf_items = cur_qbufs;
		    }
		    else
		    {
#if 0
			/* Overflow, queued buffers at maximum,
			 * unable to queue this data segment
			 */
			if(sysparm.log_errors)
			{
			    char *s = (char *)malloc(
				40 + 40 + 80
			    );
			    sprintf(
				s,
"Connection %i: Unable to queue data to send to connection,\
 queued buffers at maximum %i.\n",
				con_num,
				SWS_MAX_CON_BUF_ITEMS
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}
#endif
		    }
		    break;

		  case EPIPE:	/* Connection has been shut down */
		    con->flags |= SWS_CON_FLAG_CLOSE_WAIT;
		    NetCloseConnection(con_num, cur_ms, cur_sec, tc);
		    return(-5);
		    break;

		  case EBADF:	/* Invalid descriptor */
		  case ENOTSOCK:	/* Descriptor is not a socket */
		    con->errors_sent++;
		    NetCloseConnection(con_num, cur_ms, cur_sec, tc);
		    return(-5);
		    break;

		  case EINTR:	/* Signal occured */
		    con->errors_sent++;
		    return(-4);
		    break;

		  case ENOMEM:	/* Out of memory */
		    con->errors_sent++;
		    return(-3);
		    break;

		  default:
		    con->errors_sent++;
		    return(-1);
		    break;
		}
		break;
	    }
	    break;

	  default:	/* Low priority */
	    status = NetSend(s, data, buf_len);
	    if(status > 0)
	    {
		bytes_sent += status;
	    }
	    else
	    {
		const int error_code = errno;
		switch(error_code)
		{
		  case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
		  case EWOULDBLOCK:  
#endif
		  case ENOBUFS:	/* Out of memory, try again */
		    /* Discard if not sendable */
		    break;

		  case EPIPE:	/* Connection has been shut down */
		    con->flags |= SWS_CON_FLAG_CLOSE_WAIT;
		    NetCloseConnection(con_num, cur_ms, cur_sec, tc);
		    return(-5);
		    break;

		  case EBADF:	/* Invalid descriptor */
		  case ENOTSOCK:	/* Descriptor is not a socket */
		    con->errors_sent++;
		    NetCloseConnection(con_num, cur_ms, cur_sec, tc);
		    return(-5);
		    break;

		  case EINTR:	/* Signal occured */
		    con->errors_sent++;
		    return(-4);
		    break;

		  case ENOMEM:	/* Out of memory */
		    con->errors_sent++;
		    return(-3);
		    break;

		  default:
		    con->errors_sent++;
		    return(-1);
		    break;
		}
		break;
	    }
	    break;
	}

	/* Update number of bytes sent on that connection */
	con->bytes_sent += bytes_sent;

	return(0);
}

/*
 *      Sends sndbuf to con_num or to all connections if
 *      con_num is -1.
 * 
 *      The priority will be 1 if the con_num is not -1
 *      and the priority will be 2 if the con_num is -1.
 */
void NetSendDataToConnection(const int con_num, const char *sndbuf)
{
	const time_t	cur_ms = cur_millitime,
			cur_sec = cur_systime;
	const float tc = time_compensation;

	/* Send to all connections? */
	if(con_num == -1)
	{
	    int i;
	    sws_connection_struct *con;

	    for(i = 0; i < total_connections; i++)
	    {
		con = connection[i];
		if(con == NULL)
		    continue;

		if(con->socket < 0)
		    continue;

		NetSendDataIteration(
		    i,
		    sndbuf,
		    2,		/* Send priority bulk */
		    cur_ms, cur_sec, tc
		);
	    }
	}  
	else
	{
	    /* Send to specific connection */
	    NetSendDataIteration(
		con_num,
		sndbuf,
		1,		/* Send priority queue as needed */
		cur_ms, cur_sec, tc
	    );
	}
}

/*
 *	Sends the data to all connections that have objects that are
 *	within scanner range of the specified position and size.
 */
void NetSendDataWithInScannerRange(
	const char *sndbuf, 
	const long sect_x, const long sect_y, const long sect_z,
	const float x, const float y, const float z,
	const long size
)
{
        const time_t	cur_ms = cur_millitime,
                        cur_sec = cur_systime;
        const float tc = time_compensation;
	int i, obj_num;
	float d, size_ru = MAX(size, 0l) / 1000.0f;
	xsw_object_struct *obj;
	sws_connection_struct *con;

	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
		continue;

	    obj = xsw_object[obj_num];

	    /* In a different sector? */
	    if(!SWMath3DInSectorPtr(
		obj, sect_x, sect_y, sect_z
	    ))
		continue;

	    /* Out of scanner range? */
	    d = SWMath3DDistance(
		x - obj->x,
		y - obj->y,
		z - obj->z
	    );
	    if((d - size_ru) > obj->scanner_range)
		continue;

	    NetSendDataIteration(
		i,
		sndbuf,
		1,		/* Send priority queue as needed */
		cur_ms, cur_sec, tc
	    );
	}
}


/*
 *	Sends out schedualed information such as object poses, values,
 *	maximums, weapons, etc to all connections.
 */
void NetManageSend(
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int con_num, obj_num, obj_count;
	sw_flags_t ext_flags;
	xsw_object_struct *obj, *con_obj;
	sws_connection_struct *con;


	/* Begin sending object poses */

	/* Iterate through connections list */
	for(con_num = 0; con_num < total_connections; con_num++)
	{
	    con = connection[con_num];
	    if(con == NULL)
		continue;

	    if((con->socket < 0) ||
	       (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    )
		continue;

	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
		continue;

	    con_obj = xsw_object[obj_num];

	    /* Connection due to recieve standard object updates? */
	    if(con->obj_update_next <= cur_ms)
	    {
		/* Iterate through the objects list */
		for(obj_count = 0; obj_count < total_objects; obj_count++)
		{
		    obj = xsw_object[obj_count];
		    if(obj == NULL)
			continue;

		    /* Check if object is valid and in range */
		    if(!SWMath3DInRangePtr(
			con_obj, obj,
			con_obj->scanner_range
		    ))
			continue;

		    ext_flags = obj->ext_flags;

		    /* Hidden? */
		    if(ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* If object is not owned by the connection,
		     * then check the visibility of the object
		     */
		    if(obj->owner != obj_num)
		    {
			if((DBGetObjectVisibilityPtr(obj) <= 0.0f) &&
			   !(ext_flags & SWS_OBJ_FLAG_ON_FIRE) &&
                           !(ext_flags & SWS_OBJ_FLAG_SMOKE_TRAIL)
			)
			    continue;
		    }

		    /* Send object position */
		    NetSendObjectPose(con_num, obj_count);
		}

		/* Schedual the next object update for this connection */ 
		con->obj_update_next = cur_ms + con->obj_update_int;
	    }
	}


	/* Time to send standard object values? */
	if(next.object_values <= cur_ms)
	{
	    /* Begin sending object values */

	    /* Iterate through the connections list */
	    for(con_num = 0; con_num < total_connections; con_num++)
	    {
		con = connection[con_num];
		if(con == NULL)
		    continue;

		if((con->socket < 0) ||
		   (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
		)
		    continue;

		obj_num = con->object_num;
		if(DBIsObjectGarbage(obj_num))
		    continue;

		con_obj = xsw_object[obj_num];

		/* Iterate through the objects list */
		for(obj_count = 0; obj_count < total_objects; obj_count++)
		{
		    obj = xsw_object[obj_count];
		    if(obj == NULL)
			continue;

		    /* Check if objects are valid and in range */
		    if(!SWMath3DInRangePtr(
			con_obj, obj,
			con_obj->scanner_range
		    ))
			continue;

		    ext_flags = obj->ext_flags;

		    /* Hidden? */
		    if(ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* Not owned objects, check for visibility */
		    if(obj->owner != obj_num)
		    {
			if((DBGetObjectVisibilityPtr(obj) <= 0.0f) &&
                           !(ext_flags & SWS_OBJ_FLAG_ON_FIRE) &&
                           !(ext_flags & SWS_OBJ_FLAG_SMOKE_TRAIL)
			)
			    continue;
		    }

		    /* Send sector */
		    NetSendObjSect(con_num, obj_count, cur_ms, cur_sec, tc);

		    /* Send maximums and values */
		    NetSendObjectMaximums(con_num, obj_count);
		    NetSendObjectValues(con_num, obj_count);

		    /* Send tractor beam lock */
		    if(obj->total_tractored_objects > 0)
		    {
			int i, m = obj->total_tractored_objects;
			for(i = 0; i < m; i++)
			    NetSendTractorBeamLock(
				con_num, obj_count,
				obj->tractored_object[i]
			    );
		    }

		    /* If the object is in an asteroid field then we
		     * need to send scores for damage report
		     */
		    if(obj->loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
			NetSendScore(con_num, obj_count);
		}
	    }

	    /* Schedual next object values send */
	    next.object_values = cur_ms + sysparm.int_object_values;
	}


	/* Time to send weapon values? */
	if(next.need_weapon_values)
	{
	    /* Send weapon values */

	    /* Iterate through connections list */
	    for(con_num = 0; con_num < total_connections; con_num++)
	    {
		con = connection[con_num];
		if(con == NULL)
		    continue;

		if((con->socket < 0) ||
		   (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
		)
		    continue;

		obj_num = con->object_num;
		if(DBIsObjectGarbage(obj_num))
		    continue;

		con_obj = xsw_object[obj_num];

		/* Iterate through the objects list */
		for(obj_count = 0; obj_count < total_objects; obj_count++)
		{
		    obj = xsw_object[obj_count];
		    if(obj == NULL)
			continue;

		    /* Check if objects are valid and in range */
		    if(!SWMath3DInRangePtr(
			con_obj, obj,
			con_obj->scanner_range
		    ))
			continue;

		    ext_flags = obj->ext_flags;

		    /* Hidden? */
		    if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* Not owned objects, check for visibility */
		    if(obj->owner != obj_num)
		    {
			if((DBGetObjectVisibilityPtr(obj) <= 0.0f) &&
                           !(ext_flags & SWS_OBJ_FLAG_ON_FIRE) &&
                           !(ext_flags & SWS_OBJ_FLAG_SMOKE_TRAIL)
			)
			    continue;
		    }

		    /* Send particle trails */
		    NetSendSetObjParticleTrail(
			con_num, obj_count,
			(obj->ext_flags & SWS_OBJ_FLAG_ON_FIRE) ?
			    SW_PARTICLE_TRAIL_OP_FIRE_ON :
			    SW_PARTICLE_TRAIL_OP_FIRE_OFF,
			0.0f, 0.0f * PI, 0.0f * PI, 0.0f * PI
		    );
		    NetSendSetObjParticleTrail(
			con_num, obj_count,
			(obj->ext_flags & SWS_OBJ_FLAG_SMOKE_TRAIL) ?
			    SW_PARTICLE_TRAIL_OP_SMOKE_ON :
			    SW_PARTICLE_TRAIL_OP_SMOKE_OFF,
			0.0f, 0.0f * PI, 0.0f * PI, 0.0f * PI
		    );

		    /* Send weapons values */
		    if(obj->total_weapons > 0)
		    {
			int i, m = obj->total_weapons;
			for(i = 0; i < m; i++)
			    NetSendWeaponValues(
				con_num, obj_count, i
			    );
		    }
		}
	    }

	    /* Reset need next weapon values sent */
	    next.need_weapon_values = 0l;
	}
}


/*
 *	Handles a ShipWars extended command.
 *
 *	This function is called by NetHandleConnection().
 *
 *	Inputs assumed valid.
 */
static void NetHandleExtCmd(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int ext_cmd;
	char *ext_arg;

	/* Get the command */
	ext_cmd = StringGetNetCommand(arg);
	if(ext_cmd < 0)
	    return;

	/* Get the argument */
	ext_arg = STRDUP(StringGetNetArgument(arg));
	if(ext_arg == NULL)
	    return;

	/* Handle the extended command */
	switch(ext_cmd)
	{
	  /* System and enviroment */
	  case SWEXTCMD_SETOCSN:
	    NetHandleSetOCSN(con_num, ext_arg);
	    break;

	  case SWEXTCMD_SET_LEGEND:
	    NetHandleSetLegend(con_num, ext_arg);
	    break;

	  case SWEXTCMD_SETTITLE:
	    NetHandleSetTitle(con_num, ext_arg);
	    break;


	  /* Standards */
	  case SWEXTCMD_STDOBJVALS:
	    NetHandleObjectValues(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_STDOBJMAXS:
	    /* Server does not accept set object maximums */
	    break;
	
	  case SWEXTCMD_STDWEPVALS:
	    /* Server does not accept set weapon values */
	    break;
	    

	  /* Set */
	  case SWEXTCMD_SETOBJNAME:
	    /* Server does not accept set names */
	    break;
   
	  case SWEXTCMD_SETOBJSECT:
	  case SWEXTCMD_SETFOBJSECT:
	    NetHandleObjSect(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETTHROTTLE:
	    NetHandleObjectThrottle(con_num, ext_arg);
	    break;

	  case SWEXTCMD_SETWEAPON:
	    NetHandleSelectWeapon(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETINTERCEPT:
	    NetHandleSetIntercept(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETWEPLOCK:
	    NetHandleSetWeaponsLock(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETSHIELDS:
	    NetHandleSetShields(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETDMGCTL:
	    NetHandleSetDmgCtl(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETCLOAK:
	    NetHandleSetCloak(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETSHIELDVIS:
	    /* Server does not handle this command */
	    break;

	  case SWEXTCMD_SETLIGHTING:
	    NetHandleLighting(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETCHANNEL:
	    NetHandleSetChannel(con_num, ext_arg);
	    break;

	  case SWEXTCMD_SETSCORE:
	    NetHandleScore(con_num, ext_arg);
	    break;

	  case SWEXTCMD_SETENGINE:
	    NetHandleSetEngine(con_num, ext_arg);
	    break;

	  case SWEXTCMD_SET_OBJ_PARTICLE_TRAIL:
	    NetHandleSetObjParticleTrail(con_num, ext_arg);
	    break;

	  case SWEXTCMD_CREATE_EFFECT:
	    NetHandleCreateEffect(con_num, ext_arg);
	    break;

	  case SWEXTCMD_CREATE_SHOCKWAVE:
	    NetHandleCreateShockwave(con_num, ext_arg);
	    break;


	  /* Requests */
	  case SWEXTCMD_REQNAME:
	    NetHandleObjectName(con_num, ext_arg);
	    break;

	  case SWEXTCMD_REQSECT:
	    NetHandleReqObjSect(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;


	  /* Notifies */
	  case SWEXTCMD_NOTIFYHIT:
	    break;

	  case SWEXTCMD_NOTIFYDESTROY:
	    break;


	  /* Actions */
	  case SWEXTCMD_FIREWEAPON:
	    NetHandleFireWeapon(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_TRACTORBEAMLOCK:
	    NetHandleTractorBeamLock(con_num, ext_arg);
	    break;

	  case SWEXTCMD_HAIL:
	    NetHandleHail(con_num, ext_arg);
	    break;

	  case SWEXTCMD_COMMESSAGE:
	    NetHandleComMesg(con_num, ext_arg);
	    break;

	  case SWEXTCMD_WORMHOLEENTER:
	    NetHandleWormHoleEnter(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_ELINKENTER:
	    NetHandleELinkEnter(con_num, ext_arg);
	    break;

	  case SWEXTCMD_WEPDISARM:
	    NetHandleWeaponDisarm(con_num, ext_arg);
	    break;

	  case SWEXTCMD_AUTODESTRUCT:
	    NetHandleAutoDestruct(con_num, ext_arg, cur_ms, cur_sec, tc);
	    break;

	  case SWEXTCMD_SETDESCRIPTION:
	    break;

	  case SWEXTCMD_REQDESCRIPTION:
	    NetHandleObjectDescription(con_num, ext_arg);
	    break;


	  /* Economy */
	  case SWEXTCMD_ECO_REQVALUES:
	    NetHandleEcoGetValues(con_num, ext_arg);
	    break;

	  case SWEXTCMD_ECO_SETVALUES:
	    break;

	  case SWEXTCMD_ECO_SETPRODUCTVALUES:
	    break;

	  case SWEXTCMD_ECO_BUY:
	    NetHandleEcoBuy(con_num, ext_arg);
	    break;

	  case SWEXTCMD_ECO_SELL:
	    NetHandleEcoSell(con_num, ext_arg);
	    break;

	}   

	/* Delete the argument */
	free(ext_arg);
}


/*
 *      Checks and processes incoming network data on
 *      connection con_num.
 *
 *	Warning con_num is assumed to be valid.
 */
static void NetManageRecvConnection(
	const int con_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int socket, bytes_read;

	char recvbuf[CS_DATA_MAX_BACKLOG];
	int recvbuf_cnt;
	char *recvbuf_ptr;

	char workbuf[CS_DATA_MAX_LEN];
	int workbuf_cnt;
	char *workbuf_ptr;

	char arg[CS_DATA_MAX_LEN];
	int command;

	sws_connection_struct *con = connection[con_num];
	if(con == NULL)
	    return;

	/* No socket? */
	socket = con->socket;
	if(socket < 0)
	    return;
 
	/* Recieve incoming data */
	bytes_read = NetRecv(socket, recvbuf, sizeof(recvbuf));
	if(bytes_read == 0)
	{
	    /* recv() == 0 means that the socket has been shutdown()
	     * and all data flushed, in which case the socket should
	     * now be closed
	     */
	    NetCloseConnection(con_num, cur_ms, cur_sec, tc);
	    return;
	}

	/* Recieve error? */
	if(bytes_read < 0)
	{
	    const int error_code = errno;
	    char *error_msg;

	    /* Handle error */
	    switch(error_code)
	    {
	      case EAGAIN:	/* No data available */
#if (EWOULDBLOCK != EAGAIN)
	      case EWOULDBLOCK:
#endif
	      case EINTR:	/* Interrupted by signal */
		return;
		break;

	      case EPIPE:	/* Connection has been shut down */
		con->flags |= SWS_CON_FLAG_CLOSE_WAIT;
		NetCloseConnection(con_num, cur_ms, cur_sec, tc);
		return;
		break;

	      case EBADF:	/* Invalid descriptor */
	      case ECONNRESET:	/* Connection terminated by peer */
	      case EFAULT:	/* Receive buffer out of address space */
	      case ENOTCONN:	/* Descriptor is not connected */
	      case ENOTSOCK:	/* Descriptor is not a socket */
		error_msg = STRDUP(strerror(error_code));
		if((error_msg != NULL) && sysparm.log_errors)
		{
		    char *s = (char *)malloc(
			40 + 40 + STRLEN(error_msg) + 80
		    );
		    *error_msg = toupper(*error_msg);
		    sprintf(
			s,
"Connection %i: Receive data error: Socket %i: %s.",
			con_num, socket, error_msg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		NetCloseConnection(con_num, cur_ms, cur_sec, tc);
		free(error_msg);
		return;
		break;

	      default:
		error_msg = STRDUP(strerror(error_code));
		if((error_msg != NULL) && sysparm.log_errors)
		{
		    char *s = (char *)malloc(
			40 + 40 + STRLEN(error_msg) + 80
		    );
		    *error_msg = toupper(*error_msg);
		    sprintf(
			s,
"Connection %i: Receive data error: Socket %i: %s.",
			con_num, socket, error_msg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		free(error_msg);
		return;
		break;
	    }
	}

	/* bytes_read cannot be greater than CS_DATA_MAX_BACKLOG */
	if(bytes_read > (int)sizeof(recvbuf))
	    bytes_read = (int)sizeof(recvbuf);

	/* Increment bytes_received for this con_num */
	con->bytes_received += bytes_read;


	/* Set null character at the end of recvbuf just to be safe */
	recvbuf[sizeof(recvbuf) - 1] = '\0';
	    
	      
	/* *************************************************************** */
	/* Begin parsing recieve buffer */

	/* Reset recieve buffer count and pointer */
	recvbuf_cnt = 0;
	recvbuf_ptr = recvbuf;

	while(recvbuf_cnt < bytes_read)
	{
	    /* Do not continue to process subsequent received data
	     * if the connection is now on CLOSE_WAIT
	     */
	    if(con->flags & SWS_CON_FLAG_CLOSE_WAIT)
		break;

	    /* Reset work buffer count and pointer */
	    workbuf_cnt = 0;
	    workbuf_ptr = workbuf;

	    /* Fetch data for workbuf */
	    while(1)
	    {
		if(workbuf_cnt >= (int)sizeof(workbuf))
		{
		    /* Increment recvbuf count and ptr to next delimiter */
		    while(recvbuf_cnt < bytes_read)
		    {
			if((*recvbuf_ptr == '\n') ||
			   (*recvbuf_ptr == '\r') ||
			   (*recvbuf_ptr == '\0')
			)
			{
			    recvbuf_cnt++;
			    recvbuf_ptr++;
			    break;
			}
			recvbuf_cnt++;
			recvbuf_ptr++;
		    }
		    /* Null terminating character for workbuf will
		     * be added farther below
		     */
		    break;
		}
		/* End of recieve buffer data reached? */
		else if(recvbuf_cnt >= bytes_read)
		{
		    *workbuf_ptr = '\0';
		    break;
		}
		/* Deliminator in recieve buffer encountered? */
		else if((*recvbuf_ptr == '\n') ||
			(*recvbuf_ptr == '\r') ||
			(*recvbuf_ptr == '\0')
		)
		{   
		    *workbuf_ptr = '\0';

		    recvbuf_cnt++;
		    recvbuf_ptr++;

		    break;
		}

		/* Copy data from recieve buffer to work buffer normally */
		else
		{
		    *workbuf_ptr = *recvbuf_ptr;
		
		    recvbuf_cnt++;
		    recvbuf_ptr++;
			
		    workbuf_cnt++;
		    workbuf_ptr++;
		}
	    }

	    /* Skip if workbuf is empty */
	    if(*workbuf == '\0')
		continue;

	    /* Set null terminating character for workbuf */
	    workbuf[sizeof(workbuf) - 1] = '\0';


	    /* Get the CS message type */
	    command = StringGetNetCommand(workbuf);
	    if(command < 0)
		continue;

	    /* Get the subsequent arguments */
	    strncpy(arg, StringGetNetArgument(workbuf), sizeof(arg));
	    arg[sizeof(arg) - 1] = '\0';

	    /* Handle by the CS message type */
	    switch(command)
	    {
	      case CS_CODE_VERSION:
		NetHandleVersion(con_num, arg);
		break;

	      case CS_CODE_LOGIN:
		NetHandleLogin(con_num, arg, cur_ms, cur_sec, tc);
		break;

	      case CS_CODE_LOGOUT:
		NetHandleLogout(con_num, cur_ms, cur_sec, tc);
		break;

	      case CS_CODE_WHOAMI:
		NetHandleWhoAmI(con_num, arg);
		break;

	      case CS_CODE_REFRESH:
		NetHandleRefresh(con_num, arg);
		break;

	      case CS_CODE_INTERVAL:
		NetHandleSetInterval(con_num, arg);
		break;

	      case CS_CODE_IMAGESET:
		NetHandleSetImageSet(con_num, arg);
		break;

	      case CS_CODE_SOUNDSET:
		NetHandleSetSoundSet(con_num, arg);
		break;

	      case CS_CODE_PLAYSOUND:
		/* server dosen't play sounds */
		break;


	      case CS_CODE_LIVEMESSAGE:
		NetHandleMessage(con_num, arg);
		break;

	      case CS_CODE_SYSMESSAGE:
		NetHandleSysMessage(con_num, arg);
		break;

	      case CS_CODE_LITERALCMD:
		CmdInput(
		    con_num, arg,
		    cur_ms, cur_sec, tc
		);
		break;


	      case CS_CODE_CREATEOBJ:
		/* Server doesn't accept this command */
		break;

	      case CS_CODE_RECYCLEOBJ:
		/* Server dosen't accept this command */
		break;

	      case CS_CODE_POSEOBJ:
		NetHandleObjectPose(con_num, arg, cur_ms, cur_sec, tc);
		break;

	      case CS_CODE_FORCEPOSEOBJ:
		/* Server does not accept this */
		break;

	      case CS_CODE_EXT:
		NetHandleExtCmd(con_num, arg, cur_ms, cur_sec, tc);
		break;

	      /* Unknown command */
	      default:
#if 0
fprintf(
 stderr, "Recieved unsupported command %i.\n", command
);
fprintf(
 stderr, "Raw text: `%s'\n", workbuf
);
#endif
		con->errors_received++;
		break;
	    }

	    /* Connection got closed or deleted? */
	    if(!ConIsConnected(con_num))
		break;
	}
}


/*
 *      Procedure to check all connections for incoming network
 *      data and process the data.
 */
void NetManageRecv(
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i;

	for(i = 0; i < total_connections; i++)
	    NetManageRecvConnection(i, cur_ms, cur_sec, tc);
}                   
