#ifndef NET_H
#define NET_H

#include <sys/types.h>
#include "../include/objects.h"


extern int NetIsSocketWritable(int s, int *error_level);
extern int NetIsSocketReadable(int s, int *error_level);

extern int NetTotalGuestsConnected(void);

extern int NetManageNewConnections(
	const int socket,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

extern void NetCloseSocket(const int con_num, const int socket);
extern void NetShutdownConnection(
	const int con_num, const int how, const char is_initiator,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetCloseConnection(
	const int con_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

extern void NetSendDataToConnection(
	const int con_num, const char *sndbuf
);
extern void NetSendDataWithInScannerRange(
	const char *sndbuf,
	const long sect_x, const long sect_y, const long sect_z,
	const float x, const float y, const float z,
	const long size
);
extern void NetManageSend(
	const time_t cur_ms, const time_t cur_sec, const float tc
);

extern void NetManageRecv(
	const time_t cur_ms, const time_t cur_sec, const float tc
);


extern void NetHandleAutoDestruct(
        const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendAutoDestruct(
	const int con_num, const int obj_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
);

extern void NetHandleSetCloak(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendSetCloak(const int con_num);

extern void NetHandleSetChannel(const int con_num, const char *arg);
extern void NetSendSetChannel(
	const int con_num, const int obj_num,
	const int channel
);

extern void NetHandleComMesg(const int con_num, const char *arg);
#define NET_COM_MSG_STATIC		(1 << 0)
extern void NetSendComMesg(
	const int con_num,
	const int src_obj_num,		/* Sender */
	const int tar_obj_num,		/* Receiver or -1 for general */
	const float bearing,		/* Target to source in radians */
	const int channel,
	const char *msg,
	const unsigned int options	/* Any of NET_COM_MSG_* */
);

extern void NetHandleCreateObject(const int con_num, const char *arg);
extern void NetSendCreateObject(const int con_num, const int obj_num);

extern void NetHandleMessage(const int con_num, const char *msg);
extern void NetSendMessage(const int con_num, const char *msg);
extern void NetHandleSysMessage(const int con_num, const char *msg);
extern void NetSendSysMessage(
	const int con_num,
	const int code,		/* One of CS_SYSMESG_* */
	const char *msg
);

extern void NetHandleSetDmgCtl(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendSetDmgCtl(const int con_num, const int obj_num);

extern void NetHandleEcoGetValues(int con_num, const char *arg);
extern void NetSendEcoSetValues(int con_num, int obj_num);
extern void NetSendEcoSetProductValues(
	int con_num,
	int obj_num,
	int product_num
);
extern void NetHandleEcoBuy(int con_num, const char *arg);
extern void NetHandleEcoSell(int con_num, const char *arg);

extern void NetHandleSetImageSet(int con_num, const char *arg);
extern void NetHandleSetSoundSet(int con_num, const char *arg);
extern void NetHandleSetOCSN(int con_num, const char *arg);
extern void NetSendSetImageSet(int con_num, const char *filename);
extern void NetSendSetSoundSet(int con_num, const char *filename);
extern void NetSendSetOCSN(int con_num, const char *filename);

extern void NetHandleELinkEnter(int con_num, const char *arg);
extern void NetSendELinkFoward(int con_num, const char *url);

extern void NetHandleSetEngine(const int con_num, const char *arg);
extern void NetSendSetEngine(const int con_num, const int obj_num);

extern void NetFireWeaponSyncObjectValues(
	xsw_object_struct *obj, const int obj_num,
	const int wep_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetHandleFireWeapon(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendFireWeapon(const int con_num, const int obj_num);

extern void NetHandleHail(int con_num, const char *arg);
extern void NetSendHail(
	int con_num, 
	int src_obj, int tar_obj,
	float bearing, int channel   
);

extern void NetHandleSetInterval(int con_num, const char *arg);
extern void NetSendSetInterval(int con_num);

extern void NetHandleSetIntercept(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendSetIntercept(const int con_num, const int obj_num);

extern void NetHandleLighting(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendLighting(const int con_num, const int obj_num);

extern void NetHandleVersion(int con_num, const char *arg);
extern void NetSendVersion(int con_num);

extern void NetHandleLogin(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendLogin(const int con_num);

extern void NetHandleLogout(
	const int con_num,   
	const time_t cur_ms, const time_t cur_sec, const float tc     
);
extern void NetSendLogout(const int con_num);

extern void NetHandleNotifyDestroy(int con_num, const char *arg);
extern void NetSendNotifyDestroy(
	int con_num,
	int reason,
	int destroyed_obj_num,
	int destroyer_obj_num,
	int destroyer_obj_owner_num
);

extern void NetHandleNotifyHit(const int con_num, const char *arg);
extern void NetSendNotifyHit(
	const int con_num,
	const int wobj_num, const int tobj_num,
	const float bearing,		/* Target to weapon */
	const float dmg_structure,
	const float dmg_shield
);

extern void NetSendObjectForcePose(
	const int con_num, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

extern void NetHandleObjectMaximums(const int con_num, const char *arg);
extern void NetSendObjectMaximums(const int con_num, const int obj_num);

extern void NetHandleObjectName(int con_num, const char *arg);
extern void NetSendObjectName(int con_num, int obj_num);

extern void NetHandleObjectDescription(int con_num, const char *arg);
extern void NetSendObjectDescription(int con_num, int obj_num);

extern void NetHandleObjectPose(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendObjectPose(int con_num, int obj_num);

extern void NetHandleReqObjSect(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetHandleObjSect(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendObjSect(
	const int con_num, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendObjForceSect(
	const int con_num, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

extern void NetHandleObjectValues(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendObjectValues(const int con_num, const int obj_num);

extern int NetHandleRecycleObject(const int con_num, const char *arg);
extern int NetSendRecycleObject(const int con_num, const int obj_num);

extern void NetHandleRefresh(const int con_num, const char *arg);
extern void NetSendRefreshObject(
        const int con_num, const int obj_num,
        const char send_weapon_values
);
extern void NetSendRefresh(const int con_num);

extern void NetHandleScore(int con_num, const char *arg);
extern void NetSendScore(int con_num, int obj_num);

extern void NetHandleSetShields(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendSetShields(const int con_num, const int obj_num);

extern int NetHandleShieldVis(int con_num, char *arg);
extern int NetSendShieldVis(int con_num, int obj_num);

extern void NetHandleSetObjParticleTrail(const int con_num, const char *arg);
extern void NetSendSetObjParticleTrail(
	const int con_num, const int obj_num, const int op,
	const float offset,
	const float heading, const float pitch, const float bank
);

extern void NetHandleCreateEffect(const int con_num, const char *arg);
extern void NetSendCreateEffect(
	const int con_num,
	const unsigned int flags,	/* Any of SW_EFFECT_FLAG_* */
	const int imgref_num,
	const long sect_x, const long sect_y, const long sect_z,
	const float x, const float y, const float z,
	const long size_min, const long size_max,
	const time_t life_span,
	const int ref_obj_num
);

extern void NetHandleCreateShockwave(const int con_num, const char *arg);
extern void NetSendCreateShockwave(
	const int con_num,
	const int type,
	const long sect_x, const long sect_y, const long sect_z,
	const float x, const float y, const float z,
	const long size_min, const long size_max,
	const time_t life_span,
	const float damage_hull, const float damage_shield,
	const int ref_obj_num
);

extern void NetHandlePlaySound(const int con_num, const char *arg);
extern void NetSendPlaySound(
	const int con_num, const int sound_code,
	const float left_volume, const float right_volume
);

extern void NetSendStarChartAddObject(const int con_num, const int obj_num);
extern void NetSendStarChartSetObjectName(const int con_num, const int obj_num);
extern void NetSendStarChartSetObjectEmpire(const int con_num, const int obj_num);
extern void NetSendStarChartRecycleObject(const int con_num, const int obj_num);
extern void NetSendEntireStarChart(const int con_num);

extern void NetHandleObjectThrottle(const int con_num, const char *arg);
extern void NetSendObjectThrottle(const int con_num);

extern void NetHandleTractorBeamLock(int con_num, const char *arg);
extern void NetSendTractorBeamLock(
	int con_num, int obj_num, int tar_obj_num
);

extern void NetHandleSetLegend(const int con_num, const char *arg);
extern void NetSendLegend(const int con_num);

extern void NetHandleSetTitle(const int con_num, const char *arg);
extern void NetSendTitle(const int con_num);

extern void NetHandleWeaponDisarm(int con_num, const char *arg);
extern void NetSendWeaponDisarm(
	int con_num, int src_obj_num, int tar_obj_num
);

extern void NetHandleSetWeaponsLock(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendSetWeaponsLock(const int con_num, const int obj_num);

extern void NetHandleSelectWeapon(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendSelectWeapon(const int con_num, const int obj_num);

extern void NetHandleSetWeaponsUnlock(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
extern void NetSendSetWeaponsUnlock(const int con_num, const int obj_num);

extern void NetHandleWeaponValues(int con_num, const char *arg);
extern void NetSendWeaponValues(
	int con_num, int obj_num, int wep_num
);

extern void NetHandleWhoAmI(int con_num, const char *arg);
extern void NetSendWhoAmI(int con_num);

extern void NetHandleWormHoleEnter(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#endif	/* NET_H */
