#include <stdio.h>
#include <stdlib.h>

#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetHandleAutoDestruct(
        const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendAutoDestruct(
	const int con_num, const int obj_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleAutoDestruct(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num = -1;
	unsigned long sec = 0l;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;

	/* SWEXTCMD_AUTODESTRUCT format:
	 *
	 *	obj_num sec
	 */
	sscanf(
	    arg,
	    "%i %lu",
	    &obj_num, &sec
	);

	if(obj_num != con_obj_num)
	    return;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Schedual or cancel auto-destruct on the object */
	if(sec > 0l)
	    obj->auto_destruct_next = cur_sec + (time_t)sec;
	else
	    obj->auto_destruct_next = 0l;

	/* Send response to connection of auto destruct start or
	 * cancel
	 */
	NetSendAutoDestruct(con_num, obj_num, cur_ms, cur_sec, tc);
}


void NetSendAutoDestruct(
	const int con_num, const int obj_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	unsigned long sec;
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	if(obj->auto_destruct_next > 0l)
	    sec = (unsigned long)(
		obj->auto_destruct_next - cur_sec
	    );
	else
	    sec = 0l;

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* SWEXTCMD_AUTODESTRUCT format:
	 *             
	 *	obj_num sec
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %lu\n",
	    CS_CODE_EXT,
	    SWEXTCMD_AUTODESTRUCT,
	    obj_num, sec
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
