#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/swsoundcodes.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "net.h"
#include "swserv.h"


void NetHandleSetCloak(
        const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleSetCloak(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num = -1, con_obj_num;
        int cloak_state = SW_CLOAK_STATE_NONE;
        xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];

        con_obj_num = con->object_num;
        if(DBIsObjectGarbage(con_obj_num))
            return;

	obj = xsw_object[con_obj_num];

        /* SWEXTCMD_SETCLOAK format:
         *
         *	obj_num cloak_state
         */
        sscanf(
	    arg,
	    "%i %i",
            &obj_num, &cloak_state
        );      

        /* Connection must own object */
        if(obj_num != con_obj_num)
            return;

        /* Does object have cloak abilities? */
        if(obj->cloak_state == SW_CLOAK_STATE_NONE)
            return;

        /* Are shields up? if so, lower it */
        if(obj->shield_state == SW_SHIELD_STATE_UP)
            obj->shield_state = SW_SHIELD_STATE_DOWN;

        /* Raise cloak? */
        if((obj->cloak_state == SW_CLOAK_STATE_DOWN) &&
           (cloak_state == SW_CLOAK_STATE_UP)
        )
        {
	    /* Has the cloak refractory period passed? */
	    if((obj->last_cloak + obj->cloak_refractory_int) < cur_ms)
	    {
		/* Can only cloak if there is antimatter */
		if(obj->antimatter > 0.0f)
		{
		    obj->cloak_state = SW_CLOAK_STATE_UP;
		    obj->last_cloak = cur_ms;
                    NetSendPlaySound(
                        con_num, SW_SNDREF_CLOAK_UP,
                         1.0f, 1.0f
                    );
		}
	    }
        }
        /* Lower cloak? */
        else if((obj->cloak_state == SW_CLOAK_STATE_UP) &&
                (cloak_state == SW_CLOAK_STATE_DOWN)
        )
        {
	    int i;
	    xsw_object_struct *obj2;
	    sws_connection_struct *con2;

            obj->cloak_state = SW_CLOAK_STATE_DOWN;
	    obj->last_cloak = cur_ms;
            NetSendPlaySound(
                con_num, SW_SNDREF_CLOAK_DOWN,
                1.0f, 1.0f               
            );

	    /* Refresh this decloaked object to nearby connections */
            for(i = 0; i < total_connections; i++)
            {
                con2 = connection[i];
                if(con2 == NULL)
                    continue;

                if(DBIsObjectGarbage(con2->object_num))
                    continue;

                if(obj_num == con2->object_num)
                    continue;

                obj2 = xsw_object[con2->object_num];

                /* Not in scanner rnage? */
                if(!SWMath3DInRangePtr(
                    obj, obj2,
                    obj2->scanner_range
                ))
                    continue;

                NetSendRefreshObject(i, obj_num, 1);
            }
        }

        /* Set next object values update to now so it gets updated now */
        next.object_values = cur_ms;
}

void NetSendSetCloak(const int con_num)
{
        return;
}
