#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/objects.h"
#include "../include/unvmatch.h"
#include "../include/swnetcodes.h"

#include "eco.h"
#include "net.h"
#include "swserv.h"


void NetHandleEcoGetValues(int con_num, const char *arg);

void NetSendEcoSetValues(int con_num, int obj_num);
void NetSendEcoSetProductValues(
	int con_num,
	int obj_num, int product_num
);

void NetHandleEcoBuy(int con_num, const char *arg);
void NetHandleEcoSell(int con_num, const char *arg);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleEcoGetValues(int con_num, const char *arg)
{
	int i, obj_num = -1;
	xsw_object_struct *obj;

	/* SWEXTCMD_ECO_REQVALUES format:
	 *
	 *	obj_num
	 */
	sscanf(
	    arg,
	    "%i",
	    &obj_num
	);

	/* Is object valid? */
	if(DBIsObjectGarbage(obj_num))
	    obj = NULL;
	else
	    obj = xsw_object[obj_num];

	if((obj != NULL) ? (obj->eco != NULL) : 0)
	{
	    xsw_ecodata_struct *eco = obj->eco;

	    /* Send general eco info */
	    NetSendEcoSetValues(con_num, obj_num);

	    /* Send product info */
	    for(i = 0; i < eco->total_products; i++)
		NetSendEcoSetProductValues(
		    con_num, obj_num, i
		);

	    /* Send score for proprietor object */
	    NetSendScore(con_num, obj_num);
	}

	/* Send score for connection's object */
	NetSendScore(
	    con_num,
	    connection[con_num]->object_num
	);
}


void NetSendEcoSetValues(int con_num, int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Does object have economy data? */
	eco = obj->eco;
	if(eco == NULL)
	    return;

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 + 40 + 40 + 40 +
	    80
	);

	/* SWEXTCMD_ECO_SETVALUES format:
	 *
	 *	obj_num
	 *	flags
	 *	tax_general tax_friend tax_hostile
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %ld %.4f %.4f %.4f\n", 
	    CS_CODE_EXT,
	    SWEXTCMD_ECO_SETVALUES,
	    obj_num,
	    eco->flags,
	    eco->tax_general,
	    eco->tax_friend,
	    eco->tax_hostile
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}

void NetSendEcoSetProductValues(
	int con_num,
	int obj_num,
	int product_num
)
{
	char *sndbuf;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *product;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Does object have economy data? */
	eco = obj->eco;
	if(eco == NULL)
	    return;

	/* Is product_num valid? */
	if((product_num < 0) || (product_num >= eco->total_products))
	    return;

	product = eco->product[product_num];
	if(product == NULL)
	    return;

        sndbuf = (char *)malloc(
            40 + 40 +
            40 + 40 +
	    40 + 40 + 40 + 40 + STRLEN(product->name) +
            80
        );

	/* SWEXTCMD_ECO_SETPRODUCTVALUES format:    
	 *
	 *	obj_num ocs_code sell_price buy_price
	 *	product_amount product_max;product_name
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\
 %.4f %.4f %.4f %.4f;%s\n",
	    CS_CODE_EXT,
	    SWEXTCMD_ECO_SETPRODUCTVALUES,
	    obj_num,
	    product->ocs_code,
	    product->sell_price,
	    product->buy_price,
	    product->amount,
	    product->amount_max,
	    product->name
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}


void NetHandleEcoBuy(int con_num, const char *arg)
{
	int customer_obj_num = -1, proprietor_obj_num = -1;
	xsw_ecoproduct_struct *product;
        sws_connection_struct *con = connection[con_num];

	product = (xsw_ecoproduct_struct *)calloc(
	    1, sizeof(xsw_ecoproduct_struct)
	);

	/* SWEXTCMD_ECO_BUY format:
	 *
	 *	customer_obj_num proprietor_obj_num
	 *	ocs_code amount
	 */
	sscanf(
	    arg,
"%i %i %i %f",
	    &customer_obj_num,
	    &proprietor_obj_num,
            &product->ocs_code,
            &product->amount
	);

	/* Customer object must match that of connection */
	if(customer_obj_num != con->object_num)
	    return;

	/* Perform buy transaction */
	EcoBuy(
	    con_num,
	    customer_obj_num,
	    proprietor_obj_num,
	    product
	);

	free(product);
}

void NetHandleEcoSell(int con_num, const char *arg)
{
	int customer_obj_num = -1, proprietor_obj_num = -1;
	xsw_ecoproduct_struct *product;
	sws_connection_struct *con = connection[con_num];

	product = (xsw_ecoproduct_struct *)calloc(
	    1, sizeof(xsw_ecoproduct_struct)
	);

	/* SWEXTCMD_ECO_SELL format:
	 *
	 *	customer_obj_num proprietor_obj_num
	 *	ocs_code amount
	 */
	sscanf(
	    arg,
"%i %i %i %f",
	    &customer_obj_num,
	    &proprietor_obj_num,
	    &product->ocs_code,
	    &product->amount
	);

	/* Customer object must match that of connection */
	if(customer_obj_num != con->object_num)
	    return;

	/* Perform sell transaction */
	EcoSell(
	    con_num,
	    customer_obj_num,
	    proprietor_obj_num,
	    product
	);

	free(product);
}
