#include <stdio.h>
#include <stdlib.h>

#include "../include/unvmatch.h"
#include "../include/swsoundcodes.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetHandleSetEngine(const int con_num, const char *arg);
void NetSendSetEngine(const int con_num, const int obj_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleSetEngine(const int con_num, const char *arg)
{
	int obj_num = -1, con_obj_num;
	int engine_state = SW_ENGINE_STATE_NONE;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];

	con_obj_num = con->object_num;
	if(DBIsObjectGarbage(con_obj_num))
	    return;

	obj = xsw_object[con_obj_num];

	/* SWEXTCMD_SETENGINE format:
	 *
	 *	obj_num engine_state
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &obj_num, &engine_state
	);

	/* Connection must own object */
	if(obj_num != con_obj_num)
	    return;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

/* Consider starting as on for now */
if(engine_state == SW_ENGINE_STATE_STARTING)
 engine_state = SW_ENGINE_STATE_ON;

	if(engine_state != obj->engine_state)
	{
	    switch(engine_state)
	    {
	      case SW_ENGINE_STATE_ON:
		NetSendPlaySound(
		    con_num, SW_SNDREF_ENGINES_ON,
		    1.0f, 1.0f
		);
		break;
	      case SW_ENGINE_STATE_STARTING:
		NetSendPlaySound(
		    con_num, SW_SNDREF_ENGINES_ON,
		    1.0f, 1.0f
		);
		break;
	      case SW_ENGINE_STATE_OFF:
		NetSendPlaySound(
		    con_num, SW_SNDREF_ENGINES_OFF,
		    1.0f, 1.0f
		);
		break;
	    }
	    obj->engine_state = engine_state;

	    /* Send engine state update to all connections */
	    NetSendSetEngine(-1, obj_num);
	}
}


void NetSendSetEngine(const int con_num, const int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* SWEXTCMD_SETENGINE format:
	 *
	 *	obj_num engine_state
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETENGINE,
	    obj_num, obj->engine_state
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
