#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetHandleSetImageSet(int con_num, const char *arg);
void NetHandleSetSoundSet(int con_num, const char *arg);
void NetHandleSetOCSN(int con_num, const char *arg);

void NetSendSetImageSet(int con_num, const char *filename);
void NetSendSetSoundSet(int con_num, const char *filename);
void NetSendSetOCSN(int con_num, const char *filename);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleSetImageSet(int con_num, const char *arg)
{
	sw_legend_struct *legend = &sw_legend;
	NetSendSetImageSet(con_num, legend->imgrefs);
}

void NetHandleSetSoundSet(int con_num, const char *arg)
{
	sw_legend_struct *legend = &sw_legend;
	NetSendSetSoundSet(con_num, legend->sndrefs);
}

void NetHandleSetOCSN(int con_num, const char *arg)
{
	sw_legend_struct *legend = &sw_legend;
	NetSendSetOCSN(con_num, legend->ocsn);
}


void NetSendSetImageSet(int con_num, const char *filename)
{
	char	*sndbuf,
		lname[CS_DATA_MAX_LEN];

	if(STRISEMPTY(filename))
	    return;

	strncpy(lname, filename, sizeof(lname));
	lname[sizeof(lname) - 1] = '\0';

	sndbuf = (char *)malloc(
	    40 + STRLEN(lname) + 80
	);

	/* Format CS_CODE_IMAGESET:
	 *
	 *	path
	 */
	sprintf(
	    sndbuf,
	    "%i %s\n",
	    CS_CODE_IMAGESET,
	    lname
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}

void NetSendSetSoundSet(int con_num, const char *filename)
{
	char	*sndbuf,
		lname[CS_DATA_MAX_LEN];

	if(STRISEMPTY(filename))
	    return;

	strncpy(lname, filename, sizeof(lname));
	lname[sizeof(lname) - 1] = '\0';

	sndbuf = (char *)malloc(
	    40 + STRLEN(lname) + 80
	);

	/* Format CS_CODE_SOUNDSET:
	 *
	 *	path
	 */
	sprintf(
	    sndbuf,
	    "%i %s\n",
	    CS_CODE_SOUNDSET,
	    lname
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}

void NetSendSetOCSN(int con_num, const char *filename)
{
	char	*sndbuf,
		lname[CS_DATA_MAX_LEN];

	if(STRISEMPTY(filename))
	    return;

	strncpy(lname, filename, sizeof(lname));
	lname[sizeof(lname) - 1] = '\0';

	sndbuf = (char *)malloc(
	    40 + 40 + STRLEN(lname) + 80
	);

	/* Format SWEXTCMD_SETOCSN:
	 *
	 *	path
	 */
	sprintf(
	    sndbuf,
	    "%i %i %s\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETOCSN,
	    lname
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
