#include <stdio.h>
#include <stdlib.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#ifdef PLUGIN_SUPPORT
#include "plugins.h"
#endif
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


void NetHandleHail(int con_num, const char *arg);
void NetSendHail(
	int con_num,
	int obj_num, int tar_obj_num,
	float bearing, int channel
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleHail(int con_num, const char *arg)
{
	int	i,
		src_obj_num = -1, tar_obj_num = -1,
		channel = 0, tar_con_num;
	float bearing = 0.0f * PI;
	sw_legend_struct *legend = &sw_legend;
	xsw_object_struct *src_obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;

	/* SWEXTCMD_HAIL format:
	 *
	 *	src_obj_num tar_obj_num
	 *	bearing channel
	 */
	sscanf(
	    arg,
	    "%i %i %f %i",
	    &src_obj_num,
	    &tar_obj_num,
	    &bearing,
	    &channel
	);

	if(src_obj_num != con_obj_num)
	    return;

	if(DBIsObjectGarbage(src_obj_num))
	    return;

	src_obj = xsw_object[src_obj_num];

	for(i = 0; i < total_objects; i++)
	{
	    /* Object not valid or not in communications range? */
	    if(!SWMath3DInRange(
		xsw_object, total_objects,
		src_obj_num, i, legend->com_range
	    ))
		continue;

	    /* Is the target object specified? */
	    if((tar_obj_num > -1) && (tar_obj_num != i) &&
	       (src_obj_num != i)
	    )
	    {
		/* Skip since the hail is not ment for this object */
		continue;
	    }

	    /* Get the connection number of this object */
	    tar_con_num = ConGetByObject(i);
	    if(tar_con_num < 0)
		continue;

	    /* Get bearing, object_count to source object */
	    bearing = SWMathCoordinateDeltaVector(
		src_obj->x - xsw_object[i]->x,
		src_obj->y - xsw_object[i]->y
	    );

	    /* Send hail to that connection */
	    NetSendHail(
		tar_con_num,
		src_obj_num,
		tar_obj_num,
		bearing,
		channel
	    );
	}

#ifdef PLUGIN_SUPPORT
	/* Forward this hail to plugins */
	if(total_plugins > 0)
	{
	    int i;
	    for(i = 0; i < total_plugins; i++)
		PluginHailNotify(
		    i,
		    con_num,
		    src_obj_num,
		    tar_obj_num,
		    channel,
		    NULL	/* No message, since just a hail */
		);
	}
#endif
}


void NetSendHail(
	int con_num,
	int obj_num, int tar_obj_num,
	float bearing, int channel
)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* SWEXTCMD_HAIL format:
	 *
	 *	obj_num tar_obj_num bearing channel
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\
 %.4f %i\n",
	    CS_CODE_EXT, SWEXTCMD_HAIL,
	    obj_num, tar_obj_num,
	    bearing, channel
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
