#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


void NetHandleMessage(const int con_num, const char *msg);
void NetSendMessage(const int con_num, const char *msg);

void NetHandleSysMessage(const int con_num, const char *msg);
void NetSendSysMessage(
	const int con_num,
	const int code,		/* One of CS_SYSMESG_* */
	const char *msg
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleMessage(const int con_num, const char *msg)
{
	char *name, *dmsg, *sndbuf;
	sws_connection_struct *con = connection[con_num];
	int obj_num = con->object_num;
	if(DBIsObjectGarbage(obj_num))
	    return;

	if(STRISEMPTY(msg))
	    return;

	/* Make a copy of the message as dmsg and make sure that it is
	 * not longer than CS_MESG_MAX characters
	 */
	dmsg = STRDUP(msg);
	if(STRLEN(dmsg) > CS_MESG_MAX)
	    dmsg[CS_MESG_MAX] = '\0';

	name = STRDUP(DBGetFormalNameStr(obj_num));

	sndbuf = (char *)malloc(
	    STRLEN(name) + STRLEN(dmsg) + 80
	);
	sprintf(
	    sndbuf,
"%s: \"%s\"",
	    name,
	    dmsg
	);
	NetSendMessage(-1, sndbuf);
	free(sndbuf);

	/* Log */
	if(sysparm.log_events)
	{
	    char *s = (char *)malloc(
		STRLEN(name) + STRLEN(dmsg) + 80
	    );
	    sprintf(
		s,
"%s: Global Message: \"%s\"",
		name, dmsg
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	free(dmsg);
	free(name);
}

void NetSendMessage(const int con_num, const char *msg)
{
	char *dmsg, *sndbuf;

	if(msg == NULL)
	    return;

	/* Make a copy of the message as dmsg and make sure that it is
	 * not longer than CS_MESG_MAX characters
	 */
	dmsg = STRDUP(msg);
	if(STRLEN(dmsg) > CS_MESG_MAX)
	    dmsg[CS_MESG_MAX] = '\0';

	sndbuf = (char *)malloc(
	    40 + STRLEN(dmsg) + 80
	);

	/* CS_CODE_LIVEMESSAGE format:
	 *
	 *	msg
	 */
	sprintf(
	    sndbuf,
"%i %s\n",
	    CS_CODE_LIVEMESSAGE,
	    dmsg
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);

	free(dmsg);
}


void NetHandleSysMessage(const int con_num, const char *msg)
{
	return;
}

void NetSendSysMessage(
	const int con_num,
	const int code,		/* One of CS_SYSMESG_* */
	const char *msg
)
{
	char *dmsg, *sndbuf;

	if(STRISEMPTY(msg))
	    return;

	/* Make a copy of the message as dmsg and make sure that it is
	 * not longer than CS_MESG_MAX characters
	 */
	dmsg = STRDUP(msg);
	if(STRLEN(dmsg) > CS_MESG_MAX)
	    dmsg[CS_MESG_MAX] = '\0';

	sndbuf = (char *)malloc(
	    40 + 40 + STRLEN(dmsg) + 80
	);

	/* CS_CODE_SYSMESSAGE format:
	 *
	 *	code msg
	 */
	sprintf(
	    sndbuf,
"%i %i %s\n",
	    CS_CODE_SYSMESSAGE,
	    code, dmsg
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);

	free(dmsg);
}
