#include <stdio.h>
#include <stdlib.h>

#include "../include/unvmatch.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"
#include "sysparm.h"


static int SectLimitDistance(
	sws_connection_struct *con,
	xsw_object_struct *obj,
	long sect_x, long sect_y, long sect_z,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void NetHandleReqObjSect(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetHandleObjSect(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void NetSendObjSect(
	const int con_num, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendObjForceSect(
	const int con_num, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the object is close enough to change sectors.
 */
static int SectLimitDistance(
	sws_connection_struct *con,
	xsw_object_struct *obj,
	long sect_x, long sect_y, long sect_z,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int status = 0;
	const long	dsect_x = sect_x - obj->sect_x,
			dsect_y = sect_y - obj->sect_y,
			dsect_z = sect_z - obj->sect_z;
#if 0
	const time_t dt = CLIP(
	    (cur_ms - con->obj_last_sect),
	    0l, 10000l
	);
	const float dlimit = MAX(
	    (obj->velocity_max * (float)dt / (float)SW_CYCLE_LAPSE_MS),
	    obj->velocity_max
	);
#endif
	/* We need to use a relatively large velocity max to ensure
	 * there are no problems when crossing sectors at high
	 * velocities
	 */
	const float dlimit = MAX(
	    obj->velocity_max, 1.0f
	);
	sw_legend_struct *legend = &sw_legend;
	sws_sysparm_struct *parm = &sysparm;

	if((dsect_x == 1l) || (dsect_x == -1l))
	{
	    const float x_tolor = dlimit * parm->movement_limit_tolorance;
	    if((obj->x <= (legend->sect_x_min + x_tolor)) ||
	       (obj->x >= (legend->sect_x_max - x_tolor))
	    )
		status = 1;
	}
	if((dsect_y == 1l) || (dsect_y == -1l))
	{
	    const float y_tolor = dlimit * parm->movement_limit_tolorance;
	    if((obj->y <= (legend->sect_y_min + y_tolor)) ||
	       (obj->y >= (legend->sect_y_max - y_tolor))
	    )
		status = 1;
	}
	if((dsect_z == 1l) || (dsect_z == -1l))
	{
	    const float z_tolor = dlimit * parm->movement_limit_tolorance;
	    if((obj->z <= (legend->sect_z_min + z_tolor)) ||
	       (obj->z >= (legend->sect_z_max - z_tolor))
	    )
		status = 1;
	}

	return(status);
}


void NetHandleReqObjSect(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num = -1;

	/* SWEXTCMD_REQSECT format:
	 *
	 *	obj_num
	 */
	sscanf(
	    arg,
	    "%i",
	    &obj_num
	);

	if(DBIsObjectGarbage(obj_num))
	    return;

	NetSendObjSect(con_num, obj_num, cur_ms, cur_sec, tc);
}

void NetHandleObjSect(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num = -1, con_obj_num;  
	long sect_x = 0l, sect_y = 0l, sect_z = 0l;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];

	con_obj_num = con->object_num;

	/* SWEXTCMD_SETOBJSECT format:
	 *
	 *	obj_num
	 *	sect_x sect_y sect_z
	 */
	sscanf(
	    arg,
	    "%i %ld %ld %ld",
	    &obj_num,
	    &sect_x, &sect_y, &sect_z
	);

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Connection must own object */
	if(con_obj_num != obj_num)
	    return;

	/* Waiting for a force sect response? */
	if(con->flags & SWS_CON_FLAG_FSECT_WAIT)
	{
	    /* Is this a valid response? */
	    if((obj->sect_x == sect_x) &&
	       (obj->sect_y == sect_y) &&
	       (obj->sect_z == sect_z)
	    )
	    {
		/* Switch off force sect wait */
		con->flags &= ~SWS_CON_FLAG_FSECT_WAIT;
		con->obj_last_sect = cur_ms;
	    }
	    else
	    {
		/* Invalid response to a force sect, check if timmed
		 * out
		 */
		const time_t fsect_timeout = 3000l;
		if((con->obj_last_sect + fsect_timeout) < cur_ms)
		{
		    /* Timmed out, switch off force sect wait */
		    con->flags &= ~SWS_CON_FLAG_FSECT_WAIT;
		}
		else
		{
		    /* Resend force sect */
		    NetSendObjForceSect(
			con_num, obj_num,
			cur_ms, cur_sec, tc
		    );
		}
	    }
	    return;
	}

	/* Last sect time older (not from the same iteration)?
	 *
	 * This is to prevent the handling of subsequent sect
	 * statements that were received bunched togeather until the
	 * next main iteration
	 */
	if(con->obj_last_sect >= cur_ms)
	    return;

	/* Within distance to change sectors? */
	if(SectLimitDistance(
	    con, obj,
	    sect_x, sect_y, sect_z,
	    cur_ms, cur_sec, tc
	))
	{
	    /* Change the object's sector and position */
	    DBObjectDoSetSector(
		obj_num,
		sect_x, sect_y, sect_z,
		1			/* Allow wrapping */
	    );
	}
	else
	{
	    /* Too far away from sector boundary to change sectors,
	     * send force sect and pose to correct the client's
	     * object position
	     */
	    NetSendObjForceSect(
		con_num, obj_num,
		cur_ms, cur_sec, tc
	    );
	    NetSendObjectForcePose(con_num, obj_num, cur_ms, cur_sec, tc);
	}

	con->obj_last_sect = cur_ms;
}       


void NetSendObjSect(
	const int con_num, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 + 40 + 40 + 80
	);

	/* SWEXTCMD_SETOBJSECT format:
	 *      
	 *	obj_num
	 *	sect_x sect_y sect_z
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %ld %ld %ld\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETOBJSECT,
	    obj_num,
	    obj->sect_x, obj->sect_y, obj->sect_z
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}

void NetSendObjForceSect(
	const int con_num, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int tar_con_num;
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 + 40 + 40 + 80
	);

	/* SWEXTCMD_SETFOBJSECT format:
	 *
	 *	obj_num
	 *	sect_x sect_y sect_z
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %ld %ld %ld\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETFOBJSECT,
	    obj_num,
	    obj->sect_x, obj->sect_y, obj->sect_z
	);

	tar_con_num = ConGetByObject(obj_num);
	if(tar_con_num > -1)
	{
	    sws_connection_struct *con = connection[tar_con_num];

	    /* Mark that the object is waiting for a force sect response */
	    con->flags |= SWS_CON_FLAG_FSECT_WAIT;

	    /* Update the last sect time with the current time + 1 so that
	     * any subsequent sects from the client that are to be handled
	     * on this same iteration will get ignored
	     */
	    con->obj_last_sect = cur_ms + 1l;
	}

	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
