#include <stdio.h>
#include <stdlib.h>

#include "../include/unvmatch.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetHandleObjectValues(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendObjectValues(const int con_num, const int obj_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleObjectValues(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num = -1, con_obj_num;
	int loc_type = SW_LOC_TYPE_SPACE;
	int loc_obj_num = -1;
	int locked_on = -1;
	int intercepting_object = -1;
	int thrust_rev_state = 0;
	float thrust_dir = 0.0f * PI;
	float thrust = 0.0f;
	float throttle = 0.0f;
	int lighting = 0;		/* xswo_lighting_t */
	float hp = 0.0f;
	float power = 0.0f;
	float antimatter = 0.0f;
	int shield_state = SW_SHIELD_STATE_NONE;
	int selected_weapon = -1;
	int cloak_state = SW_CLOAK_STATE_NONE;
	float cloak_strength = 0.0f;
	float visibility = 0.0f;
	int damage_control = SW_DMGCTL_STATE_OFF;
	xsw_object_struct *obj;

	con_obj_num = connection[con_num]->object_num;
	if(DBIsObjectGarbage(con_obj_num))
	    return;

	obj = xsw_object[con_obj_num];

	/* SWEXTCMD_STDOBJVALS format:
	 *
	 *	obj_num
	 *	loc_type loc_obj_num locked_on intercepting_object
	 *	thrust_rev_state thrust_dir thrust throttle
	 *	lighting hp power
	 *	antimatter shield_state selected_weapon cloak_state
	 *	cloak_strength visibility damage_control
	 */
	sscanf(
	    arg,
"%i\
 %i %i %i %i\
 %i %f %f %f\
 %i %f %f\
 %f %i %i %i\
 %f %f %i",
	    &obj_num,
	    &loc_type, &loc_obj_num, &locked_on, &intercepting_object,
	    &thrust_rev_state, &thrust_dir, &thrust, &throttle,
	    &lighting, &hp, &power,
	    &antimatter, &shield_state, &selected_weapon, &cloak_state,
	    &cloak_strength, &visibility, &damage_control
	);

	/* Connection must own object */
	if(obj_num != con_obj_num)
	    return;

	/* Set throttle */
	if(throttle < 0.0f)
	    throttle = 0.0f;
	else if(throttle > 1.0f)
	    throttle = 1.0f;
	obj->throttle = throttle;

	/* All other members are not set from client */
}


void NetSendObjectValues(const int con_num, const int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 + 40 +
	    40 + 40 + 40 + 40 +
	    40 + 40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 + 40 +
	    80
	);

	/* SWEXTCMD_STDOBJVALS format:
	 *
	 *	obj_num
	 *	loc_type loc_obj_num locked_on intercepting_object
	 *	thrust_rev_state thrust_dir thrust throttle
	 *	lighting hp power
	 *	antimatter shield_state selected_weapon cloak_state
	 *	cloak_strength visibility damage_control
	 */
	sprintf(
	    sndbuf,
"%i %i %i\
 %i %i %i %i\
 %i %.4f %.5f %.4f\
 %lu %.4f %.4f\
 %.4f %i %i\
 %i %.4f %.4f %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_STDOBJVALS,
	    obj_num,
	    obj->loc_type, obj->loc_obj_num, obj->locked_on, obj->intercepting_object,
	    0,		/* thrust_rev_state, no longer used */
	    obj->thrust_dir, obj->thrust, obj->throttle,
	    (unsigned long)obj->lighting, obj->hp, obj->power,
	    obj->antimatter, obj->shield_state, obj->selected_weapon,
	    obj->cloak_state, obj->cloak_strength, obj->visibility, obj->damage_control
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
