#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


void NetHandleRefresh(const int con_num, const char *arg);

void NetSendRefreshObject(
        const int con_num, const int obj_num,
        const char send_weapon_values
);
void NetSendRefresh(const int con_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleRefresh(const int con_num, const char *arg)
{
	NetSendRefresh(con_num);
}


/*
 *	Refreshes the specified object.
 *
 *	Sends a create object for the specified object and then all the
 *	values of the specified object.
 */
void NetSendRefreshObject(
	const int con_num, const int obj_num,
	const char send_weapon_values
)
{
	xsw_object_struct *obj;

        if(DBIsObjectGarbage(obj_num))
            return;

        obj = xsw_object[obj_num];


        /* Send create object */
        NetSendCreateObject(con_num, obj_num);

        /* Send object maximums */
        NetSendObjectMaximums(con_num, obj_num);

        /* Send object name */
        NetSendObjectName(con_num, obj_num);

        /* Engine state */
        NetSendSetEngine(con_num, obj_num);

	/* Particle trails */
	NetSendSetObjParticleTrail(
	    con_num,
	    obj_num,
	    (obj->ext_flags & SWS_OBJ_FLAG_ON_FIRE) ?
		SW_PARTICLE_TRAIL_OP_FIRE_ON :
		SW_PARTICLE_TRAIL_OP_FIRE_OFF,
	    0.0f, 0.0f * PI, 0.0f * PI, 0.0f * PI
	);
        /* Particle trails */
        NetSendSetObjParticleTrail(
            con_num,
            obj_num,
            (obj->ext_flags & SWS_OBJ_FLAG_SMOKE_TRAIL) ?
                SW_PARTICLE_TRAIL_OP_SMOKE_ON : 
                SW_PARTICLE_TRAIL_OP_SMOKE_OFF,
	    0.0f, 0.0f * PI, 0.0f * PI, 0.0f * PI
        );

        /* Com channel */
        NetSendSetChannel(con_num, obj_num, obj->com_channel);

        /* Send tractor beam lock */
        if(obj->total_tractored_objects > 0)
        {
	    const int m = obj->total_tractored_objects;
            int i;
            for(i = 0; i < m; i++)
                NetSendTractorBeamLock(
                    con_num, obj_num, obj->tractored_object[i]
		);
	}

	/* If the connection's object is locked on to this object,
	 * then send weapons and current values of this object
	 */
	if(send_weapon_values)
        {
            /* Send values */
            NetSendObjectValues(con_num, obj_num);

            /* Send weapon values */
	    if(obj->total_weapons > 0)
	    {
		const int m = obj->total_weapons;
		int i;
		for(i = 0; i < m; i++)
		    NetSendWeaponValues(con_num, obj_num, i);
	    }
	}
}

/*
 *	Refreshes all values to the connection.
 *
 *	Including all the objects within the connection's object's
 *	scanner range.
 */
void NetSendRefresh(const int con_num)
{
	int i, obj_num;
	sw_legend_struct *legend = &sw_legend;
	xsw_object_struct *obj, *tar_obj;
	sws_connection_struct *con;

	if(!ConIsConnected(con_num))
	    return;

	con = connection[con_num];

	obj_num = con->object_num;
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];


	/* Send universe legend */
	NetSendLegend(con_num);

	/* Send universe title */
	NetSendTitle(con_num);

	/* Send file names */
	NetSendSetImageSet(con_num, legend->imgrefs);
	NetSendSetSoundSet(con_num, legend->sndrefs);
	NetSendSetOCSN(con_num, legend->ocsn);

	/* Send who am I */
	NetSendWhoAmI(con_num);

	/* Send the connection's object */
	NetSendRefreshObject(
	    con_num, obj_num,
	    1					/* Send weapon values */
	);
	NetSendObjectDescription(con_num, obj_num);
        NetSendScore(con_num, obj_num);


	/* Iterate through objects list */
	for(i = 0; i < total_objects; i++)
	{
	    tar_obj = xsw_object[i];
	    if(tar_obj == NULL)
		continue;

	    if(i == obj_num)
		continue;

	    /* Not in scanner rnage? */
	    if(!SWMath3DInRangePtr(
		obj, tar_obj,
		obj->scanner_range
	    ))
		continue;

	    /* If object is not owned by the connection, then check
	     * the visibility of the object
	     */
	    if(tar_obj->owner != obj_num)
	    {
		if(DBGetObjectVisibilityPtr(tar_obj) <= 0.0f)
		    continue;
	    }

	    /* Is this object hidden? */
	    if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		continue;

	    NetSendRefreshObject(
		con_num, i,
		(obj->locked_on == i) ? 1 : 0	/* Send weapon values? */
	    );
	}

	/* Send entire star chart listing */
	if(sysparm.send_starchart)
	    NetSendEntireStarChart(con_num);
}
