#include <stdio.h>

#include "../include/unvmatch.h"
#include "../include/swsoundcodes.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetHandleSetShields(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendSetShields(const int con_num, const int obj_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleSetShields(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int	shield_state = SW_SHIELD_STATE_NONE,
		obj_num = -1;
	float shield_frequency = 0.0f;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;

	/* NET_CMD_SETSHIELDS format:
	 *
	 *	obj_num shield_state shield_frequency
	 */
	sscanf(
	    arg,
	    "%i %i %f",
	    &obj_num, &shield_state, &shield_frequency
	);

	if(obj_num != con_obj_num)
	    return;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Does object have shields generators? */
	if(obj->shield_state == SW_SHIELD_STATE_NONE)
	    return;

	/* Set shield_frequency */
	if(shield_frequency > SW_FREQ_MAX)
	    shield_frequency = SW_FREQ_MAX;
	else if(shield_frequency < SW_FREQ_MIN)
	    shield_frequency = SW_FREQ_MIN;

	obj->shield_frequency = shield_frequency;


	/* Is the cloak up? */
	if(obj->cloak_state == SW_CLOAK_STATE_UP)
	{ 
	    obj->shield_state = SW_SHIELD_STATE_DOWN;
	    return;
	}


	/* Set shield visibility if shield is just raised */
	if((obj->shield_state == SW_SHIELD_STATE_DOWN) &&
	   (obj->power_max > 0.0f) &&
	   (shield_state == SW_SHIELD_STATE_UP)
	)
	{
	    obj->shield_visibility = obj->power / obj->power_max;

	    /* Send shield visibility to all connections */
	    NetSendShieldVis(-1, obj_num);
	}

	/* Play shield state sound as needed */
	if((obj->shield_state != SW_SHIELD_STATE_UP) &&
	   (shield_state == SW_SHIELD_STATE_UP)
	)
	    NetSendPlaySound(
		con_num, SW_SNDREF_SHIELD_UP,
		1.0f, 1.0f
	    );
	else if((obj->shield_state != SW_SHIELD_STATE_DOWN) &&
		(shield_state == SW_SHIELD_STATE_DOWN)
	)
	    NetSendPlaySound(
		con_num, SW_SNDREF_SHIELD_DOWN,
		1.0f, 1.0f      
	    );

	/* Set new shield state */
	obj->shield_state = shield_state;

	/* Set next object values update to now, so it gets updated now */
	next.object_values = cur_ms;
}


void NetSendSetShields(const int con_num, const int obj_num)
{
	return;
}
