#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetSendStarChartAddObject(const int con_num, const int obj_num);
void NetSendStarChartSetObjectName(const int con_num, const int obj_num);
void NetSendStarChartSetObjectEmpire(const int con_num, const int obj_num);
void NetSendStarChartRecycleObject(const int con_num, const int obj_num);
void NetSendEntireStarChart(const int con_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/* 
 *	Sends an add object to the star chart.
 */
void NetSendStarChartAddObject(const int con_num, const int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    80
	);

	/* SWEXTCMD_STARCHART_ADD_OBJECT format:
	 *
	 *	obj_num
	 *	type imgref_num size
	 *	sect_x sect_y sect_z
	 *	x y z
	 *	heading pitch bank
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i\
 %i %i %ld\
 %ld %ld %ld\
 %.4f %.4f %.4f\
 %.4f 0 0\n",
	    CS_CODE_EXT,
	    SWEXTCMD_STARCHART_ADD_OBJECT,
	    obj_num,
	    obj->type, obj->imgref, obj->size,
	    obj->sect_x, obj->sect_y, obj->sect_z,
	    obj->x, obj->y, obj->z,
	    obj->heading
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}

/*
 *	Sends a set object name on star chart.
 */
void NetSendStarChartSetObjectName(const int con_num, const int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num]; 

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 +
	    STRLEN(obj->name) +
	    80
	);

	/* SWEXTCMD_STARCHART_SET_OBJECT_NAME format:
	 *
	 *	obj_num;name
	 */
	sprintf(
	    sndbuf,
	    "%i %i %i;%s\n",
	    CS_CODE_EXT,
	    SWEXTCMD_STARCHART_SET_OBJECT_NAME,
	    obj_num,
	    obj->name
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}

/*
 *	Sends a set object empire on star chart.
 */
void NetSendStarChartSetObjectEmpire(const int con_num, const int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 +
	    STRLEN(obj->empire) +
	    80
	);
 
	/* SWEXTCMD_STARCHART_SET_OBJECT_EMPIRE format:
	 *
	 *	obj_num;empire
	 */     
	sprintf(
	    sndbuf,
	    "%i %i %i;%s\n",
	    CS_CODE_EXT,
	    SWEXTCMD_STARCHART_SET_OBJECT_EMPIRE,
	    obj_num, obj->empire
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}

/*
 *	Sends a recycle object on star chart.
 */
void NetSendStarChartRecycleObject(const int con_num, const int obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 +
	    40 +
	    80
	);

	/* SWEXTCMD_STARCHART_SET_OBJECT_RECYCLE format:
	 *
	 *	obj_num
	 */
	sprintf(
	    sndbuf,
	    "%i %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_STARCHART_SET_OBJECT_RECYCLE,
	    obj_num
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}


/*
 *	Sends the entire star chart to the specified connection.
 */
void NetSendEntireStarChart(const int con_num)
{
	int i;
	xsw_object_struct *obj;

	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if((obj->type != SW_OBJ_TYPE_HOME) &&
	       (obj->type != SW_OBJ_TYPE_AREA) &&
	       (obj->type != SW_OBJ_TYPE_WORMHOLE) &&
	       (obj->type != SW_OBJ_TYPE_ELINK)
	    )
		continue;

	    NetSendStarChartAddObject(con_num, i);
	    NetSendStarChartSetObjectName(con_num, i);
	    NetSendStarChartSetObjectEmpire(con_num, i);
	}
}
