#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/swsoundcodes.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetHandleTractorBeamLock(int con_num, const char *arg);
void NetSendTractorBeamLock(
	int con_num, int obj_num, int tar_obj_num
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleTractorBeamLock(int con_num, const char *arg)
{
	int obj_num = -1, tar_obj_num = -1;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];

	/* SWEXTCMD_TRACTORBEAMLOCK format:
	 *
	 *	obj_num tar_obj_num
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &obj_num, &tar_obj_num
	);
	if(DBIsObjectGarbage(obj_num) || (obj_num != con->object_num))
	    return;

	obj = xsw_object[obj_num];


	if(DBIsObjectGarbage(tar_obj_num))
	{
	    const int m = obj->total_tractored_objects;
	    int i, trac_obj_num, tcon_num;

	    for(i = 0; i < m; i++)
	    {
		trac_obj_num = obj->tractored_object[i];
		if(trac_obj_num < 0)
		    continue;

		tcon_num = ConGetByObject(trac_obj_num);
		if(tcon_num > -1)
                    NetSendPlaySound(
                        tcon_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
                        1.0f, 1.0f
                    );
	    }

	    /* Remove target object from tractor beam lock (unlock
	     * all for now)
	     */
	    DBObjectTractor(obj_num, -1);
	}
	else
	{
	    xsw_object_struct *tar_obj = xsw_object[tar_obj_num];

	    /* Target object is valid, check if is tractorable */
	    if(!DBIsObjectTractorablePtr(obj, tar_obj) ||
	       (obj->antimatter <= 0.0f)
	    )
	    {
		char *s = (char *)malloc(
		    STRLEN(tar_obj->name) + 80
		);
		sprintf(
		    s,
"Unable to engage tractor beam on `%s'.",
		    tar_obj->name
		);
		NetSendMessage(con_num, s);
		free(s);
		return;
	    }
	    DBObjectTractor(obj_num, tar_obj_num);
	}

	/* Update all connections about tractor beam lock */
	NetSendTractorBeamLock(-1, obj_num, tar_obj_num);

	/* Play the tractor beam sound */
	if(tar_obj_num > -1)
            NetSendPlaySound(
                con_num, SW_SNDREF_TRACTOR_BEAM_ENGAGE,
                1.0f, 1.0f
            );
	else
            NetSendPlaySound(
                con_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
                1.0f, 1.0f
            );

	/* Play the tractor beam engage sound to the target object
	 * as needed
	 */
	if(tar_obj_num > -1)
	{
	    const int tcon_num = ConGetByObject(tar_obj_num);
	    if(tcon_num > -1)
                NetSendPlaySound(
                    tcon_num, SW_SNDREF_TRACTOR_BEAM_ENGAGE,
                    1.0f, 1.0f
                );
	}
}


void NetSendTractorBeamLock(
	int con_num, int obj_num, int tar_obj_num
)
{
	char *sndbuf;
	xsw_object_struct *obj;  

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* SWEXTCMD_TRACTORBEAMLOCK format:
	 *
	 *	obj_num tar_obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_TRACTORBEAMLOCK,
	    obj_num, tar_obj_num
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
