#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "net.h"
#include "swserv.h"


static void NetVersionCheck(int condescriptor, int major, int minor);

void NetHandleVersion(int condescriptor, const char *arg);
void NetSendVersion(int condescriptor);


/*
 *	Checks if the specified CS network version is insufficient
 *	and reports it as needed.
 *
 *	This function should be updated on each new release in order
 *	to determine when older protocol versions become out of date.
 */
static void NetVersionCheck(int condescriptor, int major, int minor)
{
#if ((PROG_VERSION_MAJOR != 2) || (PROG_VERSION_MINOR != 5))
# warning NetVersionCheck() protocol checking needs to be updated
#endif
	if((major < PROG_VERSION_MAJOR) ||
	   ((major == PROG_VERSION_MAJOR) && (minor < PROG_VERSION_MINOR))
	)
	{
	    char *s = (char *)malloc(40 + 40 + 40 + 80);
	    sprintf(
		s,
"ShipWars network protocol version %i.%i.x or newer required.",
		PROG_VERSION_MAJOR, PROG_VERSION_MINOR
	    );
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_WARNING,
		s
	    );
	    free(s);
	}
}


void NetHandleVersion(int condescriptor, const char *arg)
{
	int status, major = 0, minor = 0, release = 0;

	status = sscanf(
	    arg,
"%i %i %i",
	    &major, &minor, &release
	);
	if(status >= 2)
	    NetVersionCheck(condescriptor, major, minor);
}

void NetSendVersion(int condescriptor)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* Format CS_CODE_VERSION:
	 *
	 *	major minor release
	 */
	sprintf(
	    sndbuf,
	    "%i %i %i %i\n",
	    CS_CODE_VERSION,
	    PROG_VERSION_MAJOR,
	    PROG_VERSION_MINOR,
	    PROG_VERSION_RELEASE
	);
	NetSendDataToConnection(condescriptor, sndbuf);
	free(sndbuf);
}
