#include <stdio.h>
#include <stdlib.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#include "weapons.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


void NetWeaponDisarmSyncObjectValues(
	xsw_object_struct *obj, int obj_num
);
void NetHandleWeaponDisarm(int con_num, const char *arg);
void NetSendWeaponDisarm(
	int con_num, int src_obj_num, int tar_obj_num
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sends new values of weapon object to all nearby connections.
 */
void NetWeaponDisarmSyncObjectValues(
	xsw_object_struct *obj, int obj_num
)
{
	int i, con_obj_num;
	sws_connection_struct *con;
	xsw_object_struct *con_obj;

	/* Check if marked hidden from connection */
	if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
	    return;

	/* Iterate through all connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    con_obj_num = con->object_num;
	    if(DBIsObjectGarbage(con_obj_num))
		continue;

	    con_obj = xsw_object[con_obj_num];

	    /* Check if objects are valid and in range */
	    if(!SWMath3DInRangePtr(
		con_obj, obj,
		con_obj->scanner_range
	    ))
		continue;

	    /* Not owned objects, check for visibility */
	    if(obj->owner != con_obj_num)
	    {
		if(DBGetObjectVisibilityPtr(obj) <= 0.00)
		    continue;
	    }

	    /* Send object values & maximums */
	    NetSendObjectMaximums(i, obj_num);
	    NetSendObjectValues(i, obj_num);
	}
}

void NetHandleWeaponDisarm(int con_num, const char *arg)
{
	int i, src_obj_num = -1, tar_obj_num = -1, con_obj_num;
	xsw_object_struct *wep_obj, *con_obj;
	sws_connection_struct *con = connection[con_num];

	con_obj_num = con->object_num;
	if(DBIsObjectGarbage(con_obj_num))
	    return;

	con_obj = xsw_object[con_obj_num];
 
	/* SWEXTCMD_WEPDISARM format:
	 *
	 *	src_obj_num tar_obj_num
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &src_obj_num, &tar_obj_num
	);

	/* Connection must own object */
	if(src_obj_num != con_obj_num)
	    return;

	if(tar_obj_num < 0)
	{
	    /* Disarm all weapons that connection object owns */
	    for(i = 0; i < total_objects; i++)
	    {
		wep_obj = xsw_object[i];
		if(wep_obj == NULL)
		    continue;

		/* Not a weapon (skip stream weapons)? */
		if(wep_obj->type != SW_OBJ_TYPE_WEAPON)
		    continue;

		/* Source object owns weapon object? */
		if(wep_obj->owner == src_obj_num)
		{
		    /* Disarm this weapon */
		    WeaponsDisarmWeapon(i);

		    NetSendObjectName(con_num, i); 
		    NetSendWeaponDisarm(
			con_num,
			src_obj_num,
			i
		    );
		    NetWeaponDisarmSyncObjectValues(
			wep_obj,
			i
		    );
		}
	    }
	}
	else
	{
	    /* Disarm specific weapon if the connection's object owns
	     * it
	     */
	    if(!DBIsObjectGarbage(tar_obj_num))
	    {
		wep_obj = xsw_object[tar_obj_num];

		/* Source object owns weapon object? */
		if((wep_obj->owner == src_obj_num) &&
		   (wep_obj->type == SW_OBJ_TYPE_WEAPON)
		)
		{
		    /* Disarm this weapon */
		    WeaponsDisarmWeapon(tar_obj_num);

		    NetSendObjectName(con_num, tar_obj_num);
		    NetSendWeaponDisarm(
			con_num,
			src_obj_num,
			tar_obj_num
		    );
		    NetWeaponDisarmSyncObjectValues(
			wep_obj,
			tar_obj_num
		    );
		}
	    }
	}
}


void NetSendWeaponDisarm(
	int con_num, int obj_num, int tar_obj_num
)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* SWEXTCMD_WEPDISARM format:
	 *
	 *	obj_num tar_obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_WEPDISARM,
	    obj_num, tar_obj_num
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
