#include <stdio.h>

#include "../include/unvmatch.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


void NetHandleWormHoleEnter(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


void NetHandleWormHoleEnter(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	float d;
	int obj_num = -1, tar_obj_num = -1, dest_obj_num;
	xsw_object_struct *obj, *tar_obj, *dest_obj;
	sw_vector_compoent_struct *src_attitude;
	sws_connection_struct *con = connection[con_num];

	/* SWEXTCMD_WORMHOLEENTER format:
	 *
	 *	obj_num tar_obj_num
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &obj_num, &tar_obj_num
	);

	/* Connection must own the source object */
	if(obj_num != con->object_num)
	    return;

	/* Get pointers to source and target objects */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	if(DBIsObjectGarbage(tar_obj_num))
	    return;

	tar_obj = xsw_object[tar_obj_num];

	/* Target is not a wormhole? */
	if(tar_obj->type != SW_OBJ_TYPE_WORMHOLE)
	    return;

	/* Get linked destination (intercepting object) of target
	 * object (the other end of the wormhole)
	 */
	dest_obj_num = tar_obj->intercepting_object;
	if(DBIsObjectGarbage(dest_obj_num))
	    return;

	dest_obj = xsw_object[dest_obj_num];

	/* Linked to destination need not be a worm hole */


	/* Move source object to destination object's position with
	 * some offsets
	 */
	src_attitude = &obj->attitude_vector_compoent;

	/* Move source object in its heading direction to
	 * away from the destination object based on both object's
	 * size (in Real units)
	 */
	d = (float)dest_obj->size / 1000.0;
	   
	obj->x = dest_obj->x + (d * src_attitude->i);
	obj->y = dest_obj->y + (d * src_attitude->j);
	obj->z = dest_obj->z;

	/* Move the object to a position that is away from the
	 * destination object
	 */
	if((obj->sect_x != dest_obj->sect_x) ||
	   (obj->sect_y != dest_obj->sect_y) ||
	   (obj->sect_z != dest_obj->sect_z)
	)
	{
	    /* Change the object's sector and update related resources */
	    DBObjectDoSetSector(
		obj_num,
		dest_obj->sect_x,
		dest_obj->sect_y,
		dest_obj->sect_z,
		0			/* No position wrapping */
	    );
	    NetSendObjForceSect(con_num, obj_num, cur_ms, cur_sec, tc);
	}
	NetSendObjectForcePose(con_num, obj_num, cur_ms, cur_sec, tc);

	/* Send create for the wormhole's destination */
	NetSendCreateObject(con_num, dest_obj_num);
}
