#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"
#include "../include/swsoundcodes.h"
#include "../include/swnetcodes.h"
#include "../include/swexporteventscodes.h"

#include "weapons.h"
#include "effects.h"
#include "net.h"
#include "sysparm.h"
#ifdef PLUGIN_SUPPORT
#include "plugins.h"
#endif
#include "swserv.h"


static int REngFreqCmp(float f1, float f2);
static float REngTurnAngle(
	float theta_cur, float theta_to, float theta_limit,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static void REngMove(
	sw_legend_struct *legend,
	const int obj_num,
	const char allow_sect_change, const char is_tractored,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static void REngHeading(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngThrottle(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngThrust(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngLocation(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngVisibility(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngShieldVisibility(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int REngMortality(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngAnimated(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int REngAutoDestruct(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngWeaponsLock(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngTractorBeam(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngIntercept(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngPowerCore(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngDamageControl(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngDoRespawn(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int REngDoHit(
	sw_legend_struct *legend,
	const int wobj_num, const int tobj_num,
	const float damage_hull,
	const float damage_shields,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int REngContactCheck(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int REngAsteroidFieldDamage(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int REngAI(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static int REngUpdateObject(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec,
	const float tc		/* Time compensation coefficient */
);

int REngInit(void);
void REngManage(
	sw_legend_struct *legend,
	const time_t cur_ms, const time_t cur_sec,
	const float tc		/* Time compensation coefficient */
);
void REngShutdown(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the frequencies are the same up to 2 decimal
 *	accuracy.
 */
static int REngFreqCmp(float f1, float f2)
{
	float df;

	/* Clip the frequency to be within the frequency range */
	if(f1 < SW_FREQ_MIN)
	    f1 = SW_FREQ_MIN;
	if(f1 > SW_FREQ_MAX)
	    f1 = SW_FREQ_MAX;

	if(f2 < SW_FREQ_MIN)
	    f2 = SW_FREQ_MIN;
	if(f2 > SW_FREQ_MAX)
	    f2 = SW_FREQ_MAX;

	/* Get absolute value difference */
	df = f2 - f1;
	if(df < 0.0f)
	    df *= -1.0f;

	/* Compare up to 2 decimal places */
	if(df < 0.01f)
	    return(1);
	else
	    return(0);
}


/*
 *	Turns the angle theta_cur towards the angle theta_to but
 *	no more than theta_limit (all units in radians).
 */
static float REngTurnAngle(
	float theta_cur, float theta_to, float theta_limit,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	const float limit_tc = theta_limit * tc;
	float delta = SWMathAngleDelta(theta_to, theta_cur);

	/* Limit delta angle */
	if(delta > limit_tc)
	    delta = limit_tc;
	else if(delta < -limit_tc) 
	    delta = -limit_tc;

	/* Turn current angle */
	theta_cur += delta;

	return(SANITIZERADIANS(theta_cur));
}


/*
 *	Moves the object by calculating one quantom shift
 *	of its 3D coordinates.
 * 
 *	If allow_sect_change is true then if the object moves out
 *	if the coordinate bounds, it will change sectors and warp to
 *	the other side.
 */
static void REngMove(
	sw_legend_struct *legend,
	const int obj_num,
	const char allow_sect_change, const char is_tractored,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char ch_sect_px, ch_sect_nx, ch_sect_py, ch_sect_ny;
	float thrust_dir, thrust_mag, thrust_magc;
	float velocity_mag, velocity_magc;
	float prev_x, prev_y, prev_z;
	float x_vel, y_vel, z_vel;	/* Velocity delta */
	float x_velc, y_velc, z_velc;	/* Velocity delta /w time compensation */
	float x_thr, y_thr, z_thr;	/* Thrust */
	float x_thrc, y_thrc, z_thrc;	/* Thrust delta /w time compensation */
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip non-moving types of object if this object is not being
	 * tractored
	 *
	 * This includes skipping players, when not tractored, their
	 * movements are updated from the client
	 */
	if(!is_tractored)
	{
	    if((obj->type == SW_OBJ_TYPE_STATIC) ||
	       (obj->type == SW_OBJ_TYPE_PLAYER) ||
	       (obj->type == SW_OBJ_TYPE_HOME) ||
	       (obj->type == SW_OBJ_TYPE_AREA) ||
	       (obj->type == SW_OBJ_TYPE_ANIMATED) ||
	       (obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	       (obj->type == SW_OBJ_TYPE_ELINK)
	    )
		return;
	}

	/* Stream weapons have the location of their owner */
	if((obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	{
	    const int owner_object = obj->owner;
	    if(!DBIsObjectGarbage(owner_object))
	    {
		xsw_object_struct *owner_obj = xsw_object[owner_object];
		obj->sect_x = owner_obj->sect_x;
		obj->sect_y = owner_obj->sect_y;
		obj->sect_z = owner_obj->sect_z;
		obj->x = owner_obj->x;
		obj->y = owner_obj->y;
		obj->z = owner_obj->z;
	    }
	    return;
	}


	/* Begin moving the object */

	/* Sanitize velocity heading */
	obj->velocity_heading = SANITIZERADIANS(obj->velocity_heading);

	/* Calculate movement from velocity (momentum)
	 * Fetch velocity_mag, with compensation for lapsed time
	 */
	velocity_magc = obj->velocity * tc;
	velocity_mag = obj->velocity;

	/* Get momentum (velocity) vector compoents (with and
	 * without time compensation) from momentum unit vector
	 * compoents.
	 */
	x_velc = obj->momentum_vector_compoent.i * velocity_magc;
	y_velc = obj->momentum_vector_compoent.j * velocity_magc;
	z_velc = obj->momentum_vector_compoent.k * velocity_magc;

	x_vel = obj->momentum_vector_compoent.i * velocity_mag;
	y_vel = obj->momentum_vector_compoent.j * velocity_mag;
	z_vel = obj->momentum_vector_compoent.k * velocity_mag;


	/* Calculate (not object relative) thrust direction */
	thrust_dir = SANITIZERADIANS(
	    obj->heading +
	    obj->thrust_dir +
	    (1.0f * PI)
	);


	/* Calculate thrust magnitude, taking into account the
	 * object's thrust power, current power, and power
	 * purity
	 *
	 * The power coeff is in the range of 0.5 to 1.0, this is
	 * due to the fact that if there is no power the vessel cannot
	 * move and many players complain that this is unfair
	 */
	if(obj->power_max > 0.0f)
	{
	    const float power_coeff = (obj->power /
		obj->power_max / 2.0f) + 0.5f;
	    thrust_mag = obj->thrust * obj->power_purity *
		power_coeff;
	}
	else
	{
	    thrust_mag = obj->thrust * obj->power_purity;
	}

	/* Calculate thrust magnitude with tc */
	thrust_magc = thrust_mag * tc;

	/* Get thrust vector compoents with and without time
	 * compensation
	 */
	x_thrc = thrust_magc * (float)sin((double)thrust_dir);
	y_thrc = thrust_magc * (float)cos((double)thrust_dir);
	z_thrc = 0.0f;

	x_thr = thrust_mag * (float)sin((double)thrust_dir);
	y_thr = thrust_mag * (float)cos((double)thrust_dir);
	z_thr = 0.0f;


	/* Move object by adding the calculated velocity and thrust
	 * vector compoents
	 */
	prev_x = obj->x;
	prev_y = obj->y;
	prev_z = obj->z;

	obj->x += x_velc + x_thrc;
	obj->y += y_velc + y_thrc;
/*	obj->z += z_velc + z_thrc; */
/* Force set z to 0 */
	obj->z = 0.0f;

	/* Update the odometer */
	obj->odometer += SWMath3DDistance(
	    obj->x - prev_x,
	    obj->y - prev_y,
	    obj->z - prev_z
	);

	/* Do sector change as needed */
	if(allow_sect_change)
	{
	    ch_sect_px = 0;
	    ch_sect_nx = 0;
	    ch_sect_py = 0;
	    ch_sect_ny = 0;

	    if(obj->x < legend->sect_x_min)
	    {
		obj->x += legend->sect_x_len;
		obj->sect_x -= 1l;
		ch_sect_nx = 1;		/* Negative sect x change */
	        NetSendObjSect(-1, obj_num, cur_ms, cur_sec, tc);
	    }
	    else if(obj->x > legend->sect_x_max)
	    {
		obj->x -= legend->sect_x_len;
		obj->sect_x += 1l;
		ch_sect_px = 1;		/* Positive sect x change */
		NetSendObjSect(-1, obj_num, cur_ms, cur_sec, tc);
	    }

	    if(obj->y < legend->sect_y_min)
	    {
		obj->y += legend->sect_y_len;
		obj->sect_y -= 1l;
		ch_sect_ny = 1;		/* Negative sect y change */
		NetSendObjSect(-1, obj_num, cur_ms, cur_sec, tc);
	    }
	    else if(obj->y > legend->sect_y_max)
	    {
		obj->y -= legend->sect_y_len;
		obj->sect_y += 1l;
		ch_sect_py = 1;		/* Positive sect y change */
		NetSendObjSect(-1, obj_num, cur_ms, cur_sec, tc);
	    }

#if 0
	    if(obj->z < legend->sect_z_min)
	    {
		obj->z += legend->sect_z_len;
		obj->sect_z -= 1;
		ch_sect_nz = 1;
		NetSendObjSect(-1, obj_num, cur_ms, cur_sec, tc);
	    }
	    else if(obj->z > legend->sect_z_max)
	    {
		obj->z -= legend->sect_z_len;
		obj->sect_z += 1;
		ch_sect_pz = 1;
		NetSendObjSect(-1, obj_num, cur_ms, cur_sec, tc);
	    }
#endif

	    /* Change sector for tractored objects */ 
	    if(ch_sect_nx || ch_sect_px ||
	       ch_sect_ny || ch_sect_py
/*
	       ch_sect_nz || ch_sect_pz
 */
	    )
	    {
		int i, trac_obj_num;
		xsw_object_struct *trac_obj;

		for(i = 0; i < obj->total_tractored_objects; i++)
		{
		    trac_obj_num = obj->tractored_object[i];
		    if(DBIsObjectGarbage(trac_obj_num))
			continue;

		    trac_obj = xsw_object[trac_obj_num];

		    trac_obj->sect_x = obj->sect_x;
		    trac_obj->sect_y = obj->sect_y;
/*		    trac_obj->sect_z = obj->sect_z; */

		    /* Move tractored object coordinates to exactly
		     * where the source object is
		     */
		    if(ch_sect_nx)
			trac_obj->x = obj->x;
		    if(ch_sect_px)
			trac_obj->x = obj->x;

		    if(ch_sect_ny)
			trac_obj->y = obj->y;
		    if(ch_sect_py)
			trac_obj->y = obj->y;


#if 0
		    if(ch_sect_nx)
			trac_obj->x += legend->sect_x_len;
		    if(ch_sect_px)
			trac_obj->x -= legend->sect_x_len;

		    if(ch_sect_ny)
			trac_obj->y += legend->sect_y_len;
		    if(ch_sect_py)
			trac_obj->y -= legend->sect_y_len;

		    if(ch_sect_nz)
			trac_obj->z += legend->sect_z_len;
		    if(ch_sect_pz)
			trac_obj->z -= legend->sect_z_len;
#endif
		    NetSendObjSect(-1, trac_obj_num, cur_ms, cur_sec, tc);
		}
	    }
	}


	/* Record new velocity and velocity_heading on object
	 *
	 * This value must not have the time compensation modifications
	 * to it
	 */
	obj->velocity_heading = SWMathCoordinateDeltaVector(
	    x_vel + x_thr,
	    y_vel + y_thr
	);
	SWMathSetUnitVector2D(
	    &obj->momentum_vector_compoent,
	    obj->velocity_heading
	);

	obj->velocity = SWMath3DDistance(
	    x_vel + x_thr,
	    y_vel + y_thr,
	    0.0f
	);
}

/*
 *	Updates the heading of the object.
 */
static void REngHeading(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char oversight = 1;
	int tar_obj_num;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip objects that do not turn */
	if((obj->type == SW_OBJ_TYPE_STATIC) ||
	   (obj->type == SW_OBJ_TYPE_HOME) ||
	   (obj->type == SW_OBJ_TYPE_AREA) ||
	   (obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (obj->type == SW_OBJ_TYPE_ELINK)
	)
	   return;

	/* Skip player objects that are not intercepting anything,
	 * when they are not intercepting, their heading is updated by
	 * the client
	 */
	if((obj->type == SW_OBJ_TYPE_PLAYER) &&
	   (obj->intercepting_object < 0)
	)
	   return;

	/* Disable oversight calculations for stream weapons */
	if(obj->type == SW_OBJ_TYPE_STREAMWEAPON)
	    oversight = 0;

	/* Is object intercepting something? */
	tar_obj_num = obj->intercepting_object;
	if(!DBIsObjectGarbage(tar_obj_num))
	{
	    /* Get rel_heading to the object we're intercepting */
	    float bearing;
	    xsw_object_struct *tar_obj = xsw_object[tar_obj_num];

	    if(SWMath3DInSameSectorPtr(obj, tar_obj))
	    {
		/* In same sector */
		float src_mx, src_my, tar_mx, tar_my, delta;

		/* Calculate the distance between the two objects */
		const float d = SWMath3DDistance(
		    tar_obj->x - obj->x,
		    tar_obj->y - obj->y,
		    tar_obj->z - obj->z
		);

		/* Calculate delta time (in cycles) for the source
		 * object to reach the target object. A bunch
		 * of errors may be introduced if this is the
		 * start of an intercept, but that shouldn't
		 * be too much as the intercept path neatens up.
		 */
		const float dt = (obj->velocity > 0.0f) ?
		    (d / obj->velocity) : 0.0f;

		/* Get current position of source object */
		src_mx = obj->x;
		src_my = obj->y;

		if(oversight && (tar_obj->velocity != 0.0f))
		{
		    /* Get calculated position of target object using
		     * the target object's current momentum and multiply
		     * that by the time (in cycles) it will take for
		     * the source object to reach the target object
		     * at both object's current positions and
		     * velocities
		     */
		    tar_mx = tar_obj->x +
		        (tar_obj->momentum_vector_compoent.i *
		        tar_obj->velocity * dt);
		    tar_my = tar_obj->y +
		        (tar_obj->momentum_vector_compoent.j *
		        tar_obj->velocity * dt);
		}
		else
		{
		    tar_mx = tar_obj->x;
		    tar_my = tar_obj->y;
		}

		/* Calculate bearing from the source object's current
		 * position to target object's predicted position
		 */
		bearing = SWMathCoordinateDeltaVector(
		    tar_mx - src_mx,
		    tar_my - src_my
		);

		if(oversight && (obj->velocity != 0.0f))
		{
		    /* Calculate the shorter delta between the bearing
		     * and the source object's velocity then apply it
		     * to the bearing so it will compensate for the
		     * source object's velocity
		     */
		    delta = SWMathAngleDelta(bearing, obj->velocity_heading);
		    if((delta <= (0.40 * PI)) && (delta >= (-0.40 * PI)))
			bearing = SANITIZERADIANS(bearing + delta);
		}
#if 0
printf("theta1=%.0f theta2=%.0f bearing=%.0f\n",
 RADTODEG(theta1), RADTODEG(theta2), RADTODEG(bearing)
);
#endif
	    }
	    else
	    {
		/* In different sectors */
		const float	sect_width = legend->sect_x_max - legend->sect_x_min,
				sect_height = legend->sect_y_max - legend->sect_y_min,
				dx = tar_obj->x - obj->x,
				dy = tar_obj->y - obj->y,
				sect_dx = (float)tar_obj->sect_x - (float)obj->sect_x,
				sect_dy = (float)tar_obj->sect_y - (float)obj->sect_y;

		bearing = SWMathCoordinateDeltaVector(
		    dx + (sect_dx * sect_width),
		    dy + (sect_dy * sect_height)
		);
	    }

	    /* Turn heading */
	    obj->heading = REngTurnAngle(
		obj->heading, bearing,
		obj->turnrate * 100.0f,
		cur_ms, cur_sec, tc
	    );
	}

	/* Sanitize heading */
	obj->heading = SANITIZERADIANS(obj->heading);
	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);
}


/*
 *	Nothing is done here, this function is just for client side.
 */
static void REngThrottle(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip static objects */
	if(obj->type == SW_OBJ_TYPE_STATIC)
	   return;

	/* Skip player objects, their throttle is updated on client */
	if(obj->type == SW_OBJ_TYPE_PLAYER)
	   return;

}

/*
 *	Thrust, antimatter consumption, drag, and other calculations
 *	of velocity are performed here.
 */
static void REngThrust(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	float throttle, thrust_power;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* We want to update all types of objects, since there is
	 * no telling if an object should have power decrementation
	 *
	 * A civilization on a planet for instance, would use up
	 * antimatter, so throttle applied would simulate antimatter
	 * usage
	 */

	/* Get throttle and thrust_power */
	throttle = obj->throttle;

	if(obj->engine_state != SW_ENGINE_STATE_ON)
	    throttle = 0.0f;

	if(throttle > 0.0f)
	{
	    const float throttle_coeff = throttle * throttle;

	    /* Get thrust power */
	    thrust_power = obj->thrust_power;

	    /* Consume antimatter based on throttle */
	    obj->antimatter = MAX(
		(obj->antimatter - (throttle_coeff *
		    legend->engine_antimatter_rate * tc)
		),
		0.0f
	    );
	}
	else
	{
	    thrust_power = 0.0f;
	    throttle = 0.0f;
	}


	/* Calculate the actual thrust output based on the calculated
	 * throttle and thrust power
	 */
	obj->thrust = MAX(
	    (throttle * thrust_power * tc),
	    0.0f
	);

	/* Reduce the thrust output if out of antimatter */
	if(obj->antimatter <= 0.0f)
	    obj->thrust *= legend->no_antimatter_thrust_coeff;


	/* Add drag to the velocity based on the object's current
	 * velocity, velocity max, and thrust power
	 */
	if((obj->velocity_max > 0.0f) && (obj->thrust_power > 0.0f))
	    obj->velocity -= (obj->velocity /
		obj->velocity_max) * obj->thrust_power * tc;
	else 
	    obj->velocity = 0.0f;

	/* If velocity is such a small number then set it to 0 */
	if(obj->velocity < 0.00003f)
	    obj->velocity = 0.0f;

	/* Consume antimatter if cloak is up */
	if(obj->cloak_state == SW_CLOAK_STATE_UP)
	{
	    const float size_ru = (float)obj->size / 1000.0f;

	    /* Consume antimatter based on the object's size */
	    obj->antimatter = MAX(
		(obj->antimatter - (size_ru *
		    legend->cloak_antimatter_rate * tc)
		),
		0.0f
	    );

	    /* Lower the cloak if out of antimatter */
	    if(obj->antimatter == 0.0f)
	    {
		int i;
		xsw_object_struct *obj2;
		sws_connection_struct *con2;
		const int con_num = ConGetByObject(obj_num);

		obj->cloak_state = SW_CLOAK_STATE_DOWN;
		obj->last_cloak = cur_ms;

		if(con_num > -1)
		    NetSendPlaySound(
			con_num, SW_SNDREF_CLOAK_DOWN,
			1.0f, 1.0f
		    );

		/* Refresh this decloaked object to nearby connections */
		for(i = 0; i < total_connections; i++)
		{
		    con2 = connection[i];
		    if(con2 == NULL)
			continue;

		    if(DBIsObjectGarbage(con2->object_num))
			continue;

		    if(obj_num == con2->object_num)
			continue;

		    obj2 = xsw_object[con2->object_num];

		    /* Not in scanner rnage? */
		    if(!SWMath3DInRangePtr(
			obj, obj2,
			obj2->scanner_range
		    ))
			continue;

		    NetSendRefreshObject(i, obj_num, 1);
		}
	    }
	}
}


/*
 *	Updates the object's location type.
 */
static void REngLocation(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, area_obj_num = -1;
	float d = 0.0f;			/* In screen units */
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj,
				*area_obj = NULL;

	/* Skip area objects */
	if(obj->type == SW_OBJ_TYPE_AREA)
	    return;

	/* Iterate through the objects list backwards and check
	 * for non-territory areas first
	 */
	for(i = total_objects - 1; i >= 0; i--)
	{
	    tar_obj = xsw_object[i];
	    if(tar_obj == NULL)
		continue;

	    /* Not an area object? */
	    if(tar_obj->type != SW_OBJ_TYPE_AREA)
		continue;

	    /* Skip territory area objects */
	    if(tar_obj->loc_type == SW_LOC_TYPE_SPACE_TERRITORY)
		continue;

	    /* Skip itself */
	    if(tar_obj == obj)
		continue;

	    /* Not in the same sector? */
	    if((obj->sect_x != tar_obj->sect_x) ||
	       (obj->sect_y != tar_obj->sect_y) ||
	       (obj->sect_z != tar_obj->sect_z)
	    )
		continue;

	    /* Calculate the distance in screen units */
	    d = SWMath3DDistance(
		tar_obj->x - obj->x,
		tar_obj->y - obj->y,
		tar_obj->z - obj->z
	    ) * 1000.0f;
	    if(d > tar_obj->size)
		continue;

	    area_obj = tar_obj;
	    area_obj_num = i;
	    break;
	}

	/* If the object was not in a non-territory area object then
	 * check if the object is in a territory area object
	 */
	if(area_obj == NULL)
	{
	    for(i = total_objects - 1; i >= 0; i--)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Not an area object? */
		if(tar_obj->type != SW_OBJ_TYPE_AREA)
		    continue;

		/* Not a territory area object? */
		if(tar_obj->loc_type != SW_LOC_TYPE_SPACE_TERRITORY)
		    continue;

		/* Skip itself */
		if(tar_obj == obj)
		    continue;

		/* Not in the same sector? */
		if((obj->sect_x != tar_obj->sect_x) ||
		   (obj->sect_y != tar_obj->sect_y) ||
		   (obj->sect_z != tar_obj->sect_z)
		)
		    continue;

		/* Calculate the distance in screen units */
		d = SWMath3DDistance(
		    tar_obj->x - obj->x,
		    tar_obj->y - obj->y,
		    tar_obj->z - obj->z
		) * 1000.0f;
		if(d > tar_obj->size)
		    continue;

		area_obj = tar_obj;
		area_obj_num = i;
		break;
	    }
	}

	/* Is this object in an area? */
	if(area_obj != NULL)
	{
	    /* Is now in an area
	     *
	     * Was the object not in an area?
	     */
	    if(obj->loc_obj_num != area_obj_num)
	    {
		switch(area_obj->loc_type)
		{
		  case SW_LOC_TYPE_SPACE_TERRITORY:
/* TODO */
		    break;
		  case SW_LOC_TYPE_NEBULA:
		    break;
		  case SW_LOC_TYPE_ASTEROID_FIELD:
		    break;
		  case SW_LOC_TYPE_PULSAR:
		    break;
		}

		/* Set the object's new location */
	        obj->loc_type = area_obj->loc_type;
		obj->loc_obj_num = area_obj_num;
	    }
	}
	else
	{
	    /* Now not in an area
	     *
	     * Was the object in an area?
	     */
	    if(obj->loc_obj_num > -1)
	    {
		obj->loc_type = SW_LOC_TYPE_SPACE;
		obj->loc_obj_num = -1;
	    }
	}
}


/*
 *	Updates the object's current visiblity by checking its
 *	shield and cloak state.
 *
 *	Does not take into account the object's current location.
 */
static void REngVisibility(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	float v;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Are the shields up? */
	if(obj->shield_state == SW_SHIELD_STATE_UP)
	{
	    if(obj->cur_visibility < 1.0f)
	    {
		obj->cur_visibility = MIN(
		    obj->cur_visibility + (0.01f * tc),
		    1.0f
		);
	    }
	}
	/* Is the cloak up? */
	else if(obj->cloak_state == SW_CLOAK_STATE_UP)
	{
	    if(obj->power_max > 0.0f)
		v = obj->visibility - (obj->cloak_strength *
		    obj->power_purity * obj->power /
		    obj->power_max);
	    else
		v = obj->visibility;
		    
	    if(obj->cur_visibility > v)
		obj->cur_visibility = MAX(
		    (obj->cur_visibility - (0.01f * tc)),
		    v
		);
	}
	/* All else adjust the visibility to normal */
	else
	{
	    if(obj->cur_visibility > obj->visibility)
	    {
		obj->cur_visibility = MAX(
		    obj->cur_visibility - (0.01f * tc),
		    obj->visibility
		);
	    }
	    else if(obj->cur_visibility < obj->visibility)
	    {   
		obj->cur_visibility = MIN(
		    obj->cur_visibility + (0.01f * tc),
		    obj->visibility
		);
	    }
	}

	/* Sanitize */
	if(obj->cur_visibility < 0.0f)
	    obj->cur_visibility = 0.0f;
 	else if(obj->cur_visibility > 1.0f)
	     obj->cur_visibility = 1.0f;
}

/*
 *	Updates the object's shield visibility.
 */
static void REngShieldVisibility(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	float x;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip if shields not visable */
	if(obj->shield_visibility <= 0.0f)
	    return;

	/* Calculate deduction */
	x = (float)((float)SW_CYCLE_LAPSE_MS /
	    (float)SW_SHIELD_VISIBILITY_INTERVAL) * tc; 

	/* Reduce shield visibility */
	obj->shield_visibility -= x;

	/* Sanitize shield_visibility */
	if(obj->shield_visibility < 0.0f)
	    obj->shield_visibility = 0.0f;
}

/*
 *	Checks if the object has lived past its life span and
 *	recycles it as needed.
 *
 *	Returns true if the object has been recycled.
 */
static int REngMortality(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char object_must_die = 0;
	xsw_object_struct *obj = xsw_object[obj_num];
	xsw_animation_struct *anim;

	/* Handle mortality by object type */
	switch(obj->type)
	{
	  case SW_OBJ_TYPE_ANIMATED:
	    /* Animated objects die when they have a limited (non zero)
	     * cycle_times and that the number of times they have
	     * cycled exceeds that
	     */
	    anim = &obj->animation;
	    if(anim->cycle_times > 0)
	    {
		if(anim->cycle_count >= anim->cycle_times)
		    object_must_die = 1;
	    }
	    break;

	  case SW_OBJ_TYPE_PLAYER:
	    /* Player objects do not die regardless */
	    break;

	  default:
	    /* All other object types, handle by life span
	     *
	     * If the object has a defined (non-zero and positive) life
	     * span then their birth time is checked with current time
	     * to see if they must die
	     */
	    if((obj->lifespan > 0l) && (obj->birth_time > 0l))
	    {
		/* Calculate age */
		const time_t dt = cur_ms - obj->birth_time;

		/* Age positive and exceeded life span? */
		if((dt < 0l) || (dt > obj->lifespan))
		    object_must_die = 1;
	    }
	    break;
	}


	/* Has this object died and we must recycle it? */
	if(object_must_die)
	{
	    /* Recycle this object */
	    DBRecycleObject(obj_num, cur_ms, cur_sec, tc);

	    /* Notify all connections about this recycle */
	    NetSendRecycleObject(-1, obj_num);

	    return(1);
	}
	else
	{
	    return(0);
	}
}

/*
 *	Updates animated object's animation frame position.
 */
static void REngAnimated(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	time_t dt;
	xsw_object_struct *obj = xsw_object[obj_num];
	xsw_animation_struct *anim = &obj->animation;

	/* Time to increment frame? */
	dt = anim->last_interval + anim->interval;
	if(dt <= cur_ms)
	{
	    anim->current_frame++;

	    /* Looped? */
	    if(anim->current_frame >= anim->total_frames)
	    {
		anim->current_frame = 0;

		/* Increment cycled times only if there is a cycle limit */
		if(anim->cycle_times > 0)
		    anim->cycle_count++;
	    }

	    anim->last_interval = cur_ms;
	}
}

/*
 *	Auto-destruct.
 *
 *	Returns true if an object has been recycled.
 */
static int REngAutoDestruct(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int objects_recycled = 0;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Auto destruct not set or this object is marked as
	 * destroyed?
	 */
	if((obj->auto_destruct_next <= 0l) ||
	   (obj->ext_flags & SWS_OBJ_FLAG_DESTROYED)
	)
	    return(objects_recycled);

	/* Time to auto-destruct? */
	if(obj->auto_destruct_next <= cur_sec)
	{
	    const float dmg_radius = legend->auto_destruct_damage_range;
	    char *obj_name;

	    /* Reset auto-destruct marker */
	    obj->auto_destruct_next = 0l;

	    /* If this object can be destroyed then mark it as been
	     * destroyed so subsequent destroy functions do not
	     * destroy it again
	     */
	    if(obj->permission.uid > SWS_ACCESS_UID_DESTROY_IMMUNE)
		obj->ext_flags |= SWS_OBJ_FLAG_DESTROYED;

	    /* If the object has power left then apply damage to
	     * nearby objects
	     */
	    if(obj->power > 0.0f)
	    {
		int i, tar_type;
		float	distance,
			power_coeff,
			power = obj->power;
		xsw_object_struct *tar_obj;

		for(i = 0; i < total_objects; i++)
		{
		    tar_obj = xsw_object[i];
		    if(tar_obj == NULL)
			continue;

		    if(i == obj_num)
			continue;

		    if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* Only hit players and controlled objects */
		    tar_type = tar_obj->type;
		    if(tar_obj->ext_flags & SWS_OBJ_FLAG_WEP_DESTROYABLE)
		    {
			if((tar_type != SW_OBJ_TYPE_WEAPON) &&
			   (tar_type != SW_OBJ_TYPE_STREAMWEAPON) &&
			   (tar_type != SW_OBJ_TYPE_SPHEREWEAPON) &&
			   (tar_type != SW_OBJ_TYPE_CONTROLLED) &&
			   (tar_type != SW_OBJ_TYPE_PLAYER)
			)
			    continue;
		    }
		    else
		    {
			if((tar_type != SW_OBJ_TYPE_CONTROLLED) &&
			   (tar_type != SW_OBJ_TYPE_PLAYER)
			)
			    continue;
		    }

		    /* PErmissions make it immune from being hit? */
		    if(tar_obj->permission.uid <= SWS_ACCESS_UID_DESTROY_IMMUNE)
			continue;

		    /* Not in the same sector? */
		    if(!SWMath3DInSameSectorPtr(obj, tar_obj))
			continue;

		    /* Calculate distance apart in real units */
		    distance = SWMath3DDistance(
			tar_obj->x - obj->x,
			tar_obj->y - obj->y,
			tar_obj->z - obj->z
		    );
		    if(distance > dmg_radius)
			continue;

		    /* Calculate power coefficient to determine how
		     * powerful is the self-destruct explosion based
		     * on the distance apart
		     */
		    power_coeff = (dmg_radius > 0.0f) ?
			sqrt(MIN(
			    (1.0f - (distance / dmg_radius)),
			    1.0f
		        )) : 0.0f;

		    /* Perform hit on this object */
		    objects_recycled += REngDoHit(
			legend,
			obj_num, i,
			power * power_coeff,
			power * power_coeff,
			cur_ms, cur_sec, tc
		    );
		    if(DBIsObjectGarbage(obj_num))
			return(objects_recycled);
		}
	    }

	    /* Get the object's name */
	    obj_name = STRDUP(DBGetFormalNameStr(obj_num));

	    /* Create debris */
	    EffectsCreateDebrisObject(obj_num, obj, cur_ms, cur_sec, tc);

	    /* Create large explosion */
	    EffectsCreateExplosionLarge(
		obj_num, obj,
		obj->sect_x, obj->sect_y, obj->sect_z,
		obj->x, obj->y, obj->z,
		obj->heading, obj->pitch, 0.0f * PI
	    );

	    /* Create shockwave */
	    EffectsCreateShockwave(
		obj_num, obj,
		obj->size,
		MAX((long)(dmg_radius * 1000.0f), obj->size),
		obj->power, obj->power
	    );

	    /* Is this object is not immune from being destroyed? */
	    if(obj->permission.uid > SWS_ACCESS_UID_DESTROY_IMMUNE)
	    {
		/* Object can be destroyed */

		if(1)
		{
		    int i;
		    char *s, loc_str[256];
		    sws_connection_struct *con;

#ifdef PLUGIN_SUPPORT
		    /* Notify all plugins */
		    for(i = 0; i < total_plugins; i++)
			PluginDestroyNotify(
			    i,
			    SW_DESTROY_REASON_SELF_DESTRUCT,
			    obj_num,	/* Destroyed object */
			    obj_num,	/* Object that caused destruction */
			    obj_num	/* Owner of weapon object */
			);
#endif

		    /* Format location string */
		    UNVLocationFormatString(
			loc_str,
			&obj->sect_x, &obj->sect_y, &obj->sect_z,
			&obj->x, &obj->y, &obj->z,
			sizeof(loc_str)
		    );

		    switch(obj->type)
		    {
		      case SW_OBJ_TYPE_CONTROLLED:
		      case SW_OBJ_TYPE_PLAYER:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(obj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed due to power-core overload (auto-destruct) at %s.",
				obj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(obj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed due to power-core overload (auto-destruct) at %s.",
			    obj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDPLAYER,
			    s
			);
			free(s);
		        break;

		      default:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(obj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed at %s.",
				obj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(obj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed at %s.",
			    obj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDOBJECT,
			    s
			);
			free(s);
			break;
		    }

		    /* Send destroy notify to all connections in
		     * the same sector
		     */
		    for(i = 0; i < total_connections; i++)
		    {
			con = connection[i];
			if(con == NULL)
			    continue;

			if((con->socket < 0) || (con->object_num < 0))
			    continue;

			/* Not in the same sector? */
			if(!SWMath3DInSameSector(
			    xsw_object, total_objects,
			    con->object_num, obj_num
			))
			    continue;

			/* Send destroy notify to this connection */
/*			NetSendObjectName(i, obj_num); */
		        NetSendNotifyDestroy(
			    i,
			    SW_DESTROY_REASON_SELF_DESTRUCT,	/* Reason is self destruct */
			    obj_num,	/* Destroyed object */
			    obj_num,	/* Object that caused destruction */
			    obj_num	/* Owner of weapon object */
			);
		    }
		}

		/* Check if the destroyed object is a player and if
		 * players should be respawned, otherwise recycle
		 * the destroyed object
		 *
		 * Note that after this point, obj should be considered
		 * invalid. It can still be used as cross reference but
		 * it must never be referenced implicitily.
		 */
		if((obj->type == SW_OBJ_TYPE_PLAYER) &&
		   sysparm.respawn_players
		)
		{
		    REngDoRespawn(legend, obj_num, cur_ms, cur_sec, tc);
		}
		else
		{
		    DBRecycleObject(obj_num, cur_ms, cur_sec, tc);
		    NetSendRecycleObject(-1, obj_num);
		    objects_recycled++;
		}
	    }
	    else
	    {
		/* Object cannot be destroyed, so sanitize values */
		if(obj->hp < 0.0f)
		    obj->hp = 0.0f;
	    }

	    free(obj_name);
	}

	return(objects_recycled);
}

/*
 *	Updates weapons lock on an object.
 */
static void REngWeaponsLock(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int tar_obj_num;
	float scanner_range;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;

	/* Get the locked on object and check if it is valid */
	tar_obj_num = obj->locked_on;
	if(DBIsObjectGarbage(tar_obj_num))
	{
	    /* Locked on object is now garbage, if it was, in fact,
	     * locked, then unlock it and schedual report to
	     * connections
	     */
	    if(tar_obj_num > -1)
	    {
		obj->locked_on = -1;
		next.object_values = cur_ms;
	    }
	    return;
	}

	tar_obj = xsw_object[tar_obj_num];

	/* Is the object marked hidden? */
	if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
	{
	    /* Locked on object is now hidden, unlock it and schedual
	     * report to connections
	     */
	    obj->locked_on = -1;
	    next.object_values = cur_ms;
	    return;
	}

	/* Weapons objects maintain their lock on the target regardless
	 * of the target's visibility and distance
	 */
	if((obj->type == SW_OBJ_TYPE_WEAPON) ||
	   (obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	    return;

	/* Get the current effective scanner range in real units */
	scanner_range = DBGetObjectEffectiveScannerRangePtr(obj);
	if(scanner_range <= 0.0f)
	{
	    /* Locked on object is no longer in scanner range, unlock
	     * it and schedual report to connections
	     */
	    obj->locked_on = -1;
	    next.object_values = cur_ms;
	    return;
	}

	/* Object not within the current effective scanner range? */
	if(!DBIsObjectInEffectiveScannerRangePtr(
	    obj, tar_obj, scanner_range
	))
	{
	    /* Locked on object is no longer in scanner range, unlock
	     * it and schedual report to connections
	     */
	    obj->locked_on = -1;
	    next.object_values = cur_ms;
	    return;
	}

	/* Object still in range */
}


/*
 *	Updates an object's tractor beam lock.
 *
 *	Checks if the object is tractoring any objects, consumes
 *	antimatter, and moves each tractored object.
 */
static void REngTractorBeam(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i;
	int trac_obj_num;
	float size_ru, tractor_beam_range = legend->tractor_beam_range;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;

	/* Skip if not tractoring anything */
	if(obj->total_tractored_objects <= 0)
	    return;

	/* Unlock tractor beam if object is out of antimatter */
	if(obj->antimatter <= 0.0f)
	{
	    const int con_num = ConGetByObject(obj_num);
	    int tcon_num;

	    if(con_num > -1)
		NetSendPlaySound(
		    con_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
		    1.0f, 1.0f
		);

	    for(i = 0; i < obj->total_tractored_objects; i++)
	    {
		trac_obj_num = obj->tractored_object[i];
		if(trac_obj_num < 0)
		    continue;

		DBObjectUntractor(obj_num, trac_obj_num);
		NetSendTractorBeamLock(-1, obj_num, -1);
		tcon_num = ConGetByObject(trac_obj_num);
		if(tcon_num > -1)
		    NetSendPlaySound(
			tcon_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
			1.0f, 1.0f
		    );
	    }
	    return;
	}

	/* Iterate through each tractored object */
	for(i = 0; i < obj->total_tractored_objects; i++)
	{
	    trac_obj_num = obj->tractored_object[i];
	    if(trac_obj_num < 0)
		continue;

	    /* Check if trac_obj_num is not -1 but garbage */
	    if(DBIsObjectGarbage(trac_obj_num))
	    {
		const int con_num = ConGetByObject(obj_num);

		DBObjectUntractor(obj_num, trac_obj_num);
		NetSendTractorBeamLock(-1, obj_num, -1);
		if(con_num > -1)
		    NetSendPlaySound(
			con_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
			1.0f, 1.0f
		    );

	        continue;
	    }

	    tar_obj = xsw_object[trac_obj_num];

	    /* Check if the target object is tractorable */
	    if(!DBIsObjectTractorablePtr(obj, tar_obj))
	    {
		const int	con_num = ConGetByObject(obj_num),
				tcon_num = ConGetByObject(trac_obj_num);

		DBObjectUntractor(obj_num, trac_obj_num);
	        NetSendTractorBeamLock(-1, obj_num, -1);
		if(con_num > -1)
		    NetSendPlaySound(
			con_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
			1.0f, 1.0f
		    );
		if(tcon_num > -1)
		    NetSendPlaySound(
			tcon_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
			1.0f, 1.0f
		    );

		continue;
	    }

	    /* Consume antimatter */
	    size_ru = (float)tar_obj->size / 1000.0f;
	    obj->antimatter = MAX(
		(obj->antimatter - (size_ru *
		    legend->tractor_beam_antimatter_rate * tc)
		),
		0.0f
	    );

	    /* Move the tractored object? */
	    if(1)
	    {
		/* Calculate the delta position behind the tractoring
		 * object as dx, dy, and dz
		 *
		 * Record the tractored object's original maximum
		 * velocity since it will have to be modified before
		 * it is moved and then restored afterwards
		 */
		float	theta1, theta2, theta3,
			mag1, mag2, mag3,
			old_velocity_max = tar_obj->velocity_max,
			d = (float)obj->size * 2.0 / 1000.0,
			tx = -SWMathRectangularToPolarX(obj->heading, d),
			ty = -SWMathRectangularToPolarY(obj->heading, d),
			dx = obj->x - tar_obj->x + tx,
			dy = obj->y - tar_obj->y + ty,
			dz = obj->z - tar_obj->z;

	        d = SWMath3DDistance(dx, dy, dz);

		/* theta1 and mag1 is the tractored object's heading
		 * and velocity
		 */
	        theta1 = tar_obj->velocity_heading;

		/* theta2 and mag2 is the heading and velocity from
		 * the tractored object to the tractoring object
		 * (or the direction of the tractor beam's pull)
		 */
	        theta2 = SWMathCoordinateDeltaVector(dx, dy);

		/* Calculate tractor beam strength mag2 by taking
		 * the coefficient of the distance away the objects
		 * are from each other
		 *
		 * The farther away they are, the greater the tractor
		 * beam strength
		 */
	        if(d < ((float)tar_obj->size * 2.0 / 1000.0))
		{
		    /* Slow down the tractored object's velocity */
		    if(obj->thrust_power > 0.0)
			mag1 = MAX(
			    tar_obj->velocity -
			    (obj->thrust_power * tc),
			    0.0
			);
		    else
			mag1 = 0.0;
		    mag2 = 0.0;
		}
	        else
		{
		    mag1 = tar_obj->velocity;
	            mag2 = obj->velocity_max *
			MIN((d * 4.0 / tractor_beam_range), 1.0);
 		}

/* Note that objects that do not move will be more difficult to
 * tractor at higher velocities since their velocity will get set to
 * 0 after this call
 */

		/* Add target object's current velocity vector with
		 * the new velocity vector caused by the tractor beam
		 * pull
		 */
	        SWMathVectorAdd(
	            theta1, mag1,
	            theta2, mag2,
	            &theta3, &mag3
	        );

	        tar_obj->velocity_heading = theta3;
		SWMathSetUnitVector2D(
		    &tar_obj->momentum_vector_compoent,
		    tar_obj->velocity_heading
		);

	        tar_obj->velocity = mag3;

		/* Turn the tractored object to face the direction
		 * it is being pulled
		 */
		tar_obj->heading = REngTurnAngle(
		    tar_obj->heading, theta3,
		    tar_obj->turnrate * 100.0f,
		    cur_ms, cur_sec, tc
		);

		/* Adjust the tractored object's maximum velocity
		 * as needed to accomidate the current velocity
		 */
	        if(tar_obj->velocity > tar_obj->velocity_max)
		    tar_obj->velocity_max = tar_obj->velocity;

	        /* Move the object */
	        REngMove(legend, trac_obj_num, 0, 1, cur_ms, cur_sec, tc);

	        /* Restore the tractored object's original maximum
		 * velocity
		 */
	        tar_obj->velocity_max = old_velocity_max;
	    }
	}
}


/*
 *	Only checks if object is in intercept range, does NOT
 *	adjust heading.
 */
static void REngIntercept(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int tar_obj_num;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip worm hole objects */
	if(obj->type == SW_OBJ_TYPE_WORMHOLE)
	    return;

	/* Get the target object */
	tar_obj_num = obj->intercepting_object;
	if(DBIsObjectGarbage(tar_obj_num))
	    return;

	/* Weapons objects maintain their lock on the target regardless
	 * of the target's visibility and distance
	 */
	if((obj->type == SW_OBJ_TYPE_WEAPON) ||
	   (obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	    return;

	/* Use check if the target object is in range */
	if(UNVMatchInterceptByNumber(
	    xsw_object, total_objects,
	    obj_num, tar_obj_num
	) < 0)
	    obj->intercepting_object = -1;
}


/*
 *	Regernerates an object's power core.
 */
static void REngPowerCore(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Does the object's power core need increasing and can be
	 * increased?
	 */
	if((obj->power < obj->power_max) &&
	   (obj->power_purity > 0.0f)
	)
	{
	    const float	prev_power = MAX(obj->power, 0.0f),
			power_purity = MIN(obj->power_purity, 1.0f);

	    /* Increase the power */
	    if(obj->antimatter > 0.0f)
		obj->power = CLIP(
		    (prev_power + (obj->core_efficency * power_purity * tc)),
		    0.0f, obj->power_max
		);
	    else
		obj->power = CLIP(
		    (prev_power +
			(obj->core_efficency * power_purity *
			    legend->no_antimatter_thrust_coeff * tc)),
		    0.0f, obj->power_max
		);

	    /* Consume antimatter as if the engine was at full
	     * throttle
	     */
	    obj->antimatter = MAX(
		(obj->antimatter - (legend->engine_antimatter_rate * tc)),
 		0.0f
	    );
	}
}


/*
 *	Repairs the object if the object's damage control is on.
 */
static void REngDamageControl(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	float power, power_purity, hp_inc;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip if damage control is not on */
	if(obj->damage_control != SW_DMGCTL_STATE_ON)
	    return;

	/* Skip if hit points are already at max */
	if(obj->hp >= obj->hp_max)
	    return;

	/* Get the power and power purity */
	power = MAX(obj->power, 0.0f);
	power_purity = CLIP(obj->power_purity, 0.0f, 1.0f);

	/* Calculate hit points increment */
	hp_inc = MAX(
	    legend->damage_control_rate * power * power_purity * tc,
	    0.0f
	);

	/* "Repair" hit points */
	obj->hp += hp_inc;
	if(obj->hp > obj->hp_max)
	{
	    hp_inc = MAX(hp_inc - obj->hp - obj->hp_max, 0.0f);
	    obj->hp = obj->hp_max;
	}

	/* Consume antimatter based on how much repairs were made and
	 * the size of the object
	 */
	obj->antimatter = MAX(
	    obj->antimatter -
		(hp_inc *
		    (MAX(((float)obj->size / 100.0f), 0.0f) + 1.0f)),
	    0.0f
	);

	/* Update the object's on fire state */
	if((obj->hp_max > 0.0f) && (obj->ext_flags & SWS_OBJ_FLAG_ON_FIRE))
	{
	    if((obj->hp / obj->hp_max) > legend->on_fire_hp_coeff)
		EffectsSetObjectOnFire(obj_num, obj, 0);
	}
}


/*
 *	Respawns the player.
 *
 *	The object will be moved back to the player starting position
 *	and its values will be restored (ie hit points and weapons).
 *
 *	Network updates, including a force pose and object values will
 *	be sent.
 */
static void REngDoRespawn(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i;
	xsw_weapon_struct *wep;
	xsw_object_struct *obj, *obj2;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	obj->ext_flags &= ~SWS_OBJ_FLAG_DESTROYED;

	/* Move to starting position */
	obj->heading = legend->player_start_heading;
	obj->pitch = legend->player_start_pitch;
	obj->bank = legend->player_start_bank;

	obj->sect_x = legend->player_start_sect_x;
	obj->sect_y = legend->player_start_sect_y;
	obj->sect_z = legend->player_start_sect_z;

	obj->x = legend->player_start_x;
	obj->y = legend->player_start_y;
	obj->z = legend->player_start_z;

	/* Restore hit points, restore the power, and clear the on
	 * fire state
	 */
	obj->hp = obj->hp_max;
	EffectsSetObjectOnFire(obj_num, obj, 0);
	obj->power = obj->power_max;
	if(obj->shield_state != SW_SHIELD_STATE_NONE)
	    obj->shield_state = SW_SHIELD_STATE_DOWN;

	obj->throttle = 0.0f;
	obj->damage_control = SW_DMGCTL_STATE_OFF;

	obj->locked_on = -1;
	obj->intercepting_object = -1;

	free(obj->tractored_object);
	obj->tractored_object = NULL;
	obj->total_tractored_objects = 0;

	/* Replenish weapons */
	obj->selected_weapon = 0;
	for(i = 0; i < obj->total_weapons; i++)
	{
	    wep = obj->weapons[i];
	    if(wep == NULL)
		continue;

/* Should wep->amount be reset to 0? */
	    wep->amount = wep->max;
	    wep->last_used = 0l;
	}

	/* Remove weapons lock of all other objects on this object */
	for(i = 0; i < total_objects; i++)
	{
	    if(DBIsObjectGarbage(i))
		continue;

	    obj2 = xsw_object[i];

	    if(obj2->locked_on == obj_num)
		obj2->locked_on = -1;
	    if(obj2->intercepting_object == obj_num)
		obj2->intercepting_object = -1;
	}

	/* Notify all connections about respawn */
	NetSendObjForceSect(-1, obj_num, cur_ms, cur_sec, tc);
	NetSendObjectForcePose(-1, obj_num, cur_ms, cur_sec, tc);
	NetSendObjectValues(-1, obj_num);
	for(i = 0; i < obj->total_weapons; i++)
	    NetSendWeaponValues(-1, obj_num, i);
	NetSendScore(-1, obj_num);
}

/*
 *	Hits the target object with the specified weapon object,
 *	updating the scores on the target object and the weapon
 *	object's owner, reducing the power and/or hit points on the
 *	target object, recycling the target object if it has been
 *	destroyed, create explosions and debris.
 *
 *	This function does not check if the weapon object actually
 *	hit the target object or not.
 *
 *	If the weapon object's type is SW_OBJ_TYPE_WEAPON
 *	(projectile or pulse weapon) then it will be recycled by this
 *	function.
 *
 *	The target object will be either respawned or recycled if its
 *	hit points goes to 0.
 *
 *	Returns the number of objects recycled.
 */
static int REngDoHit(
	sw_legend_struct *legend,
	const int wobj_num, const int tobj_num,
	const float damage_hull,
	const float damage_shields,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char	report_this = 1,	/* Log and export event */
		can_count_kill = 1;	/* Can count as a kill */
	int objects_recycled = 0, wobj_owner_num;
	char	*wobj_name,
		*wobj_owner_name,
		*tobj_name;
	float	damage_to_structure = 0.0f,
		damage_to_shield = 0.0f;
	xsw_object_struct *wobj, *tobj;


	/* Both the weapon and target objects must be valid */
	if(DBIsObjectGarbage(wobj_num))
	    return(objects_recycled);
	if(DBIsObjectGarbage(tobj_num))
	    return(objects_recycled);

	wobj = xsw_object[wobj_num];
	tobj = xsw_object[tobj_num];

	/* If the weapon has no damage ability then do not hit
	 * anything
	 */
	if((damage_hull <= 0.0f) && (damage_shields <= 0.0f))
	{
	    /* Recycle the weapon object if it is a projectile or
	     * pulse weapon
	     */
	    if(wobj->type == SW_OBJ_TYPE_WEAPON)
	    {
		DBRecycleObject(wobj_num, cur_ms, cur_sec, tc);
		NetSendRecycleObject(-1, wobj_num);
		objects_recycled++;
	    }
	    return(objects_recycled);
	}

	/* If report_destroyed_weapons is set to false, then
	 * check if target object is a weapon object.
	 * If it is, then set report_this to false so that
	 * no reporting of this hit will be made.
	 */
	if(!sysparm.report_destroyed_weapons)
	{
	    if((tobj->type == SW_OBJ_TYPE_WEAPON) ||
	       (tobj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	       (tobj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	    )
	    {
		report_this = 0;
		can_count_kill = 0;
	    }
	}


	/* Get owner of the weapon object (or the weapon object
	 * itself if it is a player or controlled object)
	 */
	if((wobj->type == SW_OBJ_TYPE_CONTROLLED) ||
	   (wobj->type == SW_OBJ_TYPE_PLAYER)
	)
	    wobj_owner_num = wobj_num;
	else
	    wobj_owner_num = wobj->owner;


	/* Get the formal names of the weapon, weapon's owner, and
	 * target objects
	 */
	wobj_name = STRDUP(DBGetFormalNameStr(wobj_num));
	wobj_owner_name = STRDUP(DBGetFormalNameStr(wobj_owner_num));
	tobj_name = STRDUP(DBGetFormalNameStr(tobj_num));


	/* Log the weapon object hitting target object? */
	if(report_this)
	{
	    char *s, loc_str[256];

	    /* Format location string */
	    UNVLocationFormatString(
		loc_str,
		&tobj->sect_x,
		&tobj->sect_y,
		&tobj->sect_z,
		&tobj->x,
		&tobj->y,
		&tobj->z,
		sizeof(loc_str)
	    );

	    /* Log hit? */
	    if(sysparm.log_events)
	    {
		s = (char *)malloc(
		    STRLEN(wobj_name) +
		    STRLEN(tobj_name) +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
		    "%s hit %s at %s.",
		    wobj_name, tobj_name, loc_str
		);
		LogAppend(fname.log, s);
		free(s);
	    }

	    /* Export hit event */
	    switch(wobj->type)
	    {
	      case SW_OBJ_TYPE_WEAPON:
		s = (char *)malloc(
		    STRLEN(tobj_name) +
		    STRLEN(wobj_name) +
		    STRLEN(wobj_owner_name) +
		    STRLEN(loc_str) +
		    80
		);
	        sprintf(
		    s,
"%s hit by %s fired by %s at %s.",
		    tobj_name, wobj_name, wobj_owner_name, loc_str
	        );
	        ExportEvents(
		    fname.events_export,
		    EVST_TYPE_PROJECTILEWEAPON_HIT,
		    s
	        );
		free(s);
	        break;

	      case SW_OBJ_TYPE_STREAMWEAPON:
	      case SW_OBJ_TYPE_SPHEREWEAPON:
		s = (char *)malloc(
		    STRLEN(tobj_name) +
		    STRLEN(wobj_name) +
		    STRLEN(wobj_owner_name) +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
"%s hit by %s fired by %s at %s.",
		    tobj_name, wobj_name, wobj_owner_name, loc_str
		);
		ExportEvents(
		    fname.events_export,
		    EVST_TYPE_STREAMWEAPON_HIT,
		    s
		);
		free(s);
		break;

	      default:
		s = (char *)malloc(
		    STRLEN(tobj_name) +
		    STRLEN(wobj_owner_name) +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
"%s hit by %s at %s.",
		    tobj_name, wobj_owner_name, loc_str
		);
		ExportEvents(
		    fname.events_export,
		    EVST_TYPE_STREAMWEAPON_HIT,
		    s
		);
		free(s);
		break;
	    }
	}


	/* Make shields visiable on target object if the target
	 * object's shields are up and the shield frequencies on the
	 * the target object and weapon object do not match
	 *
	 * Also check if the weapon object can penetrate shields
	 */
	if((tobj->shield_state == SW_SHIELD_STATE_UP) &&
	   (tobj->power_max > 0.0f) &&
	   !REngFreqCmp(tobj->shield_frequency, wobj->shield_frequency) &&
	   !(wobj->ext_flags & SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD)
	)
	{
	    tobj->shield_visibility = tobj->power / tobj->power_max;

	    /* Send shield visibility to all connections */
	    NetSendShieldVis(-1, tobj_num);
	}

	/* Should the target object receive damage? */
	if((sysparm.homes_destroyable) ? 1 :
	   ((tobj->type != SW_OBJ_TYPE_HOME) &&
	    (tobj->type != SW_OBJ_TYPE_WORMHOLE) &&
	    (tobj->type != SW_OBJ_TYPE_ELINK)
	   )
	)
	{
	    /* Does the target object have its shields up, do the
	     * shield frequencies not match, and is the weapon object
	     * not able to penetrate shields?
	     */
	    if((tobj->shield_state == SW_SHIELD_STATE_UP) &&
	       !REngFreqCmp(tobj->shield_frequency, wobj->shield_frequency) &&
	       !(wobj->ext_flags & SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD)
	    )
	    {
		float remainder;

		/* Calculate extra damage to shields due to the target
		 * object's lack of power purity
		 */
		const float extra_damage_to_shields = MAX(
		    ((damage_shields * (2.0f - tobj->power_purity)) -
			damage_shields),
		    0.0f
		);

		/* Add damage to shields from the weapon's power and the
		 * target object's lack of power purity
		 */
		damage_to_shield += damage_shields + extra_damage_to_shields;

		/* Reduce power due to damage to shields */
	        tobj->power -= damage_to_shield;

		/* Calculate remainder of damage (if any) that is to be
		 * applied to hull damage
		 */
		remainder = (tobj->power < 0.0f) ? -tobj->power : 0.0f;

		/* Shields went down? */
		if(remainder > 0.0f)
		{
		    /* Can there be damage applied to the hull? */
		    if(damage_hull > 0.0f)
		    {
			/* Calculate damage to structure using the
			 * shield damage remainder coeff multiply the
			 * hull damage
			 */
			damage_to_structure += damage_hull * remainder /
			    damage_shields;

			/* Reduce hp due to damage to hull */
			tobj->hp -= damage_to_structure;

			/* Make sure damage to structure does not
			 * exceed actual damage
			 */
			if(tobj->hp < 0.0f)
			    damage_to_structure += tobj->hp;
		    }

		    /* Undo too much damage to shields */
		    damage_to_shield += tobj->power;
	        }
	    }
	    else
	    {
		/* Target object's shields were down, no shields at
		 * all, or frequency of weapon matched frequency of
		 * target object's shields
		 */

		damage_to_structure += damage_hull;

		/* Reduce hp due to damage to hull */
	        tobj->hp -= damage_to_structure;

		/* Make sure damage to structure does not exceed actual
		 * damage
		 */
		if(tobj->hp < 0.0f)
		    damage_to_structure += tobj->hp;
	    }

	    /* If the object is tractoring other objects and its
	     * power has dropped below 25% (regardless of the
	     * tractored object's shield power) the tractor beam should
	     * be disengaged
	     */
	    if(((tobj->total_tractored_objects > 0) &&
	        ((tobj->power_max > 0.0f) ?
		   ((tobj->power / tobj->power_max) <= 0.25f) : 1)) ||
		(tobj->shield_state != SW_SHIELD_STATE_UP)
	    )
	    {
		const int con_num = ConGetByObject(tobj_num);
		int i, tcon_num, tractored_obj_num;

		if(con_num > -1)
		    NetSendPlaySound(
			con_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
			1.0f, 1.0f
		    );

		for(i = 0; i < tobj->total_tractored_objects; i++)
		{
		    tractored_obj_num = tobj->tractored_object[i];
		    if(tractored_obj_num < 0)
			continue;

		    DBObjectUntractor(tobj_num, tractored_obj_num);
		    NetSendTractorBeamLock(-1, tobj_num, -1);
		    tcon_num = ConGetByObject(tractored_obj_num);
		    if(tcon_num > -1)
			NetSendPlaySound(
			    tcon_num, SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
			    1.0f, 1.0f
			);
		}
	    }

	    /* Set the target object on fire? */
	    if((tobj->hp_max > 0.0f) &&
	       ((tobj->type == SW_OBJ_TYPE_CONTROLLED) ||
	        (tobj->type == SW_OBJ_TYPE_PLAYER)
	       )
	    )
	    {
		if((tobj->hp / tobj->hp_max) <= legend->on_fire_hp_coeff)
		    EffectsSetObjectOnFire(tobj_num, tobj, 1);
	    }


	}
	/* Sanitize the target object's power */
	if(tobj->power < 0.0f)
	    tobj->power = 0.0f;


	/* Damage to structure and shields have been calculated
	 * and processed, now add/subtract scores on each object
	 */

	/* Damage given by weapon object's owner */
	if(!DBIsObjectGarbage(wobj_owner_num))
	{
	    xsw_object_struct *wobj_owner = xsw_object[wobj_owner_num];

	    if((sysparm.homes_destroyable) ? 1 :
	       ((tobj->type != SW_OBJ_TYPE_HOME) &&
		(tobj->type != SW_OBJ_TYPE_WORMHOLE) &&
		(tobj->type != SW_OBJ_TYPE_ELINK)
	       )
	    )
	    {
		if(!UNVAllocScores(wobj_owner))
	        {
		    xsw_score_struct *score = wobj_owner->score;

		    /* Add damage_given to weapon object owner */
		    score->damage_given += damage_to_shield +
			damage_to_structure;

		    /* Add bonus credits to owner */
		    if(tobj->type == SW_OBJ_TYPE_PLAYER)
		    {
		        score->credits += (
			    (damage_to_shield + damage_to_structure) *
			    sysparm.hit_player_bonus
		        );
		    }

		    /* Add kills if target object hp's are less than 0 */
		    if((tobj->hp <= 0.0) && can_count_kill)
		    {
			score->kills++;

			/* Transfer credits on tobj to wobj_owner? */
			if(sysparm.killer_gets_credits &&
			   (tobj->score != NULL)
			)
			{
			    score->credits += tobj->score->credits;
			    tobj->score->credits = 0.0;
			}
		    }

		    /* Update scores to all connections */
		    NetSendScore(-1, wobj_owner_num);
	        }
	    }
	}

/* Checks if homes are set destroyable (hittable) and if the specified
 * object can be hit
 */
#define IS_OBJ_DESTROYABLE(_obj_)	(		\
 (sysparm.homes_destroyable) ?			\
  1 :						\
  (((_obj_)->type != SW_OBJ_TYPE_HOME) &&	\
   ((_obj_)->type != SW_OBJ_TYPE_WORMHOLE) &&	\
   ((_obj_)->type != SW_OBJ_TYPE_ELINK)		\
  )						\
)

	/* Apply damage to the target object and log the amount of
	 * damage to the target object
	 */
	if(IS_OBJ_DESTROYABLE(tobj))
	{
	    const float total_damage = damage_to_shield +
		damage_to_structure;

	    /* Add damage received to the target object's scores */
	    if(!UNVAllocScores(tobj))
	    {
		xsw_score_struct *score = tobj->score;
		score->damage_received += total_damage;
		NetSendScore(-1, tobj_num);
	    }

	    /* Log damage received? */
	    if(report_this && sysparm.log_events)
	    {
		char *s = (char *)malloc(
		    STRLEN(tobj_name) +
		    40 +
		    STRLEN(wobj_owner_name) +
		    80
		);
		if(wobj_owner_num != wobj_num)
		    sprintf(
			s,
"%s: Received %.2f damage by weapon fired by %s",
			tobj_name,
			total_damage,
			wobj_owner_name
		    );
		else
		    sprintf(
			s,
"%s: Received %.2f damage by %s",
			tobj_name,
			total_damage,
			wobj_owner_name
		    );
		LogAppend(fname.log, s);
		free(s);
	    }
	}

	/* Send hit notify to plugins and connections */
	if(1)
	{
	    int i;
	    float	reported_damage_to_structure,
			reported_damage_to_shield;
	    sws_connection_struct *con;

	    /* Get the bearing from the target object to the weapon
	     * object
	     */
	    float bearing = SWMathCoordinateDeltaVector(
		wobj->x - tobj->x,
		wobj->y - tobj->y
	    );
	    /* Adjust the bearing to be relative to the target
	     * object's heading
	     */
	    bearing = SANITIZERADIANS(bearing - tobj->heading);

	    /* Get the amount of damage to report
	     *
	     * For objects that are destroyable it is the actual
	     * value, but for objects that are not destroyable then
	     * it is 0
	     */
	    if(IS_OBJ_DESTROYABLE(tobj))
	    {
		reported_damage_to_structure = damage_to_structure;
		reported_damage_to_shield = damage_to_shield;
	    }
	    else
	    {
		reported_damage_to_structure = 0.0f;
		reported_damage_to_shield = 0.0f;
	    }


#ifdef PLUGIN_SUPPORT
	    /* Send hit notify to all plugins */
	    for(i = 0; i < total_plugins; i++)
		PluginHitNotify(
		    i,
		    wobj_num, tobj_num,
		    bearing,
		    reported_damage_to_structure,
		    reported_damage_to_shield
		);
#endif

	    /* Send hit notify to all connections that are in the
	     * same sector
	     */
	    for(i = 0; i < total_connections; i++)
	    {
		con = connection[i];
		if(con == NULL)
		    continue;

		if((con->socket < 0) || (con->object_num < 0))
		    continue;

		/* In same sector? */
		if(!SWMath3DInSameSector(
		    xsw_object, total_objects,
		    con->object_num, tobj_num
		))
		    continue;

		/* Send hit notify to this connection */
		NetSendNotifyHit(
		    i,
		    wobj_num, tobj_num,
		    bearing,
		    reported_damage_to_structure,
		    reported_damage_to_shield
		);
	    }
	}

	/* Does the weapon object produce a shockwave when it hits
	 * something?
	 */
	if(wobj->ext_flags & SWS_OBJ_FLAG_WEP_SHOCKWAVE_HIT)
	{
	    /* Create shockwave */
	    EffectsCreateShockwave(
		wobj_num, wobj,
		wobj->size, wobj->size * 3l,
		damage_hull, damage_shields
	    );
	}
#undef IS_OBJ_DESTROYABLE

	/* Target object needs to be destroyed? */
	if(((tobj->hp <= 0.0f) ||
	    (tobj->type == SW_OBJ_TYPE_WEAPON) ||
	    (tobj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	    (tobj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	   ) && !(tobj->ext_flags & SWS_OBJ_FLAG_DESTROYED)
	)
	{
	    /* If the target object can be destroyed then mark it as
	     * destroyed so that subsequent destroy functions do not
	     * destroy it again
	     */
	    if(tobj->permission.uid > SWS_ACCESS_UID_DESTROY_IMMUNE)
		tobj->ext_flags |= SWS_OBJ_FLAG_DESTROYED;

	    /* Create debris
	     *
	     * Except for weapon objects
	     */
	    if((tobj->type != SW_OBJ_TYPE_WEAPON) &&
	       (tobj->type != SW_OBJ_TYPE_STREAMWEAPON) &&
	       (tobj->type != SW_OBJ_TYPE_SPHEREWEAPON)
	    )
		EffectsCreateDebrisObject(tobj_num, tobj, cur_ms, cur_sec, tc);

	    /* Create large explosion */
	    EffectsCreateExplosionLarge(
		tobj_num, tobj,
		tobj->sect_x, tobj->sect_y, tobj->sect_z,
		tobj->x, tobj->y, tobj->z,
		tobj->heading, tobj->pitch, 0.0f * PI
	    );

	    /* Create shockwave */
	    EffectsCreateShockwave(
		tobj_num, tobj,
		tobj->size, tobj->size * 4l,
		tobj->power + damage_hull, tobj->power + damage_shields
	    );

	    /* Is the target object is not immune to be destroyed? */
	    if(tobj->permission.uid > SWS_ACCESS_UID_DESTROY_IMMUNE)
	    {
		/* Target object can be destroyed */

		if(report_this)
		{
		    int i;
		    char *s, loc_str[256];
		    sws_connection_struct *con;

#ifdef PLUGIN_SUPPORT
		    /* Notify all plugins */
		    for(i = 0; i < total_plugins; i++)
			PluginDestroyNotify(
			    i,
			    SW_DESTROY_REASON_UNSPECIFIC,
			    tobj_num,		/* Destroyed object */
			    wobj_num,		/* Object that caused destruction */
			    wobj_owner_num	/* Owner of wobj */
			);
#endif

		    /* Format location string */
		    UNVLocationFormatString(
			loc_str,
			&tobj->sect_x,
			&tobj->sect_y,
			&tobj->sect_z,
			&tobj->x,
			&tobj->y,
			&tobj->z,
			sizeof(loc_str)
		    );

		    switch(tobj->type)
		    {
		      case SW_OBJ_TYPE_PLAYER:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(tobj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed due to hull structure colapse at %s.",
				tobj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(tobj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed due to hull structure colapse at %s.",
			    tobj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDPLAYER,
			    s
			);
			free(s);
		        break;

		      case SW_OBJ_TYPE_WEAPON:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(tobj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed at %s.",
				tobj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(tobj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed at %s.",
			    tobj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDWEAPON,
			    s
			);
			free(s);
			break;

		      default:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(tobj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed at %s.",
				tobj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(tobj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed at %s.",
			    tobj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDOBJECT,
			    s
			);
			free(s);
			break;
		    }

		    /* Send destroy notify to all connections */
		    for(i = 0; i < total_connections; i++)
		    {
			con = connection[i];
			if(con == NULL)
			    continue;

			if((con->socket < 0) || (con->object_num < 0))
			    continue;

			/* Not in the same sector? */
			if(!SWMath3DInSameSector(
			    xsw_object, total_objects,
			    con->object_num, tobj_num
			))
			    continue;

			/* Send destroy notify to this connection */
			NetSendObjectName(i, tobj_num);
		        NetSendNotifyDestroy(
			    i,
			    SW_DESTROY_REASON_UNSPECIFIC,	/* Reason is hit by weapon */
			    tobj_num,		/* Destroyed object */
			    wobj_num,		/* Object that caused destruction */
			    wobj_owner_num	/* Owner of wobj */
			);
		    }
		}	/* if(report_this) */

		/* Check if the destroyed object is a player and if
		 * players should be respawned, otherwise recycle
		 * the destroyed object
		 *
		 * Note that after this point, tobj should be considered
		 * invalid. It can still be used as cross reference but
		 * it must never be referenced implicitily.
		 */
		if((tobj->type == SW_OBJ_TYPE_PLAYER) &&
		    sysparm.respawn_players
		)
		{
		    REngDoRespawn(legend, tobj_num, cur_ms, cur_sec, tc);
		}
		else
		{
		    DBRecycleObject(tobj_num, cur_ms, cur_sec, tc);
		    NetSendRecycleObject(-1, tobj_num);
		    objects_recycled++;
		}
	    }
	    else
	    {
		/* Sanitize hp on tobj (since we didn't recycle it */
		if(tobj->hp < 0.0f)
		    tobj->hp = 0.0f;
	    }
	}
	else
	{
	    /* Create small explosion */
	    float distance, bearing, size;
	    switch(wobj->type)
	    {
	      /* Stream weapon */
	      case SW_OBJ_TYPE_STREAMWEAPON:
		distance = SWMath3DDistance(
		    wobj->x - tobj->x,
		    wobj->y - tobj->y,
		    wobj->z - tobj->z
		) - (float)tobj->size / 1000.0f;
		EffectsCreateExplosionSmall(
		    wobj_num, wobj,
		    wobj->sect_x, wobj->sect_y, wobj->sect_z,
		    wobj->x + ((float)sin((double)wobj->heading) * distance),
		    wobj->y + ((float)cos((double)wobj->heading) * distance),
		    wobj->z,
		    wobj->heading, wobj->pitch, 0.0f * PI
		);
		break;

	      /* Sphere weapon */
	      case SW_OBJ_TYPE_SPHEREWEAPON:
		size = (float)tobj->size / 1000.0f;
		bearing = SWMathCoordinateDeltaVector(
		    wobj->x - tobj->x,
		    wobj->y - tobj->y
		);
		EffectsCreateExplosionSmall(
		    wobj_num, wobj,
		    wobj->sect_x, wobj->sect_y, wobj->sect_z,
		    tobj->x + ((float)sin((double)bearing) * size),
		    tobj->y + ((float)cos((double)bearing) * size),
		    tobj->z,
		    wobj->heading, wobj->pitch, 0.0f * PI
		);
		break;

	      /* Projectile/Pulse weapon */
	      case SW_OBJ_TYPE_WEAPON:
		EffectsCreateExplosionSmall(
		    wobj_num, wobj,
		    wobj->sect_x, wobj->sect_y, wobj->sect_z,
		    wobj->x, wobj->y, wobj->z,
		    wobj->heading, wobj->pitch, 0.0f * PI
		);
		break;

	      /* All else create explosion over the target object */
	      default:
		EffectsCreateExplosionSmall(
		    wobj_num, wobj,
		    tobj->sect_x, tobj->sect_y, tobj->sect_z,
		    tobj->x, tobj->y, tobj->z,
		    tobj->heading, tobj->pitch, 0.0f * PI
		);
		break;
	    }
	}

	/* The target object may be invalid if it has been destroyed
	 * and thus recycled above
	 */

	/* Recycle the weapon object if it is a projectile or pulse
	 * weapon
	 */
	if(wobj->type == SW_OBJ_TYPE_WEAPON)
	{
	    DBRecycleObject(wobj_num, cur_ms, cur_sec, tc);
	    NetSendRecycleObject(-1, wobj_num);
	    objects_recycled++;
	}

	free(wobj_name);
	free(wobj_owner_name);
	free(tobj_name);

	return(objects_recycled);
}


/*
 *	Checks if the object has come in contact with another object
 *	and then performs the appropriate collision procedure.
 *
 *	Returns true if an object has been recycled.
 */
static int REngContactCheck(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, tar_type;
	float heading_variance, distance;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;

	/* Projectile/Pulse Weapon */
	if(obj->type == SW_OBJ_TYPE_WEAPON)
	{
	    /* Begin checking if object has come in contact */

	    /* Iterate through all objects */
	    for(i = 0; i < total_objects; i++)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Skip the following types of objects */
		tar_type = tar_obj->type;
		if((tar_type <= SW_OBJ_TYPE_GARBAGE) ||
		   (tar_type == SW_OBJ_TYPE_AREA) ||
		   (tar_type == SW_OBJ_TYPE_ANIMATED)
		)
		    continue;

		/* Skip itself (owner is checked in the disarmed
		 * section below)
		 */
		if(i == obj_num)
		    continue;

		/* Skip hidden objects */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    continue;

		/* Is this a weapon that is destroyable? */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_WEP_DESTROYABLE)
		{
		    /* Do not destroy one's own weapon */
		    if(tar_obj->owner == obj->owner)
			continue;
		}
		else
		{
		    if((tar_type == SW_OBJ_TYPE_WEAPON) ||
		       (tar_type == SW_OBJ_TYPE_STREAMWEAPON) ||
		       (tar_type == SW_OBJ_TYPE_SPHEREWEAPON)
		    )
			continue;
		}

		/* Is this weapon object disarmed? */
		if(obj->antimatter <= 0.0f)
		{
		    /* No antimatter left so the weapon can be picked
		     * up
		     *
		     * Check for pick up contact
		     */ 
		    if(SWMath3DInContactPtr(obj, tar_obj))
		    {
			/* Only controlled and player objects may pick
			 * up weapons
			 */
			if((tar_type == SW_OBJ_TYPE_CONTROLLED) ||
			   (tar_type == SW_OBJ_TYPE_PLAYER)
			)
			{
			    if(WeaponsReclaimWeapon(
				i,		/* Object */
				obj_num,	/* Weapon object */
				cur_ms, cur_sec, tc
			    ))
				return(1);
			}
		    }
		}
		else
		{
		    /* This weapon has antimatter so it is armed */

		    /* Do not hit its owner */
		    if(obj->owner == i)
			continue;

		    /* Does the target object's permission level make
		     * it immune to being hit?
		     */
		    if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
			continue;

		    /* Check for contact */
		    if(SWMath3DInContactPtr(obj, tar_obj))
		    {
			/* Do hit */
		        REngDoHit(
			    legend, obj_num, i, obj->hp, obj->power,
			    cur_ms, cur_sec, tc
			);
			return(1);
		    }
		}
	    }
	}
	/* Stream Weapon */
	else if(obj->type == SW_OBJ_TYPE_STREAMWEAPON)
	{
	    /* If the stream weapon does not have power then it
	     * probably already hit something and must not hit
	     * anything else
	     */
	    if(obj->power <= 0.0f)
		return(0);

#define DO_HIT(_wobj_num_,_tobj_num_)	{	\
 REngDoHit(					\
  legend, (_wobj_num_), (_tobj_num_),		\
  obj->hp, obj->power,				\
  cur_ms, cur_sec, tc				\
 );						\
 /* Stream weapons are not recycled in		\
  * REngDoHit(), so we need to reset their	\
  * power to 0 so that they do not hit a second	\
  * object					\
  */						\
 if(!DBIsObjectGarbage(_wobj_num_)) {		\
  xsw_object_struct *obj = xsw_object[_wobj_num_]; \
  obj->hp = 0.0f;				\
  obj->power = 0.0f;				\
 }						\
}

	    /* Iterate through all home objects */
	    for(i = 0; i < total_objects; i++)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Skip non-home objects */
		tar_type = tar_obj->type;
		if(tar_type != SW_OBJ_TYPE_HOME)
		    continue;

		/* Skip itself and its owner */
		if((i == obj_num) || (i == obj->owner))
		    continue;

		/* Skip hidden objects */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    continue;

		/* Does target's permission level make it immune to
		 * being hit?
		 */
		if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
		    continue;

		/* Not in the same sector? */
	        if(!SWMath3DInSameSectorPtr(obj, tar_obj))
		    continue;

		/* Calculate the distance in screen units */
		distance = SWMath3DDistance(
		    tar_obj->x - obj->x,
		    tar_obj->y - obj->y,
		    tar_obj->z - obj->z
		) * 1000.0f;

		/* Check if the stream weapon's origin is within the
		 * target object's size
		 */
		if((long)distance <= tar_obj->size)
		{
		    DO_HIT(obj_num, i);
		    return(1);
		}

		/* Calculate the distance variance angle based on the
		 * size (radius) of the target and the distance to it
		 */
		if(distance > 0.0f)
		{
		    heading_variance = atan2(
			tar_obj->size,		/* dy */
			distance		/* dx */
		    );
		}
		else
		{
		    /* No distance away, so partical hits partical
		     * regardless of their bearings to each other
		     */
		    heading_variance = (2.0 * PI);
		}

		/* Check if in vector contact */
		if(SWMath3DInVectorContact(
		    xsw_object, total_objects,
		    obj_num,		/* Source */
		    i,			/* Target */
		    obj->heading,	/* Bearing source to target */
		    heading_variance,
		    obj->size		/* Size of source */
		))
		{
		    DO_HIT(obj_num, i);
		    return(1);
		}
	    }

	    /* Iterate through all non-home objects */ 
	    for(i = 0; i < total_objects; i++)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Skip the following types of objects */
		tar_type = tar_obj->type;
		if((tar_type <= SW_OBJ_TYPE_GARBAGE) ||
		   (tar_type == SW_OBJ_TYPE_AREA) ||
		   (tar_type == SW_OBJ_TYPE_ANIMATED)
		)
		    continue;

		/* Skip itself and its owner */
		if((i == obj_num) || (i == obj->owner))
		    continue;

		/* Skip hidden objects */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    continue;

		/* Is this a weapon that is destroyable? */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_WEP_DESTROYABLE)
		{
		    /* Do not destroy one's own weapon */
		    if(tar_obj->owner == obj->owner)
			continue;
		}
		else
		{
		    if((tar_type == SW_OBJ_TYPE_WEAPON) ||
		       (tar_type == SW_OBJ_TYPE_STREAMWEAPON) ||
		       (tar_type == SW_OBJ_TYPE_SPHEREWEAPON)
		    )
			continue;
		}

		/* Does target's permission level make it immune to
		 * being hit?
		 */
		if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
		    continue;

		/* Not in the same sector? */
	        if(!SWMath3DInSameSectorPtr(obj, tar_obj))
		    continue;

		/* Calculate the distance in screen units */
		distance = SWMath3DDistance(
		    tar_obj->x - obj->x,
		    tar_obj->y - obj->y,
		    tar_obj->z - obj->z
		) * 1000.0f;

		/* Check if the stream weapon's origin is within the
		 * target object's size
		 */
		if((long)distance <= tar_obj->size)
		{
		    DO_HIT(obj_num, i);
		    return(1);
		}

		/* Calculate the distance variance angle based on the
		 * size (radius) of the target and the distance to it
		 */
		if(distance > 0.0f)
		{
		    heading_variance = atan2(
			tar_obj->size,		/* dy */
			distance		/* dx */
		    );
		}
		else
		{
		    /* No distance away, so partical hits partical
		     * regardless of their bearings to each other
		     */
		    heading_variance = (2.0 * PI);
		}

		/* Check if in vector contact */
		if(SWMath3DInVectorContact(
		    xsw_object, total_objects,
		    obj_num,		/* Source */
		    i,			/* Target */
		    obj->heading,	/* Bearing source to target */
		    heading_variance,
		    obj->size		/* Size of source */
		))
		{
		    DO_HIT(obj_num, i);
		    return(1);
		}
	    }

#undef DO_HIT
	}
	/* Sphere Weapon */
	else if(obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	{
	    int nhits = 0;

	    /* Sphere weapon object must have power */
	    if(obj->power <= 0.0f)
		return(0);

	    /* If home objects can not be destroyed then they are safe
	     * havens in which of a sphere weapon hits a home object
	     * first then it should hit no other objects so check if
	     * the sphere weapon his a home object first
	     */
	    if(!sysparm.homes_destroyable)
	    {
		for(i = 0; i < total_objects; i++)
		{
		    tar_obj = xsw_object[i];
		    if(tar_obj == NULL)
			continue;

		    if(tar_obj->type != SW_OBJ_TYPE_HOME)
			continue;

		    /* Skip itself and its owner */  
		    if((i == obj_num) || (i == obj->owner))
			continue;

		    /* Skip hidden objects */
		    if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* Does target's permission level make it immune to
		     * being hit?
		     */
		    if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
			continue;

		    /* Did the sphere weapon hit this home object? */
		    if(SWMath3DInContactPtr(obj, tar_obj))
		    {
			/* Process hit on this home object then set the
			 * sphere weapon's power to 0 and return with a
			 * hit
			 */
			REngDoHit(
			    legend, obj_num, i, obj->hp, obj->power,
			    cur_ms, cur_sec, tc
			);
			if(!DBIsObjectGarbage(obj_num))
			{
			    xsw_object_struct *obj = xsw_object[obj_num];
			    obj->hp = 0.0f;
			    obj->power = 0.0f;
			}
			return(1);
		    }
		}
	    }

	    /* Iterate through each object
	     *
	     * Each object within the sphere weapon's radius gets an
	     * equal amount of damage
	     */
	    for(i = 0; i < total_objects; i++)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Skip the following types of objects */
		tar_type = tar_obj->type;
		if((tar_type <= SW_OBJ_TYPE_GARBAGE) ||
		   (tar_type == SW_OBJ_TYPE_AREA) ||
		   (tar_type == SW_OBJ_TYPE_ANIMATED)
		)
		    continue;

		/* Skip itself and its owner */
		if((i == obj_num) || (i == obj->owner))
		    continue;

		/* Skip hidden objects */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    continue;

		/* Is this a weapon that is destroyable? */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_WEP_DESTROYABLE)
		{
		    /* Do not destroy one's own weapon */
		    if(tar_obj->owner == obj->owner)
			continue;
		}
		else
		{
		    if((tar_type == SW_OBJ_TYPE_WEAPON) ||
		       (tar_type == SW_OBJ_TYPE_STREAMWEAPON) ||
		       (tar_type == SW_OBJ_TYPE_SPHEREWEAPON)
		    )
			continue;
		}

		/* Does target's permission level make it immune to
		 * being hit?
		 */
		if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
		    continue;

		/* Did the sphere weapon hit this object? */
		if(SWMath3DInContactPtr(obj, tar_obj))
		{
		    /* Do hit */
		    REngDoHit(
			legend, obj_num, i, obj->hp, obj->power,
			cur_ms, cur_sec, tc
		    );
		    nhits++;

		    if(DBIsObjectGarbage(obj_num))
			break;
		}
	    }

	    /* If the sphere weapon hit one or more objects then set
	     * its power to 0 so that it does not hit anything on
	     * the next call to this function
	     */
	    if(nhits > 0)
	    {
		if(!DBIsObjectGarbage(obj_num))
		{
		    xsw_object_struct *obj = xsw_object[obj_num];
		    obj->hp = 0.0f;
		    obj->power = 0.0f;
		}
		return(1);
	    }
	}

	return(0);
}


/*
 *	Applies damage to the object if it is in an asteroid field
 *	and that its current velocity has exceeded the asteroid field's
 *	velocity_max
 *
 *	Calls to this function need to be properly timmed.
 *
 *	Returns true if the object has been recycled.
 */
int REngAsteroidFieldDamage(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int objects_recycled = 0;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return(objects_recycled);

	obj = xsw_object[obj_num];

	/* Is this a controlled or player object and is it in an
	 * asteroid field? */
	if(((obj->type == SW_OBJ_TYPE_CONTROLLED) ||
	    (obj->type == SW_OBJ_TYPE_PLAYER)) &&
	    (obj->loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
	)
	{
	    int area_obj_num = obj->loc_obj_num;
	    xsw_object_struct *area_obj = DBIsObjectGarbage(area_obj_num) ?
		NULL : xsw_object[area_obj_num];
	    if(area_obj != NULL)
	    {
		int af_num = area_obj_num;

		/* Is the object going faster than the set
		 * maximum of the asteroid field and is the object
		 * able to receive damage?
		 */
		if((area_obj->velocity_max > 0.0f) &&
		   (obj->velocity >= area_obj->velocity_max) &&
		   (obj->permission.uid > SWS_ACCESS_UID_DESTROY_IMMUNE) &&
		   (legend->asteroid_field_damage_rate > 0.0f) &&
		   !(obj->ext_flags & SWS_OBJ_FLAG_DESTROYED)
		)
		{
		    float	damage_structure = 0.0f,
				damage_shield = 0.0f,
				total_damage = 0.0f,
				damage_delta = ((float)obj->size / 1000.0f) *
				    (obj->velocity / area_obj->velocity_max) *
				    legend->asteroid_field_damage_rate *
				    (legend->asteroid_field_damage_int / 1000l);

		    /* Shields up? */
		    if((obj->shield_state == SW_SHIELD_STATE_UP) &&
		       (obj->power > 0.0f) && (obj->power_max > 0.0f)
		    )
		    {
			float damage_remainder = 0.0f;

			/* Calculate extra damage to shields due to
			 * the target object's lack of power purity
			 */
			const float extra_damage_to_shields = MAX(
			    ((damage_delta * (2.0 - obj->power_purity)) -
			    damage_delta),
			    0.0f
			);

			damage_shield += damage_delta + extra_damage_to_shields;
			total_damage += damage_shield;
			obj->power -= damage_shield;

			if(obj->power < 0.0f)
			{
			    damage_remainder = -obj->power;
			    damage_shield -= damage_remainder;
			    total_damage -= damage_remainder;
			    obj->power = 0.0f;
			}

			damage_structure += damage_remainder;
			total_damage += damage_structure;
			obj->hp -= damage_structure;

			if(obj->hp < 0.0f)
			{
			    damage_structure += obj->hp;
			    total_damage += obj->hp;
			}
		    }
		    else
		    {
			damage_structure += damage_delta;
			total_damage += damage_structure;
			obj->hp -= damage_structure;

			if(obj->hp < 0.0f)
			{
			    damage_structure += obj->hp;
			    total_damage += obj->hp;
			}
		    }

		    /* Increase damage given to the object */
		    if(!UNVAllocScores(obj))
		    {
			xsw_score_struct *score = obj->score;
			score->damage_received += total_damage;
			NetSendScore(-1, obj_num);
		    }

		    /* Was the object destroyed by the damage caused
		     * by this asteroid field?
		     */
		    if(obj->hp < 0.0f)
		    {
			int i;
			char	*s,
				loc_str[256],
				obj_name[SW_OBJ_NAME_MAX + 80],
				af_name[SW_OBJ_NAME_MAX + 80];
			sws_connection_struct *con;

			/* Mark it as destroyed */
			obj->ext_flags |= SWS_OBJ_FLAG_DESTROYED;

			/* Get the object's name */
			strncpy(
			    obj_name, DBGetFormalNameStr(obj_num),
			    sizeof(obj_name)
			);
			obj_name[sizeof(obj_name) - 1] = '\0';
			strncpy(
			    af_name, DBGetFormalNameStr(af_num),
			    sizeof(af_name)
			);
			af_name[sizeof(af_name) - 1] = '\0';

			/* Create debris */
			EffectsCreateDebrisObject(obj_num, obj, cur_ms, cur_sec, tc);

			/* Create large explosion */
			EffectsCreateExplosionLarge(
			    obj_num, obj,
			    obj->sect_x, obj->sect_y, obj->sect_z,
			    obj->x, obj->y, obj->z,
			    obj->heading, obj->pitch, 0.0f * PI
			);

			/* Create shockwave */
			EffectsCreateShockwave(
			    obj_num, obj,
			    obj->size, obj->size * 4l,
			    obj->power, obj->power
			);

#ifdef PLUGIN_SUPPORT
			/* Notify all plugins */
			for(i = 0; i < total_plugins; i++)
			    PluginDestroyNotify(
				i,
				SW_DESTROY_REASON_COLLISION,
				obj_num,	/* Destroyed object */
				af_num,		/* Object that caused destruction */
				af_num		/* Owner of weapon object */
			    );
#endif

			/* Format location string */
			UNVLocationFormatString(
			    loc_str,
			    &obj->sect_x, &obj->sect_y, &obj->sect_z,
			    &obj->x, &obj->y, &obj->z,
			    sizeof(loc_str)
			);

			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(obj_name) +
				STRLEN(af_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed due to collision with %s at %s.",
				obj_name, af_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(obj_name) +
			    STRLEN(af_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed due to collision with %s at %s.",
			    obj_name, af_name, loc_str
			);
			switch(obj->type)
			{
			  case SW_OBJ_TYPE_CONTROLLED:
			  case SW_OBJ_TYPE_PLAYER:
			    ExportEvents(
				fname.events_export,
				EVST_TYPE_DESTROYEDPLAYER,
				s
			    );
			    break;
			  default:
			    ExportEvents(
				fname.events_export,
				EVST_TYPE_DESTROYEDOBJECT,
				s
			    );
			    break;
			}
			free(s);

			/* Send destroy notify to all connections in
			 * the same sector
			 */
			for(i = 0; i < total_connections; i++)
			{
			    con = connection[i];
			    if(con == NULL)
				continue;

			    if((con->socket < 0) || (con->object_num < 0))
				continue;

			    /* Not in the same sector? */
			    if(!SWMath3DInSameSector(
				xsw_object, total_objects,
				con->object_num, obj_num
			    ))
				continue;

			    /* Send destroy notify to this connection */
/*			    NetSendObjectName(i, obj_num); */
			    NetSendNotifyDestroy(
				i,
				SW_DESTROY_REASON_COLLISION,
				obj_num,	/* Destroyed object */
				af_num,		/* Object that caused destruction */
				af_num		/* Owner of weapon object */
			    );
			}

			/* Check if the destroyed object is a player
			 * and if players should be respawned,
			 * otherwise recycle the destroyed object
			 *
			 * Note that after this point, obj should be
			 * considered invalid. It can still be used
			 * as cross reference but it must never be
			 * referenced implicitily
			 */
			if((obj->type == SW_OBJ_TYPE_PLAYER) &&
			   sysparm.respawn_players
			)
			{
			    REngDoRespawn(
				legend, obj_num,
				cur_ms, cur_sec, tc
			    );
			}
			else
			{
			    DBRecycleObject(obj_num, cur_ms, cur_sec, tc);
			    NetSendRecycleObject(-1, obj_num);
			    objects_recycled++;
			}
		    }
		    else
		    {
#ifdef PLUGIN_SUPPORT
			int i;
#endif
			const int con_num = ConGetByObject(obj_num);

#ifdef PLUGIN_SUPPORT
			/* Send hit notify to all plugins */
			for(i = 0; i < total_plugins; i++)
			    PluginHitNotify(
				i,
				af_num, obj_num,
				obj->velocity_heading,
				damage_structure,
				damage_shield
			    );
#endif

			if(con_num > -1)
			{
			    /* Send hit notify to this connection */
			    NetSendNotifyHit(
				con_num,
				af_num, obj_num,
				obj->velocity_heading,
				damage_structure,
				damage_shield
			    );
 			}

			/* Create a small explosion */
			EffectsCreateExplosionSmall(
			    obj_num, obj,
			    obj->sect_x, obj->sect_y, obj->sect_z,
			    obj->x, obj->y, obj->z,
			    obj->heading, obj->pitch, 0.0f * PI
			);
		    }
		}
	    }
	}

	return(objects_recycled);
}


/*
 *	AI simulation for controlled objects.
 *
 *	If the sysparm use_server_ai is set to false then this function
 *	does nothing.
 *
 *	Returns true if an object has been created or recycled.
 */
static int REngAI(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int	status = 0,
		i, closest_fobject, closest_uobject,
		type, sel_wep, weapons_fired, iff_code;
	char *s;
	sw_flags_t ai_flags;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;
	xsw_weapon_struct *wep;

	long closest_wep_range;		/* In pixel units */
	float	scanner_range,		/* In real units */
		distance,		/* In real units */
		closest_fdistance,	/* In real units */
		closest_udistance;	/* In real units */
	float ovisibility;

	char obj_name[SW_OBJ_NAME_MAX + 80];

	const int loc_str_len = 256;
	char loc_str[loc_str_len];


	/* Server AI disabled or is the object not of the type
	 * controlled?
	 */
	if(!sysparm.use_server_ai ||
	   (obj->type != SW_OBJ_TYPE_CONTROLLED)
	)
	    return(0);


	/* Reset closest objects */
	closest_fobject = -1;
	closest_uobject = -1;

	/* Get this object's AI flags */
	ai_flags = obj->ai_flags;

	/* Get this object's scanner range */
	scanner_range = obj->scanner_range;
	closest_fdistance = scanner_range;
	closest_udistance = scanner_range;

	/* Iterate through the objects list */
	for(i = 0; i < total_objects; i++)
	{
	    tar_obj = xsw_object[i];

	    /* Skip ourselves */
	    if(obj == tar_obj)
		continue;

	    /* Checks if objects are valid and in the same sector */
/* What about objects that need to follow objects across sectors? */
	    if(!SWMath3DInSameSectorPtr(obj, tar_obj))
		continue;

	    /* React to only controlled and player objects */
	    type = tar_obj->type;
	    if((type != SW_OBJ_TYPE_CONTROLLED) &&
	       (type != SW_OBJ_TYPE_PLAYER)
	    )
		continue;

	    /* Skip objects that are marked hidden, AI's should
	     * ignore these
	     */
	    if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		continue;


	    /* Get the target object's current visibility */
	    ovisibility = DBGetObjectVisibilityPtr(tar_obj);

	    /* Calculate the distance to the target object in real
	     * units
	     */
	    distance = SWMath3DDistance(
		tar_obj->x - obj->x,
		tar_obj->y - obj->y,
		tar_obj->z - obj->z
	    );

	    /* Check IFF */
	    iff_code = UNVMatchIFFPtr(obj, tar_obj);
	    if(iff_code == SW_IFF_FRIENDLY)
	    {
		/* Friendly */
		/* Range check */
	        if((distance > (scanner_range * ovisibility)) ||
		   (distance > closest_fdistance) ||
		   (type == SW_OBJ_TYPE_CONTROLLED)
	        )
		    continue;

		closest_fdistance = distance;
		closest_fobject = i;
	    }
	    else
	    {
		/* Hostile or unknown */
		/* Range check */
		if((distance > (scanner_range * ovisibility)) ||
		   (distance > closest_udistance)
		)
		    continue;

		closest_udistance = distance;
		closest_uobject = i;
	    }
	}


	/* ******************************************************* */
	/* Follow: Adjust intercepting object */
	while((ai_flags & SW_AI_FIRE_ANY) ||
	      (ai_flags & SW_AI_FOLLOW_ANY)
	)
	{
	    /* Hostile or unknown (has precidence over friendly) */
	    if(((ai_flags & SW_AI_FIRE_UNKNOWN) ||
		(ai_flags & SW_AI_FIRE_HOSTILE) ||
		(ai_flags & SW_AI_FOLLOW_UNKNOWN) ||
		(ai_flags & SW_AI_FOLLOW_HOSTILE)
	       ) && (closest_uobject > -1)
	    )
	    {
		if(obj->intercepting_object != closest_uobject)
		{
		    obj->intercepting_object = closest_uobject;
		    obj->locked_on = closest_uobject;
		}   
	    }
	    /* Friendly */
	    else if(((ai_flags & SW_AI_FIRE_FRIEND) ||
		     (ai_flags & SW_AI_FOLLOW_FRIEND)
		     ) && (closest_fobject > -1)
	    )
	    {
		if(obj->intercepting_object != closest_fobject)
		{
		    obj->intercepting_object = closest_fobject;
		    obj->locked_on = closest_fobject;
		}
	    }
	    else
	    {
		if(obj->intercepting_object > -1)
		{
		    obj->intercepting_object = -1;
		    obj->locked_on = -1;
		}
	    }

	    break;		/* Break out of this while() loop */
	}
	/* ******************************************************* */
	/* Follow: Adjust throttle */
	while(ai_flags & SW_AI_FOLLOW_ANY)
	{
	    /* Actions in this section need to corrilate with
	     * those in the above intercepting object check
	     * since `following' implies concensus in throttle
	     * and intercept.
	     */

	    /* Hostile or unknown (has precidence over friendly) */
	    if(((ai_flags & SW_AI_FOLLOW_UNKNOWN) ||
		(ai_flags & SW_AI_FOLLOW_HOSTILE)
		) &&
		(closest_uobject > -1)
	    )
	    {
		/* Adjust throttle */
		if(closest_udistance > 0.6f)
		{
		    obj->throttle += (0.02f * tc);
		}
		else
		{
		    obj->throttle -= (0.02f * tc);

		    /* Apply external dampers (brakes), this may
		     * be a bit stronger than it is equated on the
		     * client
		     */
#if 0
		    obj->velocity -= (obj->thrust_power *
			obj->power_purity * tc
		    );
#endif
		    obj->velocity -= (0.04f * tc);
		    if(obj->velocity < 0.0f)
			obj->velocity = 0.0;
		}

		/* Sanitize throttle value */
		if(obj->throttle > 1.0f)
		    obj->throttle = 1.0f;
		else if(obj->throttle < 0.0f)
		    obj->throttle = 0.0f;
	    }
	    /* Friendly */
	    else if((ai_flags & SW_AI_FOLLOW_FRIEND) &&
		    (closest_fobject > -1)
	    )
	    {
		/* Adjust throttle */
		if(closest_fdistance > 0.6f)
		{
		    obj->throttle += (0.02f * tc);
		}
		else
		{
		    obj->throttle -= (0.02f * tc);

		    /* Apply external dampers (brakes), this may
		     * be a bit stronger than it is equated on the
		     * client
		     */
#if 0
		    obj->velocity -= (obj->thrust_power *
			obj->power_purity * tc
		    );
#endif
		    obj->velocity -= (0.04f * tc); 
		    if(obj->velocity < 0.0f)
			obj->velocity = 0.0f; 
		}

		/* Sanitize throttle value */
		if(obj->throttle > 1.0f)
		    obj->throttle = 1.0f;
		else if(obj->throttle < 0.0f)
		    obj->throttle = 0.0f;
	    }
	    else
	    {    
		obj->throttle = 0.0f;
	    }

	    break;	/* Break out of this while() loop */
	}
	/* ******************************************************* */
	/* Fire */
	while(ai_flags & SW_AI_FIRE_ANY)
	{
	    /* Same some cpu power if object has no weapons */
	    if(obj->total_weapons <= 0)
		break;

	    /* Iterate through list of weapons and set
	     * closest_wep_range to be the distance of the weapon
	     * with the longest range on the object
	     */
	    for(i = 0, closest_wep_range = 0;
		i < obj->total_weapons;
		i++
	    )
	    {
		/* Get weapon pointer and check if it's valid */
		wep = obj->weapons[i];
		if(wep == NULL)
		    continue;

		/* In pixel units */
		if(wep->range > closest_wep_range)
		    closest_wep_range = wep->range;
	    }


	    /* Check if there is a target object in range */

	    /* Hostile or unknown (precidence over friendly) */
	    if(((ai_flags & SW_AI_FIRE_UNKNOWN) ||
		(ai_flags & SW_AI_FIRE_HOSTILE)
		) && (closest_uobject > -1)
	    )
	    {
		/* Check if unknown object is out of range */
		if(closest_udistance > ((float)closest_wep_range / 1000))
		    break;

		/* Go through weapons list and select weapon with
		 * the closest range to fire on the unknown object
		 */
		for(i = 0, obj->selected_weapon = -1;
		    i < obj->total_weapons;
		    i++
		)
		{
		    /* Get weapon pointer and check if its valid */
		    wep = obj->weapons[i];
		    if(wep == NULL)
			continue;

		    /* Is the target inside this weapons range
		     * (in XSW Real units)?
		     */ 
		    if(closest_udistance > ((float)wep->range / 1000))
			continue;

		    /* Is this weapons range closer? */
		    if(wep->range <= closest_wep_range)
		    {
			/* Select this weapon */
			obj->selected_weapon = i;

			/* Remember closest weapon range */
			closest_wep_range = wep->range;
		    }
		}
	    }
	    /* Friendly */
	    else if((ai_flags & SW_AI_FIRE_FRIEND) &&
		    (closest_fobject > -1)
	    )
	    {
		/* Check if friendly object is out of range */
		if(closest_fdistance > ((float)closest_wep_range / 1000.0f))
		    break;

		/* Go through weapons list and select weapon with
		 * the closest range to fire on the friendly object
		 */
		for(i = 0, obj->selected_weapon = -1;
		    i < obj->total_weapons;
		    i++
		)
		{
		    /* Get weapon pointer and check if its valid */
		    wep = obj->weapons[i];
		    if(wep == NULL)
			continue;

		    /* Is the target inside this weapons range
		     * (in real units)?
		     */
		    if(closest_fdistance > ((float)wep->range / 1000.0f))
			continue;

		    /* Is this weapons range closer? */
		    if(wep->range <= closest_wep_range)
		    {
			/* Select this weapon */
			obj->selected_weapon = i;

			/* Remember closest weapon range */
			closest_wep_range = wep->range;
		    }
		}
	    }
	    else
	    {
		break;
	    }

	    /* Note: Procedure here is similar to the procedure
	     * in function NetHandleFireWeapon()
	     */ 

	    /* Get selected weapon pointer and check if its valid
	     * (if object has no weapons, then execution never
	     * gets past this point)
	     */
	    sel_wep = obj->selected_weapon;
	    if((sel_wep < 0) || (sel_wep >= obj->total_weapons))
		break;

	    wep = obj->weapons[sel_wep];
	    if(wep == NULL)
		break;

	    /* Too early to fire again? */
	    if((wep->delay + wep->last_used) >= cur_ms)
		break;

	    /* Weapon may not be fired while cloaked? */
	    if(!(wep->flags & SW_WEP_FLAG_FIRE_CLOAKED))
	    {
		/* Cloak raised? */
		if(obj->cloak_state == SW_CLOAK_STATE_UP)
		    break;

		/* Cloak refractory interval not passed? */
		if((obj->last_cloak + obj->cloak_refractory_int) >=
		   cur_ms
		)
		    break;
	    }

	    /* If weapon is expendable, see if object has any left */
	    if(wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
	    {
		if(wep->amount <= 0)
		    break;
	    }

	    /* Does object have enough power to fire weapon? */
	    if(obj->power < wep->create_power)
		break; 

	    /* Create (fire) the weapon object */
	    /* Stream weapons */
	    if(wep->emission_type == SW_WEP_EMISSION_STREAM)
	    {
		weapons_fired = WeaponsFire(
		    wep->ocs_code,		/* OCS */
		    obj_num,			/* Owner */
		    wep->emission_type,		/* Emission type */
		    obj->x,
		    obj->y,
		    obj->heading,
		    obj->pitch,
		    wep->damage_hull,
		    wep->damage_shields,
		    wep->range,
		    SW_FREQ_MIN,	/* Minimal frequency for now */
		    wep->flags,
		    cur_ms, cur_sec, tc
		);
	    }
	    /* Projectile */
	    else if(wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
	    {
		weapons_fired = WeaponsFire(
		    wep->ocs_code,		/* OCS */
		    obj_num,			/* Owner */
		    wep->emission_type,		/* Emission type */
		    obj->x,
		    obj->y,
		    obj->heading,
		    obj->pitch,
		    wep->damage_hull,
		    wep->damage_shields,
		    wep->range,
		    SW_FREQ_MIN,	/* Minimal frequency for now */
		    wep->flags,
		    cur_ms, cur_sec, tc
		);
	    }
	    /* All else assume pulse */
	    else
	    {
		weapons_fired = WeaponsFire(
		    wep->ocs_code,		/* OCS */
		    obj_num,			/* Owner */
		    wep->emission_type,		/* Emission type */
		    obj->x,
		    obj->y,
		    obj->heading,
		    obj->pitch,
		    wep->damage_hull,
		    wep->damage_shields,
		    wep->range,
		    SW_FREQ_MIN,	/* Minimal frequency for now */
		    wep->flags,
		    cur_ms, cur_sec, tc
		);
	    }

	    /* Reget object pointer, creating object(s) above may
	     * change the pointer value
	     */
	    obj = xsw_object[obj_num];

	    /* Update status to report memory change */
	    status = 1;

	    /* No objects created? */
	    if(weapons_fired <= 0)
		break;


	    /* Mark the last time it was fired */
	    wep->last_used = cur_ms;


	    /* Get object name for logging purposes */
	    strncpy(
		obj_name, DBGetFormalNameStr(obj_num),
		sizeof(obj_name)
	    );
	    obj_name[sizeof(obj_name) - 1] = '\0';

	    /* Format location string */
	    UNVLocationFormatString(
		loc_str,
		&obj->sect_x, &obj->sect_y, &obj->sect_z,
		&obj->x, &obj->y, &obj->z,
		loc_str_len
	    );

	    /* Consume power on object for the power that was
	     * needed to create this weapon and export event
	     */
	    switch(wep->emission_type)
	    {
	      case SW_WEP_EMISSION_PROJECTILE:
		obj->power -= (wep->create_power * weapons_fired);
		/* Decrease stock for projectiles */
		wep->amount -= weapons_fired;
		if(wep->amount < 0)
		    wep->amount = 0;

		/* Export projectile weapons fire */
		s = (char *)malloc(
		    STRLEN(obj_name) +
		    SW_OBJ_NAME_MAX +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
		    "%s fired %s at %s.",
		    obj_name,
		    DBGetOCSOPMName(wep->ocs_code),
		    loc_str
		);
		ExportEvents(
		    fname.events_export,
		    EVST_TYPE_PROJECTILEWEAPON_FIRED,
		    s
		);
		free(s);
		break;

	      case SW_WEP_EMISSION_PULSE:
		obj->power -= (wep->create_power * weapons_fired);
		break;
	    
	      /* Default to stream weapon */
	      default:
		obj->power -= wep->create_power;
 	        break;
	    }
	    /* Sanitize power on object */
	    if(obj->power < 0.0f)
		obj->power = 0.0f;

	    /* Report new object and weapons values to objects nearby */
	    NetFireWeaponSyncObjectValues(
		obj, obj_num,
		sel_wep,
		cur_ms, cur_sec, tc
	    );

#if 0
	    /* Set next object values update to now so it gets updated now */
	    next.object_values = cur_ms;
	    next.need_weapon_values = 1;
#endif

	    break;	/* Break out of this while() loop */
	}


	return(status);
}


/*
 *	Updates the object.
 *
 *	Returns true if the object was recycled during this call.
 */
static int REngUpdateObject(
	sw_legend_struct *legend, const int obj_num,
	const time_t cur_ms, const time_t cur_sec,
	const float tc		/* Time compensation coefficient */
)
{
	REngVisibility(legend, obj_num, cur_ms, cur_sec, tc);
	REngShieldVisibility(legend, obj_num, cur_ms, cur_sec, tc);

	if(REngMortality(legend, obj_num, cur_ms, cur_sec, tc))
	    return(1);

	REngAnimated(legend, obj_num, cur_ms, cur_sec, tc);

	if(REngAutoDestruct(legend, obj_num, cur_ms, cur_sec, tc))
	    return(1);

	REngPowerCore(legend, obj_num, cur_ms, cur_sec, tc);
	REngDamageControl(legend, obj_num, cur_ms, cur_sec, tc);

	REngWeaponsLock(legend, obj_num, cur_ms, cur_sec, tc);
	REngTractorBeam(legend, obj_num, cur_ms, cur_sec, tc);
	REngIntercept(legend, obj_num, cur_ms, cur_sec, tc);

	REngThrottle(legend, obj_num, cur_ms, cur_sec, tc);
	REngThrust(legend, obj_num, cur_ms, cur_sec, tc);

	REngHeading(legend, obj_num, cur_ms, cur_sec, tc);
	REngLocation(legend, obj_num, cur_ms, cur_sec, tc);
	REngMove(legend, obj_num, 1, 0, cur_ms, cur_sec, tc);

	if(REngContactCheck(legend, obj_num, cur_ms, cur_sec, tc))
	    return(1);

	if(REngAI(legend, obj_num, cur_ms, cur_sec, tc))
	    return(1);

	return(0);
}


/*
 *	Initializes the reality engine.
 */
int REngInit(void)
{
	return(0);
}

/*
 *	Manages the reality engine.
 */
void REngManage(
	sw_legend_struct *legend,
	const time_t cur_ms, const time_t cur_sec,
	const float tc		/* Time compensation coefficient */
)
{
	int obj_num;
	xsw_object_struct *obj;

	/* Iterate through the objects list */
	for(obj_num = 0; obj_num < total_objects; obj_num++)
	{
	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Update this object */
	    REngUpdateObject(
		legend, obj_num,
		cur_ms, cur_sec, tc
	    );
	}
}

/*
 *	Shuts down the reality engine.
 */
void REngShutdown(void)
{
	return;
}
