#include <stdlib.h>
#include <string.h>

#include "../include/disk.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "uew.h"
#include "aboutdlg.h"
#include "ue.h"
#include "config.h"

#include "../include/oswkeyvals.h"


static void AboutDlgOKCB(push_button_struct *btn, void *data);

about_dlg_struct *AboutDlgNew(void);
void AboutDlgDraw(about_dlg_struct *d, const int amount);
int AboutDlgManage(about_dlg_struct *d, event_t *event);          
void AboutDlgMap(about_dlg_struct *d, win_t toplevel);
void AboutDlgUnmap(about_dlg_struct *d);
void AboutDlgDelete(about_dlg_struct *d);


#define ABOUT_DLG_TITLE			"About " PROG_NAME_FULL
#define ABOUT_DLG_ICON_TITLE		"About " PROG_NAME
#define ABOUT_DLG_URL			PROG_URL

#define	ABOUT_DLG_MARGIN		10

#define ABOUT_DLG_WIDTH			350
#define ABOUT_DLG_HEIGHT		230

#define ABOUT_DLG_BUTTON_WIDTH		70
#define ABOUT_DLG_BUTTON_HEIGHT		28


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	OK button callback.
 */
static void AboutDlgOKCB(push_button_struct *btn, void *data)
{
	about_dlg_struct *d = ABOUT_DLG(data);
	if(d == NULL)
	    return;

	AboutDlgUnmap(d);
}


/*
 *	Creates a new About Dialog.
 */
about_dlg_struct *AboutDlgNew(void)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent;
	push_button_struct *btn;
        widget_global_struct *wglobal = &widget_global;
	about_dlg_struct *d = ABOUT_DLG(calloc(1, sizeof(about_dlg_struct)));

	d->map_state = False;
	d->has_focus = False;
	d->x = 0;
	d->y = 0;
	d->width = ABOUT_DLG_WIDTH;
	d->height = ABOUT_DLG_HEIGHT;

	/* Create toplevel */
	parent = gui->root_win;
	x = d->x;
	y = d->y;
	width = d->width;
	height = d->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	d->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    ABOUT_DLG_TITLE,
	    ABOUT_DLG_ICON_TITLE,
	    (ue_image.icon_unvedit_pixmap != None) ?
		ue_image.icon_unvedit_pixmap :
		wglobal->std_icon_pm,
	    False,		/* Let the WM set the position? */
	    d->x, d->y,
	    d->width, d->height,
	    d->width, d->height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);

	parent = w;

	/* Dismiss button */
	btn = &d->ok_btn;
	x = ((int)d->width - ABOUT_DLG_BUTTON_WIDTH) / 2;
	y = (int)d->height - ABOUT_DLG_BUTTON_HEIGHT - ABOUT_DLG_MARGIN;
	width = ABOUT_DLG_BUTTON_WIDTH;
	height = ABOUT_DLG_BUTTON_HEIGHT;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    AboutDlgOKCB, d
	);
	accel_keys[0] = OSW_KEY_o;
	accel_keys[1] = OSW_KEY_Escape;
	accel_keys[2] = OSW_KEY_space;
	accel_keys[3] = OSW_KEY_Return;
	PBtnSetAccelKeys(btn, accel_keys, NULL, 4);

	return(d);
}

/*
 *	Draws the about dialog.
 */
void AboutDlgDraw(about_dlg_struct *d, const int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(d == NULL)
	    return;

	if(amount == ABOUT_DLG_DRAW_AMOUNT_COMPLETE)
	{
	    int	x, y,
		lbearing, rbearing, swidth, ascent, descent,
		font_height;
	    unsigned int width, height;
	    const char *s;
	    font_t *font = wglobal->std_font;
	    win_t w = d->toplevel;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Draw the background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);

		x = ABOUT_DLG_MARGIN;
		y = ABOUT_DLG_MARGIN;

		OSWSetForeground(gui->white_pix);
		font = wglobal->std_font;
		OSWSetFont(font);
		s = PROG_NAME_FULL;
                OSWGetTextBounds(
                    font, s, -1,
                    &lbearing, &rbearing, &swidth, &ascent, &descent
                );
                font_height = ascent + descent;
                x = ((int)width - swidth) / 2;
                OSWDrawString(
                    w,
                    x - lbearing,
                    y + ascent,
                    s
                );
		y += font_height + (ABOUT_DLG_MARGIN / 2);
	    }
	    else
	    {
		image_t *img;

		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);

		y = (int)height - ABOUT_DLG_MARGIN -
		    ABOUT_DLG_BUTTON_HEIGHT - ABOUT_DLG_MARGIN;
		OSWSetForeground(wglobal->surface_shadow_pix);
		OSWDrawLine(w, 0, y, (int)width, y);

		y += 1;
		OSWSetForeground(wglobal->surface_highlight_pix);
		OSWDrawLine(w, 0, y, (int)width, y);

		x = ABOUT_DLG_MARGIN;
		y = ABOUT_DLG_MARGIN;

		img = ue_image.unvedit_logo;
		if(img != NULL)
		{
		    x = ((int)width - (int)img->width) / 2;
		    OSWPutImageToDrawablePos(img, w, x, y);
		    y += (int)(img->height * 0.75f);
		}

		OSWSetForeground(wglobal->normal_text_pix);
	    }


	    /* Version */
	    font = wglobal->std_font;
	    OSWSetFont(font);
	    s = "Version " PROG_VERSION;
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    x = ((int)width - swidth) / 2;
	    OSWDrawString(
		w,
		x - lbearing,
		y + ascent,
		s
	    );
	    y += font_height + (ABOUT_DLG_MARGIN / 2);

            font = wglobal->std_font;
            OSWSetFont(font);
	    s = ABOUT_DLG_URL;
	    OSWGetTextBounds(
                font, s, -1,
                &lbearing, &rbearing, &swidth, &ascent, &descent
            );
            font_height = ascent + descent;
            x = ((int)width - swidth) / 2;
            OSWDrawString(
                w,
                x - lbearing,
                y + ascent,
                s
            );
            y += font_height + (ABOUT_DLG_MARGIN / 2);


#if 0
	    /* Memory line 1 */
	    y += (12 * 2);
	    sprintf(text,
		"Memory:"
	    );
	    OSWDrawString(pixmap, x + 3, y + (10 / 2) + 3,
		text
	    );

	    y += 12;
	    sprintf(text,
		"    Program Used: %ld kb",
		d->mem_used / 1000
	    );
	    OSWDrawString(pixmap, x + 3, y + (10 / 2) + 3,
		text  
	    );

	    /* Memory line 2 */
	    y += 12;
	    sprintf(text,
		"           Total: %ld kb",
		d->mem_total / 1000
	    );
	    OSWDrawString(pixmap, x + 3, y + (10 / 2) + 3,
		text
	    );

	    /* Universe memory line 1 */
	    y += (12 * 2);
	    sprintf(text,
		"Universes:"
	    );
	    OSWDrawString(pixmap, x + 3, y + (10 / 2) + 3,
		text
	    );

	    /* Universe memory line 2 */
	    y += 12;
	    for(i = 0, n = 0; i < total_uews; i++)
	    {
		if(ue_uew[i] == NULL)
		    continue;

		n++;
	    }
	    sprintf(text,
		"    Loaded: %i",
		n
	    );
	    OSWDrawString(pixmap, x + 3, y + (10 / 2) + 3,
		text
	    );

	    /* Universe memory line 3 */
	    y += 12;
	    sprintf(text,
		"    Memory: %ld kb",
		d->mem_universe / 1000
	    );
	    OSWDrawString(pixmap, x + 3, y + (10 / 2) + 3,
		text
	    );


	    /* Copyright line */
	    img = ue_image.unvedit_logo;
	    if(img != NULL)
	    {
		x = 10;
		y = 10 + (int)img->height + 10;
	    }
	    else
	    {
		y += 12;
	    }

	    sprintf(text,
		ABOUT_DLG_COPYRIGHT_MESG
	    );
	    OSWDrawString(pixmap,
		x + 3 + ((int)d->width / 2) -
		    ((strlen(text) * 6) / 2),
		y + (10 / 2) + 3,
		text
	    );

	    y += 12;
	    sprintf(text,
		ABOUT_DLG_URL_MESG
	    );
	    OSWDrawString(pixmap,
		x + 3 + ((int)d->width / 2) -
		    ((strlen(text) * 6) / 2), 
		y + (10 / 2) + 3,
		text
	    );

	    y += 12;   
	    sprintf(text,
		ABOUT_DLG_LICENSE_MESG
	    );
	    OSWDrawString(pixmap,
		x + 3 + ((int)d->width / 2) -
		    ((strlen(text) * 6) / 2),
		y + (10 / 2) + 3,
		text
	    );
#endif
	}
}

/*
 *	Manages the about dialog event.
 */
int AboutDlgManage(about_dlg_struct *d, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	event_key_t *key;

	if((d == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case Expose:
	    if(w == d->toplevel)
	    {
		AboutDlgDraw(d, ABOUT_DLG_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!d->has_focus)
		return(events_handled);

	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!d->has_focus)
		return(events_handled);

	    break;

	  case FocusIn:
	    if(w == d->toplevel)
	    {
		d->has_focus = True;
		events_handled++;
	    }
	    break;
			
	  case FocusOut:
	    if(w == d->toplevel)
	    {
		d->has_focus = False;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(d->toplevel, event, &t))
	    {
		OSWGrabFocusTime(d->toplevel, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(d->toplevel, event))
	    {
		AboutDlgUnmap(d);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += PBtnManage(&d->ok_btn, event);

	return(events_handled);
}

/*
 *	Maps the about dialog.
 */
void AboutDlgMap(about_dlg_struct *d, win_t toplevel)
{
	if(d == NULL)
	    return;

	if(toplevel != None)
	{
            WidgetCenterWindowToWindow(d->toplevel, toplevel);
	    OSWSetTransientFor(d->toplevel, toplevel);
	}

	/* Map */
	OSWMapRaised(d->toplevel);
	d->map_state = True;

	PBtnMap(&d->ok_btn);

	OSWSetModal(d->toplevel);
	OSWGrabFocus(d->toplevel);
}

/*
 *	Unmaps the about dialog.
 */
void AboutDlgUnmap(about_dlg_struct *d)
{
	if(d == NULL)
	    return;

	WidgetUnmap(&d->ok_btn);

	OSWUnmapWindow(d->toplevel);
	d->map_state = False;
}

/*
 *	Deletes the about dialog.
 */
void AboutDlgDelete(about_dlg_struct *d)
{
	if(d == NULL)
	    return;

	WidgetDestroy(&d->ok_btn);
	OSWDestroyWindow(&d->toplevel);

	free(d);
}
