#include "blitting.h"


static void BlitBufBeamPixelAdditive8(void *p, const WColorStruct *c);
static void BlitBufBeamPixelAdditive15(void *p, const WColorStruct *c);
static void BlitBufBeamPixelAdditive16(void *p, const WColorStruct *c);
static void BlitBufBeamPixelAdditive24(void *p, const WColorStruct *c);
static void BlitBufBeamPixelAdditive32(void *p, const WColorStruct *c);

static void BlitBufBeamBeam(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,
	unsigned int len,
	unsigned int broadness,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const WColorStruct *)
);
void BlitBufBeam(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,			/* Bearing in radians */
	unsigned int len,		/* Length in pixels */
	unsigned int broadness,		/* Thickness in pixels */
	const WColorStruct *c
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

#define FTOI_RNDOFF(x)  (int)((x) + 0.5f)


static void BlitBufBeamPixelAdditive8(void *p, const WColorStruct *c)
{
	const u_int16_t tv = *(u_int8_t *)p;

	*(u_int8_t *)p = PACK8TO8(
	    (u_int8_t)MIN(
(u_int16_t)(tv & 0xE0) + (u_int16_t)c->r,
		0x00FF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x1C) << 3) + (u_int16_t)c->g,
		0x00FF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x03) << 6) + (u_int16_t)c->b,
		0x00FF
	    )
	);   
}

static void BlitBufBeamPixelAdditive15(void *p, const WColorStruct *c)
{
	const u_int16_t tv = *(u_int16_t *)p;

	*(u_int16_t *)p = PACK8TO15(
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x7C00) >> 7) + (u_int16_t)c->r,
		0xFF
	    ),      
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x03E0) >> 2) + (u_int16_t)c->g,
		0xFF
	    ),      
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x001F) << 3) + (u_int16_t)c->b,
		0xFF
	    )
	);   
}

static void BlitBufBeamPixelAdditive16(void *p, const WColorStruct *c)
{
	const u_int16_t tv = *(u_int16_t *)p;

	*(u_int16_t *)p = PACK8TO16(
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0xF800) >> 8) + (u_int16_t)c->r,
		0xFF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x07E0) >> 3) + (u_int16_t)c->g,
		0xFF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x001F) << 3) + (u_int16_t)c->b,
		0xFF
	    )
	);
}

static void BlitBufBeamPixelAdditive24(void *p, const WColorStruct *c)
{
	u_int8_t *p8 = (u_int8_t *)p;

	p8[2] = (u_int8_t)MIN(
	    (u_int16_t)p8[2] + (u_int16_t)c->r,      
	    0x00FF
	);
	p8[1] = (u_int8_t)MIN(
	    (u_int16_t)p8[1] + (u_int16_t)c->g,
	    0x00FF
	);        
	p8[0] = (u_int8_t)MIN(
	    (u_int16_t)p8[0] + (u_int16_t)c->b,
	    0x00FF
	);
}

static void BlitBufBeamPixelAdditive32(void *p, const WColorStruct *c)
{
	u_int8_t *p8 = (u_int8_t *)p;

	p8[2] = (u_int8_t)MIN(
	    (u_int16_t)p8[2] + (u_int16_t)c->r,
	    0x00FF
	);
	p8[1] = (u_int8_t)MIN(
	    (u_int16_t)p8[1] + (u_int16_t)c->g,
	    0x00FF
	);   
	p8[0] = (u_int8_t)MIN(
	    (u_int16_t)p8[0] + (u_int16_t)c->b,
	    0x00FF
	);
}


static void BlitBufBeamBeam(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,
	unsigned int len,
	unsigned int broadness,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const WColorStruct *)
)
{
	char quadrant;
	int	cx, cy,
		cx_end, cy_end,
		dx, dy,		/* Delta change in position per iteration */
		ray_cnt,
		ray_max;	/* Number of rays */
	float coeff, sin_theta, cos_theta;
	WColorStruct c2;

	if((tar_buf == NULL) || (tar_width == 0) || (tar_height == 0))
	    return;

	/* Entirely off the target buffer? */
	if(((tar_x + (int)len + (int)broadness) <= 0) ||
	   ((tar_y + (int)len + (int)broadness) <= 0) ||
	   ((tar_x - (int)len - (int)broadness) >= (int)tar_width) ||
	   ((tar_y - (int)len - (int)broadness) >= (int)tar_height) ||
	   (len == 0)
	)
	    return;

	/* Wrap theta to be between 0 and PI */
	while(theta >= 6.283185307)
	    theta -= 6.283185307;
	while(theta < 0)
	    theta += 6.283185307;

	/* Calculate the sin and cos values of theta in bearing */
	sin_theta = (float)sin((double)((PI / 2.0f) - theta));
	cos_theta = (float)cos((double)((PI / 2.0f) - theta));

	/* Convert all cartesian measures to * 256. The delta's
	 * dx and dy will have the proper signs for the flipped y
	 * coordinates
	 */
	tar_x <<= 8;
	tar_y <<= 8;
	tar_width <<= 8;
	tar_height <<= 8;
	dx = (int)(256 * cos_theta);
	dy = (int)(-256 * sin_theta);	/* Y coords are flipped */

	if(theta < (PI * 0.5f))
	    quadrant = 1;
	else if(theta < (PI * 1.0f))
	    quadrant = 2;
	else if(theta < (PI * 1.5f))
	    quadrant = 3;
	else
	    quadrant = 4;


	for(ray_cnt = (int)broadness / -2, ray_max = (int)broadness / 2;
	    ray_cnt < ray_max;
	    ray_cnt++
	)
	{
	    /* Calculate additive color values */
	    if(ray_max > 0)
	    {
	        if(ray_cnt < 0)
		    coeff = 1 + ((float)ray_cnt / (float)ray_max);
	        else
		    coeff = 1 - ((float)ray_cnt / (float)ray_max);
	    }
	    else
		coeff = 1;
/*	    c2.a = 0xff;	*/
	    c2.r = (u_int8_t)(c->r * coeff);
	    c2.g = (u_int8_t)(c->g * coeff);
	    c2.b = (u_int8_t)(c->b * coeff);

	    /* Calculate the current ray starting and ending positions */
	    cx = FTOI_RNDOFF(tar_x + (ray_cnt * sin_theta * 256));
	    cy = FTOI_RNDOFF(tar_y + (ray_cnt * cos_theta * 256));

	    cx_end = FTOI_RNDOFF(cx + ((float)len * cos_theta * 256));
	    cy_end = FTOI_RNDOFF(cy - ((float)len * sin_theta * 256));

	    /* Blit current line by quadrant */
	    if(quadrant == 1)		/* Quadrant 1 */
	    {
		/* Blit each point on the line */
		while((cx <= cx_end) &&
		      (cy >= cy_end)		/* Y coords are flipped */
		)
		{
		    /* Make sure point is in buffer */
		    if((cx < 0) ||
		       (cy < 0) ||
		       (cx >= (int)tar_width) ||
		       (cy >= (int)tar_height)
		    )
		    {
			cx += dx;
			cy += dy;	/* Remember Y coords are flipped */
			continue;
		    }

		    pixel_op_func(
			tar_buf +
			    ((cy >> 8) * (int)tar_bpl) +
			    ((cx >> 8) * (int)bpp),
			&c2
		    );

		    cx += dx;
		    cy += dy;	/* Remember Y coords are flipped */
		}
	    }
	    else if(quadrant == 2)	/* Quadrant 2 */
	    {
		/* Blit each point on the line */
		while((cx <= cx_end) &&
		      (cy <= cy_end)	/* Remember Y coords are flipped */
		)
		{
		    /* Make sure point is in buffer */
		    if((cx < 0) ||
		       (cy < 0) ||
		       (cx >= (int)tar_width) ||
		       (cy >= (int)tar_height)
		    ) 
		    {
			cx += dx;
			cy += dy;	/* Remember Y coords are flipped */
			continue;
		    }

		    pixel_op_func(
			tar_buf +
			    ((cy >> 8) * (int)tar_bpl) +
			    ((cx >> 8) * (int)bpp),
			&c2
		    );

		    cx += dx;
		    cy += dy;	/* Remember Y coords are flipped */
		}
	    }
	    else if(quadrant == 3)	/* Quadrant 3 */
	    {
		/* Blit each point on the line */
		while((cx >= cx_end) &&
		      (cy <= cy_end)	/* Remember Y coords are flipped */
		)
		{
		    /* Make sure point is in buffer */
		    if((cx < 0) ||
		       (cy < 0) ||
		       (cx >= (int)tar_width) ||
		       (cy >= (int)tar_height)
		    )
		    {
			cx += dx;
			cy += dy;	/* Y coords are flipped */
			continue;
		    }

		    pixel_op_func(
			tar_buf +
			    ((cy >> 8) * (int)tar_bpl) +
			    ((cx >> 8) * (int)bpp),
			&c2
		    );

		    cx += dx;
		    cy += dy;		/* Y coords are flipped */
		}
	    }
	    else if(quadrant == 4)	/* Quadrant 4 */
	    {
		/* Blit each point on the line */
		while((cx >= cx_end) &&
		      (cy >= cy_end)	/* Remember Y coords are flipped */
		)
		{
		    /* Make sure point is in buffer */
		    if((cx < 0) ||
		       (cy < 0) ||
		       (cx >= (int)tar_width) ||
		       (cy >= (int)tar_height)
		    ) 
		    {
			cx += dx;
			cy += dy;	/* Y coords are flipped */
			continue;
		    }

		    pixel_op_func(
			tar_buf +
			    ((cy >> 8) * (int)tar_bpl) +
			    ((cx >> 8) * (int)bpp),
			&c2
		    );

		    cx += dx;
		    cy += dy;		/* Y coords are flipped */
		}
	    }
	}
}

void BlitBufBeam(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,			/* Bearing in radians */
	unsigned int len,		/* Length in pixels */
	unsigned int broadness,		/* Thickness in pixels */
	const WColorStruct *c
)
{
	WColorStruct def_c;
	void (*pixel_op_func)(void *, const WColorStruct *);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

	if(c == NULL)
	{
	    def_c.a = 0x00;
	    def_c.r = 0x00;
	    def_c.g = 0x00;
	    def_c.b = 0x00;
	    c = &def_c;
	}

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufBeamPixelAdditive8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufBeamPixelAdditive15;
		break;
	      case 16:
		pixel_op_func = BlitBufBeamPixelAdditive16;
		break;   
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    pixel_op_func = BlitBufBeamPixelAdditive24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:
		pixel_op_func = BlitBufBeamPixelAdditive24;
		break;
	      case 32:
		pixel_op_func = BlitBufBeamPixelAdditive32;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufBeamBeam(
	    bpp,
	    tar_buf,
	    tar_x, tar_y,
	    tar_width, tar_height, tar_bpl,
	    theta, len, broadness, c, pixel_op_func
	);
}
