#include "blitting.h"


static int BlitBufIsTransparentPixel8(const void *p);
static int BlitBufIsTransparentPixel15(const void *p);
static int BlitBufIsTransparentPixel16(const void *p);
static int BlitBufIsTransparentPixel24(const void *p);
static int BlitBufIsTransparentPixel32(const void *p);

static void BlitBufCircleCursorPixel8(
	void *tp, const void *sp, const WColorStruct *c
);
static void BlitBufCircleCursorPixel15(
	void *tp, const void *sp, const WColorStruct *c
);
static void BlitBufCircleCursorPixel16(
	void *tp, const void *sp, const WColorStruct *c
);
static void BlitBufCircleCursorPixel24(
	void *tp, const void *sp, const WColorStruct *c
);
static void BlitBufCircleCursorPixel32(
	void *tp, const void *sp, const WColorStruct *c
);

static void BlitBufCursorCursor(
	unsigned int bpp,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const void *, const WColorStruct *c),
	int (*is_transparent_func)(const void *)
);

void BlitBufCursor(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	const WColorStruct *c
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the pixel is transparent (all defined bits are 0).
 */
static int BlitBufIsTransparentPixel8(const void *p)
{
	return(!(*(const u_int8_t *)p & 0xFF));
}

static int BlitBufIsTransparentPixel15(const void *p)
{
	return(!(*(const u_int16_t *)p & 0xEFFF));
}

static int BlitBufIsTransparentPixel16(const void *p) 
{
	return(!(*(const u_int16_t *)p & 0xFFFF));
}

static int BlitBufIsTransparentPixel24(const void *p)
{
	const u_int8_t *p8 = (const u_int8_t *)p;
	return((p8[2] == 0x00) &&
	       (p8[1] == 0x00) &&
	       (p8[0] == 0x00)
	);
}

static int BlitBufIsTransparentPixel32(const void *p)
{
	return(!(*(const u_int32_t *)p & 0x00FFFFFF));
}


/*
 *	Colors the source pixel and then averages it to the target
 *	pixel.
 */
static void BlitBufCircleCursorPixel8(
	void *tp, const void *sp, const WColorStruct *c
)
{
	const u_int8_t	tv = *(u_int8_t *)tp,
			sv = *(u_int8_t *)sp;

	 *(u_int8_t *)tp = PACK8TO8(
	    ((
((((u_int16_t)(sv & 0xE0) >> 5) * (u_int16_t)c->r) >> 8) +
  (u_int16_t)((tv & 0xE0) >> 5)
	    ) << 3) / 2,
	    ((
((((u_int16_t)(sv & 0x1C) >> 2) * (u_int16_t)c->g) >> 8) +
  (u_int16_t)((tv & 0x1C) >> 2)
	    ) << 2) / 2,
	    ((
(((u_int16_t)(sv & 0x03) * (u_int16_t)c->b) >> 8) +
  (u_int16_t)(tv & 0x03)
	    ) << 2) / 2
	);
}

static void BlitBufCircleCursorPixel15(
	void *tp, const void *sp, const WColorStruct *c
)
{
	const u_int16_t tv = *(u_int16_t *)tp,
			sv = *(u_int16_t *)sp;

	*(u_int16_t *)tp = PACK8TO15(
	    (u_int8_t)(
((((u_int16_t)((sv & 0x7C00) >> 7) * (u_int16_t)c->r) >> 8) +
(u_int16_t)((tv & 0x7C00) >> 7)) / 2
	    ),
	    (u_int8_t)(
((((u_int16_t)((sv & 0x03E0) >> 2) * (u_int16_t)c->g) >> 8) +
(u_int16_t)((tv & 0x03E0) >> 2)) / 2
	    ),
	    (u_int8_t)(
((((u_int16_t)((sv & 0x001F) << 3) * (u_int16_t)c->b) >> 8) +
(u_int16_t)((tv & 0x001F) << 3)) / 2
	    )
	);   
}

static void BlitBufCircleCursorPixel16(
	void *tp, const void *sp, const WColorStruct *c
)
{
	const u_int16_t	tv = *(u_int16_t *)tp,
			sv = *(u_int16_t *)sp;

	*(u_int16_t *)tp = PACK8TO16(
		(u_int8_t)(
((((u_int16_t)((sv & 0xF800) >> 8) * (u_int16_t)c->r) >> 8) +
(u_int16_t)((tv & 0xF800) >> 8)) / 2
		),
		(u_int8_t)(  
((((u_int16_t)((sv & 0x07E0) >> 3) * (u_int16_t)c->g) >> 8) +
(u_int16_t)((tv & 0x07E0) >> 3)) / 2
		),
		(u_int8_t)(
((((u_int16_t)((sv & 0x001F) << 3) * (u_int16_t)c->b) >> 8) +
(u_int16_t)((tv & 0x001F) << 3)) / 2
		)
	    );
}

static void BlitBufCircleCursorPixel24(
	void *tp, const void *sp, const WColorStruct *c
)
{
	u_int8_t *tp8 = (u_int8_t *)tp;
	const u_int8_t *sp8 = (const u_int8_t *)sp;

	tp8[2] = (u_int8_t)((
	    (((u_int16_t)sp8[2] * (u_int16_t)c->r) >> 8) +
	    (u_int16_t)tp8[2]
	) / 2);
	tp8[1] = (u_int8_t)((
	    (((u_int16_t)sp8[1] * (u_int16_t)c->g) >> 8) +
	    (u_int16_t)tp8[1]
	) / 2);
	tp8[0] = (u_int8_t)((
	    (((u_int16_t)sp8[0] * (u_int16_t)c->b) >> 8) +
	    (u_int16_t)tp8[0]
	) / 2);
}

static void BlitBufCircleCursorPixel32(
	void *tp, const void *sp, const WColorStruct *c
)
{
	u_int8_t *tp8 = (u_int8_t *)tp;
	const u_int8_t *sp8 = (const u_int8_t *)sp;

	tp8[2] = (u_int8_t)((
	    (((u_int16_t)sp8[2] * (u_int16_t)c->r) >> 8) +
	    (u_int16_t)tp8[2]
	) / 2);
	tp8[1] = (u_int8_t)((
	    (((u_int16_t)sp8[1] * (u_int16_t)c->g) >> 8) +
	    (u_int16_t)tp8[1]
	) / 2);
	tp8[0] = (u_int8_t)((
	    (((u_int16_t)sp8[0] * (u_int16_t)c->b) >> 8) +
	    (u_int16_t)tp8[0]
	) / 2);
}


static void BlitBufCursorCursor(
	unsigned int bpp,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const void *, const WColorStruct *c),
	int (*is_transparent_func)(const void *)
)
{
	int tar_x_start, src_x, src_y;
	const u_int8_t *src_ptr;

	if((tar_buf == NULL) || (src_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (src_width == 0) || (src_height == 0)
	)
	    return;

	/* Check if the source coordinates fall completely off
	 * the target buffer
	 */
	if(((tar_x + (int)src_width) <= 0) ||
	   ((tar_y + (int)src_height) <= 0) ||
	   (tar_x >= (int)tar_width) ||
	   (tar_y >= (int)tar_height)
	)
	    return;

	tar_x_start = tar_x;

	/* Calculate starting values */
	src_x = 0;
	src_y = 0;

	/* Blit the source buffer to the target buffer */
	while(src_y < (int)src_height)
	{
	    /* Skip if the target coordinate is off the buffer */
	    if((tar_x < 0) ||
	       (tar_y < 0) ||
	       (tar_x >= (int)tar_width) ||
	       (tar_y >= (int)tar_height)
	    )
	    {
		/* Increment x positions */
		src_x++;
		tar_x++;

		/* Go to the next line? */
		if(src_x >= (int)src_width)
		{
		    src_x = 0;
		    src_y++;
		    tar_x = tar_x_start;
		    tar_y++;
		}
		continue;
	    }

	    /* Get the pointer to the source pixel */
	    src_ptr = src_buf +
		(src_y * (int)src_bpl) + (src_x * (int)bpp);

	    /* Transparent? */
	    if(is_transparent_func(src_ptr))
	    {
		/* Increment x positions */
		src_x++;
		tar_x++;
 
		/* Go to the next line? */
		if(src_x >= (int)src_width)
		{
		    src_x = 0;
		    src_y++;
		    tar_x = tar_x_start;
		    tar_y++;
		}   
		continue;
	    }

	    pixel_op_func(
		tar_buf +
		    (tar_y * (int)tar_bpl) + (tar_x * (int)bpp),
		src_ptr,
		c
	    );

	    /* Increment x positions */
	    src_x++;
	    tar_x++;

	    /* Go to the next line? */
	    if(src_x >= (int)src_width)
	    {
		src_x = 0;
		src_y++;
		tar_x = tar_x_start;
		tar_y++;
		continue;
	    }
	}
}


void BlitBufCursor(
	unsigned int bpp, unsigned int depth, 
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	const WColorStruct *c
)
{
	int (*is_transparent_func)(const void *);
	void (*pixel_op_func)(void *, const void *, const WColorStruct *c);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;
	if(src_bpl == 0)
	    src_bpl = src_width * bpp;

	switch(bpp)   
	{
	  case 1:
	    is_transparent_func = BlitBufIsTransparentPixel8;
	    pixel_op_func = BlitBufCircleCursorPixel8;
	    break;  
	  case 2:
	    switch(depth)
	    {
	      case 15:
		is_transparent_func = BlitBufIsTransparentPixel15;
		pixel_op_func = BlitBufCircleCursorPixel15;
		break;
	      case 16:
		is_transparent_func = BlitBufIsTransparentPixel16;
		pixel_op_func = BlitBufCircleCursorPixel16;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    is_transparent_func = BlitBufIsTransparentPixel24;
	    pixel_op_func = BlitBufCircleCursorPixel24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:
		is_transparent_func = BlitBufIsTransparentPixel24;
		pixel_op_func = BlitBufCircleCursorPixel24;
		break;
	      case 32:
		is_transparent_func = BlitBufIsTransparentPixel32;
		pixel_op_func = BlitBufCircleCursorPixel32;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufCursorCursor(
	    bpp,
	    tar_buf, src_buf,
	    tar_x, tar_y,
	    tar_width, tar_height, tar_bpl,
	    src_width, src_height, src_bpl,
	    c, pixel_op_func, is_transparent_func
	);
}
