#include "blitting.h"


static void BlitBufLinePixel8(void *p, const u_int32_t v);
static void BlitBufLinePixel15(void *p, const u_int32_t v);
static void BlitBufLinePixel16(void *p, const u_int32_t v);
static void BlitBufLinePixel24(void *p, const u_int32_t v);
static void BlitBufLinePixel32(void *p, const u_int32_t v);

static void BlitBufLinePixelAdditive8(void *p, const WColorStruct *c);
static void BlitBufLinePixelAdditive15(void *p, const WColorStruct *c);
static void BlitBufLinePixelAdditive16(void *p, const WColorStruct *c);
static void BlitBufLinePixelAdditive24(void *p, const WColorStruct *c);
static void BlitBufLinePixelAdditive32(void *p, const WColorStruct *c);

static void BlitBufLineLine(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,
	unsigned int len,
	unsigned int broadness,
	const u_int32_t v,
	void (*pixel_op_func)(void *, const u_int32_t)
);
void BlitBufLine(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,			/* Bearing in radians */
	unsigned int len,		/* Length in pixels */
	unsigned int broadness,		/* Thickness in pixels */
	const WColorStruct *c
);

static void BlitBufLineLineAdditive(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,
	unsigned int len,
	unsigned int broadness,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const WColorStruct *)
);
void BlitBufLineAdditive(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,			/* Bearing in radians */
	unsigned int len,		/* Length in pixels */
	unsigned int broadness,		/* Thickness in pixels */
	const WColorStruct *c
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

#define FTOI_RNDOFF(x)	(int)((x) + 0.5f)


static void BlitBufLinePixel8(void *p, const u_int32_t v)
{
	*(u_int8_t *)p = (u_int8_t)v;
}

static void BlitBufLinePixel15(void *p, const u_int32_t v)
{
	*(u_int16_t *)p = (u_int16_t)v;
}

static void BlitBufLinePixel16(void *p, const u_int32_t v)
{
	*(u_int16_t *)p = (u_int16_t)v;
}

static void BlitBufLinePixel24(void *p, const u_int32_t v)
{
	u_int8_t *p8 = (u_int8_t *)p;
	p8[0] = (u_int8_t)((v & 0x000000FF) >> 0);
	p8[1] = (u_int8_t)((v & 0x0000FF00) >> 8);
	p8[2] = (u_int8_t)((v & 0x00FF0000) >> 16);
}
 
static void BlitBufLinePixel32(void *p, const u_int32_t v)
{
	*(u_int32_t *)p = v;
}


static void BlitBufLinePixelAdditive8(void *p, const WColorStruct *c)
{
	const u_int16_t tv = *(u_int8_t *)p;

	*(u_int8_t *)p = PACK8TO8(
	    (u_int8_t)MIN(
(u_int16_t)(tv & 0xE0) + (u_int16_t)c->r,
		0x00FF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x1C) << 3) + (u_int16_t)c->g,
		0x00FF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x03) << 6) + (u_int16_t)c->b,
		0x00FF
	    )
	);
}

static void BlitBufLinePixelAdditive15(void *p, const WColorStruct *c)
{
	const u_int16_t tv = *(u_int16_t *)p;

	*(u_int16_t *)p = PACK8TO15(
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x7C00) >> 7) + (u_int16_t)c->r,
		0xFF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x03E0) >> 2) + (u_int16_t)c->g,
		0xFF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x001F) << 3) + (u_int16_t)c->b,
		0xFF
	    )
	);
}

static void BlitBufLinePixelAdditive16(void *p, const WColorStruct *c)
{
	const u_int16_t tv = *(u_int16_t *)p;

	*(u_int16_t *)p = PACK8TO16(
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0xF800) >> 8) + (u_int16_t)c->r,
		0xFF
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x07E0) >> 3) + (u_int16_t)c->g, 
		0xFF  
	    ),
	    (u_int8_t)MIN(
(u_int16_t)((tv & 0x001F) << 3) + (u_int16_t)c->b, 
		0xFF 
	    )
	);
}

static void BlitBufLinePixelAdditive24(void *p, const WColorStruct *c)
{
	u_int8_t *p8 = (u_int8_t *)p;

	p8[2] = (u_int8_t)MIN(
	    (u_int16_t)p8[2] + (u_int16_t)c->r,
	    0x00FF
	);
	p8[1] = (u_int8_t)MIN(
	    (u_int16_t)p8[1] + (u_int16_t)c->g,
	    0x00FF
	);
	p8[0] = (u_int8_t)MIN(
	    (u_int16_t)p8[0] + (u_int16_t)c->b,
	    0x00FF
	);
}

static void BlitBufLinePixelAdditive32(void *p, const WColorStruct *c)
{
	u_int8_t *p8 = (u_int8_t *)p;

	p8[2] = (u_int8_t)MIN(
	    (u_int16_t)p8[2] + (u_int16_t)c->r,
	    0x00FF
	);
	p8[1] = (u_int8_t)MIN(
	    (u_int16_t)p8[1] + (u_int16_t)c->g,
	    0x00FF
	);
	p8[0] = (u_int8_t)MIN(
	    (u_int16_t)p8[0] + (u_int16_t)c->b,
	    0x00FF
	);
}


static void BlitBufLineLine(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,
	unsigned int len,
	unsigned int broadness,
	const u_int32_t v,
	void (*pixel_op_func)(void *, const u_int32_t)
)
{
	char quadrant;
	int	x2, y2,
		x_quad_coeff, y_quad_coeff,
		bx, by;
	float	x, y,
		dx, dy,
		yoverx, xovery,		/* y / x and x / y */
		xcount, ycount;

	if((tar_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0)
	)
	    return;
 
	/* Totally not visable on target? */
	if(((tar_x + (int)len + (int)broadness) <= 0) ||
	   ((tar_y + (int)len + (int)broadness) <= 0) ||
	   ((tar_x - (int)len + (int)broadness) >= (int)tar_width) ||
	   ((tar_y - (int)len + (int)broadness) >= (int)tar_height)
	)
	    return;

	/* Offset the center based on the broadness */
	tar_x = FTOI_RNDOFF(tar_x - ((float)broadness / 2.0f));
	tar_y = FTOI_RNDOFF(tar_y - ((float)broadness / 2.0f));

	/* Wrap theta between 0 and 2 PI */
	while(theta >= 6.283185307)
	    theta -= 6.283185307;
	while(theta < 0)
	    theta += 6.283185307;

	/* Length may not be too long */
	if(len > 10000)
	    len = 10000;

	if(broadness > 256)
	    broadness = 256;

	if(broadness == 0)
	    return;

	/* Calculate the delta change of position per iteration */
	dx = (float)sin((double)theta) * (float)len;
	dy = (float)cos((double)theta) * (float)len * -1;

	/* Calculate the quadrant coefficient (-1 or 1) */
	x_quad_coeff = (dx < 0) ? -1 : 1;
	y_quad_coeff = (dy < 0) ? -1 : 1;

	/* Determine the quadrant (clockwise from upper right) */
	if((x_quad_coeff >= 0) && (y_quad_coeff < 0))
	    quadrant = 0;
	else if((x_quad_coeff >= 0) && (y_quad_coeff >= 0))
	    quadrant = 1;
	else if((x_quad_coeff < 0) && (y_quad_coeff >= 0))
	    quadrant = 2;
	else
	    quadrant = 3;

	/* Make dy and dx positive */
	dx *= (float)x_quad_coeff;
	dy *= (float)y_quad_coeff;

	/* Calculate yoverx and xovery */
	yoverx = (dx != 0.0f) ? MIN((dy / dx), 1.0f) : 1.0f;
	xovery = (dy != 0.0f) ? MIN((dx / dy), 1.0f) : 1.0f;

	/* Either yoverx or xovery must have some value */
	if((yoverx <= 0.0f) && (xovery <= 0.0f))
	    yoverx = 0.01f;	/* Add some value to it */

	xcount = 0.0f;
	ycount = 0.0f;
	while((xcount <= dx) && (ycount <= dy))
	{
	    x = (xcount * (float)x_quad_coeff) + (float)tar_x;
	    y = (ycount * (float)y_quad_coeff) + (float)tar_y;
 
	    /* Draw by quadrant */
	    switch(quadrant)
	    {
	      case 0:   /* Upper right */
		/* Top horizontal */
		by = 0;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}  
		/* Right vertical */
		bx = broadness - 1;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
 		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )  
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}
		break;

	      case 1:	/* Lower right */
		/* Lower horizontal */
		by = broadness - 1;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}
		/* Right vertical */
		bx = broadness - 1;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}
		break;
			
	      case 2:   /* Lower left */
		/* Lower horizontal */
		by = broadness - 1;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}
		/* Left vertical */
		bx = 0;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}
		break;

	      default:	/* Upper left */
		/* Upper horizontal */
		by = 0;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}
		/* Left vertical */
		bx = 0;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    v
			);
		    }
		}
		break;
	    }

	    xcount += xovery;
	    ycount += yoverx;
	}
}

void BlitBufLine(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,			/* Bearing in radians */
	unsigned int len,		/* Length in pixels */
	unsigned int broadness,		/* Thickness in pixels */
	const WColorStruct *c
)
{
	u_int32_t v;
	WColorStruct def_c;
	void (*pixel_op_func)(void *, const u_int32_t);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

	if(c == NULL) 
	{
	    def_c.a = 0x00;
	    def_c.r = 0x00;
	    def_c.g = 0x00;
	    def_c.b = 0x00;
	    c = &def_c;
	}

	switch(depth)
	{
	  case 8:
	    v = PACK8TO8(c->r, c->g, c->b);
	    break;
	  case 15:
	    v = PACK8TO15(c->r, c->g, c->b);
	    break;
	  case 16:
	    v = PACK8TO16(c->r, c->g, c->b);
	    break;
	  case 24:
	  case 32:
	    v = PACK8TO32(c->a, c->r, c->g, c->b);
	    break;
	  default:
	    v = 0x00000000;
	    break;
	}

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufLinePixel8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufLinePixel15;
		break;
	      case 16:
		pixel_op_func = BlitBufLinePixel16;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    pixel_op_func = BlitBufLinePixel24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:
		pixel_op_func = BlitBufLinePixel24;
		break;
	      case 32:
		pixel_op_func = BlitBufLinePixel32;
		break;
	      default:
		return; 
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufLineLine(
	    bpp,
	    tar_buf,
	    tar_x, tar_y,
	    tar_width, tar_height, tar_bpl,
	    theta, len, broadness, v, pixel_op_func
	);
}


static void BlitBufLineLineAdditive(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,
	unsigned int len,
	unsigned int broadness,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const WColorStruct *)
)
{
	char quadrant;
	int	x2, y2,
		x_quad_coeff, y_quad_coeff,
		bx, by;
	float	x, y,
		dx, dy,
		yoverx, xovery,		/* y / x and x / y */
		xcount, ycount;

	if((tar_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0)
	)
	    return;
 
	/* Totally not visable on target? */
	if(((tar_x + (int)len + (int)broadness) <= 0) ||
	   ((tar_y + (int)len + (int)broadness) <= 0) ||
	   ((tar_x - (int)len + (int)broadness) >= (int)tar_width) ||
	   ((tar_y - (int)len + (int)broadness) >= (int)tar_height)
	)
	    return;

	/* Offset the center based on the broadness */
	tar_x = FTOI_RNDOFF(tar_x - ((float)broadness / 2.0f));
	tar_y = FTOI_RNDOFF(tar_y - ((float)broadness / 2.0f));

	/* Wrap theta between 0 and 2 PI */
	while(theta >= 6.283185307)
	    theta -= 6.283185307;
	while(theta < 0)
	    theta += 6.283185307;

	/* Length may not be too long */
	if(len > 10000)
	    len = 10000;

	if(broadness > 256)
	    broadness = 256;

	if(broadness == 0)
	    return;

	/* Calculate the delta change of position per iteration */
	dx = (float)sin((double)theta) * (float)len;
	dy = (float)cos((double)theta) * (float)len * -1;

	/* Calculate the quadrant coefficient (-1 or 1) */
	x_quad_coeff = (dx < 0) ? -1 : 1;
	y_quad_coeff = (dy < 0) ? -1 : 1;

	/* Determine the quadrant (clockwise from upper right) */
	if((x_quad_coeff >= 0) && (y_quad_coeff < 0))
	    quadrant = 0;
	else if((x_quad_coeff >= 0) && (y_quad_coeff >= 0))
	    quadrant = 1;
	else if((x_quad_coeff < 0) && (y_quad_coeff >= 0))
	    quadrant = 2;
	else
	    quadrant = 3;

	/* Make dy and dx positive */
	dx *= (float)x_quad_coeff;
	dy *= (float)y_quad_coeff;

	/* Calculate yoverx and xovery */
	yoverx = (dx != 0.0f) ? MIN((dy / dx), 1.0f) : 1.0f;
	xovery = (dy != 0.0f) ? MIN((dx / dy), 1.0f) : 1.0f;

	/* Either yoverx or xovery must have some value */
	if((yoverx <= 0.0f) && (xovery <= 0.0f))
	    yoverx = 0.01f;	/* Add some value to it */

	xcount = 0.0f;
	ycount = 0.0f;
	while((xcount <= dx) && (ycount <= dy))
	{
	    x = (xcount * (float)x_quad_coeff) + (float)tar_x;
	    y = (ycount * (float)y_quad_coeff) + (float)tar_y;
 
	    /* Draw by quadrant */
	    switch(quadrant)
	    {
	      case 0:   /* Upper right */
		/* Top horizontal */
		by = 0;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}  
		/* Right vertical */
		bx = broadness - 1;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
 		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )  
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}
		break;

	      case 1:	/* Lower right */
		/* Lower horizontal */
		by = broadness - 1;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}
		/* Right vertical */
		bx = broadness - 1;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}
		break;
			
	      case 2:   /* Lower left */
		/* Lower horizontal */
		by = broadness - 1;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}
		/* Left vertical */
		bx = 0;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}
		break;

	      default:	/* Upper left */
		/* Upper horizontal */
		by = 0;
		for(bx = 0; bx < (int)broadness; bx++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}
		/* Left vertical */
		bx = 0;
		for(by = 0; by < (int)broadness; by++)
		{
		    x2 = FTOI_RNDOFF(x + bx);
		    y2 = FTOI_RNDOFF(y + by);
		    if((x2 >= 0) &&
		       (y2 >= 0) &&
		       (x2 < (int)tar_width) &&
		       (y2 < (int)tar_height)
		    )
		    {
			pixel_op_func(
			    tar_buf +
				(y2 * (int)tar_bpl) +
				(x2 * (int)bpp),
			    c
			);
		    }
		}
		break;
	    }

	    xcount += xovery;
	    ycount += yoverx;
	}
}

void BlitBufLineAdditive(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float theta,			/* Bearing in radians */
	unsigned int len,		/* Length in pixels */
	unsigned int broadness,		/* Thickness in pixels */
	const WColorStruct *c
)
{
	WColorStruct def_c;
	void (*pixel_op_func)(void *, const WColorStruct *);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

	if(c == NULL)
	{
	    def_c.a = 0x00;
	    def_c.r = 0x00;
	    def_c.g = 0x00;
	    def_c.b = 0x00;
	    c = &def_c;
	}

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufLinePixelAdditive8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufLinePixelAdditive15;
		break;
	      case 16:
		pixel_op_func = BlitBufLinePixelAdditive16;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    pixel_op_func = BlitBufLinePixelAdditive24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:
		pixel_op_func = BlitBufLinePixelAdditive24;
		break;
	      case 32:
		pixel_op_func = BlitBufLinePixelAdditive32;
		break;
	      default:
		return; 
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufLineLineAdditive(
	    bpp,
	    tar_buf,
	    tar_x, tar_y,
	    tar_width, tar_height, tar_bpl,
	    theta, len, broadness, c, pixel_op_func
	);
}
