#include "blitting.h"


static void BlitBufPixelAdditive8(
	void *p, const u_int8_t v, const WColorStruct *c  
);
static void BlitBufPixelAdditive15(
	void *p, const u_int8_t v, const WColorStruct *c
);
static void BlitBufPixelAdditive16(
	void *p, const u_int8_t v, const WColorStruct *c  
);
static void BlitBufPixelAdditive24(
	void *p, const u_int8_t v, const WColorStruct *c
);
static void BlitBufPixelAdditive32(
	void *p, const u_int8_t v, const WColorStruct *c
);

static void BlitBufArea(
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpp, unsigned int tar_bpl,
	const u_int8_t *src_buf,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpp, unsigned int src_bpl,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const u_int8_t, const WColorStruct *)
);

void BlitBufPointLight(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float radius,
	const WColorStruct *c
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void BlitBufPixelAdditive8(
	void *p, const u_int8_t v, const WColorStruct *c
)
{
	/* rrrgggbb */ 
	u_int8_t *tar_pix8 = (u_int8_t *)p;

	const u_int16_t sr = (u_int16_t)(v * c->r / 0xFF),
			sg = (u_int16_t)(v * c->g / 0xFF),
			sb = (u_int16_t)(v * c->b / 0xFF),
			tv = (u_int16_t)(*tar_pix8);

	*tar_pix8 = PACK8TO8(
	    (u_int8_t)MIN(
		(((tv & 0xE0) << 0) + sr),
		0x00FF
	    ),
	    (u_int8_t)MIN(
		(((tv & 0x1C) << 3) + sg),
		0x00FF
	    ),
	    (u_int8_t)MIN(
		(((tv & 0x03) << 6) + sb),
		0x00FF
	    )
	);
}

static void BlitBufPixelAdditive15(
	void *p, const u_int8_t v, const WColorStruct *c
)
{
	/* arrrrrgg gggbbbbb */
	u_int16_t *tar_pix16 = (u_int16_t *)p;

	const u_int16_t sr = (u_int16_t)v * (u_int16_t)c->r / 0x00FF,
			sg = (u_int16_t)v * (u_int16_t)c->g / 0x00FF,
			sb = (u_int16_t)v * (u_int16_t)c->b / 0x00FF,
			tv = *tar_pix16;

	*tar_pix16 = PACK8TO15(
	    (u_int16_t)MIN(
		(((tv & 0x7C00) >> 7) + sr),
		0x00FF
	    ),
	    (u_int16_t)MIN(
		(((tv & 0x03E0) >> 2) + sg),
		0x00FF
	    ),
	    (u_int16_t)MIN(
		(((tv & 0x001F) << 3) + sb),
		0x00FF
	    )
	);
}

static void BlitBufPixelAdditive16(
	void *p, const u_int8_t v, const WColorStruct *c
)
{
	/* rrrrrggg gggbbbbb */
	u_int16_t *tar_pix16 = (u_int16_t *)p;

	const u_int16_t	sr = (u_int16_t)(v * c->r / 0xFF),
			sg = (u_int16_t)(v * c->g / 0xFF),
			sb = (u_int16_t)(v * c->b / 0xFF),
			tv = *tar_pix16;

	*tar_pix16 = PACK8TO16(
	    (u_int16_t)MIN(
		(((tv & 0xF800) >> 8) + sr),
		0x00FF
	    ),
	    (u_int16_t)MIN(
		(((tv & 0x07E0) >> 3) + sg),
		0x00FF
	    ),
	    (u_int16_t)MIN(
		(((tv & 0x001F) << 3) + sb),
		0x00FF
	    )
	);
}

static void BlitBufPixelAdditive24(
	void *p, const u_int8_t v, const WColorStruct *c
)
{
	/* rr gg bb */
	u_int8_t *tar_pix8 = (u_int8_t *)p;

	const u_int16_t sr = (u_int16_t)(v * c->r / 0xFF),
			sg = (u_int16_t)(v * c->g / 0xFF),
			sb = (u_int16_t)(v * c->b / 0xFF);

	tar_pix8[0] = (u_int8_t)MIN(
	    ((u_int16_t)(tar_pix8[0]) + sb),
	    0x00ff
	);
	tar_pix8[1] = (u_int8_t)MIN(
	    ((u_int16_t)(tar_pix8[1]) + sg),
	    0x00ff
	);
	tar_pix8[2] = (u_int8_t)MIN(
	    ((u_int16_t)(tar_pix8[2]) + sr),
	    0x00ff
	);
}

static void BlitBufPixelAdditive32(
	void *p, const u_int8_t v, const WColorStruct *c
)
{
	/* aa rr gg bb */
	u_int8_t *tar_pix8 = (u_int8_t *)p;

	const u_int16_t sr = (u_int16_t)(v * c->r / 0xFF),
			sg = (u_int16_t)(v * c->g / 0xFF),
			sb = (u_int16_t)(v * c->b / 0xFF);

	tar_pix8[0] = (u_int8_t)MIN(
	    ((u_int16_t)(tar_pix8[0]) + sb),
	    0x00ff
	);
	tar_pix8[1] = (u_int8_t)MIN(
	    ((u_int16_t)(tar_pix8[1]) + sg),
	    0x00ff
	);
	tar_pix8[2] = (u_int8_t)MIN(
	    ((u_int16_t)(tar_pix8[2]) + sr),
	    0x00ff
	);
}


/*
 *	Blits the source buffer to the target buffer.
 */
static void BlitBufArea(
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpp, unsigned int tar_bpl,
	const u_int8_t *src_buf,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpp, unsigned int src_bpl,
	const WColorStruct *c,
	void (*pixel_op_func)(void *, const u_int8_t, const WColorStruct *)
)
{
	int tx, ty, sx, sy;

	/* Check if the source coordinates fall completely off
	 * the target buffer
	 */
	if(((tar_x + (int)src_width) <= 0) ||
	   ((tar_y + (int)src_height) <= 0) ||
	   (tar_x >= (int)tar_width) ||
	   (tar_y >= (int)tar_height)
	)
	    return;

	/* Begin blitting the source buffer to the target buffer,
	 * iterating through the source buffer
	 */
	for(sy = 0, ty = tar_y; sy < (int)src_height; sy++, ty++)
	{
	    if(ty < 0)
		continue;
	    if(ty >= (int)tar_height)
		break;

	    for(sx = 0, tx = tar_x; sx < (int)src_width; sx++, tx++)
	    {
		if(tx < 0)
		    continue;
		if(tx >= (int)tar_width)
		    break;

		/* Blit this pixel */
		pixel_op_func(
		    tar_buf + (ty * (int)tar_bpl) + (tx * (int)tar_bpp),
		    *(src_buf + (sy * (int)src_bpl) + (sx * (int)src_bpp)),
		    c
		);
	    }
	}
}


void BlitBufPointLight(
        unsigned int bpp, unsigned int depth,
        u_int8_t *tar_buf,
        int tar_x, int tar_y,
        unsigned int tar_width, unsigned int tar_height,
        unsigned int tar_bpl,
        float radius,
        const WColorStruct *c
)
{
	static u_int8_t src_buf[] = {
	    0x30, 0xa0, 0x30,
	    0xa0, 0xff, 0xa0,
	    0x30, 0xa0, 0x30
	};
	const unsigned int	src_width = 3,
				src_height = 3,
				src_bpp = 1,
				src_bpl = src_width * src_bpp;

	int tx, ty;
	void (*pixel_op_func)(void *, const u_int8_t, const WColorStruct *);

	if((tar_buf == NULL) || (c == NULL))
	    return;

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

	tx = tar_x - ((int)src_width / 2);
	ty = tar_y - ((int)src_height / 2);

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufPixelAdditive8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
                pixel_op_func = BlitBufPixelAdditive15;
                break;
	      case 16:
		pixel_op_func = BlitBufPixelAdditive16;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  case 3:
	    pixel_op_func = BlitBufPixelAdditive24;
	    break;
          case 4:
            switch(depth)
            {
              case 24:
                pixel_op_func = BlitBufPixelAdditive24;
                break;
              case 32:
                pixel_op_func = BlitBufPixelAdditive32;
                break;
              default:
                return;
                break;
            }
            break;
          default:
            return;
            break;
        }

	BlitBufArea(
	    tar_buf,
	    tx, ty,
	    tar_width, tar_height,
	    bpp, tar_bpl,
	    src_buf,
	    src_width, src_height,
	    src_bpp, src_bpl,
	    c, pixel_op_func
	);
}
