#include <stdio.h>
#include <stdlib.h>

#include "../include/objects.h"
#include "../include/unvutil.h"

#include "clipboard.h"


void UEClipboardAppendObject(
        ue_clipboard_struct *cb, xsw_object_struct *obj
);
xsw_object_struct **UEClipboardPasteObjects(
        ue_clipboard_struct *cb, int *total
);
void UEClipboardClear(ue_clipboard_struct *cb);

ue_clipboard_struct *UEClipboardNew(void);
void UEClipboardDelete(ue_clipboard_struct *cb);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Copy and appends the specified object to the clipboard.
 */
void UEClipboardAppendObject(
        ue_clipboard_struct *cb, xsw_object_struct *obj
)
{
	int i;

        if((cb == NULL) || (obj == NULL))
            return;

	i = MAX(cb->total_objects, 0);
	cb->total_objects = i + 1;
	cb->object = (xsw_object_struct **)realloc(
	    cb->object,
	    cb->total_objects * sizeof(xsw_object_struct *)
	);
	cb->object[i] = UNVDupObject(obj);
}

/*
 *	Returns a coppied list of objects from the clipboard.
 */
xsw_object_struct **UEClipboardPasteObjects(
        ue_clipboard_struct *cb, int *total  
)
{
	int i;
	xsw_object_struct **list;

        if((cb == NULL) || (total == NULL))
            return(NULL);

	if(cb->total_objects <= 0)
	{
	    *total = 0;
	    return(NULL);
	}

	*total = cb->total_objects;
	list = (xsw_object_struct **)malloc(
	    cb->total_objects * sizeof(xsw_object_struct *)
	);
	for(i = 0; i < cb->total_objects; i++)
	    list[i] = UNVDupObject(cb->object[i]);

	return(list);
}

/*
 *	Deletes all the objects in the clipboard.
 */
void UEClipboardClear(ue_clipboard_struct *cb)
{
        if(cb == NULL)
            return;

        UNVDeleteAllObjects(cb->object, cb->total_objects);
	cb->object = NULL;
	cb->total_objects = 0;
}


ue_clipboard_struct *UEClipboardNew(void)
{
	return(UE_CLIPBOARD(calloc(1, sizeof(ue_clipboard_struct))));
}

void UEClipboardDelete(ue_clipboard_struct *cb)
{
	if(cb == NULL)
	    return;

	UEClipboardClear(cb);
	free(cb);
}
