#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/reality.h"
#include "../include/objects.h"
#include "../include/unvutil.h"
#include "../include/eco.h"

#include "ocsinfofio.h"
#include "uew.h"
#include "ue.h"
#include "ecow.h"

#include "../include/oswkeyvals.h"


static xsw_object_struct *EcoWGetObjectPtr(ecow_struct *ecow);

static void EcoWGetValues(ecow_struct *ecow);
static void EcoWSetValues(ecow_struct *ecow);
static void EcoWGetProductValues(ecow_struct *ecow, const int prod_num);
static void EcoWSetProductValues(ecow_struct *ecow, const int prod_num);

static void EcoWDrawCB(void *widget, void *data);

static void EcoWAddProductCB(push_button_struct *btn, void *data);
static void EcoWRemoveProductCB(push_button_struct *btn, void *data);
static void EcoWDeleteEcoCB(push_button_struct *btn, void *data);
static void EcoWCloseCB(push_button_struct *btn, void *data);

static void EcoWCListSelectCB(
	clist_struct *clist, int row_num, void *data
);
static void EcoWCListUnselectCB(
	clist_struct *clist, int row_num, void *data
);

ecow_struct *EcoWNew(uew_struct *uew, const int obj_num);
void EcoWDraw(ecow_struct *ecow, const int amount);
void EcoWQueueDraw(ecow_struct *ecow, const int amount);
int EcoWManage(ecow_struct *ecow, event_t *event);
void EcoWMap(ecow_struct *ecow);
void EcoWMapValues(ecow_struct *ecow);
void EcoWUnmap(ecow_struct *ecow);
void EcoWDelete(ecow_struct *ecow);


#define ECOW_TITLE			"Economy"
#define ECOW_ICON_TITLE			"Economy"

#define ECOW_WIDTH			400
#define ECOW_HEIGHT			380

#define ECOW_MARGIN			10

#define ECOW_PROMPT_HEIGHT		30

#define ECOW_BUTTON_WIDTH		70
#define ECOW_BUTTON_HEIGHT		28


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the pointer to the object referanced in the
 *	economy window ecow.  Returns NULL on error.
 */
static xsw_object_struct *EcoWGetObjectPtr(ecow_struct *ecow)
{
	uew_struct *uew = (ecow != NULL) ? ecow->uew : NULL;
	if(uew == NULL)
	    return(NULL);

	return(UEWGetObjectPtr(uew, ecow->obj_num));
}


/*
 *	Gets all the object's economy values to the widgets.
 */
static void EcoWGetValues(ecow_struct *ecow)
{
	int i, obj_num, row_num, total;
	char *path;
	clist_struct *clist;
	popup_list_struct *pulist;
	ocs_info_struct **ocs_info_list;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	uew_struct *uew;

	if(ecow == NULL)
	    return;

	obj_num = ecow->obj_num;
	clist = &ecow->list;
	pulist = &ecow->prod_ocs_pulist;
	uew = ecow->uew;

	/* Get OCS names */
	if(uew != NULL)
	{
	    const char *name = uew->legend.ocsn;
	    path = ISPATHABSOLUTE(name) ?
		STRDUP(name) :
		STRDUP(PrefixPaths(dname.etc, name));
	}
	else
	{
	    path = NULL;
	}
	PUListClear(pulist);
	ocs_info_list = NULL;
	total = 0;
	if(!OCSInfoFileOpen(
	    &ocs_info_list, &total, path,
	    NULL, NULL, NULL
	))
	{
	    int n;
	    ocs_info_struct *ocs_info;
	    for(i = 0; i < total; i++)
	    {
		ocs_info = ocs_info_list[i];
		if(ocs_info == NULL)
		    continue;

		n = PUListAddItem(
		    pulist, 0, ocs_info->name, None, None
		);
		PUListItemSetData(
		    pulist, n, (void *)ocs_info->code, NULL
		);

		OCSInfoDelete(ocs_info);
	    }
	    free(ocs_info_list);
	}
	free(path);

	/* Delete all existing products */
	CListClear(clist);

	obj = EcoWGetObjectPtr(ecow);
	if(obj != NULL)
	{
	    /* Create the economy data on the object as needed */
	    if(obj->eco == NULL)
	    {
		obj->eco = eco = (xsw_ecodata_struct *)calloc(
		    1, sizeof(xsw_ecodata_struct)
		);
		if(eco != NULL)
		{
		    eco->flags = ECO_FLAG_OPEN | ECO_FLAG_BUY_OK |
				 ECO_FLAG_SELL_OK | ECO_FLAG_TRADE_OK;
		    eco->tax_general = 1.0f;
		    eco->tax_friend = 1.0f;
		    eco->tax_hostile = 1.0f;
		    eco->product = NULL;
		    eco->total_products = 0;
		}

		ecow->has_changes = True;
	    }
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		/* Add eco data values to prompts */
		char *v[1];
		xsw_ecoproduct_struct *prod;

		CBtnSetActive(
		    &ecow->opened_cb,
		    (eco->flags & ECO_FLAG_OPEN) ? True : False
		);
		CBtnSetActive(
		    &ecow->buy_ok_cb,
		    (eco->flags & ECO_FLAG_BUY_OK) ? True : False
		);
		CBtnSetActive(
		    &ecow->sell_ok_cb,
		    (eco->flags & ECO_FLAG_SELL_OK) ? True : False
		);

		PromptSetF2(
		    &ecow->tax_general_prompt,
		    eco->tax_general - 1.0f, 2
		);

		PromptSetF2(
		    &ecow->tax_friend_prompt,
		    eco->tax_friend - 1.0f, 2
		);

		PromptSetF2(
		    &ecow->tax_hostile_prompt,
		    eco->tax_hostile - 1.0f, 2
		);

	        /* Add list rows */
	        for(i = 0; i < eco->total_products; i++)
	        {
		    prod = eco->product[i];
		    if(prod == NULL)
		        continue;

		    row_num = CListRowAppend(clist);
		    if(row_num < 0)
		        break;

		    v[0] = (prod->name != NULL) ?
			strdup(prod->name) : strdup("(null)");

		    CListSetRowTexts(
		        clist, row_num,
		        (const char **)v, sizeof(v) / sizeof(char *)
		    );

		    free(v[0]);
	        }
	    }
	}
	else
	{
	    /* No valid object, reset object reference */
	    ecow->obj_num = -1;
	}

	/* Reset prompts */
	PUListSetSelItem(&ecow->prod_ocs_pulist, 0);
	PromptSetS(
	    &ecow->prod_sell_prompt,
	    ""
	);
	PromptSetS(
	    &ecow->prod_buy_prompt,
	    ""
	);
	PromptSetS(
	    &ecow->prod_amount_prompt,
	    ""
	);
	PromptSetS(
	    &ecow->prod_max_prompt,
	    ""
	);

	if(obj != NULL)
	{
	    const char *name = UNVGetObjectFormalName(obj_num, obj);
	    char *s = (char *)malloc(
		STRLEN(ECOW_TITLE) +
		STRLEN(name) +
		80
	    );
	    sprintf(
		s,
		"%s: %s",
		ECOW_TITLE, name
	    );
	    OSWSetWindowTitle(ecow->toplevel, s);
	    free(s);
	}
}

/*
 *	Sets all the object's economy values from the widgets.
 */
static void EcoWSetValues(ecow_struct *ecow)
{
	clist_struct *clist;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	uew_struct *uew;

	if(ecow == NULL)
	    return;

	clist = &ecow->list;
	uew = ecow->uew;

	/* Unselect all products to apply values to the one that was
	 * last selected
	 */
	CListUnselectAllRows(clist);

	/* Apply the eco values to the object */
	obj = EcoWGetObjectPtr(ecow);
	if(obj != NULL)
	{
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		/* Flags */
		eco->flags = 0;
		if(CBtnGetActive(&ecow->opened_cb))
		    eco->flags |= ECO_FLAG_OPEN;
		if(CBtnGetActive(&ecow->buy_ok_cb))
		    eco->flags |= ECO_FLAG_BUY_OK;
		if(CBtnGetActive(&ecow->sell_ok_cb))
		    eco->flags |= ECO_FLAG_SELL_OK;

		/* Tax general */
		eco->tax_general = PromptGetF(&ecow->tax_general_prompt) + 1.0f;

		/* Tax friend */
		eco->tax_friend = PromptGetF(&ecow->tax_friend_prompt) + 1.0f;

		/* Tax hostile */
		eco->tax_hostile = PromptGetF(&ecow->tax_hostile_prompt) + 1.0f;
	    }
	}

	/* Update the has changes marker on the Universe Editor Window */
	if((uew != NULL) && ecow->has_changes)
	{
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}
}

/*
 *	Gets the specified object's economy product's values to the
 *	widgets.
 */
static void EcoWGetProductValues(ecow_struct *ecow, const int prod_num)
{
	int i;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *prod;
	xsw_object_struct *obj = EcoWGetObjectPtr(ecow);
	if(obj == NULL)
	    return;

	eco = obj->eco;
	if(eco == NULL)
	    return;

	/* Get the product */
	prod = ((prod_num >= 0) && (prod_num < eco->total_products)) ?
	    eco->product[prod_num] : NULL;
	if(prod == NULL)
	    return;

	/* Begin getting values */
	i = PUListGetItemByData(
	    &ecow->prod_ocs_pulist,
	    (void *)prod->ocs_code   
	);
	if(i < 0)
	    i = 0;
	PUListSetSelItem(&ecow->prod_ocs_pulist, i);
	PromptSetF2(
	    &ecow->prod_sell_prompt,
	    prod->sell_price, 2
	);
	PromptSetF2(
	    &ecow->prod_buy_prompt,
	    prod->buy_price, 2
	);
	PromptSetF2(
	    &ecow->prod_amount_prompt,
	    prod->amount, 2
	);
	PromptSetF2(
	    &ecow->prod_max_prompt,
	    prod->amount_max, 2
	);
}

/*
 *	Sets the specified object's economy product's values from the
 *	widgets.
 */
static void EcoWSetProductValues(ecow_struct *ecow, const int prod_num)
{
	int i;
	const char *s;
	popup_list_struct *pulist;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *prod;

	if(ecow == NULL)
	    return;

	obj = EcoWGetObjectPtr(ecow); 
	if(obj == NULL)
	    return;

	eco = obj->eco;
	if(eco == NULL)
	    return;

	/* Get the product */
	prod = ((prod_num >= 0) && (prod_num < eco->total_products)) ?
	    eco->product[prod_num] : NULL;
	if(prod == NULL)
	    return;

	/* Begin setting values */
	pulist = &ecow->prod_ocs_pulist;
	i = PUListGetSelItem(pulist);
	s = PUListGetSelItemName(pulist);
	if(s != NULL)
	{
	    strncpy(prod->name, s, sizeof(prod->name));
	    prod->name[sizeof(prod->name) - 1] = '\0';
	}
	prod->ocs_code = (int)PUListItemGetData(pulist, i);
	prod->sell_price = PromptGetF(&ecow->prod_sell_prompt);
	prod->buy_price = PromptGetF(&ecow->prod_buy_prompt);
	prod->amount = PromptGetF(&ecow->prod_amount_prompt);
	prod->amount_max = PromptGetF(&ecow->prod_max_prompt);

	/* Update the product name on the list */
	CListSetItemText(
	    &ecow->list,
	    prod_num, 0,
	    prod->name
	);
}


/*
 *	Draw callback.
 */
static void EcoWDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	ecow_struct *ecow = ECOW(widget);
	if(ecow == NULL)
	    return;

	if(ecow->map_state)
	    EcoWDraw(ecow, amount);
}


/*
 *	Create callback.
 */
static void EcoWAddProductCB(push_button_struct *btn, void *data)
{
	int i;
	clist_struct *clist;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *prod;
	uew_struct *uew;
	ecow_struct *ecow = ECOW(data);
	if(ecow == NULL)
	    return;

	clist = &ecow->list;
	uew = ecow->uew;

	/* Get the last selected product */
	i = CListGetLastSelectedRow(clist);

	/* Unselect all products to apply values to the one that was
	 * last selected
	 */
	CListUnselectAllRows(clist);

	/* Allocate a new product on object */
	obj = EcoWGetObjectPtr(ecow);
	if(obj != NULL)
	{
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		/* Create a new product */
		i = MAX(eco->total_products, 0);
		eco->total_products++;
		eco->product = (xsw_ecoproduct_struct **)realloc(
		    eco->product,
		    eco->total_products * sizeof(xsw_ecoproduct_struct *)
		);
		if(eco->product == NULL)
		{
		    eco->total_products = 0;
		    return;
		}

		eco->product[i] = prod = (xsw_ecoproduct_struct *)calloc(
		    1, sizeof(xsw_ecoproduct_struct)
		);
		if(prod == NULL)
		    return;

		strncpy(
		    prod->name, "New Product", sizeof(prod->name)
		);
		prod->name[sizeof(prod->name) - 1] = '\0';
		prod->ocs_code = 0;
		prod->amount = 0.0f;
		prod->amount_max = 0.0f;
		prod->buy_price = 0.0f;
		prod->sell_price = 0.0f;

		/* Add this product to the list */
		EcoWGetValues(ecow);

		/* Select the new product */
		CListSelectRow(clist, i);
		if(CListGetRowVisibility(clist, i) != VisibilityUnobscured)
		    CListMoveTo(clist, i, 0.5f);

		ecow->has_changes = True;
	    }
	}

	/* Update the has changes marker on the Universe Editor Window */
	if((uew != NULL) && ecow->has_changes)
	{
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}
}

/*
 *	Delete callback.
 */
static void EcoWRemoveProductCB(push_button_struct *btn, void *data)
{
	int i;
	clist_struct *clist;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	uew_struct *uew;
	ecow_struct *ecow = ECOW(data);
	if(ecow == NULL)
	    return;

	uew = ecow->uew;

	/* Get the last selected product */
	clist = &ecow->list;
	i = CListGetLastSelectedRow(clist);
	if(i < 0)
	{
	    UEDialog(
		"Remove Product",
		UE_ICON_WARNING,
		"No product selected.",
		ecow->toplevel
	    );
	    return;
	}

	obj = EcoWGetObjectPtr(ecow);
	if(obj == NULL)
	    return;

	eco = obj->eco;
	if(eco == NULL)
	    return;

	/* Does the selected product exist on the object? */
	if((i >= 0) && (i < eco->total_products))
	{
	    int j;

	    /* Delete the selected product */
	    free(eco->product[i]);

	    /* Shift pointers */
	    eco->total_products--;
	    for(j = i; j < eco->total_products; j++)
		eco->product[j] = eco->product[j + 1];

	    if(eco->total_products > 0)
	    {
	        eco->product = (xsw_ecoproduct_struct **)realloc(
		    eco->product,
		    eco->total_products * sizeof(xsw_ecoproduct_struct *)
	        );
	        if(eco->product == NULL)
		    eco->total_products = 0;
	    }
	    else
	    {
	        free(eco->product);
	        eco->product = NULL;
	        eco->total_products = 0;
	    }

	    ecow->has_changes = True;
	}

	/* Update the products list */
	EcoWGetValues(ecow);

	/* Update the has changes marker on the Universe Editor Window */
	if((uew != NULL) && ecow->has_changes)
	{
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}
}

/*
 *	Delete economy data on object callback.
 *
 *	This will delete all allocated economy stuff (both ecodata
 *	and products on the object).
 */
static void EcoWDeleteEcoCB(push_button_struct *btn, void *data)
{
	int i;
	clist_struct *clist;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	uew_struct *uew;
	ecow_struct *ecow = ECOW(data);
	if(ecow == NULL)
	    return;

	clist = &ecow->list;
	uew = ecow->uew;

	/* Get pointer to object referanced in economy window */
	obj = EcoWGetObjectPtr(ecow);
	if(obj != NULL)
	{
	    /* Confirm delete */
	    int response;
	    const char *name = UNVGetObjectFormalName(
		ecow->obj_num, obj
	    );
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Delete all economy data on `%s'?",
		name
	    );
	    response = UEGetResponse(
		"Confirm Delete",
		UE_ICON_QUESTION,
		s,
		CDlgBtnYes | CDlgBtnNo,
		ecow->toplevel
	    );
	    free(s);
	    if(response != CDlgYes)
		return;

	    /* Delete all rows in list */
	    CListClear(clist);

	    /* Get the object's economy data */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		for(i = 0; i < eco->total_products; i++)
		    free(eco->product[i]);
		free(eco->product);
		eco->product = NULL;
		eco->total_products = 0;

		free(eco);
		eco = NULL;
	    }

	    obj->eco = NULL;
	}

	ecow->has_changes = True;

	/* Update the has changes marker on the Universe Editor Window */
	if((uew != NULL) && ecow->has_changes)
	{
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}

	/* Unmap the Economy Window since the economy has been deleted
	 * on this object
	 */
	EcoWUnmap(ecow);
}

/*
 *	Close button callback.
 */
static void EcoWCloseCB(push_button_struct *btn, void *data)
{ 
	bool_t apply_to_all_selected_objects = False;
	clist_struct *clist;
	xsw_object_struct *obj;
	uew_struct *uew;
	ecow_struct *ecow = ECOW(data);  
	if(ecow == NULL)
	    return;

	uew = ecow->uew;
	if(uew == NULL)
	    return;

	/* Set all the object's economy values from the widgets */
	EcoWSetValues(ecow);

	/* Check if the universe editor has more than one object
	 * selected
	 */
	clist = &uew->objects_list;
	if(clist->total_sel_rows > 1)
	{
	    /* Query the user if the economy values should be applied
	     * to all the selected objects
	     */
	    const int response = UEGetResponse(
		"Apply Economy",
		UE_ICON_QUESTION,
"Apply the economy values to all the selected objects?",
		CDlgBtnYes | CDlgBtnNo,
		ecow->toplevel
	    );
	    if(response == CDlgYes)
		apply_to_all_selected_objects = True;
	}

	/* Apply products to all the selected objects on the Universe
	 * Editor?
	 */
	obj = EcoWGetObjectPtr(ecow);
	if(apply_to_all_selected_objects && (obj != NULL))
	{
	    int i, n, row_num, prod_num;
	    clist_struct *clist = &uew->objects_list;
	    xsw_object_struct *obj2;
	    xsw_ecodata_struct *eco, *eco2;
	    xsw_ecoproduct_struct *prod, *prod2;

	    if(UNVAllocEco(obj))
		return;

	    eco = obj->eco;

	    /* Iterate through the selected objects on the Universe
	     * Editor
	     */
	    n = clist->total_sel_rows;
	    for(i = 0; i < n; i++)
	    {
		row_num = clist->sel_row[i];
		obj2 = SW_OBJECT(CListRowGetData(clist, row_num));
		if((obj2 == NULL) || (obj2 == obj))
		    continue;

		if(UNVAllocEco(obj2))
		    continue;

		/* Copy economy values from the source object to this
		 * selected target object
		 */
		eco2 = obj2->eco;
		eco2->flags = eco->flags;
		eco2->tax_general = eco->tax_general;
		eco2->tax_friend = eco->tax_friend;
		eco2->tax_hostile = eco->tax_hostile;

		/* Delete all products on the target object and then
		 * copy the products from the source object to the
		 * target object
		 */
		for(prod_num = 0; prod_num < eco2->total_products; prod_num++)
		    free(eco2->product[prod_num]);
		eco2->total_products = eco->total_products;
		if(eco2->total_products > 0)
		{
		    eco2->product = (xsw_ecoproduct_struct **)realloc(
			eco2->product,
			eco2->total_products * sizeof(xsw_ecoproduct_struct *)
		    );
		    for(prod_num = 0; prod_num < eco2->total_products; prod_num++)
		    {
			prod = eco->product[prod_num];
			if(prod == NULL)
			{
			    eco2->product[prod_num] = NULL;
			    continue;
			}

			eco2->product[prod_num] = prod2 = (xsw_ecoproduct_struct *)malloc(
			    sizeof(xsw_ecoproduct_struct)
			);
			if(prod2 != NULL)
			    memcpy(prod2, prod, sizeof(xsw_ecoproduct_struct));
		    }
		}
		else
		{
		    free(eco2->product);
		    eco2->product = NULL;
		    eco2->total_products = 0;
		}

		if(!uew->has_changes)
		{
		    uew->has_changes = True;
		    UEWUpdateTitle(uew);
		}
	    }
	}

	/* Unmap */
	EcoWUnmap(ecow);
} 

/*
 *	Forcus prompt macro for EcoWRefocusPrompts().
 */
void EcoWDoPromptFocus(
	prompt_struct *src_prompt,
	prompt_struct *tar_prompt
)
{
	if((src_prompt == NULL) || (tar_prompt == NULL))
	    return;

	if(src_prompt == tar_prompt)
	{
	    src_prompt->has_focus = True;
	    PromptMarkBuffer(src_prompt, PROMPT_POS_END);
	}
	else
	{
	    src_prompt->has_focus = False;
	    PromptUnmarkBuffer(src_prompt, PROMPT_POS_END);
	}
	if(src_prompt->map_state)
	    PromptQueueDraw(src_prompt, PROMPT_DRAW_AMOUNT_TEXT);
}


/*
 *	Refocuses all the prompts so that only the specified prompt
 *	is in focus.
 */
void EcoWRefocusPrompts(
	ecow_struct *ecow, prompt_struct *prompt
)
{
	if((ecow == NULL) || (prompt == NULL))
	    return;

	EcoWDoPromptFocus(&ecow->tax_general_prompt, prompt);
	EcoWDoPromptFocus(&ecow->tax_friend_prompt, prompt);
	EcoWDoPromptFocus(&ecow->tax_hostile_prompt, prompt);

	EcoWDoPromptFocus(&ecow->prod_buy_prompt, prompt);
	EcoWDoPromptFocus(&ecow->prod_sell_prompt, prompt);
	EcoWDoPromptFocus(&ecow->prod_amount_prompt, prompt);
	EcoWDoPromptFocus(&ecow->prod_max_prompt, prompt);
}


/*
 *	Select row callback.
 */
static void EcoWCListSelectCB(
	clist_struct *clist, int row_num, void *data
)
{
	ecow_struct *ecow = ECOW(data);
	if(ecow == NULL)
	    return;

	EcoWGetProductValues(ecow, row_num);
}

/*
 *	Unselect row callback.
 */
static void EcoWCListUnselectCB(
	clist_struct *clist, int row_num, void *data
)
{
	ecow_struct *ecow = ECOW(data);
	if(ecow == NULL)
	    return;

	EcoWSetProductValues(ecow, row_num);
}


/*
 *	Initializes a economy window.
 */
ecow_struct *EcoWNew(uew_struct *uew, const int obj_num)
{
	int x, y, x2, y2;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	win_t w, parent = osw_gui[0].root_win;
	push_button_struct *btn;
	clist_struct *clist;
	popup_list_struct *pulist;
	widget_global_struct *wglobal = &widget_global;
	ecow_struct *ecow = ECOW(calloc(1, sizeof(ecow_struct)));
    
	ecow->map_state = False;
	ecow->has_focus = False;
	ecow->freeze_count = 0;
	ecow->x = 0;
	ecow->y = 0;
	ecow->width = ECOW_WIDTH;
	ecow->height = ECOW_HEIGHT;
	ecow->uew = uew;
	ecow->obj_num = obj_num;
	ecow->has_changes = False;

	/* Create toplevel */
	x = ecow->x;
	y = ecow->y;
	width = ecow->width;
	height = ecow->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	ecow->toplevel = w;

	OSWSetWindowWMProperties(
	    w,
	    ECOW_TITLE,
	    ECOW_ICON_TITLE,
	    (ue_image.icon_unvedit_pixmap != None) ?
		ue_image.icon_unvedit_pixmap :
		wglobal->std_icon_pm,
	    True,			/* Let WM set coordinates? */
	    ecow->x, ecow->y,
	    ecow->width, ecow->height,
	    ecow->width, ecow->height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	if(uew != NULL)
	    OSWSetTransientFor(w, uew->toplevel);

	parent = w;


	/* Opened */
	x = ECOW_MARGIN;
	y = ECOW_MARGIN;
	CBtnInit(
	    &ecow->opened_cb, parent,
	    x, y, False, "Opened"
	);
	CBtnSetTooltip(
	    &ecow->opened_cb,
"Check this to allow economy transactions (ie buy or sell) with\
 this object"
	);
	height = ecow->opened_cb.height;

	/* Buy OK */
	y += (int)height + (ECOW_MARGIN / 2);
	CBtnInit(
	    &ecow->buy_ok_cb, parent,
	    x, y, False, "Buy OK"
	);
	CBtnSetTooltip(
	    &ecow->buy_ok_cb,
"Check this to allow the buying or obtaining of products from\
 this object"
	);

	/* Sell OK */
	y += (int)height + (ECOW_MARGIN / 2);
	CBtnInit(
	    &ecow->sell_ok_cb, parent,
	    x, y, False, "Sell OK"
	);
	CBtnSetTooltip(
	    &ecow->sell_ok_cb,
"Check this to allow the selling or distributing of products to\
 this object"
	);

	/* Tax general */
	x = (int)ecow->width / 2;
	y = ECOW_MARGIN;
	width = 180;
	height = ECOW_PROMPT_HEIGHT;
	PromptInit(
	    &ecow->tax_general_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Tax General:",
	    40, 0
	);
	PromptSetTooltip(
	    &ecow->tax_general_prompt,
"Enter the tax coefficient for transactions between\
 different-empire customers (or enter 0 for no tax)"
	);

	/* Tax friend */
	y += (int)height;
	PromptInit(
	    &ecow->tax_friend_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED, 
	    "Tax Friend:",
	    40, 0
	);
	PromptSetTooltip(
	    &ecow->tax_friend_prompt,
"Enter the tax coefficient for transactions between\
 same-empire customers (or enter 0 for no tax)"
	);

	/* Tax hostile */
	y += (int)height;
	PromptInit(
	    &ecow->tax_hostile_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Tax Hostile:",
	    40, 0
	);
	PromptSetTooltip(
	    &ecow->tax_hostile_prompt,
"Enter the tax coefficient for transactions between\
 hostile-empire customers (or enter 0 for no tax)"
	);


	/* Products list */
	clist = &ecow->list;
	x = ECOW_MARGIN;
	y += (int)height + (ECOW_MARGIN / 2);
	CListInit(
	    clist, parent,
	    x, y,
	    MAX(
		(int)ecow->width - ECOW_BUTTON_WIDTH -
		    (2 * ECOW_MARGIN) - (ECOW_MARGIN / 2),
		1
	    ),
	    100
	);
	clist->select_cb = EcoWCListSelectCB;
	clist->select_data = ecow;
	clist->unselect_cb = EcoWCListUnselectCB;
	clist->unselect_data = ecow;

	CListColumnAppend(
	    clist,
	    "Product",
	    wglobal->std_font,
	    wglobal->normal_text_pix,
	    0,
	    0
	);

	/* Add Product button */
	btn = &ecow->create_prod_btn;
	x2 = (int)ecow->width - ECOW_BUTTON_WIDTH - ECOW_MARGIN;
	y2 = y;
	PBtnInit(
	    btn, parent,
	    x2, y2,
	    ECOW_BUTTON_WIDTH, ECOW_BUTTON_HEIGHT,
	    "Add",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWAddProductCB, ecow
	);

	/* Remove Product button */
	btn = &ecow->delete_prod_btn;
	y2 += ECOW_BUTTON_HEIGHT + (ECOW_MARGIN / 2);
	PBtnInit(
	    btn, parent,
	    x2, y2,
	    ECOW_BUTTON_WIDTH, ECOW_BUTTON_HEIGHT,
	    "Remove",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWRemoveProductCB, ecow
	);


	/* OCS Code */
	pulist = &ecow->prod_ocs_pulist;
	x = ECOW_MARGIN;
	y = clist->y + (int)clist->height + (ECOW_MARGIN / 2);
	PUListInit(
	    pulist, parent,
	    x, y,
	    (unsigned int)MAX(
		(int)ecow->width - (2 * ECOW_MARGIN),
		1
	    ),
	    0,
	    0,				/* Items visible */
	    PULIST_POPUP_BELOW
	);

	/* Buy Price */
	x = ECOW_MARGIN;
	y += pulist->height + (ECOW_MARGIN / 2);
	PromptInit(
	    &ecow->prod_buy_prompt, parent,
	    x, y,
	    160,
	    ECOW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Buy Price:",
	    80, 0
	);
	PromptSetTooltip(
	    &ecow->prod_buy_prompt,
"Enter the cost in credits for the customer to buy this product\
 from the proprietor (or enter 0 for free)"
	);

	/* Sell price */
	x = ECOW_MARGIN + 160 + ECOW_MARGIN;
	PromptInit(
	    &ecow->prod_sell_prompt, parent,
	    x, y,
	    160,
	    ECOW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Sell Price:",
	    80, 0
	);
	PromptSetTooltip(
	    &ecow->prod_sell_prompt,
"Enter the cost in credits for the customer to sell this product\
 to the proprietor (or enter 0 for free)"
	);

	/* Amount prompt */
	x = ECOW_MARGIN;
	y += ECOW_PROMPT_HEIGHT;
	PromptInit(
	    &ecow->prod_amount_prompt, parent,
	    x, y,
	    160,
	    ECOW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Amount:",
	    40, 0
	);
	PromptSetTooltip(
	    &ecow->prod_amount_prompt,
"Enter the quantity of this product currently in stock\
 (or enter -1 for unlimited)"
	);

	/* Amount max */
	x = ECOW_MARGIN + 160 + ECOW_MARGIN;
	PromptInit(
	    &ecow->prod_max_prompt, parent,
	    x, y,
	    160,
	    ECOW_PROMPT_HEIGHT,
	    PROMPT_STYLE_FLUSHED,
	    "Max:",
	    80, 0
	);
	PromptSetTooltip(
	    &ecow->prod_max_prompt,
"Enter the maximum quantity of this product that the proprietor\
 can stock (or enter -1 for unlimited)"
	);

	/* Link prompts togeather */
	ecow->tax_general_prompt.prev = &ecow->prod_max_prompt;
	ecow->tax_general_prompt.next = &ecow->tax_friend_prompt;

	ecow->tax_friend_prompt.prev = &ecow->tax_general_prompt;
	ecow->tax_friend_prompt.next = &ecow->tax_hostile_prompt;

	ecow->tax_hostile_prompt.prev = &ecow->tax_friend_prompt;
	ecow->tax_hostile_prompt.next = &ecow->prod_buy_prompt;


	ecow->prod_buy_prompt.next = &ecow->prod_sell_prompt;
	ecow->prod_buy_prompt.prev = &ecow->tax_hostile_prompt;

	ecow->prod_sell_prompt.prev = &ecow->prod_buy_prompt;
	ecow->prod_sell_prompt.next = &ecow->prod_amount_prompt;

	ecow->prod_amount_prompt.prev = &ecow->prod_sell_prompt;
	ecow->prod_amount_prompt.next = &ecow->prod_max_prompt;

	ecow->prod_max_prompt.prev = &ecow->prod_amount_prompt;
	ecow->prod_max_prompt.next = &ecow->tax_general_prompt;


	/* Close button */
	btn = &ecow->close_btn;
	x = (int)ecow->width - ECOW_BUTTON_WIDTH - ECOW_MARGIN;
	y = (int)ecow->height - ECOW_MARGIN - ECOW_BUTTON_HEIGHT;
	PBtnInit(
	    &ecow->close_btn, parent,
	    x, y,
	    ECOW_BUTTON_WIDTH, ECOW_BUTTON_HEIGHT,
	    "Close",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWCloseCB, ecow
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	/* Delete object's economy */
	btn = &ecow->delete_obj_eco_btn;
	x = ECOW_MARGIN;
	PBtnInit(
	    btn, ecow->toplevel,
	    x, y,
	    190, ECOW_BUTTON_HEIGHT,
	    "Delete Object's Economy",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWDeleteEcoCB, ecow
	);

	return(ecow);
}

/*
 *	Draws the Economy Window.
 */
void EcoWDraw(ecow_struct *ecow, const int amount)
{
	unsigned int width, height;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if(ecow == NULL)
	    return;

	if(True)
	{
	    w = ecow->toplevel;
	    width = ecow->width;
	    height = ecow->height;

	    /* Redraw the background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(osw_gui[0].black_pix);
		OSWDrawSolidRectangle(
		    w, 0, 0, width, height
		);
	    }
	    else
	    {
		int y;

		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);

		OSWSetForeground(wglobal->surface_shadow_pix);
		y = (int)height - 25 - ECOW_BUTTON_HEIGHT;
		OSWDrawLine(
		    w,
		    0, y,
		    (int)width, y
		);
		OSWSetForeground(wglobal->surface_highlight_pix);
		y = (int)height - 24 - ECOW_BUTTON_HEIGHT;
		OSWDrawLine(
		    w,
		    0, y,
		    (int)width, y
		);
	    }
	}
}

/*
 *	Queues the Economy Window to draw.
 */
void EcoWQueueDraw(ecow_struct *ecow, const int amount)
{
	if(ecow == NULL)
	    return;

	WidgetQueueDrawAction(
	    EcoWDrawCB, ecow, (void *)amount
	);
}

/*
 *	Manages the Economy Window.
 */
int EcoWManage(ecow_struct *ecow, event_t *event)
{
	int events_handled = 0;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w;
	event_key_t *key;

	if((ecow == NULL) || (event == NULL))
	    return(events_handled);

	if(!ecow->map_state)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case KeyPress:
	    if(!ecow->has_focus)
		break;

	    key = &event->xkey;
	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Enter */
	    if((keyval == OSW_KEY_Return) ||
	       (keyval == OSW_KEY_KP_Enter) ||
	       (keyval == OSW_KEY_ISO_Enter)
	    )
	    {
		EcoWCloseCB(&ecow->close_btn, ecow);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case KeyRelease:
	    if(!ecow->has_focus) 
		break;

	    break;

	  case Expose:
	    if(w == ecow->toplevel)
	    {
		EcoWDraw(ecow, 0);
		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == ecow->toplevel)
	    {
		ecow->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == ecow->toplevel)
	    {
		ecow->has_focus = False;
		events_handled++;
	    }   
	    break;

	  case ClientMessage:
	    if(OSWEventIsDeleteWindow(ecow->toplevel, event))
	    {
		EcoWCloseCB(&ecow->close_btn, ecow);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	/* Manage widgets */

	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->list, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->delete_obj_eco_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->close_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->create_prod_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->delete_prod_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->opened_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->buy_ok_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->sell_ok_cb, event); 

	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->tax_general_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->tax_friend_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->tax_hostile_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->prod_ocs_pulist, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->prod_buy_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->prod_sell_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->prod_amount_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ecow->prod_max_prompt, event);

	return(events_handled);
}

/*
 *	Maps the Economy Window.
 */
void EcoWMap(ecow_struct *ecow)
{
	if(ecow == NULL)
	    return;

	OSWMapRaised(ecow->toplevel);
	ecow->map_state = True;

	WidgetMap(&ecow->list);

	WidgetMap(&ecow->create_prod_btn);
	WidgetMap(&ecow->delete_prod_btn); 

	WidgetMap(&ecow->opened_cb);
	WidgetMap(&ecow->buy_ok_cb);
	WidgetMap(&ecow->sell_ok_cb);

	WidgetMap(&ecow->tax_general_prompt);
	WidgetMap(&ecow->tax_friend_prompt);
	WidgetMap(&ecow->tax_hostile_prompt);

	WidgetMap(&ecow->prod_ocs_pulist);
	WidgetMap(&ecow->prod_buy_prompt);
	WidgetMap(&ecow->prod_sell_prompt);
	WidgetMap(&ecow->prod_amount_prompt);
	WidgetMap(&ecow->prod_max_prompt);

	WidgetMap(&ecow->delete_obj_eco_btn);
	WidgetMap(&ecow->close_btn);

	OSWGrabFocus(ecow->toplevel);

	UEWDoUpdateWindowMenus();
}

/*
 *	Procedure to map economy window and fetch values.
 */
void EcoWMapValues(ecow_struct *ecow)
{
	uew_struct *uew;

	if(ecow == NULL)
	    return;

	EcoWGetValues(ecow);

	uew = ecow->uew;

	/* Select the first product */
	CListSelectRow(&ecow->list, 0);

	ecow->has_changes = False;

	/* Map window */
	EcoWMap(ecow);
}

/*
 *	Unmaps the Economy Window.
 */
void EcoWUnmap(ecow_struct *ecow)
{
	if(ecow == NULL)
	    return;

	OSWUnmapWindow(ecow->toplevel);
	ecow->map_state = False;

	UEWDoUpdateWindowMenus();
}

/*
 *	Deletes the Economy Window.
 */
void EcoWDelete(ecow_struct *ecow)
{
	if(ecow == NULL)
	    return;

	CListClear(&ecow->list);

	if(IDC())
	{
	    WidgetDestroy(&ecow->close_btn);
	    WidgetDestroy(&ecow->delete_obj_eco_btn);

	    WidgetDestroy(&ecow->opened_cb);
	    WidgetDestroy(&ecow->buy_ok_cb);
	    WidgetDestroy(&ecow->sell_ok_cb);

	    WidgetDestroy(&ecow->tax_general_prompt);
	    WidgetDestroy(&ecow->tax_friend_prompt);
	    WidgetDestroy(&ecow->tax_hostile_prompt);

	    WidgetDestroy(&ecow->prod_ocs_pulist);
	    WidgetDestroy(&ecow->prod_buy_prompt);
	    WidgetDestroy(&ecow->prod_sell_prompt);
	    WidgetDestroy(&ecow->prod_amount_prompt);
	    WidgetDestroy(&ecow->prod_max_prompt);

	    WidgetDestroy(&ecow->delete_prod_btn);
	    WidgetDestroy(&ecow->create_prod_btn);
 
	    WidgetDestroy(&ecow->list);

	    OSWDestroyWindow(&ecow->toplevel);
	}

	free(ecow);
}
