#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/tga.h"
#include "../include/osw-x.h"
#include "../include/widget.h"
#include "../include/objects.h"
#include "../include/imgrefs.h"
#include "../include/unvmain.h"
#include "../include/unvfile.h"

#include "cfgfile.h"
#include "clipboard.h"
#include "keymap.h"
#include "uew.h"
#include "propdlg.h"
#include "optwin.h"
#include "aboutdlg.h"
#include "ue.h"
#include "config.h"

#include "../include/cursors/translate.xpm"
#include "../include/cursors/xsw_scanner_lock.xpm"
#include "../include/cursors/zoom.xpm"

#include "../include/icons/info.h"
#include "../include/icons/error.h"
#include "../include/icons/question.h"
#include "../include/icons/warning.h"


int UEDoEmergencySaveAll(void);
static void UESignalCB(int s);

mode_t UECurrentFileMode(void);
const char *UECurrentDirectory(void);
bool_t UECheckFileExists(const char *path, char **error_msg_rtn);

int UECreateUEW(const char *unv_file);

/* Reset timmers */
void UEResetTimmers(const time_t cur_ms, const time_t cur_sec);

/* Manage */
static void UEMaintainImgRefLights(
	void *data,
	imgref_light_struct **list, const int total,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void UEMaintainImgRefs(
	void *data,
	imgref_struct **list, const int total,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static int UEManageGUI(
	void *data,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void UEMaintain(
	void *data,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

/* Initialize, Main, and Shutdown */
int UEInit(int argc, char **argv);
void UEMainIteration(void *data);
void UEMain(void *data);
void UEMainPop(void *data);
void UEMainPopAll(void *data);
void UEShutdown(void);


/*
 *	Maximum imgref pixels to load per iteration when using
 *	async image loading:
 */
#define IMGREF_MAX_PIXEL_LOAD_ITERATION		100000

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)



int			runlevel;
time_t			cur_millitime;

ue_clipboard_struct	*ue_clipboard;

ue_fname_struct		fname;
ue_dname_struct		dname;
ue_color_struct		ue_color;
ue_font_struct		ue_font;
ue_cursor_struct	ue_cursor;
ue_image_struct		ue_image;
ue_option_struct	option;

keymap_struct		keymap[UE_TOTAL_KEYMAPS];

dialog_struct		*ue_dlg;
cdlg_struct		*ue_cdlg;
fb_struct		*ue_file_sel;

uew_struct		**ue_uew;
int			total_uews;

optwin_struct		*optwin;
about_dlg_struct	*about_dlg;


/*
 *	Procedure to do emergancy save all.
 */
int UEDoEmergencySaveAll(void)
{
#if 0
	int i, status;
	char *dir_name, *tmp_name;
	struct stat stat_buf;
	sw_legend_struct *legend;
	uew_struct *uew;

	for(i = 0; i < total_uews; i++)
	{
	    uew = ue_uew[i];
	    if(uew == NULL)
		continue;

	    /* No objects to save? */
	    if(uew->total_objects <= 0)
		continue;

	    /* Get user's home dir */
	    dir_name = getenv("HOME");
	    if(dir_name == NULL)
		dir_name = "/tmp";
	    if(stat(dir_name, &stat_buf))
		dir_name = "/";

	    /* Create a tempory file name */
	    tmp_name = tempnam(dir_name, "univ");
	    if(tmp_name == NULL)
		continue;

	    /* Update the universe legend */
	    legend = &uew->legend;
	    strncpy(
		legend->version, PROG_VERSION,
		sizeof(legend->version)
	    );
	    legend->version[sizeof(legend->version) - 1] = '\0';

	    /* Save universe */
	    status = UNVFileSave(
		tmp_name,
		uew->object, uew->total_objects,
		legend,
		NULL, NULL
	    );
	    if(status < 0)
	    {
		fprintf(
		    stderr,
		    "%s: Emergency save failed.\n",
		    tmp_name
		);
	    }
	    else
	    {
		fprintf(
		    stderr,
		    "%s: Emergency save successful.\n",
		    tmp_name  
		);
	    } 

	    free(tmp_name);
	    tmp_name = NULL;
	}
#endif
	return(0);
}


/*
 *	Signal handler.
 */
static void UESignalCB(int s)
{
	switch(s)
	{
	  case SIGHUP:
	    break;

	  case SIGINT:
	    UEDoEmergencySaveAll();
	    UEMainPopAll(NULL);
	    break;

	  case SIGTERM:
	    UEDoEmergencySaveAll();
	    UEMainPopAll(NULL);
	    break;

	  case SIGQUIT:
	    UEDoEmergencySaveAll();
	    UEMainPopAll(NULL);
	    break;

	  case SIGABRT:
	    UEDoEmergencySaveAll();
	    UEMainPopAll(NULL);
	    break;
	    
	  case SIGPIPE:
	    break;

	  case SIGSEGV:
	    UEDoEmergencySaveAll();
	    exit(1);
	    break;

	  case SIGCONT:
	    break;
	}
}


/*
 *	Returns the current permissions for files.
 */
mode_t UECurrentFileMode(void)
{
	const mode_t m = umask(0);
        umask(m);
	return(
	    (~m) &
            (S_IRUSR | S_IWUSR |
             S_IRGRP | S_IWGRP |
             S_IROTH | S_IWOTH)
	);
}

/*
 *	Returns the current working directory.
 *
 *	The returned string must not be modified or deleted.
 */
const char *UECurrentDirectory(void)
{
	static char path[PATH_MAX];
	return(getcwd(path, sizeof(path)));
}

/*
 *	Checks if the file specified by path exists.
 *
 *	If it exists then True is returned, otherwise False is returned
 *	and if error_msg_rtn is not NULL then a string describing the
 *	error will be coppied to *error_msg_rtn.
 *
 *	The *error_msg_rtn must be deleted by the calling function.
 */
bool_t UECheckFileExists(const char *path, char **error_msg_rtn)
{
	struct stat stat_buf;

	if(STRISEMPTY(path))
	{
	    if(error_msg_rtn != NULL)
		*error_msg_rtn = STRDUP("The specified path is an empty string");
	    return(False);
	}

	/* Attempt to get the file's statistics in order to check if
	 * it exists
	 */
	if(stat(path, &stat_buf))
	{
	    /* Does not exist, get the error message describing the
	     * reason (if required)
	     */
	    const int error_code = errno;
	    if(error_msg_rtn != NULL)
	    {
		char *s = STRDUP(strerror(error_code));
		if(s != NULL)
		    *s = toupper(*s);
		*error_msg_rtn = s;
	    }
	    return(False);
	}
	else
	{
	    /* Exists, always return no error message */
	    if(error_msg_rtn != NULL)
		*error_msg_rtn = NULL;
	    return(True);
	}
}


/*
 *	Creates a new Universe Editor.
 *
 *	The unv_file specifies the path to the universe file to open,
 *	it can be NULL for no universe file.
 */
int UECreateUEW(const char *unv_file)
{
	int uew_num;
	sw_legend_struct *legend;
	uew_struct *uew = UEWNew();
	if(uew == NULL)
	    return(-1);

	for(uew_num = 0; uew_num < total_uews; uew_num++)
	{
	    if(ue_uew[uew_num] == NULL)
		break;
	}
	if(uew_num < total_uews)
	{
	    ue_uew[uew_num] = uew;
	}
	else
	{
	    uew_num = MAX(total_uews, 0);
	    total_uews = uew_num + 1;
	    ue_uew = (uew_struct **)realloc(
		ue_uew,
		total_uews * sizeof(uew_struct *)
	    );
	    if(ue_uew == NULL)
	    {
		total_uews = 0;
		UEWDelete(uew);
		return(-3);
	    }
	    ue_uew[uew_num] = uew;
	}

	UEWMap(uew);
	UEWUpdateWidgets(uew);

	/* Process all draw events so that the window is fully drawn
	 * at this point
	 */
	OSWBlockInput(True);
	while(OSWEventsPending() > 0)
	    UEMainIteration(NULL);
	OSWBlockInput(False);

	/* Grab focus */
	OSWGrabFocus(uew->toplevel);
	uew->has_focus = True;

	legend = &uew->legend;

	/* Set the OPM file */
	UEWSetOPMFile(uew, fname.opm);

	/* Open universe file on creation? */
	if(UECheckFileExists(unv_file, NULL))
	{
	    UEWOpenFile(uew, unv_file);
	}
	else
	{
	    /* Since we did not open a universe then we need to open
	     * the default reference files if available
	     */
	    UEWOpenImgRefs(uew, legend->imgrefs);
	    UEWStatus(uew, "Ready.");
	}

	return(uew_num);
}


/*
 *	Resets all timmers.
 */
void UEResetTimmers(const time_t cur_ms, const time_t cur_sec)
{
	int i, n;
	uew_struct *uew;

	/* Universe editor windows */
	n = total_uews;
	for(i = 0; i < n; i++)
	{
	    uew = ue_uew[i];
	    if(uew == NULL)
		continue;

	    if(uew->total_imgrefs > 0)
	    {
		const int n = uew->total_imgrefs;
		int i;
		imgref_struct *imgref;
		for(i = 0; i < n; i++)
		{
		    imgref = uew->imgref[i];
		    if(imgref == NULL)
			continue;

		    if(imgref->nlights > 0)
		    {
			const int n = imgref->nlights;
			int i;
			imgref_light_struct *light;
			for(i = 0; i < n; i++)
			{
			    light = imgref->light[i];
			    if(light == NULL)
				continue;

			    light->strobe_next = cur_ms;
			}
		    }
		}
	    }

	    uew->ani_next = cur_ms;
	}
}

/*
 *	Manages the ImgRef lights.
 *
 *	Updates the light states.
 */
static void UEMaintainImgRefLights(
	void *data,
	imgref_light_struct **list, const int total,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i;
	imgref_light_struct *light;

	for(i = 0; i < total; i++)
	{
	    light = list[i];
	    if(light == NULL)
		continue;

	    /* Is this light suppose to strobe? */
	    if((light->strobe_off_int > 0l) &&
	       (light->strobe_on_int > 0l)
	    )
	    {
		if(light->strobe_state)
		{
		    /* Strobe is currently on */
		    if(light->strobe_next <= cur_ms)
		    {
			/* Strobe just turned off */
			light->strobe_state = 0;

			light->strobe_next = cur_ms +
			    light->strobe_off_int;
		    }
		}
		else
		{
		    /* Strobe is currently off */
		    if(light->strobe_next <= cur_ms)
		    {
			/* Strobe just turned on */
			light->strobe_state = 1;

			light->strobe_next = cur_ms +
			    light->strobe_on_int;
		    }
		}
	    }
	}
}

/*
 *	Manages the ImgRefs.
 *
 *	Updates the light states.
 *
 *	Updates the animated frames of imgrefs displayed on various
 *	windows.
 */
void UEMaintainImgRefs(
	void *data,
	imgref_struct **list, const int total,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, status, pixels_loaded = 0;
	imgref_struct *imgref;
	tga_data_struct *td;
	osw_gui_struct *gui = &osw_gui[0];

	for(i = 0; i < total; i++)
	{
	    imgref = list[i];
	    if(imgref == NULL)
		continue;

	    /* Manage lights */
	    UEMaintainImgRefLights(
		data,
		imgref->light, imgref->nlights,
		cur_ms, cur_sec, tc
	    );

	    /* Passive image loading */

	    /* Can we still load more pixels and is this imgref's
	     * image in the loading process?
	     */
	    if((pixels_loaded < 100000) &&
	       (imgref->load_progress == IMGREF_LOAD_PROGRESS_LOADING)
	    )
	    {
		/* Library data not allocated? */
		if(imgref->lib_data == NULL)
		    continue;

		/* Continue loading by format */
		switch(imgref->img_fmt)
		{
		  case IMGREF_IMGFMT_TGA:
		    td = (tga_data_struct *)imgref->lib_data;
		    status = TgaReadPartialFromFile(
			td,
			TgaGetDepthForX(
			    gui->bits_per_pixel,
			    gui->depth
			),
			IMGREF_MAX_PIXEL_LOAD_ITERATION
		    );
		    if(status != TgaSuccess)
		    {
			/* Load error, reset imgref and load as default */
			imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		    }
		    /* Done loading? */
		    if(td->cur_load_pixel < 0)
		    {
			/* Image is done loading, mark it as such */
			imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		    }
		    /* Increment pixels loaded */
		    pixels_loaded += IMGREF_MAX_PIXEL_LOAD_ITERATION;
		    break;

		  default:
		    imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		    break;
		}
	    }
	}

	/* Check if any images' contents were loaded, if so, then
	 * redraw all windows that display them
	 */
	if(pixels_loaded > 0)
	{
	    uew_struct *uew;

	    for(i = 0; i < total_uews; i++)
	    {
		uew = ue_uew[i];
		if(uew == NULL)
		    continue;

		if(!uew->map_state)
		    continue;

		UEWQueueDraw(uew, UEW_DRAW_PREVIEW);

		if(uew->total_prop_dlgs > 0)
		{
		    const int n = uew->total_prop_dlgs;
		    int i;
		    prop_dlg_struct *d;
		    for(i = 0; i < n; i++)
		    {
			d = uew->prop_dlg[i];
			if(d == NULL)
			    continue;

			if(!d->map_state)
			    continue;

			PropDlgQueueDraw(d, PROP_DLG_DRAW_OBJECT_DISPLAY);
		    }
		}
	    }
	}
}

/*
 *	Manages the GUI and widget set.
 */
static int UEManageGUI(
	void *data,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int events_handled, total_events_handled = 0;
	event_t event;

	/* Manage GUI events */
	while(OSWEventsPending() > 0)
	{
	    events_handled = 0;
		 
	    /* Get event */
	    OSWWaitNextEvent(&event);

	    /* Manage widget event */
	    if(events_handled == 0)
		events_handled += WidgetManageGUI(&event);

	    /* Manage all universe edit windows */
	    if(events_handled == 0)
	    {
		int i;
		for(i = 0; i < total_uews; i++)
		{
		    events_handled += UEWManage(ue_uew[i], &event);
		    if(events_handled > 0)
			break;
		}
	    }
	    if(events_handled == 0)
		events_handled += OptWinManage(optwin, &event);
	    if(events_handled == 0)
		events_handled += AboutDlgManage(about_dlg, &event);
	    if(events_handled == 0)
		events_handled += CDlgManage(ue_cdlg, &event);

	    /* Manage all widget set windows */
	    if(events_handled == 0)
		events_handled += DialogManage(ue_dlg, &event);
	    if(events_handled == 0)
		events_handled += FBrowserManage(ue_file_sel, &event);

	    total_events_handled += events_handled;
	}

	/* Process queued widget events */
	WidgetMainIteration(NULL);

	return(total_events_handled);
}

/*
 *	Maintains all windows.
 */
static void UEMaintain(
	void *data,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, nuews_mapped = 0;
	uew_struct *uew;

	for(i = 0; i < total_uews; i++)
	{
	    uew = ue_uew[i];
	    if(uew == NULL)
		continue;

	    /* If the window is not mapped then delete it */
	    if(!uew->map_state)
	    {
		UEWDelete(uew);
		ue_uew[i] = NULL;
		continue;
	    }

	    nuews_mapped++;

	    UEWMaintain(data, uew, cur_ms, cur_sec, tc);
	}



	/* No universe editors mapped? */
	if(nuews_mapped == 0)
	    UEMainPop(data);
}


/*
 *	Initializes UnvEdit.
 */
int UEInit(int argc, char **argv)
{
	int i, uew_num, nfiles = 0;
	const char *s, *arg;
	char	*cwd = STRDUP(UECurrentDirectory()),
		*ue_images_dir = NULL,
		**files_list = NULL;
	rectangle_t *rect;
	ue_option_struct *opt = &option;

	/* Set the default options */
	opt->record_window_geometry = False;
	opt->show_object_labels = True;
	opt->show_geometry_labels = True;
	opt->show_grid = True;
	opt->show_contact_bounds = True;
	opt->animate_object_display = True;
	opt->grid_spacing = 200.0f;
	opt->object_marker_radius = 5;
	rect = &opt->uew_geometry;
	rect->x = 0;
	rect->y = 0;
	rect->width = 0;
	rect->height = 0;
	opt->object_label_font_name = STRDUP(
"-adobe-helvetica-medium-r-normal-*-12-*-*-*-p-*-iso8859-1"
	);
	opt->geometry_label_font_name = STRDUP(
"-adobe-helvetica-medium-r-normal-*-10-*-*-*-p-*-iso8859-1"
	);


	/* Set the default files & directories */
	strncpy(dname.toplevel, UE_DEF_XSW_TOPLEVEL_DIR, sizeof(dname.toplevel));
	dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';

	strncpy(dname.etc, UE_DEF_XSW_ETC_DIR, sizeof(dname.etc));
	dname.etc[sizeof(dname.etc) - 1] = '\0';

	strncpy(dname.images, UE_DEF_XSW_IMAGES_DIR, sizeof(dname.images));
	dname.images[sizeof(dname.images) - 1] = '\0';

	strncpy(dname.sounds, UE_DEF_XSW_SOUNDS_DIR, sizeof(dname.sounds));
	dname.sounds[sizeof(dname.sounds) - 1] = '\0';

	strncpy(dname.server, UE_DEF_SERVER_DIR, sizeof(dname.server));
	dname.server[sizeof(dname.server) - 1] = '\0';


	ue_uew = NULL;
	total_uews = 0;


	/* Initialize the universe */
	UNVInit();

	/* Initialize the clipboard */
	ue_clipboard = UEClipboardNew();

	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* All else assume argument is a file path */
	    if((*arg != '-') && (*arg != '+'))
	    {
		char *path = STRDUP(arg);
		files_list = strlistappend(files_list, &nfiles, path);
		free(path);
	    }
	}


	/* Get the configuration file path */
	s = getenv("HOME");
	if(s != NULL)
	{
	    char *full_path = STRDUP(PrefixPaths(s, UE_DEF_CONFIG_FILE));
	    strncpy(fname.cfg, full_path, sizeof(fname.cfg));
	    fname.cfg[sizeof(fname.cfg) - 1] = '\0';
	    free(full_path);
	}
	else
	{
	    strncpy(fname.cfg, UE_DEF_CONFIG_FILE, sizeof(fname.cfg));
	    fname.cfg[sizeof(fname.cfg) - 1] = '\0';
	}

	/* Open the configuration file */
	if(UECfgFileOpen(fname.cfg))
	{
	    /* The configuration file does not exist, assume this is
	     * the first time running this program
	     */
	}


	/* Connect to the GUI */
	if(OSWGUIConnect(argc, argv))
	{
	    fprintf(
		stderr,
"Unable to connect to the GUI.\n"
	    );
	    strlistfree(files_list, nfiles);
	    free(cwd);
	    return(-1);
	}

	/* Initialize the widget set */
	if(WidgetInitGlobals(argc, argv))
	{
	    fprintf(
		stderr,
"Unable to initialize the widget set.\n"  
	    );
	    strlistfree(files_list, nfiles);
	    free(cwd);
	    return(-1);
	}

	/* Load the colors */
	OSWLoadPixelRGB(
	    &ue_color.view_bg, 0x00, 0x00, 0x00
	);
	ue_color.nview_grids = 3;
	if(ue_color.nview_grids > 0)
	{
	    const int n = ue_color.nview_grids;
	    float gamma;
	    ue_color.view_grid = (pixel_t *)malloc(n * sizeof(pixel_t));
	    for(i = 0; i < n; i++)
	    {
		gamma = (float)(i + 1) / (float)n;
		OSWLoadPixelRGB(
		    &ue_color.view_grid[i],
		    (u_int8_t)(0x80 * gamma),
		    (u_int8_t)(0x19 * gamma),
		    (u_int8_t)(0x78 * gamma)
		);
	    }
	}
	else
	{
	    ue_color.view_grid = NULL;
	}
	OSWLoadPixelRGB(
	    &ue_color.view_cursor, 0xF3, 0xB3, 0xFA
	);
	OSWLoadPixelRGB(
	    &ue_color.view_label, 0xFF, 0xB3, 0xFA
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_selected, 0xFF, 0xFF, 0x40
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_static, 0xF0, 0x20, 0xF8
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_dynamic, 0xF0, 0x20, 0xF8
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_controlled, 0xFF, 0x64, 0xF8
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_player, 0xFF, 0x64, 0xF8
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_weapon, 0xFF, 0x10, 0x00
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_home, 0x90, 0x80, 0xFF
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_area, 0xE0, 0x18, 0x98
	);
	OSWLoadPixelRGB(
	    &ue_color.view_obj_wormhole, 0x70, 0xB0, 0xFF
	);


	/* Load the fonts */
	OSWLoadFont(
	    &ue_font.object_label,
	    opt->object_label_font_name
	);
	OSWLoadFont( 
	    &ue_font.geometry_label,
	    opt->geometry_label_font_name
	);


	/* Load the cursors */
#define LOAD_WCURSOR(_wcur_,_xpm_data_,_hx_,_hy_,_r_,_g_,_b_)	{ \
 WColorStruct c;				\
 c.a = 0xff;					\
 c.r = (u_int8_t)(_r_);				\
 c.g = (u_int8_t)(_g_);				\
 c.b = (u_int8_t)(_b_);				\
 (_wcur_) = WidgetCreateCursorFromData(		\
  (_xpm_data_), (_hx_), (_hy_), &c, NULL	\
 );						\
}

	ue_cursor.busy = OSWCursorOpenID(XC_watch);
	LOAD_WCURSOR(
	    ue_cursor.scanner_lock_wcur,
	    xsw_scanner_lock_xpm,
	    8, 8,
	    0xFF, 0xB3, 0xFA
	);
	LOAD_WCURSOR(
	    ue_cursor.translate_wcur,
	    translate_xpm,
	    8, 8,
	    0xFF, 0xB3, 0xFA
	);
	LOAD_WCURSOR(
	    ue_cursor.zoom_wcur,
	    zoom_xpm,
	    8, 8,
	    0xFF, 0xB3, 0xFA
	);

#undef LOAD_WCURSOR

	/* Images */
#define LOAD_PIXMAP(_pp_,_mp_,_path_)	{	\
 char *s = STRDUP(PrefixPaths(			\
  ue_images_dir, (_path_)			\
 ));						\
 *(_pp_) = WidgetLoadPixmapFromTgaFile(		\
  s, (_mp_)					\
 );						\
 free(s);					\
}
#define LOAD_IMAGE(_pimg_,_path_)	{	\
 char *s = STRDUP(PrefixPaths(			\
  ue_images_dir, (_path_)			\
 ));						\
 *(_pimg_) = WidgetLoadImageFromTgaFile(s);	\
 free(s);					\
}

	ue_images_dir = STRDUP(PrefixPaths(dname.images, "unvedit"));

	LOAD_IMAGE(
	    &ue_image.icon_unvedit,
	    UE_FILE_ICON_UNVEDIT
	);
	LOAD_IMAGE( 
	    &ue_image.unvedit_logo,
	    UE_FILE_ICON_UNVEDIT_LOGO
	);
	ue_image.icon_info = WidgetLoadImageFromTgaData(info_tga);
	ue_image.icon_warning = WidgetLoadImageFromTgaData(warning_tga);
	ue_image.icon_error = WidgetLoadImageFromTgaData(error_tga);
	ue_image.icon_question = WidgetLoadImageFromTgaData(question_tga);

	LOAD_PIXMAP(
	    &ue_image.icon_about_pixmap,
	    &ue_image.icon_about_mask,
	    UE_FILE_ICON_ABOUT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_close_pixmap,
	    &ue_image.icon_close_mask,
	    UE_FILE_ICON_CLOSE
	);
	LOAD_PIXMAP(
	    &ue_image.icon_copy_pixmap,
	    &ue_image.icon_copy_mask,
	    UE_FILE_ICON_COPY
	);
	LOAD_PIXMAP(
	    &ue_image.icon_cut_pixmap,
	    &ue_image.icon_cut_mask,
	    UE_FILE_ICON_CUT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_delete_pixmap,
	    &ue_image.icon_delete_mask,
	    UE_FILE_ICON_DELETE
	);
	LOAD_PIXMAP(
	    &ue_image.icon_export_pixmap,
	    &ue_image.icon_export_mask,
	    UE_FILE_ICON_EXPORT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_economy_pixmap,
	    &ue_image.icon_economy_mask,
	    UE_FILE_ICON_ECONOMY
	);
	LOAD_PIXMAP(
	    &ue_image.icon_import_pixmap,
	    &ue_image.icon_import_mask,
	    UE_FILE_ICON_IMPORT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_legend_pixmap,
	    &ue_image.icon_legend_mask,
	    UE_FILE_ICON_LEGEND
	);
	LOAD_PIXMAP(
	    &ue_image.icon_new_pixmap,
	    &ue_image.icon_new_mask,
	    UE_FILE_ICON_NEW
	);
	LOAD_PIXMAP(
	    &ue_image.icon_new_object_pixmap,
	    &ue_image.icon_new_object_mask,
	    UE_FILE_ICON_NEW_OBJECT
	);
	LOAD_PIXMAP( 
	    &ue_image.icon_object_properties_pixmap,
	    &ue_image.icon_object_properties_mask,
	    UE_FILE_ICON_OBJECT_PROPERTIES
	);
	LOAD_PIXMAP(
	    &ue_image.icon_open_pixmap,
	    &ue_image.icon_open_mask,
	    UE_FILE_ICON_OPEN
	);
	LOAD_PIXMAP(
	    &ue_image.icon_paste_pixmap,
	    &ue_image.icon_paste_mask,
	    UE_FILE_ICON_PASTE
	);
	LOAD_PIXMAP(
	    &ue_image.icon_opm_file_pixmap,
	    &ue_image.icon_opm_file_mask,
	    UE_FILE_ICON_FILE_OPM
	);
	LOAD_PIXMAP(
	    &ue_image.icon_remodel_object_pixmap,
	    &ue_image.icon_remodel_object_mask,
	    UE_FILE_ICON_REMODEL_OBJECT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_print_pixmap,
	    &ue_image.icon_print_mask,
	    UE_FILE_ICON_PRINT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_save_pixmap,
	    &ue_image.icon_save_mask,
	    UE_FILE_ICON_SAVE
	);
	LOAD_PIXMAP(
	    &ue_image.icon_save_as_pixmap,
	    &ue_image.icon_save_as_mask,
	    UE_FILE_ICON_SAVE_AS
	);
	LOAD_PIXMAP(
	    &ue_image.icon_screenshot_pixmap,
	    &ue_image.icon_screenshot_mask,
	    UE_FILE_ICON_SCREENSHOT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_select_pixmap,
	    &ue_image.icon_select_mask,
	    UE_FILE_ICON_SELECT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_unvedit_pixmap,
	    &ue_image.icon_unvedit_mask,
	    UE_FILE_ICON_UNVEDIT
	);
	LOAD_PIXMAP(
	    &ue_image.icon_weapons_pixmap,
	    &ue_image.icon_weapons_mask,
	    UE_FILE_ICON_WEAPONS
	);

#undef LOAD_IMAGE

	free(ue_images_dir);


	/* Reset dialogs, they will be created on an as needed basis */
	ue_dlg = NULL;
	ue_cdlg = NULL;
	ue_file_sel = NULL;
	optwin = NULL;
	about_dlg = NULL;


	/* Any files specified at startup? */
	if(nfiles > 0)
	{
	    for(i = 0; i < nfiles; i++)
		uew_num = UECreateUEW(files_list[i]);
	}
	else
	{
	    /* Create the first universe edit window */
	    uew_num = UECreateUEW(NULL);
	}

	strlistfree(files_list, nfiles);
	free(cwd);

	return(0);
}

/*
 *	Manages one iteration of UnvEdit.
 */
void UEMainIteration(void *data)
{
	float tc;
	const time_t	new_ms = OSWTimeMS(),
			new_sec = time(NULL);
	time_t cur_ms, cur_sec;

	/* Update timming */
	if(new_ms < cur_millitime)
	{
	    UEResetTimmers(new_ms, new_sec);
	}
	cur_millitime = cur_ms = new_ms;
	cur_sec = new_sec;
	tc = 1.0f;

	/* Manage the GUI */
	UEManageGUI(data, cur_ms, cur_sec, tc);

	/* Maintain all the windows */
	UEMaintain(data, cur_ms, cur_sec, tc);

}

/*
 *	Manages UnvEdit.
 *
 *	Pushes an UnvEdit main loop and manages and blocks until
 *      UEMainPop*() is called.
 */
void UEMain(void *data)
{
	const int entering_runlevel = runlevel;
	runlevel++;
	while(runlevel > entering_runlevel)
	{
	    usleep(8000);
	    UEMainIteration(data);
	}
}

/*
 *	Pops off the highest UnvEdit main loop.
 */
void UEMainPop(void *data)
{
	if(runlevel > 0)
	    runlevel--;
}

/*
 *	Pops off all UnvEdit main loops.
 */
void UEMainPopAll(void *data)
{
	runlevel = 0;
}

/*
 *	Shuts down UnvEdit.
 */
void UEShutdown(void)
{
	int i;
	ue_option_struct *opt = &option;

	/* Save configuration first */
	if(!STRISEMPTY(fname.cfg))
	    UECfgFileSave(fname.cfg);

	/* Clipboard */
	UEClipboardDelete(ue_clipboard);
	ue_clipboard = NULL;

	/* File Selector */
	UEFileSelectorDelete(ue_file_sel);
	ue_file_sel = NULL;

	/* Dialogs */
	UEConfirmationDialogDelete(ue_cdlg);
	ue_cdlg = NULL;
	UEDialogDelete(ue_dlg);
	ue_dlg = NULL;

	/* Options window */
	OptWinDelete(optwin);
	optwin = NULL;


	/* Universe editor windows */
	for(i = 0; i < total_uews; i++)
	    UEWDelete(ue_uew[i]);
	free(ue_uew);
	ue_uew = NULL;
	total_uews = 0;

	/* About window */
	AboutDlgDelete(about_dlg);
	about_dlg = NULL;


	/* Images */
	OSWDestroyImage(&ue_image.icon_unvedit);
	OSWDestroyImage(&ue_image.unvedit_logo);
	OSWDestroyImage(&ue_image.icon_info); 
	OSWDestroyImage(&ue_image.icon_warning);
	OSWDestroyImage(&ue_image.icon_error);
	OSWDestroyImage(&ue_image.icon_question);

        OSWDestroyPixmap(&ue_image.icon_about_pixmap);
        OSWDestroyBitmap(&ue_image.icon_about_mask);
	OSWDestroyPixmap(&ue_image.icon_close_pixmap);
	OSWDestroyBitmap(&ue_image.icon_close_mask);
	OSWDestroyPixmap(&ue_image.icon_copy_pixmap);
	OSWDestroyBitmap(&ue_image.icon_copy_mask);
	OSWDestroyPixmap(&ue_image.icon_cut_pixmap);
	OSWDestroyBitmap(&ue_image.icon_cut_mask);
	OSWDestroyPixmap(&ue_image.icon_delete_pixmap);
	OSWDestroyBitmap(&ue_image.icon_delete_mask);
	OSWDestroyPixmap(&ue_image.icon_export_pixmap);
	OSWDestroyBitmap(&ue_image.icon_export_mask);
	OSWDestroyPixmap(&ue_image.icon_economy_pixmap);
	OSWDestroyBitmap(&ue_image.icon_economy_mask);
	OSWDestroyPixmap(&ue_image.icon_import_pixmap);
	OSWDestroyBitmap(&ue_image.icon_import_mask);
	OSWDestroyPixmap(&ue_image.icon_legend_pixmap);
	OSWDestroyBitmap(&ue_image.icon_legend_mask);
	OSWDestroyPixmap(&ue_image.icon_new_pixmap);
	OSWDestroyBitmap(&ue_image.icon_new_mask);
	OSWDestroyPixmap(&ue_image.icon_new_object_pixmap);
	OSWDestroyBitmap(&ue_image.icon_new_object_mask);
	OSWDestroyPixmap(&ue_image.icon_object_properties_pixmap);
	OSWDestroyBitmap(&ue_image.icon_object_properties_mask);
	OSWDestroyPixmap(&ue_image.icon_open_pixmap);
	OSWDestroyBitmap(&ue_image.icon_open_mask);
	OSWDestroyPixmap(&ue_image.icon_opm_file_pixmap);
	OSWDestroyBitmap(&ue_image.icon_opm_file_mask);
	OSWDestroyPixmap(&ue_image.icon_paste_pixmap);
	OSWDestroyBitmap(&ue_image.icon_paste_mask);
	OSWDestroyPixmap(&ue_image.icon_remodel_object_pixmap);
	OSWDestroyBitmap(&ue_image.icon_remodel_object_mask);
	OSWDestroyPixmap(&ue_image.icon_print_pixmap);
	OSWDestroyBitmap(&ue_image.icon_print_mask);
	OSWDestroyPixmap(&ue_image.icon_save_pixmap);
	OSWDestroyBitmap(&ue_image.icon_save_mask);
	OSWDestroyPixmap(&ue_image.icon_save_as_pixmap);
	OSWDestroyBitmap(&ue_image.icon_save_as_mask);
	OSWDestroyPixmap(&ue_image.icon_screenshot_pixmap);
	OSWDestroyBitmap(&ue_image.icon_screenshot_mask);
        OSWDestroyPixmap(&ue_image.icon_select_pixmap);
        OSWDestroyBitmap(&ue_image.icon_select_mask);
	OSWDestroyPixmap(&ue_image.icon_unvedit_pixmap);
	OSWDestroyBitmap(&ue_image.icon_unvedit_mask);
	OSWDestroyPixmap(&ue_image.icon_weapons_pixmap);
	OSWDestroyBitmap(&ue_image.icon_weapons_mask);

	/* Cursors */
	OSWDestroyCursor(&ue_cursor.busy);
	WidgetDestroyCursor(&ue_cursor.scanner_lock_wcur);
	WidgetDestroyCursor(&ue_cursor.translate_wcur);
	WidgetDestroyCursor(&ue_cursor.zoom_wcur);

	/* Fonts */
	OSWUnloadFont(&ue_font.object_label);
	OSWUnloadFont(&ue_font.geometry_label);

	/* Colors */
	OSWDestroyPixel(&ue_color.view_bg);
	for(i = 0; i < ue_color.nview_grids; i++)
	    OSWDestroyPixel(&ue_color.view_grid[i]);
	free(ue_color.view_grid);
	ue_color.view_grid = NULL;
	ue_color.nview_grids = 0;
	OSWDestroyPixel(&ue_color.view_cursor);
	OSWDestroyPixel(&ue_color.view_label);
	OSWDestroyPixel(&ue_color.view_obj_selected);
	OSWDestroyPixel(&ue_color.view_obj_static);
	OSWDestroyPixel(&ue_color.view_obj_dynamic);
	OSWDestroyPixel(&ue_color.view_obj_controlled);
	OSWDestroyPixel(&ue_color.view_obj_player);
	OSWDestroyPixel(&ue_color.view_obj_weapon);
	OSWDestroyPixel(&ue_color.view_obj_home);
	OSWDestroyPixel(&ue_color.view_obj_area);
	OSWDestroyPixel(&ue_color.view_obj_wormhole);


	/* Widget set */
	WidgetDestroyGlobals();

	/* Disconnect from the GUI */
	OSWGUIDisconnect();


	/* Shut down the universe */
	UNVShutdown();


	free(opt->object_label_font_name);
	opt->object_label_font_name = NULL;
	free(opt->geometry_label_font_name);
	opt->geometry_label_font_name = NULL;
}


int main(int argc, char **argv)
{
	int i, status;
	const char *arg;

	runlevel = 0;

	/* Set signal callbacks */
	signal(SIGINT, UESignalCB);
	signal(SIGTERM, UESignalCB);
	signal(SIGQUIT, UESignalCB);
	signal(SIGABRT, UESignalCB);
	signal(SIGKILL, UESignalCB);
	signal(SIGPIPE, UESignalCB);
	signal(SIGSEGV, UESignalCB);
	signal(SIGCONT, UESignalCB);

	/* Parse arguments for help and version */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!strcasecmp(arg, "--help") ||
	       !strcasecmp(arg, "-help") ||
	       !strcasecmp(arg, "--h") ||
	       !strcasecmp(arg, "-h") ||
	       !strcasecmp(arg, "-?")
	    )
	    {
		UEPrintHelpStdout();
		return(0);
	    }
	    /* Version */
	    else if(!strcasecmp(arg, "--version") ||
		    !strcasecmp(arg, "-version")
	    )
	    {
		UEPrintVersionStdout();
		return(0);
	    }
	}

	/* Initialize */
	status = UEInit(argc, argv);
	if(status < 0)
	{
	    UEShutdown();
	    return(-status);
	}

	/* Manage */
	UEMain(NULL);

	/* Shut down */
	UEShutdown();

	return(0);
}
