#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/imgrefs.h"
#include "../include/reality.h"
#include "../include/objects.h"
#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "blitting.h"
#include "propdlg.h"  
#include "uew.h"
#include "ue.h"

#include "../include/oswkeyvals.h"


static void PropDlgGetValues(prop_dlg_struct *d);
static void PropDlgSetValues(prop_dlg_struct *d);
static void PropDlgSetImgRef(prop_dlg_struct *d, const int imgref_num);

static void PropDlgDrawCB(void *widget, void *data);
static void PropDlgTypeListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
);
static void PropDlgImgRefsListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
);
static void PropDlgOKCB(push_button_struct *btn, void *data);
static void PropDlgApplyCB(push_button_struct *btn, void *data);
static void PropDlgCancelCB(push_button_struct *btn, void *data);

prop_dlg_struct *PropDlgNew(uew_struct *uew, const int obj_num);
void PropDlgDraw(prop_dlg_struct *d, const int amount);
void PropDlgQueueDraw(prop_dlg_struct *d, const int amount);
int PropDlgManage(prop_dlg_struct *d, event_t *event);
void PropDlgMaintain(
	void *data, prop_dlg_struct *d,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void PropDlgUpdateWidgets(prop_dlg_struct *d);
void PropDlgMap(prop_dlg_struct *d);
void PropDlgMapValues(prop_dlg_struct *d);
void PropDlgUnmap(prop_dlg_struct *d);
void PropDlgDelete(prop_dlg_struct *d);


#define PROP_DLG_TITLE			"Properties"
#define PROP_DLG_ICON_TITLE		"Properties"

#define PROP_DLG_WIDTH			640
#define PROP_DLG_HEIGHT			480

#define PROP_DLG_MARGIN			10

#define PROP_DLG_PROMPT_HEIGHT		30
#define PROP_DLG_PULIST_HEIGHT		26

#define PROP_DLG_ICON_BTN_WIDTH		26
#define PROP_DLG_ICON_BTN_HEIGHT	26

#define PROP_DLG_BTN_WIDTH		70
#define PROP_DLG_BTN_HEIGHT		28

#define PROP_DLG_OBJECT_DISPLAY_WIDTH		150
#define PROP_DLG_OBJECT_DISPLAY_HEIGHT		120


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Gets the values from the referenced Universe Editor Window.
 */
static void PropDlgGetValues(prop_dlg_struct *d)
{
	int i, n, obj_num;
	const char *name;
	char *s;
	imgref_struct *imgref;
	popup_list_struct *pulist;
	xsw_object_struct *obj;
	xsw_score_struct *score;
	uew_struct *uew;

	if(d == NULL)
	    return;

	obj_num = d->obj_num;
	uew = d->uew;
	obj = UEWGetObjectPtr(uew, obj_num);
	if(obj == NULL)
	    return;

	d->freeze_count++;

	score = obj->score;

	/* Begin getting values */

	PromptSetS(&d->name_prompt, obj->name);

	pulist = &d->type_pulist;
	PUListSetSelItem(
	    pulist,
	    PUListGetItemByData(pulist, (void *)obj->type)
	);

	PromptSetS(&d->empire_prompt, obj->empire);
	PromptSetI(&d->uid_prompt, obj->permission.uid);
	PromptSetI(&d->gid_prompt, obj->permission.gid);
	s = (char *)malloc(40);
	sprintf(
	    s,
	    "#%i",
	    obj->owner
	);
	PromptSetS(&d->owner_prompt, s);
	free(s);

	PromptSetF2(
	    &d->credits_prompt,
	    (score != NULL) ? MAX(score->credits, 0.0f) : 0.0f,
	    2
	);
	PromptSetF2(
	    &d->rmu_prompt,
	    (score != NULL) ? MAX(score->rmu, 0.0f) : 0.0f,
	    2
	);
	PromptSetF2(
	    &d->rmu_max_prompt,
	    (score != NULL) ? MAX(score->rmu_max, 0.0f) : 0.0f,
	    2
	);
	PromptSetL(&d->size_prompt, obj->size);
	PromptSetF2(
	    &d->velocity_max_prompt,
	    MAX(obj->velocity_max *
		(1000.0f / (float)SW_CYCLE_LAPSE_MS), 0.0f),
	    6
	);
	PromptSetF2(&d->visibility_prompt, obj->visibility, 2);
	PromptSetF2(&d->hp_prompt, obj->hp, 2);
	PromptSetF2(&d->hp_max_prompt, obj->hp_max, 2);
	PromptSetF2(&d->power_prompt, obj->power, 2);
	PromptSetF2(&d->power_max_prompt, obj->power_max, 2);
	PromptSetF2(&d->power_purity_prompt, obj->power_purity, 2);
	PromptSetF2(
	    &d->core_efficency_prompt,
	    MAX(obj->core_efficency *
		(1000.0f / (float)SW_CYCLE_LAPSE_MS), 0.0f),
	    6
	);
	PromptSetF2(&d->scanner_range_prompt, obj->scanner_range, 2);

	pulist = &d->engine_pulist;
	PUListSetSelItem(
	    pulist,
	    PUListGetItemByData(pulist, (void *)obj->engine_state)
	);
	PromptSetF2(&d->throttle_prompt, obj->throttle, 2);
	PromptSetF2(&d->antimatter_prompt, obj->antimatter, 2);
	PromptSetF2(&d->antimatter_max_prompt, obj->antimatter_max, 2);
	PromptSetF2(
	    &d->turn_rate_prompt,
	    MAX(obj->turnrate * (1000.0f / (float)SW_CYCLE_LAPSE_MS) *
		(180.0f / PI), 0.0f),
	    2
	);

	pulist = &d->shields_pulist;
	PUListSetSelItem(
	    pulist,
	    PUListGetItemByData(pulist, (void *)obj->shield_state)
	);
	pulist = &d->cloak_pulist;
	PUListSetSelItem(
	    pulist,
	    PUListGetItemByData(pulist, (void *)obj->cloak_state)
	);
	PromptSetF2(&d->cloak_strength_prompt, obj->cloak_strength, 2);
	PromptSetL(&d->cloak_refractory_int_prompt, obj->cloak_refractory_int);

	CBtnSetActive(
	    &d->vector_lights_cbtn,
	    (obj->lighting & SW_OBJ_LIGHT_VECTOR) ? True : False
	);
	CBtnSetActive(
	    &d->strobes_cbtn, 
	    (obj->lighting & SW_OBJ_LIGHT_STROBE) ? True : False
	);
	CBtnSetActive(
	    &d->external_lumination_cbtn,
	    (obj->lighting & SW_OBJ_LIGHT_LUMINATION) ? True : False
	);

	CBtnSetActive(
	    &d->follow_friend_cbtn,
	    (obj->ai_flags & SW_AI_FOLLOW_FRIEND) ? True : False
	);
	CBtnSetActive(
	    &d->follow_unknown_cbtn,
	    (obj->ai_flags & SW_AI_FOLLOW_UNKNOWN) ? True : False
	);
	CBtnSetActive(
	    &d->follow_hostile_cbtn,
	    (obj->ai_flags & SW_AI_FOLLOW_HOSTILE) ? True : False
	);
	CBtnSetActive(
	    &d->fire_friend_cbtn,
	    (obj->ai_flags & SW_AI_FIRE_FRIEND) ? True : False
	);
	CBtnSetActive(
	    &d->fire_unknown_cbtn,
	    (obj->ai_flags & SW_AI_FIRE_UNKNOWN) ? True : False
	);
	CBtnSetActive(
	    &d->fire_hostile_cbtn,
	    (obj->ai_flags & SW_AI_FIRE_HOSTILE) ? True : False
	);


	/* Get the ImgRefs listing from the Universe Editor */
	pulist = &d->imgrefs_pulist;
	PUListClear(pulist);
	for(i = 0; i < uew->total_imgrefs; i++)
	{
	    imgref = uew->imgref[i];
	    if(imgref == NULL)
		continue;

	    if(!STRISEMPTY(imgref->name))
	    {
		s = STRDUP(imgref->name);
	    }
	    else
	    {
		s = (char *)malloc(40 + 80);
		sprintf(
		    s,
		    "ImgRef #%i",
		    i
		);
	    }     
	    n = PUListAddItem(
		pulist,
		0,
		s,
		None, None
	    );
	    free(s);

	    /* Set the ImgRef's number as the data */
	    PUListItemSetData(
		pulist, n,
		(void *)i, NULL
	    );
	}

	/* Display the object's image */
	PropDlgSetImgRef(d, obj->imgref);






	/* Update the Properties Dialog's title */
	name = UNVGetObjectFormalName(obj_num, obj);
	s = (char *)malloc(
	    STRLEN(PROP_DLG_TITLE) + STRLEN(name) + 80
	);
	sprintf(
	    s,
	    "%s: %s",
	    PROP_DLG_TITLE, name
	);
	OSWSetWindowTitle(d->toplevel, s);
	free(s);

	PropDlgUpdateWidgets(d);

	d->freeze_count--;
}

/*
 *	Set values.
 */
static void PropDlgSetValues(prop_dlg_struct *d)
{
	int obj_num;
	popup_list_struct *pulist;
	xsw_object_struct *obj;
	uew_struct *uew;

	if(d == NULL)
	    return;

	obj_num = d->obj_num;
	uew = d->uew;
	obj = UEWGetObjectPtr(uew, obj_num);
	if(obj == NULL)
	    return;

	/* If this object is currently displayed on the universe editor
	 * then set the existing property prompt values first
	 */
	if(uew->last_selected_object == obj_num)
	    UEWPropsDoSetValues(uew, obj_num);

	/* Begin setting values */
#define SET_STRING(_tar_string_,_src_prompt_)	{	\
 const char *s = PromptGetS(_src_prompt_);		\
 if(s != NULL) {					\
  strncpy((_tar_string_), s, sizeof(_tar_string_));	\
  (_tar_string_)[sizeof(_tar_string_) - 1] = '\0';	\
 }							\
}
	SET_STRING(obj->name, &d->name_prompt);
	pulist = &d->type_pulist;
	obj->type = (int)PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	);
	if(obj->type <= SW_OBJ_TYPE_GARBAGE)
	{
	    char *s = (char *)malloc(
		STRLEN(SW_OBJ_TYPE_NAME_GARBAGE) +
		(2 * 80)
	    );
	    sprintf(
		s,
"Setting the object to type %s\n\
effectively deletes the object.",
		SW_OBJ_TYPE_NAME_GARBAGE
	    );
	    UEDialog(
		"Setting Object Type",
		UE_ICON_WARNING,
		s,
		uew->toplevel
	    );
	    free(s);
	}
	SET_STRING(obj->empire, &d->empire_prompt);
	obj->permission.uid = PromptGetI(&d->uid_prompt);
	obj->permission.gid = PromptGetI(&d->gid_prompt);

	if(!WidgetDisabled(&d->owner_prompt))
	{
	    int i;
	    char *name = STRDUP(PromptGetS(&d->owner_prompt));
	    clist_struct *clist = &uew->objects_list;
	    xsw_object_struct *obj2;

	    if(UNVIsObjectNameIndex(name))
	    {
		const char *num_str = name;

		/* Skip leading '#' characters */
		while(*num_str == '#')
		    num_str++;

		obj->owner = ATOI(num_str);
		if((obj->owner < 0) || (obj->owner >= clist->total_rows))
		{
		    char *s = (char *)malloc(40 + (2 * 80));
		    sprintf(
			s,
"The specified owner `#%i'\n\
is invalid.",
			obj->owner
		    );
		    UEDialog(
			"Invalid Owner",
			UE_ICON_WARNING,
			s,
			uew->toplevel
		    );
		    free(s);
		    obj->owner = uew->legend.lost_found_owner;
		}
	    }
	    else if(name != NULL)
	    {
		for(i = 0; i < clist->total_rows; i++)
		{
		    obj2 = SW_OBJECT(CListRowGetData(clist, i));
		    if(obj2 == NULL)
			continue;

		    if(strcasestr(obj2->name, name) != NULL)
		    {
			obj->owner = i;
			break;
		    }
		}
		if(i >= clist->total_rows)
		{
		    char *s = (char *)malloc(
			STRLEN(name) + (2 * 80)
		    );
		    sprintf(
"The specified owner `%s'\n\
is invalid.",
			name
		    );
		    UEDialog(
			"Invalid Owner",
			UE_ICON_WARNING,
			s,
			uew->toplevel
		    );
		    free(s);
		    obj->owner = uew->legend.lost_found_owner;
		}
	    }

	    free(name);
	}

	if(PromptGetF(&d->credits_prompt) > 0.0f)
	{
	    if(!UNVAllocScores(obj))
		obj->score->credits = MAX(
		    PromptGetF(&d->credits_prompt), 0.0f
		);
	}
	if(PromptGetF(&d->rmu_prompt) > 0.0f)
	{
	    if(!UNVAllocScores(obj))
		obj->score->rmu = MAX(
		    PromptGetF(&d->rmu_prompt), 0.0f
		);
	}
	if(PromptGetF(&d->rmu_max_prompt) > 0.0f)
	{
	    if(!UNVAllocScores(obj))
		obj->score->rmu_max = MAX(
		    PromptGetF(&d->rmu_max_prompt), 0.0f
		);
	}
	obj->size = MAX(PromptGetL(&d->size_prompt), 0l);
	obj->velocity_max = MAX(
	    PromptGetF(&d->velocity_max_prompt) *
		((float)SW_CYCLE_LAPSE_MS / 1000), 0.0f
	);
	obj->visibility = CLIP(
	    PromptGetF(&d->visibility_prompt), 0.0f, 1.0f
	);
	obj->hp = PromptGetF(&d->hp_prompt);
	obj->hp_max = PromptGetF(&d->hp_max_prompt);
	obj->power = PromptGetF(&d->power_prompt);
	obj->power_max = PromptGetF(&d->power_max_prompt);
	obj->power_purity = CLIP(
	    PromptGetF(&d->power_purity_prompt), 0.0f, 1.0f
	);
	obj->core_efficency = MAX(
	    (PromptGetF(&d->core_efficency_prompt) *
		((float)SW_CYCLE_LAPSE_MS / 1000.0f)
	    ), 0.0f
	);
	obj->scanner_range = MAX(PromptGetF(&d->scanner_range_prompt), 0.0f);

	pulist = &d->engine_pulist;
	obj->engine_state = (int)PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	);
	obj->throttle = CLIP(
	    PromptGetF(&d->throttle_prompt), -1.0f, 1.0f
	);
	obj->antimatter = MAX(
	    PromptGetF(&d->antimatter_prompt), 0.0f
	);
	obj->antimatter_max = MAX(
	    PromptGetF(&d->antimatter_max_prompt), 0.0f
	);
	obj->turnrate = MAX(
	    (PromptGetF(&d->turn_rate_prompt) * 
	     ((float)SW_CYCLE_LAPSE_MS / 1000.0f) *
	     (PI / 180.0f)  
	    ),
	    0.0f
	);

	pulist = &d->shields_pulist;
	obj->shield_state = (int)PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	);
	pulist = &d->cloak_pulist;
	obj->cloak_state = (int)PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	);
	obj->cloak_strength = MAX(
	    PromptGetF(&d->cloak_strength_prompt), 0.0f
	);
	obj->cloak_refractory_int = MAX(
	    PromptGetL(&d->cloak_refractory_int_prompt), 0l
	);

	obj->lighting = 0;
	if(CBtnGetActive(&d->vector_lights_cbtn))
	    obj->lighting |= SW_OBJ_LIGHT_VECTOR;
	if(CBtnGetActive(&d->strobes_cbtn))
	    obj->lighting |= SW_OBJ_LIGHT_STROBE;
	if(CBtnGetActive(&d->external_lumination_cbtn))
	    obj->lighting |= SW_OBJ_LIGHT_LUMINATION;

	obj->ai_flags = 0;
	if(CBtnGetActive(&d->follow_friend_cbtn))
	    obj->ai_flags |= SW_AI_FOLLOW_FRIEND;
	if(CBtnGetActive(&d->follow_unknown_cbtn))
	    obj->ai_flags |= SW_AI_FOLLOW_UNKNOWN;
	if(CBtnGetActive(&d->follow_hostile_cbtn))
	    obj->ai_flags |= SW_AI_FOLLOW_HOSTILE;
	if(CBtnGetActive(&d->fire_friend_cbtn))
	    obj->ai_flags |= SW_AI_FIRE_FRIEND;
	if(CBtnGetActive(&d->fire_unknown_cbtn))
	    obj->ai_flags |= SW_AI_FIRE_UNKNOWN;
	if(CBtnGetActive(&d->fire_hostile_cbtn))
	    obj->ai_flags |= SW_AI_FIRE_HOSTILE;



	pulist = &d->imgrefs_pulist;
	obj->imgref = (int)PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	);

#undef SET_STRING

	/* If this object is currently displayed on the universe editor
	 * then update the displayed values
	 */
	if(uew->last_selected_object == obj_num)
	{
	    /* Update the display */
	    UEWDisplayObject(uew, obj_num, False);
	}

	/* Reset has changes marker on uhw */
	d->has_changes = False;

	/* Set has changes mark on the Universe Editor Window */
	if(!uew->has_changes)
	{
	    uew->has_changes = True;
	    UEWUpdateTitle(uew);
	}

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWDoUpdateWindowMenus();
}

/*
 *	Sets the ImgRef.
 *
 *	Loads the ImgRef's image as needed.
 *
 *	Updates the Properties Dialog's ImgRef animation settings.
 *
 *	Updates the Properties Dialog's ImgRefs List.
 *
 *	Draws the Properties Dialog's object display.
 */
static void PropDlgSetImgRef(prop_dlg_struct *d, const int imgref_num)
{
	int obj_num;
	popup_list_struct *pulist;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	uew_struct *uew;
	if(d == NULL)
	    return;

	obj_num = d->obj_num;
	uew = d->uew;
	if(uew == NULL)
	    return;

	obj = UEWGetObjectPtr(uew, obj_num);
	if(obj == NULL)
	    return;

	/* Get the specified imgref from the Universe Editor Window */
	imgref = ((imgref_num >= 0) && (imgref_num < uew->total_imgrefs)) ?
	    uew->imgref[imgref_num] : NULL;

	/* Load the new imgref */
	ImgRefLoad(imgref);

	/* Update animation values */
	if(obj->type == SW_OBJ_TYPE_ANIMATED)
	{
	    xsw_animation_struct *anim = &obj->animation;

	    d->ani_frame = 0;
	    d->ani_total_frames = (imgref != NULL) ?
		MAX(imgref->total_frames, 1) : 1;
	    d->ani_int = MAX(anim->interval, 10);
	    d->ani_next = cur_millitime + d->ani_int;
	}
	else
	{
	    d->ani_frame = 0;
	    d->ani_total_frames = (imgref != NULL) ?
		MAX(imgref->total_frames, 1) : 1;
	    d->ani_int = 500l;
	    d->ani_next = cur_millitime + d->ani_int;
	}

	/* Update the displayed item on the ImgRefs Popup List */
	d->freeze_count++;
	pulist = &d->imgrefs_pulist;
	PUListSetSelItem(
	    pulist,
	    PUListGetItemByData(pulist, (void *)imgref_num)
	);
	d->freeze_count--;

	if(d->map_state)
	    PropDlgQueueDraw(d, PROP_DLG_DRAW_OBJECT_DISPLAY);
}


/*
 *	Draw callback.
 */
static void PropDlgDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	prop_dlg_struct *d = PROP_DLG(widget);
	if(d == NULL)
	    return;

	if(d->map_state)
	    PropDlgDraw(d, amount);
}


/*
 *	Type popup list select callback.
 */
static void PropDlgTypeListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
)
{
	prop_dlg_struct *d = PROP_DLG(data);
	if((pulist == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	PropDlgUpdateWidgets(d);
}

/*
 *	ImgRefs popup list select callback.
 */
static void PropDlgImgRefsListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
)
{
	int imgref_num;
	prop_dlg_struct *d = PROP_DLG(data);
	if((pulist == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	imgref_num = (int)PUListItemGetData(
	    pulist, sel_item
	);

	PropDlgSetImgRef(d, imgref_num);
}

/*
 *	OK callback.
 */
static void PropDlgOKCB(push_button_struct *btn, void *data)
{
	prop_dlg_struct *d = PROP_DLG(data);
	if(d == NULL)
	    return;

	PropDlgSetValues(d);
	PropDlgUnmap(d);
}

/*
 *	Apply callback.
 */
static void PropDlgApplyCB(push_button_struct *btn, void *data)
{
	prop_dlg_struct *d = PROP_DLG(data);
	if(d == NULL)
	    return;

	PropDlgSetValues(d);
}

/*
 *	Cancel callback.
 */
static void PropDlgCancelCB(push_button_struct *btn, void *data)
{ 
	prop_dlg_struct *d = PROP_DLG(data);
	if(d == NULL)
	    return;

	PropDlgUnmap(d);
}


/*
 *	Creates a new Properties Dialog.
 */
prop_dlg_struct *PropDlgNew(uew_struct *uew, const int obj_num)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	push_button_struct *btn;
	popup_list_struct *pulist;
	widget_global_struct *wglobal = &widget_global;
	prop_dlg_struct *d = PROP_DLG(calloc(1, sizeof(prop_dlg_struct)));

	d->map_state = False;
	d->has_focus = False;
	d->freeze_count = 0;
	d->x = 0;
	d->y = 0;
	d->width = PROP_DLG_WIDTH;
	d->height = PROP_DLG_HEIGHT;

	d->uew = uew;
	d->obj_num = obj_num;

	d->ani_frame = 0;
	d->ani_total_frames = 0;
	d->ani_int = 0l;
	d->ani_next = 0l;

	d->has_changes = False;


	/* Create toplevel */
	x = d->x;
	y = d->y;
	width = d->width;
	height = d->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	d->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    PROP_DLG_TITLE,
	    PROP_DLG_ICON_TITLE,
	    (ue_image.icon_unvedit_pixmap != None) ?
		ue_image.icon_unvedit_pixmap :
		wglobal->std_icon_pm,
	    True,			/* Let WM set coordinates? */
	    x, y,
	    width, height,
	    width, height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	if(uew != NULL)
	    OSWSetTransientFor(w, uew->toplevel);

	parent = w;


	/* Name */
	x = PROP_DLG_MARGIN;
	y = PROP_DLG_MARGIN;
	width = 300;
	height = PROP_DLG_PROMPT_HEIGHT;
	PromptInit(
	    &d->name_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Name:",
	    SW_OBJ_NAME_MAX, 0
	);

	/* Type */
	x += (int)width + PROP_DLG_MARGIN;
	LabelInit(
	    &d->type_label, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    0, PROP_DLG_PULIST_HEIGHT,
	    "Type:"
	);
	x += (int)d->type_label.width;
	pulist = &d->type_pulist;
	PUListInit(
	    pulist, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    100, PROP_DLG_PULIST_HEIGHT,
	    0,
	    PULIST_POPUP_BELOW
	);
#define ADD_ITEM(_ts_,_ti_)	{	\
 const int i = PUListAddItem(		\
  pulist, 0, (_ts_), None, None		\
 );					\
 PUListItemSetData(			\
  pulist, i, (void *)(_ti_), NULL	\
 );					\
}
	ADD_ITEM(SW_OBJ_TYPE_NAME_GARBAGE, SW_OBJ_TYPE_GARBAGE);
	ADD_ITEM(SW_OBJ_TYPE_NAME_STATIC, SW_OBJ_TYPE_STATIC);
	ADD_ITEM(SW_OBJ_TYPE_NAME_DYNAMIC, SW_OBJ_TYPE_DYNAMIC);
	ADD_ITEM(SW_OBJ_TYPE_NAME_CONTROLLED, SW_OBJ_TYPE_CONTROLLED);
	ADD_ITEM(SW_OBJ_TYPE_NAME_PLAYER, SW_OBJ_TYPE_PLAYER);
	ADD_ITEM(SW_OBJ_TYPE_NAME_WEAPON, SW_OBJ_TYPE_WEAPON);
	ADD_ITEM(SW_OBJ_TYPE_NAME_STREAMWEAPON, SW_OBJ_TYPE_STREAMWEAPON);
	ADD_ITEM(SW_OBJ_TYPE_NAME_SPHEREWEAPON, SW_OBJ_TYPE_SPHEREWEAPON);
	ADD_ITEM(SW_OBJ_TYPE_NAME_HOME, SW_OBJ_TYPE_HOME);
	ADD_ITEM(SW_OBJ_TYPE_NAME_AREA, SW_OBJ_TYPE_AREA);
	ADD_ITEM(SW_OBJ_TYPE_NAME_ANIMATED, SW_OBJ_TYPE_ANIMATED);
	ADD_ITEM(SW_OBJ_TYPE_NAME_WORMHOLE, SW_OBJ_TYPE_WORMHOLE);
	ADD_ITEM(SW_OBJ_TYPE_NAME_ELINK, SW_OBJ_TYPE_ELINK);
#undef ADD_ITEM
	pulist->select_cb = PropDlgTypeListSelectCB;
	pulist->select_data = d;

	/* Empire */
	x = PROP_DLG_MARGIN;
	y += (int)height;
	width = 120;
	PromptInit(
	    &d->empire_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Empire:",
	    SW_OBJ_EMPIRE_MAX, 0
	);

	/* UID */
	x += (int)width + PROP_DLG_MARGIN;
	width = 80;
	PromptInit(
	    &d->uid_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED, 
	    "UID:",
	    40, 0
	);

	/* GID */
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 80;
	PromptInit(
	    &d->gid_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED, 
	    "GID:",
	    40, 0
	);

	/* Owner */
	x += (int)width + PROP_DLG_MARGIN;
	width = 140;
	PromptInit(
	    &d->owner_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Owner:",
	    MAX(SW_OBJ_NAME_MAX, 40), 0
	);


	/* Credits */
	x = PROP_DLG_MARGIN;
	y += (int)height + (2 * PROP_DLG_MARGIN);
	width = 140;
	PromptInit(
	    &d->credits_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Credits:",
	    40, 0
	);

	/* RMU */
	x += (int)width + PROP_DLG_MARGIN;
	width = 100;
	PromptInit(
	    &d->rmu_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "RMU:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->rmu_prompt,
"Enter the units of raw materials on the object"
	);

	/* RMU Max */
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 100;
	PromptInit( 
	    &d->rmu_max_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Max:",
	    40, 0  
	);
	PromptSetTooltip(
	    &d->rmu_max_prompt,
"Enter the object's maximum capacity for raw materials"
	);


	/* Size */
	x = PROP_DLG_MARGIN;
	y += (int)height;
	width = 100;
	PromptInit(
	    &d->size_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Size:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->size_prompt,
"Enter the size of the object in screen units"
	);

	/* Velocity Max */
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 160;
	PromptInit(
	    &d->velocity_max_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Velocity Max:",
	    40, 0  
	);
	PromptSetTooltip(
	    &d->velocity_max_prompt,
"Enter the maximum velocity in real units per second of the object"
	);

	/* Visibility */
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 100;
	PromptInit(
	    &d->visibility_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Visibility:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->visibility_prompt,
"Enter the visibility coefficient of the object,\
 where 0.0 is invisible and 1.0 is fully visible"
	);


	/* HP */
	x = PROP_DLG_MARGIN;
	y += (int)height + (2 * PROP_DLG_MARGIN);
	width = 95;
	PromptInit(
	    &d->hp_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "HP:",
	    40, 0
	);
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 100;
	PromptInit(
	    &d->hp_max_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Max:",
	    40, 0
	);

	/* Power */
	x += (int)width + PROP_DLG_MARGIN;
	width = 120;
	PromptInit(
	    &d->power_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Power:",
	    40, 0
	);
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 110;
	PromptInit(
	    &d->power_max_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Max:",
	    40, 0
	);

	/* Power Purity */
	x = PROP_DLG_MARGIN;
	y += (int)height;
	width = 130;
	PromptInit(
	    &d->power_purity_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Power Purity:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->power_purity_prompt,
"Enter the purity coefficient of the object's power,\
 where 0.0 is impure and 1.0 is purest"
	);

	/* Core Efficiency */
	x += (int)width + PROP_DLG_MARGIN;
	width = 170;
	PromptInit(
	    &d->core_efficency_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Core Efficiency:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->core_efficency_prompt,
"Enter the rate in power units per second at which the object's\
 power core regenerates power (enter 0 for no power regeneration)"
	);

	/* Scanner Range */
	x += (int)width + PROP_DLG_MARGIN;
	width = 170;
	PromptInit(
	    &d->scanner_range_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Scanner Range:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->scanner_range_prompt,
"Enter the object's scanner range in real units"
	);


	/* Engine */
	x = PROP_DLG_MARGIN + 6;
	y += (int)height + (2 * PROP_DLG_MARGIN);
	LabelInit(
	    &d->engine_label, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    0, PROP_DLG_PULIST_HEIGHT,
	    "Engine:"
	);
	x += (int)d->engine_label.width;
	pulist = &d->engine_pulist;
	PUListInit(
	    pulist, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    80, PROP_DLG_PULIST_HEIGHT,
	    4,
	    PULIST_POPUP_BELOW  
	);
#define ADD_ITEM(_ts_,_ti_)	{	\
 const int i = PUListAddItem(		\
  pulist, 0, (_ts_), None, None		\
 );					\
 PUListItemSetData(			\
  pulist, i, (void *)(_ti_), NULL	\
 );					\
}
	ADD_ITEM(SW_ENGINE_STATE_STR_NONE, SW_ENGINE_STATE_NONE);
	ADD_ITEM(SW_ENGINE_STATE_STR_OFF, SW_ENGINE_STATE_OFF);
	ADD_ITEM(SW_ENGINE_STATE_STR_STARTING, SW_ENGINE_STATE_STARTING);
	ADD_ITEM(SW_ENGINE_STATE_STR_ON, SW_ENGINE_STATE_ON);
#undef ADD_ITEM

	/* Throttle */
	x += (int)pulist->width + PROP_DLG_MARGIN;
	width = 95;
	PromptInit(
	    &d->throttle_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Throttle:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->throttle_prompt,
"Enter the coefficient of the engine's throttle (0.0 is idle, 1.0 is\
 full, and -1.0 is full reverse"
	);

	/* Antimatter */
	x += (int)width + PROP_DLG_MARGIN;
	width = 140;
	PromptInit(
	    &d->antimatter_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Antimatter:",
	    40, 0
	);
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 110;
	PromptInit(
	    &d->antimatter_max_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Max:",
	    40, 0 
	);
	  
	/* Turn Rate */
	x += (int)width + PROP_DLG_MARGIN;
	width = 80;
	PromptInit(
	    &d->turn_rate_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Turn:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->turn_rate_prompt,
"Enter the maximum turn rate of the object in degrees per second"
	);


	/* Shields */
	x = PROP_DLG_MARGIN + 6;
	y += (int)height + (PROP_DLG_MARGIN / 2);
	LabelInit(
	    &d->shields_label, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    0, PROP_DLG_PULIST_HEIGHT,
	    "Shields:"
	);
	x += (int)d->shields_label.width;
	pulist = &d->shields_pulist;
	PUListInit(
	    pulist, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    70, PROP_DLG_PULIST_HEIGHT,
	    3,
	    PULIST_POPUP_BELOW  
	);
#define ADD_ITEM(_ts_,_ti_)	{	\
 const int i = PUListAddItem(		\
  pulist, 0, (_ts_), None, None		\
 );					\
 PUListItemSetData(			\
  pulist, i, (void *)(_ti_), NULL	\
 );					\
}
	ADD_ITEM(SW_SHIELD_STATE_STR_NONE, SW_SHIELD_STATE_NONE);
	ADD_ITEM(SW_SHIELD_STATE_STR_DOWN, SW_SHIELD_STATE_DOWN);
	ADD_ITEM(SW_SHIELD_STATE_STR_UP, SW_SHIELD_STATE_UP);
#undef ADD_ITEM

	/* Cloak */
	x += (int)pulist->width + PROP_DLG_MARGIN;
	LabelInit(
	    &d->cloak_label, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    0, PROP_DLG_PULIST_HEIGHT,
	    "Cloak:"
	);
	x += (int)d->cloak_label.width;
	pulist = &d->cloak_pulist;
	PUListInit(
	    pulist, parent,
	    x, y + (((int)height - PROP_DLG_PULIST_HEIGHT) / 2),
	    70, PROP_DLG_PULIST_HEIGHT,
	    3,
	    PULIST_POPUP_BELOW  
	);
#define ADD_ITEM(_ts_,_ti_)	{	\
 const int i = PUListAddItem(		\
  pulist, 0, (_ts_), None, None		\
 );					\
 PUListItemSetData(			\
  pulist, i, (void *)(_ti_), NULL	\
 );					\
}
	ADD_ITEM(SW_CLOAK_STATE_STR_NONE, SW_CLOAK_STATE_NONE);
	ADD_ITEM(SW_CLOAK_STATE_STR_DOWN, SW_CLOAK_STATE_DOWN);
	ADD_ITEM(SW_CLOAK_STATE_STR_UP, SW_CLOAK_STATE_UP);
#undef ADD_ITEM

	/* Cloak Strength */
	x += (int)pulist->width + (PROP_DLG_MARGIN / 2);
	width = 110;
	PromptInit(
	    &d->cloak_strength_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Strength:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->cloak_strength_prompt,
"Enter the strength coefficient of the object's cloak, where\
 0.0 is powerless and 1.0 is complete effectiveness, the value\
 may be greater than 1.0 for an effective cloak under low power\
 conditions"
	);

	/* Cloak Refractory Int */
	x += (int)width + (PROP_DLG_MARGIN / 2);
	width = 140;
	PromptInit(
	    &d->cloak_refractory_int_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Refractory Int:",
	    40, 0
	);
	PromptSetTooltip(
	    &d->cloak_refractory_int_prompt,
"Enter the delay in milliseconds at which cloak-dependent\
 weapons may not be used immediately after decloaking (enter 0\
 for no delay)"
	);

	/* Vector Lights */
	x = PROP_DLG_MARGIN;
	y += (int)height + (2 * PROP_DLG_MARGIN);
	CBtnInit(
	    &d->vector_lights_cbtn, parent,
	    x, y,
	    False,
	    "Vector Lights"
	);
	width = d->vector_lights_cbtn.width;

	/* Strobes */
	x += (int)width + (PROP_DLG_MARGIN / 2);
	CBtnInit(
	    &d->strobes_cbtn, parent,
	    x, y,
	    False,
	    "Strobes"
	);
	width = d->strobes_cbtn.width;

	/* External Lumination */
	x += (int)width + (PROP_DLG_MARGIN / 2);
	CBtnInit(
	    &d->external_lumination_cbtn, parent,
	    x, y,
	    False,
	    "Lumination Lights"
	);


	/* AI Flags */
	x = PROP_DLG_MARGIN + 6;
	y += (int)height + (2 * PROP_DLG_MARGIN);
	LabelInit(
	    &d->follow_label, parent,
	    x, y + 3,
	    0, 0,
	    "Follow:"
	);
	x += (int)d->follow_label.width;
	CBtnInit(
	    &d->follow_friend_cbtn, parent,
	    x, y,
	    False,
	    "Friend"
	);
	width = d->follow_friend_cbtn.width;

	x += (int)width;
	CBtnInit(
	    &d->follow_unknown_cbtn, parent,
	    x, y,
	    False,
	    "Unknown"
	);
	width = d->follow_unknown_cbtn.width;

	x += (int)width; 
	CBtnInit(
	    &d->follow_hostile_cbtn, parent,
	    x, y,
	    False,
	    "Hostile"
	);
	width = d->follow_hostile_cbtn.width;

	x += (int)width + (2 * PROP_DLG_MARGIN);
	LabelInit(
	    &d->fire_label, parent,
	    x, y + 3,
	    0, 0,
	    "Fire:"
	);
	x += d->fire_label.width;
	CBtnInit(
	    &d->fire_friend_cbtn, parent,
	    x, y,
	    False,
	    "Friend"
	);
	width = d->fire_friend_cbtn.width;

	x += (int)width;
	CBtnInit(
	    &d->fire_unknown_cbtn, parent,
	    x, y,
	    False,
	    "Unknown"
	);
	width = d->fire_unknown_cbtn.width;

	x += (int)width;
	CBtnInit(
	    &d->fire_hostile_cbtn, parent,
	    x, y,
	    False,
	    "Hostile"
	);
	width = d->fire_hostile_cbtn.width;


	/* Link prompts togeather */
	d->name_prompt.prev = &d->cloak_refractory_int_prompt;
	d->name_prompt.next = &d->empire_prompt;

	d->empire_prompt.prev = &d->name_prompt;
	d->empire_prompt.next = &d->uid_prompt;

	d->uid_prompt.prev = &d->empire_prompt;
	d->uid_prompt.next = &d->gid_prompt;

	d->gid_prompt.prev = &d->uid_prompt;
	d->gid_prompt.next = &d->owner_prompt;

	d->owner_prompt.prev = &d->gid_prompt;
	d->owner_prompt.next = &d->credits_prompt;

	d->credits_prompt.prev = &d->owner_prompt;
	d->credits_prompt.next = &d->rmu_prompt;

	d->rmu_prompt.prev = &d->credits_prompt;
	d->rmu_prompt.next = &d->rmu_max_prompt;

	d->rmu_max_prompt.prev = &d->rmu_prompt;
	d->rmu_max_prompt.next = &d->size_prompt;

	d->size_prompt.prev = &d->rmu_max_prompt;
	d->size_prompt.next = &d->velocity_max_prompt;

	d->velocity_max_prompt.prev = &d->size_prompt;
	d->velocity_max_prompt.next = &d->visibility_prompt;

	d->visibility_prompt.prev = &d->velocity_max_prompt;
	d->visibility_prompt.next = &d->hp_prompt;

	d->hp_prompt.prev = &d->visibility_prompt;
	d->hp_prompt.next = &d->hp_max_prompt;

	d->hp_max_prompt.prev = &d->hp_prompt;
	d->hp_max_prompt.next = &d->power_prompt;

	d->power_prompt.prev = &d->hp_max_prompt;
	d->power_prompt.next = &d->power_max_prompt;

	d->power_max_prompt.prev = &d->power_prompt;
	d->power_max_prompt.next = &d->power_purity_prompt;

	d->power_purity_prompt.prev = &d->power_max_prompt;
	d->power_purity_prompt.next = &d->core_efficency_prompt;

	d->core_efficency_prompt.prev = &d->power_purity_prompt;
	d->core_efficency_prompt.next = &d->scanner_range_prompt;

	d->scanner_range_prompt.prev = &d->core_efficency_prompt;
	d->scanner_range_prompt.next = &d->throttle_prompt;

	d->throttle_prompt.prev = &d->scanner_range_prompt;
	d->throttle_prompt.next = &d->antimatter_prompt;

	d->antimatter_prompt.prev = &d->throttle_prompt;
	d->antimatter_prompt.next = &d->antimatter_max_prompt;

	d->antimatter_max_prompt.prev = &d->antimatter_prompt;
	d->antimatter_max_prompt.next = &d->turn_rate_prompt;

	d->turn_rate_prompt.prev = &d->antimatter_max_prompt;
	d->turn_rate_prompt.next = &d->cloak_strength_prompt;

	d->cloak_strength_prompt.prev = &d->turn_rate_prompt;
	d->cloak_strength_prompt.next = &d->cloak_refractory_int_prompt;

	d->cloak_refractory_int_prompt.prev = &d->cloak_strength_prompt;
	d->cloak_refractory_int_prompt.next = &d->name_prompt;


	/* Imgrefs Popup List */
	pulist = &d->imgrefs_pulist;
	x = (int)d->width - PROP_DLG_MARGIN - PROP_DLG_OBJECT_DISPLAY_WIDTH;
	y = PROP_DLG_MARGIN;
	width = PROP_DLG_OBJECT_DISPLAY_WIDTH;
	height = 0;
	PUListInit(
	    pulist, parent,
	    x, y, width, height,
	    0,
	    PULIST_POPUP_BELOW    
	);
	pulist->select_cb = PropDlgImgRefsListSelectCB;
	pulist->select_data = d;


	/* Object Display */
	y += (int)pulist->height;
	width = PROP_DLG_OBJECT_DISPLAY_WIDTH;
	height = PROP_DLG_OBJECT_DISPLAY_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	d->object_display = w;
	d->object_display_buf = None;
	OSWSetWindowInput(w, ExposureMask);

	OSWCreateSharedImage(&d->object_display_img, width, height);



	/* Buttons */
	btn = &d->ok_btn;
	width = PROP_DLG_BTN_WIDTH;
	height = PROP_DLG_BTN_HEIGHT;
	x = PROP_DLG_MARGIN;
	y = (int)d->height - PROP_DLG_MARGIN - (int)height;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    PropDlgOKCB, d
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_Return;
	accel_mods[1] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 2);

	btn = &d->apply_btn;
	x = ((int)d->width - (int)width) / 2;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Apply",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    PropDlgApplyCB, d
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	btn = &d->cancel_btn;
	x = d->width - PROP_DLG_MARGIN - (int)width;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    PropDlgCancelCB, d
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	return(d);
}

/*
 *	Draws the Properties Dialog.
 */
void PropDlgDraw(prop_dlg_struct *d, const int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(d == NULL)
	    return;

	/* Toplevel */
	if(amount == PROP_DLG_DRAW_COMPLETE)
	{
	    unsigned int	width = d->width,
				height = d->height;
	    win_t w = d->toplevel;

#define DRAW_HR(_w_,_x1_,_x2_,_y_)		{	\
 if(wglobal->force_mono) {				\
  OSWSetForeground(gui->white_pix);			\
  OSWDrawLine((_w_), (_x1_), (_y_), (_x2_), (_y_));	\
 } else {						\
  OSWSetForeground(wglobal->surface_shadow_pix);	\
  OSWDrawLine((_w_), (_x1_), (_y_), (_x2_), (_y_));	\
  OSWSetForeground(wglobal->surface_highlight_pix);	\
  OSWDrawLine((_w_), (_x1_), (_y_) + 1, (_x2_), (_y_) + 1); \
 }							\
}

	    /* Draw the background */ 
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(
		    w, 0, 0, width, height
		);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    DRAW_HR(
		w,
		0, (int)width - PROP_DLG_OBJECT_DISPLAY_WIDTH -
		    (2 * PROP_DLG_MARGIN),
		80
	    );

	    DRAW_HR(
		w,
		0, (int)width - PROP_DLG_OBJECT_DISPLAY_WIDTH -
		    (2 * PROP_DLG_MARGIN),
		158
	    );

	    DRAW_HR(
		w,
		0, (int)width,
		238
	    );

	    DRAW_HR(
		w,
		0, (int)width,
		323
	    );

	    DRAW_HR(
		w,  
		0, (int)width,
		370
	    );

	    DRAW_HR(
		w,
		0, (int)width,
		(int)height - PROP_DLG_MARGIN -
		    PROP_DLG_BTN_HEIGHT - PROP_DLG_MARGIN
	    );
#undef DRAW_HR
	}

	/* Object Display */
	if((amount == PROP_DLG_DRAW_COMPLETE) ||
	   (amount == PROP_DLG_DRAW_OBJECT_DISPLAY)
	)
	{
	    int imgref_num;
	    pixmap_t pixmap;
	    win_t w = d->object_display;
	    popup_list_struct *pulist = &d->imgrefs_pulist;
	    imgref_struct *imgref = NULL;
	    uew_struct *uew = d->uew;

	    if(d->object_display_buf == None)
	    {
		unsigned int width, height;
		OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
		OSWCreatePixmap(&d->object_display_buf, width, height);
	    }
	    pixmap = d->object_display_buf;

	    /* Get the currently displayed imgref */
	    imgref_num = (int)PUListItemGetData(
		pulist,
		PUListGetSelItem(pulist)
	    );
	    if((uew != NULL) && (imgref_num > -1))
	    {
		imgref = (imgref_num < uew->total_imgrefs) ?
		    uew->imgref[imgref_num] : NULL;
		ImgRefLoad(imgref);
	    }

	    UEWObjectDisplayDrawPixmap(
		pixmap, d->object_display_img,
		UEWGetObjectPtr(d->uew, d->obj_num),
		imgref, d->ani_frame
	    );

	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the Properties Dialog to draw.
 */
void PropDlgQueueDraw(prop_dlg_struct *d, const int amount)
{
	if(d == NULL)
	    return;

	WidgetQueueDrawAction(
	    PropDlgDrawCB, d, (void *)amount
	);
}

/*
 *	Manages the Properties Dialog event.
 */
int PropDlgManage(prop_dlg_struct *d, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;

	if((d == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case Expose:
	    if(w == d->toplevel)
	    {
		PropDlgQueueDraw(d, 0);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!d->has_focus)
		return(events_handled);;

	    break;

	  case KeyRelease:
	    if(!d->has_focus)
		return(events_handled);;

	    break;

	  case FocusIn:
	    if(w == d->toplevel)
	    {
		d->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut: 
	    if(w == d->toplevel)
	    {
		d->has_focus = False;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(d->toplevel, event, &t))
	    {
		OSWGrabFocusTime(d->toplevel, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(d->toplevel, event))
	    {
		PropDlgCancelCB(&d->cancel_btn, d);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

#define MANAGE_WIDGET(_w_)	{	\
 if(events_handled == 0)		\
  events_handled += WidgetManage(	\
   (_w_), event				\
  );					\
}

	MANAGE_WIDGET(&d->name_prompt);
	MANAGE_WIDGET(&d->type_label);
	MANAGE_WIDGET(&d->type_pulist);
	MANAGE_WIDGET(&d->empire_prompt);
	MANAGE_WIDGET(&d->uid_prompt);
	MANAGE_WIDGET(&d->gid_prompt);
	MANAGE_WIDGET(&d->owner_prompt);
	MANAGE_WIDGET(&d->credits_prompt);
	MANAGE_WIDGET(&d->rmu_prompt);
	MANAGE_WIDGET(&d->rmu_max_prompt);
	MANAGE_WIDGET(&d->size_prompt);
	MANAGE_WIDGET(&d->velocity_max_prompt);
	MANAGE_WIDGET(&d->visibility_prompt);
	MANAGE_WIDGET(&d->hp_prompt);
	MANAGE_WIDGET(&d->hp_max_prompt);
	MANAGE_WIDGET(&d->power_prompt);
	MANAGE_WIDGET(&d->power_max_prompt);
	MANAGE_WIDGET(&d->power_purity_prompt);
	MANAGE_WIDGET(&d->core_efficency_prompt);
	MANAGE_WIDGET(&d->scanner_range_prompt);
	MANAGE_WIDGET(&d->engine_label);
	MANAGE_WIDGET(&d->engine_pulist);
	MANAGE_WIDGET(&d->throttle_prompt);
	MANAGE_WIDGET(&d->antimatter_prompt);
	MANAGE_WIDGET(&d->antimatter_max_prompt);
	MANAGE_WIDGET(&d->turn_rate_prompt);
	MANAGE_WIDGET(&d->shields_label);
	MANAGE_WIDGET(&d->shields_pulist);
	MANAGE_WIDGET(&d->cloak_label);
	MANAGE_WIDGET(&d->cloak_pulist);
	MANAGE_WIDGET(&d->cloak_strength_prompt);
	MANAGE_WIDGET(&d->cloak_refractory_int_prompt);
	MANAGE_WIDGET(&d->vector_lights_cbtn);
	MANAGE_WIDGET(&d->strobes_cbtn);
	MANAGE_WIDGET(&d->external_lumination_cbtn);
	MANAGE_WIDGET(&d->follow_label);
	MANAGE_WIDGET(&d->follow_friend_cbtn);
	MANAGE_WIDGET(&d->follow_unknown_cbtn);
	MANAGE_WIDGET(&d->follow_hostile_cbtn);
	MANAGE_WIDGET(&d->fire_label);
	MANAGE_WIDGET(&d->fire_friend_cbtn);
	MANAGE_WIDGET(&d->fire_unknown_cbtn);
	MANAGE_WIDGET(&d->fire_hostile_cbtn);

	MANAGE_WIDGET(&d->imgrefs_pulist);

	MANAGE_WIDGET(&d->ok_btn);
	MANAGE_WIDGET(&d->apply_btn);
	MANAGE_WIDGET(&d->cancel_btn);

#undef MANAGE_WIDGET

	return(events_handled);
}

/*
 *	Maintains the Properties Dialog.
 */
void PropDlgMaintain(
	void *data, prop_dlg_struct *d,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	ue_option_struct *opt = &option;

	if(d == NULL)
	    return;

	/* Draw animated images */
	if((d->ani_next < cur_ms) &&
	   opt->animate_object_display
	)
	{
	    const int prev_frame = d->ani_frame;

	    d->ani_frame++;
	    if(d->ani_frame >= d->ani_total_frames)
		d->ani_frame = 0;

	    if(d->ani_frame != prev_frame)
		PropDlgQueueDraw(d, PROP_DLG_DRAW_OBJECT_DISPLAY);

	    d->ani_next = cur_ms + d->ani_int;
	}
}

/*
 *	Updates the Properties Dialog's widgets to reflect current
 *	values.
 */
void PropDlgUpdateWidgets(prop_dlg_struct *d)
{
	bool_t b;
	int obj_num, type;
	popup_list_struct *pulist;
	xsw_object_struct *obj;
	uew_struct *uew;

	if(d == NULL)
	    return;

	obj_num = d->obj_num;
	uew = d->uew;
	obj = UEWGetObjectPtr(uew, obj_num);
	if(obj == NULL)
	    return;

	d->freeze_count++;

	pulist = &d->type_pulist;
	type = (int)PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	);
	b = (type == SW_OBJ_TYPE_PLAYER) ? True : False;
	WidgetSetDisabled(&d->owner_prompt, b);

	b = (type == SW_OBJ_TYPE_CONTROLLED) ? False : True;
	WidgetSetDisabled(&d->follow_label, b);
	WidgetSetDisabled(&d->follow_friend_cbtn, b);
	WidgetSetDisabled(&d->follow_unknown_cbtn, b);
	WidgetSetDisabled(&d->follow_hostile_cbtn, b);
	WidgetSetDisabled(&d->fire_label, b);
	WidgetSetDisabled(&d->fire_friend_cbtn, b);
	WidgetSetDisabled(&d->fire_unknown_cbtn, b);
	WidgetSetDisabled(&d->fire_hostile_cbtn, b);



	d->freeze_count--;
}
	
/*
 *	Maps the Properties Dialog.
 */
void PropDlgMap(prop_dlg_struct *d)
{
	if(d == NULL)
	    return;

/*	WidgetSetFocus(&d->title_prompt, False); */

	OSWMapRaised(d->toplevel);
	d->map_state = True;

	WidgetMap(&d->name_prompt);
	WidgetMap(&d->type_label);
	WidgetMap(&d->type_pulist);
	WidgetMap(&d->empire_prompt);
	WidgetMap(&d->uid_prompt);
	WidgetMap(&d->gid_prompt);
	WidgetMap(&d->owner_prompt);
	WidgetMap(&d->credits_prompt);
	WidgetMap(&d->rmu_prompt);
	WidgetMap(&d->rmu_max_prompt);
	WidgetMap(&d->size_prompt);
	WidgetMap(&d->velocity_max_prompt);
	WidgetMap(&d->visibility_prompt);
	WidgetMap(&d->hp_prompt);
	WidgetMap(&d->hp_max_prompt);
	WidgetMap(&d->power_prompt);
	WidgetMap(&d->power_max_prompt);
	WidgetMap(&d->power_purity_prompt);
	WidgetMap(&d->core_efficency_prompt);
	WidgetMap(&d->scanner_range_prompt);
	WidgetMap(&d->engine_label);
	WidgetMap(&d->engine_pulist);
	WidgetMap(&d->throttle_prompt);
	WidgetMap(&d->antimatter_prompt);
	WidgetMap(&d->antimatter_max_prompt);
	WidgetMap(&d->turn_rate_prompt);
	WidgetMap(&d->shields_label);
	WidgetMap(&d->shields_pulist);
	WidgetMap(&d->cloak_label);
	WidgetMap(&d->cloak_pulist);
	WidgetMap(&d->cloak_strength_prompt);
	WidgetMap(&d->cloak_refractory_int_prompt);
	WidgetMap(&d->vector_lights_cbtn);
	WidgetMap(&d->strobes_cbtn);
	WidgetMap(&d->external_lumination_cbtn);
	WidgetMap(&d->follow_label);
	WidgetMap(&d->follow_friend_cbtn);
	WidgetMap(&d->follow_unknown_cbtn);
	WidgetMap(&d->follow_hostile_cbtn);
	WidgetMap(&d->fire_label);
	WidgetMap(&d->fire_friend_cbtn);
	WidgetMap(&d->fire_unknown_cbtn);
	WidgetMap(&d->fire_hostile_cbtn);

	WidgetMap(&d->imgrefs_pulist);
	OSWMapWindow(d->object_display);

	WidgetMap(&d->ok_btn);
	WidgetMap(&d->apply_btn);
	WidgetMap(&d->cancel_btn);

	OSWGrabFocus(d->toplevel);

	UEWDoUpdateWindowMenus();
}

/*
 *	Gets the values from the Universe Editor referenced by the
 *	Properties Dialog and maps the Properties Dialog.
 */
void PropDlgMapValues(prop_dlg_struct *d)
{
	if(d == NULL)
	    return;

	PropDlgGetValues(d);

	d->has_changes = False;

	PropDlgMap(d);
}

/*
 *	Unmaps the Properties Dialog.
 */
void PropDlgUnmap(prop_dlg_struct *d)
{
	if(d == NULL)
	    return;

	OSWUnmapWindow(d->toplevel);
	d->map_state = False;

	OSWDestroyPixmap(&d->object_display_buf);

	UEWDoUpdateWindowMenus();
}

/*
 *	Deletes the Properties Dialog.
 */
void PropDlgDelete(prop_dlg_struct *d)
{
	if(d == NULL)
	    return;

	PUListClear(&d->imgrefs_pulist);

	WidgetDestroy(&d->name_prompt);
	WidgetDestroy(&d->type_label);
	WidgetDestroy(&d->type_pulist);
	WidgetDestroy(&d->empire_prompt);
	WidgetDestroy(&d->uid_prompt);
	WidgetDestroy(&d->gid_prompt);
	WidgetDestroy(&d->owner_prompt);
	WidgetDestroy(&d->credits_prompt);
	WidgetDestroy(&d->rmu_prompt);
	WidgetDestroy(&d->rmu_max_prompt);
	WidgetDestroy(&d->size_prompt);
	WidgetDestroy(&d->velocity_max_prompt);
	WidgetDestroy(&d->visibility_prompt);
	WidgetDestroy(&d->hp_prompt);
	WidgetDestroy(&d->hp_max_prompt);
	WidgetDestroy(&d->power_prompt);
	WidgetDestroy(&d->power_max_prompt);
	WidgetDestroy(&d->power_purity_prompt);
	WidgetDestroy(&d->core_efficency_prompt);
	WidgetDestroy(&d->scanner_range_prompt);
	WidgetDestroy(&d->engine_label);
	WidgetDestroy(&d->engine_pulist);
	WidgetDestroy(&d->throttle_prompt);
	WidgetDestroy(&d->antimatter_prompt);
	WidgetDestroy(&d->antimatter_max_prompt);
	WidgetDestroy(&d->turn_rate_prompt);
	WidgetDestroy(&d->shields_label);
	WidgetDestroy(&d->shields_pulist);
	WidgetDestroy(&d->cloak_label);
	WidgetDestroy(&d->cloak_pulist);
	WidgetDestroy(&d->cloak_strength_prompt);
	WidgetDestroy(&d->cloak_refractory_int_prompt);
	WidgetDestroy(&d->vector_lights_cbtn);
	WidgetDestroy(&d->strobes_cbtn);
	WidgetDestroy(&d->external_lumination_cbtn);
	WidgetDestroy(&d->follow_label);
	WidgetDestroy(&d->follow_friend_cbtn);
	WidgetDestroy(&d->follow_unknown_cbtn);
	WidgetDestroy(&d->follow_hostile_cbtn);
	WidgetDestroy(&d->fire_label);
	WidgetDestroy(&d->fire_friend_cbtn);
	WidgetDestroy(&d->fire_unknown_cbtn);
	WidgetDestroy(&d->fire_hostile_cbtn);

	WidgetDestroy(&d->imgrefs_pulist);
	OSWDestroyWindow(&d->object_display);
	OSWDestroyPixmap(&d->object_display_buf);
	OSWDestroySharedImage(&d->object_display_img);

	WidgetDestroy(&d->ok_btn);
	WidgetDestroy(&d->apply_btn);
	WidgetDestroy(&d->cancel_btn);

	OSWDestroyWindow(&d->toplevel);

	free(d);
}
