#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"

#include "sndrefs.h"


sndref_struct *SndRefGetPtr(
	sndref_struct **sndrefs_list, const int nsndrefs,
	const int sndref_num
);

sndref_struct *SndRefNew(void);
int SndRefNewExplicit(
	sndref_struct ***sndrefs_list, int *nsndrefs,
	const int sndref_num
);
void SndRefDelete(
	sndref_struct **sndrefs_list, const int nsndrefs,
	const int sndref_num
);
void SndRefDeleteAll(
	sndref_struct ***sndrefs_list, int *nsndrefs
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the pointer to the sndref in the specified list.
 */
sndref_struct *SndRefGetPtr(
	sndref_struct **sndrefs_list, const int nsndrefs,
	const int sndref_num
)
{
	if((sndref_num < 0) ||
	   (sndref_num >= nsndrefs) ||
	   (sndrefs_list == NULL)
	)
	    return(NULL);
	else
	    return(sndrefs_list[sndref_num]);
}


/*
 *	Creates a new sndref.
 */
sndref_struct *SndRefNew(void)
{
	return(SNDREF(calloc(1, sizeof(sndref_struct))));
}

/*
 *	Creates a new sndref at the specified index.
 *
 *	Returns 0 on success or non-zero on error.
 */
int SndRefNewExplicit(
	sndref_struct ***sndrefs_list, int *nsndrefs,
	const int sndref_num
)
{
	sndref_struct *sndref;

	if((sndrefs_list == NULL) || (nsndrefs == NULL) ||
	   (sndref_num < 0)
	)
	    return(-2);

	/* Delete the sndref if it already exists */
	SndRefDelete(*sndrefs_list, *nsndrefs, sndref_num);

	if(*nsndrefs < 0)
	    *nsndrefs = 0;

	/* Need to allocate more pointers? */
	if(sndref_num >= *nsndrefs)
	{
	    int i;
	    const int prev_total = *nsndrefs;
	    *nsndrefs = sndref_num + 1;

	    /* Allocate more pointers */
	    *sndrefs_list = (sndref_struct **)realloc(
	        *sndrefs_list,
	        (*nsndrefs) * sizeof(sndref_struct *)
	    );
	    if(*sndrefs_list == NULL)
	    {
	        *sndrefs_list = 0;
	        return(-3);
	    }

	    /* Reset newly allocated pointers */
	    for(i = prev_total; i < *nsndrefs; i++)
		(*sndrefs_list)[i] = NULL;
	}

	(*sndrefs_list)[sndref_num] = sndref = SndRefNew();
	if(sndref == NULL)
	    return(-3);

	return(0);
}

/*
 *	Deletes the sndref.
 */
void SndRefDelete(
	sndref_struct **sndrefs_list, const int nsndrefs,
	const int sndref_num
)
{
	sndref_struct *sndref = SndRefGetPtr(
	    sndrefs_list, nsndrefs, sndref_num
	);
	if(sndref == NULL)
	    return;

	free(sndref->path);
	free(sndref->name);
	free(sndref->description);
	free(sndref);
	sndrefs_list[sndref_num] = NULL;
}

/*
 *	Deletes all the sndrefs in the specified list and then deletes
 *	the list itself.
 */
void SndRefDeleteAll(
	sndref_struct ***sndrefs_list, int *nsndrefs
)
{
	int i;

	if((sndrefs_list == NULL) || (nsndrefs == NULL))
	    return;

	for(i = 0; i < *nsndrefs; i++)
	    SndRefDelete(*sndrefs_list, *nsndrefs, i);
	free(*sndrefs_list);
	*sndrefs_list = NULL;
	*nsndrefs = 0;
}
