#include <stdlib.h>
#include <string.h>

#include "../include/widget.h"

#include "ue.h"


bool_t UESelectFile(
	const char *title,
	const char *ok_label, const char *cancel_label,
	const char *path, char ***paths_rtn, int *npaths_rtn,
	win_t toplevel
);

fb_struct *UEFileSelectorNew(void);
void UEFileSelectorDelete(fb_struct *file_sel);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : True)


/*
 *	Maps the file browser and waits for a user response.
 *
 *	Returns True if one or more files have been selected or False
 *	if the user canceled.
 *
 *	The calling function must deleted the returned list of paths.
 */
bool_t UESelectFile(
	const char *title,
	const char *ok_label, const char *cancel_label,
	const char *path, char ***paths_rtn, int *npaths_rtn,
	win_t toplevel
)
{
	bool_t response;
	fb_struct *file_sel;

	/* Not connected to GUI? */
	if(!IDC())
	    return(False);

	/* Initialize the file selector as needed */
	file_sel = ue_file_sel;
	if(file_sel == NULL)
	    ue_file_sel = file_sel = UEFileSelectorNew();
	if(file_sel == NULL)
	    return(False);

	/* Get response */
	response = FBrowserGetResponse(
	    file_sel,
	    title, ok_label, cancel_label,
	    path, paths_rtn, npaths_rtn,
	    WIDGET_WINDOW_RELATIVE_PARENT,
	    toplevel,
	    UEMain, NULL,
	    UEMainPop, NULL
	);

	/* Delete the file selector */
	UEFileSelectorDelete(file_sel);
	ue_file_sel = file_sel = NULL;

	return(response);
}


/*
 *	Creates a new file selector.
 */
fb_struct *UEFileSelectorNew(void)
{
	fb_struct *file_sel = (fb_struct *)calloc(
	    1, sizeof(fb_struct)
	);
	if(file_sel == NULL)
	    return(NULL);

	if(FBrowserInit(file_sel, FB_STYLE_SINGLE_LIST))
	{
	    free(file_sel);
	    file_sel = NULL;
	}

	file_sel->options |= FB_FLAG_WRITE_PROTECT;
        file_sel->options |= FB_FLAG_CLOSE_ON_OK;
        file_sel->options |= FB_FLAG_CLOSE_ON_CANCEL;

	return(file_sel);
}

/*
 *	Deletes the file selector.
 */
void UEFileSelectorDelete(fb_struct *file_sel)
{
	if(file_sel == NULL)
	    return;

	WidgetDestroy(file_sel);
	free(file_sel);
}

