#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/reality.h"
#include "../include/objects.h"
#include "../include/unvmain.h"
#include "../include/unvutil.h"

#include "blitting.h"
#include "keymap.h"
#include "uhw.h"
#include "propdlg.h"
#include "ecow.h"
#include "wepw.h"
#include "printdlg.h"
#include "uew.h"
#include "ue.h"
#include "config.h"

#include "../include/oswkeyvals.h"


static void UEWResizeCB(void *widget, void *data);
static void UEWDrawCB(void *widget, void *data);

void UEWSetTitle(uew_struct *uew, const char *s);
void UEWUpdateTitle(uew_struct *uew);
void UEWDisplayObject(
	uew_struct *uew, const int obj_num,
	const bool_t move_to
);
void UEWStatusMessage(uew_struct *uew, const char *msg);
void UEWStatusProgress(uew_struct *uew, const float coeff);
void UEWSetBusy(uew_struct *uew, const bool_t busy);

uew_struct *UEWNew(void);
void UEWResize(uew_struct *uew);
void UEWQueueResize(uew_struct *uew);
static void UEWDrawResizeBarV(
	uew_struct *uew,
	const int x, const int y, const unsigned int len,
	const int last_x, const bool_t has_previous
);
void UEWDraw(uew_struct *uew, const int amount);
void UEWQueueDraw(uew_struct *uew, const int amount);
int UEWManage(uew_struct *uew, event_t *event);
void UEWMaintain(
	void *data, uew_struct *uew,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void UEWUpdateWidgets(uew_struct *uew);
void UEWMap(uew_struct *uew);
void UEWUnmap(uew_struct *uew);
void UEWDelete(uew_struct *uew);


#define UEW_DEF_TITLE		PROG_NAME_FULL
#define UEW_DEF_ICON_TITLE	PROG_NAME

#define UEW_HINT_MESG_NEW	"Create a new universe"
#define UEW_HINT_MESG_OPEN	"Open a universe"
#define UEW_HINT_MESG_SAVE	"Save the current universe"
#define UEW_HINT_MESG_SAVE_AS	"Save the current universe under a new file name"
#define UEW_HINT_MESG_SAVE_AS_IMAGE	"Save a screenshot image of the current view contents"
#define UEW_HINT_MESG_CUT	"Cut the selected object(s)"
#define UEW_HINT_MESG_COPY	"Copy the selected object(s)"
#define UEW_HINT_MESG_PASTE	"Paste object(s)"
#define UEW_HINT_MESG_HEADER	"Edit the universe's legend"
#define UEW_HINT_MESG_SELECT_OPM_FILE	"Select OPM file"
#define UEW_HINT_MESG_NEW_OBJECT	"Insert a new object"
#define UEW_HINT_MESG_PROPERTIES	"Edit the selected object's properties"
#define UEW_HINT_MESG_WEAPONS	"Edit the selected object's weapons"
#define UEW_HINT_MESG_ECONOMY	"Edit the selected object's economy"
#define UEW_HINT_MESG_REMODEL	"Remodel the selected object(s) after the selected OPM"
#define UEW_HINT_MESG_DELETE	"Delete the selected object(s)"
#define UEW_HINT_MESG_PRINT	"Print the current universe"

#define UEW_DEF_WIDTH			800
#define UEW_DEF_HEIGHT			600

#define UEW_MIN_WIDTH			320
#define UEW_MIN_HEIGHT			240

#define UEW_MARGIN			10

#define UEW_TOOLBAR_MARGIN_MAJOR	5
#define UEW_TOOLBAR_MARGIN_MINOR	2
#define UEW_TOOLBAR_BTN_WIDTH		26
#define UEW_TOOLBAR_BTN_HEIGHT		26
#define UEW_TOOLBAR_HEIGHT              (UEW_TOOLBAR_BTN_HEIGHT + (2 * 2))

#define UEW_PROMPT_HEIGHT		30

#define UEW_PULIST_HEIGHT		26

#define UEW_PREVIEW_HEIGHT		150

#define UEW_SPLIT_WIDTH			5

#define UEW_VIEW_STATUS_HEIGHT		(UEW_MARGIN + \
 (3 * UEW_PROMPT_HEIGHT))

#define UEW_STATUS_HEIGHT		25
#define UEW_PBAR_WIDTH			200
#define UEW_PBAR_HEIGHT			12

#define UEW_DEF_VIEW_X_POS		190
#define UEW_DEF_PROPERTIES_X_POS	570


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define FTOI_RNDOFF(f)	((int)((f) + 0.5f))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Resize callback.
 */
static void UEWResizeCB(void *widget, void *data)
{
	uew_struct *uew = UEW(widget);
	if(uew == NULL)
	    return;

	UEWResize(uew);
}

/*
 *	Draw callback.
 */
static void UEWDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	uew_struct *uew = UEW(widget);
	if(uew == NULL)
	    return;

	if(uew->map_state)
	    UEWDraw(uew, amount);
}


/*
 *	Sets the Universe Editor Window's title.
 */
void UEWSetTitle(uew_struct *uew, const char *s)
{
	win_t w;

	if(uew == NULL)
	    return;

	w = uew->toplevel;
	if(w == None)
	    return;

	OSWSetWindowTitle(w, (s != NULL) ? s : "");
}

/*
 *	Updates the Universe Editor Window's title.
 */
void UEWUpdateTitle(uew_struct *uew)
{
	const char *title;
	char *s;
	sw_legend_struct *legend;

	if(uew == NULL)
	    return;

	legend = &uew->legend;
	title = STRISEMPTY(legend->title) ? "Untitled" : legend->title;

	s = (char *)malloc(
	    STRLEN(UEW_DEF_TITLE) + STRLEN(title) + 80
	);
	sprintf(
	    s,
	    "%s: %s%s",
	    UEW_DEF_TITLE,
	    title,
	    uew->has_changes ? " (*)" : ""
	);
	UEWSetTitle(uew, s);
	free(s);
}

/*
 *	Displays the specified object and its properties in the
 *	property and view prompts.
 *
 *	If obj_num is -1 or invalid then the object display, view
 *	prompts, and property prompts will be cleared.
 *
 *	If move_to is True then the Universe Editor's view will be
 *	moved to the center of the specified object.
 *
 *	Note that this will discard any existing data in the property
 *	and view prompts.
 */
void UEWDisplayObject(
	uew_struct *uew, const int obj_num,
	const bool_t move_to
)
{
	int imgref_num;
	imgref_struct *imgref;
	popup_list_struct *pulist;
	xsw_object_struct *obj;

	if(uew == NULL)
	    return;

	pulist = &uew->imgrefs_pulist;
	obj = UEWGetObjectPtr(uew, obj_num);
	if(obj != NULL)
	{
	    /* Unload the current imgref */
	    imgref_num = uew->cur_imgref;
	    imgref = ((imgref_num >= 0) && (imgref_num < uew->total_imgrefs)) ?
		uew->imgref[imgref_num] : NULL;
	    ImgRefUnload(imgref);

	    /* Set and load the new imgref */
	    uew->cur_imgref = imgref_num = obj->imgref;
	    imgref = ((imgref_num >= 0) && (imgref_num < uew->total_imgrefs)) ?
		uew->imgref[imgref_num] : NULL;
	    ImgRefLoad(imgref);

	    /* Update animation values */
	    if(obj->type == SW_OBJ_TYPE_ANIMATED)
	    {
		xsw_animation_struct *anim = &obj->animation;

		uew->ani_frame = 0;
		uew->ani_total_frames = (imgref != NULL) ?
		    MAX(imgref->total_frames, 1) : 1;
		uew->ani_int = MAX(anim->interval, 10);
		uew->ani_next = cur_millitime + uew->ani_int;
	    }
	    else
	    {
		uew->ani_frame = 0;
		uew->ani_total_frames = (imgref != NULL) ?
		    MAX(imgref->total_frames, 1) : 1;
		uew->ani_int = 500l;
		uew->ani_next = cur_millitime + uew->ani_int;
	    }

	    /* Update the displayed item on the ImgRefs Popup List */
	    uew->freeze_count++;
	    PUListSetSelItem(
		pulist,
		PUListGetItemByData(pulist, (void *)imgref_num)
	    );
	    uew->freeze_count--;

	    if(move_to)
	    {
		uew->cx = obj->x;
		uew->cy = obj->y;
		uew->cz = obj->z;
		uew->sect_x = obj->sect_x;
		uew->sect_y = obj->sect_y;
		uew->sect_z = obj->sect_z;
	    }

	    /* Get the view & property values */
	    UEWViewStatusGetValues(uew, obj);
	    UEWPropsDoGetValues(uew, obj_num);
	}
	else
	{
	    /* No valid object specified, clear existing values */

	    /* Unload the current imgref */
	    imgref_num = uew->cur_imgref;
	    imgref = ((imgref_num >= 0) && (imgref_num < uew->total_imgrefs)) ?
		uew->imgref[imgref_num] : NULL;
	    ImgRefUnload(imgref);

	    /* Clear the view & property prompts */
	    UEWViewStatusGetValues(uew, NULL);
	    UEWPropsDoGetValues(uew, obj_num);
	}

	if(uew->map_state)
	{
	    UEWQueueDraw(uew, UEW_DRAW_VIEW);
	    UEWQueueDraw(uew, UEW_DRAW_PREVIEW);
	}
}

/*
 *      Sets the status bar's message.
 */
void UEWStatusMessage(uew_struct *uew, const char *msg)
{ 
	if(uew == NULL)
	    return;
		   
	if(uew->status_message != msg)
	{
	    free(uew->status_message);
	    uew->status_message = STRDUP(msg);
	    if(uew->map_state)
		UEWQueueDraw(uew, UEW_DRAW_STATUS);
	}
}

/*
 *      Sets the status bar's progress.
 */
void UEWStatusProgress(uew_struct *uew, const float coeff)
{
	progress_bar_struct *pb;

	if(uew == NULL)
	    return;

	pb = &uew->pbar;
	if(coeff > pb->current)
	{
	    if((coeff - pb->current) > 0.01)
	    {
		PBarUpdate(pb, coeff, NULL);
		WidgetMainIteration(NULL);
	    }
	}
	else
	{
	    PBarUpdate(pb, coeff, NULL);
	    WidgetMainIteration(NULL);
	}
}

/*
 *	Sets the busy state.
 */
void UEWSetBusy(uew_struct *uew, const bool_t busy)
{
	if(uew == NULL)
	    return;

	if(busy)
	    uew->busy_count++;
	else
	    uew->busy_count = MAX(uew->busy_count - 1, 0);

	if(uew->busy_count == 1)
	{
	    cursor_t cursor = ue_cursor.busy;
	    OSWSetWindowCursor(uew->toplevel, cursor);
	    OSWSetWindowCursor(uew->view, cursor);
	}
	else if(uew->busy_count == 0)
	{
	    OSWSetWindowCursor(uew->toplevel, None);
	    WidgetSetWindowCursor(uew->view, ue_cursor.scanner_lock_wcur);
	}
}


/*
 *	Initializes the Universe Editor Window.
 */
uew_struct *UEWNew(void)
{
	int x, y, i, n;
	unsigned int width, height;
	win_t w, parent, parent2;
	menu_struct *menu;
	menu_bar_struct *mb;
	menu_bar_item_struct *mb_item;
	scroll_bar_struct *sb;
	clist_struct *clist;
	popup_list_struct *pulist;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;
	ue_option_struct *opt = &option;
	uew_struct *uew = UEW(calloc(1, sizeof(uew_struct)));

	uew->map_state = False;
	uew->has_focus = False;
	uew->disabled = False;
	uew->freeze_count = 0;
	uew->busy_count = 0;
	uew->view_has_focus = False;

	if(gui->def_geometry_set)
	{
	    uew->x = gui->def_toplevel_x;
	    uew->y = gui->def_toplevel_y;
	    uew->width = gui->def_toplevel_width;
	    uew->height = gui->def_toplevel_height;
	}
	else if(opt->record_window_geometry)
	{
	    rectangle_t *rect = &opt->uew_geometry;

	    uew->x = rect->x;
	    uew->y = rect->y;
	    uew->width = (rect->width > 0) ?
		rect->width : UEW_DEF_WIDTH;
	    uew->height = (rect->height > 0) ?
		rect->height : UEW_DEF_HEIGHT;

	    /* Cascade */
	    rect->x += 30;
	    rect->y += 30;
	    if((rect->x + (int)rect->width) > (int)gui->display_width)
		rect->x = 0;
	    if((rect->y + (int)rect->height) > (int)gui->display_width)
		rect->y = 0;
	}
	else
	{
	    uew->x = 0;
	    uew->y = 0;
	    uew->width = UEW_DEF_WIDTH;
	    uew->height = UEW_DEF_HEIGHT;
	}

	uew->options = 0;

	uew->unv_file = NULL;

	uew->has_changes = False;

	uew->last_selected_object = -1;

	uew->view_pos = UEW_DEF_VIEW_X_POS;
	uew->properties_pos = UEW_DEF_PROPERTIES_X_POS;

	uew->sect_x = 0l;
	uew->sect_y = 0l;
	uew->sect_z = 0l;
	uew->cx = 0.0f;
	uew->cy = 0.0f;
	uew->cz = 0.0f;
	uew->zoom = UEW_VIEW_DEF_ZOOM;

	uew->status_message = NULL;

	uew->target_string = osw_atom.target_string;
	uew->dnd_action = None;
	uew->dnd_tar_w = None;
	uew->dnd_x = 0;
	uew->dnd_y = 0;

	uew->drag_op = UEW_DRAG_OP_NONE;
	uew->button_pressed = 0;
	uew->button_pressed_time = 0l;

	uew->resize_bar_last_pos = 0;

	uew->imgref = NULL;
	uew->total_imgrefs = 0;
	uew->cur_imgref = -1;

	UNVLegendReset(&uew->legend);

	uew->uhw = NULL;

	uew->prop_dlg = NULL;
	uew->total_prop_dlgs = 0;

	uew->wepw = NULL;
	uew->total_wepws = 0;

	uew->ecow = NULL;
	uew->total_ecows = 0;

	uew->print_dlg = NULL;


	/* Create toplevel */
	parent = gui->root_win;
	OSWCreateWindow(
	    &w, parent,
	    uew->x, uew->y,
	    uew->width, uew->height
	);
	uew->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    UEW_DEF_TITLE ": Untitled",
	    UEW_DEF_ICON_TITLE,
	    (ue_image.icon_unvedit_pixmap != None) ?
		ue_image.icon_unvedit_pixmap :
		widget_global.std_icon_pm,
	    /* Let WM set coordinates? */
	    (!gui->def_geometry_set &&
		!opt->record_window_geometry) ? True : False,
	    uew->x, uew->y,
	    UEW_MIN_WIDTH, UEW_MIN_HEIGHT,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_STANDARD,
	    0, NULL
	);
	OSWSetWindowInput(
	    w,
	    OSW_EVENTMASK_TOPLEVEL
	);
	OSWDNDSetWindowTarget(w);
	parent = w;

#ifdef HAVE_X
	if(w != 0)
	{
	    gc_val_t v;
	    v.function = GXinvert;
	    v.foreground = gui->white_pix;
	    v.background = gui->black_pix;
	    v.line_width = 1;
	    v.line_style = LineSolid;
	    v.subwindow_mode = IncludeInferiors;
	    uew->invert_gc = XCreateGC(
		gui->display, w,
		GCFunction |
		GCForeground | GCBackground |
		GCLineWidth | GCLineStyle |
		GCSubwindowMode,
		&v
	    );
	}
#endif

	/* Menu bar */
#define APPEND_MENU_ITEM(_t_,_s_,_p_,_m_,_id_) {\
 n = MenuBarMenuItemAppend(			\
  mb, i, (_t_), (_s_), (_p_), (_m_), (_id_)	\
 );						\
}
#define APPEND_MENU_ITEM_SEPARATOR	{	\
 n = MenuBarMenuItemAppend(			\
  mb, i, MENU_ITEM_SEPARATOR,			\
  NULL, None, None, UEW_OP_NONE			\
 );						\
}
#define SET_MENU_ITEM_ACCEL_KEY(_k_,_m_) {	\
 MenuSetItemAccelKey(				\
  menu, n, (_k_), (_m_)				\
 );						\
}

	mb = &uew->mb;
	MenuBarInit(
	    mb, parent,
	    0, 0,
	    0, 0,
	    UEWMenuActivateCB, uew
	);

	/* File menu */
	x = 0;
	y = 0;
	i = MenuBarMenuAppend(
	    mb,
	    "File",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "New",
	    ue_image.icon_new_pixmap,
	    ue_image.icon_new_mask,
	    UEW_OP_NEW
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_n, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Open...",
	    ue_image.icon_open_pixmap,
	    ue_image.icon_open_mask,
	    UEW_OP_OPEN
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_o, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Open In New Window...",
	    ue_image.icon_open_pixmap,
	    ue_image.icon_open_mask,
	    UEW_OP_OPEN_NEW_UEW
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_O, OSW_KEY_STATE_SHIFT | OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Insert...",
	    ue_image.icon_open_pixmap,
	    ue_image.icon_open_mask,
	    UEW_OP_INSERT_UNIVERSE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_j, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Save",
	    ue_image.icon_save_pixmap,
	    ue_image.icon_save_mask,
	    UEW_OP_SAVE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_s, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Save As...",
	    ue_image.icon_save_as_pixmap,
	    ue_image.icon_save_as_mask,
	    UEW_OP_SAVE_AS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_a, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Save Selected As...",
	    ue_image.icon_save_as_pixmap,
	    ue_image.icon_save_as_mask,
	    UEW_OP_SAVE_SELECTED_AS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_A, OSW_KEY_STATE_CTRL | OSW_KEY_STATE_SHIFT
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Save As Image...",
	    ue_image.icon_screenshot_pixmap,
	    ue_image.icon_screenshot_mask,
	    UEW_OP_SAVE_AS_IMAGE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_I, OSW_KEY_STATE_CTRL | OSW_KEY_STATE_SHIFT
	);

	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Print...",
	    ue_image.icon_print_pixmap,
	    ue_image.icon_print_mask,
	    UEW_OP_PRINT
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_p, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Close",
	    ue_image.icon_close_pixmap,
	    ue_image.icon_close_mask,
	    UEW_OP_CLOSE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_w, OSW_KEY_STATE_CTRL
	);


	/* Edit menu */
	x += (int)mb_item->width;
	i = MenuBarMenuAppend(
	    mb,
	    "Edit",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Cut",
	    ue_image.icon_cut_pixmap,
	    ue_image.icon_cut_mask,
	    UEW_OP_CUT
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_x, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Copy",
	    ue_image.icon_copy_pixmap,
	    ue_image.icon_copy_mask,
	    UEW_OP_COPY
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_c, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Paste",
	    ue_image.icon_paste_pixmap,
	    ue_image.icon_paste_mask,
	    UEW_OP_PASTE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_v, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Delete",
	    ue_image.icon_delete_pixmap,
	    ue_image.icon_delete_mask,
	    UEW_OP_DELETE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_Delete, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Select All",
	    None, None,
	    UEW_OP_SELECT_ALL
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_a, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Unselect All",
	    None, None,
	    UEW_OP_UNSELECT_ALL
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_u, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Invert Selection",
	    None, None,
	    UEW_OP_INVERT_SELECTION
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_i, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Properties...",
	    ue_image.icon_object_properties_pixmap,
	    ue_image.icon_object_properties_mask,
	    UEW_OP_OBJ_PROPERTIES
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_Return, OSW_KEY_STATE_CTRL
	);

	/* View */
	x += (int)mb_item->width;
	i = MenuBarMenuAppend(
	    mb,
	    "View",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;
	APPEND_MENU_ITEM(
	    MENU_ITEM_CHECK,
	    "Object Labels",
	    None, None,
	    UEW_OP_SHOW_OBJECT_LABELS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_F9, 0
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_CHECK,
	    "Geometry Labels",
	    None, None,
	    UEW_OP_SHOW_GEOMETRY_LABELS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_F10, 0
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_CHECK,
	    "Grid",
	    None, None,
	    UEW_OP_SHOW_GRID
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_F11, 0
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_CHECK,
	    "Contact Bounds",
	    None, None,
	    UEW_OP_SHOW_CONTACT_BOUNDS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_F12, 0
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_CHECK,
	    "Animate Object Display",
	    None, None,
	    UEW_OP_ANIMATE_OBJECT_DISPLAY
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_F4, 0
	);

	/* Universe menu */
	x += (int)mb_item->width;
	i = MenuBarMenuAppend(
	    mb,
	    "Universe",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Legend...",
	    ue_image.icon_legend_pixmap,
	    ue_image.icon_legend_mask,
	    UEW_OP_LEGEND
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_l, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Insert...",
	    ue_image.icon_open_pixmap,
	    ue_image.icon_open_mask,
	    UEW_OP_INSERT_UNIVERSE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_j, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Check",
	    ue_image.icon_select_pixmap,
	    ue_image.icon_select_mask,
	    UEW_OP_CHECK_UNIVERSE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_f, OSW_KEY_STATE_CTRL
	);

	/* Objects menu */
	x += (int)mb_item->width;
	i = MenuBarMenuAppend(
	    mb,
	    "Objects",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Insert New",
	    ue_image.icon_new_object_pixmap,
	    ue_image.icon_new_object_mask,
	    UEW_OP_INSERT_OBJECT
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_i, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Properties...",
	    ue_image.icon_object_properties_pixmap,
	    ue_image.icon_object_properties_mask,
	    UEW_OP_OBJ_PROPERTIES
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_Return, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Weapons...",
	    ue_image.icon_weapons_pixmap,
	    ue_image.icon_weapons_mask,
	    UEW_OP_OBJ_WEAPONS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_t, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Economy...",
	    ue_image.icon_economy_pixmap,
	    ue_image.icon_economy_mask,
	    UEW_OP_OBJ_ECONOMY
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_e, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Remodel",
	    ue_image.icon_remodel_object_pixmap,
	    ue_image.icon_remodel_object_mask,
	    UEW_OP_OBJ_REMODEL
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_r, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Delete",
	    ue_image.icon_delete_pixmap,
	    ue_image.icon_delete_mask,
	    UEW_OP_DELETE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_Delete, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM_SEPARATOR
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Select OPM File...",
	    ue_image.icon_opm_file_pixmap,
	    ue_image.icon_opm_file_mask,
	    UEW_OP_SELECT_OPM_FILE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_m, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Export OPM File...",
	    ue_image.icon_export_pixmap,
	    ue_image.icon_export_mask,
	    UEW_OP_EXPORT_OPM_FILE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_M, OSW_KEY_STATE_CTRL | OSW_KEY_STATE_SHIFT
	);

	/* Settings menu */
	x += (int)mb_item->width;
	i = MenuBarMenuAppend(
	    mb,
	    "Settings",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Options...",
	    None, None,
	    UEW_OP_OPTIONS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_g, OSW_KEY_STATE_CTRL
	);

	/* Windows menu */
	x += (int)mb_item->width;
	uew->window_menu_num = i = MenuBarMenuAppend(
	    mb,
	    "Window",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;


	/* Help menu */
	x += (int)mb_item->width;
	uew->help_menu_num = i = MenuBarMenuAppend(
	    mb,
	    "Help",
	    x, y,
	    0, 0
	);
	mb_item = mb->item[i];
	menu = mb_item->menu;

	/* Move help menu item to the right side */
	mb_item->x = (int)uew->width - (int)mb_item->width;

	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "About",
	    ue_image.icon_about_pixmap,
	    ue_image.icon_about_mask,
	    UEW_OP_ABOUT
	);

#undef SET_MENU_ITEM_ACCEL_KEY
#undef APPEND_MENU_ITEM
#undef APPEND_MENU_ITEM_SEPARATOR


	/* Tool Bar */
	x = 0;
	y = (int)mb->height;
	width = uew->width;
	height = UEW_TOOLBAR_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->toolbar = w;
	OSWSetWindowInput(w, ExposureMask);
	parent2 = w;

	/* Tool Bar Buttons */
	x = UEW_TOOLBAR_MARGIN_MAJOR;
	y = ((int)height - UEW_TOOLBAR_BTN_HEIGHT) / 2;
	width = UEW_TOOLBAR_BTN_WIDTH;
	height = UEW_TOOLBAR_BTN_HEIGHT;
	PBtnInit(
	    &uew->tb_new_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_new_pixmap,
	    ue_image.icon_new_mask,
	    UEWTBNewPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_new_btn,
	    UEW_HINT_MESG_NEW
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_open_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_open_pixmap,
	    ue_image.icon_open_mask,
	    UEWTBOpenPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_open_btn,
	    UEW_HINT_MESG_OPEN
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_save_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_save_pixmap,
	    ue_image.icon_save_mask,
	    UEWTBSavePBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_save_btn,
	    UEW_HINT_MESG_SAVE
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_save_as_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_save_as_pixmap,
	    ue_image.icon_save_as_mask,
	    UEWTBSaveAsPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_save_as_btn,
	    UEW_HINT_MESG_SAVE_AS
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_save_as_image_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_screenshot_pixmap,
	    ue_image.icon_screenshot_mask,
	    UEWTBSaveAsImagePBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_save_as_image_btn,
	    UEW_HINT_MESG_SAVE_AS_IMAGE
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MAJOR);
	PBtnInit(
	    &uew->tb_cut_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_cut_pixmap,
	    ue_image.icon_cut_mask,
	    UEWTBCutPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_cut_btn,
	    UEW_HINT_MESG_CUT
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_copy_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_copy_pixmap,
	    ue_image.icon_copy_mask,
	    UEWTBCopyPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_copy_btn,
	    UEW_HINT_MESG_COPY
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_paste_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_paste_pixmap,
	    ue_image.icon_paste_mask,
	    UEWTBPastePBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_paste_btn,
	    UEW_HINT_MESG_PASTE
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MAJOR);
	PBtnInit(
	    &uew->tb_header_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_legend_pixmap,
	    ue_image.icon_legend_mask,
	    UEWTBHeaderPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_header_btn,
	    UEW_HINT_MESG_HEADER
	);

	pulist = &uew->opm_pulist;
	x += ((int)width + UEW_TOOLBAR_MARGIN_MAJOR);
	PUListInit(
	    pulist, parent2,
	    x, y,
	    180, UEW_PULIST_HEIGHT,
	    0,
	    PULIST_POPUP_BELOW
	);
	pulist->select_cb = UEWOPMListSelectCB;
	pulist->select_data = uew;

	x += ((int)pulist->width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_select_opm_file_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_opm_file_pixmap,
	    ue_image.icon_opm_file_mask,
	    UEWTBSelectOPMFilePBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_select_opm_file_btn,
	    UEW_HINT_MESG_SELECT_OPM_FILE
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_insert_object_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_new_object_pixmap,
	    ue_image.icon_new_object_mask,
	    UEWTBInsertObjectPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_insert_object_btn,
	    UEW_HINT_MESG_NEW_OBJECT
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_properties_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_object_properties_pixmap,
	    ue_image.icon_object_properties_mask,
	    UEWTBPropertiesPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_properties_btn,
	    UEW_HINT_MESG_PROPERTIES
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_weapons_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_weapons_pixmap,
	    ue_image.icon_weapons_mask,
	    UEWTBWeaponsPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_weapons_btn,
	    UEW_HINT_MESG_WEAPONS
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MINOR);
	PBtnInit(
	    &uew->tb_economy_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_economy_pixmap,
	    ue_image.icon_economy_mask,
	    UEWTBEconomyPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_economy_btn,
	    UEW_HINT_MESG_ECONOMY
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MAJOR);
	PBtnInit(
	    &uew->tb_remodel_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_remodel_object_pixmap,
	    ue_image.icon_remodel_object_mask,
	    UEWTBRemodelPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_remodel_btn,
	    UEW_HINT_MESG_REMODEL
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MAJOR);
	PBtnInit(
	    &uew->tb_delete_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_delete_pixmap,
	    ue_image.icon_delete_mask,
	    UEWTBDeletePBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_delete_btn,
	    UEW_HINT_MESG_DELETE
	);

	x += ((int)width + UEW_TOOLBAR_MARGIN_MAJOR);
	PBtnInit(
	    &uew->tb_print_btn, parent2,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    ue_image.icon_print_pixmap,
	    ue_image.icon_print_mask,
	    UEWTBPrintPBCB, uew
	);
	PBtnSetTooltip(
	    &uew->tb_print_btn,
	    UEW_HINT_MESG_PRINT
	);


	/* Objects List */
	clist = &uew->objects_list;
	x = 0;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = (unsigned int)MAX(
	    uew->view_pos,
	    1
	);
	height = (unsigned int)MAX(
	    ((int)uew->height - y - UEW_PULIST_HEIGHT -
		UEW_PREVIEW_HEIGHT - UEW_STATUS_HEIGHT),
	    1
	);
	CListInit(
	    clist, parent,
	    x, y, width, height
	);
	clist->activate_cb = UEWObjectsCListActivateCB;
	clist->activate_data = uew;
	clist->select_cb = UEWObjectsCListSelectCB;
	clist->select_data = uew;
	clist->unselect_cb = UEWObjectsCListUnselectCB;
	clist->unselect_data = uew;

	clist->options |= CLIST_FLAG_ALLOW_MULTI_SELECT;
	CListColumnAppend(
	    clist,
	    "Index",
	    wglobal->std_font,
	    wglobal->normal_text_pix,
	    0,
	    0
	);
	CListColumnAppend(
	    clist,
	    "Name",
	    wglobal->std_font,
	    wglobal->normal_text_pix,
	    0,
	    60
	);

	/* ImgRefs Popup List */
	pulist = &uew->imgrefs_pulist;
	y += (int)height;
	height = UEW_PULIST_HEIGHT;
	PUListInit(
	    pulist, parent,
	    x, y,
	    width, height,
	    0,
	    PULIST_POPUP_BELOW
	);
	pulist->select_cb = UEWImgRefsListSelectCB;
	pulist->select_data = uew;

	/* Preview */
	x = 0;
	y = (int)uew->height - UEW_PREVIEW_HEIGHT -
	    UEW_STATUS_HEIGHT;
	width = (unsigned int)MAX(uew->view_pos, 1);
	height = UEW_PREVIEW_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->preview = w;
	uew->preview_buf = None;
	OSWSetWindowInput(w, ExposureMask);

	OSWCreateSharedImage(
	    &uew->imgref_img, width, height
	);


	/* Objects List & View Split Bar */
	x = uew->view_pos;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = UEW_SPLIT_WIDTH;
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
	    UEW_TOOLBAR_HEIGHT - UEW_STATUS_HEIGHT,
	    1
	);
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->view_split_bar = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | ButtonPressMask | ButtonReleaseMask
	);
	WidgetSetWindowCursor(w, wglobal->h_split_wcur);

	/* View */
	x = uew->view_pos + UEW_SPLIT_WIDTH;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = (unsigned int)MAX(
	    uew->properties_pos - uew->view_pos -
		UEW_SPLIT_WIDTH,
	    1
	);
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
		UEW_TOOLBAR_HEIGHT - UEW_VIEW_STATUS_HEIGHT -
		UEW_STATUS_HEIGHT,
	    1
	);
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->view = w;
	uew->view_buf = None;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask
	);
	WidgetSetWindowCursor(w, ue_cursor.scanner_lock_wcur);


	/* View Status */
	x = uew->view_pos + UEW_SPLIT_WIDTH;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT + (int)height;
	width = (unsigned int)MAX(
	    uew->properties_pos - uew->view_pos - UEW_SPLIT_WIDTH,
	    1
	);
	height = UEW_VIEW_STATUS_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->view_status = w;
	parent2 = w;

	/* View status prompts */
	x = 4;
	y = 3;
	width = 110;
	height = UEW_PROMPT_HEIGHT;
	PromptInit(
	    &uew->obj_sect_x_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Sect X:",
	    40, 0
	);

	y += UEW_PROMPT_HEIGHT + 3;
	PromptInit(
	    &uew->obj_sect_y_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Sect Y:", 
	    40, 0
	);

	y += UEW_PROMPT_HEIGHT + 3;
	PromptInit(
	    &uew->obj_sect_z_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Sect Z:",
	    40, 0
	);

	x = 128;
	y = 3;
	width = 80;
	height = UEW_PROMPT_HEIGHT;
	PromptInit(
	    &uew->obj_cx_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "X:",
	    20, 0
	);

	y += UEW_PROMPT_HEIGHT + 3;
	PromptInit(
	    &uew->obj_cy_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Y:",
	    20, 0
	);

	y += UEW_PROMPT_HEIGHT + 3;
	PromptInit(
	    &uew->obj_cz_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Z:",
	    20, 0
	);

	x = 232;
	y = 3;
	width = 130;
	height = UEW_PROMPT_HEIGHT;
	PromptInit(
	    &uew->obj_name_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    NULL,
	    SW_OBJ_NAME_MAX + 10, 0
	);

	y += UEW_PROMPT_HEIGHT + 3;
	PromptInit(
	    &uew->obj_heading_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Heading:",
	    20, 0
	);

	y += UEW_PROMPT_HEIGHT + 3;
	PromptInit(
	    &uew->obj_size_prompt, parent2,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "   Size:",
	    20, 0
	);


	/* View & Properties Split Bar */
	x = uew->properties_pos;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = UEW_SPLIT_WIDTH;
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
		UEW_TOOLBAR_HEIGHT - UEW_STATUS_HEIGHT,
	    1
	);
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->props_split_bar = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | ButtonPressMask | ButtonReleaseMask
	); 
	WidgetSetWindowCursor(w, wglobal->h_split_wcur);

	/* Properties (parent window for the Property Prompts) */
	x = uew->properties_pos + UEW_SPLIT_WIDTH;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = (unsigned int)MAX(
	    (int)uew->width - uew->properties_pos -
		UEW_SPLIT_WIDTH,
	    1
	);
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
		UEW_TOOLBAR_HEIGHT - UEW_STATUS_HEIGHT,
	    1
	);
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->props = w;
	parent2 = w;

	/* Create the property prompts */
	UEWPropsInit(uew);

	sb = &uew->props_sb;
	SBarInit(
	    sb, parent2,
	    width, height,
	    width,
	    UEW_TOTAL_PROPS * UEW_PROP_PROMPT_HEIGHT
	);
	sb->value_changed_cb = UEWPropsSBValueChangedCB;
	sb->value_changed_data = uew;


	/* Status Bar */
	x = 0;
	y = (int)uew->height - UEW_STATUS_HEIGHT;
	width = uew->width;
	height = UEW_STATUS_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uew->status = w;
	parent2 = w;

	/* Progress Bar */
	x = (int)uew->width - UEW_PBAR_WIDTH - 10;
	y = (UEW_STATUS_HEIGHT - UEW_PBAR_HEIGHT) / 2;
	width = UEW_PBAR_WIDTH;
	height = UEW_PBAR_HEIGHT;
	PBarInit(
	    &uew->pbar, parent2,
	    x, y, width, height,
	    0.0f, 0.0f, 1.0f,
	    NULL,
	    PBAR_COMPLETION_HOLD
	);

	/* Right-Click Menu */
	menu = &uew->menu;
	MenuInit(
	    menu, gui->root_win,
	    UEWMenuActivateCB, uew
	);
#define APPEND_MENU_ITEM(_t_,_s_,_id_)	{	\
 n = MenuItemAppend(				\
  menu, (_t_), (_s_), None, None, (_id_)	\
 );						\
}
#define SET_MENU_ITEM_ACCEL_KEY(_k_,_m_) {	\
 MenuSetItemAccelKey(				\
  menu, n, (_k_), (_m_)				\
 );						\
}

	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Insert",
	    UEW_OP_INSERT_OBJECT
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_i, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    UEW_OP_NONE
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Cut",
	    UEW_OP_CUT
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_x, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Copy",
	    UEW_OP_COPY
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_c, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Paste",
	    UEW_OP_PASTE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_v, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Delete",
	    UEW_OP_DELETE
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_Delete, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    UEW_OP_NONE 
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Properties...",
	    UEW_OP_OBJ_PROPERTIES
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_Return, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Weapons...",
	    UEW_OP_OBJ_WEAPONS
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_t, OSW_KEY_STATE_CTRL
	);
	APPEND_MENU_ITEM(
	    MENU_ITEM_LABEL,
	    "Economy...",
	    UEW_OP_OBJ_ECONOMY
	);
	SET_MENU_ITEM_ACCEL_KEY(
	    OSW_KEY_e, OSW_KEY_STATE_CTRL
	);

#undef APPEND_MENU_ITEM
#undef SET_MENU_ITEM_ACCEL_KEY

	return(uew);
}

/*
 *	Resizes the universe editor window.
 */
void UEWResize(uew_struct *uew)
{
	int i, x, y;
	unsigned int width, height;
	win_t w;
	menu_bar_struct *mb;
	clist_struct *clist;
	popup_list_struct *pulist;

	if(uew == NULL)
	    return;

	/* Toplevel */
	w = uew->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	uew->x = x;
	uew->y = y;
	uew->width = width;
	uew->height = height;

	/* Clip pane positions */
	if(uew->view_pos > ((int)uew->width - 1))
	    uew->view_pos = ((int)uew->width - 1);
	if(uew->view_pos < 0)
	    uew->view_pos = 0;
	if(uew->properties_pos > ((int)uew->width - 1))
	    uew->properties_pos = ((int)uew->width - 1);
	if(uew->properties_pos < 0)
	    uew->properties_pos = 0;

	/* Menu Bar */
	mb = &uew->mb;
	x = 0;
	y = 0;
	width = uew->width;
	height = mb->height;
	OSWMoveResizeWindow(
	    mb->toplevel,
	    x, y, width, height
	);
	WidgetQueueResize(mb);

	/* Move the help menu item to the right side */
	i = uew->help_menu_num;
	if((i >= 0) && (i < mb->total_items))
	{
	    menu_bar_item_struct *item = mb->item[i];
	    if(item != NULL)
		item->x = (int)uew->width - (int)item->width;
	}


	/* Tool Bar */
	w = uew->toolbar;
	x = 0;
	y = (int)mb->height;
	width = uew->width;
	height = UEW_TOOLBAR_HEIGHT;
	OSWMoveResizeWindow(w, x, y, width, height);

	/* Objects List */
	clist = &uew->objects_list;
	w = clist->toplevel;
	x = 0;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = (unsigned int)MAX(
	    uew->view_pos,
	    1
	);
	height = (unsigned int)MAX(
	    ((int)uew->height - y - UEW_PREVIEW_HEIGHT -
		UEW_PULIST_HEIGHT - UEW_STATUS_HEIGHT),
	    1
	);
	OSWMoveResizeWindow(w, x, y, width, height);
	WidgetQueueResize(clist);

	/* ImgRefs Popup List */
	pulist = &uew->imgrefs_pulist;
	w = pulist->toplevel;
	y += (int)height;
	height = UEW_PULIST_HEIGHT;
	OSWMoveResizeWindow(w, x, y, width, height);
	WidgetQueueResize(pulist);

	/* Preview and preview imgref image */
	OSWSyncSharedImage(uew->imgref_img, uew->preview);
	OSWDestroySharedImage(&uew->imgref_img);
	OSWDestroyPixmap(&uew->preview_buf);

	w = uew->preview;
	x = 0;
	y = (int)uew->height - UEW_PREVIEW_HEIGHT -
	    UEW_STATUS_HEIGHT;
	width = (unsigned int)MAX(
	    uew->view_pos,
	    1
	);
	height = UEW_PREVIEW_HEIGHT;
	OSWMoveResizeWindow(w, x, y, width, height);

	/* Preview image */
	OSWCreateSharedImage(&uew->imgref_img, width, height);


	/* View split bar */
	w = uew->view_split_bar;
	x = uew->view_pos;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = UEW_SPLIT_WIDTH;
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
		UEW_TOOLBAR_HEIGHT - UEW_STATUS_HEIGHT,
	    1
	);
	OSWMoveResizeWindow(w, x, y, width, height);

	/* View */
	w = uew->view;
	x = uew->view_pos + UEW_SPLIT_WIDTH;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = (unsigned int)MAX(
	    uew->properties_pos - uew->view_pos -
		UEW_SPLIT_WIDTH,
	    1
	);
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
		UEW_TOOLBAR_HEIGHT - UEW_VIEW_STATUS_HEIGHT -
		UEW_STATUS_HEIGHT,
	    1
	);
	OSWMoveResizeWindow(w, x, y, width, height);

	OSWDestroyPixmap(&uew->view_buf);


	/* View status */
	w = uew->view_status;
	x = uew->view_pos + UEW_SPLIT_WIDTH;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT + (int)height;
	width = (unsigned int)MAX(
	    uew->properties_pos - uew->view_pos -
		UEW_SPLIT_WIDTH,
	    1
	);
	height = UEW_VIEW_STATUS_HEIGHT;
	OSWMoveResizeWindow(w, x, y, width, height);

	x = 4;
	y = 3;
	OSWMoveWindow(
	    uew->obj_sect_x_prompt.toplevel, x, y
	);
	y += UEW_PROMPT_HEIGHT + 3;
	OSWMoveWindow(
	    uew->obj_sect_y_prompt.toplevel, x, y
	);
	y += UEW_PROMPT_HEIGHT + 3;
	OSWMoveWindow(
	    uew->obj_sect_z_prompt.toplevel, x, y
	);

	x = 128;
	y = 3;
	OSWMoveWindow(
	    uew->obj_cx_prompt.toplevel, x, y
	);
	y += UEW_PROMPT_HEIGHT + 3;
	OSWMoveWindow(
	    uew->obj_cy_prompt.toplevel, x, y
	);
	y += UEW_PROMPT_HEIGHT + 3;
	OSWMoveWindow(
	    uew->obj_cz_prompt.toplevel, x, y
	);

	x = 232;
	y = 3;
	OSWMoveWindow(
	    uew->obj_name_prompt.toplevel, x, y
	);
	y += UEW_PROMPT_HEIGHT + 3;
	OSWMoveWindow(
	    uew->obj_heading_prompt.toplevel, x, y
	);
	y += UEW_PROMPT_HEIGHT + 3;
	OSWMoveWindow(
	    uew->obj_size_prompt.toplevel, x, y
	);


	/* Properties Split Bar */
	w = uew->props_split_bar;
	x = uew->properties_pos;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = UEW_SPLIT_WIDTH;
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
		UEW_TOOLBAR_HEIGHT - UEW_STATUS_HEIGHT,
	    1
	);
	OSWMoveResizeWindow(w, x, y, width, height);

	/* Property Prompts Parent */
	w = uew->props;
	x = uew->properties_pos + UEW_SPLIT_WIDTH;
	y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
	width = (unsigned int)MAX(
	    (int)uew->width - uew->properties_pos -
		UEW_SPLIT_WIDTH,
	    1
	);
	height = (unsigned int)MAX(
	    (int)uew->height - (int)mb->height -
		UEW_TOOLBAR_HEIGHT - UEW_STATUS_HEIGHT,
	    1
	);
	OSWMoveResizeWindow(w, x, y, width, height);

	/* Property Prompts */
	UEWPropsResize(uew);

	/* Scroll Bar */
	SBarQueueResize(
	    &uew->props_sb,
	    width, height,
	    width,
	    UEW_TOTAL_PROPS * UEW_PROP_PROMPT_HEIGHT
	);
	SBarQueueDraw(&uew->props_sb);
	UEWPropsDraw(uew, 0);


	/* Status bar */
	w = uew->status;
	x = 0;
	y = (int)uew->height - UEW_STATUS_HEIGHT;
	width = uew->width;
	height = UEW_STATUS_HEIGHT;
	OSWMoveResizeWindow(w, x, y, width, height);

	/* Progress bar */
	x = (int)uew->width - UEW_PBAR_WIDTH - 10;
	y = (UEW_STATUS_HEIGHT - UEW_PBAR_HEIGHT) / 2;
	OSWMoveWindow(
	    uew->pbar.toplevel, x, y
	);
}

/*
 *	Queues the universe editor window to resize.
 */
void UEWQueueResize(uew_struct *uew)
{
	if(uew == NULL)
	    return;

	WidgetQueueResizeAction(
	    UEWResizeCB, uew, NULL
	);
}

/*
 *	Draws a vertical resize bar.
 */
static void UEWDrawResizeBarV(
	uew_struct *uew,
	const int x, const int y, const unsigned int len,
	const int last_x, const bool_t has_previous
)
{
#if defined(HAVE_X)
	unsigned int width, height;
	win_t w;
	gc_t gc;
	osw_gui_struct *gui = &osw_gui[0];

	if(uew == NULL)
	    return;

	w = uew->toplevel;
	gc = uew->invert_gc;
	if((w == None) || (gc == None))
	    return;

	OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

	/* Need to clear previous draw? */
	if(has_previous)
	{
	    XDrawLine(
		gui->display, (Drawable)w, gc,
		last_x, y,
		last_x, y + (int)len
	    );
	}

	XDrawLine(
	    gui->display, (Drawable)w, gc,
	    x, y,
	    x, y + (int)len
	);
#endif
}

/*
 *	Draws the universe editor.
 */
void UEWDraw(uew_struct *uew, const int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(uew == NULL)
	    return;

	/* Object display */
	if((amount == UEW_DRAW_COMPLETE) ||
	   (amount == UEW_DRAW_PREVIEW)
	)
	{
	    int imgref_num;
	    unsigned int width, height;
	    pixmap_t pixmap;
	    win_t w = uew->preview;
	    imgref_struct *imgref;

	    if(uew->preview_buf == None)
	    {
		OSWGetWindowGeometry(
		    w, NULL, NULL, &width, &height, NULL
		);
		OSWCreatePixmap(&uew->preview_buf, width, height);
	    }
	    pixmap = uew->preview_buf;

	    /* Get the currently displayed imgref */
	    imgref_num = uew->cur_imgref;
	    imgref = ((imgref_num >= 0) && (imgref_num < uew->total_imgrefs)) ?
		uew->imgref[imgref_num] : NULL;

	    UEWObjectDisplayDrawPixmap(
		pixmap, uew->imgref_img,
		UEWGetObjectPtr(uew, uew->last_selected_object),
		imgref, uew->ani_frame
	    );

	    OSWPutDrawableToWindow(w, pixmap);
	}

	/* Tool Bar */
	if((amount == UEW_DRAW_COMPLETE) ||
	   (amount == UEW_DRAW_TOOLBAR)
	)
	{
	    unsigned int width, height;
	    win_t w = uew->toolbar;

	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    WidgetFrameButtonPixmap(
		w, False,
		width, height,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );
	}


	/* Objects List Split Bar */
	if(amount == UEW_DRAW_COMPLETE)
	{
	    unsigned int width, height;
	    win_t w = uew->view_split_bar;

	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    WidgetFrameButtonPixmap(
		w, False,
		width, height,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );
	}


	/* View */
	if((amount == UEW_DRAW_COMPLETE) ||
	   (amount == UEW_DRAW_VIEW)
	)
	{
	    UEWViewDraw(uew);
	}


	/* View Status */
	if((amount == UEW_DRAW_COMPLETE) ||
	   (amount == UEW_DRAW_VIEW_STATUS)
	)
	{
	    unsigned int width, height;
	    win_t w = uew->view_status;

	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile( 
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    WidgetFrameButtonPixmap(
		w, False,
		width, height,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );
	}


	/* View Split Bar */
	if(amount == UEW_DRAW_COMPLETE)
	{
	    unsigned int width, height;
	    win_t w = uew->props_split_bar;

	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    WidgetFrameButtonPixmap(
		w, False,
		width, height,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );
	}


	/* Status Bar */
	if((amount == UEW_DRAW_COMPLETE) ||
	   (amount == UEW_DRAW_STATUS)
	)
	{
	    unsigned int width, height;
	    int lbearing, ascent, descent, font_height;
	    font_t *font = wglobal->std_font;
	    win_t w = uew->status;

	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
 
	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    /* Draw status message */
	    if(!STRISEMPTY(uew->status_message))
	    {
		int x2, y2;
		const char *s = uew->status_message;
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->black_pix :
			wglobal->normal_text_pix
		);
		OSWSetFont(font);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		x2 = 2 + 5;
		y2 = ((int)height - font_height) / 2;
		OSWDrawString(w, x2 - lbearing, y2 + ascent, s);
	    }

	    WidgetFrameButtonPixmap(
	        w, False,
		width, height,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );
	}
}

/*
 *	Queues the universe editor window to draw.
 */
void UEWQueueDraw(uew_struct *uew, const int amount)
{
	if(uew == NULL)
	    return;

	WidgetQueueDrawAction(
	    UEWDrawCB, uew, (void *)amount
	);
}

/*
 *	Manages the universe editor window.
 */
int UEWManage(uew_struct *uew, event_t *event)
{
	int	x, y, root_x, root_y,
		events_handled = 0;
	time_t t;
	atom_t action;
	win_t w, src_w;
	prompt_struct *prompt;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if((uew == NULL) || (event == NULL))
	    return(events_handled);

#define MANAGE_WIDGET(_w_)	{		\
 if(events_handled == 0)			\
  events_handled += WidgetManage((_w_), event);	\
}

	/* Universe Header Window */
	if(events_handled == 0)
	    events_handled += UHWManage(uew->uhw, event);

	/* Property Dialogs */
	if(events_handled == 0)
	{
	    int i;
	    for(i = 0; i < uew->total_prop_dlgs; i++)
	    {
		events_handled += PropDlgManage(
		    uew->prop_dlg[i], event
		);
		if(events_handled > 0)
		    break;
	    }
	}

	/* Weapon Windows */
	if(events_handled == 0)
	{
	    int i;
	    for(i = 0; i < uew->total_wepws; i++)
	    {
		events_handled += WepWManage(
		    uew->wepw[i], event
		);
		if(events_handled > 0)
		    break;
	    }
	}

	/* Economy Windows */
	if(events_handled == 0)
	{
	    int i;
	    for(i = 0; i < uew->total_ecows; i++)
	    {
		events_handled += EcoWManage(
		    uew->ecow[i], event
		);
		if(events_handled > 0)
		    break;
	    }
	}

	/* Print Dialog */
	if(events_handled == 0)
	    events_handled += PrintDlgManage(uew->print_dlg, event);

	if(events_handled > 0)
	    return(events_handled);

	if(!uew->has_focus &&
	   ((event->type == KeyPress) || (event->type == KeyRelease))
	)
	    return(events_handled);

	/* Menu Bar */
	MANAGE_WIDGET(&uew->mb);
	/* Right-Click Menu */
	MANAGE_WIDGET(&uew->menu);
	if(events_handled > 0)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == uew->toplevel)
	    {
		UEWQueueResize(uew);
		if(uew->map_state)
		    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == uew->toplevel)
	    {
		UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		events_handled++;
	    }
	    else if(w == uew->toolbar)
	    {
		UEWQueueDraw(uew, UEW_DRAW_TOOLBAR);
		events_handled++;
	    }
	    else if(w == uew->preview)
	    {
		UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		events_handled++;
	    }
	    else if(w == uew->view_split_bar)
	    {
		UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		events_handled++;
	    }
	    else if(w == uew->props_split_bar)
	    {
		UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		events_handled++;
	    }
	    else if(w == uew->view)
	    {
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }
	    else if(w == uew->status)
	    {
		UEWQueueDraw(uew, UEW_DRAW_STATUS);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!uew->has_focus)
		return(events_handled);

	    key = &event->xkey;
	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Translate */
	    if(keycode == keymap[UE_KM_VIEW_TRANSLATE].keycode)
	    {
		uew->drag_op = UEW_DRAG_OP_VIEW_TRANSLATE;
		WidgetSetWindowCursor(
		    uew->view,
		    ue_cursor.translate_wcur
		);
		if(uew->view_has_focus)
		{
		    UEWStatus(
			uew,
"Press and drag the pointer on the view to translate the position."
		    );
		}
		OSWKBAutoRepeatOff();
		events_handled++;
	    }
	    /* Zoom */
	    else if(keycode == keymap[UE_KM_VIEW_ZOOM].keycode)
	    {
		uew->drag_op = UEW_DRAG_OP_VIEW_ZOOM;
		/* Set cursor on view */
		WidgetSetWindowCursor(
		    uew->view,
		    ue_cursor.zoom_wcur
		);
		if(uew->view_has_focus)
		{
		    UEWStatus(
			uew,
"Press and drag the pointer up or down on the view to adjust the zoom."
		    );
		}
		OSWKBAutoRepeatOff();
		events_handled++;
	    }
	    /* Select Rectanglular */
	    else if(keycode == keymap[UE_KM_VIEW_SELECT_RECTANGULAR].keycode)
	    {
		uew->drag_op = UEW_DRAG_OP_VIEW_SELECT_RECTANGULAR;
		if(uew->view_has_focus)
		{
		    UEWStatus(
			uew,
"Press and drag the pointer on the view to select objects in a\
 rectangular area."
		    );
		}
		OSWKBAutoRepeatOff();
		events_handled++;
	    }
	    /* Move object */
	    else if(keycode == keymap[UE_KM_VIEW_MOVE_OBJECT].keycode)
	    {
		uew->drag_op = UEW_DRAG_OP_VIEW_MOVE_OBJECT;
		if(uew->view_has_focus)
		{
		    UEWStatus(
			uew,
"Press and drag the pointer on the view to move the selected object."
		    );
		}
		OSWKBAutoRepeatOff();
		events_handled++;
	    }
	    /* Left */
	    else if((keyval == OSW_KEY_Left) &&
	            uew->view_has_focus
	    )
	    {
		int i;
		unsigned int width, height;
		const float inc_coeff =
		    (key_state & OSW_KEY_STATE_SHIFT) ? 0.1f : 0.01f;
		sw_legend_struct *legend = &uew->legend;
		OSWGetWindowGeometry(uew->view, 0, 0, &width, &height, NULL);
		if(uew->zoom > 0.0f)
		    uew->cx -= (width * inc_coeff) / uew->zoom;
		for(i = 0; uew->cx < SW_SECTOR_SIZE_X_MIN; i++)
		    uew->cx += legend->sect_x_len;
		uew->sect_x -= i;
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }
	    /* Right */
	    else if((keyval == OSW_KEY_Right) &&
	            uew->view_has_focus
	    )
	    {
		int i;
		unsigned int width, height;
		const float inc_coeff =
		    (key_state & OSW_KEY_STATE_SHIFT) ? 0.1f : 0.01f;
		sw_legend_struct *legend = &uew->legend;
		OSWGetWindowGeometry(uew->view, 0, 0, &width, &height, NULL);
		if(uew->zoom > 0.0f)
		    uew->cx += (width * inc_coeff) / uew->zoom;
		for(i = 0; uew->cx > SW_SECTOR_SIZE_X_MAX; i++)
		    uew->cx -= legend->sect_x_len;
		uew->sect_x += i;
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }
	    /* Up */
	    else if((keyval == OSW_KEY_Up) &&
		    uew->view_has_focus
	    )
	    {
		int i;
		unsigned int width, height;
		const float inc_coeff =    
		    (key_state & OSW_KEY_STATE_SHIFT) ? 0.1f : 0.01f;
		sw_legend_struct *legend = &uew->legend;
		OSWGetWindowGeometry(uew->view, 0, 0, &width, &height, NULL);
		if(uew->zoom > 0.0f)
		    uew->cy += (height * inc_coeff) / uew->zoom;
		for(i = 0; uew->cy > SW_SECTOR_SIZE_Y_MAX; i++)
		    uew->cy -= legend->sect_y_len;
		uew->sect_y += i;
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }
	    /* Down */
	    else if((keyval == OSW_KEY_Down) &&
	       uew->view_has_focus
	    )
	    {
		int i;
		unsigned int width, height;
		const float inc_coeff =
		    (key_state & OSW_KEY_STATE_SHIFT) ? 0.1f : 0.01f;
		sw_legend_struct *legend = &uew->legend;
		OSWGetWindowGeometry(uew->view, 0, 0, &width, &height, NULL);
		if(uew->zoom > 0.0f)
		    uew->cy -= (height * inc_coeff) / uew->zoom;
		for(i = 0; uew->cy < SW_SECTOR_SIZE_Y_MIN; i++)
		    uew->cy += legend->sect_y_len;
		uew->sect_y -= i;
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }
	    /* Zoom in */
	    else if(((keyval == OSW_KEY_equal) ||
		     (keyval == OSW_KEY_plus) ||
		     (keyval == OSW_KEY_KP_Add) ||
		     (keyval == OSW_KEY_Page_Down) ||
		     (keyval == OSW_KEY_KP_Page_Down)
		    ) && uew->view_has_focus
	    )
	    {
		const float zoom_effect = MAX(uew->zoom, UEW_VIEW_ZOOM_MIN) *
		    ((key_state & OSW_KEY_STATE_SHIFT) ? 5 : 1);
		uew->zoom = CLIP(
		    (uew->zoom + (0.05f * zoom_effect)),
		    UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX
		);
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }
	    /* Zoom out */
	    else if(((keyval == OSW_KEY_minus) ||
		     (keyval == OSW_KEY_underscore) ||
		     (keyval == OSW_KEY_KP_Subtract) ||
		     (keyval == OSW_KEY_Page_Up) ||
		     (keyval == OSW_KEY_KP_Page_Up)
		    ) && uew->view_has_focus
	    )
	    {
		const float zoom_effect = MAX(uew->zoom, UEW_VIEW_ZOOM_MIN) *
		    ((key_state & OSW_KEY_STATE_SHIFT) ? 5 : 1);
		uew->zoom = CLIP(
		    (uew->zoom - (0.05f * zoom_effect)),
		    UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX
		);
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }

	    /* Enter (apply changes from focused prompt) */
	    else if((keyval == OSW_KEY_Return) ||
		    (keyval == OSW_KEY_KP_Enter)
	    )
	    {
		clist_struct *clist = &uew->objects_list;

#define STATUS_PROPERTY_VALUES_APPLIED(_obj_,_obj_num_)	{	\
 const char *name = UNVGetObjectFormalName(			\
  (_obj_num_), (_obj_)						\
 );								\
 char *s = (char *)malloc(STRLEN(name) + 80);			\
 sprintf(							\
  s,								\
"Applied property values to object `%s'.",			\
  name								\
 );								\
 UEWStatus(uew, s);						\
 free(s);							\
}

		/* Apply prompt values */
		if(uew->obj_sect_x_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const long sx = PromptGetL(&uew->obj_sect_x_prompt);
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->sect_x = sx;
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_sect_y_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const long sy = PromptGetL(&uew->obj_sect_y_prompt);
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->sect_y = sy;
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_sect_z_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const long sz = PromptGetL(&uew->obj_sect_z_prompt);
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->sect_z = sz;
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_cx_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const float x = PromptGetF(&uew->obj_cx_prompt);
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->x = x;
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_cy_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const float y = PromptGetF(&uew->obj_cy_prompt);
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->y = y;
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_cz_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const float z = PromptGetF(&uew->obj_cz_prompt);
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->z = z;
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_name_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const char *name = PromptGetS(&uew->obj_name_prompt);
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if((obj != NULL) && !STRISEMPTY(name))
		    {
			UEWPropsDoSetValues(uew, obj_num);
			strncpy(obj->name, name, sizeof(obj->name));
			obj->name[sizeof(obj->name) - 1] = '\0';
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWObjectsListItemUpdateObjName(uew, obj_num);
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_heading_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const float heading = PromptGetF(
			&uew->obj_heading_prompt
		    );
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->heading = heading * (PI / 180);
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		else if(uew->obj_size_prompt.has_focus)
		{
		    const int obj_num = uew->last_selected_object;
		    const long size = PromptGetL(
			&uew->obj_size_prompt
		    );
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			UEWPropsDoSetValues(uew, obj_num);
			obj->size = size;
			UEWPropsDoGetValues(uew, obj_num);
			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		/* Property Prompts */
		else if(UEWPropsHasFocus(uew))
		{
		    const int obj_num = uew->last_selected_object;
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			int imgref_num;
			imgref_struct *imgref;
			popup_list_struct *pulist = &uew->imgrefs_pulist;

			UEWPropsDoSetValues(uew, obj_num);

			/* Unload the previous imgref */
			imgref_num = uew->cur_imgref;
			imgref = ((imgref_num >= 0) &&
			    (imgref_num < uew->total_imgrefs)) ?
				uew->imgref[imgref_num] : NULL;
			ImgRefUnload(imgref);

			/* Set the new imgref */
			uew->cur_imgref = imgref_num = obj->imgref;
			imgref = ((imgref_num >= 0) &&
			    (imgref_num < uew->total_imgrefs)) ?
				uew->imgref[imgref_num] : NULL;
			ImgRefLoad(imgref);

			/* Update the displayed item on the ImgRefs Popup List */
			uew->freeze_count++;
			PUListSetSelItem(
			   pulist,
			   PUListGetItemByData(pulist, (void *)imgref_num)
			);
			uew->freeze_count--;

		        UEWViewStatusGetValues(uew, obj);

			UEWQueueDraw(uew, UEW_DRAW_VIEW);
			UEWQueueDraw(uew, UEW_DRAW_PREVIEW);

			STATUS_PROPERTY_VALUES_APPLIED(obj, obj_num);
		    }
		    events_handled++;
		}

		if(events_handled > 0)
		{
		    if(!uew->has_changes)
		    {
			uew->has_changes = True;
			UEWUpdateTitle(uew);
		    }
		}
#undef STATUS_PROPERTY_VALUES_APPLIED
	    }
	    break;

	  case KeyRelease:
	    if(!uew->has_focus)
		return(events_handled);

	    key = &event->xkey;
	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Translate */
	    if(keycode == keymap[UE_KM_VIEW_TRANSLATE].keycode)
	    {
		if(uew->drag_op != UEW_DRAG_OP_NONE)
		{
		    uew->drag_op = UEW_DRAG_OP_NONE;
		    WidgetSetWindowCursor(
			uew->view,
			ue_cursor.scanner_lock_wcur
		    );
		    UEWStatus(uew, NULL);
		    OSWKBAutoRepeatOn();
		}
		events_handled++;
	    }
	    /* Zoom */
	    else if(keycode == keymap[UE_KM_VIEW_ZOOM].keycode)
	    {
		if(uew->drag_op != UEW_DRAG_OP_NONE)
		{
		    uew->drag_op = UEW_DRAG_OP_NONE;
		    WidgetSetWindowCursor(
			uew->view,
			ue_cursor.scanner_lock_wcur
		    );
		    UEWStatus(uew, NULL);
		    OSWKBAutoRepeatOn();
		}
		events_handled++;
	    }
	    /* Select Rectangular */
	    else if(keycode == keymap[UE_KM_VIEW_SELECT_RECTANGULAR].keycode)
	    {
		if(uew->drag_op != UEW_DRAG_OP_NONE)
		{
		    uew->drag_op = UEW_DRAG_OP_NONE;
		    WidgetSetWindowCursor(
			uew->view,
			ue_cursor.scanner_lock_wcur
		    );
		    UEWStatus(uew, NULL);
		    OSWKBAutoRepeatOn();
		}
		events_handled++;
	    }
	    /* Move object */
	    else if(keycode == keymap[UE_KM_VIEW_MOVE_OBJECT].keycode)
	    {
		if(uew->drag_op != UEW_DRAG_OP_NONE)
		{
		    uew->drag_op = UEW_DRAG_OP_NONE;
		    WidgetSetWindowCursor(
			uew->view,
			ue_cursor.scanner_lock_wcur
		    );
		    UEWStatus(uew, NULL);
		    OSWKBAutoRepeatOn();
		}
		events_handled++;
	    }
	    /* Open universe */
	    else if(keyval == OSW_KEY_F3)
	    {
		UEWMenuActivateCB(
		    NULL,
		    UEW_OP_OPEN,
		    uew
		);
		events_handled++;
	    }
	    /* Copy */
	    else if((key_state & OSW_KEY_STATE_CTRL) &&
		    (keyval == OSW_KEY_Insert)
	    )
	    {
		UEWMenuActivateCB(
		    NULL,
		    UEW_OP_COPY,
		    uew
		);
		events_handled++;
	    }
	    /* Paste */
	    else if((key_state & OSW_KEY_STATE_SHIFT) &&
		    (keyval == OSW_KEY_Insert)
	    )
	    {
		UEWMenuActivateCB(
		    NULL,
		    UEW_OP_PASTE,
		    uew
		);
		events_handled++;
	    }
	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    /* Objects list */
	    if(w == uew->objects_list.list)
	    {
		const int button_num = button->button;
		switch(button_num)
		{
		  case Button3:
		    if(True)
		    {
			int x, y;
			OSWGetPointerCoords(
			    gui->root_win,
			    &x, &y,
			    NULL, NULL
			);
			MenuMapCoordinates(&uew->menu, x, y);
			events_handled++;
		    }
		    break;
		}
	    }
	    /* View */
	    else if(w == uew->view)
	    {
		const int	button_num = button->button,
				x = button->x,
				y = button->y;
		int obj_num;
		clist_struct *clist = &uew->objects_list;

		uew->view_has_focus = True;
		uew->button_pressed = button->button;
		uew->button_last_x = x;
		uew->button_last_y = y;

		if(uew->drag_op == UEW_DRAG_OP_NONE)
		{
		    switch(button_num)
		    {
		      case Button1:
			/* Select object */
			if(uew->last_selected_object > -1)
		        {
			    obj_num = UEWViewMatchObject(
			        uew, x, y,
			        uew->last_selected_object + 1
			    );
			    if(obj_num < 0)
			        obj_num = UEWViewMatchObject(
				    uew, x, y,
				    0
			        );
		        }
		        else
		        {
			    obj_num = UEWViewMatchObject(
			        uew, x, y,
			        0
			    );
		        }
		        if((obj_num > -1) && (obj_num != uew->last_selected_object))
		        {
			    const int row_num = obj_num;
			    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
				clist, row_num
			    ));

			    CListUnselectAllRows(clist);
			    CListSelectRow(clist, row_num);

			    /* Update the view status prompts */
			    UEWViewStatusGetValues(uew, obj);
		        }
		        break;
		      case Button2:
			/* Measure */
			uew->drag_op = UEW_DRAG_OP_VIEW_MEASURE;
			OSWGrabPointerTime(
			    uew->toplevel,
			    True,
			    ButtonReleaseMask | PointerMotionMask,
			    GrabModeAsync, GrabModeAsync,
			    None,
			    None,
			    button->time
			);
			UEWQueueDraw(uew, UEW_DRAW_VIEW);
			break;
		      case Button3:
			if(True)
			{
			    int x, y;
			    OSWGetPointerCoords(
				gui->root_win,
				&x, &y,
				NULL, NULL
			    );
			    MenuMapCoordinates(&uew->menu, x, y);

			    /* Need to reset the button pressed */
			    uew->button_pressed = 0;
			}
		        break;
		      case Button4:
			if(True)
			{
			    const float zoom_effect = MAX(uew->zoom, UEW_VIEW_ZOOM_MIN) *
				((button->state & OSW_KEY_STATE_SHIFT) ? 5 : 1);
			    uew->zoom = CLIP(
				(uew->zoom + (0.05f * zoom_effect)),
				UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX
			    );
			    UEWQueueDraw(uew, UEW_DRAW_VIEW);
			}
			break;
		      case Button5:
			if(True)
			{
			    const float zoom_effect = MAX(uew->zoom, UEW_VIEW_ZOOM_MIN) *
				((button->state & OSW_KEY_STATE_SHIFT) ? 5 : 1);
			    uew->zoom = CLIP(
				(uew->zoom - (0.05f * zoom_effect)),
				UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX
			    );
			    UEWQueueDraw(uew, UEW_DRAW_VIEW);
			}
			break;
		    }
		}

		events_handled++;
	    }
	    /* View pane split position drag start */
	    else if(w == uew->view_split_bar)
	    {
		menu_bar_struct *mb = &uew->mb;
		const int	x = button->x,
				y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
		WCursor *wcur = wglobal->h_split_wcur;
		uew->view_has_focus = False;
		uew->drag_op = UEW_DRAG_OP_RESIZE_LIST;
		uew->button_pressed = button->button;
		uew->button_pressed_time = (time_t)button->time;
		OSWGrabPointerTime(
		    uew->toplevel,
		    True,
		    ButtonReleaseMask | PointerMotionMask,
		    GrabModeAsync, GrabModeAsync,
		    None,
		    (wcur != NULL) ? wcur->cursor : None,
		    button->time
		);
		UEWDrawResizeBarV(
		    uew,
		    x, y,
		    (unsigned int)MAX(
			(int)uew->height - y - UEW_STATUS_HEIGHT,
			1
		    ),
		    uew->resize_bar_last_pos, False
		);
		uew->resize_bar_last_pos = x;
		events_handled++;
	    }
	    /* Properties pane split position drag start */
	    else if(w == uew->props_split_bar)
	    {
		menu_bar_struct *mb = &uew->mb;
		const int	x = button->x,
				y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
		WCursor *wcur = wglobal->h_split_wcur;
		uew->view_has_focus = False;
		uew->drag_op = UEW_DRAG_OP_RESIZE_VIEW;
		uew->button_pressed = button->button;
		uew->button_pressed_time = (time_t)button->time;
		OSWGrabPointerTime(
		    uew->toplevel,
		    True,
		    ButtonReleaseMask | PointerMotionMask,
		    GrabModeAsync, GrabModeAsync,
		    None,
		    (wcur != NULL) ? wcur->cursor : None,
		    button->time
		);
		UEWDrawResizeBarV(
		    uew,
		    x, y,
		    (unsigned int)MAX(
			(int)uew->height - y - UEW_STATUS_HEIGHT,
			1
		    ),
		    uew->resize_bar_last_pos, False
		);
		uew->resize_bar_last_pos = x;
		events_handled++;
	    }
	    else
	    {
		uew->view_has_focus = False;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    switch(uew->drag_op)
	    {
	      case UEW_DRAG_OP_RESIZE_LIST:
		if(uew->button_pressed > 0)
		{
		    /* Resize the Objects List & View */
		    menu_bar_struct *mb = &uew->mb;
		    const int	x = button->x,
				y = (int)mb->height + UEW_TOOLBAR_HEIGHT;

		    uew->drag_op = UEW_DRAG_OP_NONE;
		    uew->view_pos = MIN(
			x,
			(uew->properties_pos - 10)
		    );

		    UEWDrawResizeBarV(
			uew,
			uew->resize_bar_last_pos, y,
			(unsigned int)MAX(
			    (int)uew->height - y - UEW_STATUS_HEIGHT,
			    1
			),
			uew->resize_bar_last_pos, False
		    );
		    uew->resize_bar_last_pos = 0;

		    UEWQueueResize(uew);
		    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_RESIZE_VIEW:
		if(uew->button_pressed > 0)
		{
		    /* Resize the View & Property Prompts */
		    menu_bar_struct *mb = &uew->mb;
		    const int	x = button->x,
				y = (int)mb->height + UEW_TOOLBAR_HEIGHT;

		    uew->drag_op = UEW_DRAG_OP_NONE;
		    uew->properties_pos = MIN(
			x,
			((int)uew->width - 10)
		    );

		    UEWDrawResizeBarV(
			uew,
			uew->resize_bar_last_pos, y,
			(unsigned int)MAX(
			    (int)uew->height - y - UEW_STATUS_HEIGHT,
			    1
			),
			uew->resize_bar_last_pos, False
		    );
		    uew->resize_bar_last_pos = 0;

		    UEWQueueResize(uew);
		    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_TRANSLATE:
		if(uew->button_pressed > 0)
		{
		    /* Translated the view */
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_ZOOM:
		if(uew->button_pressed > 0)
		{
		    /* Zoomed the view */
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_SELECT_RECTANGULAR:
		if(uew->button_pressed > 0)
		{
		    /* Select a rectangular area of objects */
		    const int	xw_min = MIN(uew->button_last_x, button->x),
				xw_max = MAX(uew->button_last_x, button->x),
				yw_min = MIN(uew->button_last_y, button->y),
				yw_max = MAX(uew->button_last_y, button->y);
		    const float	cx = uew->cx,
				cy = uew->cy;
		    const long	csect_x = uew->sect_x,
				csect_y = uew->sect_y;
		    int row_num, obj_num, obj_xw, obj_yw;
		    win_t w = uew->view;
		    unsigned int width, height;
		    clist_struct *clist = &uew->objects_list;
		    sw_legend_struct *legend = &uew->legend;
		    xsw_object_struct *obj;

		    uew->drag_op = UEW_DRAG_OP_NONE;

		    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

		    /* Unselect all existing objects */
		    CListUnselectAllRows(clist);

		    /* Iterate through the objects list and select the
		     * objects that fall within the rectangle
		     *
		     * The rectangle's bounds are in window coordinates
		     * and all checks will be done in window
		     * coordinates
		     */
		    for(obj_num = 0; obj_num < clist->total_rows; obj_num++)
		    {
			row_num = obj_num;
			obj = SW_OBJECT(CListRowGetData(clist, row_num));
			if(obj == NULL)
			    continue;

			/* Convert this object's position to window
			 * coordinates
			 */
			UEWViewConvertPositionToWindowCoordinates(
			    width, height,
			    legend->sect_x_len, legend->sect_y_len,
			    csect_x, csect_y,
			    cx, cy,
			    obj->sect_x, obj->sect_y,
			    obj->x, obj->y,
			    uew->zoom,
			    &obj_xw, &obj_yw
			);

			/* Does this object's position in window
			 * coordinates fall within the rectangle?
			 */
			if((obj_xw >= xw_min) && (obj_xw <= xw_max) &&
			   (obj_yw >= yw_min) && (obj_yw <= yw_max)
			)
			    CListSelectRow(clist, row_num);
		    }

		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_MOVE_OBJECT:
		if(uew->button_pressed > 0)
		{
		    /* Move the last selected object */
		    clist_struct *clist = &uew->objects_list;
		    const int obj_num = uew->last_selected_object;
		    xsw_object_struct *obj = SW_OBJECT(CListRowGetData(
			clist, obj_num
		    ));
		    if(obj != NULL)
		    {
			/* Record the object's current position */
			const float	cx = obj->x,
					cy = obj->y,
					cz = obj->z;
			const long	sect_x = obj->sect_x,
					sect_y = obj->sect_y,
					sect_z = obj->sect_z;

			UEWPropsDoSetValues(uew, obj_num);

			/* Restore the object's position */
			obj->x = cx;
			obj->y = cy;
			obj->z = cz;
			obj->sect_x = sect_x;
			obj->sect_y = sect_y;
			obj->sect_z = sect_z;

			UEWPropsDoGetValues(uew, obj_num);

			UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    }
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_MEASURE:
		if(uew->button_pressed > 0)
		{
		    uew->drag_op = UEW_DRAG_OP_NONE;
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	    }

	    if(uew->button_pressed == (int)button->button)
		uew->button_pressed = 0;

	    if(events_handled > 0)
	    {
		if(OSWIsPointerGrabbed())
		    OSWUngrabPointerTime(button->time);
	    }

	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    switch(uew->drag_op)
	    {
	      case UEW_DRAG_OP_RESIZE_LIST:
		if(uew->button_pressed > 0)
		{
		    menu_bar_struct *mb = &uew->mb;
		    const int	x = motion->x,
				y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
		    UEWDrawResizeBarV(
			uew,
			x, y,
			(unsigned int)MAX(
			    (int)uew->height - y - UEW_STATUS_HEIGHT,
			    1
			),
			uew->resize_bar_last_pos, True
		    );
		    uew->resize_bar_last_pos = x;
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_RESIZE_VIEW:
		if(uew->button_pressed > 0)
		{
		    menu_bar_struct *mb = &uew->mb;
		    const int   x = motion->x,
				y = (int)mb->height + UEW_TOOLBAR_HEIGHT;
		    UEWDrawResizeBarV(
			uew,
			x, y,
			(unsigned int)MAX(
			    (int)uew->height - y - UEW_STATUS_HEIGHT,
			    1
			),
			uew->resize_bar_last_pos, True
		    );
		    uew->resize_bar_last_pos = x;
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_TRANSLATE:
		if(uew->button_pressed > 0)
		{
		    const int	x = motion->x,
				y = motion->y,
				dx = x - uew->button_last_x,
				dy = y - uew->button_last_y;;
		    const float zoom = MAX(uew->zoom, UEW_VIEW_ZOOM_MIN);
		    int i;
		    sw_legend_struct *legend = &uew->legend;

		    /* Translate view */
		    uew->cx -= dx / zoom;
		    uew->cy += dy / zoom;

		    /* Change to sectors */
		    for(i = 0; uew->cx > SW_SECTOR_SIZE_X_MAX; i++)
			uew->cx -= legend->sect_x_len;
		    uew->sect_x += i;
		    for(i = 0; uew->cx < SW_SECTOR_SIZE_X_MIN; i++)
			uew->cx += legend->sect_x_len;
		    uew->sect_x -= i;

		    for(i = 0; uew->cy > SW_SECTOR_SIZE_Y_MAX; i++)
			uew->cy -= legend->sect_y_len;
		    uew->sect_y += i;
		    for(i = 0; uew->cy < SW_SECTOR_SIZE_Y_MIN; i++)
			uew->cy += legend->sect_y_len;
		    uew->sect_y -= i;

		    uew->button_last_x = x;
		    uew->button_last_y = y;
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_ZOOM:
		if(uew->button_pressed > 0)
	        {
#if 1
		    const int   x = motion->x,
				y = motion->y,
				dy = y - uew->button_last_y;
		    const float cur_zoom = MAX(uew->zoom, UEW_VIEW_ZOOM_MIN);
		    unsigned int width, height;
		    float rheight;

		    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

		    rheight = (float)((int)height - (3 * dy)) / cur_zoom;
 		    if(rheight > 0.0f)
			uew->zoom = CLIP(
			    ((float)height / rheight),
			    UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX
			);
#else
		    const int	x = motion->x,
				y = motion->y,
				dy = y - uew->button_last_y;
		    const float zoom_effect = MAX(uew->zoom, 0.00001f);

		    uew->zoom = CLIP(
			(uew->zoom + (dy * 0.01f * zoom_effect)),
			UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX
		    );
#endif
		    uew->button_last_x = x;
		    uew->button_last_y = y;
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_SELECT_RECTANGULAR:
		if(uew->button_pressed > 0)
		{
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_MOVE_OBJECT:
		if(uew->button_pressed > 0)
		{
		    /* Move all selected objects */
		    const int	x = motion->x,
				y = motion->y,
				dx = x - uew->button_last_x,
				dy = y - uew->button_last_y;
		    const float zoom = uew->zoom;
		    int i, row_num, obj_num;
		    xsw_object_struct *obj;
		    clist_struct *clist = &uew->objects_list;
		    sw_legend_struct *legend = &uew->legend;

		    for(i = 0; i < clist->total_sel_rows; i++)
		    {
		        row_num = clist->sel_row[i];
		        obj_num = row_num;
			obj = SW_OBJECT(CListRowGetData(clist, row_num));
		        if(obj == NULL)
			    continue;

		        /* Move coordinates */
		        if(zoom > 0.0f)
			{
			    obj->x += dx / zoom;
			    obj->y -= dy / zoom;
		        }

		        /* Change to sectors */
		        for(i = 0; obj->x > SW_SECTOR_SIZE_X_MAX; i++)
			    obj->x -= legend->sect_x_len;
			obj->sect_x += i;
			for(i = 0; obj->x < SW_SECTOR_SIZE_X_MIN; i++)
			    obj->x += legend->sect_x_len;
			obj->sect_x -= i;

		        for(i = 0; obj->y > SW_SECTOR_SIZE_Y_MAX; i++)
			    obj->y -= legend->sect_y_len;
			obj->sect_y += i;
			for(i = 0; obj->y < SW_SECTOR_SIZE_Y_MIN; i++)
			    obj->y += legend->sect_y_len;
			obj->sect_y -= i;

			/* Update has changes mark as needed */
			if(!uew->has_changes)
			{
			    uew->has_changes = True;
			    UEWUpdateTitle(uew);
			}

			/* Update view status prompt values */
			if(i == (clist->total_sel_rows - 1))
		        {
			    prompt = &uew->obj_sect_x_prompt;
			    PromptSetL(prompt, obj->sect_x);
			    prompt = &uew->obj_sect_y_prompt;
			    PromptSetL(prompt, obj->sect_y);
			    prompt = &uew->obj_sect_z_prompt;
			    PromptSetL(prompt, obj->sect_z);

			    prompt = &uew->obj_cx_prompt;
			    PromptSetF2(prompt, obj->x, 2);

			    prompt = &uew->obj_cy_prompt;
			    PromptSetF2(prompt, obj->y, 2);

			    prompt = &uew->obj_cz_prompt;
			    PromptSetF2(prompt, obj->z, 2);
			}
		    }

		    uew->button_last_x = x;
		    uew->button_last_y = y;
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	      case UEW_DRAG_OP_VIEW_MEASURE:
		if(uew->button_pressed > 0)
		{
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		    events_handled++;
		}
		break;
	    }
	    break;

	  case FocusIn:
	    if(w == uew->toplevel)
	    {
		uew->has_focus = True;
		events_handled++;
	    }
	    break;

	   case FocusOut:
	    if(w == uew->toplevel)
	    {
		if(uew->drag_op != UEW_DRAG_OP_NONE)
		{
		    if(OSWIsPointerGrabbed())
			OSWUngrabPointer();
		    if(!OSWGetKBAutoRepeatState())
			OSWKBAutoRepeatOn();
		    uew->drag_op = UEW_DRAG_OP_NONE;
		}
		uew->view_has_focus = False;
		uew->has_focus = False;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(uew->toplevel, event, &t))
	    {
		OSWGrabFocusTime(uew->toplevel, t);
		events_handled++;
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(uew->toplevel, event))
	    {
		/* Has changes? */
		if(uew->has_changes)
		{
		    /* Query user to save changes before existing? */
		    const int response = UEGetResponse(
			"Confirm Save Universe",
			UE_ICON_WARNING,
"The current universe has unsaved changes,\n\
do you want to save those changes?",
			CDlgBtnNo | CDlgBtnYes | CDlgBtnCancel,
			uew->toplevel
		    );
		    if(response == CDlgYes)
		    {
			if(STRISEMPTY(uew->unv_file))
			{
			    UEDialog(
"Untitled Universe",
				UE_ICON_INFO,
"The current universe has not been saved to a specified\n\
file name. To save it under a new file name, go to\n\
File->Save As...",
				uew->toplevel
			    );
			    break;
			}

			/* Save */
			if(UEWSave(uew))
			    break;
		    }
		    else if(response == CDlgNo)
		    {
			/* No save */
		    }
		    else
		    {
			/* Cancel */
			break;
		    }

		    uew->has_changes = False;
		    UEWUpdateTitle(uew);
		}

		UEWUnmap(uew);

		events_handled++;
	    }
	    /* DND Enter? */
	    else if(OSWEventIsDNDEnter(
		uew->toplevel, event,
		&src_w,
		&uew->target_string, 1
	    ))
	    {
		events_handled++;
	    }
	    /* DND Position? */
	    else if(OSWEventIsDNDPosition(
		uew->toplevel, event,
		&src_w, &w,
		&x, &y,
		&root_x, &root_y,
		&t, &action
	    ))
	    {
		win_t prev_dnd_tar_w = uew->dnd_tar_w;
		clist_struct *clist = &uew->objects_list;

		/* Objects List */
		if((w == clist->heading) ||
		   (w == clist->list)
		)
		{
		    /* Report status as accepted */
		    uew->dnd_tar_w = w;
		    uew->dnd_action = osw_atom.dnd_action_copy;
		    uew->dnd_x = x;
		    uew->dnd_y = y;
		    if(uew->map_state)
			WidgetQueueDraw(clist);
		}
		/* View */
		else if(w == uew->view)
		{
		    /* Report status as accepted */
		    uew->dnd_tar_w = w;
		    uew->dnd_action = osw_atom.dnd_action_copy;
		    uew->dnd_x = x;
		    uew->dnd_y = y;
		    if(uew->map_state)
			UEWQueueDraw(uew, UEW_DRAW_VIEW);
		}
		/* Non-DND subwindow */
		else
		{
		    /* Report status as not accepted */
		    uew->dnd_action = None;
		    uew->dnd_tar_w = None;
		}
		/* Position now over a different DND target window? */
		if(prev_dnd_tar_w != uew->dnd_tar_w)
		{
                    if(uew->map_state)
                        UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
		}
		OSWDNDSendStatus(
		    src_w, uew->toplevel, uew->dnd_action, NULL
		);
		events_handled++;
	    }
	    /* DND Leave? */
	    else if(OSWEventIsDNDLeave(
		uew->toplevel, event,
		&src_w 
	    ))
	    {
		if(uew->dnd_tar_w != None)
		{ 
		    uew->dnd_tar_w = None;
		    uew->dnd_action = None;
		    if(uew->map_state)
			UEWQueueDraw(uew, UEW_DRAW_VIEW);
		}
		events_handled++;
	    }
	    /* DND Drop? */
	    else if(OSWEventIsDNDDrop(
		uew->toplevel, event,
		&src_w, &t
	    ))
	    {
		int len;
		char *url = (char *)OSWDNDGetData(
		    uew->toplevel,
		    uew->target_string,
		    &len,
		    t
		);
		OSWDNDSendFinished(
		    src_w, uew->toplevel,
		    True,
		    uew->dnd_action
		);
		if(url != NULL)
		{
		    UEWOpenDNDURL(uew, url);
		    free(url);
		}
		uew->dnd_tar_w = None;
		uew->dnd_action = None;
		if(uew->map_state)
		    UEWQueueDraw(uew, UEW_DRAW_VIEW);
		events_handled++;
	    }
	    break;
	}


	/* Manage widgets */

	/* Objects List */
	MANAGE_WIDGET(&uew->objects_list);

	/* ImgRefs Popup List */
	MANAGE_WIDGET(&uew->imgrefs_pulist);

	/* View Value Prompts */
	MANAGE_WIDGET(&uew->obj_sect_x_prompt);
	MANAGE_WIDGET(&uew->obj_sect_y_prompt);
	MANAGE_WIDGET(&uew->obj_sect_z_prompt);
	MANAGE_WIDGET(&uew->obj_cx_prompt);
	MANAGE_WIDGET(&uew->obj_cy_prompt);
	MANAGE_WIDGET(&uew->obj_cz_prompt);
	MANAGE_WIDGET(&uew->obj_name_prompt);
	MANAGE_WIDGET(&uew->obj_heading_prompt);
	MANAGE_WIDGET(&uew->obj_size_prompt);

	/* Tool Bar */
	MANAGE_WIDGET(&uew->tb_new_btn);
	MANAGE_WIDGET(&uew->tb_open_btn);
	MANAGE_WIDGET(&uew->tb_save_btn);
	MANAGE_WIDGET(&uew->tb_save_as_btn);
	MANAGE_WIDGET(&uew->tb_save_as_image_btn);
	MANAGE_WIDGET(&uew->tb_cut_btn);
	MANAGE_WIDGET(&uew->tb_copy_btn);
	MANAGE_WIDGET(&uew->tb_paste_btn);
	MANAGE_WIDGET(&uew->tb_header_btn);
	MANAGE_WIDGET(&uew->opm_pulist);
	MANAGE_WIDGET(&uew->tb_select_opm_file_btn);
	MANAGE_WIDGET(&uew->tb_insert_object_btn);
	MANAGE_WIDGET(&uew->tb_properties_btn);
	MANAGE_WIDGET(&uew->tb_weapons_btn);
	MANAGE_WIDGET(&uew->tb_economy_btn);
	MANAGE_WIDGET(&uew->tb_remodel_btn);
	MANAGE_WIDGET(&uew->tb_delete_btn);
	MANAGE_WIDGET(&uew->tb_print_btn);

	MANAGE_WIDGET(&uew->pbar);

	/* Property Prompts */
	if(events_handled == 0)
	    events_handled += UEWPropsManage(uew, event);

#undef MANAGE_WIDGET

	return(events_handled);
}

/*
 *	Maintains the universe editor.
 *
 *	Animates the displaye imgref image.
 *
 *	Passively loads any imgref images.
 *
 *	Deletes unmapped subwindows.
 */
void UEWMaintain(
	void *data, uew_struct *uew,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	ue_option_struct *opt = &option;

	if(uew == NULL)
	    return;

	/* Imgrefs */
	UEMaintainImgRefs(
	    data,
	    uew->imgref, uew->total_imgrefs,
	    cur_ms, cur_sec, tc
	);

	/* Draw animated images */
	if((uew->ani_next < cur_ms) &&
	   opt->animate_object_display
	)
	{
	    const int prev_frame = uew->ani_frame;

	    uew->ani_frame++;
	    if(uew->ani_frame >= uew->ani_total_frames)
		uew->ani_frame = 0;

	    if(uew->ani_frame != prev_frame)
		UEWQueueDraw(uew, UEW_DRAW_PREVIEW);

	    uew->ani_next = cur_ms + uew->ani_int;
	}

	/* Maintain Property Dialogs */
	if(uew->total_prop_dlgs > 0)
	{
	    int i;
	    for(i = 0; i < uew->total_prop_dlgs; i++)
		PropDlgMaintain(
		    data, uew->prop_dlg[i],
		    cur_ms, cur_sec, tc
		);
	}

	/* Delete unused Universe Header Windows */
	if(uew->uhw != NULL)
	{
	    uhw_struct *uhw = uew->uhw;
	    if(!uhw->map_state)
	    {
		UHWDelete(uhw);
		uew->uhw = uhw = NULL;
	    }
	}

	/* Delete unused Property Dialogs */
	if(uew->total_prop_dlgs > 0)
	{
	    int i, j;
	    prop_dlg_struct *prop_dlg;
	    for(i = 0; i < uew->total_prop_dlgs; i++)
	    {
		prop_dlg = uew->prop_dlg[i];
		if((prop_dlg != NULL) ? !prop_dlg->map_state : True)
		{
		    PropDlgDelete(prop_dlg);

		    uew->total_prop_dlgs--;
		    for(j = i; j < uew->total_prop_dlgs; j++)
			uew->prop_dlg[j] = uew->prop_dlg[j + 1];

		    if(uew->total_prop_dlgs > 0)
		    {
			uew->prop_dlg = (prop_dlg_struct **)realloc(
			    uew->prop_dlg,
			    uew->total_prop_dlgs * sizeof(prop_dlg_struct *)
			);
			if(uew->prop_dlg == NULL)
			    uew->total_prop_dlgs = 0;
		    }
		    else
		    {
			free(uew->prop_dlg);
			uew->prop_dlg = NULL;
			uew->total_prop_dlgs = 0;
		    }

		    i--;
		}
	    }
	}

	/* Delete unused Weapon Windows */
	if(uew->total_wepws > 0)
	{
	    int i, j;
	    wepw_struct *wepw;
	    for(i = 0; i < uew->total_wepws; i++)
	    {
		wepw = uew->wepw[i];
		if((wepw != NULL) ? !wepw->map_state : True)
		{
		    WepWDelete(wepw);

		    uew->total_wepws--;
		    for(j = i; j < uew->total_wepws; j++)
			uew->wepw[j] = uew->wepw[j + 1];

		    if(uew->total_wepws > 0)
		    {
			uew->wepw = (wepw_struct **)realloc(
			    uew->wepw,
			    uew->total_wepws * sizeof(wepw_struct *)
			);
			if(uew->wepw == NULL)
			    uew->total_wepws = 0;
		    }
		    else
		    {
			free(uew->wepw);
			uew->wepw = NULL;
			uew->total_wepws = 0;
		    }

		    i--;
		}
	    }
	}

	/* Delete unused Economy Windows */
	if(uew->total_ecows > 0)
	{
	    int i, j;
	    ecow_struct *ecow;
	    for(i = 0; i < uew->total_ecows; i++)
	    {
		ecow = uew->ecow[i];
		if((ecow != NULL) ? !ecow->map_state : True)
		{
		    EcoWDelete(ecow);

		    uew->total_ecows--;
		    for(j = i; j < uew->total_ecows; j++)
			uew->ecow[j] = uew->ecow[j + 1];

		    if(uew->total_ecows > 0)
		    {
			uew->ecow = (ecow_struct **)realloc(
			    uew->ecow,
			    uew->total_ecows * sizeof(ecow_struct *)
			);
			if(uew->ecow == NULL)
			    uew->total_ecows = 0;
		    }
		    else
		    {
			free(uew->ecow);
			uew->ecow = NULL;
			uew->total_ecows = 0;
		    }

		    i--;
		}
	    }
	}

#if 0
/* To not delete any Print Dialogs so that we can save the last values
 * between prints
 */
	/* Delete unused Print Dialogs */
	if(uew->print_dlg != NULL)
	{
	    print_dlg_struct *d = uew->print_dlg;
	    if(!d->map_state)
	    {
		PrintDlgDelete(d);
		uew->print_dlg = d = NULL; 
	    }
	}
#endif
}

/*
 *	Updates all widgets to reflect current values.
 */
void UEWUpdateWidgets(uew_struct *uew)
{
	int i;
	menu_struct *menu;
	menu_bar_struct *mb;
	ue_option_struct *opt = &option;

	if(uew == NULL)
	    return;

	if(uew->freeze_count > 0)
	    return;

	uew->freeze_count++;

	mb = &uew->mb;

#define SET_CHECK_MENU_ITEM(_id_,_b_)	{		\
 const int i = MenuGetItemNumberByID(menu, (_id_));	\
 if(i > -1)						\
  MenuSetItemState(menu, i, (_b_));			\
}

	for(i = 0; i < mb->total_items; i++)
	{
	    menu = MenuBarGetMenuFromItem(mb, i);
	    if(menu == NULL)
		continue;

	    SET_CHECK_MENU_ITEM(
		UEW_OP_SHOW_OBJECT_LABELS, opt->show_object_labels
	    );
	    SET_CHECK_MENU_ITEM(
		UEW_OP_SHOW_GEOMETRY_LABELS, opt->show_geometry_labels
	    );
	    SET_CHECK_MENU_ITEM(
		UEW_OP_SHOW_GRID, opt->show_grid
	    );
	    SET_CHECK_MENU_ITEM(
		UEW_OP_SHOW_CONTACT_BOUNDS, opt->show_contact_bounds
	    );
	    SET_CHECK_MENU_ITEM(
		UEW_OP_ANIMATE_OBJECT_DISPLAY, opt->animate_object_display
	    );
	}

#undef SET_CHECK_MENU_ITEM

	uew->freeze_count--;
}

/*
 *	Maps the universe editor.
 */
void UEWMap(uew_struct *uew)
{
	if(uew == NULL)
	    return;

	OSWMapRaised(uew->toplevel);
	uew->map_state = True;

	WidgetMap(&uew->mb);

	OSWMapWindow(uew->toolbar);
	WidgetMap(&uew->tb_new_btn);
	WidgetMap(&uew->tb_open_btn);
	WidgetMap(&uew->tb_save_btn);
	WidgetMap(&uew->tb_save_as_btn);
	WidgetMap(&uew->tb_save_as_image_btn);
	WidgetMap(&uew->tb_cut_btn);
	WidgetMap(&uew->tb_copy_btn);
	WidgetMap(&uew->tb_paste_btn);
	WidgetMap(&uew->tb_header_btn);
	WidgetMap(&uew->opm_pulist);
	WidgetMap(&uew->tb_select_opm_file_btn);
	WidgetMap(&uew->tb_insert_object_btn);
	WidgetMap(&uew->tb_properties_btn);
	WidgetMap(&uew->tb_weapons_btn);
	WidgetMap(&uew->tb_economy_btn);
	WidgetMap(&uew->tb_remodel_btn);
	WidgetMap(&uew->tb_delete_btn);
	WidgetMap(&uew->tb_print_btn);

	WidgetMap(&uew->objects_list);
	WidgetMap(&uew->imgrefs_pulist);
	OSWMapWindow(uew->preview);

	OSWMapWindow(uew->view);
	OSWMapWindow(uew->view_split_bar);

	OSWMapWindow(uew->view_status);
	WidgetMap(&uew->obj_sect_x_prompt);
	WidgetMap(&uew->obj_sect_y_prompt);
	WidgetMap(&uew->obj_sect_z_prompt);
	WidgetMap(&uew->obj_cx_prompt);
	WidgetMap(&uew->obj_cy_prompt);
	WidgetMap(&uew->obj_cz_prompt);
	WidgetMap(&uew->obj_name_prompt);
	WidgetMap(&uew->obj_heading_prompt);
	WidgetMap(&uew->obj_size_prompt);

	OSWMapWindow(uew->props);
	OSWMapWindow(uew->props_split_bar);
	UEWPropsDraw(uew, 0);
	SBarDraw(&uew->props_sb);

	OSWMapWindow(uew->status);
	PBarMap(&uew->pbar);

	UEWDoUpdateWindowMenus();

	OSWGrabFocus(uew->toplevel);
}

/*
 *	Unmaps the Universe Editor.
 */
void UEWUnmap(uew_struct *uew)
{
	rectangle_t *rect;
	ue_option_struct *opt = &option;

	if(uew == NULL)
	    return;

	/* Record the geometry */
	rect = &opt->uew_geometry;
	OSWGetWindowRootOrigin(uew->toplevel, &uew->x, &uew->y);
	rect->x = uew->x;
	rect->y = uew->y;
	rect->width = uew->width;
	rect->height = uew->height;

	/* Is there a drag operation in progress? */
	if(uew->drag_op != UEW_DRAG_OP_NONE)
	{
	    /* Ungrab the pointer, switch back keyboard repeating,
	     * and stop the drag operation
	     */
	    if(OSWIsPointerGrabbed())
		OSWUngrabPointer();
	    OSWKBAutoRepeatOn();
	    uew->drag_op = UEW_DRAG_OP_NONE;
	}

	PBarUnmap(&uew->pbar);
	OSWUnmapWindow(uew->status);

	OSWUnmapWindow(uew->props_split_bar);
	OSWUnmapWindow(uew->props);

	WidgetUnmap(&uew->obj_cx_prompt);
	WidgetUnmap(&uew->obj_cy_prompt);
	WidgetUnmap(&uew->obj_cz_prompt);
	WidgetUnmap(&uew->obj_sect_x_prompt);
	WidgetUnmap(&uew->obj_sect_y_prompt);
	WidgetUnmap(&uew->obj_sect_z_prompt);
	WidgetUnmap(&uew->obj_name_prompt);
	WidgetUnmap(&uew->obj_heading_prompt);
	WidgetUnmap(&uew->obj_size_prompt);
	OSWUnmapWindow(uew->view_status);

	OSWUnmapWindow(uew->view_split_bar);
	OSWUnmapWindow(uew->view);

	WidgetUnmap(&uew->imgrefs_pulist);
	WidgetUnmap(&uew->objects_list);

	WidgetUnmap(&uew->tb_new_btn);
	WidgetUnmap(&uew->tb_open_btn);
	WidgetUnmap(&uew->tb_save_btn);
	WidgetUnmap(&uew->tb_save_as_btn);
	WidgetUnmap(&uew->tb_save_as_image_btn);
	WidgetUnmap(&uew->tb_cut_btn);
	WidgetUnmap(&uew->tb_copy_btn);
	WidgetUnmap(&uew->tb_paste_btn);
	WidgetUnmap(&uew->tb_header_btn);
	WidgetUnmap(&uew->opm_pulist);
	WidgetUnmap(&uew->tb_select_opm_file_btn);
	WidgetUnmap(&uew->tb_insert_object_btn);
	WidgetUnmap(&uew->tb_properties_btn);
	WidgetUnmap(&uew->tb_weapons_btn);
	WidgetUnmap(&uew->tb_economy_btn);
	WidgetUnmap(&uew->tb_remodel_btn);
	WidgetUnmap(&uew->tb_delete_btn);
	WidgetUnmap(&uew->tb_print_btn);
	OSWUnmapWindow(uew->toolbar);

	MenuBarUnmap(&uew->mb);

	OSWUnmapWindow(uew->toplevel);
	uew->map_state = False;

	UEWDoUpdateWindowMenus();
}

/*
 *	Deletes the universe editor.
 */
void UEWDelete(uew_struct *uew)
{
	int i;

	if(uew == NULL)
	    return;

	/* Universe Header Window */
	UHWDelete(uew->uhw);
	uew->uhw = NULL;

	/* Property Dialogs */
	for(i = 0; i < uew->total_prop_dlgs; i++)
	    PropDlgDelete(uew->prop_dlg[i]);
	free(uew->prop_dlg);
	uew->prop_dlg = NULL;
	uew->total_prop_dlgs = 0;

	/* Weapon Windows */
	for(i = 0; i < uew->total_wepws; i++)
	    WepWDelete(uew->wepw[i]);
	free(uew->wepw);
	uew->wepw = NULL;
	uew->total_wepws = 0;

	/* Economy Windows */
	for(i = 0; i < uew->total_ecows; i++)
	    EcoWDelete(uew->ecow[i]);
	free(uew->ecow);
	uew->ecow = NULL;
	uew->total_ecows = 0;

	/* Print Dialog */
	PrintDlgDelete(uew->print_dlg);
	uew->print_dlg = NULL;


	/* Delete all the objects */
	CListClear(&uew->objects_list);

	free(uew->status_message);
	uew->status_message = NULL;

	/* Delete the property prompts */
	UEWPropsDestroy(uew);

	/* Delete all the imgrefs */
	WidgetDestroy(&uew->imgrefs_pulist);
	ImgRefClear(&uew->imgref, &uew->total_imgrefs);

	free(uew->unv_file);
	uew->unv_file = NULL;

	/* Right-Click Menu */
	WidgetDestroy(&uew->menu);

	/* Status Bar */
	WidgetDestroy(&uew->pbar);
	OSWDestroyWindow(&uew->status);

	/* Properties */
	WidgetDestroy(&uew->props_sb);
	OSWDestroyWindow(&uew->props_split_bar);
	OSWDestroyWindow(&uew->props);

	/* View and View Status */
	OSWDestroyWindow(&uew->view_split_bar);
	WidgetDestroy(&uew->obj_size_prompt);
	WidgetDestroy(&uew->obj_heading_prompt);
	WidgetDestroy(&uew->obj_name_prompt);
	WidgetDestroy(&uew->obj_sect_z_prompt);
	WidgetDestroy(&uew->obj_sect_y_prompt);
	WidgetDestroy(&uew->obj_sect_x_prompt);
	WidgetDestroy(&uew->obj_cz_prompt);
	WidgetDestroy(&uew->obj_cy_prompt);
	WidgetDestroy(&uew->obj_cx_prompt);
	OSWDestroyWindow(&uew->view_status);

	OSWDestroyWindow(&uew->view);
	OSWDestroyPixmap(&uew->view_buf);

	/* Objects List */
	WidgetDestroy(&uew->objects_list);
	OSWSyncSharedImage(uew->imgref_img, uew->preview);
	OSWDestroySharedImage(&uew->imgref_img);
	OSWDestroyWindow(&uew->preview);
	OSWDestroyPixmap(&uew->preview_buf);

	/* Tool Bar */
	WidgetDestroy(&uew->tb_new_btn);
	WidgetDestroy(&uew->tb_open_btn);
	WidgetDestroy(&uew->tb_save_btn);
	WidgetDestroy(&uew->tb_save_as_btn);
	WidgetDestroy(&uew->tb_save_as_image_btn);
	WidgetDestroy(&uew->tb_cut_btn); 
	WidgetDestroy(&uew->tb_copy_btn);
	WidgetDestroy(&uew->tb_paste_btn);
	WidgetDestroy(&uew->tb_header_btn);
	WidgetDestroy(&uew->opm_pulist);
	WidgetDestroy(&uew->tb_select_opm_file_btn);
	WidgetDestroy(&uew->tb_insert_object_btn);
	WidgetDestroy(&uew->tb_properties_btn); 
	WidgetDestroy(&uew->tb_weapons_btn);
	WidgetDestroy(&uew->tb_economy_btn);
	WidgetDestroy(&uew->tb_remodel_btn);
	WidgetDestroy(&uew->tb_delete_btn);
	WidgetDestroy(&uew->tb_print_btn);
	OSWDestroyWindow(&uew->toolbar);

	/* Menu bar */
	WidgetDestroy(&uew->mb);

#ifdef HAVE_X
	XFreeGC(osw_gui[0].display, uew->invert_gc);
	uew->invert_gc = None;
#endif

	OSWDestroyWindow(&uew->toplevel);

	free(uew);
}
