#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/disk.h"
 
#include "../include/unvutil.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "uhw.h"
#include "propdlg.h"
#include "wepw.h"
#include "ecow.h"
#include "optwin.h"
#include "printdlg.h"
#include "aboutdlg.h"
#include "uew.h"
#include "ue.h"


void UEWMenuActivateCB(
	menu_struct *menu, int op, void *data
);

void UEWUnvFIOProgressCB(void *data, unsigned long i, unsigned long n);
void UEWImgRefsFIOProgressCB(void *data, unsigned long i, unsigned long n);
void UEWSndRefsFIOProgressCB(void *data, unsigned long i, unsigned long n);

void UEWOPMListDestroyCB(void *data);
void UEWOPMListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
);

void UEWImgRefsListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
);

void UEWObjectsCListDestroyCB(void *data);
void UEWObjectsCListActivateCB(
	clist_struct *clist, void *data
);
static void UEWSelectDisplayObject(
	uew_struct *uew,
	const int obj_num, xsw_object_struct *obj
);
void UEWObjectsCListSelectCB(
	clist_struct *clist, int row_num, void *data
);
void UEWObjectsCListUnselectCB(
	clist_struct *clist, int row_num, void *data
);

void UEWPropsSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
);

void UEWTBNewPBCB(push_button_struct *btn, void *data);
void UEWTBOpenPBCB(push_button_struct *btn, void *data);
void UEWTBInsertPBCB(push_button_struct *btn, void *data);
void UEWTBSavePBCB(push_button_struct *btn, void *data);
void UEWTBSaveAsPBCB(push_button_struct *btn, void *data);
void UEWTBSaveSelectedAsPBCB(push_button_struct *btn, void *data);
void UEWTBSaveAsImagePBCB(push_button_struct *btn, void *data);
void UEWTBCutPBCB(push_button_struct *btn, void *data);
void UEWTBCopyPBCB(push_button_struct *btn, void *data);
void UEWTBPastePBCB(push_button_struct *btn, void *data);
void UEWTBHeaderPBCB(push_button_struct *btn, void *data);
void UEWTBCheckUnvPBCB(push_button_struct *btn, void *data);
void UEWTBSelectOPMFilePBCB(push_button_struct *btn, void *data);
void UEWTBExportOPMFilePBCB(push_button_struct *btn, void *data);
void UEWTBInsertObjectPBCB(push_button_struct *btn, void *data);
void UEWTBDeletePBCB(push_button_struct *btn, void *data);
void UEWTBPropertiesPBCB(push_button_struct *btn, void *data);
void UEWTBWeaponsPBCB(push_button_struct *btn, void *data);
void UEWTBEconomyPBCB(push_button_struct *btn, void *data);
void UEWTBRemodelPBCB(push_button_struct *btn, void *data);
void UEWTBPrintPBCB(push_button_struct *btn, void *data);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Menu activate callback.
 */
void UEWMenuActivateCB(
	menu_struct *menu, int op, void *data
)
{
	ue_option_struct *opt = &option;
	uew_struct *uew = UEW(data);
	if(uew == NULL)
	    return;

	if(uew->freeze_count > 0)
	    return;

	switch(op)
	{
	  case UEW_OP_NEW:
	    UEWNewUniverse(uew);
	    break;
	  case UEW_OP_OPEN:
	    UEWOpen(uew);
	    break;
	  case UEW_OP_OPEN_NEW_UEW:
	    UEWOpenNew(uew);
	    break;
	  case UEW_OP_INSERT_UNIVERSE:
	    UEWInsert(uew);
	    break;
	  case UEW_OP_SAVE:
	    UEWSave(uew);
	    break;
	  case UEW_OP_SAVE_AS:
	    UEWSaveAs(uew);
	    break;
	  case UEW_OP_SAVE_SELECTED_AS:
	    UEWSaveSelectedAs(uew);
	    break;
          case UEW_OP_SAVE_AS_IMAGE:
	    UEWSaveAsImage(uew);
            break;
	  case UEW_OP_PRINT:
	    if(uew->print_dlg == NULL)
		uew->print_dlg = PrintDlgNew(uew);
	    PrintDlgMapValues(uew->print_dlg);
	    break;

	  case UEW_OP_CLOSE:
	  case UEW_OP_EXIT:
	    if(uew->has_changes)
	    {
	       const int response = UEGetResponse(
		    "Confirm Save Universe",
		    UE_ICON_WARNING,
"The current universe has unsaved changes,\n\
do you want to save those changes?",
		    CDlgBtnNo | CDlgBtnYes | CDlgBtnCancel,
		    uew->toplevel
		);
		if(response == CDlgYes)
		{
		    if(STRISEMPTY(uew->unv_file))
		    {
			UEDialog(
"Untitled Universe",
			    UE_ICON_INFO,
"The current universe has not been saved to a specified\n\
file name. To save it under a new file name, go to\n\
File->Save As...",
			    uew->toplevel
			);
			break;
		    }

		    /* Save */
		    if(UEWSave(uew))
			break;
		}
		else if(response == CDlgNo)
		{
		    /* No save */
		}
		else
		{
		    /* Cancel */
		    break;
		}

		uew->has_changes = False;
		UEWUpdateTitle(uew);
	    }
	    UEWUnmap(uew);
	    break;

	  case UEW_OP_UNDO:
	    break;
	  case UEW_OP_CUT:
	    UEWCut(uew);
	    break;
	  case UEW_OP_COPY:
	    UEWCopy(uew);
	    break;
	  case UEW_OP_PASTE:
	    UEWPaste(uew);
	    break;
	  case UEW_OP_DELETE:
	    UEWDeleteObject(uew);
	    break;
	  case UEW_OP_SELECT_ALL:
	    CListSelectAllRows(&uew->objects_list);
	    break;
	  case UEW_OP_UNSELECT_ALL:
	    CListUnselectAllRows(&uew->objects_list);
	    break;
	  case UEW_OP_INVERT_SELECTION:
	    CListInvertSelection(&uew->objects_list);
	    break;

	  case UEW_OP_SHOW_OBJECT_LABELS:
	    opt->show_object_labels = !opt->show_object_labels;
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
	    break;
	  case UEW_OP_SHOW_GEOMETRY_LABELS:
	    opt->show_geometry_labels = !opt->show_geometry_labels;
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
	    break;
	  case UEW_OP_SHOW_GRID:
	    opt->show_grid = !opt->show_grid;
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
	    break;
          case UEW_OP_SHOW_CONTACT_BOUNDS:
            opt->show_contact_bounds = !opt->show_contact_bounds;
            UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
            break;
	  case UEW_OP_ANIMATE_OBJECT_DISPLAY:
	    opt->animate_object_display = !opt->animate_object_display;
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE); 
	    break;

	  case UEW_OP_LEGEND:
	    if(uew->uhw == NULL)
		uew->uhw = UHWNew(uew);
	    UHWMapValues(uew->uhw);
 	    break;

	  case UEW_OP_CHECK_UNIVERSE:
	    UEWCheckUnv(uew, True, False);
	    break;

	  case UEW_OP_INSERT_OBJECT:
	    if(True)
	    {
		popup_list_struct *pulist = &uew->opm_pulist;
		const int item_num = PUListGetSelItem(pulist);
		xsw_object_struct *opm = SW_OBJECT(PUListItemGetData(
		    pulist, item_num
		));
		UEWInsertObject(uew, opm, False);
	    }
	    break;

	  case UEW_OP_OBJ_PROPERTIES:
	    if(uew->last_selected_object > -1)
	    {
		const int obj_num = uew->last_selected_object;
		prop_dlg_struct *d = PropDlgNew(uew, obj_num);
		const int i = MAX(uew->total_prop_dlgs, 0);
		uew->total_prop_dlgs = i + 1;
		uew->prop_dlg = (prop_dlg_struct **)realloc(
		    uew->prop_dlg,
		    uew->total_prop_dlgs * sizeof(prop_dlg_struct *)
		);
		uew->prop_dlg[i] = d;
		PropDlgMapValues(d);
	    }
	    else
	    {
		UEDialog(
"Edit Properties",
		    UE_ICON_WARNING,
"No object selected.",
		    uew->toplevel
		);
	    }
	    break;

	  case UEW_OP_OBJ_WEAPONS:
	    if(uew->last_selected_object > -1)
	    {
		const int obj_num = uew->last_selected_object;
		wepw_struct *wepw = WepWNew(uew, obj_num);
		const int i = MAX(uew->total_wepws, 0);
		uew->total_wepws = i + 1;
		uew->wepw = (wepw_struct **)realloc(
		    uew->wepw,
		    uew->total_wepws * sizeof(wepw_struct *)
		);
		uew->wepw[i] = wepw;
		WepWMapValues(wepw);
	    }
	    else
	    {
		UEDialog( 
"Edit Weapons",
		    UE_ICON_WARNING,
"No object selected.",
		    uew->toplevel
		);
	    }
	    break;

	  case UEW_OP_OBJ_ECONOMY:
	    if(uew->last_selected_object > -1)
	    {
		const int obj_num = uew->last_selected_object;
		ecow_struct *ecow = EcoWNew(uew, obj_num);
		const int i = MAX(uew->total_ecows, 0);
		uew->total_ecows = i + 1;
		uew->ecow = (ecow_struct **)realloc(
		    uew->ecow,
		    uew->total_ecows * sizeof(ecow_struct *)
		);
		uew->ecow[i] = ecow;
		EcoWMapValues(ecow);
	    }
	    else
	    {
		UEDialog( 
"Edit Economy",
		    UE_ICON_WARNING,
"No object selected.",
		    uew->toplevel
		);
	    }
	    break;

	  case UEW_OP_OBJ_REMODEL:
	    UEWRemodel(uew);
	    break;

	  case UEW_OP_SELECT_OPM_FILE:
	    UEWSelectOPMFile(uew);
	    break;

	  case UEW_OP_EXPORT_OPM_FILE:
	    UEWExportOPMFile(uew);
	    break;

	  case UEW_OP_OPTIONS:
	    if(optwin == NULL)
		optwin = OptWinNew();
	    OptWinMapValues(optwin, uew->toplevel);
	    break;

	  case UEW_OP_NEW_UEW:
	    if(UECreateUEW(NULL) < 0)
	    {
		UEDialog(
		    "Create Universe Editor Failed",
		    UE_ICON_ERROR,
"Unable to create a new Universe Editor.",
		    uew->toplevel
		);
	    }
	    break;

	  case UEW_OP_ABOUT:
	    if(about_dlg == NULL)
		about_dlg = AboutDlgNew();
	    AboutDlgMap(about_dlg, uew->toplevel);
	    break;
	}

	/* Switch to window */
	if((op >= UEW_OP_WINDOW_START) &&
	   (op < UEW_OP_WINDOW_END)
	)
	{
	    int	i, win_num = 0,
		menu_item_num = op - UEW_OP_WINDOW_START;
	    uew_struct *uew2;

	    /* Universe Editor Windows */
	    for(i = 0; i < total_uews; i++)
	    {
		uew2 = ue_uew[i];
		if(uew2 == NULL)
		    continue;

		if(menu_item_num == win_num)
		    UEWMap(uew2);

		win_num++;
	    }

	    /* Header Windows */
	    if(uew->uhw != NULL)
	    {
		if(menu_item_num == win_num)
		    UHWMap(uew->uhw);
		win_num++;
	    }

	    /* Property Dialogs */
	    for(i = 0; i < uew->total_prop_dlgs; i++)
	    {
		if(uew->prop_dlg[i] == NULL)
		    continue;

		if(menu_item_num == win_num)
		    PropDlgMap(uew->prop_dlg[i]);

		win_num++;
	    }

	    /* Weapons Windows */
	    for(i = 0; i < uew->total_wepws; i++)
	    {
		if(uew->wepw[i] == NULL)
		    continue;

		if(menu_item_num == win_num)
		    WepWMap(uew->wepw[i]);

		win_num++;
	    }

	    /* Economy Windows */  
	    for(i = 0; i < uew->total_ecows; i++)
	    {
		if(uew->ecow[i] == NULL)
		    continue;

		if(menu_item_num == win_num)
		    EcoWMap(uew->ecow[i]);

		win_num++;
	    }
	}
}


/*
 *	Universe file IO progress notify callback handler.
 */
void UEWUnvFIOProgressCB(void *data, unsigned long i, unsigned long n)
{
	uew_struct *uew = UEW(data);
	if(uew == NULL)
	    return;

	UEWStatusProgress(
	    uew,
	    (n > 0) ? ((float)i / (float)n) : 0.0f
	);
}

/*
 *	ImgRefs file IO progress notify callback handler.
 */
void UEWImgRefsFIOProgressCB(void *data, unsigned long i, unsigned long n)
{
	uew_struct *uew = UEW(data);
	if(uew == NULL)
	    return;

	UEWStatusProgress(
	    uew,
	    (n > 0) ? ((float)i / (float)n) : 0.0f
	);
}

/*
 *	SndRefs file IO progress notify callback handler.
 */
void UEWSndRefsFIOProgressCB(void *data, unsigned long i, unsigned long n)
{
	uew_struct *uew = UEW(data);
	if(uew == NULL)
	    return;

	UEWStatusProgress(
	    uew,
	    (n > 0) ? ((float)i / (float)n) : 0.0f
	);
}


/*
 *	OPM popup list item destroy callback.
 */
void UEWOPMListDestroyCB(void *data)
{
	xsw_object_struct *obj = SW_OBJECT(data);
	if(obj == NULL)
	    return;

	UNVDeleteObject(obj);
}

/*
 *	OPM popup list item select callback.
 */
void UEWOPMListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
)
{
	char *s;
	xsw_object_struct *opm;
	uew_struct *uew = UEW(data);
	if((pulist == NULL) || (uew == NULL))
	    return;

	if(uew->freeze_count > 0)   
	    return;

	opm = SW_OBJECT(PUListItemGetData(
	    pulist, sel_item
	));
	if(opm == NULL)
	    return;

	s = (char *)malloc(STRLEN(opm->name) + 80);
	sprintf(
	    s,
	    "Selected OPM `%s'.",
	    opm->name
	);
	UEWStatus(uew, s);
	free(s);
}

/*
 *	ImgRefs popup list item select callback.
 */
void UEWImgRefsListSelectCB(
	popup_list_struct *pulist, int sel_item, void *data
)
{
	int row_num, obj_num, imgref_num;
	clist_struct *clist;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	uew_struct *uew = UEW(data);
	if((pulist == NULL) || (uew == NULL))
	    return;

	if(uew->freeze_count > 0)
	    return;

	clist = &uew->objects_list;
	row_num = obj_num = uew->last_selected_object;
	obj = SW_OBJECT(CListRowGetData(clist, row_num));
	if(obj == NULL)
	    return;

	/* Get the selected imgref */
	imgref_num = (int)PUListItemGetData(pulist, sel_item);

	/* Set the object's current property values */
	UEWPropsDoSetValues(uew, obj_num);

	/* Set the object's imgref to the selected imgref */
	obj->imgref = imgref_num;

	/* Display the object with the new imgref */
	UEWDisplayObject(uew, obj_num, False);

	imgref = ((imgref_num >= 0) && (imgref_num < uew->total_imgrefs)) ?
	    uew->imgref[imgref_num] : NULL;
	if(imgref != NULL)
	{
	    const char *name = UNVGetObjectFormalName(obj_num, obj);
	    char *s;
	    if(!STRISEMPTY(imgref->name))
	    {
		s = (char *)malloc(
		    STRLEN(name) + STRLEN(imgref->name) + 80
		);
		sprintf(
		    s,
"Set the imgref value of object `%s' to `%s'.",
		    name, imgref->name
		);
	    }
	    else
	    {
		s = (char *)malloc(
		    STRLEN(name) + 40 + 80
		);
		sprintf(
		    s,
"Set the imgref value of object `%s' to `ImgRef #%i'.",
		    name, imgref_num
		);
	    }
	    UEWStatus(uew, s);
	    free(s);
	}

	if(!uew->has_changes)
	{
	    uew->has_changes = True;
	    UEWUpdateTitle(uew);
	}
}


/*
 *	Objects clist item destroy callback.
 */
void UEWObjectsCListDestroyCB(void *data)
{
	xsw_object_struct *obj = SW_OBJECT(data);
	if(obj == NULL)
	    return;

	UNVDeleteObject(obj);
}

/*
 *	Objects clist activate callback.
 *
 *	Updates the objects clist items to reflect
 *	current objects.
 */
void UEWObjectsCListActivateCB(
	clist_struct *clist, void *data
)
{
	int row_num, obj_num;
	uew_struct *uew = UEW(data);
	if((clist == NULL) || (uew == NULL))
	    return; 

	if(uew->freeze_count > 0)   
	    return;

	row_num = CListGetLastSelectedRow(clist);
	obj_num = row_num;

	UEWPropsDoSetValues(uew, obj_num);

	/* Redraw as needed */
	if(uew->map_state)
	{
	    UEWQueueDraw(uew, UEW_DRAW_VIEW);
	    UEWQueueDraw(uew, UEW_DRAW_PREVIEW);
	}
}

/*
 *	Gets the values for the specified object as if it were just
 *	selected.
 */
static void UEWSelectDisplayObject(
	uew_struct *uew,
	const int obj_num, xsw_object_struct *obj
)
{
	const char *name;
	char *s;

	if((uew == NULL) || (obj == NULL))
	    return;

	/* Display the object and get the view and property prompt
	 * values
	 */
	UEWDisplayObject(uew, obj_num, True);

	name = UNVGetObjectFormalName(obj_num, obj);
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Selected object `%s'.",
	    name
	);
	UEWStatus(uew, s);
	free(s);
}

/*
 *	Objects clist select callback.
 */
void UEWObjectsCListSelectCB(
	clist_struct *clist, int row_num, void *data
)
{
	int obj_num;
	xsw_object_struct *obj;
	uew_struct *uew = UEW(data);
	if((clist == NULL) || (row_num < 0) || (uew == NULL))
	    return;

	if(uew->freeze_count > 0)   
	    return;

	/* Apply values to the last selected object (if any) */
	obj_num = uew->last_selected_object;
	UEWPropsDoSetValues(uew, obj_num);

	/* Display the selected object */
	obj_num = row_num;
	obj = SW_OBJECT(CListRowGetData(clist, row_num));
	UEWSelectDisplayObject(uew, obj_num, obj);

	/* Update the last selected object */
	uew->last_selected_object = obj_num;

	if(clist->total_sel_rows > 1)
	{
	    char *s = (char *)malloc(40 + 80);
	    sprintf(
		s,  
"%i objects selected.",
		clist->total_sel_rows
	    );
	    UEWStatus(uew, s);
	    free(s);
	}
}

/*
 *	Objects clist unselect callback.
 */
void UEWObjectsCListUnselectCB(
	clist_struct *clist, int row_num, void *data
)
{
	int imgref_num, obj_num;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	uew_struct *uew = UEW(data);
	if((clist == NULL) || (row_num < 0) || (uew == NULL))
	    return;

	if(uew->freeze_count > 0)   
	    return;

	/* If this unselected object was the last selected object then
	 * apply values and clear the display
	 */
	obj_num = row_num;
	if((obj_num > -1) &&
	   (obj_num == uew->last_selected_object)
	)
	{
	    UEWPropsDoSetValues(uew, obj_num);

	    imgref_num = uew->cur_imgref;
	    imgref = ((imgref_num >= 0) && (imgref_num < uew->total_imgrefs)) ?
		uew->imgref[imgref_num] : NULL;
	    ImgRefUnload(imgref);

	    /* Reset the last selected object */
	    uew->last_selected_object = -1;

	    if(uew->map_state)
	    {
		UEWQueueDraw(uew, UEW_DRAW_VIEW);
		UEWQueueDraw(uew, UEW_DRAW_PREVIEW);
	    }  
	}

	/* Are there still objects selected? */
	if(clist->total_sel_rows >= 2)
	{
	    /* Do not apply values of the unselected object still there
	     * are still objects selected
	     */
	    char *s = (char *)malloc(40 + 80);
	    sprintf(
		s,
"%i objects selected",
		clist->total_sel_rows
	    );
	    UEWStatus(uew, s);
	    free(s);
	}
	/* Only one object still selected? */
	else if(clist->total_sel_rows == 1)
	{
	    /* Display the only object selected */
	    const int row_num = clist->sel_row[0];
	    obj_num = row_num;
	    obj = SW_OBJECT(CListRowGetData(clist, row_num));
	    UEWSelectDisplayObject(uew, obj_num, obj);

	    /* Update the last selected object */
	    uew->last_selected_object = obj_num;
	}
	else
	{
	    /* Clear the property prompts */
	    UEWPropsDoGetValues(uew, -1);

	    /* Clear the view's status prompts */
	    UEWViewStatusGetValues(uew, NULL);

	    UEWStatus(uew, "No objects selected.");
	}
}

/*
 *	Properties scroll bar value changed callback.
 */
void UEWPropsSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
)
{
	uew_struct *uew = UEW(data);
	if(uew == NULL)
	    return;

	WidgetQueueDrawAction(
	    UEWPropsDrawCB,
	    NULL, uew
	);
}


/*
 *	Tool bar button callbacks.
 */
void UEWTBNewPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_NEW, data);
}


void UEWTBOpenPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_OPEN, data);
}

void UEWTBInsertPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_INSERT_UNIVERSE, data);
}

void UEWTBSavePBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_SAVE, data);
}

void UEWTBSaveAsPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_SAVE_AS, data);
}

void UEWTBSaveSelectedAsPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_SAVE_SELECTED_AS, data);
}

void UEWTBSaveAsImagePBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_SAVE_AS_IMAGE, data);
}

void UEWTBCutPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_CUT, data);
}

void UEWTBCopyPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_COPY, data);
}

void UEWTBPastePBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_PASTE, data);
}

void UEWTBHeaderPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_LEGEND, data);
}

void UEWTBCheckUnvPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_CHECK_UNIVERSE, data);
}

void UEWTBSelectOPMFilePBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_SELECT_OPM_FILE, data);
}

void UEWTBExportOPMFilePBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_EXPORT_OPM_FILE, data);
}

void UEWTBInsertObjectPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_INSERT_OBJECT, data);
}

void UEWTBDeletePBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_DELETE, data);
}

void UEWTBPropertiesPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_OBJ_PROPERTIES, data);

}

void UEWTBWeaponsPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_OBJ_WEAPONS, data);
}

void UEWTBEconomyPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_OBJ_ECONOMY, data);
}

void UEWTBRemodelPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_OBJ_REMODEL, data);
} 

void UEWTBPrintPBCB(push_button_struct *btn, void *data)
{
	UEWMenuActivateCB(NULL, UEW_OP_PRINT, data);
}



