#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/objects.h"
#include "../include/unvutil.h"

#include "ue.h"
#include "uew.h"
#include "uewprops.h"


void UEWPropsDrawCB(void *widget, void *data);

bool_t UEWPropsHasFocus(uew_struct *uew_ptr);

void UEWPropsDoGetValues(uew_struct *uew_ptr, const int obj_num);
void UEWPropsDoSetValues(uew_struct *uew_ptr, const int obj_num);

int UEWPropsInit(uew_struct *uew_ptr);
void UEWPropsResize(uew_struct *uew_ptr);
void UEWPropsDraw(uew_struct *uew_ptr, int direction);
int UEWPropsManage(uew_struct *uew_ptr, event_t *event);
void UEWPropsDestroy(uew_struct *uew_ptr);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Draw callback.
 */
void UEWPropsDrawCB(void *widget, void *data)
{
	uew_struct *uew_ptr = (uew_struct *)data;
	if(uew_ptr == NULL)
	    return;

	UEWPropsDraw(uew_ptr, 0);
}

/*
 *	Checks if a properties prompt is in focus.
 */
bool_t UEWPropsHasFocus(uew_struct *uew_ptr)
{
	int i, n;

        if(uew_ptr == NULL)
            return(False);

        n = (int)UEW_TOTAL_PROPS;

	for(i = 0; i < n; i++)
	{
	    if(WidgetHasFocus(
		&uew_ptr->prop_prompt[i]
	    ))
		return(True);
	}

	return(False);
}


/*
 *	Gets the property values from the object and sets them to the
 *	property prompts.
 *
 *	If the object is -1 then the prompt values will be cleared.
 */
void UEWPropsDoGetValues(uew_struct *uew_ptr, const int obj_num)
{
	int i, n;
	prompt_struct *prompt;
	clist_struct *clist;
	xsw_object_struct *obj;

	if(uew_ptr == NULL)
	    return;

	clist = &uew_ptr->objects_list;
	obj = SW_OBJECT(CListRowGetData(clist, obj_num));

	n = (int)UEW_TOTAL_PROPS;

	if(obj != NULL)
	{
	    /* Get the object's properties and set them to the
	     * property prompts
	     */
	    for(i = 0; i < n; i++)
	        UEWPropsGet(&uew_ptr->prop_prompt[i], obj);
	}
	else
	{
	    /* Clear the property prompts */
	    for(i = 0; i < n; i++)
	    {
		prompt = &uew_ptr->prop_prompt[i];
		PromptSetS(prompt, "");
		PromptUnmarkBuffer(prompt, PROMPT_POS_START);
	    }
	}

}


/*
 *	Sets the values from the property prompts to the specified
 *	object.
 */
void UEWPropsDoSetValues(uew_struct *uew_ptr, const int obj_num)
{
	int i, n;
        clist_struct *clist;
	xsw_object_struct *obj;

	if(uew_ptr == NULL)
	    return;

        clist = &uew_ptr->objects_list;
        obj = SW_OBJECT(CListRowGetData(clist, obj_num));
	if(obj == NULL)
	    return;

	/* Get values from object and put them into prompts */
	n = (int)UEW_TOTAL_PROPS;
	for(i = 0; i < n; i++)
	    UEWPropsSet(&uew_ptr->prop_prompt[i], obj);

	/* Update the item name of this object on the objects list */
	UEWObjectsListItemUpdateObjName(uew_ptr, obj_num);
}


/*
 *	Initializes all the property prompts.
 */
int UEWPropsInit(uew_struct *uew_ptr)
{
	int i, n, x, y;
	win_t parent;
	unsigned int width, height;
	const char *prop_name[] = UE_OBJECT_PROPERTY_NAMES_LIST;
	prompt_struct *prompt;

	if(uew_ptr == NULL)
	    return(-2);

	parent = uew_ptr->props;
	if(parent == None)
	    return(-1);

	OSWGetWindowGeometry(
	    parent, NULL, NULL, &width, &height, NULL
	);

	x = 0;
	y = 0;
	width = (unsigned int)MAX(
	    (int)width - SCROLLBAR_YBAR_WIDTH,
	    1
	);
	height = UEW_PROP_PROMPT_HEIGHT;

	n = (int)UEW_TOTAL_PROPS;
	for(i = 0; i < n; i++)
	{
	    prompt = &uew_ptr->prop_prompt[i];
	    PromptInit(
		prompt, parent,
		x, y,
		width, height,
		PROMPT_STYLE_FLUSHED,
		prop_name[i],
		MAX(SW_OBJ_NAME_MAX, 1024), 0
	    );
	}
	return(0);
}

/*
 *	Reszes all property prompts.
 */
void UEWPropsResize(uew_struct *uew_ptr)
{
	int i, n;
	unsigned int width, height;
	win_t parent;
	prompt_struct *prompt;
	scroll_bar_struct *sb;

	if(uew_ptr == NULL)
	    return;

	parent = uew_ptr->props;
	if(parent == None)
	    return;

	OSWGetWindowGeometry(
	    parent, NULL, NULL, &width, &height, NULL
	);

	width = (unsigned int)MAX(
	    (int)width - SCROLLBAR_YBAR_WIDTH,
	    1
	);
	height = UEW_PROP_PROMPT_HEIGHT;

	n = (int)UEW_TOTAL_PROPS;
	for(i = 0; i < n; i++)
	{
	    prompt = &uew_ptr->prop_prompt[i];
	    OSWResizeWindow(prompt->toplevel, width, height);
	    WidgetQueueResize(prompt);
	}

	sb = &uew_ptr->props_sb;
	SBarSetPositionX(sb, 0.0f);
        SBarSetPositionY(sb, 0.0f);
}


/*
 *	`Redraws' the property prompts by moving them and
 *	unmapping/mapping them.
 *
 *	direction is a hint towards which direction the
 *	props list was scrolled to just before calling this function:
 *	0 = down, 1 = up.
 *
 *	If unsure, just pass 0 (for down).
 */
void UEWPropsDraw(uew_struct *uew_ptr, int direction)
{
	int i, x, y, f_vis_prompt, l_vis_prompt;
	unsigned int width, height;
	win_t parent;
	prompt_struct *prompt;
	scroll_bar_struct *sb;

	if(uew_ptr == NULL)
	    return;

	parent = uew_ptr->props;
	if(parent == None)
	    return;

	sb = &uew_ptr->props_sb;

	OSWGetWindowGeometry(
	    parent, NULL, NULL, &width, &height, NULL
	);

	/* Calculate scrolled positions */
	f_vis_prompt = MIN(
	    (int)SBarGetPositionY(sb) / UEW_PROP_PROMPT_HEIGHT,
	    UEW_TOTAL_PROPS - 1
	);
	if(f_vis_prompt < 0)
	    f_vis_prompt = 0;

	l_vis_prompt = MIN(
	    f_vis_prompt + ((int)height / UEW_PROP_PROMPT_HEIGHT) + 3,
	    UEW_TOTAL_PROPS		/* Limit to total, not index */
	);
	if(l_vis_prompt < 0)
	    l_vis_prompt = 0;

	x = 0;
	y = 0 - ((int)SBarGetPositionY(sb) % UEW_PROP_PROMPT_HEIGHT);

	/* Unmap first non-visible prompts */
	for(i = 0; i < f_vis_prompt; i++)
	    PromptUnmap(&uew_ptr->prop_prompt[i]);

	if(direction)
	{
	    /* Scrolled up */
	    y = y + ((l_vis_prompt - f_vis_prompt) *
		UEW_PROP_PROMPT_HEIGHT) - UEW_PROP_PROMPT_HEIGHT;

	    for(i = l_vis_prompt - 1; i >= f_vis_prompt; i--)
	    {
		prompt = &uew_ptr->prop_prompt[i];
		OSWMoveWindow(prompt->toplevel, x, y);
		if(!prompt->map_state)
		    PromptMap(prompt);
		y -= UEW_PROP_PROMPT_HEIGHT;
	    }
	}
	else
	{
	    /* Scrolled down */
	    for(i = f_vis_prompt; i < l_vis_prompt; i++)
	    {
		prompt = &uew_ptr->prop_prompt[i];
	        OSWMoveWindow(prompt->toplevel, x, y);
		if(!prompt->map_state)
		    PromptMap(prompt);
	        y += UEW_PROP_PROMPT_HEIGHT;
	    }
	}

        /* Map last non-visible prompts */
	for(i = l_vis_prompt; i < UEW_TOTAL_PROPS; i++)
	    PromptUnmap(&uew_ptr->prop_prompt[i]);
}

/*
 *	Manages all property prompts, plus property window scroll bar.
 */
int UEWPropsManage(uew_struct *uew_ptr, event_t *event)
{
	int i, y, events_handled = 0;

	if((uew_ptr == NULL) || (event == NULL))
	    return(events_handled);

	for(i = 0; i < (int)UEW_TOTAL_PROPS; i++)
	    events_handled += PromptManage(
		&uew_ptr->prop_prompt[i],
		event
	    );

	if(events_handled == 0)
	{
	    /* Record previous y position */
	    y = uew_ptr->props_sb.y_win_pos;

	    events_handled += SBarManage(&uew_ptr->props_sb, event);
	    if(events_handled > 0)
	    {
		if(y < uew_ptr->props_sb.y_win_pos)
		    UEWPropsDraw(uew_ptr, 0);
		else
		    UEWPropsDraw(uew_ptr, 1);
	    }
	}


	return(events_handled);
}


/*
 *	Destroys all property prompts on uew.
 */
void UEWPropsDestroy(uew_struct *uew_ptr)
{
	int i;

	if(uew_ptr == NULL)
	    return;

	for(i = 0; i < (int)UEW_TOTAL_PROPS; i++)
	    WidgetDestroy(&uew_ptr->prop_prompt[i]);
}
