#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <cmath>

#include "../include/string.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/reality.h"
#include "../include/objects.h"
#include "../include/unvfile.h"

#include "uewprops.h"
#include "config.h"


static int UEW_PROP_ALLOC_SCORES(xsw_object_struct *obj);

void UEWPropsGet(
	prompt_struct *prompt, xsw_object_struct *obj
);
void UEWPropsSet(
	prompt_struct *prompt, xsw_object_struct *obj
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)



int UEW_PROP_ALLOC_SCORES(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return(-1);

	if(obj->score == NULL)
	{
	    obj->score = (xsw_score_struct *)calloc(
		1, sizeof(xsw_score_struct)
	    );
	    if(obj->score == NULL)
		return(-3);
	}

	return(0);
}



/*
 *	Gets value from object and places it into prompt.
 */
void UEWPropsGet(
	prompt_struct *prompt, xsw_object_struct *obj
)
{
	char *parm;

	if((prompt == NULL) || (obj == NULL))
	    return;

	parm = STRDUP(PromptGetLabel(prompt));
	if(STRISEMPTY(parm))
	{
	    free(parm);
	    return;
	}

	/* Name */
	if(!strcasecmp(parm, "Name"))
	{
	    PromptSetS(prompt, obj->name);
	}
	/* Password */
	else if(!strcasecmp(parm, "Password"))
	{
	    PromptSetS(
		prompt,
		(obj->password != NULL) ?
		    obj->password : ""
	    );
	}
	/* Empire */
	else if(!strcasecmp(parm, "Empire"))
	{
	    PromptSetS(prompt, obj->empire);
	}
	/* Description */
	else if(!strcasecmp(parm, "Description"))
	{
	    PromptSetS(
		prompt,
		(obj->description != NULL) ?
		    obj->description : ""
	    );
	}     
	/* ELink */
	else if(!strcasecmp(parm, "ELink"))
	{
	    PromptSetS(
		prompt, 
		(obj->elink != NULL) ?
		    obj->elink : ""   
	    );
	}
	/* LastUpdated */
	else if(!strcasecmp(parm, "LastUpdated"))
	{
	    PromptSetL(prompt, obj->last_updated);
	}
	/* Type */
	else if(!strcasecmp(parm, "Type"))
	{
	    switch(obj->type)
	    {
	      case SW_OBJ_TYPE_GARBAGE:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_GARBAGE);
		break;
	      case SW_OBJ_TYPE_STATIC:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_STATIC);
		break;
	      case SW_OBJ_TYPE_DYNAMIC:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_DYNAMIC);
		break;
	      case SW_OBJ_TYPE_CONTROLLED:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_CONTROLLED);
		break;
	      case SW_OBJ_TYPE_PLAYER:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_PLAYER);
		break;
	      case SW_OBJ_TYPE_WEAPON:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_WEAPON);
		break;
	      case SW_OBJ_TYPE_STREAMWEAPON:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_STREAMWEAPON);
		break;
	      case SW_OBJ_TYPE_SPHEREWEAPON:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_SPHEREWEAPON);
		break;
	      case SW_OBJ_TYPE_HOME:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_HOME);
		break;
	      case SW_OBJ_TYPE_AREA:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_AREA);
		break;
	      case SW_OBJ_TYPE_ANIMATED:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_ANIMATED);
		break;
	      case SW_OBJ_TYPE_WORMHOLE:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_WORMHOLE);
		break;
	      case SW_OBJ_TYPE_ELINK:
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_ELINK);
		break;
	      default:	/* Default to static */
		PromptSetS(prompt, SW_OBJ_TYPE_NAME_STATIC);
		break;
	    }
	}
	/* LocType */
	else if(!strcasecmp(parm, "LocType"))
	{
	    switch(obj->loc_type)
	    {
	      case SW_LOC_TYPE_SPACE:
		PromptSetS(prompt, SW_LOC_TYPE_STR_SPACE);
		break;
	      case SW_LOC_TYPE_SPACE_TERRITORY:
		PromptSetS(prompt, SW_LOC_TYPE_STR_SPACE_TERRITORY);
		break;
	      case SW_LOC_TYPE_NEBULA:
		PromptSetS(prompt, SW_LOC_TYPE_STR_NEBULA);
		break;
	      case SW_LOC_TYPE_ASTEROID_FIELD:
		PromptSetS(prompt, SW_LOC_TYPE_STR_ASTEROID_FIELD);
		break;
	      case SW_LOC_TYPE_PULSAR:
		PromptSetS(prompt, SW_LOC_TYPE_STR_PULSAR);
		break;
	      default:	/* SW_LOC_TYPE_SPACE */
		PromptSetS(prompt, SW_LOC_TYPE_STR_SPACE);
		break;
	    }
	}
	/* ImgRef */
	else if(!strcasecmp(parm, "ImgRef"))
	{
	    PromptSetI(prompt, obj->imgref);
	}
	/* Owner */
	else if(!strcasecmp(parm, "Owner"))
	{
	    PromptSetI(prompt, obj->owner);
	}
	/* Size */
	else if(!strcasecmp(parm, "Size"))
	{
	    PromptSetL(prompt, obj->size);
	}
	/* LockedOn */
	else if(!strcasecmp(parm, "LockedOn"))
	{
	    PromptSetI(prompt, obj->locked_on);
	}
	/* InterceptingObject */ 
	else if(!strcasecmp(parm, "InterceptingObject"))
	{
	    PromptSetI(prompt, obj->intercepting_object);
	}
	/* ScannerRange */
	else if(!strcasecmp(parm, "ScannerRange"))
	{
	    PromptSetF2(
		prompt,
		MAX(obj->scanner_range, 0.0f),
		2
	    );
	}
	/* SectX */
	else if(!strcasecmp(parm, "SectX"))
	{
	    PromptSetL(prompt, obj->sect_x);
	}
	/* SectY */
	else if(!strcasecmp(parm, "SectY"))
	{
	    PromptSetL(prompt, obj->sect_y);
	}
	/* SectZ */
	else if(!strcasecmp(parm, "SectZ"))
	{
	    PromptSetL(prompt, obj->sect_z);
	}
	/* X */   
	else if(!strcasecmp(parm, "X"))
	{
	    PromptSetF2(prompt, obj->x, 2);
	}
	/* Y */
	else if(!strcasecmp(parm, "Y"))
	{
	    PromptSetF2(prompt, obj->y, 2);
	}
	/* Z */
	else if(!strcasecmp(parm, "Z"))
	{
	    PromptSetF2(prompt, obj->z, 2);
	}
	/* Heading */
	else if(!strcasecmp(parm, "Heading"))
	{
	    PromptSetF2(
		prompt,
		obj->heading * (180.0f / PI),
		2
	    );
	}
	/* Pitch */
	else if(!strcasecmp(parm, "Pitch"))
	{
	    PromptSetF2(
		prompt,
		obj->pitch * (180.0f / PI),
		2
	    );
	}
	/* Bank */
	else if(!strcasecmp(parm, "Bank"))
	{
	    PromptSetF2(
		prompt,
		obj->bank * (180.0f / PI),
		2
	    );
	}
	/* Velocity */
	else if(!strcasecmp(parm, "Velocity"))
	{
	    PromptSetF2(
		prompt,    
		obj->velocity *
		    (1000.0f / (float)SW_CYCLE_LAPSE_MS),
		6
	    );
	}
	/* VelocityMax */
	else if(!strcasecmp(parm, "VelocityMax"))
	{
	    PromptSetF2(
		prompt,    
		MAX(obj->velocity_max *
		    (1000.0f / (float)SW_CYCLE_LAPSE_MS), 0.0f),
		6
	    );
	}
	/* VelocityHeading */
	else if(!strcasecmp(parm, "VelocityHeading"))
	{
	    PromptSetF2(
		prompt,    
		obj->velocity_heading * (180.0f / PI),
		2
	    );
	}
	/* VelocityPitch */
	else if(!strcasecmp(parm, "VelocityPitch"))
	{
	    PromptSetF2(
		prompt,    
		obj->velocity_pitch * (180.0f / PI),
		2
	    );
	}
	/* VelocityBank */
	else if(!strcasecmp(parm, "VelocityBank"))
	{
	    PromptSetF2(
		prompt,    
		obj->velocity_bank * (180.0f / PI),
		2
	    );
	}
	/* ThrustDir */
	else if(!strcasecmp(parm, "ThrustDir"))
	{
	    PromptSetF2(
		prompt,    
		obj->thrust_dir * (180.0f / PI),
		2
	    );
	}
	/* Thrust */
	else if(!strcasecmp(parm, "Thrust"))
	{
	    PromptSetF2(
		prompt,    
		MAX(obj->thrust *
		    (1000.0f / (float)SW_CYCLE_LAPSE_MS), 0.0f),
		6
	    );
	}
	/* ThrustPower */
	else if(!strcasecmp(parm, "ThrustPower"))
	{
	    PromptSetF2(
		prompt,    
		MAX(obj->thrust_power *
		    (1000.0f / (float)SW_CYCLE_LAPSE_MS), 0.0f),
		6
	    );
	}
	/* Throttle */
	else if(!strcasecmp(parm, "Throttle"))
	{
	    PromptSetF2(
		prompt,    
		CLIP(obj->throttle, -1.0f, 1.0f),
		2
	    );
	}
	/* EngineState */
	else if(!strcasecmp(parm, "EngineState"))
	{
	    switch(obj->engine_state)
	    {
	      case SW_ENGINE_STATE_NONE:
		PromptSetS(prompt, SW_ENGINE_STATE_STR_NONE);
		break;
	      case SW_ENGINE_STATE_OFF:
		PromptSetS(prompt, SW_ENGINE_STATE_STR_OFF);
		break;
	      case SW_ENGINE_STATE_STARTING:
		PromptSetS(prompt, SW_ENGINE_STATE_STR_STARTING);
		break;
	      case SW_ENGINE_STATE_ON:
		PromptSetS(prompt, SW_ENGINE_STATE_STR_ON);
		break;
	      default:
		PromptSetS(prompt, SW_ENGINE_STATE_STR_NONE);
		break;
	    }
	}
	/* TurnRate */
	else if(!strcasecmp(parm, "TurnRate"))
	{
	    PromptSetF2(
		prompt,
		MAX(
		    obj->turnrate *
			(1000.0f / (float)SW_CYCLE_LAPSE_MS) *
			(180.0f / PI),
		    0.0f
		),
		2
	    );
	}
	/* Lighting */
	else if(!strcasecmp(parm, "Lighting"))
	{
	    char *s = strdup("");
	    if(obj->lighting & SW_OBJ_LIGHT_VECTOR)
		s = strinsstr(s, -1, SW_OBJ_LIGHT_STR_VECTOR " ");
	    if(obj->lighting & SW_OBJ_LIGHT_STROBE)
		s = strinsstr(s, -1, SW_OBJ_LIGHT_STR_STROBE " ");
	    if(obj->lighting & SW_OBJ_LIGHT_LUMINATION)
		s = strinsstr(s, -1, SW_OBJ_LIGHT_STR_LUMINATION " ");
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* Hp */
	else if(!strcasecmp(parm, "Hp"))
	{
	    PromptSetF2(
		prompt,
		MAX(obj->hp, 0.0f),
		2
	    );
	}
	/* HpMax */
	else if(!strcasecmp(parm, "HpMax"))
	{
	    PromptSetF2(
		prompt,
		MAX(obj->hp_max, 0.0f),
		2
	    );
	}
	/* Power */
	else if(!strcasecmp(parm, "Power"))
	{
	    PromptSetF2(
		prompt,
		MAX(obj->power, 0.0f),
		2
	    );
	}
	/* PowerMax */
	else if(!strcasecmp(parm, "PowerMax"))
	{
	    PromptSetF2(
		prompt,
		MAX(obj->power_max, 0.0f),
		2
	    );
	}
	/* PowerPurity */
	else if(!strcasecmp(parm, "PowerPurity"))
	{
	    PromptSetF2(
		prompt,
		CLIP(obj->power_purity, 0.0f, 1.0f),
		2
	    );
	}
	/* CoreEfficency */
	else if(!strcasecmp(parm, "CoreEfficency"))
	{
	    PromptSetF2(
		prompt,
		MAX(obj->core_efficency *
		    (1000.0f / (float)SW_CYCLE_LAPSE_MS), 0.0f),
		6
	    );
	}
	/* Antimatter */
	else if(!strcasecmp(parm, "Antimatter"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.4f", MAX(obj->antimatter, 0.0f));
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* AntimatterMax */
	else if(!strcasecmp(parm, "AntimatterMax"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.4f", MAX(obj->antimatter_max, 0.0f));
	    PromptSetS(prompt, s);                                  
	    free(s);
	}
	/* ShieldState */
	else if(!strcasecmp(parm, "ShieldState"))
	{
	    switch(obj->shield_state)
	    {
	      case SW_SHIELD_STATE_NONE:
		PromptSetS(prompt, SW_SHIELD_STATE_STR_NONE);
		break;
	      case SW_SHIELD_STATE_DOWN:
		PromptSetS(prompt, SW_SHIELD_STATE_STR_DOWN);
		break;
	      case SW_SHIELD_STATE_UP:
		PromptSetS(prompt, SW_SHIELD_STATE_STR_UP);
		break;
	      default:
		PromptSetS(prompt, SW_SHIELD_STATE_STR_NONE);
		break;
	    }
	}
	/* ShieldFrequency */
	else if(!strcasecmp(parm, "ShieldFrequency"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.2f", MAX(obj->shield_frequency, 0.0f));
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* SelectedWeapon */
	else if(!strcasecmp(parm, "SelectedWeapon"))
	{
	    PromptSetI(prompt, obj->selected_weapon);
	}
	/* TotalWeapons */
	else if(!strcasecmp(parm, "TotalWeapons"))
	{
	    PromptSetI(prompt, obj->total_weapons);
	}
	/* BirthTime */
	else if(!strcasecmp(parm, "BirthTime"))
	{
	    PromptSetL(prompt, obj->birth_time);
	}
	/* LifeSpan */
	else if(!strcasecmp(parm, "LifeSpan"))
	{
	    PromptSetL(prompt, obj->lifespan);
	}
	/* CloakState */
	else if(!strcasecmp(parm, "CloakState"))
	{
	    switch(obj->cloak_state)
	    {
	      case SW_CLOAK_STATE_NONE:
		PromptSetS(prompt, SW_CLOAK_STATE_STR_NONE);
		break;
	      case SW_CLOAK_STATE_DOWN:
		PromptSetS(prompt, SW_CLOAK_STATE_STR_DOWN);
		break;
	      case SW_CLOAK_STATE_UP:
		PromptSetS(prompt, SW_CLOAK_STATE_STR_UP);
		break;
	      default:
		PromptSetS(prompt, SW_CLOAK_STATE_STR_NONE);
		break;
	    }
	}
	/* CloakStrength */
	else if(!strcasecmp(parm, "CloakStrength"))
	{
	    PromptSetF2(
		prompt,
		MAX(obj->cloak_strength, 0.0f),
		2
	    );
	}
	/* Visibility */
	else if(!strcasecmp(parm, "Visibility"))
	{
            PromptSetF2(
                prompt,
		CLIP(obj->visibility, 0.0f, 1.0f),
		2
	    );
	}
	/* CurVisibility */
	else if(!strcasecmp(parm, "CurVisibility"))
	{
            PromptSetF2(
                prompt,
		CLIP(obj->cur_visibility, 0.0f, 1.0f),
		2
	    );
	}
	/* ShieldVisibility */
	else if(!strcasecmp(parm, "ShieldVisibility"))
	{
            PromptSetF2(
                prompt,
		MAX(obj->shield_visibility, 0.0f),
		2
	    );
	}
	/* DamageControl */
	else if(!strcasecmp(parm, "DamageControl"))
	{
	    switch(obj->damage_control)
	    {
	      case SW_DMGCTL_STATE_OFF:
		PromptSetS(prompt, SW_DMGCTL_STATE_STR_OFF);
		break;
	      case SW_DMGCTL_STATE_ON:
		PromptSetS(prompt, SW_DMGCTL_STATE_STR_ON);
		break;
	      default:
		PromptSetS(prompt, SW_DMGCTL_STATE_STR_OFF);
		break;
	    }
	}
	/* ComChannel */
	else if(!strcasecmp(parm, "ComChannel"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.2f", MAX((float)obj->com_channel / 100.0f, 0.0f));
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* AIFlags */
	else if(!strcasecmp(parm, "AIFlags"))
	{
	    char *s = strdup("");
	    if(obj->ai_flags & SW_AI_FOLLOW_FRIEND)
		s = strinsstr(s, -1, SW_AI_STR_FOLLOW_FRIEND " ");
	    if(obj->ai_flags & SW_AI_FOLLOW_UNKNOWN)
		s = strinsstr(s, -1, SW_AI_STR_FOLLOW_UNKNOWN " ");
	    if(obj->ai_flags & SW_AI_FOLLOW_HOSTILE)
		s = strinsstr(s, -1, SW_AI_STR_FOLLOW_HOSTILE " ");

	    if(obj->ai_flags & SW_AI_FIRE_FRIEND)
		s = strinsstr(s, -1, SW_AI_STR_FIRE_FRIEND " ");
	    if(obj->ai_flags & SW_AI_FIRE_UNKNOWN)
		s = strinsstr(s, -1, SW_AI_STR_FIRE_UNKNOWN " ");
	    if(obj->ai_flags & SW_AI_FIRE_HOSTILE)
		s = strinsstr(s, -1, SW_AI_STR_FIRE_HOSTILE " ");

	    PromptSetS(prompt, s);
	    free(s);
	}
	/* PermissionUID */
	else if(!strcasecmp(parm, "PermissionUID"))
	{
	    PromptSetI(prompt, obj->permission.uid);
	}
	/* PermissionGID */
	else if(!strcasecmp(parm, "PermissionGID"))
	{
	    PromptSetI(prompt, obj->permission.gid);
	}
	/* AnimationInterval */
	else if(!strcasecmp(parm, "AnimationInterval"))
	{
	    PromptSetL(prompt, MAX(obj->animation.interval, 0l));
	}
	/* AnimationLastInterval */
	else if(!strcasecmp(parm, "AnimationLastInterval"))
	{
	    PromptSetL(prompt, MAX(obj->animation.last_interval, 0l));
	}
	/* AnimationCurrentFrame */
	else if(!strcasecmp(parm, "AnimationCurrentFrame"))
	{
	    PromptSetI(prompt, MAX(obj->animation.current_frame, 0));
	}
	/* AnimationTotalFrames */
	else if(!strcasecmp(parm, "AnimationTotalFrames"))
	{
	    PromptSetI(prompt, MAX(obj->animation.total_frames, 0));
	}
	/* AnimationCycleCount */
	else if(!strcasecmp(parm, "AnimationCycleCount"))
	{
	    PromptSetI(prompt, MAX(obj->animation.cycle_count, 0));
	}
	/* AnimationCycleTimes */
	else if(!strcasecmp(parm, "AnimationCycleTimes"))
	{
	    PromptSetI(prompt, MAX(obj->animation.cycle_times, 0));
	}
	/* ScoreCredits */
	else if(!strcasecmp(parm, "ScoreCredits"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.2f",
		(obj->score != NULL) ?
		MAX(obj->score->credits, 0.0f) : 0.0f
	    );
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* ScoreRMU */
	else if(!strcasecmp(parm, "ScoreRMU"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.2f",
		(obj->score != NULL) ? 
		MAX(obj->score->rmu, 0.0f) : 0.0f
	    );
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* ScoreRMUMax */
	else if(!strcasecmp(parm, "ScoreRMUMax"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.2f",
		(obj->score != NULL) ?
		MAX(obj->score->rmu_max, 0.0f) : 0.0f
	    );
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* ScoreDamageGiven */
	else if(!strcasecmp(parm, "ScoreDamageGiven"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.2f",
		(obj->score != NULL) ? 
		MAX(obj->score->damage_given, 0.0f) : 0.0f
	    );   
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* ScoreDamageRecieved */
	else if(!strcasecmp(parm, "ScoreDamageRecieved"))
	{
	    char *s = (char *)malloc(40);
	    sprintf(s, "%.2f",
		(obj->score != NULL) ? 
		MAX(obj->score->damage_received, 0.0f) : 0.0f
	    );
	    PromptSetS(prompt, s);
	    free(s);
	}
	/* ScoreKills */
	else if(!strcasecmp(parm, "ScoreKills"))
	{
	    PromptSetI(
		prompt,
		(obj->score != NULL) ?
		    MAX(obj->score->kills, 0) : 0
	    );
	}

	free(parm);
}


/*
 *	Sets value in prompt to object.
 */
void UEWPropsSet(prompt_struct *prompt, xsw_object_struct *obj)
{
	char *parm, *val;

	if((prompt == NULL) || (obj == NULL))
	    return;

	parm = STRDUP(PromptGetLabel(prompt));
	val = STRDUP(PromptGetS(prompt));
	if(STRISEMPTY(parm) || STRISEMPTY(val))
	{
	    free(parm);
	    free(val);
	    return;
	}

	/* Name */
	if(!strcasecmp(parm, "Name"))
	{
	    strncpy(
		obj->name, val,
		sizeof(obj->name)
	    );
	    obj->name[
		sizeof(obj->name) - 1] = '\0';
	}
	/* Password */
	else if(!strcasecmp(parm, "Password"))
	{
	    free(obj->password);
	    obj->password = STRISEMPTY(val) ? NULL : strdup(val);
	}
	/* Empire */
	else if(!strcasecmp(parm, "Empire"))
	{
	    strncpy(
		obj->empire, val,
		sizeof(obj->empire)
	    );
	    obj->empire[
		sizeof(obj->empire) - 1] = '\0';
	}
	/* Description */
	else if(!strcasecmp(parm, "Description"))
	{
	    free(obj->description);
	    obj->description = STRISEMPTY(val) ? NULL : strdup(val);
	}
	/* ELink */
	else if(!strcasecmp(parm, "ELink"))
	{
	    free(obj->elink);
	    obj->elink = STRISEMPTY(val) ? NULL : strdup(val);
	}
	/* LastUpdated */
	else if(!strcasecmp(parm, "LastUpdated"))
	{
	    obj->last_updated = (time_t)MAX(atol(val), 0l);
	}
	/* Type */
	else if(!strcasecmp(parm, "Type"))
	{
	    if(!strcasecmp(val, SW_OBJ_TYPE_NAME_STATIC))
		obj->type = SW_OBJ_TYPE_STATIC;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_DYNAMIC))
		obj->type = SW_OBJ_TYPE_DYNAMIC;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_CONTROLLED)) 
		obj->type = SW_OBJ_TYPE_CONTROLLED;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_PLAYER))
		obj->type = SW_OBJ_TYPE_PLAYER;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_WEAPON))
		obj->type = SW_OBJ_TYPE_WEAPON;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_STREAMWEAPON))
		obj->type = SW_OBJ_TYPE_STREAMWEAPON;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_SPHEREWEAPON))
		obj->type = SW_OBJ_TYPE_SPHEREWEAPON;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_HOME))
		obj->type = SW_OBJ_TYPE_HOME;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_AREA))
		obj->type = SW_OBJ_TYPE_AREA;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_ANIMATED))
		obj->type = SW_OBJ_TYPE_ANIMATED;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_WORMHOLE))
		obj->type = SW_OBJ_TYPE_WORMHOLE;
	    else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_ELINK))
		obj->type = SW_OBJ_TYPE_ELINK;
	    else
		obj->type = SW_OBJ_TYPE_STATIC;
	}
	/* LocType */
	else if(!strcasecmp(parm, "LocType"))
	{
	    if(!strcasecmp(val, SW_LOC_TYPE_STR_SPACE))
		obj->loc_type = SW_LOC_TYPE_SPACE;
	    else if(!strcasecmp(val, SW_LOC_TYPE_STR_SPACE_TERRITORY))
		obj->loc_type = SW_LOC_TYPE_SPACE_TERRITORY;
	    else if(!strcasecmp(val, SW_LOC_TYPE_STR_NEBULA))
		obj->loc_type = SW_LOC_TYPE_NEBULA;
	    else if(!strcasecmp(val, SW_LOC_TYPE_STR_ASTEROID_FIELD))
		obj->loc_type = SW_LOC_TYPE_ASTEROID_FIELD;
	    else if(!strcasecmp(val, SW_LOC_TYPE_STR_PULSAR))
		obj->loc_type = SW_LOC_TYPE_PULSAR;
	    else
		obj->loc_type = SW_LOC_TYPE_SPACE;
	}
	/* ImgRef */
	else if(!strcasecmp(parm, "ImgRef"))
	{
	    obj->imgref = atoi(val);
	}
	/* Owner */
	else if(!strcasecmp(parm, "Owner"))
	{
	    obj->owner = atoi(val);
	    if(obj->owner < 0)
		obj->owner = 0;
	}
	/* Size */
	else if(!strcasecmp(parm, "Size"))
	{
	    obj->size = MAX(atol(val), 0l);
	}
	/* LockedOn */
	else if(!strcasecmp(parm, "LockedOn"))
	{
	    obj->locked_on = atoi(val);
	}
	/* InterceptingObject */
	else if(!strcasecmp(parm, "InterceptingObject"))
	{
	    obj->intercepting_object = atoi(val);
	}
	/* ScannerRange */
	else if(!strcasecmp(parm, "ScannerRange"))
	{
	    obj->scanner_range = MAX(atof(val), 0.0f);
	}
	/* SectX */
	else if(!strcasecmp(parm, "SectX"))
	{
	    obj->sect_x = atol(val);
	}
	/* SectY */
	else if(!strcasecmp(parm, "SectY"))
	{
	    obj->sect_y = atol(val);
	}
	/* SectZ */
	else if(!strcasecmp(parm, "SectZ"))
	{
	    obj->sect_z = atol(val);
	}
	/* X */
	else if(!strcasecmp(parm, "X"))
	{
	    obj->x = atof(val);
	}
	/* Y */
	else if(!strcasecmp(parm, "Y"))
	{
	    obj->y = atof(val);
	}
	/* Z */
	else if(!strcasecmp(parm, "Z"))
	{
	    obj->z = atof(val);
	}
	/* Heading */
	else if(!strcasecmp(parm, "Heading"))
	{
	    obj->heading = atof(val) * (PI / 180);
	}
	/* Pitch */
	else if(!strcasecmp(parm, "Pitch"))
	{
	    obj->pitch = atof(val) * (PI / 180);
	}
	/* Bank */
	else if(!strcasecmp(parm, "Bank"))
	{
	    obj->bank = atof(val) * (PI / 180);
	}
	/* Velocity */
	else if(!strcasecmp(parm, "Velocity"))
	{
	    obj->velocity = atof(val) *
		((float)SW_CYCLE_LAPSE_MS / 1000);
	}
	/* VelocityMax */
	else if(!strcasecmp(parm, "VelocityMax"))
	{
	    obj->velocity_max = atof(val) *
		((float)SW_CYCLE_LAPSE_MS / 1000);
	}
	/* VelocityHeading */
	else if(!strcasecmp(parm, "VelocityHeading"))
	{   
	    obj->velocity_heading = atof(val) * (PI / 180);
	}
	/* VelocityPitch */
	else if(!strcasecmp(parm, "VelocityPitch"))
	{
/*
	    obj->velocity_pitch = atof(val) * (PI / 180);
 */
	}
	/* VelocityBank */
	else if(!strcasecmp(parm, "VelocityBank"))
	{
/*
	    obj->velocity_bank = atof(val) * (PI / 180);
 */
	}
	/* ThrustDir */
	else if(!strcasecmp(parm, "ThrustDir"))
	{
	    obj->thrust_dir = atof(val) * (PI / 180.0f);  
	}
	/* Thrust */
	else if(!strcasecmp(parm, "Thrust"))
	{   
	    obj->thrust = MAX(
		(atof(val) * ((float)SW_CYCLE_LAPSE_MS / 1000.0f)),
		0.0f
	    );
	}
	/* ThrustPower */
	else if(!strcasecmp(parm, "ThrustPower"))
	{
	    obj->thrust_power = MAX(
		(atof(val) * ((float)SW_CYCLE_LAPSE_MS / 1000.0f)),
		0.0f
	    );
	}
	/* Throttle */
	else if(!strcasecmp(parm, "Throttle"))
	{
	    obj->throttle = CLIP(atof(val), -1.0f, 1.0f);
	}
	/* EngineState */
	else if(!strcasecmp(parm, "EngineState"))
	{
	    if(!strcasecmp(val, "None"))
		obj->engine_state = SW_ENGINE_STATE_NONE;
	    else if(!strcasecmp(val, "Off"))
		obj->engine_state = SW_ENGINE_STATE_OFF;
	    else if(!strcasecmp(val, "Starting"))
		obj->engine_state = SW_ENGINE_STATE_STARTING; 
	    else if(!strcasecmp(val, "On"))
		obj->engine_state = SW_ENGINE_STATE_ON;
	    else
		obj->engine_state = SW_ENGINE_STATE_ON;
	}
	/* TurnRate */
	else if(!strcasecmp(parm, "TurnRate"))
	{
	    /* Convert from degrees per second, to radians per cycle */
	    obj->turnrate = MAX(
		(atof(val) *
		 ((float)SW_CYCLE_LAPSE_MS / 1000.0f) *
		    (PI / 180.0f)
		),
		0.0f
	    );
	}
	/* Lighting */
	else if(!strcasecmp(parm, "Lighting"))
	{
	    obj->lighting = 0;
	    if(strcasestr(val, SW_OBJ_LIGHT_STR_VECTOR) != NULL)
		obj->lighting |= SW_OBJ_LIGHT_VECTOR;
	    if(strcasestr(val, SW_OBJ_LIGHT_STR_STROBE) != NULL)
		obj->lighting |= SW_OBJ_LIGHT_STROBE;
	    if(strcasestr(val, SW_OBJ_LIGHT_STR_LUMINATION) != NULL)
		obj->lighting |= SW_OBJ_LIGHT_LUMINATION;
	}
	/* Hp */
	else if(!strcasecmp(parm, "Hp"))
	{
	    obj->hp = atof(val);
	}
	/* HpMax */
	else if(!strcasecmp(parm, "HpMax"))
	{   
	    obj->hp_max = atof(val);
	}
	/* Power */
	else if(!strcasecmp(parm, "Power"))
	{
	    obj->power = atof(val);
	}
	/* PowerMax */
	else if(!strcasecmp(parm, "PowerMax"))
	{
	    obj->power_max = atof(val);
	}
	/* PowerPurity */
	else if(!strcasecmp(parm, "PowerPurity"))
	{
	    obj->power_purity = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* CoreEfficency */
	else if(!strcasecmp(parm, "CoreEfficency"))
	{
	    obj->core_efficency = MAX(
		(atof(val) * ((float)SW_CYCLE_LAPSE_MS / 1000.0f)),
		0.0f
	    );
	}
	/* Antimatter */
	else if(!strcasecmp(parm, "Antimatter"))
	{
	    obj->antimatter = atof(val);
	}
	/* AntimatterMax */
	else if(!strcasecmp(parm, "AntimatterMax"))
	{
	    obj->antimatter_max = atof(val);
	}
	/* ShieldState */
	else if(!strcasecmp(parm, "ShieldState"))
	{   
	    if(!strcasecmp(val, "None"))
		obj->shield_state = SW_SHIELD_STATE_NONE;
	    else if(!strcasecmp(val, "Down"))
		obj->shield_state = SW_SHIELD_STATE_DOWN;
	    else if(!strcasecmp(val, "Up"))
		obj->shield_state = SW_SHIELD_STATE_UP;
	    else
		obj->shield_state = SW_SHIELD_STATE_DOWN;
	}
	/* ShieldFrequency */
	else if(!strcasecmp(parm, "ShieldFrequency"))
	{   
	    obj->shield_frequency = atof(val);
	    if(obj->shield_frequency > SW_FREQ_MAX)
		obj->shield_frequency = SW_FREQ_MAX;
	    if(obj->shield_frequency < SW_FREQ_MIN)
		obj->shield_frequency = SW_FREQ_MIN;
	}
	/* SelectedWeapon */
	else if(!strcasecmp(parm, "SelectedWeapon"))
	{
	    obj->selected_weapon = atoi(val);
	}
	/* TotalWeapons */
	else if(!strcasecmp(parm, "TotalWeapons"))
	{
#if 0
/* This parameter should only be changed when weapons are allocated or
 * deallocated
 */

	    obj->total_weapons = atoi(val);
#endif
	}
	/* BirthTime */
	else if(!strcasecmp(parm, "BirthTime"))
	{
	    obj->birth_time = atol(val);
	}
	/* LifeSpan */
	else if(!strcasecmp(parm, "LifeSpan"))
	{   
	    obj->lifespan = atol(val);
	}
	/* CloakState */
	else if(!strcasecmp(parm, "CloakState"))
	{
	    if(!strcasecmp(val, "None"))
		obj->cloak_state = SW_CLOAK_STATE_NONE;
	    else if(!strcasecmp(val, "Down"))
		obj->cloak_state = SW_CLOAK_STATE_DOWN;
	    else if(!strcasecmp(val, "Up")) 
		obj->cloak_state = SW_CLOAK_STATE_UP;
	    else
		obj->cloak_state = SW_CLOAK_STATE_DOWN;
	}
	/* CloakStrength */
	else if(!strcasecmp(parm, "CloakStrength"))
	{
	    obj->cloak_strength = MAX(atof(val), 0.0f);
	}
	/* Visibility */
	else if(!strcasecmp(parm, "Visibility"))
	{
	    obj->visibility = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* CurVisibility */
	else if(!strcasecmp(parm, "CurVisibility"))
	{
	    obj->cur_visibility = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* ShieldVisibility */
	else if(!strcasecmp(parm, "ShieldVisibility"))
	{
	    obj->shield_visibility = CLIP(atof(val), 0.0f, 1.0f);
	}
	/* DamageControl */
	else if(!strcasecmp(parm, "DamageControl"))
	{
	    if(!strcasecmp(val, SW_DMGCTL_STATE_STR_OFF))
		obj->damage_control = SW_DMGCTL_STATE_OFF;
	    else if(!strcasecmp(val, SW_DMGCTL_STATE_STR_ON))
		obj->damage_control = SW_DMGCTL_STATE_ON;
	    else
		obj->damage_control = SW_DMGCTL_STATE_OFF;
	}
	/* ComChannel */
	else if(!strcasecmp(parm, "ComChannel"))
	{
	    obj->com_channel = MAX((int)(atof(val) * 100.0f), 0);
	}
	/* AIFlags */
	else if(!strcasecmp(parm, "AIFlags"))
	{
	    obj->ai_flags = 0;

	    if(strcasestr(val, SW_AI_STR_FOLLOW_FRIEND))
		obj->ai_flags |= SW_AI_FOLLOW_FRIEND;
	    if(strcasestr(val, SW_AI_STR_FOLLOW_UNKNOWN))
		obj->ai_flags |= SW_AI_FOLLOW_UNKNOWN;
	    if(strcasestr(val, SW_AI_STR_FOLLOW_HOSTILE))
		obj->ai_flags |= SW_AI_FOLLOW_HOSTILE;

	    if(strcasestr(val, SW_AI_STR_FIRE_FRIEND))
		obj->ai_flags |= SW_AI_FIRE_FRIEND;
	    if(strcasestr(val, SW_AI_STR_FIRE_UNKNOWN))
		obj->ai_flags |= SW_AI_FIRE_UNKNOWN;
	    if(strcasestr(val, SW_AI_STR_FIRE_HOSTILE))
		obj->ai_flags |= SW_AI_FIRE_HOSTILE;
	}
	/* PermissionUID */
	else if(!strcasecmp(parm, "PermissionUID"))
	{
	    obj->permission.uid = atoi(val);
	}
	/* PermissionGID */
	else if(!strcasecmp(parm, "PermissionGID"))
	{
	    obj->permission.gid = atoi(val);
	}
	/* AnimationInterval */
	else if(!strcasecmp(parm, "AnimationInterval"))
	{
	    obj->animation.interval = atol(val);
	}
	/* AnimationLastInterval */
	else if(!strcasecmp(parm, "AnimationLastInterval"))
	{
	    obj->animation.last_interval = atol(val);
	}
	/* AnimationCurrentFrame */
	else if(!strcasecmp(parm, "AnimationCurrentFrame"))
	{
	    obj->animation.current_frame = atoi(val);
	}
	/* AnimationTotalFrames */
	else if(!strcasecmp(parm, "AnimationTotalFrames"))
	{
	    obj->animation.total_frames = atoi(val);
	}
	/* AnimationCycleCount */
	else if(!strcasecmp(parm, "AnimationCycleCount"))
	{
	    obj->animation.cycle_count = atoi(val);
	}
	/* AnimationCycleTimes */
	else if(!strcasecmp(parm, "AnimationCycleTimes"))
	{
	    obj->animation.cycle_times = atoi(val);
	}
	/* ScoreCredits */
	else if(!strcasecmp(parm, "ScoreCredits"))
	{
	    if(*val != '\0')
	    {
		if(!UEW_PROP_ALLOC_SCORES(obj))
		{
		    obj->score->credits = atof(val);
		    if(obj->score->credits > SW_CREDITS_MAX)
			obj->score->credits = SW_CREDITS_MAX;
		    if(obj->score->credits < 0)
			obj->score->credits = 0;
		}
	    }
	}
	/* ScoreRMU */
	else if(!strcasecmp(parm, "ScoreRMU"))
	{
	    if(*val != '\0')
	    {
		if(!UEW_PROP_ALLOC_SCORES(obj))
		    obj->score->rmu = atof(val);
	    }
	}
	/* ScoreRMUMax */
	else if(!strcasecmp(parm, "ScoreRMUMax"))
	{
	    if(*val != '\0')
	    {
		if(!UEW_PROP_ALLOC_SCORES(obj))
		    obj->score->rmu_max = atof(val);
	    }
	}
	/* ScoreDamageGiven */
	else if(!strcasecmp(parm, "ScoreDamageGiven"))
	{
	    if(*val != '\0')
	    {
		if(!UEW_PROP_ALLOC_SCORES(obj))
		    obj->score->damage_given = atof(val);
	    }
	}
	/* ScoreDamageRecieved */
	else if(!strcasecmp(parm, "ScoreDamageRecieved"))
	{
	    if(*val != '\0')
	    {
		if(!UEW_PROP_ALLOC_SCORES(obj))
		    obj->score->damage_received = atof(val);
	    }
	}
	/* ScoreKills */
	else if(!strcasecmp(parm, "ScoreKills"))
	{
	    if(*val != '\0')
	    {
		if(!UEW_PROP_ALLOC_SCORES(obj))
		    obj->score->kills = atoi(val);
	    }
	}

	free(parm);
	free(val);
}
