#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/reality.h"
#include "../include/objects.h"

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "uhw.h"  
#include "uew.h"
#include "ue.h"

#include "../include/oswkeyvals.h"


static void UHWCreateBrowseButton(
	uhw_struct *uhw,
	prompt_struct *prompt,		/* Reference prompt */
	win_t parent,
	push_button_struct *btn,
	void (*activate_cb)(push_button_struct *, void *)
);

static void UHWGetValues(uhw_struct *uhw);
static void UHWSetValues(uhw_struct *uhw);

static void UHWDrawCB(void *widget, void *data);

static void UHWSelectOCSNsFileCB(push_button_struct *btn, void *data);
static void UHWSelectImgRefsFileCB(push_button_struct *btn, void *data);
static void UHWSelectSndRefsFileCB(push_button_struct *btn, void *data);

static void UHWOKCB(push_button_struct *btn, void *data);
static void UHWApplyCB(push_button_struct *btn, void *data);
static void UHWCancelCB(push_button_struct *btn, void *data);

uhw_struct *UHWNew(uew_struct *uew);
void UHWDraw(uhw_struct *uhw, const int amount);
void UHWQueueDraw(uhw_struct *uhw, const int amount);
int UHWManage(uhw_struct *uhw, event_t *event);
void UHWMap(uhw_struct *uhw);
void UHWMapValues(uhw_struct *uhw);
void UHWUnmap(uhw_struct *uhw);
void UHWDelete(uhw_struct *uhw);


#define UHW_DEF_TITLE				"Legend"
#define UHW_DEF_ICON_TITLE			"Legend"

#define UHW_WIDTH				640
#define UHW_HEIGHT				480

#define UHW_MARGIN				10

#define UHW_PROMPT_HEIGHT			30

#define UHW_ICON_BTN_WIDTH			26
#define UHW_ICON_BTN_HEIGHT			26

#define UHW_BTN_WIDTH				70
#define UHW_BTN_HEIGHT				28


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a browse button.
 */
static void UHWCreateBrowseButton(
	uhw_struct *uhw,
	prompt_struct *prompt,		/* Reference prompt */
	win_t parent,
	push_button_struct *btn,
	void (*activate_cb)(push_button_struct *, void *)
)
{
	int x, y;
	const unsigned int	width = UHW_ICON_BTN_WIDTH,
				height = UHW_ICON_BTN_HEIGHT;
	widget_global_struct *wglobal = &widget_global;

	if(prompt != NULL)
	{
	    x = prompt->x + (int)prompt->width + 2;
	    y = prompt->y +
		(((int)prompt->height - (int)height) / 2);
	}
	else
	{
	    x = 0;
	    y = 0;
	}

	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    NULL,
	    PBTN_LABEL_ALIGN_CENTER,
	    wglobal->browse_files_pixmap,
	    wglobal->browse_files_mask,
	    activate_cb, uhw
	);
}


/*
 *	Gets the values from the referenced Universe Editor Window.
 */
static void UHWGetValues(uhw_struct *uhw)
{
	char *s;
	sw_legend_struct *legend;
	uew_struct *uew;

	if(uhw == NULL)
	    return;

	uew = uhw->uew;
	if(uew == NULL)
	    return;

	/* Begin getting values */

	legend = &uew->legend;

	PromptSetS(&uhw->title_prompt, legend->title);
	PromptSetF2(&uhw->ru_to_au_prompt, legend->ru_to_au, 4);

	s = (char *)malloc(40);
	sprintf(s, "#%i", legend->lost_found_owner);
	PromptSetS(&uhw->lost_found_owner_prompt, s);
	free(s);

	PromptSetS(&uhw->ocsn_file_prompt, legend->ocsn);
	PromptSetS(&uhw->imgrefs_file_prompt, legend->imgrefs);
	PromptSetS(&uhw->sndrefs_file_prompt, legend->sndrefs);

	/* Player start position */
	s = (char *)malloc(256);
	UNVLocationFormatString(
	    s,
	    &legend->player_start_sect_x,
	    &legend->player_start_sect_y,
	    &legend->player_start_sect_z,
	    &legend->player_start_x,
	    &legend->player_start_y,
	    &legend->player_start_z,
	    256
	);
	PromptSetS(&uhw->player_start_pos_prompt, s);
	free(s);

	/* Player start direction */
	s = (char *)malloc(256);
	UNVDirectionFormatString(
	    s,
	    &legend->player_start_heading,
	    &legend->player_start_pitch,
	    &legend->player_start_bank,
	    256
	);
	PromptSetS(&uhw->player_start_dir_prompt, s);
	free(s);

	/* Guest start position */
	s = (char *)malloc(256);
	UNVLocationFormatString(
	    s,
	    &legend->guest_start_sect_x,
	    &legend->guest_start_sect_y,
	    &legend->guest_start_sect_z,
	    &legend->guest_start_x,
	    &legend->guest_start_y,
	    &legend->guest_start_z,
	    256
	);
	PromptSetS(&uhw->guest_start_pos_prompt, s);
	free(s);

	/* Guest start direction */
	s = (char *)malloc(256);
	UNVDirectionFormatString(
	    s,
	    &legend->guest_start_heading,
	    &legend->guest_start_pitch,
	    &legend->guest_start_bank,
	    256
	);
	PromptSetS(&uhw->guest_start_dir_prompt, s);
	free(s);

	PromptSetF2(
	   &uhw->antimatter_low_warning_prompt,
	   legend->antimatter_low_warning, 2
	);
	PromptSetL(
	   &uhw->asteroid_field_damage_int_prompt,
	   legend->asteroid_field_damage_int
	);
	PromptSetF2(
	   &uhw->asteroid_field_damage_rate_prompt,
	   legend->asteroid_field_damage_rate, 4
	);
	PromptSetF2(
	   &uhw->auto_destruct_damage_range_prompt,
	   legend->auto_destruct_damage_range, 2
	);
	PromptSetF2(
	   &uhw->cloak_antimatter_rate_prompt,
	   legend->cloak_antimatter_rate, 4
	);
	PromptSetF2(
	   &uhw->com_range_prompt,
	   legend->com_range, 2
	);
	PromptSetF2(
	   &uhw->damage_control_rate_prompt,
	   legend->damage_control_rate, 4
	);
	PromptSetL(
	   &uhw->disarmed_weapon_lifespan_prompt,
	   legend->disarmed_weapon_lifespan
	);
	PromptSetF2(
	   &uhw->engine_antimatter_rate_prompt,
	   legend->engine_antimatter_rate, 4
	);
	PromptSetF2(
	   &uhw->no_antimatter_thrust_coeff_prompt,
	   legend->no_antimatter_thrust_coeff, 2
	);
	PromptSetF2(
	   &uhw->on_fire_hp_coeff_prompt,
	   legend->on_fire_hp_coeff, 2
	);
	PromptSetF2(
	   &uhw->tractor_beam_antimatter_rate_prompt,
	   legend->tractor_beam_antimatter_rate, 4
	);
	PromptSetF2(
	   &uhw->tractor_beam_range_prompt,
	   legend->tractor_beam_range, 2
	);
	PromptSetF2(
	   &uhw->transporter_range_prompt,
	   legend->transporter_range, 2
	);
}

/*
 *	Set values.
 */
static void UHWSetValues(uhw_struct *uhw)
{
	int total_objects;
	char *error_msg;
	const char *s;
	char *path;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct **objects_list;
	uew_struct *uew;

	if(uhw == NULL)
	    return;

	uew = uhw->uew;
	if(uew == NULL)
	    return;

	clist = &uew->objects_list;
	legend = &uew->legend;

	total_objects = clist->total_rows;
	if(total_objects > 0)
	{
	    int i;
	    objects_list = (xsw_object_struct **)malloc(
		total_objects * sizeof(xsw_object_struct *)
	    );
	    for(i = 0; i < total_objects; i++)
		objects_list[i] = SW_OBJECT(CListRowGetData(clist, i));
	}
	else
	{
	    objects_list = NULL;
	}

	/* Begin setting values */

	/* Title */
	s = PromptGetS(&uhw->title_prompt);
	if(s != NULL)
	{
	    strncpy(legend->title, s, sizeof(legend->title));
	    legend->title[sizeof(legend->title) - 1] = '\0';
	}

	/* RU to AU */
	legend->ru_to_au = PromptGetF(
	    &uhw->ru_to_au_prompt
	);

	/* Lost and found owner */
	legend->lost_found_owner = UNVMatchObjectByName(
	    objects_list, total_objects,
	    PromptGetS(&uhw->lost_found_owner_prompt), -1
	);

	/* OCSN File */
	s = PromptGetS(&uhw->ocsn_file_prompt);
	if(s != NULL)
	{
	    strncpy(legend->ocsn, s, sizeof(legend->ocsn));
	    legend->ocsn[sizeof(legend->ocsn) - 1] = '\0';
	}

	/* ImgRefs File */
	s = PromptGetS(&uhw->imgrefs_file_prompt);
	if(s != NULL)
	{
	    strncpy(legend->imgrefs, s, sizeof(legend->imgrefs));
	    legend->imgrefs[sizeof(legend->imgrefs) - 1] = '\0';
	}

	/* SndRefs File */
	s = PromptGetS(&uhw->sndrefs_file_prompt);
	if(s != NULL)
	{
	    strncpy(legend->sndrefs, s, sizeof(legend->sndrefs));
	    legend->sndrefs[sizeof(legend->sndrefs) - 1] = '\0';
	}

	/* Player start position */
	UNVParseLocation(
	    PromptGetS(&uhw->player_start_pos_prompt),
	    &legend->player_start_sect_x,
	    &legend->player_start_sect_y,
	    &legend->player_start_sect_z,
	    &legend->player_start_x,
	    &legend->player_start_y,
	    &legend->player_start_z
	);

	/* Player start direction */
	UNVParseDirection(
	    PromptGetS(&uhw->player_start_dir_prompt),
	    &legend->player_start_heading,
	    &legend->player_start_pitch,
	    &legend->player_start_bank
	);

	/* Guest start position */
	UNVParseLocation(
	    PromptGetS(&uhw->guest_start_pos_prompt),
	    &legend->guest_start_sect_x,
	    &legend->guest_start_sect_y, 
	    &legend->guest_start_sect_z,
	    &legend->guest_start_x,
	    &legend->guest_start_y,
	    &legend->guest_start_z 
	);
 
	/* Guest start direction */
	UNVParseDirection(
	    PromptGetS(&uhw->guest_start_dir_prompt),
	    &legend->guest_start_heading,
	    &legend->guest_start_pitch,
	    &legend->guest_start_bank
	);

	legend->antimatter_low_warning = MAX(PromptGetF(
	   &uhw->antimatter_low_warning_prompt
	), 0.0f);
	legend->asteroid_field_damage_int = MAX(PromptGetL(
	   &uhw->asteroid_field_damage_int_prompt
	), 0l);
	legend->asteroid_field_damage_rate = MAX(PromptGetF( 
	   &uhw->asteroid_field_damage_rate_prompt
	), 0.0f); 
	legend->auto_destruct_damage_range = MAX(PromptGetF( 
	   &uhw->auto_destruct_damage_range_prompt
	), 0.0f);
	legend->cloak_antimatter_rate = MAX(PromptGetF( 
	   &uhw->cloak_antimatter_rate_prompt
	), 0.0f); 
	legend->com_range = MAX(PromptGetF( 
	   &uhw->com_range_prompt
	), 0.0f);
	legend->damage_control_rate = MAX(PromptGetF(
	   &uhw->damage_control_rate_prompt
	), 0.0f);
	legend->disarmed_weapon_lifespan = MAX(PromptGetL(
	   &uhw->disarmed_weapon_lifespan_prompt
	), 0l); 
	legend->engine_antimatter_rate = MAX(PromptGetF( 
	   &uhw->engine_antimatter_rate_prompt
	), 0.0f);
	legend->no_antimatter_thrust_coeff = CLIP(PromptGetF( 
	   &uhw->no_antimatter_thrust_coeff_prompt
	), 0.0f, 1.0f);
	legend->on_fire_hp_coeff = CLIP(PromptGetF(
	   &uhw->on_fire_hp_coeff_prompt
	), 0.0f, 1.0f);
	legend->tractor_beam_antimatter_rate = MAX(PromptGetF( 
	   &uhw->tractor_beam_antimatter_rate_prompt
	), 0.0f);
	legend->tractor_beam_range = MAX(PromptGetF( 
	   &uhw->tractor_beam_range_prompt
	), 0.0f);
	legend->transporter_range = MAX(PromptGetF( 
	   &uhw->transporter_range_prompt
	), 0.0f);


	free(objects_list);


	/* Check if any files are missing */

	/* OCS Names */
	path = ISPATHABSOLUTE(legend->ocsn) ?
	    STRDUP(legend->ocsn) :
	    STRDUP(PrefixPaths(dname.etc, legend->ocsn));
	if(!UECheckFileExists(path, &error_msg))
	{
	    char *s = (char *)malloc(
		STRLEN(error_msg) + STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s",
		error_msg, path
	    );
	    UEDialog(
		"Missing OCS Names File",
		UE_ICON_WARNING,
		s,
		uew->toplevel
	    );
	    free(s);
	}
	free(error_msg);
	free(path);

	/* Imgrefs */
	path = ISPATHABSOLUTE(legend->imgrefs) ?
	    STRDUP(legend->imgrefs) :
	    STRDUP(PrefixPaths(dname.images, legend->imgrefs));
	if(!UECheckFileExists(path, &error_msg))
	{
	    char *s = (char *)malloc(
		STRLEN(error_msg) + STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s",
		error_msg, path
	    );
	    UEDialog(
		"Missing ImgRefs File",
		UE_ICON_WARNING,
		s,
		uew->toplevel
	    );
	    free(s);
	}
	free(error_msg);
	free(path);

	/* SndRefs */
	path = ISPATHABSOLUTE(legend->sndrefs) ?
	    STRDUP(legend->sndrefs) :
	    STRDUP(PrefixPaths(dname.sounds, legend->sndrefs));
	if(!UECheckFileExists(path, &error_msg))
	{
	    char *s = (char *)malloc(
		STRLEN(error_msg) + STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s",
		error_msg, path
	    );
	    UEDialog(
		"Missing SndRefs File",
		UE_ICON_WARNING,
		s,
		uew->toplevel
	    );
	    free(s);
	}
	free(error_msg);
	free(path);


	/* Begin updating displayed data to reflect changes */

	/* Reset has changes marker on uhw */
	uhw->has_changes = False;

	/* Set has changes mark on the Universe Editor Window */
	if(!uew->has_changes)
	{
	    uew->has_changes = True;
	    UEWUpdateTitle(uew);
	}

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	/* Open the new ImgRefs file */
	UEWOpenImgRefs(uew, legend->imgrefs);

	UEWDoUpdateWindowMenus();
}


/*
 *	Draw callback.
 */
static void UHWDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	uhw_struct *uhw = UHW(widget);
	if(uhw == NULL)
	    return;

	if(uhw->map_state)
	    UHWDraw(uhw, amount);
}

/*
 *	Browse OCSNs file callback.
 */
static void UHWSelectOCSNsFileCB(push_button_struct *btn, void *data)
{
	const char *mask = "*.ocsn";
	int npaths;
	const char *s;
	char *path, **paths_list;
	prompt_struct *prompt;
	uhw_struct *uhw = UHW(data);
	if(uhw == NULL)
	    return;

	prompt = &uhw->ocsn_file_prompt;
	s = PromptGetS(prompt);
	if(ISPATHABSOLUTE(s))
	{
	    char *parent = STRDUP(GetParentDir(PromptGetS(prompt)));
	    path = STRDUP(PrefixPaths(parent, mask));
	    free(parent);
	}
	else
	{
	    path = STRDUP(PrefixPaths(dname.etc, mask));
	}

	if(UESelectFile(
	    "Select OCSNs File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    uhw->toplevel
	))
	{
	    char *new_path = (npaths > 0) ?
		STRDUP(paths_list[npaths - 1]) : NULL;
	    if(new_path != NULL)
	    {
		StripParentPath(new_path, dname.etc);
		PromptSetS(prompt, new_path);
		free(new_path);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
}

/*
 *	Browse ImgRefs file callback.
 */
static void UHWSelectImgRefsFileCB(push_button_struct *btn, void *data)
{
	const char *mask = "*.isr";
	int npaths;
	const char *s;
	char *path, **paths_list;
	prompt_struct *prompt;
	uhw_struct *uhw = UHW(data);
	if(uhw == NULL)
	    return;

	prompt = &uhw->imgrefs_file_prompt;
	s = PromptGetS(prompt);
	if(ISPATHABSOLUTE(s))
	{
	    char *parent = STRDUP(GetParentDir(PromptGetS(prompt)));
	    path = STRDUP(PrefixPaths(parent, mask));
	    free(parent);
	}
	else
	{
	    path = STRDUP(PrefixPaths(dname.images, mask));
	}

	if(UESelectFile(
	    "Select ImgRefs File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    uhw->toplevel
	))
	{
	    char *new_path = (npaths > 0) ?
		STRDUP(paths_list[npaths - 1]) : NULL;
	    if(new_path != NULL)
	    {
		StripParentPath(new_path, dname.images);
		PromptSetS(prompt, new_path);
		free(new_path);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
}

/*
 *	Browse SndRefs file callback.
 */
static void UHWSelectSndRefsFileCB(push_button_struct *btn, void *data)
{
	const char *mask = "*.ss";
	int npaths;
	const char *s;
	char *path, **paths_list;
	prompt_struct *prompt;
	uhw_struct *uhw = UHW(data);
	if(uhw == NULL)
	    return;

	prompt = &uhw->sndrefs_file_prompt;
	s = PromptGetS(prompt);
	if(ISPATHABSOLUTE(s))
	{
	    char *parent = STRDUP(GetParentDir(PromptGetS(prompt)));
	    path = STRDUP(PrefixPaths(parent, mask));
	    free(parent);
	}
	else
	{
	    path = STRDUP(PrefixPaths(dname.sounds, mask));
	}

	if(UESelectFile(
	    "Select SndRefs File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    uhw->toplevel
	))
	{
	    char *new_path = (npaths > 0) ?
		STRDUP(paths_list[npaths - 1]) : NULL;
	    if(new_path != NULL)
	    {
		StripParentPath(new_path, dname.sounds);
		PromptSetS(prompt, new_path);
		free(new_path);
	    }
	}

	strlistfree(paths_list, npaths);
	free(path);
}


/*
 *	OK callback.
 */
static void UHWOKCB(push_button_struct *btn, void *data)
{
	uhw_struct *uhw = UHW(data);
	if(uhw == NULL)
	    return;

	UHWSetValues(uhw);
	UHWUnmap(uhw);
}

/*
 *	Apply callback.
 */
static void UHWApplyCB(push_button_struct *btn, void *data)
{
	uhw_struct *uhw = UHW(data);
	if(uhw == NULL)
	    return;

	UHWSetValues(uhw);
}

/*
 *	Cancel callback.
 */
static void UHWCancelCB(push_button_struct *btn, void *data)
{ 
	uhw_struct *uhw = UHW(data);
	if(uhw == NULL)
	    return;

	UHWUnmap(uhw);
}


/*
 *	Creates a new Universe Header Window.
 */
uhw_struct *UHWNew(uew_struct *uew)
{
	int x, y;
	unsigned int width, height, width2, height2;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	push_button_struct *btn;
	widget_global_struct *wglobal = &widget_global;
	uhw_struct *uhw = UHW(calloc(1, sizeof(uhw_struct)));

	uhw->map_state = False;
	uhw->has_focus = False;
	uhw->x = 0;
	uhw->y = 0;
	uhw->width = UHW_WIDTH;
	uhw->height = UHW_HEIGHT;

	uhw->uew = uew;

	uhw->has_changes = False;


	/* Create toplevel */
	x = uhw->x;
	y = uhw->y;
	width = uhw->width;
	height = uhw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	uhw->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    UHW_DEF_TITLE,
	    UHW_DEF_ICON_TITLE,
	    (ue_image.icon_unvedit_pixmap != None) ?
		ue_image.icon_unvedit_pixmap :   
		wglobal->std_icon_pm,
	    False,			/* Let WM set coordinates? */
	    x, y,
	    width, height,
	    width, height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	if(uew != NULL)                          
	    OSWSetTransientFor(w, uew->toplevel);

	parent = w;

	/* Title prompt */
	width = (unsigned int)MAX(
	    (((int)uhw->width / 2) - (int)(1.5f * UHW_MARGIN)),
	    1
	);
	height = UHW_PROMPT_HEIGHT;
	x = UHW_MARGIN;
	y = UHW_MARGIN;
	PromptInit(
	    &uhw->title_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Title:",
	    SW_UNV_TITLE_MAX, 0
	);
	PromptSetTooltip(
	   &uhw->title_prompt,
"Enter the title of this universe"
	);

	/* Convert ru to au prompt */
	y += (int)height + (UHW_MARGIN / 2);
	PromptInit(
	    &uhw->ru_to_au_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "RU To AU:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->ru_to_au_prompt,
"Enter the coefficient that is to be multiplied with all real unit (ru)\
 values in order to convert them to astronomical unit (au) values"
	);

	/* Lost and found owner prompt */
	y += (int)height + (UHW_MARGIN / 2);
	PromptInit(
	    &uhw->lost_found_owner_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Lost/Found Owner:",
	    SW_OBJ_NAME_MAX + 20, 0
	);
	PromptSetTooltip(
	   &uhw->lost_found_owner_prompt,
"Enter the name or number of the object that will be the\
 owner of all unclaimed objects"
	);

	/* OSCN File prompt */
	y += (int)height + (UHW_MARGIN / 2);
	width2 = (unsigned int)MAX(
	    (int)width - 30,
	    1
	);
	height2 = height;
	PromptInit(
	    &uhw->ocsn_file_prompt, parent,
	    x, y, width2, height2,
	    PROMPT_STYLE_FLUSHED,
	    "OCSN File:",
	    PATH_MAX + NAME_MAX, 0
	);
	PromptSetTooltip(
	   &uhw->ocsn_file_prompt,
"Enter the file name of the Object Create Script Names (OCSN) file that\
 all clients will be instructed to use for this universe"
	);
	UHWCreateBrowseButton(
	    uhw,
	    &uhw->ocsn_file_prompt,
	    parent,
	    &uhw->ocsn_file_btn,
	    UHWSelectOCSNsFileCB
	);

	/* ImgRefs File prompt */
	y += (int)height;
	PromptInit(
	    &uhw->imgrefs_file_prompt, parent,
	    x, y, width2, height2,
	    PROMPT_STYLE_FLUSHED,
	    "ImgRef File:",
	    PATH_MAX + NAME_MAX, 0
	);
	PromptSetTooltip(
	   &uhw->imgrefs_file_prompt,
"Enter the file name of the Image References (ImgRefs) file that\
 all clients will be instructed to use for this universe"
	);
	UHWCreateBrowseButton(
	    uhw,
	    &uhw->imgrefs_file_prompt,
	    parent,
	    &uhw->imgrefs_file_btn,
	    UHWSelectImgRefsFileCB
	);

	/* SndRefs File prompt */
	y += (int)height;
	PromptInit(
	    &uhw->sndrefs_file_prompt, parent,
	    x, y, width2, height2,
	    PROMPT_STYLE_FLUSHED,
	    "SndRefs File:",
	    PATH_MAX + NAME_MAX, 0
	);
	PromptSetTooltip(
	   &uhw->sndrefs_file_prompt,
"Enter the file name of the Sound References (SndRefs) file that\
 all clients will be instructed to use for this universe"
	);
	UHWCreateBrowseButton(
	    uhw,
	    &uhw->sndrefs_file_prompt,
	    parent,
	    &uhw->sndrefs_file_btn,
	    UHWSelectSndRefsFileCB
	);


	/* Player start position */
	y += (int)height + (UHW_MARGIN / 2);
	PromptInit(
	    &uhw->player_start_pos_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Player Start Position:",
	    256, 0
	);
	PromptSetTooltip(
	   &uhw->player_start_pos_prompt,
"Enter the position of the default player starting position"
	);

	/* Player start direction */
	y += (int)height;
	PromptInit(
	    &uhw->player_start_dir_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Player Start Direction:", 
	    256, 0
	);
	PromptSetTooltip(
	   &uhw->player_start_dir_prompt,
"Enter the direction of the default player starting direction"
	);

	/* Guest start position */
	y += (int)height + (UHW_MARGIN / 2);
	PromptInit(
	    &uhw->guest_start_pos_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Guest Start Position:",
	    256, 0
	);
	PromptSetTooltip(
	   &uhw->guest_start_pos_prompt,
"Enter the position of where guest vessels will be placed at"
	);

	/* Guest start direction */
	y += (int)height;
	PromptInit(
	    &uhw->guest_start_dir_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Guest Start Direction:",
	    256, 0
	);
	PromptSetTooltip(
	   &uhw->guest_start_dir_prompt,
"Enter the direction of where guest vessels will be directed towards"
	);


	/* Antimatter Low Warning */
	x += (int)width + UHW_MARGIN;
	y = UHW_MARGIN;
	width = (unsigned int)MAX(
	    (((int)uhw->width / 2) - (int)(1.5f * UHW_MARGIN)),
	    1
	);
	PromptInit(
	    &uhw->antimatter_low_warning_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Antimatter Low Warning:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->antimatter_low_warning_prompt,
"Enter the units of antimatter at which the player will be warned\
 about when the amount of antimatter falls below this value (set\
 this value to 0 to never warn)"
	);

	/* Asteroid Field Damage Int */
	y += (int)height;
	PromptInit(
	    &uhw->asteroid_field_damage_int_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Asteroid Field Damage Int:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->asteroid_field_damage_int_prompt,
"Enter the interval in milliseconds at which damage is applied\
 when a vessel flies above the velocity of the asteroid fueld's\
 velocity_max value"
	);

	/* Asteroid Field Damage Rate */
	y += (int)height;
	PromptInit(
	    &uhw->asteroid_field_damage_rate_prompt, parent,
	    x, y, width, height, 
	    PROMPT_STYLE_FLUSHED,
	    "Asteroid Field Damage Rate:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->asteroid_field_damage_rate_prompt,
"Enter the rate (hp -= this_value * size_ru * velocity_coeff / cycle)\
 at which objects will receive damage when they fly above the velocity\
 of the asteroid field's velocity_max value"
	);

	/* Auto Destruct Damage Range */
	y += (int)height;
	PromptInit(
	    &uhw->auto_destruct_damage_range_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Auto Destruct Damage Range:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->auto_destruct_damage_range_prompt,
"Enter the range (in real units) at which other objects within this\
 range of an object detonated by self-destruct will receive damage\
 (set this value to 0 to specify that auto-destruct never causes\
 damage to other objects)"
	);

	/* Cloak Antimatter Rate */
	y += (int)height;
	PromptInit(
	    &uhw->cloak_antimatter_rate_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Cloak Antimatter Rate:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->cloak_antimatter_rate_prompt,
"Enter the rate (antimatter -= this_value * size_ru / cycle) at which\
 the cloak consumes antimatter"
	);

	/* Com Range */
	y += (int)height;
	PromptInit(
	    &uhw->com_range_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Com Range:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->com_range_prompt,
"Enter the maximum range (in real units) of hails and com\
 messages in normal space"
	);

	/* Damage Control Rate */
	y += (int)height;
	PromptInit(
	    &uhw->damage_control_rate_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Damage Control Rate:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->damage_control_rate_prompt,
"Enter the rate (hp += this_value * power * power_purity / cycle)\
 at which damage control repairs hitpoints"
	);

	/* Disarmed Weapon Lifespan */
	y += (int)height;
	PromptInit(
	    &uhw->disarmed_weapon_lifespan_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Disarmed Weapon Lifespan:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->disarmed_weapon_lifespan_prompt,
"Enter the time (in milliseconds) at which disarmed weapons will\
 remain before they are automatically recycled (set this value\
 to 0 to never recycle disarmed weapons)"
	);

	/* Engine Antimatter Rate */
	y += (int)height;
	PromptInit(
	    &uhw->engine_antimatter_rate_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Engine Antimatter Rate:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->engine_antimatter_rate_prompt,
"Enter the rate (antimatter -= this_value * throttle / cycle) at\
 which antimatter is consumed by the engine"
	);

	/* No Antimatter Thrust Coeff */
	y += (int)height;
	PromptInit(
	    &uhw->no_antimatter_thrust_coeff_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "No Antimatter Thrust Coeff:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->no_antimatter_thrust_coeff_prompt,
"Enter the coefficient that will be multiplied to the thrust\
 output when antimatter runs out (set this value to 0 to\
 completely disable thrust output when antimatter runs out)"
	);

	/* On Fire HP Coeff */
	y += (int)height;
	PromptInit(
	    &uhw->on_fire_hp_coeff_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "On Fire HP Coeff:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->on_fire_hp_coeff_prompt,
"Enter the coefficient (coeff = hp / hp_max) at which\
 a controlled or player object will catch on fire when its\
 hitpoints coefficient drops below this value"
	);

	/* Tractor Beam Antimatter Rate */
	y += (int)height;
	PromptInit(
	    &uhw->tractor_beam_antimatter_rate_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Tractor Beam Antimatter Rate:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->tractor_beam_antimatter_rate_prompt,
"Enter the rate (antimatter -= this_value * size_ru / cycle) at which\
 antimatter is consumed by the tractor beam (set this value to 0 to\
 specify that the tractor beam never consumes antimatter)"
	);

	/* Tractor Beam Range */
	y += (int)height;
	PromptInit(      
	    &uhw->tractor_beam_range_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Tractor Beam Range:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->tractor_beam_range_prompt,
"Enter the maximum range (in real units) of the tractor beam"
	);

	/* Transporter Range */
	y += (int)height;
	PromptInit(
	    &uhw->transporter_range_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Transporter Range:",
	    40, 0
	);
	PromptSetTooltip(
	   &uhw->transporter_range_prompt,
"Enter the maximum range (in real units) of the transporter\
 (for economy transactions and transfering items between vessels)"
	);


	/* Link prompts togeather */
	uhw->title_prompt.prev = &uhw->transporter_range_prompt;
	uhw->title_prompt.next = &uhw->ru_to_au_prompt;

	uhw->ru_to_au_prompt.prev = &uhw->title_prompt;
	uhw->ru_to_au_prompt.next = &uhw->lost_found_owner_prompt;

	uhw->lost_found_owner_prompt.prev = &uhw->ru_to_au_prompt;
	uhw->lost_found_owner_prompt.next = &uhw->ocsn_file_prompt;

	uhw->ocsn_file_prompt.prev = &uhw->lost_found_owner_prompt;
	uhw->ocsn_file_prompt.next = &uhw->imgrefs_file_prompt;

	uhw->imgrefs_file_prompt.prev = &uhw->ocsn_file_prompt;
	uhw->imgrefs_file_prompt.next = &uhw->sndrefs_file_prompt;

	uhw->sndrefs_file_prompt.prev = &uhw->imgrefs_file_prompt;
	uhw->sndrefs_file_prompt.next = &uhw->player_start_pos_prompt;

	uhw->player_start_pos_prompt.prev = &uhw->sndrefs_file_prompt;
	uhw->player_start_pos_prompt.next = &uhw->player_start_dir_prompt;

	uhw->player_start_dir_prompt.prev = &uhw->player_start_pos_prompt;
	uhw->player_start_dir_prompt.next = &uhw->guest_start_pos_prompt;

	uhw->guest_start_pos_prompt.prev = &uhw->player_start_dir_prompt;
	uhw->guest_start_pos_prompt.next = &uhw->guest_start_dir_prompt;

	uhw->guest_start_dir_prompt.prev = &uhw->guest_start_pos_prompt;
	uhw->guest_start_dir_prompt.next = &uhw->antimatter_low_warning_prompt;


	uhw->antimatter_low_warning_prompt.prev = &uhw->guest_start_dir_prompt;
	uhw->antimatter_low_warning_prompt.next = &uhw->asteroid_field_damage_rate_prompt;

	uhw->asteroid_field_damage_int_prompt.prev = &uhw->antimatter_low_warning_prompt;
	uhw->asteroid_field_damage_int_prompt.next = &uhw->asteroid_field_damage_rate_prompt;

	uhw->asteroid_field_damage_rate_prompt.prev = &uhw->asteroid_field_damage_int_prompt;
	uhw->asteroid_field_damage_rate_prompt.next = &uhw->auto_destruct_damage_range_prompt;

	uhw->auto_destruct_damage_range_prompt.prev = &uhw->asteroid_field_damage_rate_prompt;
	uhw->auto_destruct_damage_range_prompt.next = &uhw->cloak_antimatter_rate_prompt;

	uhw->cloak_antimatter_rate_prompt.prev = &uhw->auto_destruct_damage_range_prompt;
	uhw->cloak_antimatter_rate_prompt.next = &uhw->com_range_prompt;

	uhw->com_range_prompt.prev = &uhw->cloak_antimatter_rate_prompt;
	uhw->com_range_prompt.next = &uhw->damage_control_rate_prompt;

	uhw->damage_control_rate_prompt.prev = &uhw->com_range_prompt;
	uhw->damage_control_rate_prompt.next = &uhw->disarmed_weapon_lifespan_prompt;

	uhw->disarmed_weapon_lifespan_prompt.prev = &uhw->damage_control_rate_prompt;
	uhw->disarmed_weapon_lifespan_prompt.next = &uhw->engine_antimatter_rate_prompt;

	uhw->engine_antimatter_rate_prompt.prev = &uhw->disarmed_weapon_lifespan_prompt;
	uhw->engine_antimatter_rate_prompt.next = &uhw->no_antimatter_thrust_coeff_prompt;

	uhw->no_antimatter_thrust_coeff_prompt.prev = &uhw->engine_antimatter_rate_prompt;
	uhw->no_antimatter_thrust_coeff_prompt.next = &uhw->on_fire_hp_coeff_prompt;

	uhw->on_fire_hp_coeff_prompt.prev = &uhw->no_antimatter_thrust_coeff_prompt;
	uhw->on_fire_hp_coeff_prompt.next = &uhw->tractor_beam_antimatter_rate_prompt;

	uhw->tractor_beam_antimatter_rate_prompt.prev = &uhw->on_fire_hp_coeff_prompt;
	uhw->tractor_beam_antimatter_rate_prompt.next = &uhw->tractor_beam_range_prompt;

	uhw->tractor_beam_range_prompt.prev = &uhw->tractor_beam_antimatter_rate_prompt;
	uhw->tractor_beam_range_prompt.next = &uhw->transporter_range_prompt;

	uhw->transporter_range_prompt.prev = &uhw->tractor_beam_range_prompt;
	uhw->transporter_range_prompt.next = &uhw->title_prompt;


	/* Buttons */
	btn = &uhw->ok_btn;
	width = UHW_BTN_WIDTH;
	height = UHW_BTN_HEIGHT;
	x = UHW_MARGIN;
	y = (int)uhw->height - UHW_MARGIN - (int)height;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    UHWOKCB, uhw
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_Return;
	accel_mods[1] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 2);

	btn = &uhw->apply_btn;
	x = ((int)uhw->width - (int)width) / 2;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Apply",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    UHWApplyCB, uhw
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	btn = &uhw->cancel_btn;
	x = uhw->width - UHW_MARGIN - (int)width;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    UHWCancelCB, uhw
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	return(uhw);
}

/*
 *	Draws the Universe Header Window.
 */
void UHWDraw(uhw_struct *uhw, const int amount)
{
	unsigned int width, height;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(uhw == NULL)
	    return;

	if(uhw->toplevel != None)
	{
	    int y;

	    width = uhw->width;
	    height = uhw->height;
	    w = uhw->toplevel;

	    /* Draw the background */ 
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(
		    w, 0, 0, width, height
		);

		OSWSetForeground(gui->white_pix);
		y = (int)height - UHW_MARGIN - UHW_BTN_HEIGHT -
		    UHW_MARGIN;
		OSWDrawLine(
		    w, 0, y, (int)width, y
		);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);

		OSWSetForeground(wglobal->surface_shadow_pix);
		y = (int)height - UHW_MARGIN - UHW_BTN_HEIGHT -
		    UHW_MARGIN;
		OSWDrawLine(w, 0, y, (int)width, y);
		OSWSetForeground(wglobal->surface_highlight_pix);
		y += 1;
		OSWDrawLine(w, 0, y, (int)width, y);
	    }
	}
}

/*
 *	Queues the Universe Header Window to draw.
 */
void UHWQueueDraw(uhw_struct *uhw, const int amount)
{
	if(uhw == NULL)
	    return;

	WidgetQueueDrawAction(
	    UHWDrawCB, uhw, (void *)amount
	);
}

/*
 *	Manages the Universe Header Window event.
 */
int UHWManage(uhw_struct *uhw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;

	if((uhw == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case Expose:
	    if(w == uhw->toplevel)
	    {
		UHWQueueDraw(uhw, 0);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!uhw->has_focus)
		return(events_handled);;

	    break;

	  case KeyRelease:
	    if(!uhw->has_focus)
		return(events_handled);;

	    break;

	  case FocusIn:
	    if(w == uhw->toplevel)
	    {
		uhw->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut: 
	    if(w == uhw->toplevel)
	    {
		uhw->has_focus = False;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(uhw->toplevel, event, &t))
	    {
		OSWGrabFocusTime(uhw->toplevel, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */  
	    else if(OSWEventIsDeleteWindow(uhw->toplevel, event))
	    {
		UHWCancelCB(&uhw->cancel_btn, uhw);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->title_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->ru_to_au_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->lost_found_owner_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->ocsn_file_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->ocsn_file_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->imgrefs_file_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->imgrefs_file_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->sndrefs_file_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->sndrefs_file_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->player_start_pos_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->player_start_dir_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->guest_start_pos_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->guest_start_dir_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->antimatter_low_warning_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->asteroid_field_damage_int_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->asteroid_field_damage_rate_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->auto_destruct_damage_range_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->cloak_antimatter_rate_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->com_range_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->damage_control_rate_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->disarmed_weapon_lifespan_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->engine_antimatter_rate_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->no_antimatter_thrust_coeff_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->on_fire_hp_coeff_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->tractor_beam_antimatter_rate_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->tractor_beam_range_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->transporter_range_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->ok_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->apply_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&uhw->cancel_btn, event);

	return(events_handled);
}
	
/*
 *	Maps the Universe Header Window.
 */
void UHWMap(uhw_struct *uhw)
{
	if(uhw == NULL)
	    return;

	WidgetSetFocus(&uhw->title_prompt, False);
	WidgetSetFocus(&uhw->ru_to_au_prompt, False);
	WidgetSetFocus(&uhw->lost_found_owner_prompt, False);
	WidgetSetFocus(&uhw->ocsn_file_prompt, False);
	WidgetSetFocus(&uhw->imgrefs_file_prompt, False);
	WidgetSetFocus(&uhw->sndrefs_file_prompt, False);
	WidgetSetFocus(&uhw->player_start_pos_prompt, False);
	WidgetSetFocus(&uhw->player_start_dir_prompt, False);
	WidgetSetFocus(&uhw->guest_start_pos_prompt, False);
	WidgetSetFocus(&uhw->guest_start_dir_prompt, False);

	OSWMapRaised(uhw->toplevel);
	uhw->map_state = True;

	WidgetMap(&uhw->title_prompt);
	WidgetMap(&uhw->ru_to_au_prompt);
	WidgetMap(&uhw->lost_found_owner_prompt);
	WidgetMap(&uhw->ocsn_file_prompt);
	WidgetMap(&uhw->ocsn_file_btn);
	WidgetMap(&uhw->imgrefs_file_prompt);
	WidgetMap(&uhw->imgrefs_file_btn);
	WidgetMap(&uhw->sndrefs_file_prompt);
	WidgetMap(&uhw->sndrefs_file_btn);
	WidgetMap(&uhw->player_start_pos_prompt);
	WidgetMap(&uhw->player_start_dir_prompt);
	WidgetMap(&uhw->guest_start_pos_prompt);
	WidgetMap(&uhw->guest_start_dir_prompt);

	WidgetMap(&uhw->antimatter_low_warning_prompt);
	WidgetMap(&uhw->asteroid_field_damage_int_prompt);
	WidgetMap(&uhw->asteroid_field_damage_rate_prompt);
	WidgetMap(&uhw->auto_destruct_damage_range_prompt);
	WidgetMap(&uhw->cloak_antimatter_rate_prompt);
	WidgetMap(&uhw->com_range_prompt);
	WidgetMap(&uhw->damage_control_rate_prompt);
	WidgetMap(&uhw->disarmed_weapon_lifespan_prompt);
	WidgetMap(&uhw->engine_antimatter_rate_prompt);
	WidgetMap(&uhw->no_antimatter_thrust_coeff_prompt);
	WidgetMap(&uhw->on_fire_hp_coeff_prompt);
	WidgetMap(&uhw->tractor_beam_antimatter_rate_prompt);
	WidgetMap(&uhw->tractor_beam_range_prompt);
	WidgetMap(&uhw->transporter_range_prompt);
 
	WidgetMap(&uhw->ok_btn);
	WidgetMap(&uhw->apply_btn);
	WidgetMap(&uhw->cancel_btn);

	OSWGrabFocus(uhw->toplevel);

	PromptMarkBuffer(&uhw->title_prompt, PROMPT_POS_END);
	WidgetSetFocus(&uhw->title_prompt, True);

	UEWDoUpdateWindowMenus();
}

/*
 *	Gets the values from the Universe Editor referenced by the
 *	Universe Header Window and maps the Universe Header Window.
 */
void UHWMapValues(uhw_struct *uhw)
{
	uew_struct *uew;

	if(uhw == NULL)
	    return;

	uew = uhw->uew;

	UHWGetValues(uhw);

	if(uew != NULL)
	    WidgetCenterWindowToWindow(uhw->toplevel, uew->toplevel);

	uhw->has_changes = False;

	UHWMap(uhw);
}

/*
 *	Unmaps the Universe Header Window.
 */
void UHWUnmap(uhw_struct *uhw)
{
	if(uhw == NULL)
	    return;

	OSWUnmapWindow(uhw->toplevel);
	uhw->map_state = False;

	UEWDoUpdateWindowMenus();
}

/*
 *	Deletes the Universe Header Window.
 */
void UHWDelete(uhw_struct *uhw)
{
	if(uhw == NULL)
	    return;

	if(IDC())
	{
	    WidgetDestroy(&uhw->ok_btn);
	    WidgetDestroy(&uhw->apply_btn);
	    WidgetDestroy(&uhw->cancel_btn);

	    WidgetDestroy(&uhw->title_prompt);
	    WidgetDestroy(&uhw->ru_to_au_prompt);
	    WidgetDestroy(&uhw->lost_found_owner_prompt);
	    WidgetDestroy(&uhw->ocsn_file_prompt);
	    WidgetDestroy(&uhw->ocsn_file_btn);
	    WidgetDestroy(&uhw->imgrefs_file_prompt);
	    WidgetDestroy(&uhw->imgrefs_file_btn);
	    WidgetDestroy(&uhw->sndrefs_file_prompt);
	    WidgetDestroy(&uhw->sndrefs_file_btn);
	    WidgetDestroy(&uhw->player_start_pos_prompt);
	    WidgetDestroy(&uhw->player_start_dir_prompt);
	    WidgetDestroy(&uhw->guest_start_pos_prompt);
	    WidgetDestroy(&uhw->guest_start_dir_prompt);

	    WidgetDestroy(&uhw->antimatter_low_warning_prompt);
	    WidgetDestroy(&uhw->asteroid_field_damage_int_prompt);
	    WidgetDestroy(&uhw->asteroid_field_damage_rate_prompt);
	    WidgetDestroy(&uhw->auto_destruct_damage_range_prompt);
	    WidgetDestroy(&uhw->cloak_antimatter_rate_prompt);
	    WidgetDestroy(&uhw->com_range_prompt);
	    WidgetDestroy(&uhw->damage_control_rate_prompt);
	    WidgetDestroy(&uhw->disarmed_weapon_lifespan_prompt);
	    WidgetDestroy(&uhw->engine_antimatter_rate_prompt);
	    WidgetDestroy(&uhw->no_antimatter_thrust_coeff_prompt);
	    WidgetDestroy(&uhw->on_fire_hp_coeff_prompt);
	    WidgetDestroy(&uhw->tractor_beam_antimatter_rate_prompt);
	    WidgetDestroy(&uhw->tractor_beam_range_prompt);
	    WidgetDestroy(&uhw->transporter_range_prompt);

	    OSWDestroyWindow(&uhw->toplevel);
	}

	free(uhw);
}
