#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "../include/osw-x.h"
#include "../include/widget.h"
#include "../include/objects.h"
#include "../include/unvutil.h"

#include "ocsinfofio.h"
#include "sndrefs.h"
#include "wepw.h"
#include "uew.h"
#include "ue.h"

#include "../include/oswkeyvals.h"


static xsw_object_struct *WepWGetObjectPtr(wepw_struct *wepw);
static xsw_weapon_struct *WepWGetWeaponPtr(
	wepw_struct *wepw, const int wep_num
);

static void WepWGetSndRefsToPUList(
	wepw_struct *wepw, popup_list_struct *pulist
);

static void WepWGetValues(wepw_struct *wepw);
static void WepWSetValues(wepw_struct *wepw);
static void WepWGetWeaponValues(wepw_struct *wepw, const int wep_num);
static void WepWSetWeaponValues(wepw_struct *wepw, const int wep_num);

static void WepWDrawCB(void *widget, void *data);

static void WepWCListSelectCB(
	clist_struct *clist, int row_num, void *data
);
static void WepWCListUnselectCB(
	clist_struct *clist, int row_num, void *data
);

static void WepWWeaponAddCB(push_button_struct *btn, void *data);
static void WepWWeaponRemoveCB(push_button_struct *btn, void *data);
static void WepWCloseCB(push_button_struct *btn, void *data);

wepw_struct *WepWNew(uew_struct *uew, const int obj_num);
void WepWDraw(wepw_struct *wepw, const int amount);
void WepWQueueDraw(wepw_struct *wepw, const int amount);
int WepWManage(wepw_struct *wepw, event_t *event);
void WepWMap(wepw_struct *wepw);
void WepWMapValues(wepw_struct *wepw);
void WepWUnmap(wepw_struct *wepw);
void WepWDelete(wepw_struct *wepw);


#define WEPW_TITLE			"Weapons"
#define WEPW_ICON_TITLE			"Weapons"

#define WEPW_MARGIN			10

#define WEPW_WIDTH			400
#define WEPW_HEIGHT			505

#define WEPW_PROMPT_HEIGHT		30
#define WEPW_PULIST_HEIGHT		26

#define WEPW_BUTTON_WIDTH		70
#define WEPW_BUTTON_HEIGHT		28


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Gets the pointer to the object referenced on the Weapons
 *	Window.
 */
static xsw_object_struct *WepWGetObjectPtr(wepw_struct *wepw)
{
	uew_struct *uew = (wepw != NULL) ? wepw->uew : NULL;
	if(uew == NULL)
	    return(NULL);

	return(UEWGetObjectPtr(uew, wepw->obj_num));
}

/*
 *	Gets the pointer to the weapon.
 */
static xsw_weapon_struct *WepWGetWeaponPtr(
	wepw_struct *wepw, const int wep_num
)
{
	xsw_object_struct *obj = WepWGetObjectPtr(wepw);
	if(obj == NULL)
	    return(NULL);

	if((wep_num < 0) || (wep_num >= obj->total_weapons))
	    return(NULL);

	return(obj->weapons[wep_num]);
}


/*
 *	Opens the SndRefs file to the specified Popup List.
 */
static void WepWGetSndRefsToPUList(
	wepw_struct *wepw, popup_list_struct *pulist
)
{
	int total;
	const char *name;
	char *path;
	sndref_struct **sndrefs_list;
	uew_struct *uew;

	if((wepw == NULL) || (pulist == NULL))
	    return;

	uew = wepw->uew;
	if(uew == NULL)
	    return;

	name = uew->legend.sndrefs;
	path = ISPATHABSOLUTE(name) ?
	    STRDUP(name) :
	    STRDUP(PrefixPaths(dname.sounds, name));

	PUListClear(pulist);

	sndrefs_list = NULL;
	total = 0;
	if(!SndRefsFileOpen(
	    &sndrefs_list, &total, path,
	    NULL,
	    NULL, NULL, NULL
	))
	{
	    int i, n;
	    char *s;
	    sndref_struct *sndref;
	    for(i = 0; i < total; i++)
	    {
		sndref = sndrefs_list[i];
		if(sndref == NULL)
		    continue;

		if(STRISEMPTY(sndref->name))
		{
		    s = (char *)malloc(40 + 80);
		    sprintf(
			s,
			"SndRef #%i",
			i
		    );
		}
		else
		{
		    s = STRDUP(sndref->name);
		}
		n = PUListAddItem(
		    pulist, 0, s, None, None
		);
		free(s);
		PUListItemSetData(
		    pulist, n, (void *)i, NULL
		);

		SndRefDelete(sndrefs_list, total, i);
	    }
	    free(sndrefs_list);
	}
	free(path);
}


/*
 *	Gets all the object's weapon values to the widgets.
 */
static void WepWGetValues(wepw_struct *wepw)
{
	int obj_num, total;
	char *path;
	clist_struct *clist;
	popup_list_struct *pulist;
	ocs_info_struct **ocs_info_list;
	xsw_object_struct *obj;
	uew_struct *uew;

	if(wepw == NULL)
	    return;

	uew = wepw->uew;
	obj_num = wepw->obj_num;
	obj = WepWGetObjectPtr(wepw);

	/* Get OCS names */
	pulist = &wepw->ocs_pulist;
	if(uew != NULL)
	{
	    const char *name = uew->legend.ocsn;
	    path = ISPATHABSOLUTE(name) ?
		STRDUP(name) :
		STRDUP(PrefixPaths(dname.etc, name));
	}
	else
	{
	    path = NULL;
	}
	PUListClear(pulist);
	ocs_info_list = NULL;
	total = 0;
	if(!OCSInfoFileOpen(
	    &ocs_info_list, &total, path,
	    NULL, NULL, NULL
	))
	{
	    int i, n;
	    ocs_info_struct *ocs_info;
	    for(i = 0; i < total; i++)
	    {
		ocs_info = ocs_info_list[i];
		if(ocs_info == NULL)
		    continue;

		n = PUListAddItem(
		    pulist, 0, ocs_info->name, None, None
		);
		PUListItemSetData(
		    pulist, n, (void *)ocs_info->code, NULL
		);

		OCSInfoDelete(ocs_info);
	    }
	    free(ocs_info_list);
	}
	free(path);


	/* Update the weapons list */
	clist = &wepw->list;
	CListClear(clist);
	if(obj != NULL)
	{
	    int i, n, row_num;
	    char *s, *v[1];
	    popup_list_struct *pulist = &wepw->ocs_pulist;
	    popup_list_item_struct *pulist_item;
	    xsw_weapon_struct *wep;

	    /* Add weapons to the list */
	    for(i = 0; i < obj->total_weapons; i++)
	    {
		wep = obj->weapons[i];
		if(wep == NULL)
		    continue;

		row_num = CListRowAppend(clist);
		if(row_num < 0)
		    break;

		n = PUListGetItemByData(pulist, (void *)wep->ocs_code);
		pulist_item = PUListGetItem(pulist, n);
		if((pulist_item != NULL) ? !STRISEMPTY(pulist_item->text) : False)
		{
		    v[0] = STRDUP(pulist_item->text);
		}
		else
		{
		    v[0] = s = (char *)malloc(40 + 80);
		    sprintf(s, "Weapon #%i", i + 1);
		}
		CListSetRowTexts(
		    clist, row_num,
		    (const char **)v, sizeof(v) / sizeof(char *)
		);
		free(v[0]);
	    }
	}
	else
	{
	    /* No valid object */
	    wepw->obj_num = -1;
	}

	/* Reset prompts */
	PUListSetSelItem(&wepw->ocs_pulist, 0);
	PUListSetSelItem(&wepw->emission_type_pulist, 0);

	PromptSetS(&wepw->amount_prompt, "");
	PromptSetS(&wepw->max_prompt, "");
	PromptSetS(&wepw->damage_hull_prompt, "");
	PromptSetS(&wepw->damage_shields_prompt, "");
	PromptSetS(&wepw->range_prompt, "");
	PromptSetS(&wepw->create_power_prompt, "");
	PromptSetS(&wepw->delay_prompt, "");
	PUListSetSelItem(&wepw->fire_sndref_pulist, 0);
	WepWGetSndRefsToPUList(wepw, &wepw->fire_sndref_pulist);

	CBtnSetActive(&wepw->no_lock_cb, False);
	CBtnSetActive(&wepw->fire_cloaked_cb, False);
	CBtnSetActive(&wepw->no_fire_sound_cb, False);
	CBtnSetActive(&wepw->fixed_cb, False);
	CBtnSetActive(&wepw->penetrate_shield_cb, False);
	CBtnSetActive(&wepw->destroyable_cb, False);
	CBtnSetActive(&wepw->smoke_trail_cb, False);
	CBtnSetActive(&wepw->smoke_puff_fired_cb, False);
	CBtnSetActive(&wepw->shockwave_hit_cb, False);

	if(obj != NULL)
	{
	    const char *name = UNVGetObjectFormalName(obj_num, obj);
	    char *s = (char *)malloc(
		STRLEN(WEPW_TITLE) +
		STRLEN(name) +
		80
	    );
	    sprintf(
		s,  
		"%s: %s",
		WEPW_TITLE, name
	    );
	    OSWSetWindowTitle(wepw->toplevel, s);
	    free(s);
	}
}

/*
 *	Sets all the object's weapon values from the widgets.
 */
static void WepWSetValues(wepw_struct *wepw)
{
	clist_struct *clist;
	uew_struct *uew;

	if(wepw == NULL)
	    return;

	clist = &wepw->list;
	uew = wepw->uew;

	/* Apply changes to the current selected weapon */
	CListUnselectAllRows(clist);

	/* Update has changes marker on the Universe Editor */
	if((uew != NULL) && wepw->has_changes)
	{
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}
}

/*
 *	Gets the specified object's weapon values to the widgets.
 */
static void WepWGetWeaponValues(wepw_struct *wepw, const int wep_num)
{
	int i;
	popup_list_struct *pulist;
	xsw_weapon_struct *wep;

	if(wepw == NULL)
	    return;

	wep = WepWGetWeaponPtr(wepw, wep_num);
	if(wep == NULL)
	    return;

	/* OCS Code */
	pulist = &wepw->ocs_pulist;
	PUListSetSelItem(
	    pulist,
	    PUListGetItemByData(
		pulist, (void *)wep->ocs_code
	    )
	);

	/* Emission Type */
	switch(wep->emission_type)
	{
	  case SW_WEP_EMISSION_STREAM:
	    i = 0;
	    break;
	  case SW_WEP_EMISSION_PROJECTILE:
	    i = 1;
	    break;
	  case SW_WEP_EMISSION_PULSE:
	    i = 2;
	    break;
	  default:
	    i = 0;
	    break;
	}
	PUListSetSelItem(&wepw->emission_type_pulist, i);

	PromptSetI(&wepw->amount_prompt, wep->amount);
	PromptSetI(&wepw->max_prompt, wep->max);
	PromptSetF2(&wepw->damage_hull_prompt, wep->damage_hull, 2);
	PromptSetF2(&wepw->damage_shields_prompt, wep->damage_shields, 2);
 	PromptSetI(&wepw->range_prompt, wep->range);
	PromptSetF2(&wepw->create_power_prompt, wep->create_power, 2);
	PromptSetI(&wepw->delay_prompt, wep->delay);
	PromptSetI(&wepw->range_prompt, wep->range);
	pulist = &wepw->fire_sndref_pulist;
	PUListSetSelItem(
	    pulist,
	    PUListGetItemByData(
		pulist, (void *)wep->fire_sndref
	    )
	);

	/* Flags */
	CBtnSetActive(
	    &wepw->no_lock_cb,
	    (wep->flags & SW_WEP_FLAG_NO_LOCK) ? True : False
	);
	CBtnSetActive(
	    &wepw->fire_cloaked_cb,
	    (wep->flags & SW_WEP_FLAG_FIRE_CLOAKED) ? True : False
	);
	CBtnSetActive(
	    &wepw->no_fire_sound_cb,
	    (wep->flags & SW_WEP_FLAG_NO_FIRE_SOUND) ? True : False
	);
	CBtnSetActive(
	    &wepw->fixed_cb,
	    (wep->flags & SW_WEP_FLAG_FIXED) ? True : False
	);
	CBtnSetActive(
	    &wepw->penetrate_shield_cb,
	    (wep->flags & SW_WEP_FLAG_PENETRATE_SHIELD) ? True : False
	);
	CBtnSetActive(
	    &wepw->destroyable_cb,
	    (wep->flags & SW_WEP_FLAG_DESTROYABLE) ? True : False
	);
	CBtnSetActive(
	    &wepw->smoke_trail_cb,
	    (wep->flags & SW_WEP_FLAG_SMOKE_TRAIL) ? True : False
	);
	CBtnSetActive(
	    &wepw->smoke_puff_fired_cb,
	    (wep->flags & SW_WEP_FLAG_SMOKE_PUFF_FIRED) ? True : False
	);
	CBtnSetActive(
	    &wepw->shockwave_hit_cb,
	    (wep->flags & SW_WEP_FLAG_SHOCKWAVE_HIT) ? True : False
	);
}

/*
 *	Sets the specified object's weapon values from the widgets.
 */
static void WepWSetWeaponValues(wepw_struct *wepw, const int wep_num)
{
	int i;
	popup_list_struct *pulist;
	popup_list_item_struct *pulist_item;
	xsw_weapon_struct *wep = WepWGetWeaponPtr(wepw, wep_num);
	if(wep == NULL)
	    return;

	/* Get values from prompts and set to weapon */

	/* OCS Code */
	pulist = &wepw->ocs_pulist;
	i = PUListGetSelItem(pulist);
	pulist_item = PUListGetItem(pulist, i);
	if(pulist_item != NULL)
	{
	    const char *name = pulist_item->text;

	    wep->ocs_code = (int)PUListItemGetData(pulist, i);

	    if(!STRISEMPTY(name))
	    {
		char *v[1];

		v[0] = STRDUP(name);
		CListSetRowTexts(
		    &wepw->list, wep_num,
		    (const char **)v, sizeof(v) / sizeof(char *)
		);
		free(v[0]);
	    }
	}

	/* Emission Type */
	pulist = &wepw->emission_type_pulist;
	switch(PUListGetSelItem(pulist))
	{
	  case 0:
	    i = SW_WEP_EMISSION_STREAM;
	    break;
	  case 1:
	    i = SW_WEP_EMISSION_PROJECTILE;
	    break;
	  case 2:
	    i = SW_WEP_EMISSION_PULSE;
	    break;
	  default:
	    i = SW_WEP_EMISSION_STREAM;
	    break;
	}
	wep->emission_type = i;

	wep->amount = PromptGetI(&wepw->amount_prompt);
	wep->max = PromptGetI(&wepw->max_prompt);
	wep->damage_hull = PromptGetF(&wepw->damage_hull_prompt);
	wep->damage_shields = PromptGetF(&wepw->damage_shields_prompt);
	wep->range = PromptGetL(&wepw->range_prompt);
	wep->create_power = PromptGetF(&wepw->create_power_prompt);
	wep->delay = PromptGetL(&wepw->delay_prompt);
	pulist = &wepw->fire_sndref_pulist;
	wep->fire_sndref = (int)PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	);

	/* Flags */
	wep->flags = 0;
	if(CBtnGetActive(&wepw->no_lock_cb))
	    wep->flags |= SW_WEP_FLAG_NO_LOCK;
	if(CBtnGetActive(&wepw->fire_cloaked_cb))
	    wep->flags |= SW_WEP_FLAG_FIRE_CLOAKED;
	if(CBtnGetActive(&wepw->no_fire_sound_cb))
	    wep->flags |= SW_WEP_FLAG_NO_FIRE_SOUND;
	if(CBtnGetActive(&wepw->fixed_cb))
	    wep->flags |= SW_WEP_FLAG_FIXED;
	if(CBtnGetActive(&wepw->penetrate_shield_cb))
	    wep->flags |= SW_WEP_FLAG_PENETRATE_SHIELD;
	if(CBtnGetActive(&wepw->destroyable_cb))
	    wep->flags |= SW_WEP_FLAG_DESTROYABLE;
	if(CBtnGetActive(&wepw->smoke_trail_cb))
	    wep->flags |= SW_WEP_FLAG_SMOKE_TRAIL;
	if(CBtnGetActive(&wepw->smoke_puff_fired_cb))
	    wep->flags |= SW_WEP_FLAG_SMOKE_PUFF_FIRED;
	if(CBtnGetActive(&wepw->shockwave_hit_cb))
	    wep->flags |= SW_WEP_FLAG_SHOCKWAVE_HIT;
}


/*
 *	Draw callback.
 */
static void WepWDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	wepw_struct *wepw = WEPW(widget);        
	if(wepw == NULL)
	    return;

	if(wepw->map_state)
	    WepWDraw(wepw, amount);
}

/*
 *	Weapons list select callback.
 */
static void WepWCListSelectCB(
	clist_struct *clist, int row_num, void *data
)
{
	wepw_struct *wepw = WEPW(data);
	if((clist == NULL) || (row_num < 0) || (wepw == NULL))
	    return;

	WepWGetWeaponValues(wepw, row_num);
}

/*
 *	Weapons list unselect callback.
 */
static void WepWCListUnselectCB(
	clist_struct *clist, int row_num, void *data
)
{
	wepw_struct *wepw = WEPW(data);
	if((clist == NULL) || (row_num < 0) || (wepw == NULL))
	    return;

	WepWSetWeaponValues(wepw, row_num);
}

/*
 *	Add weapon callback.
 */
static void WepWWeaponAddCB(push_button_struct *btn, void *data)
{
	int row_num;
	clist_struct *clist;
	xsw_object_struct *obj;
	uew_struct *uew;
	wepw_struct *wepw = WEPW(data);
	if(wepw == NULL)
	    return;

	clist = &wepw->list;
	row_num = CListGetLastSelectedRow(clist);
	uew = wepw->uew;

	/* Apply values to current selected weapon */
	CListUnselectAllRows(clist);

	/* Add a new weapon on the object */
	obj = WepWGetObjectPtr(wepw);
	if(obj != NULL)
	{
	    if(obj->total_weapons >= SW_OBJ_WEAPONS_MAX)
	    {
		char *s = (char *)malloc(40 + (2 * 80));
		sprintf(
		    s,
"The maximum of %i weapons exists for this object,\n\
unable to create additional weapon.",
		    SW_OBJ_WEAPONS_MAX
		);
		UEDialog(
		    "Create Weapon Failed",
		    UE_ICON_WARNING,
		    s,
		    wepw->toplevel
		);
		free(s);
	    }
	    else
	    {
		/* Allocate a new weapon */
		xsw_weapon_struct *wep;
		int i = MAX(obj->total_weapons, 0);
		obj->total_weapons = i + 1;
		obj->weapons = (xsw_weapon_struct **)realloc(
		    obj->weapons,
		    obj->total_weapons * sizeof(xsw_weapon_struct *)
		);
		obj->weapons[i] = wep = (xsw_weapon_struct *)calloc(
		    1, sizeof(xsw_weapon_struct)
		);

		/* Update the weapons list */
		WepWGetValues(wepw);

		wepw->has_changes = True;
	    }
	}

	/* Update the has changes marker on the Universe Editor Window */
	if((uew != NULL) && wepw->has_changes)
	{
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}
}

/*
 *	Remove weapon callback.
 */
static void WepWWeaponRemoveCB(push_button_struct *btn, void *data)
{
	int row_num, wep_num;
	clist_struct *clist;
	xsw_object_struct *obj;
	uew_struct *uew;
	wepw_struct *wepw = WEPW(data);
	if(wepw == NULL)
	    return;

	clist = &wepw->list;
	row_num = CListGetLastSelectedRow(clist);
	uew = wepw->uew;

	/* Nothing selected? */
	if(clist->total_sel_rows <= 0)
	{
	    UEDialog(
		"Remove Weapon",
		UE_ICON_WARNING,
		"No weapon selected.",
		wepw->toplevel
	    );
	    return;
	}

	obj = WepWGetObjectPtr(wepw);
	if(obj == NULL)
	    return;

	/* Apply values to current selected weapon */
	CListUnselectAllRows(clist);

	wep_num = row_num;
	if((wep_num < 0) || (wep_num >= obj->total_weapons))
	    return;

	/* Delete the weapon on the object */
	if(obj->weapons != NULL)
	{
	    int i;

	    free(obj->weapons[wep_num]);

	    obj->total_weapons--;
	    for(i = wep_num; i < obj->total_weapons; i++)
		obj->weapons[i] = obj->weapons[i + 1];

	    if(obj->total_weapons > 0)
	    {
		obj->weapons = (xsw_weapon_struct **)realloc(
		    obj->weapons,
		    obj->total_weapons * sizeof(xsw_weapon_struct *)
		);
		if(obj->weapons == NULL)
		    obj->total_weapons = 0;
	    }
	    else
	    {
		free(obj->weapons);
		obj->weapons = NULL;
		obj->total_weapons = 0;
	    }

	    wepw->has_changes = True;
	}

	/* Update the weapons list */
	WepWGetValues(wepw);

	/* Update the has changes marker on the Universe Editor Window */
	if((uew != NULL) && wepw->has_changes)
	{
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}
}

/*
 *	Close button callback.
 */
static void WepWCloseCB(push_button_struct *btn, void *data)
{
	bool_t apply_to_all_selected_objects = False;
	clist_struct *clist;
	xsw_object_struct *obj;
	uew_struct *uew;
	wepw_struct *wepw = WEPW(data);
	if(wepw == NULL)
	    return;

	uew = wepw->uew;
	if(uew == NULL)
	    return;

	/* Set all the object's weapons from the widgets */
	WepWSetValues(wepw);

	/* Check if the universe editor has more than one object
	 * selected
	 */
	clist = &uew->objects_list;
	if(clist->total_sel_rows > 1)
	{
	    /* Query the user if the weapon values should be applied
	     * to all the selected objects
	     */
	    const int response = UEGetResponse(
		"Apply Weapons",
		UE_ICON_QUESTION,
"Apply weapon values to all the selected objects?",
		CDlgBtnYes | CDlgBtnNo,
		wepw->toplevel
	    );
	    if(response == CDlgYes)
		apply_to_all_selected_objects = True;
	}

	/* Apply weapons to all the selected objects on the Universe
	 * Editor?
	 */
	obj = WepWGetObjectPtr(wepw);
	if(apply_to_all_selected_objects && (obj != NULL))
	{
	    int i, n, row_num, wep_num;
	    clist_struct *clist = &uew->objects_list;
	    xsw_object_struct *obj2;
	    xsw_weapon_struct *wep, *wep2;

	    /* Iterate through the selected objects on the Universe
	     * Editor
	     */
	    n = clist->total_sel_rows;
	    for(i = 0; i < n; i++)
	    {
		row_num = clist->sel_row[i];
		obj2 = SW_OBJECT(CListRowGetData(clist, row_num));
		if((obj2 == NULL) || (obj2 == obj))
		    continue;

		/* Delete all weapons on the target object */
		for(wep_num = 0; wep_num < obj2->total_weapons; wep_num++)
		    free(obj2->weapons[wep_num]);

		/* Copy the weapons from the source to the target object */
		obj2->total_weapons = obj->total_weapons;
		if(obj2->total_weapons > 0)
		{
		    obj2->weapons = (xsw_weapon_struct **)realloc(
			obj2->weapons,
			obj2->total_weapons * sizeof(xsw_weapon_struct *)
		    );
		    for(wep_num = 0; wep_num < obj2->total_weapons; wep_num++)
		    {
			wep = obj->weapons[wep_num];
			if(wep == NULL)
			{
			    obj2->weapons[wep_num] = NULL;
			    continue;
			}

			obj2->weapons[wep_num] = wep2 = (xsw_weapon_struct *)malloc(
			    sizeof(xsw_weapon_struct)
			);
			if(wep2 != NULL)
			    memcpy(wep2, wep, sizeof(xsw_weapon_struct));
		    }
		}
		else
		{
		    free(obj2->weapons);
		    obj2->weapons = NULL;
		    obj2->total_weapons = 0;
		}

		if(!uew->has_changes)
		{
		    uew->has_changes = True;
		    UEWUpdateTitle(uew);
		}  
	    }
	}

	/* Unmap */
	WepWUnmap(wepw);

}

/*
 *	Creates a new Weapons Window.
 */
wepw_struct *WepWNew(uew_struct *uew, const int obj_num)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	pixel_t pix;
	font_t *font;
	win_t w, parent = osw_gui[0].root_win;
	push_button_struct *btn;
	clist_struct *clist;
	popup_list_struct *pulist;
	widget_global_struct *wglobal = &widget_global;
	wepw_struct *wepw = WEPW(calloc(1, sizeof(wepw_struct)));

	wepw->map_state = False;
	wepw->has_focus = False;
	wepw->freeze_count = 0;
	wepw->x = 0;
	wepw->y = 0;
	wepw->width = WEPW_WIDTH;
	wepw->height = WEPW_HEIGHT;
	wepw->uew = uew;
	wepw->obj_num = obj_num;
	wepw->has_changes = False;

	/* Create toplevel */
	x = wepw->x;
	y = wepw->y;
	width = wepw->width;
	height = wepw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	wepw->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    WEPW_TITLE,
	    WEPW_ICON_TITLE,
	    (ue_image.icon_unvedit_pixmap != None) ?
		ue_image.icon_unvedit_pixmap :
		wglobal->std_icon_pm,
	    True,			/* Let WM set coordinates? */
	    x, y,
	    width, height,
	    width, height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	if(uew != NULL)
	    OSWSetTransientFor(w, uew->toplevel);

	parent = w;

	/* Weapons list */
	clist = &wepw->list;
	width = (unsigned int)MAX(
	    (int)wepw->width - WEPW_BUTTON_WIDTH -
	    (2 * WEPW_MARGIN) - 5,
	    100
	);
	height = 100;
	x = WEPW_MARGIN;
	y = WEPW_MARGIN;
	CListInit(
	    clist, parent,
	    x, y, width, height
	);
	clist->select_cb = WepWCListSelectCB;
	clist->select_data = wepw;
	clist->unselect_cb = WepWCListUnselectCB;
	clist->unselect_data = wepw;

	font = wglobal->std_font;
	pix = wglobal->normal_text_pix;
	CListColumnAppend(
	    clist,
	    "Weapon",
	    font, pix,
	    0,
	    0
	);

	/* Add weapon button */
	btn = &wepw->weapon_add_btn;
	width = WEPW_BUTTON_WIDTH;
	height = WEPW_BUTTON_HEIGHT;
	x = (int)wepw->width - (int)width - WEPW_MARGIN;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Add",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    WepWWeaponAddCB, wepw
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Remove weapon button */
	btn = &wepw->weapon_remove_btn;
	y += (int)height + (WEPW_MARGIN / 2);
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Remove",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    WepWWeaponRemoveCB, wepw
	);
	accel_keys[0] = OSW_KEY_r;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);


	/* OCS Code */
	pulist = &wepw->ocs_pulist;
	x = WEPW_MARGIN;
	y = clist->y + clist->height + (WEPW_MARGIN / 2);
	PUListInit(
	    pulist, parent,
	    x, y,
	    (unsigned int)MAX(
		(int)wepw->width - (2 * WEPW_MARGIN),
		1
	    ),
	    0,
	    7,				/* Items visible */
	    PULIST_POPUP_BELOW
	);

	/* Emission Type */
	y += (int)pulist->height + (WEPW_MARGIN / 2);
	pulist = &wepw->emission_type_pulist;
	PUListInit(
	    pulist, parent,
	    x, y,
	    (unsigned int)MAX(
		(int)wepw->width - (2 * WEPW_MARGIN),
		1 
	    ),
	    0,
	    3,				/* Items visible */
	    PULIST_POPUP_BELOW
	);
	PUListAddItem(
	    pulist,
	    0,
	    SW_WEP_EMISSION_STR_STREAM,
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    SW_WEP_EMISSION_STR_PROJECTILE,
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    SW_WEP_EMISSION_STR_PULSE,
	    None, None
	);


	/* Amount */
	width = (unsigned int)MAX(
	    ((int)wepw->width / 2) - (int)(1.5 * WEPW_MARGIN),
	    1
	);
	height = WEPW_PROMPT_HEIGHT;
	x = WEPW_MARGIN;
	y += (int)pulist->height + (WEPW_MARGIN / 2);
	PromptInit(
	    &wepw->amount_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Amount:",
	    80, 0
	);
	PromptSetTooltip(
	    &wepw->amount_prompt,
"Enter the quantity of this weapon (projectile weapons only)"
	);

	/* Max */
	x += (int)width + WEPW_MARGIN;
	PromptInit(
	    &wepw->max_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Max:",
	    80, 0
	);
	PromptSetTooltip(
	    &wepw->max_prompt,
"Enter the maximum quantity of this weapon (projectile weapons only)"
	);

	/* Damage Hull */
	x = WEPW_MARGIN;
	y += (int)height;
	PromptInit(
	    &wepw->damage_hull_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Damage Hull:",
	    80, 0
	);
	PromptSetTooltip(
	    &wepw->damage_hull_prompt,
"Enter the amount of damage that this weapon causes to the structure\
 of the object that it hits"
	);

	/* Damage Shields */
	x += (int)width + WEPW_MARGIN;
	PromptInit(
	    &wepw->damage_shields_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Damage Shield:",
	    80, 0    
	);
	PromptSetTooltip(
	    &wepw->damage_shields_prompt,
"Enter the amount of damage that this weapon causes to the shields\
 of the object that it hits"
	);

	/* Create Power */
	x = WEPW_MARGIN;
	y += (int)height;
	PromptInit(
	    &wepw->create_power_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Create Power:",
	    80, 0
	);
	PromptSetTooltip(
	    &wepw->create_power_prompt,
"Enter the minimum amount of power that is needed to fire this weapon"
	);

	/* Range */
	x += (int)width + WEPW_MARGIN;
	PromptInit(
	    &wepw->range_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Range (su):",
	    80, 0
	);
	PromptSetTooltip(
	    &wepw->range_prompt,
"Enter the range of this weapon in screen units. For stream\
 weapons, this is the length of the beam"
	);

	/* Delay */
	x = WEPW_MARGIN;
	y += (int)height;
	PromptInit(
	    &wepw->delay_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Delay (ms):",
	    80, 0
	);
	PromptSetTooltip(
	    &wepw->delay_prompt,
"Enter the minimum amount of time in milliseconds that must pass\
 between each use of this weapon"
	);


	/* Link prompts togeather */
	wepw->amount_prompt.prev = &wepw->delay_prompt;
	wepw->amount_prompt.next = &wepw->max_prompt;

	wepw->max_prompt.prev = &wepw->amount_prompt;
	wepw->max_prompt.next = &wepw->damage_hull_prompt;

	wepw->damage_hull_prompt.prev = &wepw->max_prompt;
	wepw->damage_hull_prompt.next = &wepw->damage_shields_prompt;

	wepw->damage_shields_prompt.prev = &wepw->damage_hull_prompt;
	wepw->damage_shields_prompt.next = &wepw->create_power_prompt;

	wepw->create_power_prompt.prev = &wepw->damage_shields_prompt;
	wepw->create_power_prompt.next = &wepw->range_prompt;

	wepw->range_prompt.prev = &wepw->create_power_prompt;
	wepw->range_prompt.next = &wepw->delay_prompt;

	wepw->delay_prompt.prev = &wepw->range_prompt;
	wepw->delay_prompt.next = &wepw->amount_prompt;


	/* Fire SndRef */
	x = WEPW_MARGIN + 6;
	y += (int)height;
	width = 0;
	height = WEPW_PULIST_HEIGHT;
	LabelInit(
	    &wepw->fire_sndref_label, parent,
	    x, y,
	    width, height,
	    "Fire SndRef:"
	);
	x += (int)wepw->fire_sndref_label.width;
	pulist = &wepw->fire_sndref_pulist;
	width = (unsigned int)MAX(
	    (int)wepw->width - x - WEPW_MARGIN,
	    1
	);
	PUListInit(
	    pulist, parent,
	    x, y + (((int)height - WEPW_PULIST_HEIGHT) / 2),
	    width, height,
	    7,
	    PULIST_POPUP_BELOW
	);
	height = pulist->height;


	/* No Lock */
	x = WEPW_MARGIN;
	y += (int)height + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->no_lock_cb, parent,
	    x, y, False, "No Lock"
	);
	CBtnSetTooltip(
	    &wepw->no_lock_cb,
"Check this to specify that this weapon is not to lock on to the\
 target when fired (useful for pulse weapons, mines, buoys, etc)"
	);
	height = wepw->no_lock_cb.height;

	/* Fire Cloaked */
	x = ((int)wepw->width / 2) + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->fire_cloaked_cb, parent,
	    x, y, False, "Fire Cloaked"
	);
	CBtnSetTooltip(
	    &wepw->fire_cloaked_cb,
"Check this to allow the firing of this weapon when cloaked"
	);

	/* No Fire Sound */
	x = WEPW_MARGIN;
	y += (int)height + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->no_fire_sound_cb, parent,
	    x, y, False, "No Fire Sound"   
	);
	CBtnSetTooltip(
	    &wepw->no_fire_sound_cb,
"Check this to specify that no SndRef is to be played when this\
 weapon is fired"
	);

	/* Fixed */
	x = ((int)wepw->width / 2) + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->fixed_cb, parent,
	    x, y, False, "Fixed"   
	);
	CBtnSetTooltip(
	    &wepw->fixed_cb,
"Check this to specify that this weapon may not be aimed at any\
 other angle other than the vessel's heading"
	);

	/* Penetrate Shield */
	x = WEPW_MARGIN;
	y += (int)height + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->penetrate_shield_cb, parent,
	    x, y, False, "Penetrate Shield"
	);
	CBtnSetTooltip(
	    &wepw->penetrate_shield_cb,
"Check this to specify that this weapon will penetrate the shields\
 of any object regardless of if the two frequencies match or not"
	);

	/* Destroyable */
	x = ((int)wepw->width / 2) + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->destroyable_cb, parent,
	    x, y, False, "Destroyable"
	);
	CBtnSetTooltip(
	    &wepw->destroyable_cb,
"Check this to specify that this weapon can be destroyed by\
 another weapon"
	);

	/* Smoke Trail */
	x = WEPW_MARGIN;
	y += (int)height + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->smoke_trail_cb, parent,
	    x, y, False, "Smoke Trail"
	);
	CBtnSetTooltip(
	    &wepw->smoke_trail_cb,
"Check this to specify that this weapon produces a smoke trail as\
 it fires"
	);

	/* Smoke Puff Fired */
	x = ((int)wepw->width / 2) + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->smoke_puff_fired_cb, parent,
	    x, y, False, "Smoke Puff Fired"
	);
	CBtnSetTooltip(
	    &wepw->smoke_puff_fired_cb,
"Check this to produce a smoke puff at the point where this\
 weapon is fired"
	);

	/* Shockwave Hit */
	x = WEPW_MARGIN;
	y += (int)height + (WEPW_MARGIN / 2);
	CBtnInit(
	    &wepw->shockwave_hit_cb, parent,
	    x, y, False, "Shockwave Hit"
	);
	CBtnSetTooltip(
	    &wepw->shockwave_hit_cb,
"Check this to specify that this weapon produces a shockwave\
 when it hits something"
	);


	/* Close button */
	btn = &wepw->close_btn;
	width = WEPW_BUTTON_WIDTH;
	height = WEPW_BUTTON_HEIGHT;
	x = (int)wepw->width - WEPW_MARGIN - (int)width;
	y = (int)wepw->height - WEPW_MARGIN - (int)height;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Close",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    WepWCloseCB, wepw
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL; 
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	return(wepw);
}

/*
 *	Draws the Weapons Window.
 */
void WepWDraw(wepw_struct *wepw, const int amount)
{
	unsigned int width, height;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if(wepw == NULL)
	    return;

	if(wepw->toplevel != None)
	{
	    w = wepw->toplevel;
	    width = wepw->width;
	    height = wepw->height;

	    /* Redraw the background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(osw_gui[0].black_pix);
		OSWDrawSolidRectangle(
		    w, 0, 0, width, height
		);
	    }
	    else
	    {
		int y;

		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);

		OSWSetForeground(wglobal->surface_shadow_pix);
		y = (int)height - 25 - WEPW_BUTTON_HEIGHT;
		OSWDrawLine(
		    w,
		    0, y,
		    (int)width, y
		);
		OSWSetForeground(wglobal->surface_highlight_pix);
		y = (int)height - 24 - WEPW_BUTTON_HEIGHT;
		OSWDrawLine(
		    w,
		    0, y,
		    (int)width, y
		);
	    }
	}
}

/*
 *	Queues the Weapon Window to draw.
 */
void WepWQueueDraw(wepw_struct *wepw, const int amount) 
{
	if(wepw == NULL)
	    return;

	WidgetQueueDrawAction(
	    WepWDrawCB, wepw, (void *)amount
	);
}

/*
 *	Manages the Weapons Window event.
 */
int WepWManage(wepw_struct *wepw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;

	if((wepw == NULL) || (event == NULL))
	    return(events_handled);

	if(!wepw->map_state &&
	   (event->type != MapNotify)
	)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case Expose:
	    if(w == wepw->toplevel)
	    {
		WepWQueueDraw(wepw, 0);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!wepw->has_focus)
		return(events_handled);

	    break;

	  case KeyRelease:
	    if(!wepw->has_focus) 
		return(events_handled);

	    break;

	  case FocusIn:
	    if(w == wepw->toplevel)
	    {
		wepw->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == wepw->toplevel)
	    {
		wepw->has_focus = False;
		events_handled++;
	    }   
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(wepw->toplevel, event, &t))
	    {
		OSWGrabFocusTime(wepw->toplevel, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */  
	    else if(OSWEventIsDeleteWindow(wepw->toplevel, event))
	    {
		WepWCloseCB(&wepw->close_btn, wepw);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	/* Manage widgets */

	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->list, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->weapon_add_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->weapon_remove_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->close_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->ocs_pulist, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->emission_type_pulist, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->amount_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->max_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->damage_hull_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->damage_shields_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->range_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->create_power_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->delay_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->fire_sndref_label, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->fire_sndref_pulist, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->no_lock_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->fire_cloaked_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->no_fire_sound_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->fixed_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->penetrate_shield_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->destroyable_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->smoke_trail_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->smoke_puff_fired_cb, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&wepw->shockwave_hit_cb, event);

	return(events_handled);
}

/*
 *	Maps the Weapons Window.
 */
void WepWMap(wepw_struct *wepw)
{
	if(wepw == NULL)
	    return;

	OSWMapRaised(wepw->toplevel);
	wepw->map_state = True;

	WidgetMap(&wepw->list);
	WidgetMap(&wepw->weapon_add_btn);
	WidgetMap(&wepw->weapon_remove_btn); 

	WidgetMap(&wepw->ocs_pulist);
	WidgetMap(&wepw->emission_type_pulist);

	WidgetMap(&wepw->amount_prompt);
	WidgetMap(&wepw->max_prompt);
	WidgetMap(&wepw->damage_hull_prompt);
	WidgetMap(&wepw->damage_shields_prompt);
	WidgetMap(&wepw->create_power_prompt);
	WidgetMap(&wepw->range_prompt);
	WidgetMap(&wepw->delay_prompt);
	WidgetMap(&wepw->fire_sndref_label);
	WidgetMap(&wepw->fire_sndref_pulist);

	WidgetMap(&wepw->no_lock_cb);
	WidgetMap(&wepw->fire_cloaked_cb);
	WidgetMap(&wepw->no_fire_sound_cb);
	WidgetMap(&wepw->fixed_cb);
	WidgetMap(&wepw->penetrate_shield_cb);
	WidgetMap(&wepw->destroyable_cb);
	WidgetMap(&wepw->smoke_trail_cb);
	WidgetMap(&wepw->smoke_puff_fired_cb);
	WidgetMap(&wepw->shockwave_hit_cb);

	WidgetMap(&wepw->close_btn);

	OSWGrabFocus(wepw->toplevel);

	UEWDoUpdateWindowMenus();
}

/*
 *	Maps the Weapons Window gets the values from the specified
 *	object.
 */
void WepWMapValues(wepw_struct *wepw)
{
	if(wepw == NULL)
	    return;

	/* Get values from the referenced object */
	WepWGetValues(wepw);

	/* Select the first weapon */
	CListSelectRow(&wepw->list, 0);

	/* Reset the has changes marker */
	wepw->has_changes = False;

	/* Map the Weapons Window */
	WepWMap(wepw);
}

/*
 *	Unmaps the Weapons Window.
 */
void WepWUnmap(wepw_struct *wepw)
{
	if(wepw == NULL)
	    return;

	WidgetUnmap(&wepw->list);
	WidgetUnmap(&wepw->weapon_add_btn);
	WidgetUnmap(&wepw->weapon_remove_btn);

	WidgetUnmap(&wepw->ocs_pulist);
	WidgetUnmap(&wepw->emission_type_pulist);

	WidgetUnmap(&wepw->amount_prompt);
	WidgetUnmap(&wepw->max_prompt);
	WidgetUnmap(&wepw->damage_hull_prompt);
	WidgetUnmap(&wepw->damage_shields_prompt);
	WidgetUnmap(&wepw->range_prompt);
	WidgetUnmap(&wepw->create_power_prompt);
	WidgetUnmap(&wepw->delay_prompt);
	WidgetUnmap(&wepw->fire_sndref_label);
	WidgetUnmap(&wepw->fire_sndref_pulist);

	WidgetUnmap(&wepw->no_lock_cb);
	WidgetUnmap(&wepw->fire_cloaked_cb);
	WidgetUnmap(&wepw->no_fire_sound_cb);
	WidgetUnmap(&wepw->fixed_cb);
	WidgetUnmap(&wepw->penetrate_shield_cb);
	WidgetUnmap(&wepw->destroyable_cb);
	WidgetUnmap(&wepw->smoke_trail_cb);
	WidgetUnmap(&wepw->smoke_puff_fired_cb);
	WidgetUnmap(&wepw->shockwave_hit_cb);
 
	WidgetUnmap(&wepw->close_btn);

	OSWUnmapWindow(wepw->toplevel);
	wepw->map_state = False;

	UEWDoUpdateWindowMenus();
}

/*
 *	Deletes the Weapons Window.
 */
void WepWDelete(wepw_struct *wepw)
{
	if(wepw == NULL)
	    return;

	CListClear(&wepw->list);

	WidgetDestroy(&wepw->close_btn);

	WidgetDestroy(&wepw->ocs_pulist);
	WidgetDestroy(&wepw->emission_type_pulist);

	WidgetDestroy(&wepw->amount_prompt);
	WidgetDestroy(&wepw->max_prompt);
	WidgetDestroy(&wepw->damage_hull_prompt);
	WidgetDestroy(&wepw->damage_shields_prompt);
	WidgetDestroy(&wepw->range_prompt);
	WidgetDestroy(&wepw->create_power_prompt);
	WidgetDestroy(&wepw->delay_prompt);
	WidgetDestroy(&wepw->fire_sndref_label);
	WidgetDestroy(&wepw->fire_sndref_pulist);

	WidgetDestroy(&wepw->no_lock_cb);
	WidgetDestroy(&wepw->fire_cloaked_cb);
	WidgetDestroy(&wepw->no_fire_sound_cb);
	WidgetDestroy(&wepw->fixed_cb);
	WidgetDestroy(&wepw->penetrate_shield_cb);
	WidgetDestroy(&wepw->destroyable_cb);
	WidgetDestroy(&wepw->smoke_trail_cb);
	WidgetDestroy(&wepw->smoke_puff_fired_cb);
	WidgetDestroy(&wepw->shockwave_hit_cb);

	WidgetDestroy(&wepw->weapon_add_btn);
	WidgetDestroy(&wepw->weapon_remove_btn);
	WidgetDestroy(&wepw->list);

	OSWDestroyWindow(&wepw->toplevel);

	free(wepw);
}
