#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/cfgfmt.h"
#include "../include/fio.h"
#include "../include/reality.h"
#include "../include/objects.h"

#include "../include/unvmain.h"
#include "../include/unvutil.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvfile.h"


xsw_object_struct **UNVFileOpen(
	const char *path,
	int *total,
	sw_legend_struct *legend,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data
);
int UNVFileSave(
	const char *path, const mode_t mode,
	xsw_object_struct **list, int total,
	sw_legend_struct *legend,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Loads a list of objects from the specified file.
 */
xsw_object_struct **UNVFileOpen(
	const char *path,
	int *total,
	sw_legend_struct *legend,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data
)
{
	int i, prev;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	unsigned long file_size;
	FILE *fp;
	struct stat stat_buf;
	int	lines_read = 0,
		obj_num = -1,
		weapons_count = 0,
		product_num = 0;

	xsw_ecodata_struct *eco;
	xsw_object_struct *obj, **list = NULL;

	if(total == NULL)
	    return(list);

	*total = 0;

	if(STRISEMPTY(path))
	    return(list);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(
		    STRLEN(path) + STRLEN(error_msg) + 80
		);
		sprintf(
		    s,
"%s: %s.",
		    path, error_msg
		);
		fprintf(stderr, "%s\n", s);
		free(s);
	    }
	    free(error_msg);
	    return(list);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;
	}
	else
	{
	    file_size = 0l;
	}


	/* Reset the legend */
	UNVLegendReset(legend);

	/* Call notify progress function */
	if(progress_cb != NULL)
	    progress_cb(
		progress_data,
		(unsigned long)ftell(fp), file_size
	    );


	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';

	    /* Report progress */
	    if(progress_cb != NULL)
		progress_cb(
		    progress_data,
		    (unsigned long)ftell(fp), file_size
		);

	    /* BeginHeader */
	    if(!strcasecmp(parm, "BeginHeader"))
	    {
		while(1)
		{
		    /* Get the next parameter and value */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';


		    /* Title */
		    if(!strcasecmp(parm, "Title"))
		    {
			if(legend != NULL)
			{
			    strncpy(
				legend->title,
				val,
				sizeof(legend->title)
			    );
			    legend->title[sizeof(legend->title) - 1] = '\0';
			}
		    }
		    /* Version */
		    else if(!strcasecmp(parm, "Version"))
		    {
			if(legend != NULL)
			{
			    strncpy(
				legend->version,
				val,
				sizeof(legend->version)
			    );
			    legend->version[sizeof(legend->version) - 1] = '\0';
			}
		    }
		    /* OCSNFile */
		    else if(!strcasecmp(parm, "OCSNFile"))
		    {
			if(legend != NULL)
			{   
			    strncpy(
				legend->ocsn, val, 
				sizeof(legend->ocsn)  
			    );
			    legend->ocsn[
				sizeof(legend->ocsn) - 1] = '\0';
			}   
		    }    
		    /* ImgRefsFile */
		    else if(!strcasecmp(parm, "ImgRefsFile") ||
			    !strcasecmp(parm, "ImageSetFile")
		    )
		    {
			if(legend != NULL)
			{
			    strncpy(
				legend->imgrefs, val,
				sizeof(legend->imgrefs)
			    );
			    legend->imgrefs[
				sizeof(legend->imgrefs) - 1] = '\0';
			}
		    }
		    /* SndRefsFile */
		    else if(!strcasecmp(parm, "SndRefsFile") ||
			    !strcasecmp(parm, "SoundSchemeFile")
		    )
		    {
			if(legend != NULL)
			{
			    strncpy(
				legend->sndrefs, val,
				sizeof(legend->sndrefs)
			    );
			    legend->sndrefs[
				sizeof(legend->sndrefs) - 1] = '\0';
			}
		    }

		    /* RUToAU */
		    else if(!strcasecmp(parm, "RUToAU") ||
			    !strcasecmp(parm, "UnitConversionRUTOAU")
		    )
		    {
			if(legend != NULL)
			    legend->ru_to_au = MAX(atof(val), 0.0f);
		    }

		    /* LostFoundOwner */
		    else if(!strcasecmp(parm, "LostFoundOwner"))
		    {
			if(legend != NULL)
			    legend->lost_found_owner = atoi(val);
		    }

		    /* PlayerStartSectX */
		    else if(!strcasecmp(parm, "PlayerStartSectX"))
		    {
			if(legend != NULL)
			    legend->player_start_sect_x = atol(val);
		    }
		    /* PlayerStartSectY */
		    else if(!strcasecmp(parm, "PlayerStartSectY"))
		    {
			if(legend != NULL)
			    legend->player_start_sect_y = atol(val);
		    }
		    /* PlayerStartSectZ */
		    else if(!strcasecmp(parm, "PlayerStartSectZ"))
		    {
			if(legend != NULL)
			    legend->player_start_sect_z = atol(val);
		    }
		    /* PlayerStartX */
		    else if(!strcasecmp(parm, "PlayerStartX"))
		    {
			if(legend != NULL)
			    legend->player_start_x = atof(val);
		    }
		    /* PlayerStartY */
		    else if(!strcasecmp(parm, "PlayerStartY"))
		    {
			if(legend != NULL)
			    legend->player_start_y = atof(val);
		    }
		    /* PlayerStartZ */
		    else if(!strcasecmp(parm, "PlayerStartZ"))
		    {
			if(legend != NULL)
			    legend->player_start_z = atof(val);
		    }

		    /* PlayerStartHeading */
		    else if(!strcasecmp(parm, "PlayerStartHeading"))
		    {
			if(legend != NULL)
			    legend->player_start_heading = atof(val);
		    }
		    /* PlayerStartPitch */
		    else if(!strcasecmp(parm, "PlayerStartPitch"))
		    {
			if(legend != NULL)
			    legend->player_start_pitch = atof(val);
		    }
		    /* PlayerStartBank */
		    else if(!strcasecmp(parm, "PlayerStartBank"))
		    {
			if(legend != NULL)
			    legend->player_start_bank = atof(val);
		    }

		    /* GuestStartSectX */
		    else if(!strcasecmp(parm, "GuestStartSectX"))
		    {
			if(legend != NULL)
			    legend->guest_start_sect_x = atol(val);
		    }
		    /* GuestStartSectY */
		    else if(!strcasecmp(parm, "GuestStartSectY"))
		    {
			if(legend != NULL)
			    legend->guest_start_sect_y = atol(val);
		    }
		    /* GuestStartSectZ */
		    else if(!strcasecmp(parm, "GuestStartSectZ"))
		    {
			if(legend != NULL)
			    legend->guest_start_sect_z = atol(val);
		    }
		    /* GuestStartX */
		    else if(!strcasecmp(parm, "GuestStartX"))
		    {
			if(legend != NULL)
			    legend->guest_start_x = atof(val);
		    }
		    /* GuestStartY */
		    else if(!strcasecmp(parm, "GuestStartY"))
		    {
			if(legend != NULL)
			    legend->guest_start_y = atof(val);
		    }
		    /* GuestStartZ */
		    else if(!strcasecmp(parm, "GuestStartZ"))
		    {
			if(legend != NULL)
			    legend->guest_start_z = atof(val);
		    }

		    /* GuestStartHeading */
		    else if(!strcasecmp(parm, "GuestStartHeading"))
		    {
			if(legend != NULL)
			    legend->guest_start_heading = atof(val);
		    }
		    /* GuestStartPitch */
		    else if(!strcasecmp(parm, "GuestStartPitch"))
		    {
			if(legend != NULL)
			    legend->guest_start_pitch = atof(val);
		    }
		    /* GuestStartBank */
		    else if(!strcasecmp(parm, "GuestStartBank"))
		    {
			if(legend != NULL)
			    legend->guest_start_bank = atof(val);
		    }

		    /* AntimatterLowWarning */
		    else if(!strcasecmp(parm, "AntimatterLowWarning"))
		    {
			if(legend != NULL)
			    legend->antimatter_low_warning = MAX(
				atof(val), 0.0f
			    );
		    }
		    /* AsteroidFieldDamageInt */
		    else if(!strcasecmp(parm, "AsteroidFieldDamageInt"))
		    {
			if(legend != NULL)
			    legend->asteroid_field_damage_int = MAX(
				atol(val), 1000l
			    );
		    }
		    /* AsteroidFieldDamageRate */
		    else if(!strcasecmp(parm, "AsteroidFieldDamageRate"))
		    {
			if(legend != NULL)
			    legend->asteroid_field_damage_rate = MAX(
				atof(val), 0.0f
			    );
		    }
		    /* AutoDestructDamageRange */
		    else if(!strcasecmp(parm, "AutoDestructDamageRange"))
		    {
			if(legend != NULL)
			    legend->auto_destruct_damage_range = MAX(
				atof(val), 0.0f
			    );
		    }
		    /* CloakAntimatterRate */
		    else if(!strcasecmp(parm, "CloakAntimatterRate"))
		    {
			if(legend != NULL)
			    legend->cloak_antimatter_rate = MAX(
				atof(val), 0.0f
			    );
		    }
		    /* ComRange */
		    else if(!strcasecmp(parm, "ComRange"))
		    {
			if(legend != NULL)
			    legend->com_range = MAX(
				atof(val), 0.0f
			    );
		    }
		    /* DamageControlRate */
		    else if(!strcasecmp(parm, "DamageControlRate"))
		    {
			if(legend != NULL)
			    legend->damage_control_rate = MAX(             
				atof(val), 0.0f
			    );
		    }
		    /* DisarmedWeaponLifespan */
		    else if(!strcasecmp(parm, "DisarmedWeaponLifespan"))
		    {
			if(legend != NULL)
			    legend->disarmed_weapon_lifespan = (time_t)atol(val);
		    }
		    /* EngineAntimatterRate */
		    else if(!strcasecmp(parm, "EngineAntimatterRate"))
		    {
			if(legend != NULL)
			    legend->engine_antimatter_rate = MAX(             
				atof(val), 0.0f
			    );
		    }
		    /* NoAntimatterThrustCoeff */
		    else if(!strcasecmp(parm, "NoAntimatterThrustCoeff"))
		    {
			if(legend != NULL)
			    legend->no_antimatter_thrust_coeff = MAX(
				atof(val), 0.0f
			    );      
		    }
		    /* OnFireHPCoeff */
		    else if(!strcasecmp(parm, "OnFireHPCoeff"))
		    {
			if(legend != NULL)
			    legend->on_fire_hp_coeff = CLIP(
				atof(val), 0.0f, 1.0f
			    );
		    }
		    /* TractorBeamAntimatterRate */
		    else if(!strcasecmp(parm, "TractorBeamAntimatterRate"))
		    {
			if(legend != NULL)
			    legend->tractor_beam_antimatter_rate = MAX(
				atof(val), 0.0f
			    );
		    }
		    /* TractorBeamRange */
		    else if(!strcasecmp(parm, "TractorBeamRange"))
		    {
			if(legend != NULL)
			    legend->tractor_beam_range = MAX(             
				atof(val), 0.0f
			    );
		    }
		    /* TransporterRange */
		    else if(!strcasecmp(parm, "TransporterRange"))
		    {
			if(legend != NULL)
			    legend->transporter_range = MAX(
				atof(val), 0.0f
			    );
		    }

		    /* EndHeader */
		    else if(!strcasecmp(parm, "EndHeader"))
		    {
			break;
		    }
		}
		continue;
	    }
	    /* ***************************************************** */
	    /* BeginXSWObject */
	    else if(!strcasecmp(parm, "BeginXSWObject") ||
	            !strcasecmp(parm, "BeginObject")
	    )
	    {
		/* Get object number */
		obj_num = atol(val);
		if((obj_num < 0) || (obj_num >= SW_OBJECTS_MAX))
		{
		    fprintf(
			stderr,
"%s: Line %i: Object number `%i' out of range [%i, %i].\n",
			path, lines_read,
			obj_num,
			0, SW_OBJECTS_MAX - 1
		    );
		    continue;
		}

		/* Allocate a new object in memory */

		/* Need to allocate more object pointers? */
		if(obj_num >= *total)
		{
		    prev = *total;
		    *total = obj_num + 1;

		    list = (xsw_object_struct **)realloc(
			list,
			*total * sizeof(xsw_object_struct *)
		    );
		    if(list == NULL)
		    {
			*total = 0;
			continue;
		    }

		    /* Allocate and reset each new object */
		    for(i = prev; i < *total; i++)
		    {
			list[i] = (xsw_object_struct *)calloc(
			    1, sizeof(xsw_object_struct)
			);
			UNVResetObject(list[i]);
		    }
		}
		/* Allocate object structure as needed */
		else if(list[obj_num] == NULL)
		{
		    list[obj_num] = (xsw_object_struct *)calloc(
			1, sizeof(xsw_object_struct)
		    );
		    if(list[obj_num] == NULL)
			continue;
		}
		else
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: Redefining object #%i.\n",
			path, lines_read, obj_num
		    );
		}

		/* Get pointer to object */
		obj = list[obj_num];

		UNVResetObject(obj);

		/* Reset weapons count */
		weapons_count = 0;

		while(1)
		{
		    /* Free previous line and allocate/read next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';


		    /* Name */
		    if(!strcasecmp(parm, "Name"))
		    {
			strncpy(
			    obj->name,
			    val,
			    sizeof(obj->name)
			);
			obj->name[sizeof(obj->name) - 1] = '\0';
		    }
		    /* Password */
		    else if(!strcasecmp(parm, "Password"))
		    {
			if(*val != '\0')
			{
			    free(obj->password);
			    obj->password = STRDUP(val);
			}
		    }
		    /* Empire */
		    else if(!strcasecmp(parm, "Empire"))
		    {
			strncpy(
			    obj->empire,
			    val,
			    sizeof(obj->empire)
			);
			obj->empire[sizeof(obj->empire) - 1] = '\0';

			/* Make sure upper case */
			strtoupper(obj->empire);
		    }
		    /* Description */
		    else if(!strcasecmp(parm, "Description"))
		    {  
			free(obj->description);
			obj->description = STRDUP(val);
		    }
		    /* ELink */
		    else if(!strcasecmp(parm, "ELink"))
		    {
			free(obj->elink);
			obj->elink = STRDUP(val);
		    }
		    /* Type */  
		    else if(!strcasecmp(parm, "Type"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->type = atoi(val);
			}
			else
			{
			   if(!strcasecmp(val, SW_OBJ_TYPE_NAME_STATIC))
				obj->type = SW_OBJ_TYPE_STATIC;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_DYNAMIC))
				obj->type = SW_OBJ_TYPE_DYNAMIC;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_CONTROLLED))
				obj->type = SW_OBJ_TYPE_CONTROLLED;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_PLAYER))
				obj->type = SW_OBJ_TYPE_PLAYER;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_WEAPON))
				obj->type = SW_OBJ_TYPE_WEAPON;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_STREAMWEAPON))
				obj->type = SW_OBJ_TYPE_STREAMWEAPON;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_SPHEREWEAPON))
				obj->type = SW_OBJ_TYPE_SPHEREWEAPON;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_HOME))
				obj->type = SW_OBJ_TYPE_HOME;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_AREA))
				obj->type = SW_OBJ_TYPE_AREA;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_ANIMATED))
				obj->type = SW_OBJ_TYPE_ANIMATED;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_WORMHOLE))
				obj->type = SW_OBJ_TYPE_WORMHOLE;
			   else if(!strcasecmp(val, SW_OBJ_TYPE_NAME_ELINK))
				obj->type = SW_OBJ_TYPE_ELINK;
			}
			if(obj->type <= SW_OBJ_TYPE_GARBAGE)
			    obj->type = SW_OBJ_TYPE_STATIC;
		    }
		    /* LastConnected */
		    else if(!strcasecmp(parm, "LastConnected"))
		    {
			obj->last_connected = atol(val);
		    }
		    /* LocType */
		    else if(!strcasecmp(parm, "LocType"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->loc_type = atoi(val);
			}
			else
			{
			    if(!strcasecmp(val, SW_LOC_TYPE_STR_SPACE))
				obj->loc_type = SW_LOC_TYPE_SPACE;
			    else if(!strcasecmp(val, SW_LOC_TYPE_STR_SPACE_TERRITORY))
				obj->loc_type = SW_LOC_TYPE_SPACE_TERRITORY;
			    else if(!strcasecmp(val, SW_LOC_TYPE_STR_NEBULA))
				obj->loc_type = SW_LOC_TYPE_NEBULA;
			    else if(!strcasecmp(val, SW_LOC_TYPE_STR_ASTEROID_FIELD))
				obj->loc_type = SW_LOC_TYPE_ASTEROID_FIELD;
			    else if(!strcasecmp(val, SW_LOC_TYPE_STR_PULSAR))
				obj->loc_type = SW_LOC_TYPE_PULSAR;
			}
			if(obj->loc_type < SW_LOC_TYPE_SPACE)
			    obj->loc_type = SW_LOC_TYPE_SPACE;
		    }
		    /* LocObjNum */
		    else if(!strcasecmp(parm, "LocObjNum"))
		    {
			obj->loc_obj_num = atoi(val);
		    }
		    /* ImgRef */
		    else if(!strcasecmp(parm, "ImgRef") ||
			    !strcasecmp(parm, "ImageSet")
		    )
		    {
			obj->imgref = atoi(val);
		    }
		    /* Owner */
		    else if(!strcasecmp(parm, "Owner"))
		    {
			obj->owner = atoi(val);
			if(obj->owner >= SW_OBJECTS_MAX)
			    obj->owner = -1;
		    }
		    /* Size */
		    else if(!strcasecmp(parm, "Size"))
		    {
			obj->size = atol(val);
			if(obj->size < 1)
			    obj->size = 1;
		    }
		    /* LockedOn */
		    else if(!strcasecmp(parm, "LockedOn"))
		    {
			obj->locked_on = atoi(val);
			if(obj->locked_on >= SW_OBJECTS_MAX)
			    obj->locked_on = -1;
		    }
		    /* InterceptingObject */
		    else if(!strcasecmp(parm, "InterceptingObject"))
		    {
			obj->intercepting_object = atoi(val);
			if(obj->intercepting_object >=
			    SW_OBJECTS_MAX)
			    obj->intercepting_object = -1;
		    }
		    /* ScannerRange */
		    else if(!strcasecmp(parm, "ScannerRange"))
		    {
			obj->scanner_range =
			    atof(val);
			if(obj->scanner_range < 1)
			    obj->scanner_range = 1;
		    }
		    /* SectX */
		    else if(!strcasecmp(parm, "SectX"))
		    {
			obj->sect_x = atol(val);
		    }
		    /* SectY */
		    else if(!strcasecmp(parm, "SectY"))
		    {
			obj->sect_y = atol(val);
		    }
		    /* SectZ */
		    else if(!strcasecmp(parm, "SectZ"))
		    {
			obj->sect_z = atol(val);
		    }
		    /* CoordX */
		    else if(!strcasecmp(parm, "CoordX"))
		    {
			obj->x = atof(val);
		    }
		    /* CoordY */
		    else if(!strcasecmp(parm, "CoordY"))
		    {
			obj->y = atof(val);
		    }
		    /* CoordZ */
		    else if(!strcasecmp(parm, "CoordZ"))
		    {
			obj->z = atof(val);
		    }
		    /* ObjectHeading (in radians) */
		    else if(!strcasecmp(parm, "ObjectHeading"))
		    {
			obj->heading = atof(val);
			SWMathSetUnitVector2D(
			    &obj->attitude_vector_compoent,
			    obj->heading
			);
		    }
		    /* ObjectPitch (in radians) */
		    else if(!strcasecmp(parm, "ObjectPitch"))
		    {
			obj->pitch = atof(val);
/* Update attitude vector compoents? */
		    }
		    /* ObjectBank (in radians) */
		    else if(!strcasecmp(parm, "ObjectBank"))
		    {
			obj->bank = atof(val);
/* Update attitude vector compoents? */
		    }
		    /* VelocityHeading (in radians) */
		    else if(!strcasecmp(parm, "VelocityHeading"))
		    {
			obj->velocity_heading = atof(val);
			SWMathSetUnitVector2D(
			    &obj->momentum_vector_compoent,
			    obj->velocity_heading
			);
		    }
		    /* VelocityPitch (in radians) */
		    else if(!strcasecmp(parm, "VelocityPitch"))
		    {
			obj->velocity_pitch = atof(val);
/* Update momentum vector compoents? */
		    }
		    /* VelocityBank (in radians) */
		    else if(!strcasecmp(parm, "VelocityBank"))
		    {
			obj->velocity_bank = atof(val);
/* Update momentum vector compoents? */
		    }
		    /* VelocityMax */
		    else if(!strcasecmp(parm, "VelocityMax"))
		    {
			obj->velocity_max = atof(val);
		    }
		    /* Velocity */
		    else if(!strcasecmp(parm, "Velocity"))
		    {
			obj->velocity = atof(val);
		    }
		    /* ThrustDir */
		    else if(!strcasecmp(parm, "ThrustDir"))
		    {
			obj->thrust_dir = atof(val);
		    }
		    /* ThrustPower */
		    else if(!strcasecmp(parm, "ThrustPower"))
		    {
			obj->thrust_power = atof(val);
		    }
		    /* Thrust */
		    else if(!strcasecmp(parm, "Thrust"))
		    {
			obj->thrust = atof(val);
		    }
		    /* Throttle */
		    else if(!strcasecmp(parm, "Throttle"))
		    {
			obj->throttle = atof(val);
		    }
		    /* EngineState */
		    else if(!strcasecmp(parm, "EngineState"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->engine_state = atoi(val);
			}
			else   
			{   
			    if(!strcasecmp(val, SW_ENGINE_STATE_STR_NONE))
				obj->engine_state = SW_ENGINE_STATE_NONE; 
			    else if(!strcasecmp(val, SW_ENGINE_STATE_STR_OFF))
				obj->engine_state = SW_ENGINE_STATE_OFF;
			    else if(!strcasecmp(val, SW_ENGINE_STATE_STR_STARTING))
				obj->engine_state = SW_ENGINE_STATE_STARTING;
			    else if(!strcasecmp(val, SW_ENGINE_STATE_STR_ON))
				obj->engine_state = SW_ENGINE_STATE_ON;
			}
		    }
		    /* TurnRate */
		    else if(!strcasecmp(parm, "TurnRate"))
		    {
			obj->turnrate = atof(val);
		    }
		    /* OdoMeter */
		    else if(!strcasecmp(parm, "OdoMeter"))
		    {
			obj->odometer = MAX(atof(val), 0.0f);
		    }
		    /* Lighting */
		    else if(!strcasecmp(parm, "Lighting"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->lighting = atol(val);
			}
			else
			{   
			    int i, strc;
			    char *s, **strv = strexp(val, &strc);
			    obj->lighting = 0;
			    for(i = 0; i < strc; i++)
			    {
				s = strv[i];
				if(s == NULL)
				    continue;

				if(*s == '!')
				    continue;

				if(!strcasecmp(s, SW_OBJ_LIGHT_STR_VECTOR))
				    obj->lighting |= SW_OBJ_LIGHT_VECTOR;
				else if(!strcasecmp(s, SW_OBJ_LIGHT_STR_STROBE))
				    obj->lighting |= SW_OBJ_LIGHT_STROBE;
				else if(!strcasecmp(s, SW_OBJ_LIGHT_STR_LUMINATION))
				    obj->lighting |= SW_OBJ_LIGHT_LUMINATION;
			    }
			    strlistfree(strv, strc);
			}
		    }
		    /* BrakesState */
		    else if(!strcasecmp(parm, "BrakesState"))
		    {
			/* No longer supported */
		    }
		    /* BrakesPower */
		    else if(!strcasecmp(parm, "BrakesPower"))
		    {
			/* No longer supported */
		    }
		    /* HitPointsMax */
		    else if(!strcasecmp(parm, "HitPointsMax"))
		    {
			obj->hp_max = atof(val);
		    }
		    /* HitPoints */
		    else if(!strcasecmp(parm, "HitPoints"))
		    {
			obj->hp = atof(val);
		    }
		    /* PowerPurity */
		    else if(!strcasecmp(parm, "PowerPurity"))
		    {
			obj->power_purity = atof(val);
		    }
		    /* PowerMax */
		    else if(!strcasecmp(parm, "PowerMax"))
		    {
			obj->power_max = atof(val);
		    }
		    /* Power */
		    else if(!strcasecmp(parm, "Power"))
		    {
			obj->power = atof(val);
		    }
		    /* CoreEfficency */
		    else if(!strcasecmp(parm, "CoreEfficency"))
		    {
			obj->core_efficency = atof(val);
		    }
		    /* AntimatterMax */
		    else if(!strcasecmp(parm, "AntimatterMax"))
		    {
			obj->antimatter_max = atof(val);
		    }
		    /* Antimatter */
		    else if(!strcasecmp(parm, "Antimatter"))
		    {
			obj->antimatter = atof(val);
		    }
		    /* ShieldState */
		    else if(!strcasecmp(parm, "ShieldState"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->shield_state = atoi(val);
			}
			else
			{
			    if(!strcasecmp(val, SW_SHIELD_STATE_STR_NONE))
				obj->shield_state = SW_SHIELD_STATE_NONE;
			    else if(!strcasecmp(val, SW_SHIELD_STATE_STR_DOWN))
				obj->shield_state = SW_SHIELD_STATE_DOWN;
			    else if(!strcasecmp(val, SW_SHIELD_STATE_STR_UP))
				obj->shield_state = SW_SHIELD_STATE_UP;
			}
		    }
		    /* ShieldFrequency */
		    else if(!strcasecmp(parm, "ShieldFrequency"))
		    {
			obj->shield_frequency = atof(val);
		    }
		    /* SelectedWeapon */
		    else if(!strcasecmp(parm, "SelectedWeapon"))
		    {
			obj->selected_weapon = atoi(val);
		    }
		    /* TotalWeapons */
		    else if(!strcasecmp(parm, "TotalWeapons"))
		    {
			if(UNVAllocObjectWeapons(obj, atoi(val)))
			{
			    fprintf(
				stderr,
"%s: Line %i: Error allocating %s weapons.\n",
				path, lines_read, val
			    );
			    obj->total_weapons = 0;
			    obj->selected_weapon = -1;
			}
		    }
		    /* BirthTime */
		    else if(!strcasecmp(parm, "BirthTime"))
		    {
			obj->birth_time = atol(val);
		    }
		    /* LifeSpan */
		    else if(!strcasecmp(parm, "LifeSpan"))
		    {
			obj->lifespan = atol(val);
		    }
		    /* CreationOCS */
		    else if(!strcasecmp(parm, "CreationOCS"))
		    {
			obj->creation_ocs = atoi(val);
		    }
		    /* CloakState */
		    else if(!strcasecmp(parm, "CloakState"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->cloak_state = atoi(val);
			}
			else
			{
			    if(!strcasecmp(val, SW_CLOAK_STATE_STR_NONE))
				obj->cloak_state = SW_CLOAK_STATE_NONE;
			    else if(!strcasecmp(val, SW_CLOAK_STATE_STR_DOWN))
				obj->cloak_state = SW_CLOAK_STATE_DOWN;
			    else if(!strcasecmp(val, SW_CLOAK_STATE_STR_UP))
				obj->cloak_state = SW_CLOAK_STATE_UP;
			}
		    }
		    /* CloakStrength */
		    else if(!strcasecmp(parm, "CloakStrength"))
		    {
			obj->cloak_strength = MAX(atof(val), 0.0f);
		    }
		    /* CloakRefractoryInt */
		    else if(!strcasecmp(parm, "CloakRefractoryInt"))
		    {
			obj->cloak_refractory_int = MAX(atol(val), 0l);
		    }
		    /* ShieldVisibility */
		    else if(!strcasecmp(parm, "ShieldVisibility"))
		    {
			obj->shield_visibility = MAX(atof(val), 0.0f);
		    }
		    /* CurrentVisibility */
		    else if(!strcasecmp(parm, "CurrentVisibility"))
		    {
			obj->cur_visibility = MAX(atof(val), 0.0f);
		    }
		    /* Visibility */
		    else if(!strcasecmp(parm, "Visibility"))
		    {
			obj->visibility = MAX(atof(val), 0.0f);
		    }
		    /* DamageControl */
		    else if(!strcasecmp(parm, "DamageControl"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->damage_control = atoi(val);
			}
			else
			{
			    if(!strcasecmp(val, SW_DMGCTL_STATE_STR_OFF))
				obj->damage_control = SW_DMGCTL_STATE_OFF;
			    else if(!strcasecmp(val, SW_DMGCTL_STATE_STR_ON))
				obj->damage_control = SW_DMGCTL_STATE_ON;
			}
		    }
		    /* ComChannel */
		    else if(!strcasecmp(parm, "ComChannel"))
		    {
			obj->com_channel = atoi(val);
			if(((float)obj->com_channel / 100) > SW_FREQ_MAX)
			    obj->com_channel = static_cast<int>(SW_FREQ_MAX / 100);
			else if(((float)obj->com_channel / 100) < SW_FREQ_MIN)
			    obj->com_channel = static_cast<int>(SW_FREQ_MIN / 100);
		    }
		    /* AutoDestructNext */
		    else if(!strcasecmp(parm, "AutoDestructNext"))
		    {
			obj->auto_destruct_next = atol(val);
		    }
		    /* AIFlags */
		    else if(!strcasecmp(parm, "AIFlags"))
		    {
			if(isdigit(*val) || (*val == '-'))
			{
			    obj->ai_flags = atol(val);
			}
			else
			{
			    int i, strc;
			    char *s, **strv = strexp(val, &strc);
			    obj->ai_flags = 0;
			    for(i = 0; i < strc; i++)
			    {
				s = strv[i];
				if(s == NULL)
				    continue;

				if(*s == '!')
				    continue;

				if(!strcasecmp(s, SW_AI_STR_FOLLOW_FRIEND))
				    obj->ai_flags |= SW_AI_FOLLOW_FRIEND;
				else if(!strcasecmp(s, SW_AI_STR_FOLLOW_UNKNOWN))
				    obj->ai_flags |= SW_AI_FOLLOW_UNKNOWN;
				else if(!strcasecmp(s, SW_AI_STR_FOLLOW_HOSTILE))
				    obj->ai_flags |= SW_AI_FOLLOW_HOSTILE;
				else if(!strcasecmp(s, SW_AI_STR_FIRE_FRIEND))
				    obj->ai_flags |= SW_AI_FIRE_FRIEND;
				else if(!strcasecmp(s, SW_AI_STR_FIRE_UNKNOWN))
				    obj->ai_flags |= SW_AI_FIRE_UNKNOWN;
				else if(!strcasecmp(s, SW_AI_STR_FIRE_HOSTILE))
				    obj->ai_flags |= SW_AI_FIRE_HOSTILE;
			    }
			    strlistfree(strv, strc);
			}
		    }

		    /* *************************************************** */
		    /* ScoreCredits */
		    else if(!strcasecmp(parm, "ScoreCredits"))
		    {
			if(!UNVAllocScores(obj))
			{
			    obj->score->credits = atof(val);
			}
		    }
		    /* ScoreRMUMax */
		    else if(!strcasecmp(parm, "ScoreRMUMax"))
		    {
			if(!UNVAllocScores(obj))
			{
			    obj->score->rmu_max = atof(val);
			}
		    }
		    /* ScoreRMU */
		    else if(!strcasecmp(parm, "ScoreRMU"))
		    {
			if(!UNVAllocScores(obj))
			{   
			    obj->score->rmu = atof(val);
			}
		    }
		    /* ScoreDamageGiven */
		    else if(!strcasecmp(parm, "ScoreDamageGiven") ||
			    !strcasecmp(parm, "ScoreDamageGivin")
		    )
		    {
			if(!UNVAllocScores(obj))
			{
			    obj->score->damage_given = atof(val);
			}
		    }
		    /* ScoreDamageRecieved */
		    else if(!strcasecmp(parm, "ScoreDamageRecieved"))
		    {
			if(!UNVAllocScores(obj))
			{   
			    obj->score->damage_received = atof(val);
			}
		    }
		    /* ScoreKills */
		    else if(!strcasecmp(parm, "ScoreKills"))
		    {
			if(!UNVAllocScores(obj))
			{
			    obj->score->kills = atoi(val);
			}
		    }
		    /* *************************************************** */
		    /* PermissionUID */
		    else if(!strcasecmp(parm, "PermissionUID"))
		    {
			obj->permission.uid = MAX(atoi(val), 0);
		    }
		    /* PermissionGID */
		    else if(!strcasecmp(parm, "PermissionGID"))
		    {
			obj->permission.gid = MAX(atoi(val), 0);
		    }
		    /* *************************************************** */
		    /* AnimationLastInterval */
		    else if(!strcasecmp(parm, "AnimationLastInterval"))
		    {
			obj->animation.last_interval = (time_t)MAX(atol(val), 0l);
		    }
		    /* AnimationInterval */
		    else if(!strcasecmp(parm, "AnimationInterval"))
		    {
			obj->animation.interval = (time_t)MAX(atol(val), 0l);
		    }
		    /* AnimationCurrentFrame */
		    else if(!strcasecmp(parm, "AnimationCurrentFrame"))
		    {
			obj->animation.current_frame = MAX(atoi(val), 0);
		    }
		    /* AnimationTotalFrames */
		    else if(!strcasecmp(parm, "AnimationTotalFrames"))
		    {
			obj->animation.total_frames = MAX(atoi(val), 1);
		    }
		    /* AnimationCycleCount */
		    else if(!strcasecmp(parm, "AnimationCycleCount"))
		    {
			obj->animation.cycle_count = MAX(atoi(val), 0);
		    }
		    /* AnimationCycleTimes */
		    else if(!strcasecmp(parm, "AnimationCycleTimes"))
		    {
			obj->animation.cycle_times = MAX(atoi(val), 0);
		    }
		    /* **************************************************** */
		    /* BeginWeapon */
		    else if(!strcasecmp(parm, "BeginWeapon"))
		    {
			xsw_weapon_struct *wep = (weapons_count < obj->total_weapons) ?
			    obj->weapons[weapons_count] : NULL;

			while(1)
			{
			    /* Get the next line */
			    free(s);
			    s = FReadNextLineAllocCount(
				fp, UNIXCFG_COMMENT_CHAR, &lines_read
			    );
			    if(s == NULL)
				break;

			    /* Get the parameter from the line */
			    s2 = StringCfgParseParm(s);
			    if(s2 == NULL)
				continue;
			    strncpy(parm, s2, sizeof(parm));
			    parm[sizeof(parm) - 1] = '\0';

			    /* Get the value from the line */
			    s2 = StringCfgParseValue(s); 
			    if(s2 == NULL)
				s2 = "0";
			    strncpy(val, s2, sizeof(val));
			    val[sizeof(val) - 1] = '\0';


			    /* Flags */
			    if(!strcasecmp(parm, "Flags"))
			    {
				if(wep != NULL)
				{
				    if(isdigit(*val) || (*val == '-'))
				    {
					wep->flags = atol(val);
				    }
				    else
				    {
					int i, strc;
					char *s, **strv = strexp(val, &strc);
					wep->flags = 0;
					for(i = 0; i < strc; i++)
					{
					    s = strv[i];
					    if(s == NULL)
						continue;

					    if(*s == '!')
						continue;

			if(!strcasecmp(s, SW_WEP_FLAG_STR_NO_LOCK))
			    wep->flags |= SW_WEP_FLAG_NO_LOCK;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_FIRE_CLOAKED))
			    wep->flags |= SW_WEP_FLAG_FIRE_CLOAKED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_NO_FIRE_SOUND))
			    wep->flags |= SW_WEP_FLAG_NO_FIRE_SOUND;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_FIXED))
			    wep->flags |= SW_WEP_FLAG_FIXED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_PENETRATE_SHIELD))
			    wep->flags |= SW_WEP_FLAG_PENETRATE_SHIELD;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_DESTROYABLE))
			    wep->flags |= SW_WEP_FLAG_DESTROYABLE;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SMOKE_TRAIL))
			    wep->flags |= SW_WEP_FLAG_SMOKE_TRAIL;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SMOKE_PUFF_FIRED))
			    wep->flags |= SW_WEP_FLAG_SMOKE_PUFF_FIRED;
			else if(!strcasecmp(s, SW_WEP_FLAG_STR_SHOCKWAVE_HIT))
			    wep->flags |= SW_WEP_FLAG_SHOCKWAVE_HIT;

					}
					strlistfree(strv, strc);
				    }
				}
			    }
			    /* EmissionType */
			    else if(!strcasecmp(parm, "EmissionType"))
			    {
				if(wep != NULL)
				{
				    if(isdigit(*val) || (*val == '-'))
				    {
					wep->emission_type = atoi(val);
				    }
				    else
				    {
			if(!strcasecmp(val, SW_WEP_EMISSION_STR_STREAM))
			    wep->emission_type = SW_WEP_EMISSION_STREAM;
			else if(!strcasecmp(val, SW_WEP_EMISSION_STR_PROJECTILE))
			    wep->emission_type = SW_WEP_EMISSION_PROJECTILE;
			else if(!strcasecmp(val, SW_WEP_EMISSION_STR_PULSE))
			    wep->emission_type = SW_WEP_EMISSION_PULSE;
				    }
				}
			    }
			    /* OCSCode */
			    else if(!strcasecmp(parm, "OCSCode"))
			    {
				if(wep != NULL)
				    wep->ocs_code = atoi(val);
			    }
			    /* Amount */
			    else if(!strcasecmp(parm, "Amount"))
			    {
				if(wep != NULL)
				    wep->amount	= atoi(val);
			    }
			    /* Max */
			    else if(!strcasecmp(parm, "Max"))
			    {
				if(wep != NULL)
				    wep->max = atoi(val);
			    }
			    /* CreatePower */
			    else if(!strcasecmp(parm, "CreatePower"))
			    {
				if(wep != NULL)
				    wep->create_power = atof(val);
			    }
			    /* Power (pre 2.3.x compatability) */
			    else if(!strcasecmp(parm, "Power"))
			    {
				if(wep != NULL)
				    wep->damage_hull =
					wep->damage_shields = atof(val);
			    }
			    /* DamageHull */
			    else if(!strcasecmp(parm, "DamageHull"))
			    {
				if(wep != NULL)
				    wep->damage_hull = atof(val);
			    }
			    /* DmgShields */
			    else if(!strcasecmp(parm, "DamageShields"))
			    {
				if(wep != NULL)
				    wep->damage_shields = atof(val);
			    }
			    /* Range */
			    else if(!strcasecmp(parm, "Range"))
			    {
				if(wep != NULL)
				    wep->range = atol(val);
			    }
			    /* Delay */
			    else if(!strcasecmp(parm, "Delay"))
			    {
				if(wep != NULL)
				    wep->delay = atol(val);
			    }
			    /* LastUsed */
			    else if(!strcasecmp(parm, "LastUsed"))
			    {
				if(wep != NULL)
				    wep->last_used = atol(val);
			    }
			    /* FireSndRef */
			    else if(!strcasecmp(parm, "FireSndRef") ||
				    !strcasecmp(parm, "FireSoundCode")
			    )
			    {
				if(wep != NULL)
				    wep->fire_sndref = atoi(val);
			    }
			    /* EndWeapon */
			    else if(!strcasecmp(parm, "EndWeapon"))
			    {
			        weapons_count++;
			        break;
			    }
			    /* Unsupported property */
			    else
			    {
				fprintf(
				    stderr,
"%s: Line %i: Warning: Unsupported parameter `%s'\n",
				    path, lines_read, parm
				);
			    }
		        }
		    }
		    /* ************************************************ */
		    /* BeginEcoData */
		    else if(!strcasecmp(parm, "BeginEcoData"))
		    {
			/* Allocate economy structure on object as needed */
		        if(UNVAllocEco(obj))
			    continue;

		        eco = obj->eco;
		        if(eco == NULL)
			    continue;

			while(1)
			{
			    /* Get the next line */
			    free(s);
			    s = FReadNextLineAllocCount(
				fp, UNIXCFG_COMMENT_CHAR, &lines_read
			    );
			    if(s == NULL)
				break;

			    /* Get the parameter from the line */
			    s2 = StringCfgParseParm(s);
			    if(s2 == NULL)
				continue;
			    strncpy(parm, s2, sizeof(parm));
			    parm[sizeof(parm) - 1] = '\0';

			    /* Get the value from the line */
			    s2 = StringCfgParseValue(s);
			    if(s2 == NULL)
				s2 = "0";
			    strncpy(val, s2, sizeof(val));
			    val[sizeof(val) - 1] = '\0';


			    /* Flags */
		            if(!strcasecmp(parm, "Flags"))
			    {
				if(isdigit(*val) || (*val == '-'))
				{
				    eco->flags = atol(val);
				}
				else
				{
				    int i, strc;
				    char *s, **strv = strexp(val, &strc);
				    eco->flags = 0;
				    for(i = 0; i < strc; i++)
				    {
					s = strv[i];
					if(s == NULL)
					    continue;

					if(*s == '!')
					    continue;

					if(!strcasecmp(s, ECO_FLAG_NAME_OPEN))
					    eco->flags |= ECO_FLAG_OPEN;
					else if(!strcasecmp(s, ECO_FLAG_NAME_BUY_OK))
					    eco->flags |= ECO_FLAG_BUY_OK;
					else if(!strcasecmp(s, ECO_FLAG_NAME_SELL_OK))
					    eco->flags |= ECO_FLAG_SELL_OK;
					else if(!strcasecmp(s, ECO_FLAG_NAME_TRADE_OK))
					    eco->flags |= ECO_FLAG_TRADE_OK;
					else if(!strcasecmp(s, ECO_FLAG_NAME_INTRODUCE_OK))
					    eco->flags |= ECO_FLAG_INTRODUCE_OK;
				    }
				    strlistfree(strv, strc);
				}
			    }
			    /* TaxGeneral */
			    else if(!strcasecmp(parm, "TaxGeneral"))
			    {
				eco->tax_general = MAX(atof(val), 0.0f);
			    }
			    /* TaxFriend */
			    else if(!strcasecmp(parm, "TaxFriend"))
			    {
				eco->tax_friend = MAX(atof(val), 0.0f);
			    }
			    /* TaxHostile */
			    else if(!strcasecmp(parm, "TaxHostile"))
			    {
				eco->tax_hostile = MAX(atof(val), 0.0f);
			    }
			    /* **************************************** */
			    /* BeginEcoProduct */
			    else if(!strcasecmp(parm, "BeginEcoProduct"))
			    {
				xsw_ecoproduct_struct *product;

				/* Allocate an eco product */
				if(eco->total_products < 0)
				    eco->total_products = 0;
				product_num = eco->total_products;
				eco->total_products = product_num + 1;

				eco->product = (xsw_ecoproduct_struct **)realloc(
				    eco->product,
				    eco->total_products * sizeof(xsw_ecoproduct_struct *)
				);
				if(eco->product == NULL)
				{
				    eco->total_products = 0;
				    continue;
				}

				eco->product[product_num] = product = (xsw_ecoproduct_struct *)
				    calloc(1, sizeof(xsw_ecoproduct_struct));
				if(product == NULL)
				{
				    product_num = -1;
				    continue;
				}

				while(1)
				{
				    /* Get the next line */
				    free(s);
				    s = FReadNextLineAllocCount(
				        fp, UNIXCFG_COMMENT_CHAR, &lines_read
				    );
				    if(s == NULL)
					break;

				    /* Get the parameter from the line */
				    s2 = StringCfgParseParm(s);
				    if(s2 == NULL)
					continue;
				    strncpy(parm, s2, sizeof(parm));
				    parm[sizeof(parm) - 1] = '\0';

				    /* Get the value from the line */
				    s2 = StringCfgParseValue(s);
				    if(s2 == NULL)
					s2 = "0";
				    strncpy(val, s2, sizeof(val));
				    val[sizeof(val) - 1] = '\0';


				    /* Name */
				    if(!strcasecmp(parm, "Name"))
				    {
					strncpy(
					    product->name, val,
					    sizeof(product->name)
					);
					product->name[sizeof(product->name) - 1] = '\0';
				    }
				    /* OCSCode */
				    else if(!strcasecmp(parm, "OCSCode"))
				    {
					product->ocs_code = atoi(val);
				    }
				    /* SellPrice */
				    else if(!strcasecmp(parm, "SellPrice"))
				    {
					product->sell_price = atof(val);
				    }
				    /* BuyPrice */
				    else if(!strcasecmp(parm, "BuyPrice"))
				    {
				        product->buy_price = atof(val);
				    }
				    /* Amount */
				    else if(!strcasecmp(parm, "Amount"))
				    {
					product->amount = atof(val);
				    }
				    /* AmountMax */
				    else if(!strcasecmp(parm, "AmountMax"))
				    {
					product->amount_max = atof(val);
				    }
				    /* EndEcoProduct */
				    else if(!strcasecmp(parm, "EndEcoProduct"))
				    {
					break;
				    }
				}
			    }	/* BeginEcoProduct */
 
		            /* EndEcoData */
		            else if(!strcasecmp(parm, "EndEcoData"))
		            {
/*
				DBSortEconomyProducts(obj_num);
 */
			        break;
		            }
		        }
		    }

		    /* ***************************************************** */
		    /* EndXSWObject */
		    else if(!strcasecmp(parm, "EndXSWObject"))
		    {
			break;
		    }
		    /* Unsupported property */
		    else
		    {
		        fprintf(
			    stderr,
"%s: Line %i: Warning: Unsupported parameter `%s'\n",
			    path, lines_read, parm
	                );
		    }
		}
	    }
	}


	/* Close the file */
	fclose(fp);

	/* Make sure the lost and found owner is valid */
	if(legend != NULL)
	{
	    char invalid = 0;

	    i = legend->lost_found_owner;

	    if((i < 0) || (i >= *total))
	        invalid = 1;
	    else if(list[i] == NULL)
		invalid = 1;
	    else if(list[i]->type != SW_OBJ_TYPE_PLAYER)
		invalid = 0;	/* Let this be okay for now, StarCharts
				 * may have no players */
	    else
		invalid = 0;

	    if(invalid)
	    {
	        fprintf(
		    stderr,
    "%s: Warning: Lost and found owner object #%i is invalid.\n",
		    path, i
	        );
	        legend->lost_found_owner = 0;
	    }
	}

	/* Call notify progress function */
	if(progress_cb != NULL)
	    progress_cb(
		progress_data,
		file_size, file_size
	    );

	return(list);
}


/*
 *	Writes the list of objects to the specified file.
 */
int UNVFileSave(
	const char *path, const mode_t mode,
	xsw_object_struct **list, int total,
	sw_legend_struct *legend,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data
)
{
	int i, n;
	FILE *fp;
	xsw_weapon_struct *wep;
	xsw_ecodata_struct *eco;
	xsw_object_struct *obj;

	if(progress_cb != NULL)
	    progress_cb(progress_data, 0l, (unsigned long)total);

	if(STRISEMPTY(path))
	    return(-2);

	/* Open file for writing */
	fp = fopen(path, "w");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(
		    STRLEN(path) + STRLEN(error_msg) + 80
		);
		sprintf(
		    s,
"%s: %s.",
		    path, error_msg
		);
		fprintf(stderr, "%s\n", s);
		free(s);
	    }
	    free(error_msg);
	    return(-1);
	}

	/* Set the file's permissions */
	fchmod(fileno(fp), mode);

	/* Write header */
	fprintf(fp, "# ShipWars Universe File\n");
	fprintf(fp, "#\n\n");

	if(legend != NULL)
	{
	    fprintf(fp, "# Header\n");
	    fprintf(fp, "BeginHeader\n");
	    fprintf(fp, "\tTitle = %s\n", legend->title);
	    fprintf(fp, "\tVersion = %s\n", legend->version);

	    fprintf(fp, "\tOCSNFile = %s\n", legend->ocsn);
	    fprintf(fp, "\tImgRefsFile = %s\n", legend->imgrefs);
	    fprintf(fp, "\tSndRefsFile = %s\n", legend->sndrefs);

	    if(legend->ru_to_au < 1.0f)
	        fprintf(
		    fp,
		    "\tRUToAU = %.6f\n",
		    legend->ru_to_au
		);
	    else
	        fprintf(
		    fp,
		    "\tRUToAU = %.2f\n",
		    legend->ru_to_au
		);

	    fprintf(
		fp,
		"\tLostFoundOwner = %i\n",
		legend->lost_found_owner
	    );

	    fprintf(
		fp,
		"\tPlayerStartSectX = %ld\n",
		legend->player_start_sect_x
	    );
	    fprintf(
		fp,
		"\tPlayerStartSectY = %ld\n",
		legend->player_start_sect_y
	    );
	    fprintf(
		fp,
		"\tPlayerStartSectZ = %ld\n",
		legend->player_start_sect_z
	    );
	    fprintf(
		fp,
		"\tPlayerStartX = %f\n",
		legend->player_start_x
	    );
	    fprintf(
		fp,
		"\tPlayerStartY = %f\n",
		legend->player_start_y
	    );
	    fprintf(
		fp,
		"\tPlayerStartZ = %f\n",
		legend->player_start_z
	    );
	    fprintf(
		fp,
		"\tPlayerStartHeading = %f\n",
		legend->player_start_heading
	    );
	    fprintf(
		fp,
		"\tPlayerStartPitch = %f\n",
		legend->player_start_pitch
	    );
	    fprintf(
		fp,
		"\tPlayerStartBank = %f\n",
		legend->player_start_bank
	    );

	    fprintf(
		fp,
		"\tGuestStartSectX = %ld\n",
		legend->guest_start_sect_x
	    );
	    fprintf(
		fp,
		"\tGuestStartSectY = %ld\n",
		legend->guest_start_sect_y
	    );
	    fprintf(
		fp,
		"\tGuestStartSectZ = %ld\n",
		legend->guest_start_sect_z
	    );
	    fprintf(
		fp,
		"\tGuestStartX = %f\n",
		legend->guest_start_x
	    );
	    fprintf(
		fp,
		"\tGuestStartY = %f\n",
		legend->guest_start_y
	    );
	    fprintf(
		fp,
		"\tGuestStartZ = %f\n",
		legend->guest_start_z
	    );
	    fprintf(
		fp,
		"\tGuestStartHeading = %f\n",
		legend->guest_start_heading
	    );
	    fprintf(
		fp,
		"\tGuestStartPitch = %f\n",
		legend->guest_start_pitch
	    );
	    fprintf(
		fp,
		"\tGuestStartBank = %f\n",
		legend->guest_start_bank
	    );

	    fprintf(
		fp, 
		"\tAntimatterLowWarning = %f\n",
		legend->antimatter_low_warning
	    );
	    fprintf(
		fp,
		"\tAsteroidFieldDamageInt = %ld\n",
		legend->asteroid_field_damage_int
	    );
	    fprintf(
		fp,
		"\tAsteroidFieldDamageRate = %f\n",
		legend->asteroid_field_damage_rate
	    );
	    fprintf(
		fp,  
		"\tAutoDestructDamageRange = %f\n",
		legend->auto_destruct_damage_range
	    );
	    fprintf(
		fp,
		"\tCloakAntimatterRate = %f\n",
		legend->cloak_antimatter_rate
	    );  
	    fprintf(
		fp,
		"\tComRange = %f\n",
		legend->com_range
	    );  
	    fprintf(
		fp,
		"\tDamageControlRate = %f\n",
		legend->damage_control_rate
	    );
	    fprintf(
		fp,
		"\tDisarmedWeaponLifespan = %ld\n",
		legend->disarmed_weapon_lifespan
	    );
	    fprintf(
		fp,
		"\tEngineAntimatterRate = %f\n",
		legend->engine_antimatter_rate
	    );
	    fprintf(
		fp,  
		"\tNoAntimatterThrustCoeff = %f\n",
		legend->no_antimatter_thrust_coeff
	    );
	    fprintf(
		fp,
		"\tOnFireHPCoeff = %f\n",
		legend->on_fire_hp_coeff
	    );  
	    fprintf(
		fp,
		"\tTractorBeamAntimatterRate = %f\n",
		legend->tractor_beam_antimatter_rate
	    );
	    fprintf(
		fp,
		"\tTractorBeamRange = %f\n",
		legend->tractor_beam_range
	    );
	    fprintf(
		fp,
		"\tTransporterRange = %f\n",
		legend->transporter_range
	    );

	    fprintf(fp, "EndHeader\n\n");
	}


	/* ********************************************************** */
	/* Begin writing objects to file */
	for(i = 0; i < total; i++)
	{
	    obj = list[i];
	    if(obj == NULL)
		continue;

	    /* Call notify progress function */   
	    if(progress_cb != NULL)
		progress_cb(
		    progress_data,
		    (unsigned long)i, (unsigned long)total
		);

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;


	    /* object_count is now of a valid object that needs
	     * to be saved, begin saving xsw object number
	     * object_count
	     */

	    /* Comment */
	    fprintf(fp, "# Object #%i\n", i);

	    /* BeginXSWObject */
	    fprintf(fp, "BeginXSWObject = %i\n", i);

	    /* Name */
	    fprintf(fp, "\tName = %s\n", obj->name);

	    /* Password */
	    if(!STRISEMPTY(obj->password))
		fprintf(fp, "\tPassword = %s\n", obj->password);

	    /* Empire */
	    strtoupper(obj->empire);	/* Make sure upper case */
	    fprintf(fp, "\tEmpire = %s\n", obj->empire);

	    /* Description */
	    if(obj->description != NULL)
		fprintf(fp, "\tDescription = %s\n", obj->description);

	    /* ELink */
	    if(obj->elink != NULL)
		fprintf(fp, "\tELink = %s\n", obj->elink);

	    /* Type */
	    fprintf(fp, "\tType = %i\n", obj->type);

	    /* LastConnected */
	    if(obj->last_connected > 0l)
		fprintf(fp, "\tLastConnected = %ld\n", obj->last_connected);

	    /* LocType */
	    fprintf(fp, "\tLocType = %i\n", obj->loc_type);

	    /* LocObjNum */
	    fprintf(fp, "\tLocObjNum = %i\n", obj->loc_obj_num);

	    /* ImgRef */
	    fprintf(fp, "\tImgRef = %i\n", obj->imgref);

	    /* Owner */
	    fprintf(fp, "\tOwner = %i\n", obj->owner);

	    /* Size */
	    fprintf(fp, "\tSize = %ld\n", obj->size);

	    /* LockedOn */
	    fprintf(fp, "\tLockedOn = %i\n", obj->locked_on);

	    /* InterceptingObject */
	    fprintf(fp, "\tInterceptingObject = %i\n",
		obj->intercepting_object
	    );
 
	    /* ScannerRange */
	    fprintf(fp, "\tScannerRange = %.4f\n",
		obj->scanner_range
	    );

	    /* SectX */
	    fprintf(fp, "\tSectX = %ld\n", obj->sect_x);
	    /* SectY */
	    fprintf(fp, "\tSectY = %ld\n", obj->sect_y);
	    /* SectZ */
	    fprintf(fp, "\tSectZ = %ld\n", obj->sect_z);

	    /* CoordX */
	    fprintf(fp, "\tCoordX = %.4f\n", obj->x);
	    /* CoordY */
	    fprintf(fp, "\tCoordY = %.4f\n", obj->y);
	    /* CoordZ */
	    fprintf(fp, "\tCoordZ = %.4f\n", obj->z);

	    /* ObjectHeading */
	    fprintf(fp, "\tObjectHeading = %.5f\n", obj->heading);
	    /* ObjectPitch */
	    fprintf(fp, "\tObjectPitch = %.5f\n", obj->pitch);
	    /* ObjectBank */
	    fprintf(fp, "\tObjectBank = %.5f\n", obj->bank);

	    /* VelocityHeading */
	    fprintf(fp, "\tVelocityHeading = %.5f\n",
		obj->velocity_heading
	    );
	    /* VelocityPitch */
	    fprintf(fp, "\tVelocityPitch = %.5f\n",
		obj->velocity_pitch
	    );
	    /* VelocityBank */
	    fprintf(fp, "\tVelocityBank = %.5f\n",
		obj->velocity_bank
	    );
		   
	    /* VelocityMax */
	    fprintf(fp, "\tVelocityMax = %.5f\n",
		obj->velocity_max
	    );
	    /* Velocity */
	    fprintf(fp, "\tVelocity = %.5f\n",
		obj->velocity
	    );
	    /* ThrustDir */
	    fprintf(fp, "\tThrustDir = %.4f\n",
		obj->thrust_dir
	    );
	    /* ThrustPower */
	    fprintf(fp, "\tThrustPower = %.5f\n",
		obj->thrust_power
	    );
	    /* Thrust */
	    fprintf(fp, "\tThrust = %.5f\n",
		obj->thrust
	    );
	    /* Throttle */
	    fprintf(fp, "\tThrottle = %.5f\n",
		obj->throttle
	    );
	    /* EngineState */
	    fprintf(fp, "\tEngineState = %i\n",
		obj->engine_state
	    );
	    /* TurnRate */
	    fprintf(fp, "\tTurnRate = %.5f\n",
		obj->turnrate
	    );
	    /* OdoMeter */
	    if(obj->odometer > 0.0f)
		fprintf(
		    fp,
		    "\tOdoMeter = %f\n",
		    obj->odometer
		);
	    /* Lighting */
	    fprintf(fp, "\tLighting = %lu\n",
		(unsigned long)obj->lighting
	    );
	    /* HitPointsMax */
	    fprintf(fp, "\tHitPointsMax = %.4f\n",
		obj->hp_max
	    );
	    /* HitPoints */
	    fprintf(fp, "\tHitPoints = %.4f\n",
		obj->hp
	    );
	    /* PowerPurity */
	    fprintf(fp, "\tPowerPurity = %.4f\n",
		obj->power_purity
	    );
	    /* PowerMax */
	    fprintf(fp, "\tPowerMax = %.4f\n",
		obj->power_max
	    );
	    /* Power */
	    fprintf(fp, "\tPower = %.4f\n",
		obj->power
	    );
	    /* CoreEfficency */
	    fprintf(fp, "\tCoreEfficency = %.4f\n",
		obj->core_efficency
	    );
	    /* AntimatterMax */
	    fprintf(fp, "\tAntimatterMax = %.4f\n",
		obj->antimatter_max
	    );
	    /* Antimatter */
	    fprintf(fp, "\tAntimatter = %.4f\n",
		obj->antimatter
	    );
	    /* ShieldState */
	    fprintf(fp, "\tShieldState = %i\n",
		obj->shield_state
	    );
	    /* ShieldFrequency */
	    fprintf(fp, "\tShieldFrequency = %.4f\n",
		obj->shield_frequency
	    );

	    /* SelectedWeapon */
	    if(obj->selected_weapon >= SW_OBJ_WEAPONS_MAX)
		obj->selected_weapon = SW_OBJ_WEAPONS_MAX - 1;
	    if(obj->selected_weapon < -1)
		obj->selected_weapon = -1;
	    fprintf(fp, "\tSelectedWeapon = %i\n",
		obj->selected_weapon
	    );

	    /* TotalWeapons */ 
	    if(obj->total_weapons > SW_OBJ_WEAPONS_MAX)
		obj->total_weapons = SW_OBJ_WEAPONS_MAX;
	    if(obj->total_weapons < 0)
		obj->total_weapons = 0;
	    fprintf(fp, "\tTotalWeapons = %i\n",
		obj->total_weapons
	    );

	    /* BirthTime */
	    if(obj->birth_time > 0l)
		fprintf(
		    fp,
		    "\tBirthTime = %ld\n",
		    obj->birth_time
		);
	    /* LifeSpan */
	    if(obj->lifespan > 0l)
		fprintf(
		    fp,
		    "\tLifeSpan = %ld\n",
		    obj->lifespan
		);
	    /* CreationOCS */
	    fprintf(
		fp,
		"\tCreationOCS = %i\n",
		obj->creation_ocs
	    );
	    /* CloakState */
	    fprintf(
		fp,
		"\tCloakState = %i\n",
		obj->cloak_state
	    );
	    /* CloakStrength */
	    if(obj->cloak_strength > 0.0f)
		fprintf(
		    fp,
		    "\tCloakStrength = %.4f\n",
		    obj->cloak_strength
		);
	    /* CloakRefractoryInt */
	    if(obj->cloak_refractory_int > 0l)
		fprintf(
		    fp,
		    "\tCloakRefractoryInt = %ld\n",
		    obj->cloak_refractory_int
		);
	    /* ShieldVisibility */
	    if(obj->shield_visibility > 0.0f)
		fprintf(
		    fp,
		    "\tShieldVisibility = %.4f\n",
		    obj->shield_visibility
		);
	    /* CurrentVisibility */
	    fprintf(
		fp,
		"\tCurrentVisibility = %.4f\n",
		obj->cur_visibility
	    );
	    /* Visibility */
	    fprintf(
		fp,
		"\tVisibility = %.4f\n",
		obj->visibility
	    );
	    /* DamageControl */
	    fprintf(fp, "\tDamageControl = %i\n",
		obj->damage_control
	    );
	    /* ComChannel */
	    fprintf(fp, "\tComChannel = %i\n",
		obj->com_channel
	    );
	    /* AutoDestructNext */
	    if(obj->auto_destruct_next > 0l)
		fprintf(
		    fp,
		    "\tAutoDestructNext = %ld\n",
		    obj->auto_destruct_next
		);
	    /* AIFlags */
	    fprintf(
		fp,
		"\tAIFlags = %lu\n",
		(unsigned long)obj->ai_flags
	    );

	    /* Scores */
	    if(obj->score != NULL)
	    {
	        /* ScoreCredits */
		fprintf(fp, "\tScoreCredits = %.4f\n",
		    obj->score->credits
		);
		/* ScoreRMUMax */
		fprintf(fp, "\tScoreRMUMax = %.4f\n",
		    obj->score->rmu_max
		);
		/* ScoreRMU */
		fprintf(fp, "\tScoreRMU = %.4f\n",
		    obj->score->rmu
		);
		/* ScoreDamageGiven */
		fprintf(fp, "\tScoreDamageGiven = %.4f\n",
		    obj->score->damage_given
		);
		/* ScoreDamageRecieved */
		fprintf(fp, "\tScoreDamageRecieved = %.4f\n",
		    obj->score->damage_received
		);
		/* ScoreKills */
		fprintf(fp, "\tScoreKills = %i\n",
		    obj->score->kills
		);
	    }

	    /* Permission stuff */
	    /* PermissionUID */
	    fprintf(fp, "\tPermissionUID = %i\n",
		obj->permission.uid
	    );
	    /* PermissionGID */
	    fprintf(fp, "\tPermissionGID = %i\n",
		obj->permission.gid
	    ); 

	    /* Animation stuff */
	    /* AnimationLastInterval */
	    fprintf(fp, "\tAnimationLastInterval = %ld\n",
		obj->animation.last_interval
	    );

	    /* AnimationInterval */
	    fprintf(fp, "\tAnimationInterval = %ld\n",
		obj->animation.interval
	    );
	    /* AnimationCurrentFrame */
	    fprintf(fp, "\tAnimationCurrentFrame = %i\n",
		obj->animation.current_frame
	    );
	    /* AnimationTotalFrames */
	    fprintf(fp, "\tAnimationTotalFrames = %i\n",
		obj->animation.total_frames
	    );
	    /* AnimationCycleCount */
	    fprintf(fp, "\tAnimationCycleCount = %i\n",
		obj->animation.cycle_count
	    );
	    /* AnimationCycleTimes */
	    fprintf(fp, "\tAnimationCycleTimes = %i\n",
		obj->animation.cycle_times
	    );

 	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;

		fprintf(fp, "\tBeginWeapon\n");

		fprintf(fp, "\t\tFlags = %lu\n",
		    (unsigned long)wep->flags
		);
		fprintf(fp, "\t\tEmissionType = %i\n",
		    wep->emission_type
		);
		fprintf(fp, "\t\tOCSCode = %i\n",
		    wep->ocs_code
		);
		fprintf(fp, "\t\tAmount = %i\n",
		    wep->amount
		);
		fprintf(fp, "\t\tMax = %i\n",
		    wep->max
		);
		fprintf(fp, "\t\tCreatePower = %.4f\n",
		    wep->create_power
		);
		fprintf(fp, "\t\tDamageHull = %.4f\n",
		    wep->damage_hull
		);
		fprintf(fp, "\t\tDamageShields = %.4f\n",
		    wep->damage_shields
		);
		fprintf(fp, "\t\tRange = %ld\n",
		    wep->range
		);
		fprintf(fp, "\t\tDelay = %ld\n",
		    wep->delay
		);
		fprintf(fp, "\t\tLastUsed = 0\n");	/* Always 0 */
		fprintf(fp, "\t\tFireSndRef = %i\n",
		    wep->fire_sndref
		);

		fprintf(fp, "\tEndWeapon\n");
	    }

	    /* Economy */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		xsw_ecoproduct_struct *product;

		fprintf(fp, "\tBeginEcoData\n");
		fprintf(fp, "\t\tFlags = %lu\n",
		    (unsigned long)eco->flags
		);
		fprintf(fp, "\t\tTaxGeneral = %.4f\n",
		    eco->tax_general
		);
		fprintf(fp, "\t\tTaxFriend = %.4f\n",
		    eco->tax_friend
		);
		fprintf(fp, "\t\tTaxHostile = %.4f\n",
		    eco->tax_hostile
		);

		/* Products */
		for(n = 0; n < eco->total_products; n++)
		{
		    product = eco->product[n];
		    if(product == NULL)
			continue;

		    fprintf(fp, "\t\tBeginEcoProduct\n");
		    fprintf(fp, "\t\t\tName = %s\n",
			product->name
		    );
		    fprintf(fp, "\t\t\tOCSCode = %i\n",
			product->ocs_code
		    );
		    fprintf(fp, "\t\t\tSellPrice = %.4f\n", 
			product->sell_price
		    );
		    fprintf(fp, "\t\t\tBuyPrice = %.4f\n",
			product->buy_price
		    );
		    fprintf(fp, "\t\t\tAmount = %.4f\n",
			product->amount
		    );
		    fprintf(fp, "\t\t\tAmountMax = %.4f\n",
			product->amount_max
		    );

		    fprintf(fp, "\t\tEndEcoProduct\n");
		}

		fprintf(fp, "\tEndEcoData\n");
	    }

	    /* EndXSWObject */
	    fprintf(fp, "EndXSWObject\n\n");
	}

	/* Close the file */
	fclose(fp);

	/* Report final progress */
	if(progress_cb != NULL)
	    progress_cb(
		progress_data,
		(unsigned long)total, (unsigned long)total
	    );

	return(0);
}
