/*
		   Universe Objects Management Main
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
	  
#include "../include/string.h"
#include "../include/reality.h"   
#include "../include/objects.h"
#include "../include/imgrefs.h"

#include "../include/unvmain.h"
#include "../include/unvutil.h"
#include "../include/unvmatch.h"
#include "../include/unvfile.h"


xsw_object_struct unv_garbage_object;


void UNVObjSetGarbage(xsw_object_struct *obj);
void UNVLegendReset(sw_legend_struct *legend);

int UNVInit(void);
void UNVShutdown(void);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Resets the object's values.
 *
 *	Existing values are assumed to be garbage and will not be
 *	deleted.
 */
void UNVObjSetGarbage(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return;

	obj->type = SW_OBJ_TYPE_GARBAGE;

	obj->ext_flags = 0;

	memset(obj->name, '\0', sizeof(obj->name));

	obj->password = NULL;

	strncpy(
	    obj->empire, SW_DEF_EMPIRE,
	    sizeof(obj->empire)
	);
	obj->empire[sizeof(obj->empire) - 1] = '\0';

	obj->description = NULL;

	obj->elink = NULL;

	obj->last_updated = 0l;
	obj->last_connected = 0l;

	obj->loc_type = SW_LOC_TYPE_SPACE;
	obj->loc_obj_num = -1;
	obj->imgref = IMGREF_DEFAULT_IMGREF;
	obj->owner = -1;
	obj->size = 1l;
	obj->locked_on = -1;
	obj->intercepting_object = -1;

	obj->scanner_range = 10.0f;
	obj->sect_x = 0l;
	obj->sect_y = 0l;
	obj->sect_z = 0l;
	obj->x = 0.0f;
	obj->y = 0.0f;
	obj->z = 0.0f;
	obj->heading = 0.0f * PI;
	obj->pitch = 0.0f * PI;
	obj->bank = 0.0f * PI;
	memset(
	    &obj->attitude_vector_compoent,
	    0x00,
	    sizeof(sw_vector_compoent_struct)
	);
	obj->velocity = 0.0f;
	obj->velocity_max = 0.0f;
	obj->velocity_heading = 0.0f * PI;
	obj->velocity_pitch = 0.0f * PI;
	obj->velocity_bank = 0.0f * PI;
	memset(
	    &obj->momentum_vector_compoent,
	    0x00,
	    sizeof(sw_vector_compoent_struct)
	);
	obj->thrust_dir = 1.0f * PI;
	obj->thrust = 0.0f;
	obj->thrust_power = 0.0f;
	obj->throttle = 0.0f;
	obj->engine_state = SW_ENGINE_STATE_ON;
	obj->turnrate = 0.0003f;
	obj->odometer = 0.0f;
	obj->hp = 1.0f;
	obj->hp_max = 1.0f;
	obj->power = 0.0f;
	obj->power_max = 0.0f;
	obj->power_purity = 1.0f;
	obj->core_efficency = 0.0f;
	obj->antimatter = 0.0f;
	obj->antimatter_max = 0.0f;
	obj->shield_state = SW_SHIELD_STATE_NONE;
	obj->shield_frequency = SW_DEF_SHIELD_FREQ;
	obj->selected_weapon = -1;
	obj->total_weapons = 0;
	obj->birth_time = 0l;
	obj->lifespan = 0l;
	obj->creation_ocs = 0;
	obj->cloak_state = SW_CLOAK_STATE_NONE;
	obj->cloak_strength = 0.0f;
	obj->last_cloak = 0l;
	obj->cloak_refractory_int = 0l;
	obj->shield_visibility = 0.0f;
	obj->visibility = 1.0f;
	obj->cur_visibility = 1.0f;
	obj->damage_control = SW_DMGCTL_STATE_OFF;
	obj->com_channel = SW_DEF_COM_CHANNEL;
	obj->auto_destruct_next = 0l;
	obj->ai_flags = 0;

	obj->tractored_object = NULL;
	obj->total_tractored_objects = 0;
 
	obj->permission.uid = SW_DEF_UID;
	obj->permission.gid = SW_DEF_GID;

	obj->animation.interval = 500l;
	obj->animation.last_interval = 0l;
	obj->animation.current_frame = 0;
	obj->animation.total_frames = 1;
	obj->animation.cycle_count = 0;
	obj->animation.cycle_times = 0;

	obj->weapons = NULL;
	obj->score = NULL;
	obj->eco = NULL;
}

/*
 *	Resets the universe legend values to defaults.
 *
 *	Existing values are assumed to be garbage.
 */
void UNVLegendReset(sw_legend_struct *legend)
{
	if(legend == NULL)
	    return;

	memset(legend, 0x00, sizeof(sw_legend_struct));

	strncpy(legend->title, "Untitled", sizeof(legend->title));
	strncpy(legend->ocsn, "default.ocsn", sizeof(legend->ocsn));
	strncpy(legend->imgrefs, "default.isr", sizeof(legend->imgrefs));
	strncpy(legend->sndrefs, "default.ss", sizeof(legend->sndrefs));

	legend->ru_to_au = SW_DEF_RU_TO_AU;
	legend->sect_x_max = SW_SECTOR_SIZE_X_MAX;
	legend->sect_x_min = SW_SECTOR_SIZE_X_MIN;
	legend->sect_y_max = SW_SECTOR_SIZE_Y_MAX;
	legend->sect_y_min = SW_SECTOR_SIZE_Y_MIN;
	legend->sect_z_max = SW_SECTOR_SIZE_Z_MAX;
	legend->sect_z_min = SW_SECTOR_SIZE_Z_MIN;
	legend->sect_x_len = legend->sect_x_max - legend->sect_x_min;
	legend->sect_y_len = legend->sect_y_max - legend->sect_y_min;
	legend->sect_z_len = legend->sect_z_max - legend->sect_z_min;

	legend->lost_found_owner = -1;

	legend->antimatter_low_warning = SW_DEF_ANTIMATTER_LOW_WARNING;
	legend->asteroid_field_damage_int = SW_DEF_ASTEROID_FIELD_DAMAGE_INT;
	legend->asteroid_field_damage_rate = SW_DEF_ASTEROID_FIELD_DAMAGE_RATE;
	legend->auto_destruct_damage_range = SW_DEF_AUTO_DESTRUCT_DAMAGE_RANGE;
	legend->cloak_antimatter_rate = SW_DEF_CLOAK_ANTIMATTER_RATE;
	legend->com_range = SW_DEF_COM_RANGE;
	legend->damage_control_rate = SW_DEF_DAMAGE_CONTROL_RATE;
	legend->disarmed_weapon_lifespan = SW_DEF_DISARMED_WEAPON_LIFESPAN;
	legend->engine_antimatter_rate = SW_DEF_ENGINE_ANTIMATTER_RATE;
	legend->no_antimatter_thrust_coeff = SW_DEF_NO_ANTIMATTER_THRUST_COEFF;
	legend->on_fire_hp_coeff = SW_DEF_ON_FIRE_HP_COEFF;
	legend->tractor_beam_antimatter_rate = SW_DEF_TRACTOR_BEAM_ANTIMATTER_RATE;
	legend->tractor_beam_range = SW_DEF_TRACTOR_BEAM_RANGE;
	legend->transporter_range = SW_DEF_TRANSPORTER_RANGE;
}


/*
 *	Initializes the universe.
 */
int UNVInit(void)
{
	/* Reset the global garbage object values */
	UNVObjSetGarbage(&unv_garbage_object);


	return(0);
}


/*
 *	Shuts down and deallocates universe object management
 *	resources.
 */
void UNVShutdown(void)
{
	return;
}




