#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/widget.h"

#include "../include/oswkeyvals.h"


static int PROMPT_CHARS_VIS(prompt_struct *prompt);

static void PromptDDEStoreASCIISeg(
	win_t w, const char *buf, int len
);

static void PromptResizeCB(void *widget, void *data);
static void PromptDrawCB(void *widget, void *data);
static void PromptScrollTimeoutCB(void *widget, void *data);

static void PromptAddScrollTimeout(
	prompt_struct *prompt, const time_t interval_ms
);
static void PromptRemoveScrollTimeout(prompt_struct *prompt);

void PromptMarkBuffer(prompt_struct *prompt, int cursor_pos);
void PromptUnmarkBuffer(prompt_struct *prompt, int cursor_pos);

float PromptGetF(prompt_struct *prompt);
int PromptGetI(prompt_struct *prompt);
long PromptGetL(prompt_struct *prompt);
char *PromptGetS(prompt_struct *prompt);
void PromptSetF(prompt_struct *prompt, float val);
void PromptSetF2(
	prompt_struct *prompt,
	const float val, const unsigned int ndecimals
);
void PromptSetI(prompt_struct *prompt, int val);
void PromptSetL(prompt_struct *prompt, long val);
void PromptSetUL(prompt_struct *prompt, unsigned long val);
void PromptSetS(prompt_struct *prompt, const char *val);

int PromptInit(
	prompt_struct *prompt, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	int style,
	const char *label,
	unsigned int buf_len, int hist_bufs
);
const char *PromptGetLabel(prompt_struct *prompt);
void PromptSetLabel(prompt_struct *prompt, const char *label);
void PromptSetTooltip(
	prompt_struct *prompt, const char *msg
);
void PromptResize(prompt_struct *prompt);
void PromptQueueResize(prompt_struct *prompt);
static void PROMPT_UNFOCUS_REDRAW_ALL(prompt_struct *except_prompt);
void PromptDraw(prompt_struct *prompt, int amount);
void PromptQueueDraw(prompt_struct *prompt, int amount);
int PromptManage(prompt_struct *prompt, event_t *event);
void PromptMap(prompt_struct *prompt);
void PromptUnmap(prompt_struct *prompt);
void PromptDestroy(prompt_struct *prompt);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : True)

#define INRANGEINC(a,min,max) ( (a >= min) && (a <= max) )


#define PROMPT_DEF_WIDTH		100
#define PROMPT_DEF_HEIGHT		30

#define PROMPT_CHAR_WIDTH		7
#define PROMPT_CHAR_HEIGHT		14

#define PROMPT_MARGIN			4

#define PROMPT_MIN_WIDTH		10
#define PROMPT_MIN_HEIGHT		10


/*
 *	Returns the number of characters visiable at one time
 *	on the prompt.
 */
static int PROMPT_CHARS_VIS(prompt_struct *prompt)
{
	unsigned int width, height;

	if(prompt == NULL)
	    return(0);

	if((prompt->buf == NULL) ||
	   (prompt->text_area == None)
	)
	    return(0);

	OSWGetWindowGeometry(
	    prompt->text_area, NULL, NULL, &width, &height, NULL
	);
	return(MAX(
	    (((int)width - (PROMPT_MARGIN * 2)) / PROMPT_CHAR_WIDTH),
 	    0
	));
}


/*
 *	Macro to put buf of length len into DDE
 */
static void PromptDDEStoreASCIISeg(
	win_t w, const char *buf, int len
)
{
	char *s;

	if((buf == NULL) || (len <= 0))
	    return;

	s = (char *)malloc((len + 1) * sizeof(char));
	if(s == NULL)
	    return;

	memcpy(s, buf, len);
	s[len] = '\0';

	OSWDDESetString(w, s);
	free(s);
}


/*
 *	Resize callback.
 */
static void PromptResizeCB(void *widget, void *data)
{
	prompt_struct *prompt = (prompt_struct *)widget;
	if(prompt == NULL)
	    return;

	PromptResize(prompt);
}

/*
 *	Draw callback.
 */
static void PromptDrawCB(void *widget, void *data)
{
	int amount = (int)data;
	prompt_struct *prompt = (prompt_struct *)widget;
	if(prompt == NULL)
	    return;

	if(prompt->map_state)
	    PromptDraw(prompt, amount);
}

/*
 *	Scroll timeout callback.
 */
static void PromptScrollTimeoutCB(void *widget, void *data)
{
	int x, y, len, prev_pos, prev_vis_pos;
	unsigned int width, height;
	win_t w;
	widget_global_struct *wglobal = &widget_global;
	prompt_struct *prompt = (prompt_struct *)widget;

	if(prompt == NULL)
	    return;

	w = prompt->text_area;
	if(w == None)
	    return;

	len = MIN(STRLEN(prompt->buf), prompt->buf_len);
	prev_pos = prompt->buf_pos;
	prev_vis_pos = prompt->buf_vis_pos;

	OSWGetPointerCoords(w, NULL, NULL, &x, &y);
	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);

	/* Scroll left? */
	if(x < 0)
	{
	    const int dc = MAX(
		((0 - x) / PROMPT_CHAR_WIDTH),
		1
	    );

	    /* Scroll */
	    prompt->buf_vis_pos = MAX(
		prompt->buf_vis_pos - dc,
		0
	    );

	    /* Set cursor position */
	    prompt->buf_pos = prompt->buf_vis_pos;

	    /* Update selection bounds? */
	    if(prompt->selection_in_progress)     
	    {
		if(prompt->buf_pos < prompt->buf_sel_origin)
		{
		    /* Marking left */
		    prompt->buf_sel_start = prompt->buf_pos;
		    prompt->buf_sel_end = prompt->buf_sel_origin;
		}
		else 
		{   
		    /* Marking right */
		    prompt->buf_sel_end = prompt->buf_pos;
		    prompt->buf_sel_start = prompt->buf_sel_origin;
		}
	    }

	}
	/* Scroll right? */
	else if(x >= (int)width)
	{
	    const int chars_visible = PROMPT_CHARS_VIS(prompt);
	    const int dc = MAX(
		((x + 1 - (int)width) / PROMPT_CHAR_WIDTH),  
		1
	    );

	    /* Scroll */
	    prompt->buf_vis_pos = MIN(
		prompt->buf_vis_pos + dc,
		MAX(len - chars_visible, 0)
	    );

	    /* Set cursor position */
	    prompt->buf_pos = MIN(
		prompt->buf_vis_pos + chars_visible,
		len - 1
	    );

	    /* Update selection bounds? */
	    if(prompt->selection_in_progress)
	    {
		if(prompt->buf_pos < prompt->buf_sel_origin)
		{
		    /* Marking left */
		    prompt->buf_sel_start = prompt->buf_pos;
		    prompt->buf_sel_end = prompt->buf_sel_origin;
		}
		else 
		{
		    /* Marking right */
		    prompt->buf_sel_end = prompt->buf_pos;
		    prompt->buf_sel_start = prompt->buf_sel_origin;
		}
	    }
	}

	/* Clip cursor and selection positions */
	if(prompt->buf_pos > len)
	    prompt->buf_pos = len;
	if(prompt->buf_pos < 0)
	    prompt->buf_pos = 0;

	if(prompt->buf_sel_end >= len)
	    prompt->buf_sel_end = len - 1;
	if(prompt->buf_sel_end < 0)
	    prompt->buf_sel_end = 0;

	if(prompt->buf_sel_start >= len)
	    prompt->buf_sel_start = len - 1;
	if(prompt->buf_sel_start < 0)
	    prompt->buf_sel_start = 0;

	if((prev_vis_pos != prompt->buf_vis_pos) ||
	   (prev_pos != prompt->buf_pos)
	)
	    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);

	/* Schedual next timeout */
	PromptAddScrollTimeout(
	    prompt, wglobal->prompt_scroll_repeat_int
	);
}


/*
 *      Scheduals a scroll timeout.  
 */
static void PromptAddScrollTimeout(
	prompt_struct *prompt, const time_t interval_ms
)
{
	if(prompt == NULL)
	    return;

	WidgetTimeoutAdd(
	    PromptScrollTimeoutCB,
	    prompt, NULL,
	    interval_ms
	);
}

/* 
 *	Removes any scroll timeouts for the specified prompt.
 */
static void PromptRemoveScrollTimeout(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return;

	WidgetTimeoutRemoveByFuncWidget(
	    PromptScrollTimeoutCB, prompt
	);
}


/*
 *	Selects the prompt's text.
 *
 *	The cursor_pos can be one of the following:
 *
 *	PROMPT_POS_CUR
 *	PROMPT_POS_START
 *	PROMPT_POS_END
 */
void PromptMarkBuffer(prompt_struct *prompt, int cursor_pos)
{
	int len, chars_visible;
	const char *buf;

	if(prompt == NULL)
	    return;

	buf = prompt->buf;
	len = STRLEN(buf);
	if(len <= 0)
	    return;
	if(len > prompt->buf_len)
	    len = prompt->buf_len;

	/* Select all the text */
	prompt->buf_sel_origin = 0;
	prompt->buf_sel_start = 0;
	prompt->buf_sel_end = len;

	/* Store the text to the DDE */
	PromptDDEStoreASCIISeg(prompt->text_area, buf, len);

	switch(cursor_pos)
	{
	  case PROMPT_POS_START:
	    prompt->buf_pos = 0;
	    break;
	  case PROMPT_POS_END:
	    /* Calculate total characters visable on text area */
	    chars_visible = PROMPT_CHARS_VIS(prompt);
	    prompt->buf_vis_pos = MAX(len - chars_visible, 0);
	    prompt->buf_pos = len;
	    break;
	  case PROMPT_POS_CUR:
	    break;
	}
}

/*
 *	Unselects the prompt's text.
 *
 *	The cursor_pos can be one of the following:
 *
 *	PROMPT_POS_CUR
 *	PROMPT_POS_START
 *	PROMPT_POS_END
 */
void PromptUnmarkBuffer(prompt_struct *prompt, int cursor_pos)
{
	int len, chars_visible;
	const char *buf;

	if(prompt == NULL)
	    return;

	buf = prompt->buf;
	len = STRLEN(buf);
	if(len <= 0)
	    return;
	if(len > prompt->buf_len)
	    len = prompt->buf_len;

	/* Unselect the text */
	prompt->buf_sel_origin = -1;
	prompt->buf_sel_start = -1;
	prompt->buf_sel_end = -1;

	switch(cursor_pos)
	{
	  case PROMPT_POS_START:
	    prompt->buf_pos = 0;     
	    break;
	  case PROMPT_POS_END:
	    /* Calculate total characters visable on text area */
	    chars_visible = PROMPT_CHARS_VIS(prompt);
	    prompt->buf_vis_pos = MAX(len - chars_visible, 0);
	    prompt->buf_pos = len;
	    break;
	  case PROMPT_POS_CUR:
	    break;
	}
}


/*
 *	Get float value from prompt.
 */
float PromptGetF(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return(0);

	if(prompt->buf == NULL)
	    return(0);
	    
	return(atof(prompt->buf));
}

/*
 *	Get int value from prompt.
 */
int PromptGetI(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return(0);

	if(prompt->buf == NULL)
	    return(0);

	return(atoi(prompt->buf));
}

/*
 *      Get long value from prompt.
 */
long PromptGetL(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return(0);
	if(prompt->buf == NULL)
	    return(0);

	return(atol(prompt->buf));
}

/*      
 *	Get pointer to prompt's buffer (can return NULL).
 */
char *PromptGetS(prompt_struct *prompt)
{
	return((prompt != NULL) ? prompt->buf : NULL);
}


/*
 *	Set prompt to float value, redraws if mapped.
 */
void PromptSetF(prompt_struct *prompt, float val)
{
	char buf[40];
	sprintf(buf, "%f", val);
	PromptSetS(prompt, buf);
}

/*
 *      Set prompt to float value with the specified number of
 *	decimal places, redraws if mapped.
 */
void PromptSetF2(
	prompt_struct *prompt,
	const float val, const unsigned int ndecimals
)
{
	char fmt[40], buf[40];
	sprintf(fmt, "%%.%if", (int)MIN(ndecimals, 20));
	sprintf(buf, fmt, val);
	PromptSetS(prompt, buf);
}

/*
 *	Set prompt to int value, redraws if mapped.
 */
void PromptSetI(prompt_struct *prompt, int val)
{
	char buf[40];
	sprintf(buf, "%i", val);
	PromptSetS(prompt, buf);
}

/*
 *      Set prompt to long value, redraws if mapped.
 */
void PromptSetL(prompt_struct *prompt, long val)
{
	char buf[40];
	sprintf(buf, "%ld", val);
	PromptSetS(prompt, buf);
}

/*
 *      Set prompt to unsigned long value, redraws if mapped.
 */
void PromptSetUL(prompt_struct *prompt, unsigned long val)
{
	char buf[40];
	sprintf(buf, "%lu", val);
	PromptSetS(prompt, buf);
}

/*
 *	Coppies string value prompt buffer, redraws if mapped.
 */
void PromptSetS(prompt_struct *prompt, const char *val) 
{
	if(prompt == NULL)
	    return;

	if(prompt->buf == NULL)
	    return;

	if(val == NULL)
	    val = "";

	prompt->buf_vis_pos = 0;
	prompt->buf_pos = 0;
	strncpy(prompt->buf, val, prompt->buf_len);

	if(prompt->map_state)
	    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
}


/*
 *	Initializes the prompt.
 */
int PromptInit(
	prompt_struct *prompt, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	int style,
	const char *label,
	unsigned int buf_len, int hist_bufs
)
{
	int i;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if((prompt == NULL) || (parent == None))
	    return(-2);

	if(width == 0)
	    width = PROMPT_DEF_WIDTH;
	if(height == 0)
	    height = PROMPT_DEF_HEIGHT;

	prompt->type = WIDGET_TYPE_PROMPT;
	prompt->map_state = False;
	prompt->has_focus = False;
	prompt->disabled = False;
	prompt->visibility_state = VisibilityFullyObscured;
	prompt->x = x;
	prompt->y = y;
	prompt->width = MAX((int)width, PROMPT_MIN_WIDTH);
	prompt->height = MAX((int)height, PROMPT_MIN_HEIGHT);
	prompt->font = wglobal->prompt_label_font;
	prompt->next = NULL;
	prompt->prev = NULL;
	prompt->data = NULL;

	prompt->options = 0;
	prompt->style = style;

	prompt->label = STRDUP(label);

	prompt->text_font = wglobal->prompt_text_font;
	prompt->buf = NULL;
	prompt->buf_len = 0;
	prompt->buf_pos = 0;
	prompt->buf_vis_pos = 0;

	prompt->selection_in_progress = False;
	prompt->buf_sel_origin = -1;
	prompt->buf_sel_start = -1;
	prompt->buf_sel_end = -1;

	prompt->hist_buf = NULL;
	prompt->total_hist_bufs = 0;
	prompt->hist_buf_pos = 0;

	prompt->button_pressed = 0;
	prompt->button_pressed_count = 0;
	prompt->button_pressed_time = 0l;

	prompt->activate_cb = NULL;
	prompt->activate_data = NULL;
	prompt->cancel_cb = NULL;
	prompt->cancel_data = NULL;

	/* Create toplevel */
	x = prompt->x;
	y = prompt->y;
	width = prompt->width;
	height = prompt->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	prompt->toplevel = w;
	prompt->toplevel_buf = None;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    VisibilityChangeMask
	);
	parent = w;

	/* Create text area */
	x = 0;
	y = 0;
	width = prompt->width;
	height = prompt->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	prompt->text_area = w;
	prompt->text_area_buf = None;

	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    KeyPressMask | KeyReleaseMask |
	    ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask
	);
	WidgetSetWindowCursor(w, wglobal->text_wcur);

	/* Set the label */
	prompt->label = STRDUP(label);

	/* Create the text buffer */
	prompt->buf_len = buf_len;
	if(prompt->buf_len > 0)
	{
	    prompt->buf = (char *)calloc(
		prompt->buf_len + 1,
		sizeof(char)
	    );
	    if(prompt->buf == NULL)
		prompt->buf_len = 0;
	}
	else
	{
	    prompt->buf = NULL;
	}

	/* History buffers */
	prompt->total_hist_bufs = hist_bufs;
	if((prompt->total_hist_bufs > 0) && (prompt->buf_len > 0))
	{
	    prompt->hist_buf = (char **)calloc(
		prompt->total_hist_bufs,
		sizeof(char *)
	    );
	    if(prompt->hist_buf == NULL)
	    {
	        prompt->total_hist_bufs = 0;
		return(-3);
	    }
	    for(i = 0; i < prompt->total_hist_bufs; i++)
	    {
		prompt->hist_buf[i] = (char *)calloc(
		    prompt->buf_len + 1,
		    sizeof(char)
		);
		if(prompt->hist_buf[i] == NULL)
		{
		    return(-3);
		}
	    }
	}
	else
	{
	    prompt->hist_buf = NULL;
	}

	/* Resize prompt initially */
	PromptResize(prompt);

	/* Add prompt to wdiget regeristry */
	WidgetRegister(prompt, prompt->type);

	return(0);
}

/*
 *	Gets the prompt's label.
 */
const char *PromptGetLabel(prompt_struct *prompt)
{
	return((prompt != NULL) ? prompt->label : NULL);
}

/*
 *	Sets the prompt's label.
 */
void PromptSetLabel(prompt_struct *prompt, const char *label)
{
	if(prompt == NULL)
	    return;

	if(prompt->label != label)
	{
	    free(prompt->label);
	    prompt->label = STRDUP(label);

	    PromptQueueResize(prompt);
	    if(prompt->map_state)
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_COMPLETE);
	}
}

/*
 *	Sets the prompt's hint message
 */
void PromptSetTooltip(
	prompt_struct *prompt, const char *msg   
)
{
	if(prompt == NULL)
	    return;

	TooltipsSetMessage(NULL, prompt->text_area, msg);
}

/*
 *	Resizes the prompt.
 */
void PromptResize(prompt_struct *prompt)
{
	int x, y, len, chars_visible, new_vis_pos;
	unsigned int width, height;
	win_t w;

	if(prompt == NULL)
	    return;

	w = prompt->toplevel;
	OSWGetWindowGeometry(w,  &x, &y, &width, &height, NULL);
	prompt->x = x;
	prompt->y = y;
	prompt->width = width;
	prompt->height = height;

	OSWDestroyPixmap(&prompt->toplevel_buf);

	/* Text area */
	if((prompt->style == PROMPT_STYLE_FLUSHED) ||
	   (prompt->style == PROMPT_STYLE_RAISED)
	)
	{
	    if((prompt->font != NULL) && !STRISEMPTY(prompt->label))
	    {
		int swidth;

		OSWGetTextBounds(
		    prompt->font, prompt->label, -1,
		    NULL, NULL, &swidth, NULL, NULL
		);

		x = swidth + (2 * PROMPT_MARGIN);
		y = PROMPT_MARGIN;
		width = (unsigned int)MAX(
		    (int)prompt->width - x - PROMPT_MARGIN,
		    2
		);
		height = (unsigned int)MAX(
		    (int)prompt->height - (2 * PROMPT_MARGIN),
		    2
		);
	    }
	    else
	    {
		x = PROMPT_MARGIN;
		y = PROMPT_MARGIN;
		width = (unsigned int)MAX(
		    (int)prompt->width - (2 * PROMPT_MARGIN),
		    2
		);
		height = (unsigned int)MAX(
		    (int)prompt->height - (2 * PROMPT_MARGIN),
		    2
		);   
	    }
	}
	else	/* PROMPT_STYLE_NOBORDER */
	{
	    x = 0;
	    y = 0;
	    width = prompt->width;
	    height = prompt->height;
	}
	OSWMoveResizeWindow(prompt->text_area, x, y, width, height);
	OSWDestroyPixmap(&prompt->text_area_buf);

	/* Make sure that the scroll and cursor positions are visible */
	len = STRLEN(prompt->buf);
	chars_visible = PROMPT_CHARS_VIS(prompt);
	new_vis_pos = MAX(len - chars_visible, 0);
	if(new_vis_pos < prompt->buf_vis_pos)
	    prompt->buf_vis_pos = new_vis_pos;

	if(prompt->buf_pos >= (prompt->buf_vis_pos + chars_visible))
	    prompt->buf_pos = prompt->buf_vis_pos + chars_visible - 1;
	if(prompt->buf_pos < prompt->buf_vis_pos)
	    prompt->buf_vis_pos = 0;
}

/*
 *	Queues the prompt to resize.
 */
void PromptQueueResize(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return;

	WidgetQueueResizeAction(PromptResizeCB, prompt, NULL);
}

/*
 *	Infocuses and draws all prompts except the specified
 *	prompt except_prompt.
 */
static void PROMPT_UNFOCUS_REDRAW_ALL(prompt_struct *except_prompt)
{
	int i;
	prompt_struct *prompt;
	widget_register_struct *reg;
	widget_global_struct *wglobal = &widget_global;

	for(i = 0; i < wglobal->nwidgets; i++)
	{
	    reg = wglobal->widget[i];
	    if(reg == NULL)
		continue;
   
	    if((reg->type != WIDGET_TYPE_PROMPT) ||
	       (reg->widget == NULL) ||
	       (reg->widget == (void *)except_prompt)
	    )
		continue;

	    prompt = (prompt_struct *)reg->widget;

	    /* Unfocus */
	    prompt->has_focus = False;

	    /* Redraw if already mapped */
	    if(prompt->map_state)
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_COMPLETE);
	}
}

/*
 *	Draws the prompt.
 */
void PromptDraw(prompt_struct *prompt, int amount)
{
	pixmap_t pixmap;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(prompt == NULL)
	    return;

	/* Recreate buffers as needed */
	if(prompt->toplevel_buf == None)
	{
	    if(OSWCreatePixmap(
		&prompt->toplevel_buf, prompt->width, prompt->height
	    ))
		return;
	}
	if(prompt->text_area_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		prompt->text_area, NULL, NULL, &width, &height, NULL
	    );
	    if(OSWCreatePixmap(
		&prompt->text_area_buf, width, height
	    ))
		return;
	}


	/* Draw background and name label */
	if((amount == PROMPT_DRAW_AMOUNT_COMPLETE) &&
	   (prompt->style != PROMPT_STYLE_NOBORDER)
	)
	{
	    unsigned int width, height;

	    w = prompt->toplevel;
	    pixmap = prompt->toplevel_buf;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    if(wglobal->force_mono)
		OSWClearPixmap(
		    pixmap, width, height,
		    gui->black_pix
		);
	    else
	        WidgetPutImageTile(
		    pixmap,
		    wglobal->std_bkg_img,
		    width, height
	        );

	    /* Draw label */
	    if(prompt->label != NULL)
	    {
		int lbearing, rbearing, ascent, descent;
		font_t *font = prompt->font;

	        if(wglobal->force_mono)
	            OSWSetForeground(gui->white_pix);
	        else
		    OSWSetForeground(
			prompt->disabled ?
			    wglobal->disabled_text_pix :
			    wglobal->normal_text_pix
		    );

		OSWSetFont(font);
		OSWGetTextBounds(
		    font, prompt->label, -1,
		    &lbearing, &rbearing,
		    NULL,
		    &ascent, &descent
		);
	        OSWDrawString(
		    pixmap,
		    (2 * PROMPT_MARGIN) - lbearing,
		    (((int)height - (ascent + descent)) / 2) + ascent,
		    prompt->label
	        );
	    }

	    /* Draw frame on toplevel */
	    if(prompt->style == PROMPT_STYLE_RAISED)
	    {
		WidgetFrameButtonPixmap(
		    pixmap, False,
		    width, height,
		    (wglobal->force_mono) ? gui->white_pix :
			wglobal->surface_highlight_pix,
		    (wglobal->force_mono) ? gui->white_pix :
			wglobal->surface_shadow_pix
		);
	    }

	    OSWPutDrawableToWindow(w, pixmap);
	}

	/* Draw text area */
	if(True)
	{
	    int	i, j, x, len,
		chars_visible, chars_drawn,
		lbearing, rbearing, swidth, ascent, descent;
	    unsigned int width, height; 
	    font_t *font;

	    w = prompt->text_area;
	    pixmap = prompt->text_area_buf;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Draw background */
	    OSWClearPixmap(
		pixmap, width, height,
		wglobal->force_mono ?
		    gui->black_pix : wglobal->surface_editable_pix
	    );

	    /* Calculate the number of characters visible plus 1
	     * to ensure any partially-off edge will be drawn
	     */
	    chars_visible = MAX(
	        ((int)width - (PROMPT_MARGIN * 2)) / PROMPT_CHAR_WIDTH,
	        0
	    ) + 1;

	    /* Get length of buffer */
	    len = STRLEN(prompt->buf);

	    /* Sanitize selection range */
	    if(prompt->buf_sel_start > prompt->buf_sel_end)
	    {
	        prompt->buf_sel_start = -1;
	        prompt->buf_sel_end = -1;
	    }
	    if(prompt->buf_sel_start >= len)
	    {
	        prompt->buf_sel_start = -1;
	        prompt->buf_sel_end = -1;
	    }
	    if(prompt->buf_sel_end >= len)
	    {
		prompt->buf_sel_end = MAX(len - 1, 0);
	    }

	    /* Clip the buffer scroll position */
	    if((prompt->buf_len - chars_visible) >= 0)
	        prompt->buf_vis_pos = MIN(
		    prompt->buf_vis_pos,
		    (prompt->buf_len - chars_visible)
		);
	    else
		prompt->buf_vis_pos = 0;

	    /* Draw text */
	    i = 0;		/* Buffer index */
	    j = 0;		/* Current buffer segment length */
	    x = PROMPT_MARGIN;
	    chars_drawn = 0;
	    /* Is a segment of the text selected? */
	    if((prompt->buf_sel_start >= 0) &&
	       (prompt->buf_sel_end >= 0) &&
	       (prompt->buf_sel_start <= prompt->buf_sel_end)
	    )
	    {
	        /* Drawing with selection */

	        /* Draw the first unselected segment */
	        i = prompt->buf_vis_pos;
	        j = MAX(
		    prompt->buf_sel_start - prompt->buf_vis_pos,
		    0
	        );
	        if(j >= (chars_visible - chars_drawn))
		    j = chars_visible - chars_drawn - 1;
	        if((i + j) >= prompt->buf_len)
		    j = (prompt->buf_len - 1) - i;

	        if(INRANGEINC(i, 0, prompt->buf_len - 1) &&
	           INRANGEINC(i + j, 0, prompt->buf_len - 1) &&
	           (j > 0) &&
	           (prompt->buf != NULL) &&
	           (chars_drawn <= (int)chars_visible)
	        )
	        {
		    if(wglobal->force_mono)
		        OSWSetForeground(gui->white_pix);
		    else
		        OSWSetForeground(
			    prompt->disabled ?
				wglobal->disabled_text_pix :
				wglobal->editable_text_pix
			);

		    font = prompt->text_font;
		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, prompt->buf + i, j,
			&lbearing, &rbearing,
			&swidth,
			&ascent, &descent
		    );
		    OSWDrawStringLimited(
			pixmap,
			x - lbearing,
			(((int)height - (ascent + descent)) / 2) + ascent,
			prompt->buf + i, j
		    );
		    x += swidth - lbearing;
		}
		chars_drawn += j;

		/* Draw the selected segment */
		i = MAX(prompt->buf_sel_start, prompt->buf_vis_pos);
		j = MAX(
		    (prompt->buf_sel_end - i) + 1,
		    0
	        );
	        if(j > ((int)chars_visible - (int)chars_drawn))
		    j = (int)chars_visible - (int)chars_drawn;
	        if((i + j) >= prompt->buf_len)
		    j = (prompt->buf_len - 1) - i;

		if(INRANGEINC(i, 0, prompt->buf_len - 1) &&
		   INRANGEINC(i + j, 0, prompt->buf_len - 1) &&
		   (j > 0) &&
		   (prompt->buf != NULL) &&
		   (chars_drawn <= chars_visible)
	        )
	        {
		    if(wglobal->force_mono)
		        OSWSetForeground(gui->white_pix);
		    else
		        OSWSetForeground(
			    prompt->disabled ?
				wglobal->disabled_text_pix :
				wglobal->surface_selected_pix
			);

		    font = prompt->text_font;
		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, prompt->buf + i, j,
			&lbearing, &rbearing,
			&swidth,
			&ascent, &descent
		    );

		    /* Apply lbearing from swidth if the first
		     * unselected segment was not drawn
		     */
		    if(chars_drawn == 0)
			swidth -= lbearing;

		    OSWDrawSolidRectangle(
			pixmap,
			x,
			PROMPT_MARGIN,
			swidth,
			(int)height - (2 * PROMPT_MARGIN)
		    );

		    if(wglobal->force_mono)
			OSWSetForeground(gui->black_pix);
		    else
			OSWSetForeground(
			    prompt->disabled ?
				wglobal->selected_text_pix :
				wglobal->selected_text_pix
			);
		    OSWDrawStringLimited(
			pixmap,
			x - ((chars_drawn > 0) ? 0 : lbearing),
			(((int)height - (ascent + descent)) / 2) + ascent,
		        prompt->buf + i, j
		    );
		    x += swidth;
	        }
		chars_drawn += j;

		/* Draw the last unselected segment */
		i = MAX(prompt->buf_sel_end + 1, 0);
		j = MAX(chars_visible - chars_drawn, 0);
		if((i + j) > len)
		    j = len - i;

		if(INRANGEINC(i, 0, prompt->buf_len - 1) &&
		   INRANGEINC(i + j, 0, prompt->buf_len - 1) &&
		   (j > 0) &&
		   (prompt->buf != NULL) &&
		   (chars_drawn <= chars_visible)
	        )
	        {
		    font = prompt->text_font;
		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, prompt->buf + i, j,
			&lbearing, &rbearing,    
			&swidth,
			&ascent, &descent
		    );
		    if(wglobal->force_mono)
		        OSWSetForeground(gui->white_pix);
		    else
		        OSWSetForeground(wglobal->editable_text_pix);
		    OSWDrawStringLimited(
			pixmap,
		        x - ((chars_drawn > 0) ? 0 : lbearing),
		        (((int)height - (ascent + descent)) / 2) + ascent,
			prompt->buf + i, j
		    );
	        }
	    }
	    else
	    {
		/* No segment of the text is selected */
	        i = prompt->buf_vis_pos;
	        j = MIN(chars_visible, (int)strlen(prompt->buf + i));
	        if((i + j) >= prompt->buf_len)
		    j = (prompt->buf_len - 1) - i;

	        if(INRANGEINC(i, 0, prompt->buf_len - 1) &&
	           INRANGEINC(i + j, 0, prompt->buf_len - 1) &&
	           (j > 0) &&
	           (prompt->buf != NULL) &&
	           (chars_drawn <= chars_visible)
	        )
	        {
		    font = prompt->text_font;
		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, prompt->buf + i, j,
			&lbearing, &rbearing,    
			&swidth,
			&ascent, &descent
		    );
		    if(wglobal->force_mono)
			OSWSetForeground(gui->white_pix);
		    else
		        OSWSetForeground(
			    prompt->disabled ?
				wglobal->disabled_text_pix :
				wglobal->editable_text_pix
			);
		    OSWDrawStringLimited(
			pixmap,
			x - ((chars_drawn > 0) ? 0 : lbearing),
			(((int)height - (ascent + descent)) / 2) + ascent,
			prompt->buf + i, j
		    );
	        }
	    }

	    /* Draw text cursor if in focus and no selection */
	    if(INRANGEINC(prompt->buf_pos, 0, prompt->buf_len - 1) &&
	       prompt->has_focus &&
	       (prompt->buf_sel_start < 0)
	    )
	    {
		int y;

		x = ((prompt->buf_pos - prompt->buf_vis_pos) *
		    PROMPT_CHAR_WIDTH) + PROMPT_MARGIN;

	        if(wglobal->force_mono)
		    OSWSetForeground(gui->white_pix);
	        else
	            OSWSetForeground(
			prompt->disabled ?
			    wglobal->disabled_text_pix :
			    wglobal->editable_text_pix
		    );
	        OSWDrawLine(
		    pixmap,
		    x, PROMPT_MARGIN + 1,
		    x, (int)height - PROMPT_MARGIN - 1
	        );
		y = PROMPT_MARGIN;
	        OSWDrawLine(
		    pixmap,
		    x - 2, y,
		    x + 2, y
	        );
		y = (int)height - PROMPT_MARGIN - 1;
	        OSWDrawLine(
		    pixmap,
		    x - 2, y,
		    x + 2, y
	        );
	    }

	    /* Draw the frame around the text area */
	    if(prompt->style != PROMPT_STYLE_NOBORDER)
	    {
	        WidgetFrameButtonPixmap(
		    pixmap, True, width, height,
		    wglobal->force_mono ? gui->white_pix :
			wglobal->surface_highlight_pix,
		    wglobal->force_mono ? gui->white_pix :
			wglobal->surface_shadow_pix
	        );
	    }

	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the prompt to draw.
 */
void PromptQueueDraw(prompt_struct *prompt, int amount)
{
	if(prompt == NULL)
	    return;

	WidgetQueueDrawAction(PromptDrawCB, prompt, (void *)amount);
}

/*
 *	Manages the prompt event.
 */
int PromptManage(prompt_struct *prompt, event_t *event)
{
	int events_handled = 0;
	win_t w;
	keycode_t keycode;
	keyval_t keyval;
	key_state_t key_state;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	widget_global_struct *wglobal = &widget_global;

	if((prompt == NULL) || (event == NULL))
	    return(events_handled);

	if(!prompt->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == prompt->toplevel)
	    {
		if(prompt->map_state)
		    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == prompt->toplevel)
	    {
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    else if(w == prompt->text_area)
	    {
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!prompt->has_focus)
		break;

	    TooltipsUnmap(NULL);

	    /* Skip modifier keys */
	    keycode = key->keycode;
	    if(OSWIsModifierKey(keycode))
		break;

	    /* Skip if there is text buffer */
	    if((prompt->buf == NULL) || (prompt->buf_len <= 0))
		break;

	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Clip position */
	    if(prompt->buf_pos < 0)
		prompt->buf_pos = 0;
	    else if(prompt->buf_pos >= prompt->buf_len)
		prompt->buf_pos = prompt->buf_len - 1;

	    /* Tab */
	    if((keyval == OSW_KEY_Tab) ||
	       (keyval == OSW_KEY_KP_Tab) ||
	       (keyval == OSW_KEY_ISO_Left_Tab)
	    )
	    {
		prompt_struct *prompt2 = (key_state & OSW_KEY_STATE_SHIFT) ?
		    (prompt_struct *)prompt->prev :
		    (prompt_struct *)prompt->next;
		if((prompt2 != NULL) ?
		   WidgetIsRegistered(prompt2) : False
		)
		{
		    prompt->has_focus = False;
		    prompt2->has_focus = True;
		    PromptMarkBuffer(
			prompt2,
			PROMPT_POS_END
		    );
		    if(prompt2->map_state)
			PromptQueueDraw(prompt2, PROMPT_DRAW_AMOUNT_COMPLETE);
		}
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    /* BackSpace */
	    else if((keyval == OSW_KEY_BackSpace) &&
		    !prompt->disabled
	    )
	    {
		int	len = MIN(
			    STRLEN(prompt->buf),
			    prompt->buf_len
			),
			chars_visible = PROMPT_CHARS_VIS(prompt);

		/* Selected? */
		if((prompt->buf_sel_start > -1) &&
		   (prompt->buf_sel_end > -1)
		)
		{
		    /* Delete selected text */
		    int x = MIN((int)prompt->buf_sel_start, len),
			y = MIN((int)prompt->buf_sel_end + 1, len);
		    char *buf = prompt->buf;
		    while(y < len)
		    {
			if(buf[y] == '\0')
			    break;

			buf[x] = buf[y];

			x++;
			y++;
		    }
		    buf[x] = '\0';	/* Null terminate */

		    /* Get new length */
		    len = MIN(
			STRLEN(prompt->buf),
			prompt->buf_len
		    );

		    /* Move cursor to position where selection started */
		    prompt->buf_pos = CLIP(
			prompt->buf_sel_start,
			0, len
		    );

		    /* Scroll as needed */
		    if((prompt->buf_vis_pos + chars_visible) > len)
			prompt->buf_vis_pos = MAX(
			    (len - chars_visible), 0
			);

		    /* Clear selection */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;
		}
		else if((prompt->buf_pos > 0) &&
		        (prompt->buf_pos <= len)
		)
		{
		    /* Erase previous character */
		    int i, m = prompt->buf_len;
		    char *buf = prompt->buf;

		    /* Adjust position */
		    prompt->buf_pos -= 1;

		    /* Scroll as needed */
		    if(prompt->buf_pos < prompt->buf_vis_pos)
			prompt->buf_vis_pos = prompt->buf_pos;

		    /* Shift all characters */
		    for(i = prompt->buf_pos; i < m; i++)
		    {
			if(buf[i] == '\0')
			    break;

			buf[i] = buf[i + 1];
		    }
		}
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    /* Delete a character */
	    else if(((keyval == OSW_KEY_Delete) ||
		    (!(key_state & OSW_KEY_STATE_NUM_LOCK) &&
		     (keyval == OSW_KEY_KP_Delete))
		    ) && !prompt->disabled
	    )
	    {
		int	len = MIN(
				STRLEN(prompt->buf),
				prompt->buf_len
			),
			chars_visible = PROMPT_CHARS_VIS(prompt);

		/* Selected? */
		if((prompt->buf_sel_start > -1) &&
		   (prompt->buf_sel_end > -1)
		)
		{
		    /* Delete selected text */
		    int	x = MIN(prompt->buf_sel_start, len),
			y = MIN(prompt->buf_sel_end + 1, len);
		    char *buf = prompt->buf;
		    while(y < len)
		    {
			if(buf[y] == '\0')
			    break;

			buf[x] = buf[y];

			x++;
			y++;
		    }
		    buf[x] = '\0';	/* Null terminate */

		    /* Get new length */
		    len = MIN(
			STRLEN(prompt->buf),
			prompt->buf_len
		    );

		    /* Move cursor to position where selection started */
		    prompt->buf_pos = CLIP(
			prompt->buf_sel_start,
			0, len
		    );

		    /* Scroll as needed */
		    if((prompt->buf_vis_pos + chars_visible) > len)
			prompt->buf_vis_pos = MAX(
			    (len - chars_visible), 0
			);

		    /* Clear selection */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;
		}
		else if((prompt->buf_pos >= 0) &&
		        (prompt->buf_pos < len)
		)
		{
		    /* Erase next character */
		    int i, m = prompt->buf_len;
		    char *buf = prompt->buf;

		    /* Shift all characters */
		    for(i = prompt->buf_pos; i < m; i++)
		    {
			if(buf[i] == '\0')
			    break;

			buf[i] = buf[i + 1];
		    }
		}
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    /* Enter */
	    else if(((keyval == OSW_KEY_Return) ||
		     (keyval == OSW_KEY_KP_Enter) ||
		     (keyval == OSW_KEY_ISO_Enter)
		    ) &&
		    !prompt->disabled
	    )
	    {
		char *value;

		/* Record buffer to history */
		if((prompt->hist_buf != NULL) &&
		   (prompt->total_hist_bufs > 0) &&
		   (prompt->buf_len > 0)
		)
		{
		    int i;

		    /* Shift the history buffers */
		    for(i = prompt->total_hist_bufs - 1; i > 0; i--)
		    {
			if((prompt->hist_buf[i] == NULL) ||
			   (prompt->hist_buf[i - 1] == NULL)
			)
			    continue;

			memcpy(
			    prompt->hist_buf[i],
			    prompt->hist_buf[i - 1],
			    prompt->buf_len + 1	/* Yes that much is allocated */
			);
		    }

		    if(prompt->buf != NULL)
		        memcpy(
			    prompt->hist_buf[0], prompt->buf,
			    prompt->buf_len + 1
			);

		    /* Adjust history buffer position to -1 */
		    prompt->hist_buf_pos = -1;
		}

		/* Make a copy of the value for passing to the
		 * activate callback so that any changes to the prompt
		 * made by the activate callback will not modify the
		 * the value that was passed to it
		 */
		value = STRDUP(prompt->buf);

		/* Clear the value? */
		if(!(prompt->options & PROMPT_FLAG_KEEP_VALUE))
		{
		    memset(
			prompt->buf, 0x00,
			prompt->buf_len + 1
		    );
		    prompt->buf_pos = 0;
		    prompt->buf_vis_pos = 0;
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;
		}

		/* Call the activate callback */
		if(prompt->activate_cb != NULL)
		    prompt->activate_cb(   
			prompt,
			value,
			prompt->activate_data
		    );

		if(!WidgetIsRegistered(prompt))
		{
		    free(value);
		    events_handled++;
		    return(events_handled);
		}

		free(value);	/* Delete the coppied value */

		events_handled++;
	    }
	    /* Cancel */
	    else if((keyval == OSW_KEY_Escape) &&
		    !prompt->disabled
	    )
	    {
		/* Call the cancel callback */
		if(prompt->cancel_cb != NULL)
		    prompt->cancel_cb(
			prompt,
			prompt->cancel_data
		    );

		events_handled++;
	    }
	    /* Cursor up */
	    else if(((keyval == OSW_KEY_Up) ||
		    (!(key_state & OSW_KEY_STATE_NUM_LOCK) &&
		     ((keyval == OSW_KEY_KP_Up) ||
		      (keyval == OSW_KEY_KP_Page_Up)
		    ))
		    ) && !prompt->disabled
	    )
	    {
		/* Recall older history buffer */
		if(prompt->total_hist_bufs > 0)
		{
		    /* Change last recalled history buffer position */
		    prompt->hist_buf_pos++;
		    if(prompt->hist_buf_pos < 0)
		        prompt->hist_buf_pos = 0;
		    if(prompt->hist_buf_pos >= prompt->total_hist_bufs)
		        prompt->hist_buf_pos = prompt->total_hist_bufs - 1;

		    /* Recall that history buffer */
		    if(prompt->hist_buf_pos < prompt->total_hist_bufs)
		    {
			memcpy(
			    prompt->buf,
			    prompt->hist_buf[prompt->hist_buf_pos],
			    prompt->buf_len
			);
			prompt->buf_pos = strlen(prompt->buf);
		    }
		    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		    events_handled++;
		}

	    }
	    /* Cursor down */
	    else if(((keyval == OSW_KEY_Down) ||
		     (!(key_state & OSW_KEY_STATE_NUM_LOCK) &&
		      ((keyval == OSW_KEY_KP_Down) ||
		       (keyval == OSW_KEY_KP_Page_Down)
		     ))
		    ) && !prompt->disabled
	    )
	    {
		/* Recall newer history buffer */
		if(prompt->total_hist_bufs > 0)
		{
		    /* Change last recalled history buffer position */
		    prompt->hist_buf_pos--;
		    if(prompt->hist_buf_pos < 0)
			prompt->hist_buf_pos = 0;
		    if(prompt->hist_buf_pos >= prompt->total_hist_bufs)
			prompt->hist_buf_pos = prompt->total_hist_bufs - 1;

		    /* Recall that history buffer */
		    if(prompt->hist_buf_pos < prompt->total_hist_bufs)
		    {
		        memcpy(
			    prompt->buf,
			    prompt->hist_buf[prompt->hist_buf_pos],
			    prompt->buf_len
		        );
		        prompt->buf_pos = strlen(prompt->buf);
		    }
		    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		    events_handled++;
		}

	    }
	    /* Cursor left */
	    else if((keyval == OSW_KEY_Left) ||
		    (!(key_state & OSW_KEY_STATE_NUM_LOCK) &&
		     (keyval == OSW_KEY_KP_Left))
	    )
	    {
		/* Check if shift key is pressed */
		if((prompt->buf_pos > 0) && (key_state & OSW_KEY_STATE_SHIFT))
		{
		    /* Set select marks */
		    int sel_len;

		    /* Record origin if previously not marked */
		    if((prompt->buf_sel_start < 0) &&
		       (prompt->buf_sel_end < 0)
		    )
		    {
			/* Previously not marked, so make it so that
			 * only the current character gets marked
			 */
			prompt->buf_pos -= 1;

			prompt->buf_sel_start = prompt->buf_pos;
			prompt->buf_sel_end = prompt->buf_pos;
			prompt->buf_sel_origin = prompt->buf_pos;

			prompt->buf_pos += 1;	/* Need to add 1 */
		    }

		    /* Move the cursor to the left */
		    prompt->buf_pos -= 1;

		    /* Move selection mark position */
		    if(prompt->buf_pos < prompt->buf_sel_origin)
			prompt->buf_sel_start = prompt->buf_pos;
		    else
			prompt->buf_sel_end = prompt->buf_pos;


		    /* Store the bytes of the selected buffer range */
		    sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
		    if(sel_len >= 0)
			PromptDDEStoreASCIISeg(
			    prompt->text_area,
			    prompt->buf + prompt->buf_sel_start,
			    sel_len + 1
			);
		}
		else if(prompt->buf_pos > 0)
		{
		    /* Clear selection */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;

		    /* Move cursor to the right */
		    prompt->buf_pos -= 1;
		}

		/* Scroll as needed */
		if(prompt->buf_pos < prompt->buf_vis_pos)
		    prompt->buf_vis_pos = prompt->buf_pos;

		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    /* Cursor right */
	    else if((keyval == OSW_KEY_Right) ||
		    (!(key_state & OSW_KEY_STATE_NUM_LOCK) &&
		     (keyval == OSW_KEY_KP_Right))
	    )
	    {
		/* Get length of text value in buffer */
		int	len = MIN(
			    STRLEN(prompt->buf),
			    prompt->buf_len
			),
			chars_visible = PROMPT_CHARS_VIS(prompt);

		/* Mark selection right */
		if((prompt->buf_pos < len) && (key_state & OSW_KEY_STATE_SHIFT))
		{
		    int sel_len;

		    /* Record origin if previously not marked */
		    if((prompt->buf_sel_start < 0) &&
		       (prompt->buf_sel_end < 0)
		    )
		    {
			prompt->buf_sel_start = prompt->buf_pos;
			prompt->buf_sel_end = prompt->buf_pos;
			prompt->buf_sel_origin = prompt->buf_pos;

			/* Previously not marked, make it so only
			 * current character gets marked
			 */
			prompt->buf_pos -= 1;
		    }

		    /* Move cursor to the right */
		    prompt->buf_pos += 1;
		    if(prompt->buf_pos >= len)
			prompt->buf_pos = len - 1;

		    /* Move selection mark position */
		    if(prompt->buf_pos <= prompt->buf_sel_origin)
			prompt->buf_sel_start = prompt->buf_pos;
		    else
			prompt->buf_sel_end = prompt->buf_pos;

		    /* Store the bytes of the selected buffer range */
		    sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
		    if(sel_len >= 0)
			PromptDDEStoreASCIISeg(
			    prompt->text_area,
			    prompt->buf + prompt->buf_sel_start,
			    sel_len + 1
			);
		}
		else if(prompt->buf_pos < len)
		{
		    /* Unset selection mark */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;

		    /* Move cursor to the right */
		    prompt->buf_pos += 1;
		}

		/* Scroll as needed */
		if(prompt->buf_pos >=
		   (prompt->buf_vis_pos + chars_visible)
		)
		    prompt->buf_vis_pos = MAX(
			(prompt->buf_pos - chars_visible), 0
		    );

		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    /* Home */
	    else if((keyval == OSW_KEY_Home) ||
		    (!(key_state & OSW_KEY_STATE_NUM_LOCK) &&
		     ((keyval == OSW_KEY_KP_Home) ||
		      (keyval == OSW_KEY_KP_Begin)
		    ))
	    )
	    {
		if(key_state & OSW_KEY_STATE_SHIFT)
		{
		    int sel_len;

		    /* Record origin if previously not marked */
		    if((prompt->buf_sel_start < 0) &&
		       (prompt->buf_sel_end < 0)
		    )
		    {
			prompt->buf_pos -= 1;

			prompt->buf_sel_start = prompt->buf_pos;
			prompt->buf_sel_end = prompt->buf_pos;
			prompt->buf_sel_origin = prompt->buf_pos;
		    }

		    prompt->buf_sel_end = prompt->buf_sel_origin;
		    prompt->buf_sel_start = 0;

		    /* Store the bytes of the selected buffer range */
		    sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
		    if(sel_len >= 0)
			PromptDDEStoreASCIISeg(
			    prompt->text_area,
			    prompt->buf + prompt->buf_sel_start,
			    sel_len + 1
			);
		}
		else
		{
		    /* Clear selection */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;
		}

		/* Set cursor position to 0 */
		prompt->buf_pos = 0;

		/* Scroll as needed */
		if(prompt->buf_pos < prompt->buf_vis_pos)
		    prompt->buf_vis_pos = prompt->buf_pos;

		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    /* End */
	    else if((keyval == OSW_KEY_End) ||
		    (!(key_state & OSW_KEY_STATE_NUM_LOCK) &&
		     (keyval == OSW_KEY_KP_End))
	    )
	    {
		int	len = MIN(
			    STRLEN(prompt->buf),
			    prompt->buf_len
			),
			chars_visible = PROMPT_CHARS_VIS(prompt);

		if(key_state & OSW_KEY_STATE_SHIFT)
		{
		    int sel_len;

		    /* Record origin if previously not marked */
		    if((prompt->buf_sel_start < 0) &&
		       (prompt->buf_sel_end < 0)
		    )
		    {
			prompt->buf_sel_start = prompt->buf_pos;
			prompt->buf_sel_end = prompt->buf_pos;
			prompt->buf_sel_origin = prompt->buf_pos;
		    }

		    prompt->buf_sel_start = prompt->buf_sel_origin;
		    prompt->buf_sel_end = len - 1;

		    /* Store the bytes of the selected buffer range */
		    sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
		    if(sel_len >= 0)
			PromptDDEStoreASCIISeg(
			    prompt->text_area,
			    prompt->buf + prompt->buf_sel_start,
			    sel_len + 1
			);
		}
		else
		{
		    /* Clear selection */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;
		}

		/* Set cursor position to end */
		prompt->buf_pos = len;

		/* Scroll as needed */
		if(prompt->buf_pos > (prompt->buf_vis_pos + chars_visible)) 
		    prompt->buf_vis_pos = MAX(
			(prompt->buf_pos - chars_visible), 0
		    );

		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    /* Shift+insert: Paste DDE buffer */
	    else if(prompt->disabled ?
		False :
		(((keyval == OSW_KEY_Insert) &&
		  (key_state & OSW_KEY_STATE_SHIFT)) ||
		 ((keyval == OSW_KEY_v) &&
		  (key_state & OSW_KEY_STATE_CTRL))
		)
	    )
	    {
		char *tmp_buf = OSWDDEGetString(prompt->text_area);
		const int tmp_buf_len = STRLEN(tmp_buf);

		int	len = MIN(
			    STRLEN(prompt->buf),
			    prompt->buf_len
			),
			chars_visible;

		/* Unmark selection */
		prompt->buf_sel_origin = -1;
		prompt->buf_sel_start = -1;
		prompt->buf_sel_end = -1; 

		/* Sanitize position. (Remember we have one extra byte.) */
		if(prompt->buf_pos > len)
		    prompt->buf_pos = len;
		if(prompt->buf_pos < 0)
		    prompt->buf_pos = 0;

		/* Check if there is text selection marked */
		if((tmp_buf != NULL) &&
		   (tmp_buf_len > 0)
		)
		{
		    /* Shift tailing segment of prompt buffer */
		    int	x = len,
			y = len + tmp_buf_len,
			z;
		    unsigned int width, height;

		    if(y > prompt->buf_len)
		    {
			z = y - prompt->buf_len;
			y = MAX(y - z, 0);
			x = MAX(x - z, 0);
		    }

		    while(x >= prompt->buf_pos)
		    {
			prompt->buf[y] = prompt->buf[x];
			x--;
			y--;
		    }

		    /* Paste buffer in */
		    y = MAX(prompt->buf_pos, 0);
		    for(x = 0; x < tmp_buf_len; x++, y++)
		    {
			if(y >= (prompt->buf_len - 1))
			{
			    prompt->buf[y] = '\0';
			    break;
			}

			prompt->buf[y] = tmp_buf[x];
		    }
		    /* Null terminate */
		    prompt->buf[prompt->buf_len] = '\0';

		    /* Deleted the received buffer */
		    free(tmp_buf);
		    tmp_buf = NULL;

		    /* Must move cursor ahead */
		    prompt->buf_pos = MIN(
			prompt->buf_pos + tmp_buf_len,
			prompt->buf_len - 1
		    );

		    /* Scroll as needed */
		    OSWGetWindowGeometry(
			prompt->text_area,
			NULL, NULL, &width, &height, NULL
		    );
		    chars_visible = MAX(
			((int)width - (PROMPT_MARGIN * 2)) /
			PROMPT_CHAR_WIDTH,
			0
		    );
		    if(prompt->buf_pos >
			(prompt->buf_vis_pos + chars_visible)
		    )
		    {
			prompt->buf_vis_pos = MAX(
			    (int)prompt->buf_pos - chars_visible,
			    0
			);
		    }
		}
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }

	    /* Add a character */
	    else if(!prompt->disabled)
	    {
		int len = MIN(
		    STRLEN(prompt->buf),
		    prompt->buf_len
		);

		/* Check if there are any selected text */
		if((prompt->buf_sel_start >= 0) &&
		   (prompt->buf_sel_end >= 0)
		)
		{
		    /* Delete the selected text */
		    int	i = MIN(prompt->buf_sel_start, len),
			j = MIN(prompt->buf_sel_end + 1, len);
		    while(j < len)
		    {      
			if(prompt->buf[j] == '\0')
			    break;

			prompt->buf[i] = prompt->buf[j];

			i++;
			j++;
		    }
		    /* Null terminate */
		    prompt->buf[i] = '\0';

		    /* Get the new content length */
		    len = MIN(
			STRLEN(prompt->buf),
			prompt->buf_len
		    );

		    /* Move cursor to position where selection started */
		    prompt->buf_pos = CLIP(
			prompt->buf_sel_start,
			0, len
		    );

		    /* Unmark selection */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;
		}   

		/* Add character */
		if((prompt->buf_pos >= 0) &&
		   (prompt->buf_pos < (prompt->buf_len - 1))
		)
		{
		    char c;
		    int i;
		    char *buf = prompt->buf;

		    /* Shift all the characters from buf_pos */
		    for(i = prompt->buf_len; i > prompt->buf_pos; i--)
			buf[i] = buf[i - 1];
		    buf[prompt->buf_len] = '\0';

		    /* Add a character */
		    c = OSWGetASCIIFromKeyCode(keycode, key_state);
		    if(c != '\0')
		    {
			const int chars_visible = PROMPT_CHARS_VIS(prompt);

		        prompt->buf[prompt->buf_pos] = c;

			i = prompt->buf_len;
			prompt->buf[i] = '\0';

		        /* Move cursor ahead one character */
		        prompt->buf_pos += 1;

			/* Scroll as needed */
			if(prompt->buf_pos >
			   (prompt->buf_vis_pos + chars_visible)
			)
			    prompt->buf_vis_pos = MAX(
				(prompt->buf_pos - chars_visible),
				0
			    );
			if(prompt->buf_pos < prompt->buf_vis_pos)
			    prompt->buf_vis_pos = prompt->buf_pos;
		    }
		}
		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!prompt->has_focus)
		break;

	    break;

	  case ButtonPress:
	    button = &event->xbutton;

	    /* Skip if there is no text buffer */
	    if((prompt->buf == NULL) || (prompt->buf_len <= 0))   
		break;

	    /* Focus as needed */
	    if((w == prompt->toplevel) ||
	       (w == prompt->text_area)
	    )
	    {
		PROMPT_UNFOCUS_REDRAW_ALL(prompt);
	        prompt->has_focus = True;
	    }

	    /* Text area */
	    if(w == prompt->text_area)
	    {
		const int	button_num = button->button,
				x = (int)button->x;

		prompt->button_pressed = button_num;

		TooltipsUnmap(NULL);

		/* Button1: Place cursor and start selection */
		if(button_num == Button1)
		{
		    const int len = STRLEN(prompt->buf);
		    const time_t	cur_ms = OSWTimeMS(),
					last_pressed_time = prompt->button_pressed_time,
					dt = cur_ms - last_pressed_time;

		    prompt->selection_in_progress = True;

		    /* Set new position of cursor */
		    prompt->buf_pos = CLIP(
			(((x - PROMPT_MARGIN) /
			    PROMPT_CHAR_WIDTH) + prompt->buf_vis_pos),
			0, len
		    );

		    /* Start selection */
		    prompt->buf_sel_origin = prompt->buf_pos;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;

		    /* Double click? */
		    if((last_pressed_time > 0l) && (dt >= 0l) &&
		       (dt <= wglobal->double_click_int) &&
		       (prompt->button_pressed_count == 1)
		    )
		    {
			/* Select a "word" */
			char c;
			int i, sel_len;

			/* Seek i to the start of the selection */
			for(i = MIN(prompt->buf_pos,
				    (prompt->buf_len - 1)
			    ); i > 0; i--
			)
			{
			    c = prompt->buf[i];

			    /* Check if char is not a letter, number, or '_' */
			    if(!((c >= 0x30) && (c <= 0x39)) &&
			       !((c >= 0x41) && (c <= 0x5a)) &&
			       !((c >= 0x61) && (c <= 0x7a)) &&
			       (c != '_')
			    )
			    {
				i++;
				break;
			    }
			}
			prompt->buf_sel_start = CLIP(i, 0, len);

			/* Seek i to the end of the selection */
			for(i = MAX(prompt->buf_pos, 0);
			    i < prompt->buf_len; i++
			)
			{
			    c = prompt->buf[i];

			    if(c == '\0')
			    {
				i--;
				break;
			    }

			    /* Check if char is not a letter, number, or '_' */
			    if(!((c >= 0x30) && (c <= 0x39)) &&
			       !((c >= 0x41) && (c <= 0x5a)) &&
			       !((c >= 0x61) && (c <= 0x7a)) &&
			       (c != '_')
			    )
			    {
			        i--;
				break;
			    }
			}
			prompt->buf_sel_end = CLIP(i, 0, len);

			/* Store the bytes of the selected buffer range */
			sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
			if(sel_len >= 0)
			    PromptDDEStoreASCIISeg(
				w,
				prompt->buf + prompt->buf_sel_start,
				sel_len + 1
			    );

			/* Increment button press count */
			prompt->button_pressed_count++;
		    }
		    /* Triple click? */
		    else if((last_pressed_time > 0l) && (dt >= 0l) &&
			    (dt <= wglobal->double_click_int) &&
			    (prompt->button_pressed_count == 2)
		    )
		    {
			/* Select the entire contents */
			int sel_len;

			prompt->buf_sel_start = 0;
			prompt->buf_sel_end = len;

			sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
			if(sel_len >= 0)
			    PromptDDEStoreASCIISeg(
				w,
				prompt->buf + prompt->buf_sel_start,
				sel_len + 1
			    );

			/* Increment button press count */
			prompt->button_pressed_count++;
 		    }
		    else
		    {
			/* Reset the last button press record */
			prompt->button_pressed_count = 1;
			prompt->button_pressed_time = cur_ms;
		    }

		    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		    events_handled++;
	        }
		/* Button2: Paste DDE buffer */
		else if((button->button == Button2) &&
			!prompt->disabled
		)
		{
		    char *tmp_buf = OSWDDEGetString(prompt->text_area);
		    const int tmp_buf_len = STRLEN(tmp_buf);
		    const int len = MIN(STRLEN(prompt->buf), prompt->buf_len); 

		    /* Unmark selection */
		    prompt->buf_sel_origin = -1;
		    prompt->buf_sel_start = -1;
		    prompt->buf_sel_end = -1;

		    /* Calculate and set new position of cursor */
		    prompt->buf_pos = ((x - PROMPT_MARGIN) /
			PROMPT_CHAR_WIDTH) + prompt->buf_vis_pos;

		    /* Clip position */
		    if(prompt->buf_pos > len)
			prompt->buf_pos = len;
		    if(prompt->buf_pos < 0)
			prompt->buf_pos = 0;

		    /* Check if there is text selection marked */
		    if((tmp_buf != NULL) && (tmp_buf_len > 0))
		    {
			/* Shift tailing segment of prompt buffer */
			int	x = len,
				y = len + tmp_buf_len,
				z, chars_visible;
			unsigned int width, height;

			if(y > prompt->buf_len)
			{
			    z = y - prompt->buf_len;
			    y = MAX(y - z, 0);
			    x = MAX(x - z, 0);
		        }

		        while(x >= prompt->buf_pos)
		        {
			    prompt->buf[y] = prompt->buf[x];
			    x--;
			    y--;
		        }

			/* Paste buffer in */
			y = MAX(prompt->buf_pos, 0);
			for(x = 0; x < tmp_buf_len; x++, y++)
			{
			    if(y >= (prompt->buf_len - 1))
			    {
				prompt->buf[y] = '\0';
				break;
			    }

			    prompt->buf[y] = tmp_buf[x];
		        }
		        /* Null terminate */
		        prompt->buf[prompt->buf_len] = '\0';

		        /* Delete the received buffer */
		        free(tmp_buf);
		        tmp_buf = NULL;

		        /* Move the cursor ahead */
		        prompt->buf_pos = MIN(
			    (prompt->buf_pos + tmp_buf_len),
			    (prompt->buf_len - 1)
		        );

		        /* Scroll as needed */
		        OSWGetWindowGeometry(
			    w,
			    NULL, NULL, &width, &height, NULL
		        );
		        chars_visible = MAX(
			    (((int)width - (PROMPT_MARGIN * 2)) /
				PROMPT_CHAR_WIDTH),
			    0
		        );
		        if(prompt->buf_pos >
			    (prompt->buf_vis_pos + chars_visible)
			)
			    prompt->buf_vis_pos = MAX(
			        (prompt->buf_pos - chars_visible),
			        0
			    );
		    }
		    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		    events_handled++;
		}
		/* Button3: Set selection end mark (if there was a start) */
		else if(button->button == Button3)
	        {
		    const int	prev_buf_pos = CLIP(
			prompt->buf_pos, 0, prompt->buf_len
		    ),
				m = STRLEN(prompt->buf);
		    int sel_len;

		    prompt->selection_in_progress = True;

		    /* Calculate and set new position of cursor */
		    prompt->buf_pos = CLIP(
			(((x - PROMPT_MARGIN) /
			    PROMPT_CHAR_WIDTH) + prompt->buf_vis_pos),
			0, m
		    );

		    /* Is there an existing selection? */
		    if(prompt->buf_sel_end > -1)
		    {
			/* Extend the existing selection */
			if(prompt->buf_pos > MIN(prompt->buf_sel_start,
						 prompt->buf_sel_end)
			)
			{
			    prompt->buf_sel_start = MIN(
				prompt->buf_sel_start,
				prompt->buf_sel_end
			    );
			    prompt->buf_sel_end = prompt->buf_pos;
			}
			else if(prompt->buf_pos < MAX(prompt->buf_sel_start,
						      prompt->buf_sel_end)
			)
			{
			    prompt->buf_sel_end = MAX(
				prompt->buf_sel_start,
				prompt->buf_sel_end
			    );
			    prompt->buf_sel_start = prompt->buf_pos;
			}
		    }
		    else
		    {
			/* Set the selection from the previous cursor
			 * position to the new cursor position
			 */
			if(prompt->buf_pos > prev_buf_pos)
			{
			    prompt->buf_sel_start = prev_buf_pos;
			    prompt->buf_sel_end = prompt->buf_pos;
		        }
		        else if(prompt->buf_pos < prev_buf_pos)
		        {
			    prompt->buf_sel_start = prompt->buf_pos;
			    prompt->buf_sel_end = prev_buf_pos;
			}
		    }
		    /* Clip selection */
		    if(prompt->buf_sel_start > -1)
		    {
			prompt->buf_sel_start = CLIP(
			    prompt->buf_sel_start,
			    0, m
			);
			prompt->buf_sel_end = CLIP(
			    prompt->buf_sel_end,
			    0, m           
			);
		    }

		    /* No need to unmark other prompts, they already were
		     * on the start of marking
		     */

		    /* Store the bytes of the selected buffer range */
		    sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
		    if(sel_len >= 0)
			PromptDDEStoreASCIISeg(
			    w,
			    prompt->buf + prompt->buf_sel_start,
			    sel_len + 1
			);

		    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		    events_handled++;
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;

	    /* Skip if there is no text buffer */
	    if((prompt->buf == NULL) || (prompt->buf_len == 0))
		break;

	    /* Toplevel or text area */
	    if((w == prompt->toplevel) ||
	       (w == prompt->text_area)
	    )
	    {
		prompt->button_pressed = 0;

		/* Was there a selection in progress? */
		if(prompt->selection_in_progress)
		{
		    prompt->selection_in_progress = False;

		    /* Store the bytes of the selected buffer range */
		    if((prompt->buf_sel_end > -1) &&
		       (prompt->buf_sel_start > -1)
		    )
		    {
			const int sel_len = prompt->buf_sel_end - prompt->buf_sel_start;
			if(sel_len >= 0)
			    PromptDDEStoreASCIISeg(
				prompt->text_area,
				prompt->buf + prompt->buf_sel_start,
				sel_len + 1
			    );

			PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		    }

		    /* Remove any scroll timeouts for this prompt */
		    PromptRemoveScrollTimeout(prompt);
		}

		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;

	    /* Skip if there is no text buffer */
	    if((prompt->buf == NULL) || (prompt->buf_len == 0))
		break;

	    /* Skip if there is no selection in progress */
	    if(!prompt->selection_in_progress)
		break;

	    /* Handle mark drag */
	    if(w == prompt->text_area)
	    {
		int	x = (int)motion->x,
			len = strlen(prompt->buf),
			new_pos = MAX(
			    (((x - PROMPT_MARGIN) / PROMPT_CHAR_WIDTH) +
			    prompt->buf_vis_pos),
			prompt->buf_vis_pos
		);

		if(new_pos < prompt->buf_sel_origin)
		{
		    /* Marking left */
		    prompt->buf_sel_start = new_pos;
		    prompt->buf_sel_end = prompt->buf_sel_origin;
		}
		else
		{
		    /* Marking right */
		    prompt->buf_sel_end = new_pos;
		    prompt->buf_sel_start = prompt->buf_sel_origin;
		}
		prompt->buf_pos = new_pos;

		/* Clip positions */
		if(prompt->buf_pos > len)
		    prompt->buf_pos = len;
		if(prompt->buf_pos < 0)
		    prompt->buf_pos = 0;

		if(prompt->buf_sel_end >= len)
		    prompt->buf_sel_end = len - 1;
		if(prompt->buf_sel_end < 0)
		    prompt->buf_sel_end = 0;
	  
		if(prompt->buf_sel_start >= len)
		    prompt->buf_sel_start = len - 1;
		if(prompt->buf_sel_start < 0)
		    prompt->buf_sel_start = 0;

		PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_TEXT);
		events_handled++;
	    }
	    break;

	  case EnterNotify:
	    if(w == prompt->text_area)
	    {
		TooltipsSchedualMap(
		    NULL, (time_t)-1, prompt->text_area
		);
		events_handled++;
	    }
	    break;

	  case LeaveNotify:
	    if(w == prompt->text_area)
	    {
		TooltipsUnmap(NULL);
		if(prompt->selection_in_progress)
		{
		    /* Schedual scroll repeat */
		    PromptRemoveScrollTimeout(prompt);
		    PromptAddScrollTimeout(prompt, 0l);
		}
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    if(w == prompt->toplevel)
	    {
		prompt->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;

	  case SelectionClear:
	    w = event->xselectionclear.window;
	    if((w == prompt->toplevel) ||
	       (w == prompt->text_area)
	    )
	    {
		/* Unmark selection */
		prompt->buf_sel_origin = -1;
		prompt->buf_sel_start = -1;
		prompt->buf_sel_end = -1;

		if(prompt->map_state)
		    PromptQueueDraw(prompt, PROMPT_DRAW_AMOUNT_COMPLETE);

		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the prompt.
 */
void PromptMap(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return;

	if(!prompt->map_state)
	{
	    int	len = STRLEN(prompt->buf),
		chars_visible = PROMPT_CHARS_VIS(prompt);

	    /* Scroll to end */
	    prompt->buf_vis_pos = MAX((len - chars_visible), 0);
	    if(prompt->buf_pos < prompt->buf_vis_pos)
		prompt->buf_pos = prompt->buf_vis_pos;

	    /* Map the prompt */
	    OSWMapRaised(prompt->toplevel);
	    prompt->map_state = True;
	    OSWMapWindow(prompt->text_area);
	}
}

/*
 *	Unmaps the prompt.
 */
void PromptUnmap(prompt_struct *prompt)
{
	PromptClose(prompt);
}


/*
 *	Destroy prompt.
 */
void PromptDestroy(prompt_struct *prompt)
{
	int i;

	if(prompt == NULL)
	    return;

	/* Remove any scroll timeouts for this prompt */
	PromptRemoveScrollTimeout(prompt);

	/* Delete the associated hint message */
	TooltipsDeleteMessage(NULL, prompt->text_area);

	OSWDestroyPixmap(&prompt->text_area_buf);
	OSWDestroyWindow(&prompt->text_area);

	OSWDestroyPixmap(&prompt->toplevel_buf);
	OSWDestroyWindow(&prompt->toplevel);

	prompt->type = WIDGET_TYPE_UNKNOWN;
	prompt->map_state = False;
	prompt->has_focus = False;
	prompt->disabled = False;
	prompt->visibility_state = VisibilityFullyObscured;
	prompt->x = 0;
	prompt->y = 0;
	prompt->width = 0;
	prompt->height = 0;
	prompt->font = NULL;
	prompt->next = NULL;
	prompt->prev = NULL;

	prompt->options = 0;
	prompt->style = PROMPT_STYLE_FLUSHED;

	prompt->text_font = NULL;

	prompt->activate_cb = NULL;
	prompt->activate_data = NULL;
	prompt->cancel_cb = NULL;
	prompt->cancel_data = NULL;

	/* Delete the label */
	free(prompt->label);
	prompt->label = NULL;

	/* Delete the buffer */
	free(prompt->buf);
	prompt->buf = NULL;
	prompt->buf_len = 0;

	prompt->buf_pos = 0;
	prompt->buf_vis_pos = 0;

	prompt->selection_in_progress = False;
	prompt->buf_sel_origin = -1;
	prompt->buf_sel_start = -1;
	prompt->buf_sel_end = -1;

	/* Delete the history buffers */
	for(i = 0; i < prompt->total_hist_bufs; i++)
	{
	    if(prompt->hist_buf[i] == NULL)
		continue;

	    free(prompt->hist_buf[i]);
	}
	free(prompt->hist_buf);
	prompt->hist_buf = NULL;
	prompt->total_hist_bufs = 0;
	prompt->hist_buf_pos = 0;

	prompt->button_pressed = 0;
	prompt->button_pressed_count = 0;
	prompt->button_pressed_time = 0l;
}

/*
 *	Closes the prompt.
 */
void PromptClose(prompt_struct *prompt)
{
	if(prompt == NULL)
	    return;

	/* Remove any scroll timeouts for this prompt */
	PromptRemoveScrollTimeout(prompt);

	OSWUnmapWindow(prompt->toplevel);
	prompt->map_state = False;

	OSWDestroyPixmap(&prompt->text_area_buf);
	OSWDestroyPixmap(&prompt->toplevel_buf);
}
