#include <stdlib.h>
#include <string.h>
#include "../include/widget.h"


static void CBtnResizeCB(void *widget, void *data);
static void CBtnDrawCB(void *widget, void *data);

bool_t CBtnGetActive(check_button_struct *cb);
void CBtnSetActive(check_button_struct *cb, bool_t state);

int CBtnInit(
	check_button_struct *cb, win_t parent,
	int x, int y,
	bool_t state,
	const char *label
);
void CBtnSetTooltip(
	check_button_struct *cb, const char *msg
);
void CBtnResize(check_button_struct *cb);
void CBtnQueueResize(check_button_struct *cb);
void CBtnDraw(check_button_struct *cb, int amount);
void CBtnQueueDraw(check_button_struct *cb, int amount);
int CBtnManage(check_button_struct *cb, event_t *event);
void CBtnMap(check_button_struct *cb);
void CBtnUnmap(check_button_struct *cb);
void CBtnDestroy(check_button_struct *cb);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define TB_MARGIN		2


/*
 *	Resize callback.
 */
static void CBtnResizeCB(void *widget, void *data)
{
	check_button_struct *cb = (check_button_struct *)widget;
	if(cb == NULL)
	    return;

	CBtnResize(cb);
}

/*
 *	Draw callback.
 */
static void CBtnDrawCB(void *widget, void *data)
{
	int amount = (int)data;
	check_button_struct *cb = (check_button_struct *)widget;
	if(cb == NULL)
	    return;

	if(cb->map_state)
	    CBtnDraw(cb, amount);
}


/*
 *	Gets the check_button's toggled state.
 */
bool_t CBtnGetActive(check_button_struct *cb)
{
	return((cb != NULL) ? cb->state : False);
}

/*
 *	Sets the check_button's toggled state.
 */
void CBtnSetActive(check_button_struct *cb, bool_t state)
{
	if(cb == NULL)
	    return;

	if(cb->state != state)
	{
	    cb->state = state;
	    if(cb->map_state)
		CBtnQueueDraw(cb, CBTN_DRAW_AMOUNT_BUTTON);
	}
}


/*
 *	Initializes the check_button.
 */
int CBtnInit(
	check_button_struct *cb, win_t parent,
	int x, int y,
	bool_t state,
	const char *label
)
{
	font_t *font;
	image_t *img;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if((cb == NULL) || (parent == None))
	    return(-2);

	cb->type = WIDGET_TYPE_CHECK_BUTTON;
	cb->map_state = False;
	cb->has_focus = False;
	cb->disabled = False;
	cb->visibility_state = VisibilityFullyObscured;
	cb->x = x;
	cb->y = y;
	cb->width = 0;
	cb->height = 0;
	cb->font = font = wglobal->std_font;
	cb->next = NULL;
	cb->prev = NULL;
        cb->data = NULL;

	cb->state = state;
	cb->highlighted = False;
	cb->label = STRDUP(label);

	cb->checked_cb = NULL;
	cb->checked_data = NULL;

	/* Calculate width and height */
	img = wglobal->check_button_unarmed_img;
	if(img != NULL)
	{
	    if(label != NULL)
	    {
		int swidth, ascent, descent;
		OSWGetTextBounds(
		    font, label, -1,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		cb->width = img->width + (unsigned int)swidth +
		    (3 * TB_MARGIN);
		cb->height = MAX(
		    (int)img->height,
		    (ascent + descent)
		) + (2 * TB_MARGIN);
	    }
	    else
	    {
		cb->width = img->width + (2 * TB_MARGIN);
		cb->height = img->height + (2 * TB_MARGIN);
	    }
	}
	else
	{
	    if(label != NULL)
	    {
		int swidth, ascent, descent;
		OSWGetTextBounds(
		    font, label, -1,
		    NULL, NULL, &swidth, &ascent, &descent  
		);
		cb->width = (unsigned int)swidth + (2 * TB_MARGIN);
		cb->height = (unsigned int)(ascent + descent) +
		    (2 * TB_MARGIN);
	    }
	    else
	    {
		cb->width = (2 * TB_MARGIN);
		cb->height = (2 * TB_MARGIN);
	    }
	}

	/* Create toplevel */
	OSWCreateWindow(
	    &w, parent,
	    cb->x, cb->y,
	    cb->width, cb->height
	);
	cb->toplevel = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    ButtonPressMask | ButtonReleaseMask |
	    EnterWindowMask | LeaveWindowMask |
	    VisibilityChangeMask
	);

	/* Add the widget to the regeristry */
	WidgetRegister(cb, cb->type);

	return(0);
}

/*
 *	Sets the check_button's hint message.
 */
void CBtnSetTooltip(
	check_button_struct *cb, const char *msg
)
{
	if(cb == NULL)
	    return;

	TooltipsSetMessage(NULL, cb->toplevel, msg);
}

/*
 *	Resizes the check_button.
 */
void CBtnResize(check_button_struct *cb)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(cb == NULL)
	    return;

	w = cb->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	cb->x = x;
	cb->y = y;
	cb->width = width;
	cb->height = height;
}

/*
 *	Queues the check_button to resize.
 */
void CBtnQueueResize(check_button_struct *cb)
{
	if(cb == NULL)
	    return;

	WidgetQueueResizeAction(CBtnResizeCB, cb, NULL);
}

/*
 *	Draws the check_button.
 */
void CBtnDraw(check_button_struct *cb, int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(cb == NULL)
	    return;

	if(True)
	{
	    int x, y;
	    unsigned int width, height;
	    win_t w = cb->toplevel;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Draw the background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
	        OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    /* Draw label */
	    if(cb->label != NULL)
	    {
		int	x, y,
			lbearing, rbearing, ascent, descent,
			font_height;
		const char *label = cb->label;
		font_t *font = cb->font;
		image_t *img = wglobal->check_button_unarmed_img;

		OSWSetFont(font);
		OSWGetTextBounds(
		    font, label, -1,
		    &lbearing, &rbearing, NULL,
		    &ascent, &descent
		);
		font_height = ascent + descent;
		if(img != NULL)
		{
		    x = TB_MARGIN + (int)img->width + TB_MARGIN;
		    y = ((int)height - font_height) / 2;
		}
		else
		{
		    x = TB_MARGIN;
		    y = ((int)height - font_height) / 2;
		}
		if(wglobal->force_mono)
		    OSWSetForeground(gui->white_pix);
		else if(cb->disabled)
		    OSWSetForeground(wglobal->disabled_text_pix);
		else
		    OSWSetForeground(wglobal->normal_text_pix);
		OSWDrawString(
		    w,
		    x - lbearing, y + ascent,
		    label
		);
	    }

	    /* Draw check_button */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->white_pix);
		OSWDrawSolidRectangle(
		    w,
		    TB_MARGIN, TB_MARGIN,
		    16, 16
		);
		OSWSetForeground(gui->black_pix);
		OSWDrawRectangle(
		    w,
		    TB_MARGIN + 1, TB_MARGIN + 1,
		    13, 13
		);
		if(cb->state)
		{
		    OSWDrawLine(
			w,
			TB_MARGIN + 1, TB_MARGIN + 1,
			TB_MARGIN + 14, TB_MARGIN + 14
		    );
		    OSWDrawLine(
			w,
			TB_MARGIN + 1, TB_MARGIN + 14,
			TB_MARGIN + 14, TB_MARGIN + 1
		    );
		}
	    }
	    else
	    {
		x = TB_MARGIN;
		y = TB_MARGIN;
		WidgetPutImageNormal(
		    w,
		    cb->state ?
			wglobal->check_button_armed_img :
			(cb->highlighted ?
			    wglobal->check_button_highlighted_img :
			    wglobal->check_button_unarmed_img),
		    x, y,
		    True
		);
	    }
	}
}

/*
 *	Queues the check_button to draw.
 */
void CBtnQueueDraw(check_button_struct *cb, int amount)
{
	if(cb == NULL)
	    return;

	WidgetQueueDrawAction(CBtnDrawCB, cb, (void *)amount);
}

/*
 *	Manages the check_button event.
 */
int CBtnManage(check_button_struct *cb, event_t *event)
{
	int events_handled = 0;
	win_t w;

	if((cb == NULL) || (event == NULL))
	    return(events_handled);

	if(!cb->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == cb->toplevel)
	    {
		if(cb->map_state)
		    CBtnQueueDraw(cb, CBTN_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == cb->toplevel)
	    {
		CBtnQueueDraw(cb, CBTN_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case ButtonPress:
	    if(w == cb->toplevel)
	    {
		if(!cb->disabled)
		{
		    const bool_t prev_state = cb->state;

		    cb->has_focus = True;

		    TooltipsUnmap(NULL);

		    if(cb->state)
			cb->state = False;
		    else
			cb->state = True;

		    /* Call toggled callback */
		    if((prev_state != cb->state) &&
		       (cb->checked_cb != NULL)
		    )
		        cb->checked_cb(
			    cb,
			    cb->state,
			    cb->checked_data
		        );

		    if(!WidgetIsRegistered(cb))
		    {
		        events_handled++;
		        return(events_handled);
		    }

		    CBtnQueueDraw(cb, CBTN_DRAW_AMOUNT_BUTTON);
		}

		events_handled++;
	    }
	    else
	    {
		if(cb->has_focus)
		{
		    cb->has_focus = False;
		    CBtnQueueDraw(cb, CBTN_DRAW_AMOUNT_BUTTON);
		}
	    }
	    break;

	  case ButtonRelease:
	    if(w == cb->toplevel)
	    {
		events_handled++;
	    }
	    break;

	  case EnterNotify:
	    if(w == cb->toplevel)
	    {
		if(!cb->disabled)
		{
		    TooltipsSchedualMap(
			NULL, (time_t)-1, cb->toplevel
		    );
		    cb->highlighted = True;
		    CBtnQueueDraw(cb, CBTN_DRAW_AMOUNT_COMPLETE);
		}
		events_handled++;
	    }
	    break;

	  case LeaveNotify:
	    if(w == cb->toplevel)
	    {
		if(!cb->disabled)
		{
		    TooltipsUnmap(NULL);
		    cb->highlighted = False;
		    CBtnQueueDraw(cb, CBTN_DRAW_AMOUNT_COMPLETE);
		}
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    if(w == cb->toplevel)
	    {
		cb->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the check_button.
 */
void CBtnMap(check_button_struct *cb)
{
	if(cb == NULL)
	    return;

	if(!cb->map_state)
	{
	    OSWMapRaised(cb->toplevel);
	    cb->map_state = True;
	}
}

/*
 *	Unmaps the check_button.
 */
void CBtnUnmap(check_button_struct *cb)
{
	if(cb == NULL)
	    return;

	OSWUnmapWindow(cb->toplevel);
	cb->map_state = False;
}

/*
 *	Destroys the check_button.
 */
void CBtnDestroy(check_button_struct *cb)
{
	if(cb == NULL)
	    return;

	/* Delete the associated hint message */
	TooltipsDeleteMessage(NULL, cb->toplevel);

	OSWDestroyWindow(&cb->toplevel);

	cb->type = WIDGET_TYPE_UNKNOWN;
	cb->map_state = False;
	cb->has_focus = False;
	cb->disabled = False;
	cb->visibility_state = VisibilityFullyObscured;
	cb->x = 0;
	cb->y = 0;
	cb->width = 0;
	cb->height = 0;
	cb->disabled = False;
	cb->next = NULL;
	cb->prev = NULL;

	free(cb->label);
	cb->label = NULL;
	cb->state = False;
	cb->highlighted = False;

	cb->checked_cb = NULL;
	cb->checked_data = NULL;
}
